Package: nhlscraper
Title: Scraper for National Hockey League Data
Version: 0.6.1
Authors@R: c(person('Rento', 'Saijo', email = 'rentosaijo0527@gmail.com', role = c('aut', 'cre', 'cph'), comment = c(ORCID = '0009-0008-4919-7349')), person('Lars', 'Skytte', email = 'lars.sunesen.skytte@gmail.com', role = 'ctb'), person('Jack', 'Pallotta', email = 'jackjpallotta@gmail.com', role = 'ctb', comment = c(ORCID = '0009-0004-0744-6977')))
Description: Scrapes and cleans data from the 'NHL' and 'ESPN' APIs into data.frames and lists. Wraps 125+ endpoints documented in <https://github.com/RentoSaijo/nhlscraper/wiki> from high-level multi-season summaries and award winners to low-level decisecond replays and bookmakers' odds, making them more accessible. Features cleaning and visualization tools, primarily for play-by-plays.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: httr2 (>= 1.2.0), jsonlite (>= 2.0.0), xml2 (>= 1.5.0), arrow
        (>= 23.0.0)
Suggests: testthat (>= 3.0.0), knitr (>= 1.50.0), rmarkdown (>= 2.29.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://rentosaijo.github.io/nhlscraper/,
        https://github.com/RentoSaijo/nhlscraper
BugReports: https://github.com/RentoSaijo/nhlscraper/issues
Copyright: Copyright: NHL and the NHL Shield are registered trademarks
        of the National Hockey League. NHL and NHL team marks are the
        property of the NHL and its teams.
NeedsCompilation: yes
Packaged: 2026-05-24 13:44:48 UTC; rsai_91
Author: Rento Saijo [aut, cre, cph] (ORCID:
    <https://orcid.org/0009-0008-4919-7349>),
  Lars Skytte [ctb],
  Jack Pallotta [ctb] (ORCID: <https://orcid.org/0009-0004-0744-6977>)
Maintainer: Rento Saijo <rentosaijo0527@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-24 14:10:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-24 18:28:36 UTC; unix
Archs: nhlscraper.so.dSYM

Package: nhlscraper
Title: Scraper for National Hockey League Data
Version: 0.6.1
Authors@R: c(person('Rento', 'Saijo', email = 'rentosaijo0527@gmail.com', role = c('aut', 'cre', 'cph'), comment = c(ORCID = '0009-0008-4919-7349')), person('Lars', 'Skytte', email = 'lars.sunesen.skytte@gmail.com', role = 'ctb'), person('Jack', 'Pallotta', email = 'jackjpallotta@gmail.com', role = 'ctb', comment = c(ORCID = '0009-0004-0744-6977')))
Description: Scrapes and cleans data from the 'NHL' and 'ESPN' APIs into data.frames and lists. Wraps 125+ endpoints documented in <https://github.com/RentoSaijo/nhlscraper/wiki> from high-level multi-season summaries and award winners to low-level decisecond replays and bookmakers' odds, making them more accessible. Features cleaning and visualization tools, primarily for play-by-plays.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: httr2 (>= 1.2.0), jsonlite (>= 2.0.0), xml2 (>= 1.5.0), arrow
        (>= 23.0.0)
Suggests: testthat (>= 3.0.0), knitr (>= 1.50.0), rmarkdown (>= 2.29.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://rentosaijo.github.io/nhlscraper/,
        https://github.com/RentoSaijo/nhlscraper
BugReports: https://github.com/RentoSaijo/nhlscraper/issues
Copyright: Copyright: NHL and the NHL Shield are registered trademarks
        of the National Hockey League. NHL and NHL team marks are the
        property of the NHL and its teams.
NeedsCompilation: yes
Packaged: 2026-05-24 13:44:48 UTC; rsai_91
Author: Rento Saijo [aut, cre, cph] (ORCID:
    <https://orcid.org/0009-0008-4919-7349>),
  Lars Skytte [ctb],
  Jack Pallotta [ctb] (ORCID: <https://orcid.org/0009-0004-0744-6977>)
Maintainer: Rento Saijo <rentosaijo0527@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-24 14:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-24 18:43:28 UTC; unix
Archs: nhlscraper.so.dSYM

# Generated by roxygen2: do not edit by hand

export(add_deltas)
export(add_goalie_biometrics)
export(add_shift_times)
export(add_shooter_biometrics)
export(attendance)
export(award_winners)
export(awards)
export(boxscore)
export(bracket)
export(calculate_expected_goals)
export(calculate_xG)
export(coach_career_statistics)
export(coach_career_stats)
export(coach_franchise_statistics)
export(coach_franchise_stats)
export(coaches)
export(combine_reports)
export(contracts)
export(countries)
export(draft_picks)
export(draft_prospects)
export(draft_rankings)
export(draft_tracker)
export(drafts)
export(draw_NHL_rink)
export(espn_futures)
export(espn_game_odds)
export(espn_game_summary)
export(espn_games)
export(espn_injuries)
export(espn_pbp)
export(espn_play_by_play)
export(espn_player_summary)
export(espn_players)
export(espn_team_summary)
export(espn_teams)
export(espn_transactions)
export(expansion_draft_picks)
export(expansion_drafts)
export(franchise_playoff_situational_results)
export(franchise_season_statistics)
export(franchise_season_stats)
export(franchise_statistics)
export(franchise_stats)
export(franchise_team_statistics)
export(franchise_team_stats)
export(franchise_versus_franchise)
export(franchise_vs_franchise)
export(franchises)
export(game_odds)
export(game_rosters)
export(game_type_now)
export(games)
export(gc_pbp)
export(gc_pbp_raw)
export(gc_pbps)
export(gc_pbps_raw)
export(gc_play_by_play)
export(gc_play_by_play_raw)
export(gc_play_by_plays)
export(gc_play_by_plays_raw)
export(gc_summary)
export(general_managers)
export(get_attendance)
export(get_award_winners)
export(get_awards)
export(get_bracket)
export(get_configuration)
export(get_countries)
export(get_draft_picks)
export(get_draft_rankings)
export(get_draft_tracker)
export(get_drafts)
export(get_espn_athlete)
export(get_espn_athletes)
export(get_espn_coach)
export(get_espn_coach_career)
export(get_espn_coaches)
export(get_espn_event)
export(get_espn_event_odds)
export(get_espn_event_officials)
export(get_espn_event_play_by_play)
export(get_espn_event_stars)
export(get_espn_events)
export(get_espn_futures)
export(get_espn_injuries)
export(get_espn_team)
export(get_espn_teams)
export(get_espn_transactions)
export(get_franchise_season_by_season)
export(get_franchise_team_totals)
export(get_franchise_vs_franchise)
export(get_franchises)
export(get_game_boxscore)
export(get_game_landing)
export(get_game_story)
export(get_games)
export(get_gc_play_by_play)
export(get_glossary)
export(get_goalie_leaders)
export(get_goalie_milestones)
export(get_goalie_statistics)
export(get_goalies)
export(get_officials)
export(get_partner_odds)
export(get_player_game_log)
export(get_player_landing)
export(get_players)
export(get_schedule)
export(get_scoreboards)
export(get_scores)
export(get_season_now)
export(get_seasons)
export(get_series)
export(get_series_schedule)
export(get_shift_charts)
export(get_skater_leaders)
export(get_skater_milestones)
export(get_skater_statistics)
export(get_skaters)
export(get_spotlight_players)
export(get_standings)
export(get_standings_information)
export(get_streams)
export(get_team_prospects)
export(get_team_roster)
export(get_team_roster_statistics)
export(get_team_schedule)
export(get_team_scoreboard)
export(get_team_seasons)
export(get_team_statistics)
export(get_teams)
export(get_tv_schedule)
export(get_venues)
export(get_wsc_play_by_play)
export(glossary)
export(gms)
export(goalie_edge_5_vs_5)
export(goalie_edge_five_versus_five)
export(goalie_edge_leaders)
export(goalie_edge_save_percentage)
export(goalie_edge_seasons)
export(goalie_edge_shot_location)
export(goalie_edge_summary)
export(goalie_game_report)
export(goalie_game_scoring)
export(goalie_game_statistics)
export(goalie_game_stats)
export(goalie_leaders)
export(goalie_milestones)
export(goalie_regular_statistics)
export(goalie_regular_stats)
export(goalie_report_configs)
export(goalie_report_configurations)
export(goalie_scoring)
export(goalie_season_report)
export(goalie_season_statistics)
export(goalie_season_stats)
export(goalie_series_statistics)
export(goalie_series_stats)
export(goalie_statistics)
export(goalie_stats)
export(ig_game_cum_xG)
export(ig_game_cumulative_expected_goals)
export(ig_game_shot_locations)
export(ig_game_shot_locs)
export(location)
export(lottery_odds)
export(officials)
export(penalty_shots)
export(ping)
export(player_game_log)
export(player_seasons)
export(player_summary)
export(players)
export(playoff_season_statistics)
export(playoff_season_stats)
export(ps)
export(pss)
export(replay)
export(replays)
export(roster)
export(roster_statistics)
export(roster_stats)
export(schedule)
export(scores)
export(season_now)
export(seasons)
export(series)
export(series_schedule)
export(shift_chart)
export(shift_chart_summaries)
export(shift_chart_summary)
export(shift_charts)
export(shifts)
export(skater_edge_leaders)
export(skater_edge_seasons)
export(skater_edge_shot_location)
export(skater_edge_shot_speed)
export(skater_edge_skating_distance)
export(skater_edge_skating_speed)
export(skater_edge_summary)
export(skater_edge_zone_time)
export(skater_game_report)
export(skater_leaders)
export(skater_milestones)
export(skater_playoff_statistics)
export(skater_playoff_stats)
export(skater_regular_statistics)
export(skater_regular_stats)
export(skater_report_configs)
export(skater_report_configurations)
export(skater_season_report)
export(skater_season_statistics)
export(skater_season_stats)
export(skater_series_statistics)
export(skater_series_stats)
export(skater_statistics)
export(skater_stats)
export(spotlight_players)
export(standings)
export(standings_rules)
export(streams)
export(team_edge_leaders)
export(team_edge_seasons)
export(team_edge_shot_location)
export(team_edge_shot_speed)
export(team_edge_skating_distance)
export(team_edge_skating_speed)
export(team_edge_summary)
export(team_edge_zone_time)
export(team_game_report)
export(team_logos)
export(team_month_schedule)
export(team_prospects)
export(team_report_configs)
export(team_report_configurations)
export(team_season_report)
export(team_season_schedule)
export(team_season_statistics)
export(team_season_stats)
export(team_seasons)
export(team_week_schedule)
export(teams)
export(tv_schedule)
export(venues)
export(wsc_pbp)
export(wsc_pbp_raw)
export(wsc_pbps)
export(wsc_pbps_raw)
export(wsc_play_by_play)
export(wsc_play_by_play_raw)
export(wsc_play_by_plays)
export(wsc_play_by_plays_raw)
export(wsc_summary)
export(x_game_cum_xG)
export(x_game_cumulative_expected_goals)
export(x_game_shot_locations)
export(x_game_shot_locs)
useDynLib(nhlscraper, .registration = TRUE)

# nhlscraper <a href="https://rentosaijo.github.io/nhlscraper/"><img src="man/figures/logo.png" align="right" height="138" alt="nhlscraper website" /></a>
[![CRAN Status](https://www.r-pkg.org/badges/version/nhlscraper)](https://CRAN.R-project.org/package=nhlscraper)
[![Dev Version](https://img.shields.io/badge/dev%20ver-0.6.0.9000-red.svg)](https://github.com/RentoSaijo/nhlscraper)
![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/nhlscraper)

### Overview

nhlscraper is a minimum-dependency R package to scrape, clean, and visualize NHL data via the NHL and ESPN APIs. It primarily wraps 125+ [endpoints](https://github.com/RentoSaijo/nhlscraper/wiki) from high-level multi-season summaries and award winners to low-level decisecond replays and bookmakers' odds, making them significantly more accessible. It also features cleaning and visualization functions, primarily for play-by-plays, to help analyze the data.

### Prerequisite

- R/RStudio; you can check out my [tutorial](https://youtu.be/CUXzeuXJCPs?si=bkWFAm-hSXhUsxaN) if you are not familiar!

### Installation

Install the official version from [CRAN](https://cran.r-project.org) with:
```r
install.packages('nhlscraper')
```

Install the development version from [GitHub](https://github.com/) with:
```r
install.packages('devtools')
devtools::install_github('RentoSaijo/nhlscraper')
```

### Disclosure

Detailed documentation for each scraping function will be released gradually over time. Because each function can return a large amount of information, users are encouraged to explore the provided examples to discover what is available. Most, if not all, of the endpoints accessed by this package are unofficially documented (i.e., hidden), so it is important to use them responsibly and with respect for the NHL’s data servers. Endpoints serving historical or otherwise mostly static data should ideally be queried once and then stored locally (for example, in a MySQL database) for further analysis, rather than being called repeatedly; the [load functions](https://rentosaijo.github.io/nhlscraper/reference/index.html#load) attempt to mitigate this risk by providing pre-scraped data, so please make use of them. The exact rate limits for these APIs are not publicly known, so users are asked to avoid excessive or abusive querying to help ensure continued access for everyone.

### History

Prior to the NHL API rework in 2023, Drew Hynes documented a comprehensive list of known [endpoints](https://gitlab.com/dword4/nhlapi/), and several R packages, such as [nhlapi](https://github.com/jozefhajnala/nhlapi) and [hockeyR](https://github.com/danmorse314/hockeyR), were built to access them. However, after the NHL completely transformed its API structure, all of these packages became mostly defunct as their authors understandably chose not to continue maintaining them. The community gathered around work by Zachary Maludzinski to discover and share new [endpoints](https://github.com/Zmalski/NHL-API-Reference), but progress naturally slowed once most of the "main" endpoints were identified. Over the summer of 2025, I began reverse-engineering many of the remaining, undocumented endpoints while searching for additional data for future research, with a particular focus on NHL EDGE and Records data. After I shared those [findings](https://github.com/Zmalski/NHL-API-Reference/issues/69), the effort to expand and refine this map of the APIs accelerated, and with the addition of David Fleischer's [effort](https://github.com/dfleis/nhl-api-docs), we ultimately identified 400+ new endpoints. In parallel, I also discovered many new endpoints for the ESPN API, extending beyond what Joseph Wilson had already [compiled](https://github.com/pseudo-r/Public-ESPN-API).

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}

## ----data---------------------------------------------------------------------
# Pull draft picks and keep skaters with measured size.
draft_tbl <- nhlscraper::draft_picks()
draft_tbl <- draft_tbl[
  draft_tbl[['draftYear']] >= 1979 &
    draft_tbl[['positionCode']] != 'G' &
    !is.na(draft_tbl[['height']]) &
    !is.na(draft_tbl[['weight']]),
]

# Create era, round, and position buckets.
draft_tbl[['roundBucket']] <- ifelse(
  draft_tbl[['roundNumber']] == 1,
  'Round 1',
  'Rounds 2-7'
)
draft_tbl[['era']] <- cut(
  draft_tbl[['draftYear']],
  breaks = c(1978, 1989, 1999, 2009, 2019, Inf),
  labels = c(
    '1979-1989',
    '1990-1999',
    '2000-2009',
    '2010-2019',
    '2020-2025'
  )
)
draft_tbl[['positionBucket']] <- ifelse(
  draft_tbl[['positionCode']] == 'D',
  'Defense',
  'Forward'
)
draft_tbl[['tallSkater']] <- draft_tbl[['height']] >= 74
nrow(draft_tbl)

## ----era-table----------------------------------------------------------------
# Summarize size by era and round bucket.
era_summary <- aggregate(
  cbind(height, weight) ~ era + roundBucket,
  data = draft_tbl,
  FUN = mean
)
make_table(
  era_summary,
  caption = 'Average drafted skater size by era and draft bucket.'
)

## ----rolling-data-------------------------------------------------------------
# Compute annual mean height by round bucket.
round1_annual <- aggregate(
  height ~ draftYear,
  data = draft_tbl[draft_tbl[['roundBucket']] == 'Round 1', ],
  FUN = mean
)
later_annual <- aggregate(
  height ~ draftYear,
  data = draft_tbl[draft_tbl[['roundBucket']] == 'Rounds 2-7', ],
  FUN = mean
)

# Smooth annual means with five-draft rolling averages.
round1_annual[['rollHeight']] <- as.numeric(stats::filter(
  round1_annual[['height']],
  rep(1 / 5, 5),
  sides = 2
))
later_annual[['rollHeight']] <- as.numeric(stats::filter(
  later_annual[['height']],
  rep(1 / 5, 5),
  sides = 2
))

## ----rolling-plot, fig.cap = 'Five-draft rolling average height for first-round skaters and later-round skaters.'----
graphics::plot(
  round1_annual[['draftYear']],
  round1_annual[['rollHeight']],
  type = 'l',
  lwd = 2,
  col = '#0f4c5c',
  ylim = range(
    c(round1_annual[['rollHeight']], later_annual[['rollHeight']]),
    na.rm = TRUE
  ),
  xlab = 'Draft Year',
  ylab = 'Average Height (Inches)'
)
graphics::lines(
  later_annual[['draftYear']],
  later_annual[['rollHeight']],
  lwd = 2,
  col = '#e36414'
)
graphics::legend(
  'topright',
  legend = c('Round 1', 'Rounds 2-7'),
  col = c('#0f4c5c', '#e36414'),
  lwd = 2,
  bty = 'n'
)

## ----tall-share---------------------------------------------------------------
# Summarize share of taller skaters by era.
tall_share <- aggregate(
  tallSkater ~ era + roundBucket,
  data = draft_tbl,
  FUN = mean
)
tall_counts <- aggregate(
  height ~ era + roundBucket,
  data = draft_tbl,
  FUN = length
)
names(tall_counts)[names(tall_counts) == 'height'] <- 'n'
tall_share <- merge(tall_share, tall_counts, by = c('era', 'roundBucket'))

make_table(
  tall_share,
  caption = 'Share of drafted skaters measuring at least 6-foot-2.'
)

## ----tall-share-plot, fig.cap = 'Share of drafted skaters at least 6-foot-2 by era and round bucket.'----
# Plot tall-skater shares by era.
tall_matrix <- rbind(
  tall_share[['tallSkater']][tall_share[['roundBucket']] == 'Round 1'],
  tall_share[['tallSkater']][tall_share[['roundBucket']] == 'Rounds 2-7']
)
graphics::barplot(
  tall_matrix,
  beside = TRUE,
  col = c('#1b4332', '#95d5b2'),
  ylim = c(0, 0.7),
  names.arg = levels(draft_tbl[['era']]),
  ylab = 'Share At Least 6-Foot-2',
  xlab = 'Draft Era'
)
graphics::legend(
  'topright',
  legend = c('Round 1', 'Rounds 2-7'),
  fill = c('#1b4332', '#95d5b2'),
  bty = 'n'
)

## ----position-table-----------------------------------------------------------
# Summarize size by era and position family.
position_summary <- aggregate(
  cbind(height, weight) ~ era + positionBucket,
  data = draft_tbl,
  FUN = mean
)
make_table(
  position_summary,
  caption = 'Average drafted skater size by era and position family.'
)

## ----model--------------------------------------------------------------------
# Fit simple draft-height model.
draft_fit <- stats::lm(
  height ~ draftYear + I(roundNumber == 1) + I(positionCode == 'D'),
  data = draft_tbl
)
draft_fit_tbl <- as.data.frame(summary(draft_fit)$coefficients)
draft_fit_tbl[['term']] <- rownames(draft_fit_tbl)
rownames(draft_fit_tbl) <- NULL
draft_fit_tbl[['term']] <- c(
  'Intercept',
  'Draft year',
  'First-round indicator',
  'Defense indicator'
)
draft_fit_tbl <- draft_fit_tbl[, c(
  'term',
  'Estimate',
  'Std. Error',
  't value',
  'Pr(>|t|)'
)]
make_table(
  draft_fit_tbl,
  caption = 'Linear model of drafted skater height.',
  digits = 4
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}

## ----data---------------------------------------------------------------------
# Pull regular-season game and team records.
games_tbl <- nhlscraper::games()
teams_tbl <- nhlscraper::teams()

# Keep salary-cap regular-season games.
games_tbl <- games_tbl[
  games_tbl[['seasonId']] >= 20052006 &
    games_tbl[['gameTypeId']] == 2,
  c(
    'gameId',
    'seasonId',
    'gameDate',
    'homeTeamId',
    'visitingTeamId',
    'homeScore',
    'visitingScore'
  )
]

# Expand games into team-game rows.
home_games <- data.frame(
  gameId = games_tbl[['gameId']],
  seasonId = games_tbl[['seasonId']],
  gameDate = as.Date(games_tbl[['gameDate']]),
  teamId = games_tbl[['homeTeamId']],
  isHome = TRUE,
  goalsFor = games_tbl[['homeScore']],
  goalsAgainst = games_tbl[['visitingScore']]
)
away_games <- data.frame(
  gameId = games_tbl[['gameId']],
  seasonId = games_tbl[['seasonId']],
  gameDate = as.Date(games_tbl[['gameDate']]),
  teamId = games_tbl[['visitingTeamId']],
  isHome = FALSE,
  goalsFor = games_tbl[['visitingScore']],
  goalsAgainst = games_tbl[['homeScore']]
)
team_games <- rbind(home_games, away_games)

# Sort within team and compute off-days since previous game.
team_games <- team_games[order(
  team_games[['teamId']],
  team_games[['gameDate']],
  team_games[['gameId']]
), ]
team_games[['previousGameDate']] <- ave(
  team_games[['gameDate']],
  team_games[['teamId']],
  FUN = function(x) c(as.Date(NA), utils::head(x, -1))
)
team_games[['restDays']] <-
  as.integer(team_games[['gameDate']] - team_games[['previousGameDate']]) - 1L
team_games <- team_games[!is.na(team_games[['restDays']]), ]

# Bucket rest and compute result metrics.
team_games[['restBucket']] <- ifelse(
  team_games[['restDays']] >= 3,
  '3+',
  as.character(team_games[['restDays']])
)
team_games[['restBucket']] <- factor(
  team_games[['restBucket']],
  levels = c('0', '1', '2', '3+')
)
team_games[['win']] <- team_games[['goalsFor']] > team_games[['goalsAgainst']]
team_games[['goalDiff']] <-
  team_games[['goalsFor']] - team_games[['goalsAgainst']]

## ----rest-table---------------------------------------------------------------
# Summarize results by rest bucket.
rest_summary <- aggregate(
  cbind(win, goalDiff) ~ restBucket,
  data = team_games,
  FUN = mean
)
rest_counts <- as.data.frame(table(team_games[['restBucket']]))
names(rest_counts) <- c('restBucket', 'games')
rest_summary <- merge(rest_summary, rest_counts, by = 'restBucket')
rest_summary <- rest_summary[
  match(levels(team_games[['restBucket']]), rest_summary[['restBucket']]),
  c('restBucket', 'games', 'win', 'goalDiff')
]

make_table(
  rest_summary,
  caption = 'Win rate and average goal differential by rest bucket.'
)

## ----rest-plot, fig.cap = 'Win rate across rest buckets in the salary-cap era.'----
graphics::barplot(
  rest_summary[['win']],
  names.arg = rest_summary[['restBucket']],
  col = c('#d62828', '#f77f00', '#fcbf49', '#90be6d'),
  border = NA,
  ylim = c(0, 0.6),
  xlab = 'Days of Rest',
  ylab = 'Win Rate'
)
graphics::abline(
  h = mean(team_games[['win']]),
  lty = 2,
  col = '#4d4d4d'
)

## ----home-road-table----------------------------------------------------------
# Summarize rest effect by venue.
home_road_summary <- aggregate(
  win ~ restBucket + isHome,
  data = team_games,
  FUN = mean
)
home_wins <- home_road_summary[
  home_road_summary[['isHome']],
  c('restBucket', 'win')
]
away_wins <- home_road_summary[
  !home_road_summary[['isHome']],
  c('restBucket', 'win')
]
names(home_wins)[names(home_wins) == 'win'] <- 'homeWinRate'
names(away_wins)[names(away_wins) == 'win'] <- 'awayWinRate'
home_road_table <- merge(home_wins, away_wins, by = 'restBucket')
home_road_table <- home_road_table[
  match(levels(team_games[['restBucket']]), home_road_table[['restBucket']]),
]
make_table(
  home_road_table,
  caption = 'Home and road win rate by rest bucket.'
)

## ----home-road-plot, fig.cap = 'Home and road win rate across rest buckets.'----
graphics::plot(
  seq_len(nrow(home_road_table)),
  home_road_table[['homeWinRate']],
  type = 'b',
  pch = 19,
  lwd = 2,
  col = '#1d3557',
  xaxt = 'n',
  ylim = c(0.35, 0.60),
  xlab = 'Days of Rest',
  ylab = 'Win Rate'
)
graphics::lines(
  seq_len(nrow(home_road_table)),
  home_road_table[['awayWinRate']],
  type = 'b',
  pch = 19,
  lwd = 2,
  col = '#e63946'
)
graphics::axis(
  side = 1,
  at = seq_len(nrow(home_road_table)),
  labels = home_road_table[['restBucket']]
)
graphics::legend(
  'bottomright',
  legend = c('Home', 'Away'),
  col = c('#1d3557', '#e63946'),
  pch = 19,
  lwd = 2,
  bty = 'n'
)

## ----model--------------------------------------------------------------------
# Fit simple win model with rest and venue.
rest_fit <- stats::glm(
  as.integer(win) ~ restBucket + isHome,
  data = team_games,
  family = stats::binomial()
)
rest_fit_tbl <- as.data.frame(summary(rest_fit)$coefficients)
rest_fit_tbl[['term']] <- rownames(rest_fit_tbl)
rownames(rest_fit_tbl) <- NULL
rest_fit_tbl[['term']] <- c(
  'Intercept',
  'One day versus zero',
  'Two days versus zero',
  'Three-plus days versus zero',
  'Home indicator'
)
rest_fit_tbl <- rest_fit_tbl[, c(
  'term',
  'Estimate',
  'Std. Error',
  'z value',
  'Pr(>|z|)'
)]
make_table(
  rest_fit_tbl,
  caption = 'Logistic model of wins on rest and venue.',
  digits = 4
)

## ----team-table---------------------------------------------------------------
# Rank teams by back-to-back win rate.
zero_rest_tbl <- team_games[team_games[['restDays']] == 0, c('teamId', 'win')]
zero_rest_tbl <- aggregate(
  win ~ teamId,
  data = zero_rest_tbl,
  FUN = function(x) c(winRate = mean(x), games = length(x))
)
zero_rest_tbl <- data.frame(
  teamId = zero_rest_tbl[['teamId']],
  winRate = zero_rest_tbl[['win']][, 'winRate'],
  games = zero_rest_tbl[['win']][, 'games']
)
zero_rest_tbl <- zero_rest_tbl[zero_rest_tbl[['games']] >= 50, ]
zero_rest_tbl <- merge(
  zero_rest_tbl,
  teams_tbl[, c('teamId', 'teamTriCode')],
  by = 'teamId',
  all.x = TRUE
)
zero_rest_tbl <- zero_rest_tbl[
  order(-zero_rest_tbl[['winRate']]),
  c('teamTriCode', 'games', 'winRate')
]
zero_rest_tbl <- utils::head(zero_rest_tbl, 10)
make_table(
  zero_rest_tbl,
  caption = 'Best back-to-back win rates among teams with at least 50 zero-rest games.'
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}

## ----data---------------------------------------------------------------------
# Pull playoff, regular-season, and biometric records.
playoff_stats <- nhlscraper::skater_playoff_statistics()
career_stats <- nhlscraper::skater_statistics()[, c(
  'playerId',
  'rsGamesPlayed',
  'rsPoints',
  'positionCode'
)]
player_bios <- nhlscraper::players()[, c(
  'playerId',
  'playerFullName',
  'height',
  'weight'
)]

# Join player-level tables.
analysis_tbl <- merge(
  playoff_stats,
  career_stats,
  by = c('playerId', 'positionCode'),
  all.x = TRUE
)
analysis_tbl <- merge(
  analysis_tbl,
  player_bios,
  by = 'playerId',
  all.x = TRUE
)

# Keep salary-cap skaters with meaningful samples.
analysis_tbl <- analysis_tbl[
  !is.na(analysis_tbl[['height']]) &
    !is.na(analysis_tbl[['weight']]) &
    analysis_tbl[['firstSeasonForGameType']] >= 20052006 &
    analysis_tbl[['gamesPlayed']] >= 20 &
    analysis_tbl[['rsGamesPlayed']] >= 200,
]

# Fill missing names and compute scoring rates.
analysis_tbl[['playerFullName']] <- ifelse(
  is.na(analysis_tbl[['playerFullName']]) |
    analysis_tbl[['playerFullName']] == '',
  paste(
    analysis_tbl[['skaterFirstName']],
    analysis_tbl[['skaterLastName']]
  ),
  analysis_tbl[['playerFullName']]
)
analysis_tbl[['regularPPG']] <-
  analysis_tbl[['rsPoints']] / analysis_tbl[['rsGamesPlayed']]
analysis_tbl[['playoffPPG']] <-
  analysis_tbl[['points']] / analysis_tbl[['gamesPlayed']]
analysis_tbl[['playoffLift']] <-
  analysis_tbl[['playoffPPG']] - analysis_tbl[['regularPPG']]
analysis_tbl[['positionBucket']] <- ifelse(
  analysis_tbl[['positionCode']] == 'D',
  'Defense',
  'Forward'
)

# Assign equal-count weight quartiles.
weight_share <- rank(
  analysis_tbl[['weight']],
  ties.method = 'first'
) / nrow(analysis_tbl)
analysis_tbl[['weightQuartile']] <- cut(
  weight_share,
  breaks = c(0, 0.25, 0.50, 0.75, 1),
  include.lowest = TRUE,
  labels = c('Lightest', 'Second', 'Third', 'Heaviest')
)
nrow(analysis_tbl)

## ----sample-table-------------------------------------------------------------
# Show sample of strongest playoff scorers.
sample_tbl <- analysis_tbl[
  order(-analysis_tbl[['playoffPPG']], -analysis_tbl[['gamesPlayed']]),
  c(
    'playerFullName',
    'positionCode',
    'weight',
    'rsGamesPlayed',
    'gamesPlayed',
    'regularPPG',
    'playoffPPG',
    'playoffLift'
  )
]
sample_tbl <- utils::head(sample_tbl, 8)
make_table(
  sample_tbl,
  caption = 'Top playoff scoring rates among skaters in the working sample.'
)

## ----quartiles----------------------------------------------------------------
# Summarize scoring levels and playoff lift by quartile.
quartile_summary <- aggregate(
  cbind(regularPPG, playoffPPG, playoffLift) ~ weightQuartile,
  data = analysis_tbl,
  FUN = mean
)
quartile_counts <- as.data.frame(table(analysis_tbl[['weightQuartile']]))
names(quartile_counts) <- c('weightQuartile', 'n')
quartile_summary <- merge(
  quartile_summary,
  quartile_counts,
  by = 'weightQuartile'
)
quartile_summary <- quartile_summary[
  match(levels(analysis_tbl[['weightQuartile']]), quartile_summary[['weightQuartile']]),
  c('weightQuartile', 'n', 'regularPPG', 'playoffPPG', 'playoffLift')
]
make_table(
  quartile_summary,
  caption = 'Regular-season scoring, playoff scoring, and playoff lift by weight quartile.'
)

## ----quartile-plot, fig.cap = 'Playoff scoring level and playoff lift by weight quartile.'----
# Plot playoff scoring distribution and average lift.
old_par <- graphics::par(no.readonly = TRUE)
graphics::par(mfrow = c(1, 2), mar = c(8, 4, 3, 1))
graphics::boxplot(
  playoffPPG ~ weightQuartile,
  data = analysis_tbl,
  col = c('#d9ed92', '#b5e48c', '#76c893', '#34a0a4'),
  border = '#3a5a40',
  las = 2,
  ylab = 'Playoff Points Per Game',
  xlab = ''
)
graphics::barplot(
  quartile_summary[['playoffLift']],
  names.arg = quartile_summary[['weightQuartile']],
  col = c('#f4d35e', '#ee964b', '#f95738', '#7b2cbf'),
  border = NA,
  las = 2,
  ylab = 'Playoff Lift',
  xlab = ''
)
graphics::abline(h = 0, lty = 2, col = '#4d4d4d')
graphics::par(old_par)

## ----risers-------------------------------------------------------------------
# Show largest positive playoff lifts among larger-sample skaters.
risers_tbl <- analysis_tbl[
  analysis_tbl[['gamesPlayed']] >= 40,
  c(
    'playerFullName',
    'positionBucket',
    'weight',
    'regularPPG',
    'playoffPPG',
    'playoffLift'
  )
]
risers_tbl <- risers_tbl[order(-risers_tbl[['playoffLift']]), ]
risers_tbl <- utils::head(risers_tbl, 10)
make_table(
  risers_tbl,
  caption = 'Largest playoff scoring lifts among skaters with at least 40 playoff games.'
)

## ----model--------------------------------------------------------------------
# Fit simple playoff-lift model.
lift_fit <- stats::lm(
  playoffLift ~ height + weight + I(positionCode == 'D'),
  data = analysis_tbl
)
lift_fit_tbl <- as.data.frame(summary(lift_fit)$coefficients)
lift_fit_tbl[['term']] <- rownames(lift_fit_tbl)
rownames(lift_fit_tbl) <- NULL
lift_fit_tbl[['term']] <- c(
  'Intercept',
  'Height',
  'Weight',
  'Defense indicator'
)
lift_fit_tbl <- lift_fit_tbl[, c(
  'term',
  'Estimate',
  'Std. Error',
  't value',
  'Pr(>|t|)'
)]
make_table(
  lift_fit_tbl,
  caption = 'Linear model of playoff scoring lift on height, weight, and position.',
  digits = 4
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}

## ----data---------------------------------------------------------------------
# Pull summary, play-by-play, and roster context.
game_summary <- nhlscraper::gc_summary(2023030417)
pbp_xg <- nhlscraper::calculate_expected_goals(
  nhlscraper::gc_pbp(2023030417)
)
if (!('xG' %in% names(pbp_xg))) {
  stop(
    paste(
      "calculate_expected_goals() did not return an xG column during article build.",
      "This usually means the loaded nhlscraper namespace is stale or the scorer hit",
      "a runtime error and returned the input unchanged."
    )
  )
}
rosters <- nhlscraper::game_rosters(2023030417)

# Build team and player labels.
home_id <- game_summary[['homeTeam']][['id']]
away_id <- game_summary[['awayTeam']][['id']]
home_abbrev <- game_summary[['homeTeam']][['abbrev']]
away_abbrev <- game_summary[['awayTeam']][['abbrev']]

rosters[['playerFullName']] <- paste(
  rosters[['playerFirstName']],
  rosters[['playerLastName']]
)
rosters[['teamTriCode']] <- ifelse(
  rosters[['teamId']] == home_id,
  home_abbrev,
  away_abbrev
)

# Keep shot events with positive xG.
shots <- pbp_xg[!is.na(pbp_xg[['xG']]) & pbp_xg[['xG']] > 0, , drop = FALSE]
roster_match <- match(shots[['shootingPlayerId']], rosters[['playerId']])
shots[['playerFullName']] <- rosters[['playerFullName']][roster_match]
shots[['teamTriCode']] <- rosters[['teamTriCode']][roster_match]
shots[['timeInPeriod']] <- sprintf(
  '%02d:%02d',
  shots[['secondsElapsedInPeriod']] %/% 60,
  shots[['secondsElapsedInPeriod']] %% 60
)

## ----team-table---------------------------------------------------------------
# Summarize team-level scoreboard and xG results.
team_table <- data.frame(
  team = c(home_abbrev, away_abbrev),
  goals = c(
    game_summary[['homeTeam']][['score']],
    game_summary[['awayTeam']][['score']]
  ),
  shotsOnGoal = c(
    game_summary[['homeTeam']][['sog']],
    game_summary[['awayTeam']][['sog']]
  ),
  xG = c(
    sum(shots[['xG']][shots[['eventOwnerTeamId']] == home_id], na.rm = TRUE),
    sum(shots[['xG']][shots[['eventOwnerTeamId']] == away_id], na.rm = TRUE)
  )
)
team_table[['xGPerShot']] <-
  team_table[['xG']] / team_table[['shotsOnGoal']]
make_table(
  team_table,
  caption = 'Game 7 team context: score, shots on goal, and expected goals.'
)

## ----period-table-------------------------------------------------------------
# Summarize xG by period and team.
period_data <- data.frame(
  periodNumber = shots[['periodNumber']],
  eventOwnerTeamId = shots[['eventOwnerTeamId']],
  xG = shots[['xG']]
)
period_summary <- aggregate(
  xG ~ periodNumber + eventOwnerTeamId,
  data = period_data,
  FUN = sum
)
period_ids <- sort(unique(shots[['periodNumber']]))
period_table <- data.frame(period = period_ids)
home_match <- match(
  period_ids,
  period_summary[['periodNumber']][period_summary[['eventOwnerTeamId']] == home_id]
)
away_match <- match(
  period_ids,
  period_summary[['periodNumber']][period_summary[['eventOwnerTeamId']] == away_id]
)
period_table[[paste0(home_abbrev, '_xG')]] <-
  period_summary[['xG']][period_summary[['eventOwnerTeamId']] == home_id][home_match]
period_table[[paste0(away_abbrev, '_xG')]] <-
  period_summary[['xG']][period_summary[['eventOwnerTeamId']] == away_id][away_match]
period_table[is.na(period_table)] <- 0
make_table(
  period_table,
  caption = 'Expected goals by period in Game 7.'
)

## ----chance-table-------------------------------------------------------------
# Show largest individual shot-quality events.
chance_idx <- order(-shots[['xG']])
chance_table <- data.frame(
  player = shots[['playerFullName']][chance_idx],
  team = shots[['teamTriCode']][chance_idx],
  period = shots[['periodNumber']][chance_idx],
  timeInPeriod = shots[['timeInPeriod']][chance_idx],
  xCoordNorm = shots[['xCoordNorm']][chance_idx],
  yCoordNorm = shots[['yCoordNorm']][chance_idx],
  xG = shots[['xG']][chance_idx]
)
chance_table <- utils::head(chance_table, 10)
make_table(
  chance_table,
  caption = 'Highest-xG individual chances in Game 7.',
  digits = 3
)

## ----cum-data-----------------------------------------------------------------
# Build cumulative xG paths for both teams.
build_cum_path <- function(team_id) {
  team_idx <- shots[['eventOwnerTeamId']] == team_id
  team_shots <- data.frame(
    eventId = shots[['eventId']][team_idx],
    secondsElapsedInGame = shots[['secondsElapsedInGame']][team_idx],
    xG = shots[['xG']][team_idx]
  )
  team_shots <- team_shots[order(
    team_shots[['secondsElapsedInGame']],
    team_shots[['eventId']]
  ), ]
  data.frame(
    minutes = c(0, team_shots[['secondsElapsedInGame']] / 60),
    cumXG = c(0, cumsum(team_shots[['xG']]))
  )
}
home_path <- build_cum_path(home_id)
away_path <- build_cum_path(away_id)

## ----cum-plot, fig.cap = 'Cumulative expected goals in Game 7 of the 2024 Stanley Cup Final.'----
graphics::plot(
  home_path[['minutes']],
  home_path[['cumXG']],
  type = 's',
  lwd = 2,
  col = '#c1121f',
  xlim = c(0, 60),
  ylim = c(0, max(c(home_path[['cumXG']], away_path[['cumXG']])) * 1.05),
  xlab = 'Minutes Elapsed',
  ylab = 'Cumulative Expected Goals'
)
graphics::lines(
  away_path[['minutes']],
  away_path[['cumXG']],
  type = 's',
  lwd = 2,
  col = '#003049'
)
graphics::abline(v = c(20, 40), lty = 3, col = '#8d99ae')
graphics::legend(
  'topleft',
  legend = c(home_abbrev, away_abbrev),
  col = c('#c1121f', '#003049'),
  lwd = 2,
  bty = 'n'
)

## ----rink-plot, fig.cap = 'Shot-quality map for Game 7. Point size scales with expected goals.'----
# Split shot map inputs by team.
home_shots <- shots[shots[['eventOwnerTeamId']] == home_id, ]
away_shots <- shots[shots[['eventOwnerTeamId']] == away_id, ]
nhlscraper::draw_NHL_rink()
graphics::points(
  home_shots[['xCoordNorm']],
  home_shots[['yCoordNorm']],
  pch = 19,
  col = grDevices::rgb(0.76, 0.07, 0.12, 0.55),
  cex = 0.6 + 7 * sqrt(home_shots[['xG']])
)
graphics::points(
  away_shots[['xCoordNorm']],
  away_shots[['yCoordNorm']],
  pch = 19,
  col = grDevices::rgb(0.00, 0.19, 0.29, 0.55),
  cex = 0.6 + 7 * sqrt(away_shots[['xG']])
)
graphics::legend(
  'topright',
  legend = c(home_abbrev, away_abbrev),
  pch = 19,
  col = c(
    grDevices::rgb(0.76, 0.07, 0.12, 0.75),
    grDevices::rgb(0.00, 0.19, 0.29, 0.75)
  ),
  bty = 'n'
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}

## ----leaders------------------------------------------------------------------
# Pull 2024-25 team EDGE leaders.
edge_leaders <- nhlscraper::team_edge_leaders(
  season = 20242025,
  game_type = 2
)

# Build compact leader table.
leader_table <- data.frame(
  metric = c(
    'Shots over 90 mph',
    'Bursts over 22 mph',
    'Distance per 60',
    'High-danger shots on goal',
    'Offensive-zone time',
    'Neutral-zone time',
    'Defensive-zone time'
  ),
  team = c(
    edge_leaders[['shotAttemptsOver90']][['team']][['abbrev']],
    edge_leaders[['burstsOver22']][['team']][['abbrev']],
    edge_leaders[['distancePer60']][['team']][['abbrev']],
    edge_leaders[['highDangerSOG']][['team']][['abbrev']],
    edge_leaders[['offensiveZoneTime']][['team']][['abbrev']],
    edge_leaders[['neutralZoneTime']][['team']][['abbrev']],
    edge_leaders[['defensiveZoneTime']][['team']][['abbrev']]
  ),
  value = c(
    as.character(edge_leaders[['shotAttemptsOver90']][['attempts']]),
    as.character(edge_leaders[['burstsOver22']][['bursts']]),
    sprintf('%.2f miles', edge_leaders[['distancePer60']][['distanceSkated']][['imperial']]),
    as.character(edge_leaders[['highDangerSOG']][['sog']]),
    sprintf('%.3f', edge_leaders[['offensiveZoneTime']][['zoneTime']]),
    sprintf('%.3f', edge_leaders[['neutralZoneTime']][['zoneTime']]),
    sprintf('%.3f', edge_leaders[['defensiveZoneTime']][['zoneTime']])
  ),
  stringsAsFactors = FALSE
)
make_table(
  leader_table,
  caption = 'League-leading 2024-25 team EDGE categories.'
)

## ----profiles-----------------------------------------------------------------
# Define selected teams and robust fetch helpers.
team_ids <- c(CAR = 12, COL = 21, EDM = 22, FLA = 13, WSH = 15)
fetch_with_retry <- function(fetch_fun, validator, tries = 3) {
  for (i in seq_len(tries)) {
    value <- try(fetch_fun(), silent = TRUE)
    if (!inherits(value, 'try-error') && validator(value)) {
      return(value)
    }
    Sys.sleep(i / 4)
  }
  NULL
}
valid_df <- function(x, required_cols) {
  is.data.frame(x) && nrow(x) > 0 && all(required_cols %in% names(x))
}
extract_name <- function(first_name, last_name) {
  if (is.na(first_name) || is.na(last_name) || first_name == '' || last_name == '') {
    return(NA_character_)
  }
  paste(first_name, last_name)
}
build_team_profile <- function(team_code, team_id) {
  team_summary <- fetch_with_retry(
    function() nhlscraper::team_edge_summary(
      team = team_id,
      season = 20242025,
      game_type = 2
    ),
    function(x) is.list(x) && 'team' %in% names(x)
  )
  zone_rows <- fetch_with_retry(
    function() nhlscraper::team_edge_zone_time(
      team = team_id,
      season = 20242025,
      game_type = 2,
      category = 'details'
    ),
    function(x) valid_df(x, c('strengthCode', 'offensiveZonePctg'))
  )
  skating_rows <- fetch_with_retry(
    function() nhlscraper::team_edge_skating_speed(
      team = team_id,
      season = 20242025,
      game_type = 2,
      category = 'details'
    ),
    function(x) {
      valid_df(x, c(
        'positionCode',
        'maxSkatingSpeed.imperial',
        'burstsOver22.value'
      ))
    }
  )
  shot_speed_rows <- fetch_with_retry(
    function() nhlscraper::team_edge_shot_speed(
      team = team_id,
      season = 20242025,
      game_type = 2,
      category = 'details'
    ),
    function(x) {
      valid_df(x, c(
        'position',
        'topShotSpeed.imperial',
        'shotAttempts90To100.value'
      ))
    }
  )
  shot_location_rows <- fetch_with_retry(
    function() nhlscraper::team_edge_shot_location(
      team = team_id,
      season = 20242025,
      game_type = 2,
      category = 'details'
    ),
    function(x) valid_df(x, c('area', 'sog'))
  )
  if (
    is.null(zone_rows) ||
      is.null(skating_rows) ||
      is.null(shot_speed_rows) ||
      is.null(shot_location_rows)
  ) {
    return(data.frame(
      team = team_code,
      points = if (is.null(team_summary)) NA_real_ else as.numeric(team_summary[['team']][['points']]),
      wins = if (is.null(team_summary)) NA_real_ else as.numeric(team_summary[['team']][['wins']]),
      offensiveZonePctg = NA_real_,
      maxSkatingSpeed = NA_real_,
      burstsOver22 = NA_real_,
      shotAttemptsOver90 = NA_real_,
      hardestShot = NA_real_,
      interiorShare = NA_real_,
      circleShare = NA_real_,
      pointShare = NA_real_,
      otherShare = NA_real_,
      fastestSkater = NA_character_,
      hardestShooter = NA_character_,
      stringsAsFactors = FALSE
    ))
  }
  zone_row <- zone_rows[zone_rows[['strengthCode']] == 'all', , drop = FALSE]
  skating_row <- skating_rows[skating_rows[['positionCode']] == 'all', , drop = FALSE]
  shot_speed_row <- shot_speed_rows[shot_speed_rows[['position']] == 'all', , drop = FALSE]
  if (!nrow(zone_row)) zone_row <- zone_rows[1, , drop = FALSE]
  if (!nrow(skating_row)) skating_row <- skating_rows[1, , drop = FALSE]
  if (!nrow(shot_speed_row)) shot_speed_row <- shot_speed_rows[1, , drop = FALSE]
  interior_mask <- shot_location_rows[['area']] %in% c(
    'Crease',
    'Low Slot',
    'L Net Side',
    'R Net Side'
  )
  circle_mask <- shot_location_rows[['area']] %in% c(
    'High Slot',
    'L Circle',
    'R Circle'
  )
  point_mask <- shot_location_rows[['area']] %in% c(
    'Center Point',
    'L Point',
    'R Point',
    'Outside L',
    'Outside R',
    'Beyond Red Line'
  )
  total_shots <- sum(shot_location_rows[['sog']])
  data.frame(
    team = team_code,
    points = if (is.null(team_summary)) NA_real_ else as.numeric(team_summary[['team']][['points']]),
    wins = if (is.null(team_summary)) NA_real_ else as.numeric(team_summary[['team']][['wins']]),
    offensiveZonePctg = as.numeric(zone_row[['offensiveZonePctg']][1]),
    maxSkatingSpeed = as.numeric(skating_row[['maxSkatingSpeed.imperial']][1]),
    burstsOver22 = as.numeric(skating_row[['burstsOver22.value']][1]),
    shotAttemptsOver90 = as.numeric(
      shot_speed_row[['shotAttemptsOver100.value']][1] +
        shot_speed_row[['shotAttempts90To100.value']][1]
    ),
    hardestShot = as.numeric(shot_speed_row[['topShotSpeed.imperial']][1]),
    interiorShare = sum(shot_location_rows[['sog']][interior_mask]) / total_shots,
    circleShare = sum(shot_location_rows[['sog']][circle_mask]) / total_shots,
    pointShare = sum(shot_location_rows[['sog']][point_mask]) / total_shots,
    otherShare = sum(shot_location_rows[['sog']][!(interior_mask | circle_mask | point_mask)]) / total_shots,
    fastestSkater = extract_name(
      skating_row[['maxSkatingSpeed.overlay.player.firstName.default']][1],
      skating_row[['maxSkatingSpeed.overlay.player.lastName.default']][1]
    ),
    hardestShooter = extract_name(
      shot_speed_row[['topShotSpeed.overlay.player.firstName.default']][1],
      shot_speed_row[['topShotSpeed.overlay.player.lastName.default']][1]
    ),
    stringsAsFactors = FALSE
  )
}
team_profiles <- Map(
  build_team_profile,
  team_code = names(team_ids),
  team_id = unname(team_ids)
)
team_profiles <- do.call(rbind, team_profiles)
rownames(team_profiles) <- NULL
profile_table <- team_profiles[, c(
  'team',
  'points',
  'wins',
  'offensiveZonePctg',
  'maxSkatingSpeed',
  'burstsOver22',
  'shotAttemptsOver90',
  'hardestShot',
  'interiorShare'
)]
make_table(
  profile_table,
  caption = 'Five-team 2024-25 EDGE profile comparison.'
)

## ----pace-plot, fig.cap = 'Territorial control and pace look different across elite 2024-25 teams.'----
# Plot territorial control and burst volume.
old_par <- graphics::par(no.readonly = TRUE)
graphics::par(mfrow = c(1, 2), mar = c(5, 7, 3, 1))
ordered_zone <- team_profiles[order(team_profiles[['offensiveZonePctg']]), ]
graphics::barplot(
  ordered_zone[['offensiveZonePctg']],
  names.arg = ordered_zone[['team']],
  horiz = TRUE,
  las = 1,
  col = '#2a9d8f',
  border = NA,
  xlab = 'Offensive-Zone Share'
)
ordered_bursts <- team_profiles[order(team_profiles[['burstsOver22']]), ]
graphics::barplot(
  ordered_bursts[['burstsOver22']],
  names.arg = ordered_bursts[['team']],
  horiz = TRUE,
  las = 1,
  col = '#e76f51',
  border = NA,
  xlab = 'Bursts Over 22 MPH'
)
graphics::par(old_par)

## ----shot-mix-plot, fig.cap = 'Shot-geography mix for five elite 2024-25 teams.'----
# Plot shot mix shares.
shot_mix <- t(as.matrix(team_profiles[, c(
  'interiorShare',
  'circleShare',
  'pointShare',
  'otherShare'
)]))
colnames(shot_mix) <- team_profiles[['team']]
rownames(shot_mix) <- c(
  'Interior',
  'Circles and slot',
  'Points and perimeter',
  'Other'
)
graphics::barplot(
  shot_mix,
  beside = FALSE,
  col = c('#1b4332', '#40916c', '#74c69d', '#d8f3dc'),
  ylim = c(0, 1),
  ylab = 'Share of Tracked Shots',
  xlab = 'Team'
)
graphics::legend(
  'topright',
  legend = rownames(shot_mix),
  fill = c('#1b4332', '#40916c', '#74c69d', '#d8f3dc'),
  bty = 'n'
)

## ----player-table-------------------------------------------------------------
# Show players behind each team's most extreme speed and shot events.
player_table <- team_profiles[, c(
  'team',
  'fastestSkater',
  'maxSkatingSpeed',
  'hardestShooter',
  'hardestShot'
)]
make_table(
  player_table,
  caption = "Players behind each team's fastest burst and hardest shot."
)

#' Access all the seasons
#' 
#' `get_seasons()` is deprecated. Use [seasons()] instead.
#' 
#' @returns data.frame with one row per season
#' @export

get_seasons <- function() {
  .Deprecated(
    new     = 'seasons()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_seasons()` is deprecated.',
      'Use `seasons()` instead.'
    )
  )
  seasons()
}

#' Access the season and game type as of now
#' 
#' `get_season_now()` is defunct. Use [season_now()] and/or [game_type_now()] 
#' instead.
#' 
#' @export

get_season_now <- function() {
  .Defunct(
    new     = 'season_now()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_season_now()` is defunct.',
      'Use `season_now()` and/or `game_type_now()` instead.'
    )
  )
}

#' Access the standings rules by season
#' 
#' `get_standings_information()` is deprecated. Use [standings_rules()] 
#' instead.
#' 
#' @returns data.frame with one row per season
#' @export

get_standings_information <- function() {
  .Deprecated(
    new     = 'standings_rules()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_standings_information()` is deprecated.',
      'Use `standings_rules()` instead.'
    )
  )
  standings_rules()
}

#' Access the standings for a date
#' 
#' `get_standings()` is deprecated. Use [standings()] instead.
#' 
#' @inheritParams standings
#' @returns data.frame with one row per team
#' @export

get_standings <- function(date = '2025-01-01') {
  .Deprecated(
    new     = 'standings()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_standings()` is deprecated.',
      'Use `standings()` instead.'
    )
  )
  standings(date)
}

#' Access the schedule for a date
#' 
#' `get_schedule()` is deprecated. Use [schedule()] instead.
#' 
#' @inheritParams standings
#' @returns data.frame with one row per game
#' @export

get_schedule <- function(date = '2025-01-01') {
  .Deprecated(
    new     = 'schedule()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_schedule()` is deprecated.',
      'Use `schedule()` instead.'
    )
  )
  schedule(date)
}

#' Access all the venues
#' 
#' `get_venues()` is deprecated. Use [venues()] instead.
#' 
#' @returns data.frame with one row per venue
#' @export

get_venues <- function() {
  .Deprecated(
    new     = 'venues()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_venues()` is deprecated.',
      'Use `venues()` instead.'
    )
  )
  venues()
}

#' Access the attendance by season and game type
#' 
#' `get_attendance()` is deprecated. Use [attendance()] instead.
#' 
#' @returns data.frame with one row per season
#' @export

get_attendance <- function() {
  .Deprecated(
    new     = 'attendance()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_attendance()` is deprecated.',
      'Use `attendance()` instead.'
    )
  )
  attendance()
}

#' Access the goalie statistics leaders for a season, game type, and category
#' 
#' `get_goalie_leaders()` is deprecated. Use [goalie_leaders()] instead.
#'
#' @inheritParams roster_statistics
#' @param category character of 'wins', 'shutouts', 'savePctg', or 
#' 'goalsAgainstAverage'
#' @returns data.frame with one row per player
#' @export

get_goalie_leaders <- function(
  season    = 'current',
  game_type = '',
  category  = 'wins'
) {
  .Deprecated(
    new     = 'goalie_leaders()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_goalie_leaders()` is deprecated.',
      'Use `goalie_leaders()` instead.'
    )
  )
  category <- switch(
    category,
    wins                = 'w',
    shutouts            = 's',
    savePctg            = 's%',
    goalsAgainstAverage = 'gaa'
  )
  goalie_leaders(season, game_type, category)
}

#' Access the goalies on milestone watch
#' 
#' `get_goalie_milestones()` is deprecated. Use [goalie_milestones()] 
#' instead.
#'
#' @returns data.frame with one row per player
#' @export

get_goalie_milestones <- function() {
  .Deprecated(
    new     = 'goalie_milestones()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_goalie_milestones()` is deprecated.',
      'Use `goalie_milestones()` instead.'
    )
  )
  goalie_milestones()
}

#' Access all the goalies for a range of seasons
#' 
#' `get_goalies()` is defunct. Use [players()] instead.
#'
#' @export

get_goalies <- function() {
  .Defunct(
    new     = 'players()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_goalies()` is defunct',
      'Use `players()` instead.'
    )
  )
}

#' Access various reports for all the goalies by season or game
#' 
#' `get_goalie_statistics()` is defunct. Use [goalie_season_report()] or 
#' [goalie_game_report()] instead.
#' 
#' @export

get_goalie_statistics <- function() {
  .Defunct(
    new     = 'goalie_season_report()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_goalie_statistics()` is defunct.',
      'Use `goalie_season_report()` and/or `goalie_game_report()` instead.'
    )
  )
}

# Public helpers ----------------------------------------------------------

#' Add event-to-event deltas to a play-by-play
#'
#' `add_deltas()` adds event-to-event deltas in raw and normalized x/y,
#' distance, and angle for a play-by-play. Sequences are bounded by faceoffs:
#' each sequence begins at a faceoff, faceoff rows do not look backward across
#' the boundary, and subsequent events are compared to the most recent prior
#' valid-spatial event in the same faceoff-bounded sequence. Shootout and
#' penalty-shot rows (`0101`/`1010`) are left as `NA` and do not serve as
#' anchors for later rows. When multiple events in a sequence share the same
#' recorded second, zero-time denominators used internally are replaced by
#' `1 / n`, where `n` is the number of events in that same second within the
#' sequence.
#'
#' @param play_by_play data.frame of play-by-play(s) using the current public schema returned by [gc_play_by_play()], [gc_play_by_plays()], [wsc_play_by_play()], or [wsc_play_by_plays()]
#' @returns data.frame with one row per event (play) and added columns: `eventIdPrev`, `secondsElapsedInSequence`, `dSecondsElapsedInSequence`, `dXCoord`, `dYCoord`, `dXCoordNorm`, `dYCoordNorm`, `dDistance`, `dAngle`, `dXCoordPerSecond`, `dYCoordPerSecond`, `dXCoordNormPerSecond`, `dYCoordNormPerSecond`, `dDistancePerSecond`, `dAnglePerSecond`
#' @export
add_deltas <- function(play_by_play) {
  delta_ctx <- .compute_pbp_deltas(play_by_play)
  out <- .apply_pbp_delta_columns(play_by_play, delta_ctx)
  attr(out, 'delta_context') <- delta_ctx
  out
}

#' Build an empty play-by-play delta context
#'
#' `.empty_pbp_delta_context()` allocates an empty delta context sized to the
#' number of play-by-play rows.
#'
#' @param n integer row count
#' @returns named list of empty delta vectors
#' @keywords internal
.empty_pbp_delta_context <- function(n) {
  list(
    eventIdPrev = rep(NA_integer_, n),
    secondsElapsedInSequence = rep(NA_real_, n),
    dSecondsElapsedInSequence = rep(NA_real_, n),
    dXCoord = rep(NA_real_, n),
    dYCoord = rep(NA_real_, n),
    dXCoordNorm = rep(NA_real_, n),
    dYCoordNorm = rep(NA_real_, n),
    dDistance = rep(NA_real_, n),
    dAngle = rep(NA_real_, n),
    dXCoordPerSecond = rep(NA_real_, n),
    dYCoordPerSecond = rep(NA_real_, n),
    dXCoordNormPerSecond = rep(NA_real_, n),
    dYCoordNormPerSecond = rep(NA_real_, n),
    dDistancePerSecond = rep(NA_real_, n),
    dAnglePerSecond = rep(NA_real_, n)
  )
}

#' List public delta column names
#'
#' `.pbp_delta_public_cols()` returns the public-facing delta columns inserted
#' by [add_deltas()].
#'
#' @returns character vector of public delta column names
#' @keywords internal
.pbp_delta_public_cols <- function() {
  c(
    'eventIdPrev',
    'secondsElapsedInSequence',
    'dSecondsElapsedInSequence',
    'dXCoord',
    'dYCoord',
    'dXCoordNorm',
    'dYCoordNorm',
    'dDistance',
    'dAngle',
    'dXCoordPerSecond',
    'dYCoordPerSecond',
    'dXCoordNormPerSecond',
    'dYCoordNormPerSecond',
    'dDistancePerSecond',
    'dAnglePerSecond'
  )
}

#' Compute play-by-play deltas in R
#'
#' `.compute_pbp_deltas_in_r()` is the pure-R fallback for event-to-event delta
#' calculations.
#'
#' @inheritParams add_deltas
#' @returns named list containing delta context vectors
#' @keywords internal
.compute_pbp_deltas_in_r <- function(play_by_play) {
  n <- nrow(play_by_play)
  out <- .empty_pbp_delta_context(n)
  if (!n) {
    return(out)
  }

  t <- play_by_play$secondsElapsedInGame
  x_raw <- play_by_play$xCoord
  y_raw <- play_by_play$yCoord
  x_norm <- play_by_play$xCoordNorm
  y_norm <- play_by_play$yCoordNorm
  distance <- play_by_play$distance
  angle <- play_by_play$angle
  event_id <- as.integer(play_by_play$eventId)
  situation_code <- as.character(play_by_play$situationCode)
  is_faceoff <- play_by_play$eventTypeDescKey == 'faceoff'
  is_ps_so <- !is.na(situation_code) & situation_code %in% c('0101', '1010')

  for (g in unique(play_by_play$gameId)) {
    idx <- which(play_by_play$gameId == g)
    idx <- idx[order(t[idx], play_by_play$sortOrder[idx], na.last = TRUE)]
    if (!length(idx)) {
      next
    }
    m <- length(idx)
    t_ord <- t[idx]
    valid_spatial <- !is.na(x_norm[idx]) & !is.na(y_norm[idx]) &
      !is.na(distance[idx]) & !is.na(angle[idx]) &
      !is_ps_so[idx]
    prev_pos <- rep(NA_integer_, m)
    seq_id <- rep(NA_integer_, m)
    same_second_count <- rep(NA_integer_, m)
    last_valid_since_faceoff <- NA_integer_
    current_seq <- 0L
    seq_start_time <- NA_real_

    for (k in seq_len(m)) {
      i <- idx[k]
      if (is_faceoff[i]) {
        current_seq <- current_seq + 1L
        seq_start_time <- t[i]
        seq_id[k] <- current_seq
        prev_pos[k] <- NA_integer_
        if (!is_ps_so[i] && !is.na(seq_start_time) && !is.na(t[i])) {
          out$secondsElapsedInSequence[i] <- t[i] - seq_start_time
        }
      } else if (current_seq > 0L) {
        seq_id[k] <- current_seq
        if (!is_ps_so[i]) {
          prev_pos[k] <- last_valid_since_faceoff
        }
        if (!is_ps_so[i] && !is.na(seq_start_time) && !is.na(t[i])) {
          out$secondsElapsedInSequence[i] <- t[i] - seq_start_time
        }
      }
      if (is_faceoff[i]) {
        last_valid_since_faceoff <- if (valid_spatial[k]) k else NA_integer_
      } else if (valid_spatial[k]) {
        last_valid_since_faceoff <- k
      }
    }

    k <- 1L
    while (k <= m) {
      same_seq <- if (is.na(seq_id[k])) is.na(seq_id) else seq_id == seq_id[k]
      same_time <- if (is.na(t_ord[k])) is.na(t_ord) else t_ord == t_ord[k]
      same_second <- same_seq & same_time
      same_second_count[same_second] <- sum(same_second, na.rm = TRUE)
      k <- max(which(same_second)) + 1L
    }

    curr_k <- which(!is.na(prev_pos))
    if (!length(curr_k)) {
      next
    }
    curr <- idx[curr_k]
    prev <- idx[prev_pos[curr_k]]
    dt <- t[curr] - t[prev]
    dx_raw <- x_raw[curr] - x_raw[prev]
    dy_raw <- y_raw[curr] - y_raw[prev]
    dx_norm <- x_norm[curr] - x_norm[prev]
    dy_norm <- y_norm[curr] - y_norm[prev]
    dd <- distance[curr] - distance[prev]
    da <- angle[curr] - angle[prev]

    out$dSecondsElapsedInSequence[curr] <- dt
    out$dXCoord[curr] <- dx_raw
    out$dYCoord[curr] <- dy_raw
    out$dXCoordNorm[curr] <- dx_norm
    out$dYCoordNorm[curr] <- dy_norm
    out$dDistance[curr] <- dd
    out$dAngle[curr] <- da
    out$eventIdPrev[curr] <- event_id[prev]

    ok <- !is.na(dt) & dt >= 0
    j <- which(ok)
    if (length(j)) {
      cc <- curr[j]
      denom <- dt[j]
      zero_dt <- which(denom == 0)
      if (length(zero_dt)) {
        denom[zero_dt] <- 1 / same_second_count[curr_k[j[zero_dt]]]
      }
      out$dXCoordPerSecond[cc] <- dx_raw[j] / denom
      out$dYCoordPerSecond[cc] <- dy_raw[j] / denom
      out$dXCoordNormPerSecond[cc] <- dx_norm[j] / denom
      out$dYCoordNormPerSecond[cc] <- dy_norm[j] / denom
      out$dDistancePerSecond[cc] <- dd[j] / denom
      out$dAnglePerSecond[cc] <- da[j] / denom
    }
  }

  out
}

#' Compute play-by-play deltas with native fallback
#'
#' `.compute_pbp_deltas()` validates the public schema, then uses the native
#' delta routine when available and falls back to R otherwise.
#'
#' @inheritParams add_deltas
#' @returns named list containing delta context vectors
#' @keywords internal
.compute_pbp_deltas <- function(play_by_play) {
  .require_public_pbp_columns(
    play_by_play,
    c(
      'gameId', 'eventId', 'sortOrder', 'secondsElapsedInGame',
      'eventTypeDescKey', 'situationCode', 'xCoord', 'yCoord',
      'xCoordNorm', 'yCoordNorm', 'distance', 'angle'
    ),
    'add_deltas'
  )
  n <- nrow(play_by_play)
  if (!n) {
    return(.empty_pbp_delta_context(0))
  }
  situation_code <- as.character(play_by_play$situationCode)
  if (!.ensure_local_native_symbol('nhlscraper_pbp_deltas')) {
    return(.compute_pbp_deltas_in_r(play_by_play))
  }
  tryCatch(
    .Call('nhlscraper_pbp_deltas', list(
      as.integer(order(
        play_by_play$gameId,
        play_by_play$secondsElapsedInGame,
        play_by_play$sortOrder,
        seq_len(n),
        na.last = TRUE
      )),
      as.integer(play_by_play$gameId),
      as.integer(play_by_play$eventId),
      as.double(play_by_play$secondsElapsedInGame),
      as.double(play_by_play$xCoord),
      as.double(play_by_play$yCoord),
      as.double(play_by_play$xCoordNorm),
      as.double(play_by_play$yCoordNorm),
      as.double(play_by_play$distance),
      as.double(play_by_play$angle),
      as.logical(play_by_play$eventTypeDescKey == 'faceoff'),
      as.logical(!is.na(situation_code) & situation_code %in% c('0101', '1010'))
    )),
    error = function(e) .compute_pbp_deltas_in_r(play_by_play)
  )
}

#' Attach delta context columns to play-by-play
#'
#' `.apply_pbp_delta_columns()` inserts computed delta columns into a
#' play-by-play and reorders them into the public schema.
#'
#' @inheritParams add_deltas
#' @param delta_ctx named list returned by `.compute_pbp_deltas()`
#' @returns data.frame with public delta columns inserted
#' @keywords internal
.apply_pbp_delta_columns <- function(play_by_play, delta_ctx) {
  public_cols <- .pbp_delta_public_cols()
  legacy_cols <- c(
    'dXN', 'dYN', 'dD', 'dA', 'dT',
    'dXNdT', 'dYNdT', 'dDdT', 'dAdT'
  )
  keep <- setdiff(names(play_by_play), c(public_cols, legacy_cols))
  play_by_play <- play_by_play[, keep, drop = FALSE]
  for (nm in public_cols) {
    play_by_play[[nm]] <- delta_ctx[[nm]]
  }
  after <- match('angle', names(play_by_play))
  insert <- public_cols
  nms <- names(play_by_play)
  play_by_play <- play_by_play[, c(
    nms[seq_len(after)],
    insert,
    setdiff(nms[-seq_len(after)], insert)
  ), drop = FALSE]
  nms <- names(play_by_play)
  anchor_cols <- grep(
    '(^secondsElapsedInPeriodSinceLastShiftAgainst$|SecondsElapsedInPeriodSinceLastShiftAgainst$)',
    nms,
    value = TRUE
  )
  if (length(anchor_cols)) {
    keep <- nms[nms != 'secondsElapsedInSequence']
    after <- max(match(anchor_cols, keep, nomatch = 0L))
    tail_cols <- if (after < length(keep)) {
      keep[(after + 1L):length(keep)]
    } else {
      character()
    }
    play_by_play <- play_by_play[, c(
      keep[seq_len(after)],
      'secondsElapsedInSequence',
      tail_cols
    ), drop = FALSE]
  }
  play_by_play
}

#' Add shooter biometrics to (a) play-by-play(s)
#'
#' `add_shooter_biometrics()` adds shooter biometrics (height, weight, hand, age at game date, and position) to (a) play-by-play(s) for shot attempts.
#'
#' @param play_by_play data.frame of play-by-play(s) using the current public schema returned by [gc_play_by_play()], [gc_play_by_plays()], [wsc_play_by_play()], or [wsc_play_by_plays()]
#' @returns data.frame with one row per event (play) and added columns: `shooterHeight`, `shooterWeight`, `shooterHandCode`, `shooterAge`, and `shooterPositionCode`
#' @export

add_shooter_biometrics <- function(play_by_play) {
  .require_public_pbp_columns(
    play_by_play,
    c('gameId', 'shootingPlayerId', 'scoringPlayerId'),
    'add_shooter_biometrics'
  )
  bios  <- players()
  games <- games()
  shooterId <- ifelse(!is.na(play_by_play$scoringPlayerId), play_by_play$scoringPlayerId, play_by_play$shootingPlayerId)
  gidx <- match(play_by_play$gameId, games$gameId)
  gd   <- as.Date(games$gameDate[gidx])
  pidx <- match(shooterId, bios$playerId)
  ht   <- bios$height[pidx]
  wt   <- bios$weight[pidx]
  hand <- bios$handCode[pidx]
  pos  <- bios$positionCode[pidx]
  bd   <- as.Date(bios$birthDate[pidx])
  gy   <- as.integer(format(gd, '%Y'))
  by   <- as.integer(format(bd, '%Y'))
  gm   <- as.integer(format(gd, '%m'))
  bm   <- as.integer(format(bd, '%m'))
  gD   <- as.integer(format(gd, '%d'))
  bD   <- as.integer(format(bd, '%d'))
  had_bday <- (gm > bm) | (gm == bm & gD >= bD)
  age  <- (gy - by) - ifelse(had_bday, 0L, 1L)
  age[is.na(gd) | is.na(bd)] <- NA_integer_
  play_by_play$shooterHeight       <- ht
  play_by_play$shooterWeight       <- wt
  play_by_play$shooterHandCode     <- hand
  play_by_play$shooterPositionCode <- pos
  play_by_play$shooterAge          <- age
  play_by_play
}

#' Add goalie biometrics to (a) play-by-play(s)
#'
#' `add_goalie_biometrics()` adds goalie biometrics (height, weight, hand, and age at game date) to (a) play-by-play(s) for shot attempts. Goalie identity is resolved from `goalieInNetId` first and `goaliePlayerIdAgainst` second. If neither is available on a row, the added goalie biometrics are left as `NA`.
#'
#' @param play_by_play data.frame of play-by-play(s) using the current public schema returned by [gc_play_by_play()], [gc_play_by_plays()], [wsc_play_by_play()], or [wsc_play_by_plays()]
#' @returns data.frame with one row per event (play) and added columns: `goalieHeight`, `goalieWeight`, `goalieHandCode`, and `goalieAge`
#' @export

add_goalie_biometrics <- function(play_by_play) {
  .require_public_pbp_columns(play_by_play, 'gameId', 'add_goalie_biometrics')
  if (!('goalieInNetId' %in% names(play_by_play)) &&
      !('goaliePlayerIdAgainst' %in% names(play_by_play))) {
    stop(
      "add_goalie_biometrics() requires public play-by-play column(s): goalieInNetId or goaliePlayerIdAgainst",
      call. = FALSE
    )
  }
  bios  <- players()
  games <- games()
  goalieId <- if ('goalieInNetId' %in% names(play_by_play)) {
    goalie_in_net <- suppressWarnings(as.integer(play_by_play$goalieInNetId))
    if ('goaliePlayerIdAgainst' %in% names(play_by_play)) {
      goalie_against <- suppressWarnings(as.integer(play_by_play$goaliePlayerIdAgainst))
      goalie_in_net[is.na(goalie_in_net)] <- goalie_against[is.na(goalie_in_net)]
      goalie_in_net
    } else {
      goalie_in_net
    }
  } else {
    suppressWarnings(as.integer(play_by_play$goaliePlayerIdAgainst))
  }
  gidx <- match(play_by_play$gameId, games$gameId)
  gd   <- as.Date(games$gameDate[gidx])
  pidx <- match(goalieId, bios$playerId)
  ht   <- bios$height[pidx]
  wt   <- bios$weight[pidx]
  hand <- bios$handCode[pidx]
  bd   <- as.Date(bios$birthDate[pidx])
  gy   <- as.integer(format(gd, '%Y'))
  by   <- as.integer(format(bd, '%Y'))
  gm   <- as.integer(format(gd, '%m'))
  bm   <- as.integer(format(bd, '%m'))
  gD   <- as.integer(format(gd, '%d'))
  bD   <- as.integer(format(bd, '%d'))
  had_bday <- (gm > bm) | (gm == bm & gD >= bD)
  age <- (gy - by) - ifelse(had_bday, 0L, 1L)
  age[is.na(gd) | is.na(bd)]  <- NA_integer_
  play_by_play$goalieHeight   <- ht
  play_by_play$goalieWeight   <- wt
  play_by_play$goalieHandCode <- hand
  play_by_play$goalieAge      <- age
  play_by_play
}

#' Add on-ice shift times to a play-by-play
#'
#' `add_shift_times()` adds `SecondsRemainingInShift`,
#' `SecondsElapsedInShift`, and `SecondsElapsedInPeriodSinceLastShift` columns
#' for the on-ice goalies and skaters already present in a public play-by-play.
#' It accepts either a single game play-by-play plus [shift_chart()] data or a
#' season aggregate plus [shift_charts()] data.
#'
#' @param play_by_play data.frame of play-by-play(s) using the current public schema returned by [gc_play_by_play()], [gc_play_by_plays()], [wsc_play_by_play()], or [wsc_play_by_plays()]
#' @param shift_chart data.frame returned by [shift_chart()] or [shift_charts()]
#' @returns data.frame with one row per event (play) and added or updated scalar on-ice shift timing columns
#' @examples
#' \donttest{
#'   pbp <- gc_play_by_play(game = 2023030417)
#'   sc  <- shift_chart(game = 2023030417)
#'   pbp <- add_shift_times(pbp, sc)
#' }
#' @export
add_shift_times <- function(play_by_play, shift_chart) {
  slot_count <- .on_ice_skater_slots(play_by_play = play_by_play)
  required_skater_cols <- unique(c(
    grep('^(?:home|away)Skater[0-9]+PlayerId$', names(play_by_play), value = TRUE),
    paste0('homeSkater', seq_len(slot_count), 'PlayerId'),
    paste0('awaySkater', seq_len(slot_count), 'PlayerId')
  ))
  .require_public_pbp_columns(
    play_by_play,
    c(
      'gameId',
      'periodNumber',
      'secondsElapsedInPeriod',
      'isHome',
      'homeGoaliePlayerId',
      'awayGoaliePlayerId',
      required_skater_cols
    ),
    'add_shift_times'
  )
  .require_shift_chart_columns(
    shift_chart,
    c(
      'gameId',
      'period',
      'playerId',
      'startSecondsElapsedInPeriod',
      'endSecondsElapsedInPeriod'
    ),
    'add_shift_times'
  )
  timing_cols <- c(
    .on_ice_timing_scalar_column_names(
      'SecondsRemainingInShift',
      play_by_play = play_by_play
    ),
    .on_ice_timing_scalar_column_names(
      'SecondsElapsedInShift',
      play_by_play = play_by_play
    ),
    .on_ice_timing_scalar_column_names(
      'SecondsElapsedInPeriodSinceLastShift',
      play_by_play = play_by_play
    )
  )
  relocate_timing_cols <- function(play_by_play) {
    timing_present <- intersect(timing_cols, names(play_by_play))
    id_present <- intersect(
      .on_ice_id_scalar_column_names(play_by_play = play_by_play),
      names(play_by_play)
    )
    if (!length(timing_present) || !length(id_present)) {
      return(play_by_play)
    }
    non_timing <- setdiff(names(play_by_play), timing_present)
    insert_after <- max(match(id_present, non_timing, nomatch = 0L))
    if (insert_after <= 0L) {
      return(play_by_play)
    }
    tail_cols <- if (insert_after < length(non_timing)) {
      non_timing[(insert_after + 1L):length(non_timing)]
    } else {
      character()
    }
    play_by_play[, c(
      non_timing[seq_len(insert_after)],
      timing_present,
      tail_cols
    ), drop = FALSE]
  }
  for (nm in timing_cols) {
    if (!(nm %in% names(play_by_play))) {
      play_by_play[[nm]] <- rep(NA_real_, nrow(play_by_play))
    }
  }
  play_by_play <- relocate_timing_cols(play_by_play)
  if (!nrow(play_by_play) || !nrow(shift_chart)) {
    return(play_by_play)
  }
  timing <- tryCatch(
    .compute_on_ice_shift_timing_matrices(play_by_play, shift_chart),
    error = function(e) NULL
  )
  if (is.null(timing)) {
    return(play_by_play)
  }
  play_by_play <- .assign_on_ice_shift_metric(
    play_by_play,
    home_matrix = timing$homeRemaining,
    away_matrix = timing$awayRemaining,
    metric_suffix = 'SecondsRemainingInShift'
  )
  play_by_play <- .assign_on_ice_shift_metric(
    play_by_play,
    home_matrix = timing$homeElapsed,
    away_matrix = timing$awayElapsed,
    metric_suffix = 'SecondsElapsedInShift'
  )
  play_by_play <- .assign_on_ice_shift_metric(
    play_by_play,
    home_matrix = timing$homeSinceLast,
    away_matrix = timing$awaySinceLast,
    metric_suffix = 'SecondsElapsedInPeriodSinceLastShift'
  )
  relocate_timing_cols(play_by_play)
}

# Internal helpers --------------------------------------------------------

#' Strip the game ID into the season ID, game type ID, and game number for all the events (plays) in a play-by-play
#' 
#' `.strip_game_id()` strips the game ID into season ID, game type ID, and game number for all the events (plays) in a play-by-play.
#' 
#' @param play_by_play data.frame of play-by-play(s); see [gc_play_by_play()] and/or [wsc_play_by_play()] for reference
#' @returns data.frame with one row per event (play) and added columns: `seasonId`, `gameTypeId`, and `gameNumber`
#' @keywords internal

.strip_game_id <- function(play_by_play) {
  game                    <- unique(play_by_play$gameId)
  play_by_play$seasonId   <- game %/% 1e6 * 1e4 + game %/% 1e6 + 1
  play_by_play$gameTypeId <- game %/% 1e4 %% 1e2
  play_by_play$gameNumber <- game %% 1e4
  front                   <- c('gameId', 'eventId', 'seasonId', 'gameTypeId', 'gameNumber', 'sortOrder')
  play_by_play[, c(front, setdiff(names(play_by_play), front))]
}

#' Normalize raw/internal play-by-play names to the public schema
#'
#' `.normalize_public_pbp_names()` renames the raw GameCenter/World Showcase
#' column names used during ingest into the current public play-by-play schema.
#' This helper is for the package's own ingest pipeline only and is not a
#' legacy-input compatibility layer for downstream callers.
#'
#' @param play_by_play data.frame play-by-play object
#' @returns data.frame with current public column names where available
#' @keywords internal
.normalize_public_pbp_names <- function(play_by_play) {
  if (!is.data.frame(play_by_play) || !ncol(play_by_play)) {
    return(play_by_play)
  }
  rename_to_public <- c(
    period = 'periodNumber',
    typeCode = 'eventTypeCode',
    typeDescKey = 'eventTypeDescKey',
    homeSOG = 'homeShots',
    awaySOG = 'awayShots',
    descKey = 'penaltyTypeDescKey',
    duration = 'penaltyDuration'
  )
  for (old in names(rename_to_public)) {
    new <- rename_to_public[[old]]
    if (old %in% names(play_by_play)) {
      if (new %in% names(play_by_play)) {
        fill <- is.na(play_by_play[[new]])
        play_by_play[[new]][fill] <- play_by_play[[old]][fill]
        play_by_play[[old]] <- NULL
      } else {
        names(play_by_play)[names(play_by_play) == old] <- new
      }
    }
  }
  play_by_play
}

#' Identify Return-to-Play round-robin shootout games
#'
#' `.covid_round_robin_shootout_game_ids()` returns the 2019-20 round-robin
#' seeding games that used regular-season overtime/shootout rules despite
#' carrying playoff game IDs.
#'
#' @returns integer vector of game IDs
#' @keywords internal
.covid_round_robin_shootout_game_ids <- function() {
  c(2019030002L, 2019030016L)
}

#' Identify shootout-eligible games for public play-by-play cleanup
#'
#' `.public_pbp_is_shootout_eligible()` returns `TRUE` for games whose overtime
#' rules follow the regular-season shootout path.
#'
#' @param game_id integer game IDs
#' @param game_type_id integer game type IDs
#' @returns logical vector
#' @keywords internal
.public_pbp_is_shootout_eligible <- function(game_id, game_type_id) {
  (!is.na(game_type_id) & game_type_id == 2L) |
    (!is.na(game_id) & game_id %in% .covid_round_robin_shootout_game_ids())
}

#' Derive the legal period length in seconds
#'
#' `.public_pbp_legal_period_seconds()` returns the nominal legal length for each
#' play-by-play row's period under the relevant game context.
#'
#' @param game_id integer game IDs
#' @param game_type_id integer game type IDs
#' @param period integer period numbers
#' @returns integer vector of legal period lengths in seconds
#' @keywords internal
.public_pbp_legal_period_seconds <- function(game_id, game_type_id, period) {
  shootout_eligible <- .public_pbp_is_shootout_eligible(game_id, game_type_id)
  out <- ifelse(
    is.na(period),
    NA_integer_,
    ifelse(
      period <= 3L,
      1200L,
      ifelse(
        shootout_eligible & period == 4L,
        300L,
        ifelse(shootout_eligible & period >= 5L, 0L, 1200L)
      )
    )
  )
  as.integer(out)
}

#' Format elapsed seconds as an `MM:SS` clock
#'
#' `.format_elapsed_clock()` converts elapsed-in-period seconds back into
#' zero-padded `MM:SS` strings.
#'
#' @param seconds integer/numeric vector of elapsed seconds
#' @returns character vector
#' @keywords internal
.format_elapsed_clock <- function(seconds) {
  seconds <- suppressWarnings(as.integer(seconds))
  out <- rep(NA_character_, length(seconds))
  ok <- !is.na(seconds) & seconds >= 0L
  if (any(ok)) {
    mins <- seconds[ok] %/% 60L
    secs <- seconds[ok] %% 60L
    out[ok] <- sprintf('%02d:%02d', mins, secs)
  }
  out
}

#' Strip the timestamp and period number into the time elapsed in the period and game for all the events (plays) in a play-by-play
#' 
#' `.strip_time_period()` strips the timestamp and period number into the time elapsed in the period and game for all the events (plays) in a play-by-play.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added columns `secondsElapsedInPeriod` and `secondsElapsedInGame`
#' @keywords internal

.strip_time_period <- function(play_by_play) {
  isPlayoffs <- play_by_play$gameTypeId == 3
  time_in_period <- as.character(play_by_play$timeInPeriod)
  valid_format <- !is.na(time_in_period) &
    grepl('^[0-9]{1,2}:[0-9]{2}$', time_in_period)
  elp <- rep(NA_integer_, length(time_in_period))
  if (any(valid_format)) {
    tp <- strsplit(time_in_period[valid_format], ':', fixed = TRUE)
    mins <- suppressWarnings(as.integer(vapply(tp, `[`, '', 1L)))
    secs <- suppressWarnings(as.integer(vapply(tp, `[`, '', 2L)))
    valid_clock <- !is.na(mins) & !is.na(secs) & secs >= 0L & secs <= 59L
    parsed <- rep(NA_integer_, length(tp))
    parsed[valid_clock] <- 60L * mins[valid_clock] + secs[valid_clock]
    legal_max <- .public_pbp_legal_period_seconds(
      game_id = play_by_play$gameId[valid_format],
      game_type_id = play_by_play$gameTypeId[valid_format],
      period = play_by_play$periodNumber[valid_format]
    )
    parsed[!is.na(parsed) & !is.na(legal_max) & parsed > legal_max] <- NA_integer_
    elp[valid_format] <- parsed
  }
  play_by_play$secondsElapsedInPeriod <- elp
  base <- ifelse(
    play_by_play$periodNumber <= 3L,
    (play_by_play$periodNumber - 1L) * 1200L,
    ifelse(
      isPlayoffs,
      3600L + (play_by_play$periodNumber - 4L) * 1200L,
      3600L + (play_by_play$periodNumber - 4L) * 300L
    )
  )
  play_by_play$secondsElapsedInGame   <- base + elp
  insert <- c('periodNumber', 'timeInPeriod', 'secondsElapsedInPeriod', 'secondsElapsedInGame')
  keep   <- setdiff(names(play_by_play), insert)
  after  <- match('sortOrder', keep)
  play_by_play[, c(keep[seq_len(after)], insert, keep[(after + 1L):length(keep)])]
}

#' Repair public play-by-play clock/order defects in boundary-row feeds
#'
#' `.repair_public_pbp_sequence()` removes rows with impossible clocks and
#' reorders explicit-boundary feeds into a more coherent public sequence. The
#' repair is intentionally scoped to the modern boundary-row era, where the
#' audited issues are dominated by stale sort-order markers, early `period-end`
#' rows, and opening faceoffs that arrive after live-play rows.
#'
#' @inheritParams .strip_game_id
#' @returns data.frame with repaired clocks and ordering where feasible
#' @keywords internal
.repair_public_pbp_sequence <- function(play_by_play) {
  n <- nrow(play_by_play)
  if (
    !n ||
      !all(c(
        'gameId', 'seasonId', 'gameTypeId', 'periodNumber', 'sortOrder',
        'timeInPeriod', 'secondsElapsedInPeriod', 'eventTypeDescKey'
      ) %in% names(play_by_play))
  ) {
    return(play_by_play)
  }

  type_desc_key <- as.character(play_by_play$eventTypeDescKey)
  has_boundary_rows <- any(
    !is.na(type_desc_key) &
      type_desc_key %in% c('period-start', 'period-end', 'game-end', 'shootout-complete')
  )
  if (!has_boundary_rows || all(play_by_play$seasonId < 20092010L, na.rm = TRUE)) {
    return(play_by_play)
  }

  has_clock_string <- !is.na(play_by_play$timeInPeriod) & nzchar(play_by_play$timeInPeriod)
  valid_clock <- !is.na(play_by_play$secondsElapsedInPeriod)
  drop_idx <- has_clock_string & !valid_clock
  if (any(drop_idx)) {
    play_by_play <- play_by_play[!drop_idx, , drop = FALSE]
  }
  if (!nrow(play_by_play)) {
    return(play_by_play)
  }

  row_id <- seq_len(nrow(play_by_play))
  original_sort <- suppressWarnings(as.integer(play_by_play$sortOrder))
  original_sort[is.na(original_sort)] <- seq_len(nrow(play_by_play))[is.na(original_sort)]
  new_sort <- original_sort
  new_seconds <- suppressWarnings(as.integer(play_by_play$secondsElapsedInPeriod))

  for (g in unique(play_by_play$gameId)) {
    idx_game <- which(play_by_play$gameId == g)
    idx_game <- idx_game[order(original_sort[idx_game], row_id[idx_game], na.last = TRUE)]
    if (!length(idx_game)) {
      next
    }

    periods <- sort(unique(play_by_play$periodNumber[idx_game]))
    periods <- periods[!is.na(periods)]
    idx_game_new <- integer()

    for (p in periods) {
      idx_period <- idx_game[play_by_play$periodNumber[idx_game] == p]
      if (!length(idx_period)) {
        next
      }

      period_types <- as.character(play_by_play$eventTypeDescKey[idx_period])
      period_secs <- new_seconds[idx_period]
      period_sort <- original_sort[idx_period]
      period_game_type <- suppressWarnings(as.integer(play_by_play$gameTypeId[idx_period][1L]))
      legal_max <- .public_pbp_legal_period_seconds(g, period_game_type, p)[1L]

      is_period_start <- !is.na(period_types) & period_types == 'period-start'
      is_period_end <- !is.na(period_types) & period_types == 'period-end'
      is_game_end <- !is.na(period_types) & period_types == 'game-end'
      is_shootout_complete <- !is.na(period_types) & period_types == 'shootout-complete'
      faceoff_pos <- which(!is.na(period_types) & period_types == 'faceoff')
      opening_faceoff_pos <- if (length(faceoff_pos)) faceoff_pos[1L] else integer()

      if (length(opening_faceoff_pos)) {
        pre_faceoff_pos <- seq_len(opening_faceoff_pos - 1L)
        pre_faceoff_pos <- pre_faceoff_pos[!is_period_start[pre_faceoff_pos]]
        if (length(pre_faceoff_pos)) {
          earliest_pre_sec <- suppressWarnings(min(period_secs[pre_faceoff_pos], na.rm = TRUE))
          if (is.finite(earliest_pre_sec)) {
            target_sec <- if (is.na(period_secs[opening_faceoff_pos])) {
              as.integer(earliest_pre_sec)
            } else {
              min(as.integer(period_secs[opening_faceoff_pos]), as.integer(earliest_pre_sec))
            }
            period_secs[opening_faceoff_pos] <- as.integer(target_sec)
          }
        }
      }

      non_terminal_pos <- which(!(is_period_end | is_game_end | is_shootout_complete))
      observed_period_max <- suppressWarnings(max(period_secs[non_terminal_pos], na.rm = TRUE))
      if (!is.finite(observed_period_max)) {
        observed_period_max <- NA_integer_
      } else {
        observed_period_max <- as.integer(observed_period_max)
      }
      has_goal <- any(!is.na(period_types) & period_types == 'goal')
      has_later_period <- any(play_by_play$periodNumber[idx_game] > p, na.rm = TRUE)
      use_legal_terminal <- !is.na(legal_max) && (
        p <= 3L ||
          (.public_pbp_is_shootout_eligible(g, period_game_type) &&
             p == 4L &&
             !has_goal &&
             has_later_period)
      )
      terminal_sec <- suppressWarnings(max(
        c(
          period_secs[is_period_end],
          observed_period_max,
          if (use_legal_terminal) legal_max else NA_integer_
        ),
        na.rm = TRUE
      ))
      if (!is.finite(terminal_sec)) {
        terminal_sec <- if (!is.na(legal_max)) legal_max else 0L
      }
      terminal_sec <- as.integer(max(0L, terminal_sec))
      if (any(is_period_start)) {
        period_secs[is_period_start] <- 0L
      }
      if (any(is_period_end)) {
        period_secs[is_period_end] <- terminal_sec
      }
      if (any(is_game_end)) {
        period_secs[is_game_end] <- terminal_sec
      }
      if (any(is_shootout_complete) && is.na(legal_max)) {
        period_secs[is_shootout_complete] <- terminal_sec
      }

      regular_pos <- which(!(is_period_start | is_period_end | is_game_end | is_shootout_complete))
      if (length(opening_faceoff_pos)) {
        regular_pos <- setdiff(regular_pos, opening_faceoff_pos)
      }
      regular_pos <- regular_pos[order(
        period_secs[regular_pos],
        period_sort[regular_pos],
        row_id[idx_period[regular_pos]],
        na.last = TRUE
      )]

      idx_period_new <- c(
        idx_period[which(is_period_start)],
        if (length(opening_faceoff_pos)) idx_period[opening_faceoff_pos] else integer(),
        idx_period[regular_pos],
        idx_period[which(is_shootout_complete)],
        idx_period[which(is_period_end)],
        idx_period[which(is_game_end)]
      )
      idx_game_new <- c(idx_game_new, idx_period_new)
      new_seconds[idx_period] <- period_secs
    }

    if (length(idx_game_new) != length(idx_game)) {
      missing_idx <- setdiff(idx_game, idx_game_new)
      idx_game_new <- c(idx_game_new, missing_idx)
    }
    idx_game_new <- idx_game_new[!duplicated(idx_game_new)]
    new_sort[idx_game_new] <- sort(original_sort[idx_game], na.last = TRUE)
  }

  play_by_play$sortOrder <- new_sort
  play_by_play$secondsElapsedInPeriod <- new_seconds
  play_by_play$timeInPeriod <- .format_elapsed_clock(play_by_play$secondsElapsedInPeriod)
  play_by_play <- .strip_time_period(play_by_play)
  play_by_play[order(play_by_play$sortOrder, row_id, na.last = TRUE), , drop = FALSE]
}

#' Remove illogically ordered boundary events from a play-by-play
#'
#' `.drop_illogical_ordered_events()` removes stray boundary rows that can
#' appear between a period's `period-end` and the following period's
#' `period-start`. In practice, these are most often duplicated `0:00`
#' `faceoff`/`stoppage` rows that are still tagged to the previous period even
#' though the actual opening sequence for the next period follows immediately
#' after.
#'
#' @inheritParams .strip_game_id
#' @returns data.frame with illogically ordered boundary events removed
#' @keywords internal

.drop_illogical_ordered_events <- function(play_by_play) {
  n <- nrow(play_by_play)
  if (n < 3L) {
    return(play_by_play)
  }
  keep <- rep(TRUE, n)
  game_id <- as.integer(play_by_play$gameId)
  for (g in unique(game_id)) {
    idx <- which(game_id == g)
    idx <- idx[order(play_by_play$sortOrder[idx], na.last = TRUE)]
    if (length(idx) < 3L) {
      next
    }
    for (k in 2:(length(idx) - 1L)) {
      prev_i <- idx[k - 1L]
      cur_i <- idx[k]
      next_i <- idx[k + 1L]
      if (
        identical(play_by_play$eventTypeDescKey[prev_i], 'period-end') &&
          identical(play_by_play$eventTypeDescKey[next_i], 'period-start') &&
          play_by_play$eventTypeDescKey[cur_i] %in% c('faceoff', 'stoppage') &&
          !is.na(play_by_play$secondsElapsedInPeriod[cur_i]) &&
          play_by_play$secondsElapsedInPeriod[cur_i] == 0L &&
          !is.na(play_by_play$secondsElapsedInPeriod[next_i]) &&
          play_by_play$secondsElapsedInPeriod[next_i] == 0L &&
          !is.na(play_by_play$periodNumber[prev_i]) &&
          !is.na(play_by_play$periodNumber[cur_i]) &&
          !is.na(play_by_play$periodNumber[next_i]) &&
          play_by_play$periodNumber[cur_i] == play_by_play$periodNumber[prev_i] &&
          play_by_play$periodNumber[next_i] == play_by_play$periodNumber[prev_i] + 1L
      ) {
        keep[cur_i] <- FALSE
      }
    }
  }
  play_by_play <- play_by_play[keep, , drop = FALSE]
  n <- nrow(play_by_play)
  if (
    n < 3L ||
      !all(c('gameId', 'periodNumber', 'sortOrder', 'eventTypeDescKey') %in% names(play_by_play))
  ) {
    return(play_by_play)
  }

  row_id <- seq_len(n)
  game_id <- as.integer(play_by_play$gameId)
  period <- as.integer(play_by_play$periodNumber)
  type_desc_key <- as.character(play_by_play$eventTypeDescKey)
  sort_order <- as.integer(play_by_play$sortOrder)

  for (g in unique(game_id)) {
    idx_game <- which(game_id == g)
    idx_sorted <- idx_game[order(sort_order[idx_game], row_id[idx_game], na.last = TRUE)]
    if (length(idx_sorted) < 3L) {
      next
    }

    periods <- unique(period[idx_sorted])
    periods <- periods[!is.na(periods)]
    for (p in periods) {
      idx_period <- idx_sorted[period[idx_sorted] == p]
      if (length(idx_period) < 3L) {
        next
      }
      period_types <- type_desc_key[idx_period]
      period_start_pos <- which(period_types == 'period-start')
      faceoff_pos <- which(period_types == 'faceoff')
      if (!length(period_start_pos) || !length(faceoff_pos)) {
        next
      }
      period_start_pos <- period_start_pos[1L]
      faceoff_pos <- faceoff_pos[faceoff_pos > period_start_pos]
      if (!length(faceoff_pos)) {
        next
      }
      faceoff_pos <- faceoff_pos[1L]
      if (faceoff_pos == period_start_pos + 1L) {
        next
      }
      intervening_pos <- seq.int(period_start_pos + 1L, faceoff_pos - 1L)
      intervening_types <- period_types[intervening_pos]
      if (!any(!is.na(intervening_types))) {
        next
      }
      idx_period <- c(
        idx_period[seq_len(period_start_pos)],
        idx_period[faceoff_pos],
        idx_period[intervening_pos],
        if (faceoff_pos < length(idx_period)) {
          idx_period[(faceoff_pos + 1L):length(idx_period)]
        } else {
          integer(0)
        }
      )
      idx_sorted[period[idx_sorted] == p] <- idx_period
    }

    sort_order[idx_sorted] <- sort(sort_order[idx_game], na.last = TRUE)
  }

  play_by_play$sortOrder <- sort_order
  play_by_play[order(sort_order, row_id, na.last = TRUE), , drop = FALSE]
}

#' Ensure a native symbol is available
#'
#' `.ensure_local_native_symbol()` checks whether a native symbol is already
#' registered and, if needed, tries to load the installed package shared object.
#' It does not trust the current working directory.
#'
#' @param symbol native symbol name
#' @returns logical scalar indicating whether the symbol is available
#' @keywords internal
.ensure_local_native_symbol <- function(symbol) {
  if (is.loaded(symbol)) {
    return(TRUE)
  }
  dlls <- getLoadedDLLs()
  if ('nhlscraper' %in% names(dlls)) {
    return(is.loaded(symbol, PACKAGE = 'nhlscraper'))
  }
  pkg_path <- tryCatch(
    getNamespaceInfo(asNamespace('nhlscraper'), 'path'),
    error = function(e) ''
  )
  if (!nzchar(pkg_path)) {
    return(FALSE)
  }
  dll_path <- file.path(
    pkg_path,
    'libs',
    paste0('nhlscraper', .Platform$dynlib.ext)
  )
  if (!file.exists(dll_path)) {
    return(FALSE)
  }
  tryCatch(
    {
      dyn.load(dll_path)
      is.loaded(symbol, PACKAGE = 'nhlscraper')
    },
    error = function(e) FALSE
  )
}

#' Validate required public play-by-play columns
#'
#' `.require_public_pbp_columns()` errors when a public helper is called with a
#' play-by-play missing required public schema columns.
#'
#' @param play_by_play data.frame play-by-play object
#' @param cols character vector of required columns
#' @param fn_name calling function name used in the error message
#' @returns `NULL`, invisibly, or an error if required columns are missing
#' @keywords internal
.require_public_pbp_columns <- function(play_by_play, cols, fn_name) {
  missing_cols <- setdiff(cols, names(play_by_play))
  if (length(missing_cols)) {
    stop(
      sprintf(
        "%s() requires public play-by-play column(s): %s",
        fn_name,
        paste(missing_cols, collapse = ", ")
      ),
      call. = FALSE
    )
  }
}

#' Validate required shift-chart columns
#'
#' `.require_shift_chart_columns()` errors when a helper is called with a shift
#' chart missing required public schema columns.
#'
#' @param shift_chart data.frame shift chart object
#' @param cols character vector of required columns
#' @param fn_name calling function name used in the error message
#' @returns `NULL`, invisibly, or an error if required columns are missing
#' @keywords internal
.require_shift_chart_columns <- function(shift_chart, cols, fn_name) {
  missing_cols <- setdiff(cols, names(shift_chart))
  if (length(missing_cols)) {
    stop(
      sprintf(
        "%s() requires shift chart column(s): %s",
        fn_name,
        paste(missing_cols, collapse = ", ")
      ),
      call. = FALSE
    )
  }
}

#' Build a shot-event mask
#'
#' `.shot_event_mask()` flags play-by-play rows matching selected event types.
#'
#' @param play_by_play data.frame play-by-play object
#' @param types character vector of event type keys to keep
#' @returns logical vector
#' @keywords internal
.shot_event_mask <- function(play_by_play, types) {
  n <- nrow(play_by_play)
  if (!n) return(logical(0))
  eventTypeDescKey <- if ('eventTypeDescKey' %in% names(play_by_play)) {
    as.character(play_by_play$eventTypeDescKey)
  } else {
    rep(NA_character_, n)
  }
  !is.na(eventTypeDescKey) & eventTypeDescKey %in% types
}

#' Compute shot context summaries
#'
#' `.compute_shot_context()` derives rush and rebound flags plus running goal,
#' shot, Fenwick, and Corsi counts, using native code when available.
#'
#' @param play_by_play data.frame play-by-play object
#' @returns named list of logical and integer shot-context vectors
#' @keywords internal
.compute_shot_context <- function(play_by_play) {
  n <- nrow(play_by_play)
  if (!n) {
    return(list(
      isRush = logical(0),
      isRebound = logical(0),
      createdRebound = logical(0),
      homeGoals = integer(0),
      awayGoals = integer(0),
      homeShots = integer(0),
      awayShots = integer(0),
      homeFenwick = integer(0),
      awayFenwick = integer(0),
      homeCorsi = integer(0),
      awayCorsi = integer(0)
    ))
  }

  type_desc_key <- if ('eventTypeDescKey' %in% names(play_by_play)) {
    as.character(play_by_play$eventTypeDescKey)
  } else {
    rep(NA_character_, n)
  }
  zone_code <- if ('zoneCode' %in% names(play_by_play)) {
    as.character(play_by_play$zoneCode)
  } else {
    rep(NA_character_, n)
  }
  game_id <- if ('gameId' %in% names(play_by_play)) {
    as.integer(play_by_play$gameId)
  } else {
    seq_len(n)
  }
  sort_order <- if ('sortOrder' %in% names(play_by_play)) {
    as.integer(play_by_play$sortOrder)
  } else {
    seq_len(n)
  }
  seconds_elapsed <- if ('secondsElapsedInGame' %in% names(play_by_play)) {
    as.integer(play_by_play$secondsElapsedInGame)
  } else {
    rep(NA_integer_, n)
  }
  is_home <- if ('isHome' %in% names(play_by_play)) {
    ifelse(
      play_by_play$isHome %in% TRUE,
      1L,
      ifelse(play_by_play$isHome %in% FALSE, 0L, NA_integer_)
    )
  } else {
    rep(NA_integer_, n)
  }
  valid_type <- !is.na(type_desc_key)
  is_attempt <- valid_type & type_desc_key %in% c(
    'goal', 'shot-on-goal', 'missed-shot', 'blocked-shot'
  )
  is_source <- valid_type & type_desc_key %in% c(
    'shot-on-goal', 'missed-shot', 'blocked-shot'
  )
  is_goal <- valid_type & type_desc_key == 'goal'
  is_sog <- valid_type & type_desc_key %in% c('goal', 'shot-on-goal')
  is_fenwick <- valid_type & type_desc_key %in% c(
    'goal', 'shot-on-goal', 'missed-shot'
  )
  is_corsi <- valid_type & type_desc_key %in% c(
    'goal', 'shot-on-goal', 'missed-shot', 'blocked-shot'
  )
  is_stop <- valid_type & type_desc_key %in% c(
    'stoppage', 'faceoff', 'period-start', 'period-end', 'game-end'
  )
  is_nz_dz <- !is.na(zone_code) & zone_code %in% c('N', 'D')
  situation_code <- if ('situationCode' %in% names(play_by_play)) {
    as.character(play_by_play$situationCode)
  } else {
    rep(NA_character_, n)
  }
  is_ps_so <- !is.na(situation_code) & situation_code %in% c('0101', '1010')
  order_time <- order(game_id, seconds_elapsed, sort_order, seq_len(n), na.last = TRUE)
  order_sort <- order(game_id, sort_order, seq_len(n), na.last = TRUE)

  compute_in_r <- function() {
    event_owner_team_id <- if ('eventOwnerTeamId' %in% names(play_by_play)) {
      as.integer(play_by_play$eventOwnerTeamId)
    } else {
      rep(NA_integer_, n)
    }
    is_rush <- rep(NA, n)
    is_rush[is_attempt] <- FALSE
    is_rebound <- rep(NA, n)
    is_rebound[is_attempt] <- FALSE
    created_rebound <- rep(NA, n)
    created_rebound[is_attempt] <- FALSE

    for (g in unique(game_id)) {
      idx_time <- order_time[game_id[order_time] == g]
      last_nz_dz_time <- NA_integer_
      last_time <- integer(0)
      last_idx <- integer(0)
      names(last_time) <- character(0)
      names(last_idx) <- character(0)
      for (i in idx_time) {
        if (is_stop[i]) {
          last_nz_dz_time <- NA_integer_
          last_time <- integer(0)
          last_idx <- integer(0)
          names(last_time) <- character(0)
          names(last_idx) <- character(0)
          next
        }
        t <- seconds_elapsed[i]
        if (!is.na(t)) {
          if (is_attempt[i] && !is.na(last_nz_dz_time) && t - last_nz_dz_time >= 0L &&
              t - last_nz_dz_time <= 4L) {
            is_rush[i] <- TRUE
          }
          tid <- as.character(event_owner_team_id[i])
          if (is_attempt[i] && length(last_time) && tid %in% names(last_time)) {
            dt <- t - last_time[[tid]]
            if (!is.na(dt) && dt >= 0L && dt <= 3L) {
              is_rebound[i] <- TRUE
              created_rebound[last_idx[[tid]]] <- TRUE
            }
          }
          if (is_nz_dz[i]) {
            last_nz_dz_time <- t
          }
          if (is_source[i] && !is.na(event_owner_team_id[i])) {
            last_time[tid] <- t
            last_idx[tid] <- i
          }
        }
      }
    }

    is_rush[is_attempt & is_ps_so] <- FALSE
    is_rebound[is_attempt & is_ps_so] <- FALSE
    created_rebound[is_attempt & is_ps_so] <- FALSE

    home_goals <- integer(n)
    away_goals <- integer(n)
    home_sog <- integer(n)
    away_sog <- integer(n)
    home_fenwick <- integer(n)
    away_fenwick <- integer(n)
    home_corsi <- integer(n)
    away_corsi <- integer(n)

    for (g in unique(game_id)) {
      idx_sort <- order_sort[game_id[order_sort] == g]
      hG <- 0L
      aG <- 0L
      hS <- 0L
      aS <- 0L
      hF <- 0L
      aF <- 0L
      hC <- 0L
      aC <- 0L
      for (i in idx_sort) {
        home_goals[i] <- hG
        away_goals[i] <- aG
        home_sog[i] <- hS
        away_sog[i] <- aS
        home_fenwick[i] <- hF
        away_fenwick[i] <- aF
        home_corsi[i] <- hC
        away_corsi[i] <- aC
        if (!is.na(is_home[i]) && is_home[i] == 1L) {
          hG <- hG + as.integer(is_goal[i])
          hS <- hS + as.integer(is_sog[i])
          hF <- hF + as.integer(is_fenwick[i])
          hC <- hC + as.integer(is_corsi[i])
        } else if (!is.na(is_home[i]) && is_home[i] == 0L) {
          aG <- aG + as.integer(is_goal[i])
          aS <- aS + as.integer(is_sog[i])
          aF <- aF + as.integer(is_fenwick[i])
          aC <- aC + as.integer(is_corsi[i])
        }
      }
    }

    list(
      isRush = is_rush,
      isRebound = is_rebound,
      createdRebound = created_rebound,
      homeGoals = home_goals,
      awayGoals = away_goals,
      homeShots = home_sog,
      awayShots = away_sog,
      homeFenwick = home_fenwick,
      awayFenwick = away_fenwick,
      homeCorsi = home_corsi,
      awayCorsi = away_corsi
    )
  }

  if (!.ensure_local_native_symbol('nhlscraper_pbp_shot_context')) {
    return(compute_in_r())
  }

  tryCatch(
    .Call('nhlscraper_pbp_shot_context', list(
      as.integer(order_time),
      as.integer(order_sort),
      as.integer(game_id),
      as.integer(seconds_elapsed),
      if ('eventOwnerTeamId' %in% names(play_by_play)) as.integer(play_by_play$eventOwnerTeamId) else rep(NA_integer_, n),
      as.integer(is_home),
      as.logical(is_attempt),
      as.logical(is_source),
      as.logical(is_goal),
      as.logical(is_sog),
      as.logical(is_fenwick),
      as.logical(is_corsi),
      as.logical(is_stop),
      as.logical(is_nz_dz),
      as.logical(is_ps_so)
    )),
    error = function(e) compute_in_r()
  )
}

#' Attach shot context columns to play-by-play
#'
#' `.apply_shot_context()` appends rush, rebound, and running score/attempt
#' context columns to a play-by-play.
#'
#' @param play_by_play data.frame play-by-play object
#' @returns data.frame with shot context columns added
#' @keywords internal
.apply_shot_context <- function(play_by_play) {
  ctx <- .compute_shot_context(play_by_play)
  if (is.null(ctx$homeShots) && !is.null(ctx$homeSOG)) {
    ctx$homeShots <- ctx$homeSOG
  }
  if (is.null(ctx$awayShots) && !is.null(ctx$awaySOG)) {
    ctx$awayShots <- ctx$awaySOG
  }
  is_home <- if ('isHome' %in% names(play_by_play)) {
    ifelse(
      play_by_play$isHome %in% TRUE,
      1L,
      ifelse(play_by_play$isHome %in% FALSE, 0L, NA_integer_)
    )
  } else {
    rep(NA_integer_, nrow(play_by_play))
  }

  play_by_play$isRush <- ctx$isRush
  play_by_play$isRebound <- ctx$isRebound
  play_by_play$createdRebound <- ctx$createdRebound
  play_by_play$homeGoals <- ctx$homeGoals
  play_by_play$awayGoals <- ctx$awayGoals
  play_by_play$homeShots <- ctx$homeShots
  play_by_play$awayShots <- ctx$awayShots
  play_by_play$homeFenwick <- ctx$homeFenwick
  play_by_play$awayFenwick <- ctx$awayFenwick
  play_by_play$homeCorsi <- ctx$homeCorsi
  play_by_play$awayCorsi <- ctx$awayCorsi
  play_by_play$goalsFor <- ifelse(
    is_home == 1L,
    play_by_play$homeGoals,
    ifelse(is_home == 0L, play_by_play$awayGoals, NA_integer_)
  )
  play_by_play$goalsAgainst <- ifelse(
    is_home == 1L,
    play_by_play$awayGoals,
    ifelse(is_home == 0L, play_by_play$homeGoals, NA_integer_)
  )
  play_by_play$shotsFor <- ifelse(
    is_home == 1L,
    play_by_play$homeShots,
    ifelse(is_home == 0L, play_by_play$awayShots, NA_integer_)
  )
  play_by_play$shotsAgainst <- ifelse(
    is_home == 1L,
    play_by_play$awayShots,
    ifelse(is_home == 0L, play_by_play$homeShots, NA_integer_)
  )
  play_by_play$fenwickFor <- ifelse(
    is_home == 1L,
    play_by_play$homeFenwick,
    ifelse(is_home == 0L, play_by_play$awayFenwick, NA_integer_)
  )
  play_by_play$fenwickAgainst <- ifelse(
    is_home == 1L,
    play_by_play$awayFenwick,
    ifelse(is_home == 0L, play_by_play$homeFenwick, NA_integer_)
  )
  play_by_play$corsiFor <- ifelse(
    is_home == 1L,
    play_by_play$homeCorsi,
    ifelse(is_home == 0L, play_by_play$awayCorsi, NA_integer_)
  )
  play_by_play$corsiAgainst <- ifelse(
    is_home == 1L,
    play_by_play$awayCorsi,
    ifelse(is_home == 0L, play_by_play$homeCorsi, NA_integer_)
  )
  play_by_play$goalDifferential <- play_by_play$goalsFor - play_by_play$goalsAgainst
  play_by_play$shotDifferential <- play_by_play$shotsFor - play_by_play$shotsAgainst
  play_by_play$fenwickDifferential <- play_by_play$fenwickFor - play_by_play$fenwickAgainst
  play_by_play$corsiDifferential <- play_by_play$corsiFor - play_by_play$corsiAgainst
  play_by_play
}

#' Normalize situation codes for parsing
#'
#' `.normalize_situation_code_for_parse()` pads 1-4 digit situation codes to
#' four characters for internal parsing without rewriting the source column.
#'
#' @param situation_code vector of raw situation codes
#' @returns character vector of parse-ready situation codes
#' @keywords internal
.normalize_situation_code_for_parse <- function(situation_code) {
  sc_chr <- as.character(situation_code)
  ok <- !is.na(sc_chr) & grepl('^[0-9]{1,4}$', sc_chr)
  out <- rep(NA_character_, length(sc_chr))
  if (any(ok)) {
    out[ok] <- sprintf('%04s', sc_chr[ok])
    out[ok] <- chartr(' ', '0', out[ok])
  }
  out
}

#' Parse situation-code components
#'
#' `.parse_situation_code_components()` converts situation codes into away/home
#' goalie and skater counts without mutating the original source vector.
#'
#' @param situation_code vector of raw situation codes
#' @returns integer matrix with away/home goalie and skater counts
#' @keywords internal
.parse_situation_code_components <- function(situation_code) {
  sc_chr <- .normalize_situation_code_for_parse(situation_code)
  cbind(
    awayGoalieCount = suppressWarnings(as.integer(substr(sc_chr, 1L, 1L))),
    awaySkaterCount = suppressWarnings(as.integer(substr(sc_chr, 2L, 2L))),
    homeSkaterCount = suppressWarnings(as.integer(substr(sc_chr, 3L, 3L))),
    homeGoalieCount = suppressWarnings(as.integer(substr(sc_chr, 4L, 4L)))
  )
}

#' Strip the situation code into goalie and skater counts, man differential, and strength state for all the events (plays) in a play-by-play by perspective
#' 
#' `.strip_situation_code()` strips the situation code into goalie and skater counts for home and away teams, then (from the event owner's perspective) computes man differential and a strength state classification without rewriting the source `situationCode` column.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added columns: `homeIsEmptyNet`, `awayIsEmptyNet`, `homeSkaterCount`, `awaySkaterCount`, `isEmptyNetFor`, `isEmptyNetAgainst`, `skaterCountFor`, `skaterCountAgainst`, `manDifferential`, and `strengthState`
#' @keywords internal
.strip_situation_code <- function(play_by_play) {
  sc_raw <- if ('situationCode' %in% names(play_by_play)) {
    play_by_play$situationCode
  } else {
    rep(NA_character_, nrow(play_by_play))
  }
  sc_parts <- .parse_situation_code_components(sc_raw)
  aG <- sc_parts[, 'awayGoalieCount']
  aS <- sc_parts[, 'awaySkaterCount']
  hS <- sc_parts[, 'homeSkaterCount']
  hG <- sc_parts[, 'homeGoalieCount']
  play_by_play$homeIsEmptyNet  <- hG == 0L
  play_by_play$awayIsEmptyNet  <- aG == 0L
  play_by_play$homeSkaterCount <- hS
  play_by_play$awaySkaterCount <- aS
  play_by_play$isEmptyNetFor     <- ifelse(
    play_by_play$isHome,
    hG == 0L,
    ifelse(!is.na(play_by_play$isHome), aG == 0L, NA)
  )
  play_by_play$isEmptyNetAgainst <- ifelse(
    play_by_play$isHome,
    aG == 0L,
    ifelse(!is.na(play_by_play$isHome), hG == 0L, NA)
  )
  play_by_play$skaterCountFor     <- ifelse(
    play_by_play$isHome,
    hS,
    ifelse(!is.na(play_by_play$isHome), aS, NA)
  )
  play_by_play$skaterCountAgainst <- ifelse(
    play_by_play$isHome,
    aS,
    ifelse(!is.na(play_by_play$isHome), hS, NA)
  )
  forMen                       <- play_by_play$skaterCountFor + ifelse(play_by_play$isHome, hG, aG)
  agMen                        <- play_by_play$skaterCountAgainst + ifelse(play_by_play$isHome, aG, hG)
  play_by_play$manDifferential <- ifelse(is.na(play_by_play$isHome), NA, forMen - agMen)
  play_by_play$strengthState   <- ifelse(
    is.na(play_by_play$manDifferential),
    NA,
    ifelse(
      play_by_play$manDifferential > 0L,
      'power-play',
      ifelse(play_by_play$manDifferential < 0L, 'penalty-kill', 'even-strength')
    )
  )
  after  <- match('situationCode', names(play_by_play))
  insert <- c(
    'homeIsEmptyNet',
    'awayIsEmptyNet',
    'homeSkaterCount',
    'awaySkaterCount',
    'isEmptyNetFor',
    'isEmptyNetAgainst',
    'skaterCountFor',
    'skaterCountAgainst',
    'manDifferential',
    'strengthState'
  )
  nms <- names(play_by_play)
  play_by_play[, c(nms[seq_len(after)], insert, setdiff(nms[-seq_len(after)], insert))]
}

#' Flag if the event belongs to the home team or not for all the events (plays) in a play-by-play
#' 
#' `.flag_is_home()` flags if the event belongs to the home team or not for all the events (plays) in a play-by-play.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added `isHome` column
#' @keywords internal

.flag_is_home <- function(play_by_play) {
  pbp <- play_by_play[order(play_by_play$sortOrder), ]
  n   <- nrow(pbp)
  fill_down <- function(x) {
    if (is.null(x)) return(rep(NA_real_, n))
    x[1L] <- ifelse(is.na(x[1L]), 0, x[1L])
    if (length(x) > 1L) for (i in 2L:length(x)) if (is.na(x[i])) x[i] <- x[i - 1L]
    x
  }
  hs  <- fill_down(pbp$homeScore)
  as  <- fill_down(pbp$awayScore)
  ho  <- fill_down(pbp$homeShots)
  ao  <- fill_down(pbp$awayShots)
  dhs <- c(0, diff(hs))
  das <- c(0, diff(as))
  dho <- c(0, diff(ho))
  dao <- c(0, diff(ao))
  inc_home  <- (!is.na(dhs) & dhs > 0) | (!is.na(dho) & dho > 0)
  inc_away  <- (!is.na(das) & das > 0) | (!is.na(dao) & dao > 0)
  homeVotes <- pbp$eventOwnerTeamId[inc_home & !is.na(pbp$eventOwnerTeamId)]
  awayVotes <- pbp$eventOwnerTeamId[inc_away & !is.na(pbp$eventOwnerTeamId)]
  homeId    <- as.integer(names(which.max(table(homeVotes))))
  awayId    <- as.integer(names(which.max(table(awayVotes))))
  play_by_play$isHome <- ifelse(
    play_by_play$eventOwnerTeamId == homeId,
    TRUE,
    ifelse(play_by_play$eventOwnerTeamId == awayId, FALSE, NA)
  )
  insert <- c('eventOwnerTeamId', 'isHome', 'eventTypeCode', 'eventTypeDescKey')
  keep   <- setdiff(names(play_by_play), insert)
  after  <- match('secondsElapsedInGame', keep)
  play_by_play[, c(keep[seq_len(after)], insert, keep[(after + 1L):length(keep)])]
}

#' Flag if the shot attempt is a rush attempt or not for all the shots in a play-by-play
#' 
#' `.flag_is_rush()` flags whether a shot attempt is a rush attempt, defined as any shot attempt occurring within 4 seconds of a prior event in the neutral or defensive zone with no stoppage in play in between.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added `isRush` column
#' @keywords internal

.flag_is_rush <- function(play_by_play) {
  play_by_play$isRush <- .compute_shot_context(play_by_play)$isRush
  after               <- match('angle', names(play_by_play))
  insert              <- c('shotType', 'isRush')
  nms                 <- names(play_by_play)
  play_by_play[, c(nms[seq_len(after)], insert, setdiff(nms[-seq_len(after)], insert))]
}

#' Flag if the shot attempt is a rebound attempt or creates a rebound for all the shots in a play-by-play
#' 
#' `.flag_is_rebound()` flags whether a shot attempt is a rebound attempt (i.e., taken within 3 seconds of a prior blocked, missed, or saved attempt with no stoppage in between), and whether a shot attempt creates a rebound under the same definition.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added columns: `createdRebound` and `isRebound`
#' @keywords internal

.flag_is_rebound <- function(play_by_play) {
  ctx <- .compute_shot_context(play_by_play)
  play_by_play$createdRebound <- ctx$createdRebound
  play_by_play$isRebound      <- ctx$isRebound
  after  <- match('isRush', names(play_by_play))
  insert <- c('isRebound', 'createdRebound')
  nms    <- names(play_by_play)
  play_by_play[, c(nms[seq_len(after)], insert, setdiff(nms[-seq_len(after)], insert))]
}

#' Normalize the x and y coordinates for all the events (plays) in a play-by-play
#' 
#' `.normalize_coordinates()` normalizes the x and y coordinates for all the events (plays) in a play-by-play such that they all attack towards +x. If `homeTeamDefendingSide` is not available, the home defending side in period 1 is inferred using `zoneCode`, `isHome`, and `xCoord`.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added columns `xCoordNorm` and `yCoordNorm`
#' @keywords internal

.normalize_coordinates <- function(play_by_play) {
  n                       <- nrow(play_by_play)
  play_by_play$xCoordNorm <- NA_real_
  play_by_play$yCoordNorm <- NA_real_
  has_side                <- 'homeTeamDefendingSide' %in% names(play_by_play)
  for (g in unique(play_by_play$gameId)) {
    idx <- which(play_by_play$gameId == g)
    if (has_side) {
      side          <- play_by_play$homeTeamDefendingSide[idx][!is.na(play_by_play$homeTeamDefendingSide[idx])][1L]
      home_def_left <- side == 'left'
    } else {
      d_idx         <- idx[
        play_by_play$isHome[idx] & 
          play_by_play$zoneCode[idx] == 'D' &
          play_by_play$periodNumber[idx] == 1L & 
          !is.na(play_by_play$xCoord[idx])
      ]
      o_idx         <- idx[
        play_by_play$isHome[idx] & 
          play_by_play$zoneCode[idx] == 'O' &
          play_by_play$periodNumber[idx] == 1L & 
          !is.na(play_by_play$xCoord[idx])
      ]
      home_def_left <- if (length(d_idx)) {
        mean(play_by_play$xCoord[d_idx] < 0) >= 0.5
      } else {
        mean(play_by_play$xCoord[o_idx] > 0) >= 0.5
      }
      play_by_play$homeTeamDefendingSide      <- NA_character_
      play_by_play$homeTeamDefendingSide[idx] <- ifelse(home_def_left, 'left', 'right')
    }
    home_att_pos_p1 <- home_def_left
    home_att_pos    <- ifelse(play_by_play$periodNumber[idx] %% 2L == 1L, home_att_pos_p1, !home_att_pos_p1)
    att_pos         <- ifelse(play_by_play$isHome[idx], home_att_pos, !home_att_pos)
    flip            <- !att_pos
    play_by_play$xCoordNorm[idx] <- ifelse(flip, -play_by_play$xCoord[idx], play_by_play$xCoord[idx])
    play_by_play$yCoordNorm[idx] <- ifelse(flip, -play_by_play$yCoord[idx], play_by_play$yCoord[idx])
  }
  insert <- c('homeTeamDefendingSide', 'zoneCode', 'xCoord', 'yCoord', 'xCoordNorm', 'yCoordNorm')
  keep   <- setdiff(names(play_by_play), insert)
  after  <- match('strengthState', keep)
  play_by_play[, c(keep[seq_len(after)], insert, keep[(after + 1L):length(keep)])]
}

#' Calculate the Euclidean distance from the attacking net for all the events (plays) in a play-by-play
#' 
#' `.calculate_distance()` calculates the Euclidean distance from the attacking net for all the events (plays) in a play-by-play.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added `distance` column
#' @keywords internal

.calculate_distance <- function(play_by_play) {
  net_x                 <- 89
  play_by_play$distance <- sqrt((net_x - play_by_play$xCoordNorm)^2 + (play_by_play$yCoordNorm)^2)
  after                 <- match('yCoordNorm', names(play_by_play))
  insert                <- 'distance'
  nms                   <- names(play_by_play)
  play_by_play[, c(nms[seq_len(after)], insert, setdiff(nms[-seq_len(after)], insert))]
}

#' Calculate the Euclidean angle from the attacking net for all the events (plays) in a play-by-play
#' 
#' `.calculate_angle()` calculates the Euclidean angle from the attacking net for all the events (plays) in a play-by-play.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added `angle` column
#' @keywords internal

.calculate_angle <- function(play_by_play) {
  net_x            <- 89
  dx               <- net_x - play_by_play$xCoordNorm
  play_by_play$angle <- atan2(abs(play_by_play$yCoordNorm), dx) * 180 / pi
  after            <- match('distance', names(play_by_play))
  insert           <- 'angle'
  nms              <- names(play_by_play)
  play_by_play[, c(nms[seq_len(after)], insert, setdiff(nms[-seq_len(after)], insert))]
}

#' Count the as-of-event goal, shots on goal, Fenwick, and Corsi attempts and differentials for all the events (plays) in a play-by-play by perspective
#' 
#' `.count_goals_shots()` counts the as-of-event goal, shots on goal, Fenwick, and Corsi attempts and differentials for all the events (plays) in a play-by-play by perspective.
#' 
#' @inheritParams .strip_game_id
#' @returns data.frame with one row per event (play) and added columns: `homeGoals`, `awayGoals`, `homeShots`, `awayShots`, `homeFenwick`, `awayFenwick`, `homeCorsi`, `awayCorsi`, `goalsFor`, `goalsAgainst`, `shotsFor`, `shotsAgainst`, `fenwickFor`, `fenwickAgainst`, `corsiFor`, `corsiAgainst`, `goalDifferential`, `shotDifferential`, `fenwickDifferential`, and `corsiDifferential`
#' @keywords internal

.count_goals_shots <- function(play_by_play) {
  play_by_play <- .apply_shot_context(play_by_play)
  insert <- c(
    'homeGoals', 'awayGoals', 'homeShots', 'awayShots', 'homeFenwick', 'awayFenwick', 'homeCorsi', 'awayCorsi',
    'goalsFor', 'goalsAgainst', 'shotsFor', 'shotsAgainst', 'fenwickFor', 'fenwickAgainst', 'corsiFor', 'corsiAgainst',
    'goalDifferential', 'shotDifferential', 'fenwickDifferential', 'corsiDifferential'
  )
  keep   <- setdiff(names(play_by_play), insert)
  after  <- match('createdRebound', keep)
  play_by_play[, c(keep[seq_len(after)], insert, keep[(after + 1L):length(keep)])]
}

#' Extract an integer on-ice column
#'
#' `.on_ice_int_col()` safely extracts an integer column from a data frame and
#' returns `NA` values when the column is absent.
#'
#' @param df data.frame source object
#' @param name column name to extract
#' @returns integer vector
#' @keywords internal
.on_ice_int_col <- function(df, name) {
  if (name %in% names(df)) as.integer(df[[name]]) else rep(NA_integer_, nrow(df))
}

#' Build scalar on-ice timing column names
#'
#' `.on_ice_timing_scalar_column_names()` returns the scalar goalie and skater
#' timing column names for a timing metric suffix.
#'
#' @param metric_suffix scalar timing suffix
#' @param play_by_play optional data.frame whose existing on-ice columns should
#'   be inspected
#' @param slot_count optional integer scalar minimum slot count requested by the
#'   caller
#' @returns character vector of column names
#' @keywords internal
.on_ice_timing_scalar_column_names <- function(
  metric_suffix,
  play_by_play = NULL,
  slot_count = NULL
) {
  skater_slots <- seq_len(.on_ice_skater_slots(
    play_by_play = play_by_play,
    slot_count = slot_count
  ))
  c(
    paste0('homeGoalie', metric_suffix),
    paste0('awayGoalie', metric_suffix),
    paste0('goalie', metric_suffix, 'For'),
    paste0('goalie', metric_suffix, 'Against'),
    as.vector(rbind(
      paste0('homeSkater', skater_slots, metric_suffix),
      paste0('awaySkater', skater_slots, metric_suffix),
      paste0('skater', skater_slots, metric_suffix, 'For'),
      paste0('skater', skater_slots, metric_suffix, 'Against')
    ))
  )
}

#' Sort shift data for timing lookups
#'
#' `.sort_shift_chart_for_timing()` filters incomplete shift rows and sorts the
#' remainder into the order expected by the timing resolvers.
#'
#' @param shift_data data.frame shift chart data
#' @returns filtered and sorted shift data.frame
#' @keywords internal
.sort_shift_chart_for_timing <- function(shift_data) {
  shift_data <- shift_data[
    !is.na(shift_data$gameId) &
      !is.na(shift_data$period) &
      !is.na(shift_data$playerId) &
      !is.na(shift_data$startSecondsElapsedInPeriod) &
      !is.na(shift_data$endSecondsElapsedInPeriod),
    ,
    drop = FALSE
  ]
  if (!nrow(shift_data)) {
    return(shift_data)
  }
  ord <- order(
    as.integer(shift_data$gameId),
    as.integer(shift_data$period),
    as.integer(shift_data$playerId),
    as.integer(shift_data$startSecondsElapsedInPeriod),
    as.integer(shift_data$endSecondsElapsedInPeriod)
  )
  shift_data[ord, , drop = FALSE]
}

#' Compute on-ice shift timings in R
#'
#' `.compute_on_ice_shift_timing_in_r()` is the pure-R fallback for resolving
#' on-ice remaining-shift, elapsed-shift, and time-since-last-shift matrices.
#'
#' @param play_by_play data.frame play-by-play object with on-ice player IDs
#' @param shift_data data.frame shift chart data
#' @returns list containing home and away timing matrices
#' @keywords internal
.compute_on_ice_shift_timing_in_r <- function(play_by_play, shift_data) {
  n <- nrow(play_by_play)
  slot_count <- .on_ice_skater_slots(play_by_play = play_by_play)
  home_request <- cbind(
    as.integer(play_by_play$homeGoaliePlayerId),
    as.matrix(play_by_play[, paste0('homeSkater', seq_len(slot_count), 'PlayerId'), drop = FALSE])
  )
  away_request <- cbind(
    as.integer(play_by_play$awayGoaliePlayerId),
    as.matrix(play_by_play[, paste0('awaySkater', seq_len(slot_count), 'PlayerId'), drop = FALSE])
  )
  storage.mode(home_request) <- 'integer'
  storage.mode(away_request) <- 'integer'

  home_elapsed <- matrix(NA_real_, nrow = n, ncol = slot_count + 1L)
  away_elapsed <- matrix(NA_real_, nrow = n, ncol = slot_count + 1L)
  home_remaining <- matrix(NA_real_, nrow = n, ncol = slot_count + 1L)
  away_remaining <- matrix(NA_real_, nrow = n, ncol = slot_count + 1L)
  home_since <- matrix(NA_real_, nrow = n, ncol = slot_count + 1L)
  away_since <- matrix(NA_real_, nrow = n, ncol = slot_count + 1L)
  if (!nrow(shift_data) || !n) {
    return(list(
      homeRemaining = home_remaining,
      awayRemaining = away_remaining,
      homeElapsed = home_elapsed,
      awayElapsed = away_elapsed,
      homeSinceLast = home_since,
      awaySinceLast = away_since
    ))
  }

  prev_end <- rep(NA_integer_, nrow(shift_data))
  keys <- paste(shift_data$gameId, shift_data$period, shift_data$playerId, sep = ':')
  if (nrow(shift_data) > 1L) {
    for (i in 2:nrow(shift_data)) {
      if (keys[i] == keys[i - 1L]) {
        prev_end[i] <- as.integer(shift_data$endSecondsElapsedInPeriod[i - 1L])
      }
    }
  }
  split_idx <- split(seq_len(nrow(shift_data)), keys)

  lookup_player <- function(game_id, period, seconds_elapsed, player_id) {
    if (
      is.na(game_id) ||
      is.na(period) ||
      is.na(seconds_elapsed) ||
      is.na(player_id)
    ) {
      return(c(NA_real_, NA_real_, NA_real_))
    }
    idx <- split_idx[[paste(game_id, period, player_id, sep = ':')]]
    if (is.null(idx) || !length(idx)) {
      return(c(NA_real_, NA_real_, NA_real_))
    }
    starts <- as.integer(shift_data$startSecondsElapsedInPeriod[idx])
    ends <- as.integer(shift_data$endSecondsElapsedInPeriod[idx])
    pos <- findInterval(as.integer(seconds_elapsed), starts)
    if (pos < 1L || pos > length(idx)) {
      return(c(NA_real_, NA_real_, NA_real_))
    }
    row_idx <- idx[pos]
    if (seconds_elapsed > ends[pos]) {
      return(c(NA_real_, NA_real_, NA_real_))
    }
    remaining <- as.numeric(ends[pos] - seconds_elapsed)
    elapsed <- as.numeric(seconds_elapsed - starts[pos])
    since <- if (is.na(prev_end[row_idx])) {
      as.numeric(300L + seconds_elapsed)
    } else {
      as.numeric(seconds_elapsed - prev_end[row_idx])
    }
    c(remaining, elapsed, since)
  }

  for (i in seq_len(n)) {
    for (j in seq_len(slot_count + 1L)) {
      home_vals <- lookup_player(
        play_by_play$gameId[i],
        play_by_play$periodNumber[i],
        play_by_play$secondsElapsedInPeriod[i],
        home_request[i, j]
      )
      away_vals <- lookup_player(
        play_by_play$gameId[i],
        play_by_play$periodNumber[i],
        play_by_play$secondsElapsedInPeriod[i],
        away_request[i, j]
      )
      home_remaining[i, j] <- home_vals[1L]
      home_elapsed[i, j] <- home_vals[2L]
      home_since[i, j] <- home_vals[3L]
      away_remaining[i, j] <- away_vals[1L]
      away_elapsed[i, j] <- away_vals[2L]
      away_since[i, j] <- away_vals[3L]
    }
  }

  list(
    homeRemaining = home_remaining,
    awayRemaining = away_remaining,
    homeElapsed = home_elapsed,
    awayElapsed = away_elapsed,
    homeSinceLast = home_since,
    awaySinceLast = away_since
  )
}

#' Compute on-ice shift timing matrices
#'
#' `.compute_on_ice_shift_timing_matrices()` uses the native timing resolver
#' when available and falls back to the R implementation otherwise.
#'
#' @param play_by_play data.frame play-by-play object with on-ice player IDs
#' @param shift_data data.frame shift chart data
#' @returns list containing home and away timing matrices
#' @keywords internal
.compute_on_ice_shift_timing_matrices <- function(play_by_play, shift_data) {
  shift_data <- .sort_shift_chart_for_timing(shift_data)
  if (
    !.ensure_local_native_symbol('nhlscraper_on_ice_shift_timings') ||
      !nrow(play_by_play)
  ) {
    return(.compute_on_ice_shift_timing_in_r(play_by_play, shift_data))
  }

  slot_count <- .on_ice_skater_slots(play_by_play = play_by_play)
  home_request <- cbind(
    as.integer(play_by_play$homeGoaliePlayerId),
    as.matrix(play_by_play[, paste0('homeSkater', seq_len(slot_count), 'PlayerId'), drop = FALSE])
  )
  away_request <- cbind(
    as.integer(play_by_play$awayGoaliePlayerId),
    as.matrix(play_by_play[, paste0('awaySkater', seq_len(slot_count), 'PlayerId'), drop = FALSE])
  )
  storage.mode(home_request) <- 'integer'
  storage.mode(away_request) <- 'integer'

  tryCatch(
    .Call('nhlscraper_on_ice_shift_timings', list(
      as.integer(play_by_play$gameId),
      as.integer(play_by_play$periodNumber),
      as.integer(play_by_play$secondsElapsedInPeriod),
      home_request,
      away_request,
      as.integer(shift_data$gameId),
      as.integer(shift_data$period),
      as.integer(shift_data$playerId),
      as.integer(shift_data$startSecondsElapsedInPeriod),
      as.integer(shift_data$endSecondsElapsedInPeriod)
    )),
    error = function(e) .compute_on_ice_shift_timing_in_r(play_by_play, shift_data)
  )
}

#' Assign scalar on-ice timing metrics
#'
#' `.assign_on_ice_shift_metric()` copies goalie and skater timing matrices into
#' home, away, for, and against scalar play-by-play columns.
#'
#' @param play_by_play data.frame play-by-play object
#' @param home_matrix numeric matrix for home on-ice players
#' @param away_matrix numeric matrix for away on-ice players
#' @param metric_suffix scalar timing suffix
#' @returns data.frame with timing columns assigned
#' @keywords internal
.assign_on_ice_shift_metric <- function(
  play_by_play,
  home_matrix,
  away_matrix,
  metric_suffix
) {
  slot_count <- max(0L, ncol(home_matrix) - 1L)
  home_cols <- c(
    paste0('homeGoalie', metric_suffix),
    paste0('homeSkater', seq_len(slot_count), metric_suffix)
  )
  away_cols <- c(
    paste0('awayGoalie', metric_suffix),
    paste0('awaySkater', seq_len(slot_count), metric_suffix)
  )
  for_cols <- c(
    paste0('goalie', metric_suffix, 'For'),
    paste0('skater', seq_len(slot_count), metric_suffix, 'For')
  )
  against_cols <- c(
    paste0('goalie', metric_suffix, 'Against'),
    paste0('skater', seq_len(slot_count), metric_suffix, 'Against')
  )
  for (j in seq_along(home_cols)) {
    play_by_play[[home_cols[j]]] <- home_matrix[, j]
    play_by_play[[away_cols[j]]] <- away_matrix[, j]
    vals_for <- rep(NA_real_, nrow(play_by_play))
    vals_against <- rep(NA_real_, nrow(play_by_play))
    home_idx <- play_by_play$isHome %in% TRUE
    away_idx <- play_by_play$isHome %in% FALSE
    vals_for[home_idx] <- home_matrix[home_idx, j]
    vals_for[away_idx] <- away_matrix[away_idx, j]
    vals_against[home_idx] <- away_matrix[home_idx, j]
    vals_against[away_idx] <- home_matrix[away_idx, j]
    play_by_play[[for_cols[j]]] <- vals_for
    play_by_play[[against_cols[j]]] <- vals_against
  }
  play_by_play
}

#' Add on-ice shift timing context
#'
#' `.add_on_ice_shift_timing_context()` enriches play-by-play rows with scalar
#' shift elapsed and since-last-shift columns for the on-ice goalies and
#' skaters.
#'
#' @param play_by_play data.frame play-by-play object with on-ice player IDs
#' @param game game ID
#' @param shift_data optional shift chart data.frame
#' @returns data.frame with scalar on-ice timing columns added
#' @keywords internal
.add_on_ice_shift_timing_context <- function(
  play_by_play,
  game,
  shift_data = NULL
) {
  if (is.null(shift_data)) {
    return(play_by_play)
  }
  .require_public_pbp_columns(play_by_play, 'periodNumber', 'add_shift_times')
  add_shift_times(play_by_play, shift_data)
}

test_that("gc_play_by_play() returns the public schema and fills goal shooters", {
  gc_raw <- data.frame(
    eventId = 1:3,
    timeInPeriod = c("00:10", "00:20", "00:30"),
    timeRemaining = c("19:50", "19:40", "19:30"),
    situationCode = c("1551", "1551", "1551"),
    homeTeamDefendingSide = c("left", "left", "left"),
    typeCode = c(502L, 505L, 506L),
    typeDescKey = c("faceoff", "goal", "shot-on-goal"),
    sortOrder = 1:3,
    pptReplayUrl = c(NA_character_, "https://example.com/ppt", NA_character_),
    periodDescriptor.number = c(1L, 1L, 1L),
    periodDescriptor.periodType = c("REG", "REG", "REG"),
    periodDescriptor.maxRegulationPeriods = c(3L, 3L, 3L),
    details.eventOwnerTeamId = c(10L, 10L, 20L),
    details.winningPlayerId = c(11L, NA_integer_, NA_integer_),
    details.losingPlayerId = c(21L, NA_integer_, NA_integer_),
    details.xCoord = c(0, 60, -55),
    details.yCoord = c(0, 10, -8),
    details.zoneCode = c("N", "O", "O"),
    details.shootingPlayerId = c(NA_integer_, NA_integer_, 99L),
    details.reason = c(NA_character_, "wrist", "snap"),
    details.shotType = c(NA_character_, "wrist", "snap"),
    details.goalieInNetId = c(NA_integer_, 30L, 40L),
    details.awaySOG = c(0L, 0L, 1L),
    details.homeSOG = c(0L, 1L, 1L),
    details.typeCode = c(NA_integer_, NA_integer_, NA_integer_),
    details.descKey = c(NA_character_, NA_character_, NA_character_),
    details.duration = c(NA_integer_, NA_integer_, NA_integer_),
    details.committedByPlayerId = c(NA_integer_, NA_integer_, NA_integer_),
    details.drawnByPlayerId = c(NA_integer_, NA_integer_, NA_integer_),
    details.scoringPlayerId = c(NA_integer_, 88L, NA_integer_),
    details.scoringPlayerTotal = c(NA_integer_, 1L, NA_integer_),
    details.assist1PlayerId = c(NA_integer_, 77L, NA_integer_),
    details.assist1PlayerTotal = c(NA_integer_, 1L, NA_integer_),
    details.assist2PlayerId = c(NA_integer_, 66L, NA_integer_),
    details.assist2PlayerTotal = c(NA_integer_, 1L, NA_integer_),
    details.awayScore = c(0L, 0L, 0L),
    details.homeScore = c(0L, 1L, 1L),
    details.highlightClipSharingUrl = c(NA_character_, "https://example.com/share", NA_character_),
    details.highlightClipSharingUrlFr = c(NA_character_, NA_character_, NA_character_),
    details.highlightClip = c(NA_character_, "https://example.com/highlight", NA_character_),
    details.highlightClipFr = c(NA_character_, NA_character_, NA_character_),
    details.discreteClip = c(NA_character_, "https://example.com/discrete", NA_character_),
    details.discreteClipFr = c(NA_character_, NA_character_, NA_character_),
    stringsAsFactors = FALSE
  )

  local_mocked_bindings(
    .perform_parallel_requests = function(reqs, on_error = "return") {
      expect_named(reqs, c("pbp_meta", "html_pbp"))
      list(
        pbp_meta = structure(list(kind = "pbp_meta"), class = "mock_resp"),
        html_pbp = structure(list(kind = "html_pbp"), class = "mock_resp")
      )
    },
    .parallel_request_failed = function(resp) FALSE,
    .nhl_json_from_response = function(resp) {
      if (identical(resp$kind, "pbp_meta")) {
        return(list(
          plays = gc_raw,
          rosterSpots = data.frame(),
          homeTeam = list(id = 10L, abbrev = "HOM"),
          awayTeam = list(id = 20L, abbrev = "AWY")
        ))
      }
      stop("Unexpected response kind")
    },
    .optional_html_pbp_rows_from_response = function(...) data.frame(),
    .add_html_on_ice_players = function(play_by_play, ...) {
      play_by_play <- .add_empty_html_on_ice_columns(play_by_play)
      play_by_play$homeGoaliePlayerId <- c(NA_integer_, 30L, 40L)
      play_by_play$awayGoaliePlayerId <- c(NA_integer_, 50L, 60L)
      play_by_play$goaliePlayerIdFor <- c(NA_integer_, 30L, 60L)
      play_by_play$goaliePlayerIdAgainst <- c(NA_integer_, 50L, 40L)
      play_by_play$homeSkater1PlayerId <- c(NA_integer_, 101L, 102L)
      play_by_play$awaySkater1PlayerId <- c(NA_integer_, 201L, 202L)
      play_by_play$skater1PlayerIdFor <- c(NA_integer_, 101L, 202L)
      play_by_play$skater1PlayerIdAgainst <- c(NA_integer_, 201L, 102L)
      play_by_play$homeSkater6PlayerId <- c(NA_integer_, 106L, NA_integer_)
      play_by_play$awaySkater6PlayerId <- c(NA_integer_, NA_integer_, 206L)
      play_by_play$skater6PlayerIdFor <- c(NA_integer_, 106L, 206L)
      play_by_play$skater6PlayerIdAgainst <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$homeSkater7PlayerId <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$awaySkater7PlayerId <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$skater7PlayerIdFor <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$skater7PlayerIdAgainst <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$homeSkater8PlayerId <- c(NA_integer_, 108L, NA_integer_)
      play_by_play$awaySkater8PlayerId <- c(NA_integer_, NA_integer_, 208L)
      play_by_play$skater8PlayerIdFor <- c(NA_integer_, 108L, 208L)
      play_by_play$skater8PlayerIdAgainst <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play
    },
    .add_on_ice_shift_timing_context = function(...) stop(".add_on_ice_shift_timing_context() should not be called"),
    .package = "nhlscraper"
  )

  out <- gc_play_by_play(2010020001)

  expect_false(any(c(
    "period", "typeCode", "typeDescKey", "homeSOG", "awaySOG", "SOGFor",
    "SOGAgainst", "SOGDifferential", "descKey", "duration", "timeInPeriod",
    "timeRemaining", "awayScore", "homeScore"
  ) %in% names(out)))
  expect_true(all(c(
    "periodNumber", "periodType", "eventTypeCode", "eventTypeDescKey",
    "homeShots", "awayShots", "shotsFor", "shotsAgainst", "shotDifferential",
    "penaltyTypeDescKey", "penaltyDuration", "pptReplayUrl", "goalieInNetId",
    "homeGoaliePlayerId", "awayGoaliePlayerId", "goaliePlayerIdFor",
    "goaliePlayerIdAgainst", "homeSkater1PlayerId", "awaySkater1PlayerId",
    "skater1PlayerIdFor", "skater1PlayerIdAgainst", "homeSkater6PlayerId",
    "awaySkater6PlayerId", "skater6PlayerIdFor", "skater6PlayerIdAgainst",
    "homeSkater7PlayerId", "awaySkater7PlayerId", "skater7PlayerIdFor",
    "skater7PlayerIdAgainst", "homeSkater8PlayerId", "awaySkater8PlayerId",
    "skater8PlayerIdFor", "skater8PlayerIdAgainst"
  ) %in% names(out)))
  strength_idx <- match("strengthState", names(out))
  expect_equal(
    names(out)[(strength_idx + 1L):(strength_idx + 4L)],
    c(
      "homeGoaliePlayerId",
      "awayGoaliePlayerId",
      "goaliePlayerIdFor",
      "goaliePlayerIdAgainst"
    )
  )
  expect_equal(
    out$shootingPlayerId[out$eventTypeDescKey == "goal"],
    out$scoringPlayerId[out$eventTypeDescKey == "goal"]
  )
  expect_true(all(c("playerId", "blockingPlayerId", "servedByPlayerId") %in% names(out)))
  expect_true(all(is.na(out$playerId)))
  expect_true(all(is.na(out$blockingPlayerId)))
  expect_true(all(is.na(out$servedByPlayerId)))
  expect_equal(out$goalieInNetId, c(NA_integer_, 30L, 40L))
  expect_false("homeGoalieSecondsElapsedInShift" %in% names(out))
  expect_false("homeGoalieSecondsElapsedInPeriodSinceLastShift" %in% names(out))
  expect_equal(
    names(out)[match("drawnByPlayerId", names(out)) + 1L],
    "servedByPlayerId"
  )
  expect_equal(
    names(out)[match("blockingPlayerId", names(out)) + 1L],
    "goalieInNetId"
  )
  expect_equal(
    names(out)[match("goalieInNetId", names(out)) + 1L],
    "shootingPlayerId"
  )
})

test_that("wsc_play_by_play() returns the public schema with utc and no clip fields", {
  wsc_raw <- data.frame(
    id = 999L,
    eventId = 1:3,
    period = c(1L, 1L, 1L),
    timeInPeriod = c("00:10", "00:20", "00:30"),
    secondsRemaining = c(1190L, 1180L, 1170L),
    situationCode = c("1551", "1551", "1551"),
    typeCode = c(502L, 505L, 506L),
    typeDescKey = c("faceoff", "goal", "shot-on-goal"),
    homeTeamDefendingSide = c("left", "left", "left"),
    sortOrder = 1:3,
    utc = c(
      "2010-10-07T00:00:10Z",
      "2010-10-07T00:00:20Z",
      "2010-10-07T00:00:30Z"
    ),
    eventOwnerTeamId = c(10L, 10L, 20L),
    losingPlayerId = c(21L, NA_integer_, NA_integer_),
    winningPlayerId = c(11L, NA_integer_, NA_integer_),
    xCoord = c(0, 60, -55),
    yCoord = c(0, 10, -8),
    zoneCode = c("N", "O", "O"),
    shootingPlayerId = c(NA_integer_, NA_integer_, 99L),
    reason = c(NA_character_, "wrist", "snap"),
    shotType = c(NA_character_, "wrist", "snap"),
    goalieInNetId = c(NA_integer_, 30L, 40L),
    awaySOG = c(0L, 0L, 1L),
    homeSOG = c(0L, 1L, 1L),
    penaltyTypeCode = c(NA_integer_, NA_integer_, NA_integer_),
    descKey = c(NA_character_, NA_character_, NA_character_),
    duration = c(NA_integer_, NA_integer_, NA_integer_),
    committedByPlayerId = c(NA_integer_, NA_integer_, NA_integer_),
    drawnByPlayerId = c(NA_integer_, NA_integer_, NA_integer_),
    strength = c(NA_character_, NA_character_, NA_character_),
    strengthCode = c(NA_character_, NA_character_, NA_character_),
    goalCode = c(NA_character_, NA_character_, NA_character_),
    scoringPlayerId = c(NA_integer_, 88L, NA_integer_),
    assist1PlayerId = c(NA_integer_, 77L, NA_integer_),
    assist2PlayerId = c(NA_integer_, 66L, NA_integer_),
    awayScore = c(0L, 0L, 0L),
    homeScore = c(0L, 1L, 1L),
    scoringPlayerTotal = c(NA_integer_, 1L, NA_integer_),
    assist1PlayerTotal = c(NA_integer_, 1L, NA_integer_),
    assist2PlayerTotal = c(NA_integer_, 1L, NA_integer_),
    stringsAsFactors = FALSE
  )

  local_mocked_bindings(
    .perform_parallel_requests = function(reqs, on_error = "return") {
      expect_named(reqs, c("pbp_meta", "wsc", "html_pbp"))
      list(
        pbp_meta = structure(list(kind = "pbp_meta"), class = "mock_resp"),
        wsc = structure(list(kind = "wsc"), class = "mock_resp"),
        html_pbp = structure(list(kind = "html_pbp"), class = "mock_resp")
      )
    },
    .parallel_request_failed = function(resp) FALSE,
    .nhl_json_from_response = function(resp) {
      if (identical(resp$kind, "pbp_meta")) {
        return(list(
          rosterSpots = data.frame(),
          homeTeam = list(id = 10L, abbrev = "HOM"),
          awayTeam = list(id = 20L, abbrev = "AWY")
        ))
      }
      if (identical(resp$kind, "wsc")) {
        return(wsc_raw)
      }
      stop("Unexpected response kind")
    },
    .optional_html_pbp_rows_from_response = function(...) data.frame(),
    .add_html_on_ice_players = function(play_by_play, ...) {
      play_by_play <- .add_empty_html_on_ice_columns(play_by_play)
      play_by_play$homeGoaliePlayerId <- c(NA_integer_, 30L, 40L)
      play_by_play$awayGoaliePlayerId <- c(NA_integer_, 50L, 60L)
      play_by_play$goaliePlayerIdFor <- c(NA_integer_, 30L, 60L)
      play_by_play$goaliePlayerIdAgainst <- c(NA_integer_, 50L, 40L)
      play_by_play$homeSkater1PlayerId <- c(NA_integer_, 101L, 102L)
      play_by_play$awaySkater1PlayerId <- c(NA_integer_, 201L, 202L)
      play_by_play$skater1PlayerIdFor <- c(NA_integer_, 101L, 202L)
      play_by_play$skater1PlayerIdAgainst <- c(NA_integer_, 201L, 102L)
      play_by_play$homeSkater6PlayerId <- c(NA_integer_, 106L, NA_integer_)
      play_by_play$awaySkater6PlayerId <- c(NA_integer_, NA_integer_, 206L)
      play_by_play$skater6PlayerIdFor <- c(NA_integer_, 106L, 206L)
      play_by_play$skater6PlayerIdAgainst <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$homeSkater7PlayerId <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$awaySkater7PlayerId <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$skater7PlayerIdFor <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$skater7PlayerIdAgainst <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play$homeSkater8PlayerId <- c(NA_integer_, 108L, NA_integer_)
      play_by_play$awaySkater8PlayerId <- c(NA_integer_, NA_integer_, 208L)
      play_by_play$skater8PlayerIdFor <- c(NA_integer_, 108L, 208L)
      play_by_play$skater8PlayerIdAgainst <- c(NA_integer_, NA_integer_, NA_integer_)
      play_by_play
    },
    .add_on_ice_shift_timing_context = function(...) stop(".add_on_ice_shift_timing_context() should not be called"),
    .package = "nhlscraper"
  )

  out <- wsc_play_by_play(2010020001)

  expect_false(any(c(
    "period", "typeCode", "typeDescKey", "homeSOG", "awaySOG", "SOGFor",
    "SOGAgainst", "SOGDifferential", "descKey", "duration", "timeInPeriod",
    "secondsRemaining", "awayScore", "homeScore", "periodType",
    "discreteClip", "highlightClip", "highlightClipSharingUrl", "pptReplayUrl"
  ) %in% names(out)))
  expect_true(all(c(
    "periodNumber", "utc", "eventTypeCode", "eventTypeDescKey", "homeShots",
    "awayShots", "shotsFor", "shotsAgainst", "shotDifferential",
    "penaltyTypeDescKey", "penaltyDuration", "goalieInNetId", "homeGoaliePlayerId",
    "awayGoaliePlayerId", "goaliePlayerIdFor", "goaliePlayerIdAgainst",
    "homeSkater1PlayerId", "awaySkater1PlayerId", "skater1PlayerIdFor",
    "skater1PlayerIdAgainst", "homeSkater6PlayerId", "awaySkater6PlayerId",
    "skater6PlayerIdFor", "skater6PlayerIdAgainst", "homeSkater7PlayerId",
    "awaySkater7PlayerId", "skater7PlayerIdFor", "skater7PlayerIdAgainst",
    "homeSkater8PlayerId", "awaySkater8PlayerId", "skater8PlayerIdFor",
    "skater8PlayerIdAgainst"
  ) %in% names(out)))
  expect_match(names(out)[10], "^utc$")
  strength_idx <- match("strengthState", names(out))
  expect_equal(
    names(out)[(strength_idx + 1L):(strength_idx + 4L)],
    c(
      "homeGoaliePlayerId",
      "awayGoaliePlayerId",
      "goaliePlayerIdFor",
      "goaliePlayerIdAgainst"
    )
  )
  expect_equal(
    out$shootingPlayerId[out$eventTypeDescKey == "goal"],
    out$scoringPlayerId[out$eventTypeDescKey == "goal"]
  )
  expect_true(all(c("playerId", "blockingPlayerId", "servedByPlayerId") %in% names(out)))
  expect_true(all(is.na(out$playerId)))
  expect_true(all(is.na(out$blockingPlayerId)))
  expect_true(all(is.na(out$servedByPlayerId)))
  expect_equal(out$goalieInNetId, c(NA_integer_, 30L, 40L))
  expect_false("homeGoalieSecondsElapsedInShift" %in% names(out))
  expect_false("homeGoalieSecondsElapsedInPeriodSinceLastShift" %in% names(out))
  expect_equal(
    names(out)[match("drawnByPlayerId", names(out)) + 1L],
    "servedByPlayerId"
  )
  expect_equal(
    names(out)[match("blockingPlayerId", names(out)) + 1L],
    "goalieInNetId"
  )
  expect_equal(
    names(out)[match("goalieInNetId", names(out)) + 1L],
    "shootingPlayerId"
  )
})

test_that("raw play-by-play helpers return flattened raw source rows", {
  gc_raw <- data.frame(
    eventId = 1:2,
    sortOrder = 1:2,
    periodDescriptor.number = c(1L, 1L),
    stringsAsFactors = FALSE
  )
  wsc_raw <- data.frame(
    id = c(10L, 11L),
    eventId = 1:2,
    sortOrder = 1:2,
    stringsAsFactors = FALSE
  )

  local_mocked_bindings(
    nhl_api = function(path, type, ...) {
      expect_identical(type, "w")
      if (identical(path, "v1/gamecenter/2010020001/play-by-play")) {
        return(list(plays = gc_raw))
      }
      if (identical(path, "v1/wsc/play-by-play/2010020001")) {
        return(wsc_raw)
      }
      stop("Unexpected path")
    },
    .package = "nhlscraper"
  )

  gc_out <- gc_play_by_play_raw(2010020001)
  wsc_out <- wsc_play_by_play_raw(2010020001)

  expect_equal(names(gc_out)[1], "gameId")
  expect_equal(gc_out$gameId, rep(2010020001L, 2L))
  expect_true("periodDescriptor.number" %in% names(gc_out))

  expect_equal(names(wsc_out)[1], "gameId")
  expect_equal(wsc_out$gameId, rep(2010020001L, 2L))
  expect_false("id" %in% names(wsc_out))
})

test_that("repair_public_pbp_sequence drops invalid clocks and repairs boundary order", {
  pbp <- data.frame(
    gameId = rep(2019030016L, 6L),
    seasonId = rep(20192020L, 6L),
    gameTypeId = rep(3L, 6L),
    eventId = c(6L, 9L, 10L, 11L, 12L, 13L),
    sortOrder = c(6L, 9L, 10L, 11L, 12L, 13L),
    periodNumber = rep(1L, 6L),
    timeInPeriod = c("00:00", "00:00", "00:00", "00:09", "00:08", "08:70"),
    eventTypeDescKey = c("period-start", "faceoff", "period-end", "shot-on-goal", "hit", "penalty"),
    stringsAsFactors = FALSE
  )

  out <- pbp |>
    .strip_time_period() |>
    .repair_public_pbp_sequence()

  expect_false(13L %in% out$eventId)
  expect_equal(out$eventTypeDescKey, c("period-start", "faceoff", "hit", "shot-on-goal", "period-end"))
  expect_equal(out$timeInPeriod[out$eventTypeDescKey == "period-end"], "20:00")
  expect_true(all(diff(out$secondsElapsedInPeriod) >= 0))
})

test_that("illogically ordered boundary faceoffs are removed", {
  pbp <- data.frame(
    gameId = rep(1L, 5L),
    eventId = c(10L, 11L, 12L, 13L, 14L),
    sortOrder = c(190L, 201L, 206L, 209L, 211L),
    periodNumber = c(1L, 1L, 1L, 2L, 2L),
    timeInPeriod = c("18:24", "20:00", "00:00", "00:00", "00:00"),
    eventTypeDescKey = c("shot-on-goal", "period-end", "faceoff", "period-start", "faceoff"),
    stringsAsFactors = FALSE
  )

  out <- pbp |>
    .strip_game_id() |>
    .strip_time_period() |>
    .drop_illogical_ordered_events()

  expect_false(12L %in% out$eventId)
  expect_equal(out$eventId, c(10L, 11L, 13L, 14L))
})

test_that("add_deltas(gc_pbp()) returns non-empty data.frame", {
  skip_if_offline()
  test <- add_deltas(gc_pbp())
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

test_that("add_deltas() starts new sequences at faceoffs", {
  pbp <- data.frame(
    gameId = rep(1L, 11),
    eventId = seq_len(11),
    secondsElapsedInGame = c(9, 10, 10, 10, 11, 20, 20, 20, 20, 20, 21),
    xCoord = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    yCoord = rep(0, 11),
    xCoordNorm = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    yCoordNorm = rep(0, 11),
    distance = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    angle = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    eventTypeDescKey = c(
      "shot", "faceoff", "shot", "shot", "shot",
      "shot", "faceoff", "shot", "shot", "shot", "shot"
    ),
    situationCode = rep("1551", 11),
    sortOrder = seq_len(11)
  )

  out <- add_deltas(pbp)

  expect_true(is.na(out$dSecondsElapsedInSequence[2]))
  expect_true(is.na(out$dXCoordNorm[2]))
  expect_true(is.na(out$dSecondsElapsedInSequence[7]))
  expect_true(is.na(out$dXCoordNorm[7]))
  expect_true(is.na(out$secondsElapsedInSequence[1]))
  expect_equal(out$secondsElapsedInSequence[2], 0)
  expect_equal(out$secondsElapsedInSequence[5], 1)
  expect_equal(out$secondsElapsedInSequence[8], 0)
  expect_true(is.na(out$eventIdPrev[2]))
  expect_true(is.na(out$eventIdPrev[7]))
  expect_equal(out$dXCoord[8], 10)
  expect_equal(out$dXCoordNorm[8], 10)
  expect_equal(out$dSecondsElapsedInSequence[8], 0)
  expect_equal(out$eventIdPrev[8], 7)
})

test_that("add_deltas() preserves same-second internal rate scaling", {
  pbp <- data.frame(
    gameId = rep(1L, 11),
    eventId = seq_len(11),
    secondsElapsedInGame = c(9, 10, 10, 10, 11, 20, 20, 20, 20, 20, 21),
    xCoord = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    yCoord = rep(0, 11),
    xCoordNorm = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    yCoordNorm = rep(0, 11),
    distance = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    angle = c(1, 0, 3, 6, 10, 20, 50, 60, 70, 80, 90),
    eventTypeDescKey = c(
      "shot", "faceoff", "shot", "shot", "shot",
      "shot", "faceoff", "shot", "shot", "shot", "shot"
    ),
    situationCode = rep("1551", 11),
    sortOrder = seq_len(11)
  )

  out <- .compute_pbp_deltas(pbp)

  expect_equal(out$dSecondsElapsedInSequence[3], 0)
  expect_equal(out$dXCoordPerSecond[3], 9)
  expect_equal(out$dXCoordPerSecond[4], 9)
  expect_equal(out$dXCoordNormPerSecond[3], 9)
  expect_equal(out$dXCoordNormPerSecond[4], 9)
  expect_equal(out$dXCoordPerSecond[8], 40)
  expect_equal(out$dXCoordPerSecond[9], 40)
  expect_equal(out$dXCoordPerSecond[10], 40)
  expect_equal(out$dXCoordPerSecond[11], 10)
  expect_equal(out$dXCoordNormPerSecond[8], 40)
  expect_equal(out$dXCoordNormPerSecond[9], 40)
  expect_equal(out$dXCoordNormPerSecond[10], 40)
  expect_equal(out$dXCoordNormPerSecond[11], 10)
  expect_equal(out$dYCoordPerSecond[c(3, 4, 8, 9, 10, 11)], rep(0, 6))
})

test_that("add_deltas() respects game boundaries on aggregated inputs", {
  pbp <- data.frame(
    gameId = c(1L, 1L, 2L, 2L),
    eventId = c(101L, 102L, 201L, 202L),
    secondsElapsedInGame = c(10, 11, 10, 11),
    xCoord = c(0, 10, 5, 15),
    yCoord = c(0, 0, 1, 1),
    xCoordNorm = c(0, 10, 5, 15),
    yCoordNorm = c(0, 0, 1, 1),
    distance = c(0, 10, 5, 15),
    angle = c(0, 10, 5, 15),
    eventTypeDescKey = c("faceoff", "shot", "faceoff", "shot"),
    situationCode = rep("1551", 4),
    sortOrder = c(1L, 2L, 1L, 2L)
  )

  out <- add_deltas(pbp)

  expect_true(all(is.na(out$eventIdPrev[c(1L, 3L)])))
  expect_equal(out$eventIdPrev[c(2L, 4L)], c(101L, 201L))
  expect_equal(out$secondsElapsedInSequence[c(2L, 4L)], c(1, 1))
  expect_equal(out$dXCoord[c(2L, 4L)], c(10, 10))
})

test_that("add_deltas() leaves shootout and penalty-shot rows as NA", {
  pbp <- data.frame(
    gameId = rep(1L, 5),
    eventId = seq_len(5),
    secondsElapsedInGame = c(10, 11, 11, 12, 13),
    xCoord = c(0, 10, 20, 30, 40),
    yCoord = rep(0, 5),
    xCoordNorm = c(0, 10, 20, 30, 40),
    yCoordNorm = rep(0, 5),
    distance = c(0, 10, 20, 30, 40),
    angle = c(0, 10, 20, 30, 40),
    eventTypeDescKey = c("faceoff", "shot", "shot-on-goal", "goal", "shot"),
    situationCode = c("1551", "1551", "1010", "0101", "1551"),
    sortOrder = seq_len(5)
  )

  out <- add_deltas(pbp)

  expect_true(all(is.na(out[3, c(
    "dSecondsElapsedInSequence", "dXCoord", "dYCoord", "dXCoordNorm",
    "dYCoordNorm", "dDistance", "dAngle"
  )])))
  expect_true(all(is.na(out[4, c(
    "dSecondsElapsedInSequence", "dXCoord", "dYCoord", "dXCoordNorm",
    "dYCoordNorm", "dDistance", "dAngle"
  )])))
  expect_true(is.na(out$secondsElapsedInSequence[3]))
  expect_true(is.na(out$secondsElapsedInSequence[4]))
  expect_true(is.na(out$eventIdPrev[3]))
  expect_true(is.na(out$eventIdPrev[4]))
  expect_equal(out$dXCoord[5], 30)
  expect_equal(out$dSecondsElapsedInSequence[5], 2)
  expect_equal(out$eventIdPrev[5], 2)
  expect_equal(out$secondsElapsedInSequence[5], 3)
})

test_that("add_deltas() places secondsElapsedInSequence after shift-rest-against", {
  pbp <- data.frame(
    gameId = rep(1L, 3),
    eventId = seq_len(3),
    secondsElapsedInGame = c(10, 11, 12),
    xCoord = c(0, 10, 20),
    yCoord = c(0, 0, 0),
    xCoordNorm = c(0, 10, 20),
    yCoordNorm = c(0, 0, 0),
    distance = c(0, 10, 20),
    angle = c(0, 10, 20),
    secondsElapsedInPeriodSinceLastShiftAgainst = I(list(1L, 2L, 3L)),
    eventTypeDescKey = c("faceoff", "shot", "shot"),
    situationCode = rep("1551", 3),
    sortOrder = seq_len(3)
  )

  out <- add_deltas(pbp)
  cols <- names(out)

  expect_equal(
    match("secondsElapsedInSequence", cols),
    match("secondsElapsedInPeriodSinceLastShiftAgainst", cols) + 1L
  )
  expect_false(any(c(
    "dXN", "dYN", "dD", "dA", "dT", "dXNdT", "dYNdT", "dDdT", "dAdT"
  ) %in% cols))
  expect_true(all(c(
    "eventIdPrev", "dSecondsElapsedInSequence", "dXCoord", "dYCoord",
    "dXCoordNorm", "dYCoordNorm", "dDistance", "dAngle",
    "dXCoordPerSecond", "dYCoordPerSecond", "dXCoordNormPerSecond",
    "dYCoordNormPerSecond", "dDistancePerSecond", "dAnglePerSecond"
  ) %in% cols))
  expect_equal(
    cols[(match("dAngle", cols) + 1L):(match("dAngle", cols) + 6L)],
    c(
      "dXCoordPerSecond", "dYCoordPerSecond", "dXCoordNormPerSecond",
      "dYCoordNormPerSecond", "dDistancePerSecond", "dAnglePerSecond"
    )
  )
})

test_that("add_shooter_biometrics() keeps blocked shots and returns simplified columns", {
  local_mocked_bindings(
    players = function() {
      data.frame(
        playerId = 10L,
        height = 72L,
        weight = 190L,
        handCode = "L",
        positionCode = "C",
        birthDate = "2000-01-01",
        stringsAsFactors = FALSE
      )
    },
    games = function() {
      data.frame(
        gameId = 1L,
        gameDate = "2025-01-10",
        stringsAsFactors = FALSE
      )
    },
    .package = "nhlscraper"
  )

  play_by_play <- data.frame(
    gameId = 1L,
    scoringPlayerId = NA_integer_,
    shootingPlayerId = 10L,
    stringsAsFactors = FALSE
  )

  out <- add_shooter_biometrics(play_by_play)

  expect_equal(out$shooterHeight, 72L)
  expect_equal(out$shooterWeight, 190L)
  expect_equal(out$shooterHandCode, "L")
  expect_equal(out$shooterAge, 25L)
  expect_equal(out$shooterPositionCode, "C")
  expect_false("shooterSide" %in% names(out))
})

test_that("add_goalie_biometrics() leaves missing blocked-shot goalies as NA", {
  local_mocked_bindings(
    players = function() {
      data.frame(
        playerId = 30L,
        height = 74L,
        weight = 185L,
        handCode = "R",
        birthDate = "1998-01-01",
        stringsAsFactors = FALSE
      )
    },
    games = function() {
      data.frame(
        gameId = 1L,
        gameDate = "2025-01-10",
        stringsAsFactors = FALSE
      )
    },
    .package = "nhlscraper"
  )

  play_by_play <- data.frame(
    gameId = c(1L, 1L),
    eventTypeDescKey = c("blocked-shot", "blocked-shot"),
    goaliePlayerIdAgainst = c(NA_integer_, 30L),
    stringsAsFactors = FALSE
  )

  out <- add_goalie_biometrics(play_by_play)

  expect_true(all(is.na(out$goalieHeight[1])))
  expect_true(all(is.na(out$goalieWeight[1])))
  expect_true(all(is.na(out$goalieHandCode[1])))
  expect_true(all(is.na(out$goalieAge[1])))
  expect_equal(out$goalieHeight[2], 74L)
  expect_equal(out$goalieWeight[2], 185L)
  expect_equal(out$goalieHandCode[2], "R")
  expect_equal(out$goalieAge[2], 27L)
  expect_false("goalieSide" %in% names(out))
})

test_that("biometrics helpers respect game dates on aggregated inputs", {
  local_mocked_bindings(
    players = function() {
      data.frame(
        playerId = c(10L, 30L),
        height = c(72L, 74L),
        weight = c(190L, 185L),
        handCode = c("L", "R"),
        positionCode = c("C", NA_character_),
        birthDate = c("2000-01-01", "1998-01-01"),
        stringsAsFactors = FALSE
      )
    },
    games = function() {
      data.frame(
        gameId = c(1L, 2L),
        gameDate = c("2025-01-10", "2026-01-10"),
        stringsAsFactors = FALSE
      )
    },
    .package = "nhlscraper"
  )

  shooter_pbp <- data.frame(
    gameId = c(1L, 2L),
    scoringPlayerId = c(NA_integer_, NA_integer_),
    shootingPlayerId = c(10L, 10L),
    stringsAsFactors = FALSE
  )
  goalie_pbp <- data.frame(
    gameId = c(1L, 2L),
    goalieInNetId = c(30L, 30L),
    goaliePlayerIdAgainst = c(31L, 31L),
    stringsAsFactors = FALSE
  )

  shooter_out <- add_shooter_biometrics(shooter_pbp)
  goalie_out <- add_goalie_biometrics(goalie_pbp)

  expect_equal(shooter_out$shooterAge, c(25L, 26L))
  expect_equal(goalie_out$goalieAge, c(27L, 28L))
  expect_equal(shooter_out$shooterHeight, c(72L, 72L))
  expect_equal(goalie_out$goalieHeight, c(74L, 74L))
})

test_that("HTML PBP parser resolves on-ice players and owner perspective", {
  roster_lookup <- .build_game_roster_lookup(data.frame(
    teamId = c(8L, 8L, 8L, 10L, 10L, 10L),
    playerId = c(801L, 802L, 803L, 1001L, 1002L, 1003L),
    sweaterNumber = c(11L, 31L, 76L, 37L, 35L, 2L),
    positionCode = c("C", "G", "D", "C", "G", "D"),
    playerLastName = c("Gomez", "Price", "Subban", "Brent", "Giguere", "Schenn"),
    playerFirstName = c("Scott", "Carey", "P.K.", "Clarke", "Jean-Sebastien", "Luke"),
    stringsAsFactors = FALSE
  ))

  html <- paste(
    "<table>",
    "<tr><th>#</th><th>Per</th><th>Str</th><th>Time:ElapsedGame</th><th>Event</th><th>Description</th><th>MTL On Ice</th><th>TOR On Ice</th></tr>",
    "<tr><td>2</td><td>1</td><td>EV</td><td>0:0020:00</td><td>FAC</td><td>MTL won Neu. Zone - MTL #11 GOMEZ vs TOR #37 BRENT</td><td>11 C 76 D 31 G</td><td>37 C 2 D 35 G</td></tr>",
    "<tr><td>7</td><td>1</td><td>EV</td><td>1:1318:47</td><td>BLOCK</td><td>MTL #76 SUBBAN BLOCKED BY TOR #2 SCHENN, Wrist, Def. Zone</td><td>11 C 76 D 31 G</td><td>37 C 2 D 35 G</td></tr>",
    "<tr><td>82</td><td>1</td><td>EV</td><td>14:025:58</td><td>PENL</td><td>MTL #81 ELLER Hooking(2 min), Def. Zone Drawn By: TOR #37 BRENT</td><td>11 C 76 D 31 G</td><td>37 C 2 D 35 G</td></tr>",
    "</table>"
  )
  doc <- xml2::read_html(html)
  out <- .parse_html_pbp_doc(
    doc = doc,
    roster_lookup = roster_lookup,
    home_team_id = 10L,
    away_team_id = 8L,
    home_abbrev = "TOR",
    away_abbrev = "MTL",
    is_playoffs = FALSE
  )

  expect_equal(out$typeDescKey, c("faceoff", "blocked-shot", "penalty"))
  expect_equal(out$ownerTeamId, c(8L, 8L, 8L))
  expect_equal(out$primaryPlayerId[1], 801L)
  expect_equal(out$secondaryPlayerId[1], 1001L)
  expect_equal(out$primaryPlayerId[2], 803L)
  expect_equal(out$secondaryPlayerId[2], 1003L)
  expect_equal(out$homeGoaliePlayerId, c(1002L, 1002L, 1002L))
  expect_equal(out$awayGoaliePlayerId, c(802L, 802L, 802L))
  expect_equal(out$homeSkaterPlayerIds[[1]], c(1001L, 1003L))
  expect_equal(out$awaySkaterPlayerIds[[1]], c(801L, 803L))
})

test_that("HTML PBP parser preserves six-skater empty-net rows", {
  roster_lookup <- .build_game_roster_lookup(data.frame(
    teamId = c(rep(8L, 6L), rep(10L, 7L)),
    playerId = c(801:806, 1001:1007),
    sweaterNumber = c(11L, 12L, 13L, 14L, 15L, 16L, 31L, 32L, 33L, 34L, 35L, 36L, 37L),
    positionCode = c(rep("F", 6L), "G", rep("F", 6L)),
    playerLastName = paste0("P", c(801:806, 1001:1007)),
    playerFirstName = "A",
    stringsAsFactors = FALSE
  ))

  html <- paste(
    "<table>",
    "<tr><th>#</th><th>Per</th><th>Str</th><th>Time:ElapsedGame</th><th>Event</th><th>Description</th><th>MTL On Ice</th><th>TOR On Ice</th></tr>",
    "<tr><td>301</td><td>3</td><td>6v5</td><td>19:100:50</td><td>SHOT</td><td>TOR ONGOAL - #32 P1002, Wrist, Off. Zone, 20 ft.</td><td>11 F 12 F 13 F 14 F 15 F 16 F</td><td>31 G 32 F 33 F 34 F 35 F 36 F 37 F</td></tr>",
    "</table>"
  )
  doc <- xml2::read_html(html)
  out <- .parse_html_pbp_doc(
    doc = doc,
    roster_lookup = roster_lookup,
    home_team_id = 10L,
    away_team_id = 8L,
    home_abbrev = "TOR",
    away_abbrev = "MTL",
    is_playoffs = FALSE
  )

  expect_true(is.na(out$awayGoaliePlayerId[1]))
  expect_equal(out$awaySkaterPlayerIds[[1]], 801:806)
  expect_equal(out$homeGoaliePlayerId[1], 1001L)
  expect_equal(out$homeSkaterPlayerIds[[1]], 1002:1007)
})

test_that("HTML PBP parser handles dotted team abbreviations", {
  roster_lookup <- .build_game_roster_lookup(data.frame(
    teamId = c(28L, 28L, 22L, 22L),
    playerId = c(2801L, 2802L, 2201L, 2202L),
    sweaterNumber = c(19L, 31L, 93L, 40L),
    positionCode = c("C", "G", "C", "G"),
    playerLastName = c("Thornton", "Niemi", "Nugent-Hopkins", "Dubnyk"),
    playerFirstName = c("Joe", "Antti", "Ryan", "Devan"),
    stringsAsFactors = FALSE
  ))

  html <- paste(
    "<table>",
    "<tr><th>#</th><th>Per</th><th>Str</th><th>Time:ElapsedGame</th><th>Event</th><th>Description</th><th>S.J On Ice</th><th>EDM On Ice</th></tr>",
    "<tr><td>2</td><td>1</td><td>EV</td><td>0:0020:00</td><td>FAC</td><td>S.J won Neu. Zone - EDM #93 NUGENT-HOPKINS vs S.J #19 THORNTON</td><td>19 C 31 G</td><td>93 C 40 G</td></tr>",
    "</table>"
  )
  doc <- xml2::read_html(html)
  out <- .parse_html_pbp_doc(
    doc = doc,
    roster_lookup = roster_lookup,
    home_team_id = 22L,
    away_team_id = 28L,
    home_abbrev = "EDM",
    away_abbrev = "SJS",
    is_playoffs = FALSE
  )

  expect_equal(out$typeDescKey, "faceoff")
  expect_equal(out$ownerTeamId, 28L)
  expect_equal(out$homeGoaliePlayerId, 2202L)
  expect_equal(out$awayGoaliePlayerId, 2802L)
})

test_that("HTML actor parser handles reversed faceoff player order", {
  roster_lookup <- .build_game_roster_lookup(data.frame(
    teamId = c(8L, 10L),
    playerId = c(801L, 1001L),
    sweaterNumber = c(15L, 42L),
    positionCode = c("C", "C"),
    playerLastName = c("Halpern", "Bozak"),
    playerFirstName = c("Jeff", "Tyler"),
    stringsAsFactors = FALSE
  ))

  actors <- .html_extract_actor_player_ids(
    description = "TOR won Off. Zone - MTL #15 HALPERN vs TOR #42 BOZAK",
    type_desc_key = "faceoff",
    owner_team_id = 10L,
    home_team_id = 10L,
    away_team_id = 8L,
    home_abbrev = "TOR",
    away_abbrev = "MTL",
    roster_lookup = roster_lookup
  )

  expect_equal(actors$primaryPlayerId, 1001L)
  expect_equal(actors$secondaryPlayerId, 801L)
})

test_that("HTML PBP matcher aligns events when same-second rows differ locally", {
  play_by_play <- data.frame(
    gameId = rep(2010020001L, 3L),
    eventId = c(10L, 11L, 12L),
    periodNumber = c(1L, 1L, 1L),
    secondsElapsedInPeriod = c(100L, 100L, 101L),
    sortOrder = c(10L, 11L, 12L),
    eventTypeDescKey = c("blocked-shot", "penalty", "faceoff"),
    eventOwnerTeamId = c(8L, 8L, 10L),
    isHome = c(FALSE, FALSE, TRUE),
    situationCode = c("1551", "1551", "1451"),
    shootingPlayerId = c(803L, NA_integer_, NA_integer_),
    blockingPlayerId = c(1003L, NA_integer_, NA_integer_),
    committedByPlayerId = c(NA_integer_, 803L, NA_integer_),
    drawnByPlayerId = c(NA_integer_, 1001L, NA_integer_),
    winningPlayerId = c(NA_integer_, NA_integer_, 1001L),
    losingPlayerId = c(NA_integer_, NA_integer_, 801L),
    playerId = c(NA_integer_, NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = c(50L, 51L, 52L),
    period = c(1L, 1L, 1L),
    strengthCodeHtml = c("EV", "EV", "PP"),
    secondsElapsedInPeriod = c(100L, 100L, 101L),
    htmlEventCode = c("BLOCK", "PENL", "FAC"),
    typeDescKey = c("blocked-shot", "penalty", "faceoff"),
    description = c(
      "MTL #76 SUBBAN BLOCKED BY TOR #2 SCHENN",
      "MTL #76 SUBBAN Hooking(2 min), Drawn By: TOR #37 BRENT",
      "TOR won Neu. Zone - TOR #37 BRENT vs MTL #11 GOMEZ"
    ),
    ownerTeamId = c(8L, 8L, 10L),
    primaryPlayerId = c(803L, 803L, 1001L),
    secondaryPlayerId = c(1003L, 1001L, 801L),
    tertiaryPlayerId = c(NA_integer_, NA_integer_, NA_integer_),
    homeGoaliePlayerId = c(1002L, 1002L, 1002L),
    awayGoaliePlayerId = c(802L, 802L, 802L),
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(c(1001L, 1003L), c(1001L, 1003L), c(1001L, 1003L))
  html_rows$awaySkaterPlayerIds <- list(c(801L, 803L), c(801L, 803L), c(801L, 803L))

  matched <- .match_html_pbp_to_api(play_by_play, html_rows)

  expect_equal(matched$apiIndex, c(1L, 2L, 3L))
})

test_that("HTML PBP matcher rescues duplicate shot clusters after greedy backtracking", {
  filler_n <- 26L
  n <- filler_n + 4L

  play_by_play <- data.frame(
    gameId = rep(2015020150L, n),
    eventId = seq_len(n),
    periodNumber = c(3L, rep(1L, filler_n), 3L, 3L, 3L),
    secondsElapsedInPeriod = c(765L, seq_len(filler_n), 766L, 765L, 766L),
    sortOrder = seq_len(n),
    eventTypeDescKey = c("shot-on-goal", rep("faceoff", filler_n), "shot-on-goal", "shot-on-goal", "shot-on-goal"),
    eventOwnerTeamId = c(6L, rep(10L, filler_n), 6L, 6L, 6L),
    isHome = c(FALSE, rep(TRUE, filler_n), FALSE, FALSE, FALSE),
    situationCode = rep("1551", n),
    shootingPlayerId = c(8474625L, rep(NA_integer_, filler_n), 8474625L, 8474625L, 8474625L),
    blockingPlayerId = rep(NA_integer_, n),
    committedByPlayerId = rep(NA_integer_, n),
    drawnByPlayerId = rep(NA_integer_, n),
    winningPlayerId = c(NA_integer_, 1001L + seq_len(filler_n) - 1L, NA_integer_, NA_integer_, NA_integer_),
    losingPlayerId = c(NA_integer_, 2001L + seq_len(filler_n) - 1L, NA_integer_, NA_integer_, NA_integer_),
    playerId = rep(NA_integer_, n),
    stringsAsFactors = FALSE
  )

  html_rows <- data.frame(
    htmlEventNumber = seq_len(n),
    period = c(rep(1L, filler_n), 3L, 3L, 3L, 3L),
    strengthCodeHtml = rep("EV", n),
    secondsElapsedInPeriod = c(seq_len(filler_n), 765L, 765L, 766L, 766L),
    htmlEventCode = c(rep("FAC", filler_n), rep("SHOT", 4L)),
    typeDescKey = c(rep("faceoff", filler_n), rep("shot-on-goal", 4L)),
    description = "",
    ownerTeamId = c(rep(10L, filler_n), rep(6L, 4L)),
    primaryPlayerId = c(1001L + seq_len(filler_n) - 1L, rep(8474625L, 4L)),
    secondaryPlayerId = c(2001L + seq_len(filler_n) - 1L, rep(NA_integer_, 4L)),
    tertiaryPlayerId = rep(NA_integer_, n),
    homeGoaliePlayerId = rep(10L, n),
    awayGoaliePlayerId = rep(20L, n),
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- rep(list(101:105), n)
  html_rows$awaySkaterPlayerIds <- rep(list(201:205), n)

  matched <- .match_html_pbp_to_api(play_by_play, html_rows)

  expect_equal(nrow(matched), n)
  shot_api_idx <- matched$apiIndex[match((filler_n + 1L):n, matched$htmlEventNumber)]
  expect_equal(shot_api_idx[1:2], c(29L, 1L))
  expect_equal(sort(shot_api_idx), c(1L, 28L, 29L, 30L))
})

test_that("HTML PBP matcher does not discard candidates when API actors are NA", {
  play_by_play <- data.frame(
    gameId = c(1L, 1L),
    eventId = c(10L, 11L),
    periodNumber = c(2L, 2L),
    secondsElapsedInPeriod = c(1167L, 1167L),
    sortOrder = c(10L, 11L),
    eventTypeDescKey = c("penalty", "penalty"),
    eventOwnerTeamId = c(9L, 9L),
    isHome = c(FALSE, FALSE),
    situationCode = c("1551", "1551"),
    committedByPlayerId = c(8476999L, 8482092L),
    drawnByPlayerId = c(NA_integer_, 8480796L),
    playerId = c(NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = c(218L, 222L),
    period = c(2L, 2L),
    strengthCodeHtml = c("EV", "EV"),
    secondsElapsedInPeriod = c(1167L, 1167L),
    htmlEventCode = c("PENL", "PENL"),
    typeDescKey = c("penalty", "penalty"),
    description = c(
      "OTT #35 ULLMARK Goalie leave crease(2 min) Served By: #21 COUSINS, Off. Zone",
      "OTT #71 GREIG Roughing(2 min), Off. Zone Drawn By: WSH #42 FEHERVARY"
    ),
    ownerTeamId = c(9L, 9L),
    primaryPlayerId = c(8476999L, 8482092L),
    secondaryPlayerId = c(8481656L, 8480796L),
    tertiaryPlayerId = c(NA_integer_, NA_integer_),
    homeGoaliePlayerId = c(1L, 1L),
    awayGoaliePlayerId = c(2L, 2L),
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(c(101L, 102L, 103L, 104L, 105L), c(101L, 102L, 103L, 104L, 105L))
  html_rows$awaySkaterPlayerIds <- list(c(201L, 202L, 203L, 204L, 205L), c(201L, 202L, 203L, 204L, 205L))

  matched <- .match_html_pbp_to_api(play_by_play, html_rows)

  expect_equal(matched$apiIndex[match(c(218L, 222L), matched$htmlEventNumber)], c(1L, 2L))
})

test_that("HTML on-ice enrichment keeps penalty on-ice rows and rescues matched count mismatches", {
  play_by_play <- data.frame(
    gameId = rep(1L, 3L),
    eventId = c(10L, 11L, 12L),
    periodNumber = c(1L, 1L, 1L),
    secondsElapsedInPeriod = c(100L, 101L, 102L),
    sortOrder = c(10L, 11L, 12L),
    eventTypeDescKey = c("penalty", "shot-on-goal", "faceoff"),
    eventOwnerTeamId = c(8L, 8L, 10L),
    isHome = c(FALSE, FALSE, TRUE),
    situationCode = c("1551", "1451", "1551"),
    committedByPlayerId = c(803L, NA_integer_, NA_integer_),
    drawnByPlayerId = c(1001L, NA_integer_, NA_integer_),
    shootingPlayerId = c(NA_integer_, 803L, NA_integer_),
    winningPlayerId = c(NA_integer_, NA_integer_, 1001L),
    losingPlayerId = c(NA_integer_, NA_integer_, 801L),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = c(1L, 2L, 3L),
    period = c(1L, 1L, 1L),
    strengthCodeHtml = c("EV", "EV", "EV"),
    secondsElapsedInPeriod = c(100L, 101L, 102L),
    htmlEventCode = c("PENL", "SHOT", "FAC"),
    typeDescKey = c("penalty", "shot-on-goal", "faceoff"),
    description = c(
      "MTL #76 SUBBAN Hooking(2 min), Drawn By: TOR #37 BRENT",
      "MTL ONGOAL - #76 SUBBAN, Wrist, Off. Zone",
      "TOR won Neu. Zone - TOR #37 BRENT vs MTL #11 GOMEZ"
    ),
    ownerTeamId = c(8L, 8L, 10L),
    primaryPlayerId = c(803L, 803L, 1001L),
    secondaryPlayerId = c(1001L, NA_integer_, 801L),
    tertiaryPlayerId = c(NA_integer_, NA_integer_, NA_integer_),
    homeGoaliePlayerId = c(1002L, 1002L, 1002L),
    awayGoaliePlayerId = c(802L, 802L, 802L),
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(
    c(1001L, 1003L, 1004L, 1005L, 1006L),
    c(1001L, 1003L, 1004L, 1005L, 1006L),
    c(1001L, 1003L, 1004L, 1005L, 1006L)
  )
  html_rows$awaySkaterPlayerIds <- list(
    c(801L, 803L, 804L, 805L, 806L),
    c(801L, 803L, 804L, 805L, 806L),
    c(801L, 803L, 804L, 805L, 806L)
  )

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeGoaliePlayerId[1], 1002L)
  expect_equal(out$awayGoaliePlayerId[1], 802L)
  expect_equal(out$homeSkater1PlayerId[1], 1001L)
  expect_equal(out$awaySkater1PlayerId[1], 801L)
  expect_equal(out$homeGoaliePlayerId[2], 1002L)
  expect_equal(out$awayGoaliePlayerId[2], 802L)
  expect_equal(out$awaySkater2PlayerId[2], 803L)
  expect_equal(out$homeGoaliePlayerId[3], 1002L)
  expect_equal(out$awayGoaliePlayerId[3], 802L)
  expect_equal(out$homeSkater1PlayerId[3], 1001L)
  expect_equal(out$awaySkater1PlayerId[3], 801L)
})

test_that("drop_illogical_ordered_events moves the opening faceoff ahead of later plays", {
  play_by_play <- data.frame(
    gameId = rep(1L, 4L),
    eventId = c(1L, 2L, 3L, 4L),
    sortOrder = c(5L, 6L, 7L, 8L),
    periodNumber = rep(1L, 4L),
    secondsElapsedInPeriod = c(0L, 34L, 0L, 40L),
    eventTypeDescKey = c("period-start", "blocked-shot", "faceoff", "shot-on-goal"),
    stringsAsFactors = FALSE
  )

  out <- .drop_illogical_ordered_events(play_by_play)

  expect_equal(out$eventId, c(1L, 3L, 2L, 4L))
  expect_equal(out$sortOrder, c(5L, 6L, 7L, 8L))
})

test_that("HTML on-ice enrichment populates one-on-one shootout rows", {
  play_by_play <- data.frame(
    gameId = rep(1L, 2L),
    eventId = c(10L, 11L),
    periodNumber = c(5L, 5L),
    secondsElapsedInPeriod = c(0L, 0L),
    sortOrder = c(10L, 11L),
    eventTypeDescKey = c("shot-on-goal", "goal"),
    eventOwnerTeamId = c(10L, 8L),
    isHome = c(TRUE, FALSE),
    situationCode = c("1010", "0101"),
    shootingPlayerId = c(910L, 920L),
    scoringPlayerId = c(NA_integer_, 920L),
    goalieInNetId = c(NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = c(1L, 2L),
    period = c(5L, 5L),
    strengthCodeHtml = c("SO", "SO"),
    secondsElapsedInPeriod = c(0L, 0L),
    htmlEventCode = c("SHOT", "GOAL"),
    typeDescKey = c("shot-on-goal", "goal"),
    description = c("HOME shooter", "AWAY scorer"),
    ownerTeamId = c(10L, 8L),
    primaryPlayerId = c(910L, 920L),
    secondaryPlayerId = c(NA_integer_, NA_integer_),
    tertiaryPlayerId = c(NA_integer_, NA_integer_),
    homeGoaliePlayerId = c(1001L, 1001L),
    awayGoaliePlayerId = c(2001L, 2001L),
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(integer(), integer())
  html_rows$awaySkaterPlayerIds <- list(integer(), integer())

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_true(is.na(out$homeGoaliePlayerId[1]))
  expect_equal(out$awayGoaliePlayerId[1], 2001L)
  expect_equal(out$homeSkater1PlayerId[1], 910L)
  expect_true(is.na(out$awaySkater1PlayerId[1]))
  expect_true(is.na(out$goaliePlayerIdFor[1]))
  expect_equal(out$goaliePlayerIdAgainst[1], 2001L)
  expect_equal(out$skater1PlayerIdFor[1], 910L)
  expect_true(is.na(out$skater1PlayerIdAgainst[1]))

  expect_equal(out$homeGoaliePlayerId[2], 1001L)
  expect_true(is.na(out$awayGoaliePlayerId[2]))
  expect_true(is.na(out$homeSkater1PlayerId[2]))
  expect_equal(out$awaySkater1PlayerId[2], 920L)
  expect_true(is.na(out$goaliePlayerIdFor[2]))
  expect_equal(out$goaliePlayerIdAgainst[2], 1001L)
  expect_equal(out$skater1PlayerIdFor[2], 920L)
  expect_true(is.na(out$skater1PlayerIdAgainst[2]))
})

test_that("HTML on-ice enrichment accepts penalty-reconstructed skater counts", {
  play_by_play <- data.frame(
    gameId = rep(1L, 2L),
    gameTypeId = rep(2L, 2L),
    eventId = c(10L, 11L),
    periodNumber = c(1L, 1L),
    secondsElapsedInPeriod = c(51L, 56L),
    secondsElapsedInGame = c(51L, 56L),
    sortOrder = c(10L, 11L),
    eventTypeDescKey = c("penalty", "shot-on-goal"),
    eventOwnerTeamId = c(8L, 10L),
    isHome = c(FALSE, TRUE),
    situationCode = c("1551", "1551"),
    homeIsEmptyNet = c(FALSE, FALSE),
    awayIsEmptyNet = c(FALSE, FALSE),
    homeSkaterCount = c(5L, 5L),
    awaySkaterCount = c(5L, 5L),
    isEmptyNetFor = c(FALSE, FALSE),
    isEmptyNetAgainst = c(FALSE, FALSE),
    skaterCountFor = c(5L, 5L),
    skaterCountAgainst = c(5L, 5L),
    manDifferential = c(0L, 0L),
    strengthState = c("even-strength", "even-strength"),
    penaltyTypeCode = c("MIN", NA_character_),
    duration = c(2L, NA_integer_),
    committedByPlayerId = c(801L, NA_integer_),
    drawnByPlayerId = c(1001L, NA_integer_),
    shootingPlayerId = c(NA_integer_, 1002L),
    scoringPlayerId = c(NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = 1L,
    period = 1L,
    strengthCodeHtml = "PP",
    secondsElapsedInPeriod = 56L,
    htmlEventCode = "SHOT",
    typeDescKey = "shot-on-goal",
    description = "HOME shot",
    ownerTeamId = 10L,
    primaryPlayerId = 1002L,
    secondaryPlayerId = NA_integer_,
    tertiaryPlayerId = NA_integer_,
    homeGoaliePlayerId = 1001L,
    awayGoaliePlayerId = 2001L,
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(1002:1006)
  html_rows$awaySkaterPlayerIds <- list(2002:2005)

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeSkaterCount[2], 5L)
  expect_equal(out$awaySkaterCount[2], 4L)
  expect_false(out$homeIsEmptyNet[2])
  expect_false(out$awayIsEmptyNet[2])
  expect_equal(out$manDifferential[2], 1L)
  expect_equal(out$strengthState[2], "power-play")
  expect_equal(out$homeGoaliePlayerId[2], 1001L)
  expect_equal(out$awayGoaliePlayerId[2], 2001L)
  expect_equal(out$homeSkater1PlayerId[2], 1002L)
  expect_equal(out$awaySkater4PlayerId[2], 2005L)
})

test_that("HTML on-ice enrichment accepts delayed-penalty shots before the call is assessed", {
  play_by_play <- data.frame(
    gameId = rep(1L, 2L),
    gameTypeId = rep(2L, 2L),
    eventId = c(10L, 11L),
    periodNumber = c(1L, 1L),
    secondsElapsedInPeriod = c(720L, 722L),
    secondsElapsedInGame = c(720L, 722L),
    sortOrder = c(10L, 11L),
    eventTypeDescKey = c("shot-on-goal", "penalty"),
    eventOwnerTeamId = c(8L, 10L),
    isHome = c(FALSE, TRUE),
    situationCode = c("0651", "0651"),
    homeIsEmptyNet = c(FALSE, FALSE),
    awayIsEmptyNet = c(TRUE, TRUE),
    homeSkaterCount = c(5L, 5L),
    awaySkaterCount = c(6L, 6L),
    isEmptyNetFor = c(TRUE, FALSE),
    isEmptyNetAgainst = c(FALSE, TRUE),
    skaterCountFor = c(6L, 5L),
    skaterCountAgainst = c(5L, 6L),
    manDifferential = c(0L, -1L),
    strengthState = c("even-strength", "penalty-kill"),
    penaltyTypeCode = c(NA_character_, "MIN"),
    duration = c(NA_integer_, 2L),
    committedByPlayerId = c(NA_integer_, 1003L),
    drawnByPlayerId = c(NA_integer_, 2003L),
    shootingPlayerId = c(2002L, NA_integer_),
    scoringPlayerId = c(NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = 1L,
    period = 1L,
    strengthCodeHtml = "EV",
    secondsElapsedInPeriod = 720L,
    htmlEventCode = "SHOT",
    typeDescKey = "shot-on-goal",
    description = "AWAY shot",
    ownerTeamId = 8L,
    primaryPlayerId = 2002L,
    secondaryPlayerId = NA_integer_,
    tertiaryPlayerId = NA_integer_,
    homeGoaliePlayerId = 1001L,
    awayGoaliePlayerId = 2001L,
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(1002:1006)
  html_rows$awaySkaterPlayerIds <- list(2002:2006)

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeSkaterCount[1], 5L)
  expect_equal(out$awaySkaterCount[1], 5L)
  expect_false(out$homeIsEmptyNet[1])
  expect_false(out$awayIsEmptyNet[1])
  expect_equal(out$manDifferential[1], 0L)
  expect_equal(out$strengthState[1], "even-strength")
  expect_equal(out$homeGoaliePlayerId[1], 1001L)
  expect_equal(out$awayGoaliePlayerId[1], 2001L)
  expect_equal(out$awaySkater1PlayerId[1], 2002L)
  expect_equal(out$homeSkater5PlayerId[1], 1006L)
})

test_that("HTML on-ice enrichment accepts late empty-net pulls for the trailing team", {
  play_by_play <- data.frame(
    gameId = 1L,
    gameTypeId = 2L,
    eventId = 10L,
    periodNumber = 3L,
    secondsElapsedInPeriod = 1190L,
    secondsElapsedInGame = 3590L,
    sortOrder = 10L,
    eventTypeDescKey = "shot-on-goal",
    eventOwnerTeamId = 8L,
    isHome = FALSE,
    situationCode = "1551",
    homeIsEmptyNet = FALSE,
    awayIsEmptyNet = FALSE,
    homeSkaterCount = 5L,
    awaySkaterCount = 5L,
    isEmptyNetFor = FALSE,
    isEmptyNetAgainst = FALSE,
    skaterCountFor = 5L,
    skaterCountAgainst = 5L,
    manDifferential = 0L,
    strengthState = "even-strength",
    homeGoals = 3L,
    awayGoals = 2L,
    shootingPlayerId = 2002L,
    scoringPlayerId = NA_integer_,
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = 1L,
    period = 3L,
    strengthCodeHtml = "6v5",
    secondsElapsedInPeriod = 1190L,
    htmlEventCode = "SHOT",
    typeDescKey = "shot-on-goal",
    description = "AWAY extra-attacker shot",
    ownerTeamId = 8L,
    primaryPlayerId = 2002L,
    secondaryPlayerId = NA_integer_,
    tertiaryPlayerId = NA_integer_,
    homeGoaliePlayerId = 1001L,
    awayGoaliePlayerId = NA_integer_,
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(1002:1006)
  html_rows$awaySkaterPlayerIds <- list(2002:2007)

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeSkaterCount, 5L)
  expect_equal(out$awaySkaterCount, 6L)
  expect_false(out$homeIsEmptyNet)
  expect_true(out$awayIsEmptyNet)
  expect_equal(out$manDifferential, 0L)
  expect_equal(out$homeGoaliePlayerId, 1001L)
  expect_true(is.na(out$awayGoaliePlayerId))
  expect_equal(out$awaySkater1PlayerId, 2002L)
  expect_equal(out$awaySkater6PlayerId, 2007L)
})

test_that("HTML on-ice enrichment backfills unmatched delayed-penalty rows from prior state", {
  play_by_play <- data.frame(
    gameId = rep(1L, 2L),
    gameTypeId = rep(2L, 2L),
    eventId = c(10L, 11L),
    periodNumber = c(1L, 1L),
    secondsElapsedInPeriod = c(100L, 108L),
    secondsElapsedInGame = c(100L, 108L),
    sortOrder = c(10L, 11L),
    eventTypeDescKey = c("faceoff", "delayed-penalty"),
    eventOwnerTeamId = c(10L, 8L),
    isHome = c(TRUE, FALSE),
    situationCode = c("1551", "1551"),
    homeIsEmptyNet = c(FALSE, FALSE),
    awayIsEmptyNet = c(FALSE, FALSE),
    homeSkaterCount = c(5L, 5L),
    awaySkaterCount = c(5L, 5L),
    isEmptyNetFor = c(FALSE, FALSE),
    isEmptyNetAgainst = c(FALSE, FALSE),
    skaterCountFor = c(5L, 5L),
    skaterCountAgainst = c(5L, 5L),
    manDifferential = c(0L, 0L),
    strengthState = c("even-strength", "even-strength"),
    winningPlayerId = c(1002L, NA_integer_),
    losingPlayerId = c(2002L, NA_integer_),
    committedByPlayerId = c(NA_integer_, NA_integer_),
    drawnByPlayerId = c(NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = 1L,
    period = 1L,
    strengthCodeHtml = "EV",
    secondsElapsedInPeriod = 100L,
    htmlEventCode = "FAC",
    typeDescKey = "faceoff",
    description = "HOME faceoff",
    ownerTeamId = 10L,
    primaryPlayerId = 1002L,
    secondaryPlayerId = 2002L,
    tertiaryPlayerId = NA_integer_,
    homeGoaliePlayerId = 1001L,
    awayGoaliePlayerId = 2001L,
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(1002:1006)
  html_rows$awaySkaterPlayerIds <- list(2002:2006)

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeGoaliePlayerId[2], 1001L)
  expect_equal(out$awayGoaliePlayerId[2], 2001L)
  expect_equal(out$homeSkater1PlayerId[2], 1002L)
  expect_equal(out$awaySkater5PlayerId[2], 2006L)
  expect_equal(out$homeSkaterCount[2], 5L)
  expect_equal(out$awaySkaterCount[2], 5L)
})

test_that("actor-based HTML on-ice rescues update derived strength context when HTML is plausible", {
  play_by_play <- data.frame(
    gameId = 1L,
    gameTypeId = 2L,
    eventId = 10L,
    periodNumber = 4L,
    secondsElapsedInPeriod = 5L,
    secondsElapsedInGame = 3605L,
    sortOrder = 10L,
    eventTypeDescKey = "shot-on-goal",
    eventOwnerTeamId = 10L,
    isHome = TRUE,
    situationCode = "1451",
    homeIsEmptyNet = FALSE,
    awayIsEmptyNet = FALSE,
    homeSkaterCount = 5L,
    awaySkaterCount = 4L,
    isEmptyNetFor = FALSE,
    isEmptyNetAgainst = FALSE,
    skaterCountFor = 5L,
    skaterCountAgainst = 4L,
    manDifferential = 1L,
    strengthState = "power-play",
    shootingPlayerId = 1002L,
    scoringPlayerId = NA_integer_,
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = 1L,
    period = 4L,
    strengthCodeHtml = "EV",
    secondsElapsedInPeriod = 5L,
    htmlEventCode = "SHOT",
    typeDescKey = "shot-on-goal",
    description = "HOME shot",
    ownerTeamId = 10L,
    primaryPlayerId = 1002L,
    secondaryPlayerId = NA_integer_,
    tertiaryPlayerId = NA_integer_,
    homeGoaliePlayerId = 1001L,
    awayGoaliePlayerId = 2001L,
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(1002:1006)
  html_rows$awaySkaterPlayerIds <- list(2002:2006)

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeGoaliePlayerId, 1001L)
  expect_equal(out$awayGoaliePlayerId, 2001L)
  expect_equal(out$awaySkater5PlayerId, 2006L)
  expect_equal(out$homeSkaterCount, 5L)
  expect_equal(out$awaySkaterCount, 5L)
  expect_equal(out$manDifferential, 0L)
  expect_equal(out$strengthState, "even-strength")
})

test_that("HTML on-ice enrichment preserves seven-skater overflow rows", {
  play_by_play <- data.frame(
    gameId = 1L,
    gameTypeId = 2L,
    eventId = 10L,
    periodNumber = 2L,
    secondsElapsedInPeriod = 209L,
    secondsElapsedInGame = 1409L,
    sortOrder = 10L,
    eventTypeDescKey = "shot-on-goal",
    eventOwnerTeamId = 10L,
    isHome = TRUE,
    situationCode = "1551",
    homeIsEmptyNet = FALSE,
    awayIsEmptyNet = FALSE,
    homeSkaterCount = 5L,
    awaySkaterCount = 5L,
    isEmptyNetFor = FALSE,
    isEmptyNetAgainst = FALSE,
    skaterCountFor = 5L,
    skaterCountAgainst = 5L,
    manDifferential = 0L,
    strengthState = "even-strength",
    shootingPlayerId = 1002L,
    scoringPlayerId = NA_integer_,
    stringsAsFactors = FALSE
  )
  html_rows <- data.frame(
    htmlEventNumber = 1L,
    period = 2L,
    strengthCodeHtml = "EV",
    secondsElapsedInPeriod = 209L,
    htmlEventCode = "SHOT",
    typeDescKey = "shot-on-goal",
    description = "HOME shot",
    ownerTeamId = 10L,
    primaryPlayerId = 1002L,
    secondaryPlayerId = NA_integer_,
    tertiaryPlayerId = NA_integer_,
    homeGoaliePlayerId = 1001L,
    awayGoaliePlayerId = 2001L,
    stringsAsFactors = FALSE
  )
  html_rows$homeSkaterPlayerIds <- list(1002:1009)
  html_rows$awaySkaterPlayerIds <- list(2002:2006)

  local_mocked_bindings(
    .fetch_html_pbp_on_ice = function(...) html_rows,
    .package = "nhlscraper"
  )

  out <- .add_html_on_ice_players(
    play_by_play,
    game = 1L,
    rosters = data.frame(),
    home_team = list(id = 10L, abbrev = "TOR"),
    away_team = list(id = 8L, abbrev = "MTL")
  )

  expect_equal(out$homeSkater7PlayerId, 1008L)
  expect_equal(out$homeSkater8PlayerId, 1009L)
  expect_equal(out$skater7PlayerIdFor, 1008L)
  expect_equal(out$skater8PlayerIdFor, 1009L)
  expect_equal(out$homeSkaterCount, 5L)
  expect_equal(out$awaySkaterCount, 5L)
})

test_that("strip_situation_code preserves original values and available counts", {
  play_by_play <- data.frame(
    gameId = c(1L, 1L, 1L),
    eventId = c(1L, 2L, 3L),
    sortOrder = c(1L, 2L, 3L),
    secondsElapsedInGame = c(1L, 2L, 3L),
    eventTypeDescKey = c("period-start", "penalty", "shot-on-goal"),
    eventOwnerTeamId = c(NA_integer_, 8L, 10L),
    isHome = c(NA, FALSE, TRUE),
    situationCode = c(NA_character_, "0651", "1451"),
    stringsAsFactors = FALSE
  )

  out <- .strip_situation_code(play_by_play)

  expect_identical(out$situationCode, play_by_play$situationCode)
  expect_true(is.na(out$homeSkaterCount[1]))
  expect_equal(out$awaySkaterCount[2], 6L)
  expect_equal(out$homeSkaterCount[2], 5L)
  expect_equal(out$awaySkaterCount[3], 4L)
  expect_equal(out$homeSkaterCount[3], 5L)
})

test_that("add_shift_times() populates scalar goalie and skater timing columns", {
  play_by_play <- data.frame(
    gameId = c(1L, 1L),
    periodNumber = c(1L, 1L),
    secondsElapsedInPeriod = c(40L, 80L),
    eventTypeDescKey = c("shot-on-goal", "hit"),
    isHome = c(TRUE, FALSE),
    homeGoaliePlayerId = c(101L, 101L),
    awayGoaliePlayerId = c(201L, 201L),
    homeSkater1PlayerId = c(111L, 111L),
    awaySkater1PlayerId = c(211L, 211L),
    homeSkater2PlayerId = c(NA_integer_, NA_integer_),
    awaySkater2PlayerId = c(NA_integer_, NA_integer_),
    homeSkater3PlayerId = c(NA_integer_, NA_integer_),
    awaySkater3PlayerId = c(NA_integer_, NA_integer_),
    homeSkater4PlayerId = c(NA_integer_, NA_integer_),
    awaySkater4PlayerId = c(NA_integer_, NA_integer_),
    homeSkater5PlayerId = c(NA_integer_, NA_integer_),
    awaySkater5PlayerId = c(NA_integer_, NA_integer_),
    homeSkater6PlayerId = c(NA_integer_, NA_integer_),
    awaySkater6PlayerId = c(NA_integer_, NA_integer_),
    homeSkater7PlayerId = c(117L, 117L),
    awaySkater7PlayerId = c(217L, 217L),
    homeSkater8PlayerId = c(118L, 118L),
    awaySkater8PlayerId = c(218L, 218L),
    stringsAsFactors = FALSE
  )
  shift_data <- data.frame(
    gameId = rep(1L, 10L),
    teamId = c(1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L),
    playerId = c(101L, 111L, 111L, 117L, 118L, 201L, 211L, 211L, 217L, 218L),
    period = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
    startSecondsElapsedInPeriod = c(0L, 0L, 70L, 0L, 0L, 0L, 0L, 70L, 0L, 0L),
    endSecondsElapsedInPeriod = c(120L, 50L, 120L, 120L, 120L, 120L, 60L, 120L, 120L, 120L),
    stringsAsFactors = FALSE
  )

  out <- add_shift_times(play_by_play, shift_data)

  last_id_col <- tail(intersect(
    .on_ice_id_scalar_column_names(play_by_play = out),
    names(out)
  ), 1L)
  timing_start <- match(last_id_col, names(out)) + 1L
  remaining_cols <- .on_ice_timing_scalar_column_names(
    "SecondsRemainingInShift",
    play_by_play = out
  )
  elapsed_cols <- .on_ice_timing_scalar_column_names(
    "SecondsElapsedInShift",
    play_by_play = out
  )
  expect_equal(
    names(out)[timing_start:(timing_start + length(remaining_cols) - 1L)],
    remaining_cols
  )
  expect_equal(out$homeGoalieSecondsRemainingInShift, c(80, 40))
  expect_equal(out$awayGoalieSecondsRemainingInShift, c(80, 40))
  expect_equal(out$homeSkater1SecondsRemainingInShift, c(10, 40))
  expect_equal(out$awaySkater1SecondsRemainingInShift, c(20, 40))
  expect_equal(out$homeSkater7SecondsRemainingInShift, c(80, 40))
  expect_equal(out$awaySkater7SecondsRemainingInShift, c(80, 40))
  expect_equal(out$homeSkater8SecondsRemainingInShift, c(80, 40))
  expect_equal(out$awaySkater8SecondsRemainingInShift, c(80, 40))
  expect_equal(out$goalieSecondsRemainingInShiftFor, c(80, 40))
  expect_equal(out$goalieSecondsRemainingInShiftAgainst, c(80, 40))
  expect_equal(out$skater1SecondsRemainingInShiftFor, c(10, 40))
  expect_equal(out$skater1SecondsRemainingInShiftAgainst, c(20, 40))
  expect_equal(out$skater7SecondsRemainingInShiftFor, c(80, 40))
  expect_equal(out$skater7SecondsRemainingInShiftAgainst, c(80, 40))
  expect_equal(out$skater8SecondsRemainingInShiftFor, c(80, 40))
  expect_equal(out$skater8SecondsRemainingInShiftAgainst, c(80, 40))
  expect_equal(
    names(out)[
      (timing_start + length(remaining_cols)):
        (timing_start + length(remaining_cols) + length(elapsed_cols) - 1L)
    ],
    elapsed_cols
  )
  expect_equal(out$homeGoalieSecondsElapsedInShift, c(40, 80))
  expect_equal(out$awayGoalieSecondsElapsedInShift, c(40, 80))
  expect_equal(out$homeSkater1SecondsElapsedInShift, c(40, 10))
  expect_equal(out$awaySkater1SecondsElapsedInShift, c(40, 10))
  expect_equal(out$homeSkater7SecondsElapsedInShift, c(40, 80))
  expect_equal(out$awaySkater7SecondsElapsedInShift, c(40, 80))
  expect_equal(out$homeSkater8SecondsElapsedInShift, c(40, 80))
  expect_equal(out$awaySkater8SecondsElapsedInShift, c(40, 80))
  expect_equal(out$homeSkater1SecondsElapsedInPeriodSinceLastShift, c(340, 30))
  expect_equal(out$awaySkater1SecondsElapsedInPeriodSinceLastShift, c(340, 20))
  expect_equal(out$goalieSecondsElapsedInShiftFor, c(40, 80))
  expect_equal(out$goalieSecondsElapsedInShiftAgainst, c(40, 80))
  expect_equal(out$skater1SecondsElapsedInShiftFor, c(40, 10))
  expect_equal(out$skater1SecondsElapsedInShiftAgainst, c(40, 10))
  expect_equal(out$skater7SecondsElapsedInShiftFor, c(40, 80))
  expect_equal(out$skater7SecondsElapsedInShiftAgainst, c(40, 80))
  expect_equal(out$skater8SecondsElapsedInShiftFor, c(40, 80))
  expect_equal(out$skater8SecondsElapsedInShiftAgainst, c(40, 80))
})

test_that("add_shift_times() handles multi-game aggregates", {
  play_by_play <- data.frame(
    gameId = c(1L, 2L),
    periodNumber = c(1L, 1L),
    secondsElapsedInPeriod = c(40L, 50L),
    eventTypeDescKey = c("shot-on-goal", "shot-on-goal"),
    isHome = c(TRUE, FALSE),
    homeGoaliePlayerId = c(101L, 301L),
    awayGoaliePlayerId = c(201L, 401L),
    homeSkater1PlayerId = c(111L, 311L),
    awaySkater1PlayerId = c(211L, 411L),
    homeSkater2PlayerId = c(NA_integer_, NA_integer_),
    awaySkater2PlayerId = c(NA_integer_, NA_integer_),
    homeSkater3PlayerId = c(NA_integer_, NA_integer_),
    awaySkater3PlayerId = c(NA_integer_, NA_integer_),
    homeSkater4PlayerId = c(NA_integer_, NA_integer_),
    awaySkater4PlayerId = c(NA_integer_, NA_integer_),
    homeSkater5PlayerId = c(NA_integer_, NA_integer_),
    awaySkater5PlayerId = c(NA_integer_, NA_integer_),
    homeSkater6PlayerId = c(NA_integer_, NA_integer_),
    awaySkater6PlayerId = c(NA_integer_, NA_integer_),
    homeSkater7PlayerId = c(NA_integer_, NA_integer_),
    awaySkater7PlayerId = c(NA_integer_, NA_integer_),
    stringsAsFactors = FALSE
  )
  shift_data <- data.frame(
    gameId = c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
    teamId = c(1L, 1L, 2L, 2L, 3L, 3L, 4L, 4L),
    playerId = c(101L, 111L, 201L, 211L, 301L, 311L, 401L, 411L),
    period = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
    startSecondsElapsedInPeriod = c(0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L),
    endSecondsElapsedInPeriod = c(120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L),
    stringsAsFactors = FALSE
  )

  out <- add_shift_times(play_by_play, shift_data)

  expect_equal(out$homeGoalieSecondsRemainingInShift, c(80, 70))
  expect_equal(out$awayGoalieSecondsRemainingInShift, c(80, 70))
  expect_equal(out$goalieSecondsRemainingInShiftFor, c(80, 70))
  expect_equal(out$goalieSecondsRemainingInShiftAgainst, c(80, 70))
  expect_equal(out$homeGoalieSecondsElapsedInShift, c(40, 50))
  expect_equal(out$awayGoalieSecondsElapsedInShift, c(40, 50))
  expect_equal(out$goalieSecondsElapsedInShiftFor, c(40, 50))
  expect_equal(out$goalieSecondsElapsedInShiftAgainst, c(40, 50))
})

test_that("public helpers require the new public schema names", {
  legacy_speed <- data.frame(
    gameId = 1L,
    eventId = 1L,
    sortOrder = 1L,
    secondsElapsedInGame = 1L,
    typeDescKey = "shot-on-goal",
    situationCode = "1551",
    xCoordNorm = 0,
    yCoordNorm = 0,
    distance = 10,
    angle = 0,
    stringsAsFactors = FALSE
  )
  expect_error(add_deltas(legacy_speed), "eventTypeDescKey")

  legacy_goalie <- data.frame(
    gameId = 1L,
    goalieInNetId = 99L,
    stringsAsFactors = FALSE
  )
  expect_no_error(add_goalie_biometrics(legacy_goalie))
})

test_that("goalie biometrics prefer goalieInNetId over goaliePlayerIdAgainst", {
  play_by_play <- data.frame(
    gameId = 1L,
    goalieInNetId = 99L,
    goaliePlayerIdAgainst = 88L,
    stringsAsFactors = FALSE
  )
  local_mocked_bindings(
    players = function() data.frame(
      playerId = c(88L, 99L),
      height = c(72L, 74L),
      weight = c(180L, 190L),
      handCode = c("R", "L"),
      birthDate = c("1991-01-15", "1990-01-15"),
      stringsAsFactors = FALSE
    ),
    games = function() data.frame(
      gameId = 1L,
      gameDate = "2020-02-01",
      stringsAsFactors = FALSE
    ),
    .package = "nhlscraper"
  )

  out <- add_goalie_biometrics(play_by_play)
  expect_equal(out$goalieHeight, 74L)
  expect_equal(out$goalieWeight, 190L)
  expect_equal(out$goalieHandCode, "L")
  expect_equal(out$goalieAge, 30L)
})

test_that(".shot_event_mask() includes short missed shots", {
  play_by_play <- data.frame(
    eventTypeDescKey = c("missed-shot", "shot-on-goal", "blocked-shot"),
    reason = c("short", NA, NA),
    stringsAsFactors = FALSE
  )

  out <- nhlscraper:::.shot_event_mask(
    play_by_play,
    c("missed-shot", "shot-on-goal")
  )

  expect_equal(out, c(TRUE, TRUE, FALSE))
})

test_that(".count_goals_shots() includes short missed shots in Fenwick and Corsi", {
  play_by_play <- data.frame(
    gameId = rep(1L, 4L),
    eventId = 1:4,
    sortOrder = 1:4,
    isHome = c(TRUE, TRUE, TRUE, FALSE),
    eventTypeDescKey = c("missed-shot", "shot-on-goal", "missed-shot", "blocked-shot"),
    reason = c("short", NA, "wide", NA),
    createdRebound = c(FALSE, FALSE, FALSE, FALSE),
    marker = 1:4,
    stringsAsFactors = FALSE
  )

  out <- nhlscraper:::.count_goals_shots(play_by_play)

  expect_equal(out$reason[1], "short")
  expect_equal(out$homeFenwick, c(0L, 1L, 2L, 3L))
  expect_equal(out$homeCorsi, c(0L, 1L, 2L, 3L))
  expect_equal(out$awayCorsi, c(0L, 0L, 0L, 0L))
})

test_that(".flag_is_rebound() treats short missed shots as rebound sources", {
  play_by_play <- data.frame(
    gameId = rep(1L, 2L),
    eventId = 1:2,
    sortOrder = 1:2,
    secondsElapsedInGame = c(10L, 12L),
    eventTypeDescKey = c("missed-shot", "shot-on-goal"),
    reason = c("short", NA),
    situationCode = c("1551", "1551"),
    eventOwnerTeamId = c(10L, 10L),
    isRush = c(NA, FALSE),
    marker = 1:2,
    stringsAsFactors = FALSE
  )

  out <- nhlscraper:::.flag_is_rebound(play_by_play)

  expect_false(out$isRebound[1])
  expect_true(out$createdRebound[1])
  expect_true(out$isRebound[2])
  expect_false(out$createdRebound[2])
})

test_that("calculate_expected_goals() scores short missed shots with xG", {
  local_mocked_bindings(
    .xg_warn_ignored_model = function(model, fn_name) invisible(NULL),
    .xg_prepare_play_by_play = function(play_by_play) play_by_play,
    .xg_build_model_frame = function(shots, play_by_play) {
      shots$shootingPlayerId <- rep(NA_integer_, nrow(shots))
      shots
    },
    .xg_partition_shots = function(shots) rep("sd", nrow(shots)),
    .xg_score_partition = function(df, spec) rep(0.2, nrow(df)),
    .package = "nhlscraper"
  )

  play_by_play <- data.frame(
    gameId = rep(1L, 3L),
    eventId = 1:3,
    sortOrder = 1:3,
    eventTypeDescKey = c("missed-shot", "missed-shot", "shot-on-goal"),
    goaliePlayerIdAgainst = c(NA_integer_, NA_integer_, NA_integer_),
    reason = c("short", "wide", NA),
    stringsAsFactors = FALSE
  )

  out <- calculate_expected_goals(play_by_play)

  expect_true(is.finite(out$xG[1]) && out$xG[1] > 0)
  expect_true(is.finite(out$xG[2]) && out$xG[2] > 0)
  expect_true(is.finite(out$xG[3]) && out$xG[3] > 0)
})

#' Access all the ESPN athletes (players)
#' 
#' `get_espn_athletes()` is deprecated. Use [espn_players()] instead.
#' 
#' @returns data.frame with one row per ESPN athlete (player)
#' @export

get_espn_athletes <- function() {
  .Deprecated(
    new     = 'espn_players()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_athletes()` is deprecated.',
      'Use `espn_players()` instead.'
    )
  )
  espn_players()
}

#' Access the ESPN summary for an athlete (player) and season
#' 
#' `get_espn_athlete()` is defunct. Use [espn_player_summary()] instead.
#' 
#' @export

get_espn_athlete <- function() {
  .Defunct(
    new     = 'espn_player_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_athlete()` is defunct.',
      'Use `espn_player_summary()` instead.'
    )
  )
}

#' Access all the players
#'
#' `players()` retrieves all the players as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player
#' @examples
#' # May take >5s, so skip.
#' \donttest{all_players <- players()}
#' @export

players <- function() {
  tryCatch({
    players <- nhl_api(
      path = 'player',
      type = 'r'
    )$data
    players <- players[order(players$id), ]
    names(players)[names(players) == 'id']            <- 'playerId'
    names(players)[names(players) == 'addNames']      <- 'playerAddNames'
    names(players)[names(players) == 'firstName']     <- 'playerFirstName'
    names(players)[names(players) == 'fullName']      <- 'playerFullName'
    names(players)[names(players) == 'lastName']      <- 'playerLastName'
    names(players)[names(players) == 'middleName']    <- 'playerMiddleName'
    names(players)[names(players) == 'position']      <- 'positionCode'
    names(players)[names(players) == 'prName']        <- 'playerPrName'
    names(players)[names(players) == 'shootsCatches'] <- 'handCode'
    names(players)[names(players) == 'centralRegistryPosition'] <- 'centralRegistryPositionCode'
    players
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the season(s) and game type(s) in which a player played
#'
#' `player_seasons()` retrieves the season(s) and game type(s) in which a player played as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context.
#'
#' @param player integer ID (e.g., 8480039); see [players()] for reference
#'
#' @returns data.frame with one row per season
#' @examples
#' Martin_Necas_seasons <- player_seasons(player = 8480039)
#' @export

player_seasons <- function(player = 8478402) {
  tryCatch(
    expr = {
      seasons <- nhl_api(
        path = sprintf('v1/player/%s/game-log/now', player),
        type = 'w'
      )$playerStatsSeasons
      seasons[[1]]
      names(seasons)[names(seasons) == 'season']    <- 'seasonId'
      names(seasons)[names(seasons) == 'gameTypes'] <- 'gameTypeIds'
      seasons
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the summary for a player
#'
#' `player_summary()` retrieves the summary for a player as a nested `list` that separates summary and detail blocks for player identity, role, handedness, and biographical profile.
#'
#' @inheritParams player_seasons
#'
#' @returns list with various items
#' @examples
#' Martin_Necas_summary <- player_summary(player = 8480039)
#' @export

player_summary <- function(player = 8478402) {
  tryCatch(
    expr = {
      summary <- nhl_api(
        path = sprintf('v1/player/%s/landing', player),
        type = 'w'
      )
      names(summary)[names(summary) == 'firstName']    <- 'playerFirstName'
      names(summary)[names(summary) == 'lastName']     <- 'playerLastName'
      names(summary)[names(summary) == 'position']     <- 'positionCode'
      names(summary)[names(summary) == 'fullTeamName'] <- 'teamFullName'
      names(summary) <- normalize_team_abbrev_cols(names(summary))
      summary
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the game log for a player, season, and game type
#'
#' `player_game_log()` retrieves the game log for a player, season, and game type as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow plus production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams player_seasons
#' @inheritParams roster_statistics
#'
#' @returns data.frame with one row per game
#' @examples
#' Martin_Necas_game_log_regular_20242025 <- player_game_log(
#'   player    = 8480039,
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

player_game_log <- function(
  player    = 8478402, 
  season    = 'now', 
  game_type = ''
) {
  tryCatch(
    expr = {
      log <- nhl_api(
        path = sprintf(
          'v1/player/%s/game-log/%s/%s', 
          player, 
          season, 
          game_type
        ),
        type = 'w'
      )
      log$playerStatsSeasons[0, ]
      out <- log$gameLog
      names(out) <- normalize_locale_names(names(out))
      names(out) <- normalize_team_abbrev_cols(names(out))
      out
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the spotlight players
#'
#' `spotlight_players()` retrieves the spotlight players as a `data.frame` where each row represents player and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player
#' @examples
#' spotlight_players <- spotlight_players()
#' @export

spotlight_players <- function() {
  tryCatch({
    players <- nhl_api(
      path = 'v1/player-spotlight',
      type = 'w'
    )
    names(players)[names(players) == 'position'] <- 'positionCode'
    names(players) <- normalize_locale_names(names(players))
    names(players) <- scope_person_name_cols(names(players), 'player')
    players
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

test_that("contracts() returns non-empty data.frame", {
  skip_if_offline()
  test <- contracts()
  expect_true(is.data.frame(test) && nrow(test) > 0)
  expect_false(any(c("twoYearCash", "threeYearCash") %in% names(test)))
})

test_that("contracts() returns matched player name and position from players()", {
  ns <- asNamespace("nhlscraper")
  old_locked <- bindingIsLocked(".contracts_base", ns)
  if (old_locked) {
    unlockBinding(".contracts_base", ns)
  }
  old_contracts_base <- get(".contracts_base", envir = ns, inherits = FALSE)
  assign(
    ".contracts_base",
    data.frame(
      playerFullName = "Jon Example",
      positionCode = "LW",
      teamId = 5L,
      signedWithTeamId = 5L,
      signedWithTriCode = "ABC",
      ageAtSigning = 25L,
      startSeasonId = 20242025L,
      endSeasonId = 20252026L,
      contractYears = 2L,
      contractAAV = 1.5,
      contractValue = 3,
      signingBonus = 0.5,
      stringsAsFactors = FALSE
    ),
    envir = ns
  )
  on.exit({
    assign(".contracts_base", old_contracts_base, envir = ns)
    if (old_locked) {
      lockBinding(".contracts_base", ns)
    }
  }, add = TRUE)

  local_mocked_bindings(
    players = function() {
      data.frame(
        playerId = 42L,
        playerFullName = "Jonathan Example",
        playerFirstName = "Jon",
        playerLastName = "Example",
        positionCode = "L",
        birthDate = "1999-07-01",
        currentTeamId = 5L,
        careerTeamId = 5L,
        firstSignedByTeamId = 5L,
        lastNHLTeamId = 5L,
        onRoster = "Y",
        stringsAsFactors = FALSE
      )
    },
    .package = "nhlscraper"
  )

  test <- get("contracts", envir = ns)()

  expect_equal(test$playerId, 42L)
  expect_equal(test$playerFullName, "Jonathan Example")
  expect_equal(test$positionCode, "L")
})

#' Access the ESPN coaches for a season
#' 
#' `get_espn_coaches()` is defunct. Use [coaches()] instead.
#' 
#' @export

get_espn_coaches <- function() {
  .Defunct(
    new     = 'coaches()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_coaches()` is defunct.',
      'Use `coaches()` instead.'
    )
  )
}

#' Access the ESPN statistics for a coach and (multiple) season(s)
#' 
#' `get_espn_coach()` is defunct. Use [coach_career_statistics()] instead.
#'
#' @export

get_espn_coach <- function() {
  .Defunct(
    new     = 'coach_career_statistics()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_coach()` is defunct.',
      'Use `coach_career_statistics()` instead.'
    )
  )
}

#' Access the career ESPN statistics for a coach
#'
#' `get_espn_coach_career()` is defunct. Use [coach_career_statistics()] instead.
#'
#' @export

get_espn_coach_career <- function() {
  .Defunct(
    new     = 'coach_career_statistics()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_coach_career()` is defunct.',
      'Use `coach_career_statistics()` instead.'
    )
  )
}

#' Access the season(s) and game type(s) in which there exists skater EDGE 
#' statistics
#'
#' `skater_edge_seasons()` retrieves the season(s) and game type(s) in which there exists skater EDGE statistics as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @returns data.frame with one row per season
#' @examples
#' skater_EDGE_seasons <- skater_edge_seasons()
#' @export

skater_edge_seasons <- function() {
  tryCatch(
    expr = {
      seasons <- nhl_api(
        path = sprintf('v1/edge/skater-landing/now'),
        type = 'w'
      )$seasonsWithEdgeStats
      names(seasons)[names(seasons) == 'id']        <- 'seasonId'
      names(seasons)[names(seasons) == 'gameTypes'] <- 'gameTypeIds'
      seasons
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the skater EDGE statistics leaders for a season and game type
#'
#' `skater_edge_leaders()` retrieves the skater EDGE statistics leaders for a season and game type as a nested `list` that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
#'
#' @param season integer in YYYYYYYY (e.g., 20242025); see 
#' [skater_edge_seasons()] for reference
#' @param game_type integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season) OR character of 'pre', 'regular', or 
#' 'playoff'/'post'; see [skater_edge_seasons()] for reference; most functions 
#' will NOT support pre-season
#'
#' @returns list of various items
#' @examples
#' skater_EDGE_leaders_regular_20242025 <- skater_edge_leaders(
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

skater_edge_leaders <- function(season = 'now', game_type = '') {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf(
          'v1/edge/skater-landing/%s/%s', 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )$leaders
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the EDGE summary for a skater, season, and game type
#'
#' `skater_edge_summary()` retrieves the EDGE summary for a skater, season, and game type as a nested `list` that separates summary and detail blocks for player identity, role, handedness, and biographical profile plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams player_seasons
#' @inheritParams skater_edge_leaders
#'
#' @returns list of various items
#' @examples
#' Martin_Necas_EDGE_summary_regular_20242025 <- skater_edge_summary(
#'   player    = 8480039, 
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

skater_edge_summary <- function(
  player    = 8478402, 
  season    = 'now', 
  game_type = ''
) {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf(
          'v1/edge/skater-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the EDGE zone time statistics for a skater, season, game type, and 
#' category
#'
#' `skater_edge_zone_time()` retrieves the EDGE zone time statistics for a skater, season, game type, and category as a `data.frame` where each row represents strength state and includes detail on NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams skater_edge_summary
#' @param category character of 'd'/'details' or 's'/'starts'
#'
#' @returns data.frame with one row per strength state (category = 'details') 
#' or list with six items (category = 'starts')
#' @examples
#' Martin_Necas_starts_regular_20242025 <- skater_edge_zone_time(
#'   player    = 8480039,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'S'
#' )
#' @export

skater_edge_zone_time <- function(
  player    = 8478402,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'zoneTimeDetails',
        s = 'zoneStarts'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/skater-zone-time/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE skating distance statistics for a skater, season, game type, 
#' and category
#'
#' `skater_edge_skating_distance()` retrieves the EDGE skating distance statistics for a skater, season, game type, and category as a `data.frame` where each row represents strength state and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and tracking/spatial detail such as location, speed, distance, and zone distribution.
#'
#' @inheritParams skater_edge_summary
#' @param category character of 'd'/'details' or 'l'/'l10'/'last 10'
#'
#' @returns data.frame with one row per strength state (category = 'details') 
#' or game (category = 'last 10')
#' @examples
#' Martin_Necas_L10_skating_distance_regular_20242025 <- 
#'   skater_edge_skating_distance(
#'     player    = 8480039,
#'     season    = 20242025,
#'     game_type = 2,
#'     category  = 'L'
#'   )
#' @export

skater_edge_skating_distance <- function(
  player    = 8478402,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'skatingDistanceDetails',
        l = 'skatingDistanceLast10'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/skater-skating-distance-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE skating speed statistics for a skater, season, game type, and 
#' category
#'
#' `skater_edge_skating_speed()` retrieves the EDGE skating speed statistics for a skater, season, game type, and category as a nested `list` that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams skater_edge_summary
#' @param category character of 'd'/'details' or 't'/'top'/'top speeds'
#'
#' @returns list with four items (category = 'details') or data.frame with 
#' one row per burst (category = 'top speeds')
#' @examples
#' Martin_Necas_top_speeds_regular_20242025 <- skater_edge_skating_speed(
#'   player    = 8480039,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'T'
#' )
#' @export

skater_edge_skating_speed <- function(
  player    = 8478402,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'skatingSpeedDetails',
        t = 'topSkatingSpeeds'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/skater-skating-speed-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE shot location statistics for a skater, season, game type, and 
#' category
#'
#' `skater_edge_shot_location()` retrieves the EDGE shot location statistics for a skater, season, game type, and category as a `data.frame` where each row represents shot location and includes detail on production, workload, efficiency, and result-level performance outcomes plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams skater_edge_summary
#' @param category character of 'd'/details' or 't'/'totals'
#'
#' @returns data.frame with one row per shot location
#' @examples
#' Martin_Necas_shot_location_totals_regular_20242025 <- 
#'   skater_edge_shot_location(
#'     player    = 8480039,
#'     season    = 20242025,
#'     game_type = 2,
#'     category  = 'T'
#'   )
#' @export

skater_edge_shot_location <- function(
  player    = 8478402,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'shotLocationDetails',
        t = 'shotLocationTotals'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/skater-shot-location-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE shot speed statistics for a skater, season, game type, and 
#' category
#'
#' `skater_edge_shot_speed()` retrieves the EDGE shot speed statistics for a skater, season, game type, and category as a nested `list` that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams skater_edge_summary
#' @param category character of 'd'/'details' or 'h'/'hardest'
#'
#' @returns list with six items (category = 'details') or data.frame with one 
#' row per shot (category = 'hardest')
#' @examples
#' Martin_Necas_hardest_shots_regular_20242025 <- skater_edge_shot_speed(
#'   player    = 8480039,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'H'
#' )
#' @export

skater_edge_shot_speed <- function(
    player    = 8478402,
    season    = 'now', 
    game_type = '', 
    category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'shotSpeedDetails',
        h = 'hardestShots',
      )
      nhl_api(
        path = sprintf(
          'v1/edge/skater-shot-speed-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

test_that("skater_edge_zone_time() returns non-empty data.frame", {
  skip_if_offline()
  test <- skater_edge_zone_time()
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

test_that("skater_edge_zone_time(0) returns message and empty data.frame", {
  skip_if_offline()
  expect_message(
    test <- skater_edge_zone_time(0),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.data.frame(test) && nrow(test) == 0)
})

#' Access all the ESPN players
#'
#' `espn_players()` retrieves all the ESPN players as a `data.frame` where each row represents ESPN player and includes detail on person-level profile context and performance history with situational splits.
#'
#' @returns data.frame with one row per ESPN player
#' @examples
#' all_ESPN_players <- espn_players()
#' @export

espn_players <- function() {
  tryCatch({
    page <- 1
    all_players <- list()
    repeat {
      players <- espn_api(
        path  = 'athletes',
        query = list(limit = 1000, page = page),
        type  = 'c'
      )
      df   <- as.data.frame(players$items, stringsAsFactors = FALSE)
      all_players[[length(all_players) + 1]] <- df
      if (nrow(df) < 1000) break
      page <- page + 1
    }
    out <- do.call(rbind, all_players)
    id  <- sub('.*athletes/([0-9]+)\\?lang.*', '\\1', out[[1]])
    data.frame(espnPlayerId = id, stringsAsFactors = FALSE)
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the ESPN summary for a player
#'
#' `espn_player_summary()` retrieves the ESPN summary for a player as a `data.frame` where each row represents one result and includes detail on game timing, matchup state, scoring flow, and situational event detail.
#'
#' @param player integer ID (e.g., 3988803); see [espn_players()] for 
#' reference
#'
#' @returns data.frame with one row
#' @examples
#' ESPN_summary_Charlie_McAvoy <- espn_player_summary(player = 3988803)
#' @export

espn_player_summary <- function(player = 3988803) {
  get_or_na <- function(x, ...) {
    tryCatch({
      for (nm in list(...)) {
        if (is.null(x)) return(NA)
        x <- x[[nm]]
      }
      if (is.null(x)) NA else x
    }, error = function(e) NA)
  }
  player <- tryCatch(
    espn_api(
      path = sprintf('athletes/%s', player),
      type = 'c'
    ),
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      NULL
    }
  )
  if (is.null(player)) {
    return(data.frame())
  }
  data.frame(
    playerFirstName = get_or_na(player, 'firstName'),
    playerLastName  = get_or_na(player, 'lastName'),
    playerFullName = get_or_na(player, 'fullName'),
    dateOfBirth  = get_or_na(player, 'dateOfBirth'),
    birthCountry = get_or_na(player, 'birthPlace', 'country'),
    birthCity    = get_or_na(player, 'birthPlace', 'city'),
    age          = get_or_na(player, 'age'),
    height       = get_or_na(player, 'height'),
    weight       = get_or_na(player, 'weight'),
    handCode     = get_or_na(player, 'hand', 'abbreviation'),
    draftYear    = get_or_na(player, 'draft', 'year'),
    draftRound   = get_or_na(player, 'draft', 'round'),
    draftPick    = get_or_na(player, 'draft', 'selection'),
    debutYear    = get_or_na(player, 'debutYear'),
    positionCode = get_or_na(player, 'position', 'abbreviation'),
    experience   = get_or_na(player, 'experience', 'years'),
    isActive     = get_or_na(player, 'active'),
    stringsAsFactors = FALSE
  )
}

#' Access the playoff bracket for a season
#' 
#' `get_bracket()` is deprecated. Use [bracket()] instead.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per series
#' @export

get_bracket <- function(season = season_now()) {
  .Deprecated(
    new     = 'bracket()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_bracket()` is deprecated.',
      'Use `bracket()` instead.'
    )
  )
  bracket(season)
}

#' Access the playoff schedule for a season and series
#' 
#' `get_series_schedule()` is deprecated. Use [series_schedule()] instead.
#' 
#' @inheritParams series_schedule
#' @returns data.frame with one row per game
#' @export

get_series_schedule <- function(season = season_now(), series = 'a') {
  .Deprecated(
    new     = 'series_schedule()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_series_schedule()` is deprecated.',
      'Use `series_schedule()` instead.'
    )
  )
  series_schedule(season, series)
}

#' Access the playoff series for a season and round
#' 
#' `get_series()` is defunct.
#' 
#' @export

get_series <- function() {
  .Defunct(
    msg = paste(
      '`get_series()` is defunct.'
    )
  )
}

#' Access all the teams
#'
#' `teams()` retrieves all the teams as a `data.frame` where each row represents team and includes detail on team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per team
#' @examples
#' all_teams <- teams()
#' @export

teams <- function() {
  tryCatch({
    teams <- nhl_api(
      path = 'en/team',
      type = 's'
    )$data
    teams <- teams[order(teams$id), ]
    names(teams)[names(teams) == 'id']         <- 'teamId'
    names(teams)[names(teams) == 'fullName']   <- 'teamFullName'
    names(teams)[names(teams) == 'triCode']    <- 'teamTriCode'
    names(teams)[names(teams) == 'rawTricode'] <- 'teamTriCodeRaw'
    teams
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the season(s) and game type(s) in which a team played
#'
#' `team_seasons()` retrieves the season(s) and game type(s) in which a team played as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context.
#'
#' @param team integer ID (e.g., 21), character full name (e.g., 'Colorado 
#' Avalanche'), OR three-letter code (e.g., 'COL'); see [teams()] for 
#' reference; ID is preferable as there now exists duplicate three-letter codes 
#' (i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')
#'
#' @returns data.frame with one row per season
#' @examples
#' COL_seasons <- team_seasons(team = 21)
#' @export

team_seasons <- function(team = 1) {
  tryCatch(
    expr = {
      seasons <- nhl_api(
        path = sprintf('v1/club-stats-season/%s', to_team_tri_code(team)),
        type = 'w'
      )
      names(seasons)[names(seasons) == 'season'] <- 'seasonId'
      names(seasons)[names(seasons) == 'gameTypes'] <- 'gameTypeIds'
      seasons
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the configurations for team reports
#'
#' `team_report_configurations()` retrieves the configurations for team reports as a nested `list` that separates summary and detail blocks for situational splits across home/road, strength state, and overtime/shootout states plus configuration catalogs for valid report categories and filters.
#'
#' @returns list with various items
#' @examples
#' team_report_configs <- team_report_configurations()
#' @export

team_report_configurations <- function() {
  tryCatch({
    nhl_api(
      path = 'en/config',
      type = 's'
    )$teamReportData
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname team_report_configurations
#' @export

team_report_configs <- function() {
  team_report_configurations()
}

#' Access various reports for a season, game type, and category for all 
#' the teams by season
#'
#' `team_season_report()` retrieves various reports for a season, game type, and category for all the teams by season as a `data.frame` where each row represents team and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams roster_statistics
#' @param category character (e.g., 'leadingtrailing'); see 
#' [team_report_configurations()] for reference
#'
#' @returns data.frame with one row per team
#' @examples
#' situational_team_season_report_playoffs_20212022 <- team_season_report(
#'   season    = 20212022, 
#'   game_type = 3, 
#'   category  = 'leadingtrailing'
#' )
#' @export

team_season_report <- function(
  season    = season_now(), 
  game_type = game_type_now(), 
  category  = 'summary'
) {
  tryCatch(
    expr = {
      report <- nhl_api(
        path  = sprintf('en/team/%s', category),
        query = list(
          limit       = -1,
          isAggregate = FALSE,
          isGame      = FALSE,
          cayenneExp  = sprintf(
            'seasonId = %s and gameTypeId = %s', 
            season,
            game_type
          )
        ),
        type  = 's'
      )$data
      names(report) <- normalize_team_abbrev_cols(names(report))
      report[order(report$teamId), ]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access various reports for a season, game type, and category for all 
#' the teams by game
#'
#' `team_game_report()` retrieves various reports for a season, game type, and category for all the teams by game as a `data.frame` where each row represents game per team and includes detail on game timeline state, period/clock progression, and matchup flow, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams team_season_report
#'
#' @returns data.frame with one row per game per team
#' @examples
#' situational_team_game_report_playoffs_20212022 <- team_game_report(
#'   season    = 20212022, 
#'   game_type = 3, 
#'   category  = 'leadingtrailing'
#' )
#' @export

team_game_report <- function(
  season    = season_now(), 
  game_type = game_type_now(), 
  category  = 'summary'
) {
  tryCatch(
    expr = {
      report <- nhl_api(
        path  = sprintf('en/team/%s', category),
        query = list(
          limit       = -1,
          isAggregate = FALSE,
          isGame      = TRUE,
          cayenneExp  = sprintf(
            'seasonId = %s and gameTypeId = %s', 
            season,
            game_type
          )
        ),
        type  = 's'
      )$data
      names(report) <- normalize_team_abbrev_cols(names(report))
      report[order(report$teamId, report$gameId), ]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the statistics for all the teams by season and game type
#'
#' `team_season_statistics()` retrieves the statistics for all the teams by season and game type as a `data.frame` where each row represents team per season per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per team per season per game type
#' @examples
#' # May take >5s, so skip.
#' \donttest{team_season_statistics <- team_season_statistics()}
#' @export

team_season_statistics <- function() {
  tryCatch({
    stats <- nhl_api(
      path = 'team-stats',
      type = 'r'
    )$data
    stats <- stats[order(
      stats$`id.db:TEAMID`, 
      stats$`id.db:SEASON`, 
      stats$`id.db:GAMETYPE`
    ), ]
    names(stats)[names(stats) == 'id.db:TEAMID']   <- 'teamId'
    names(stats)[names(stats) == 'id.db:SEASON']   <- 'seasonId'
    names(stats)[names(stats) == 'id.db:GAMETYPE'] <- 'gameTypeId'
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname team_season_statistics
#' @export

team_season_stats <- function() {
  team_season_statistics()
}

#' Access the roster for a team, season, and position
#'
#' `roster()` retrieves the roster for a team, season, and position as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @inheritParams team_seasons
#' @param season integer in YYYYYYYY (e.g., 20242025); see [seasons()] for 
#' reference
#' @param position character of 'f'/'forwards', 'd'/'defensemen', or 
#' 'g'/'goalies'
#'
#' @returns data.frame with one row per player
#' @examples
#' COL_defensemen_20242025 <- roster(
#'   team     = 21,
#'   season   = 20242025,
#'   position = 'D'
#' )
#' @export

roster <- function(
  team     = 1,
  season   = 'current',
  position = 'forwards'
) {
  tryCatch(
    expr = {
      position <- switch(
        substring(tolower(position), 1, 1),
        f = 'forwards',
        d = 'defensemen',
        g = 'goalies'
      )
      players <- nhl_api(
        path = sprintf('v1/roster/%s/%s', to_team_tri_code(team), season),
        type = 'w'
      )[[position]]
      names(players)[names(players) == 'id'] <- 'playerId'
      names(players) <- normalize_locale_names(names(players))
      names(players) <- scope_person_name_cols(names(players), 'player')
      players
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the roster statistics for a team, season, game type, and position
#'
#' `roster_statistics()` retrieves the roster statistics for a team, season, game type, and position as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams roster
#' @param game_type integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season) OR character of 'pre', 'regular', or 
#' playoff'/'post'; see [seasons()] for reference; most functions will NOT 
#' support pre-season
#' @param position character of 's'/'skaters' or 'g'/'goalies'
#'
#' @returns data.frame with one row per player
#' @examples
#' COL_goalies_statistics_regular_20242025 <- roster_statistics(
#'   team      = 21,
#'   season    = 20242025,
#'   game_type = 2,
#'   position  = 'G'
#' )
#' @export

roster_statistics <- function(
  team      = 1,
  season    = 'now',
  game_type = '',
  position  = 'skaters'
) {
  tryCatch(
    expr = {
      position <- switch(
        substring(tolower(position), 1, 1),
        s = 'skaters',
        g = 'goalies'
      )
      players <- nhl_api(
        path = sprintf(
          'v1/club-stats/%s/%s/%s', 
          to_team_tri_code(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[position]]
      names(players) <- normalize_locale_names(names(players))
      names(players) <- scope_person_name_cols(names(players), 'player')
      players
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname roster_statistics
#' @export
roster_stats <- function(
  team      = 1,
  season    = 'now',
  game_type = '',
  position  = 'skaters'
) {
  roster_statistics(team, season, game_type, position)
}

#' Access the prospects for a team and position
#'
#' `team_prospects()` retrieves the prospects for a team and position as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @inheritParams roster
#'
#' @returns data.frame with one row per player
#' @examples
#' COL_forward_prospects <- team_prospects(
#'   team     = 21,
#'   position = 'F'
#' )
#' @export

team_prospects <- function(team = 1, position = 'forwards') {
  tryCatch(
    expr = {
      position <- switch(
        substring(tolower(position), 1, 1),
        f = 'forwards',
        d = 'defensemen',
        g = 'goalies'
      )
      players <- nhl_api(
        path = sprintf('v1/prospects/%s', to_team_tri_code(team)),
        type = 'w'
      )[[position]]
      names(players)[names(players) == 'id'] <- 'playerId'
      names(players) <- normalize_locale_names(names(players))
      names(players) <- scope_person_name_cols(names(players), 'player')
      players
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the schedule for a team and season
#'
#' `team_season_schedule()` retrieves the schedule for a team and season as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @inheritParams roster
#'
#' @returns data.frame with one row per game
#' @examples
#' COL_schedule_20252026 <- team_season_schedule(
#'   team   = 21, 
#'   season = 20252026
#' )
#' @export

team_season_schedule <- function(team = 1, season = 'now') {
  tryCatch(
    expr = {
      games <- nhl_api(
        path = sprintf(
          'v1/club-schedule-season/%s/%s', 
          to_team_tri_code(team), 
          season
        ),
        type = 'w'
      )$games
      names(games)[names(games) == 'id']       <- 'gameId'
      names(games)[names(games) == 'season']   <- 'seasonId'
      names(games)[names(games) == 'gameType'] <- 'gameTypeId'
      names(games) <- normalize_locale_names(names(games))
      names(games) <- normalize_team_abbrev_cols(names(games))
      games
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the schedule for a team and month
#'
#' `team_month_schedule()` retrieves the schedule for a team and month as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @inheritParams team_seasons
#' @param month character in 'YYYY-MM' (e.g., '2025-01'); see [seasons()] 
#' for reference
#'
#' @returns data.frame with one row per game
#' @examples
#' COL_schedule_December_2025 <- team_month_schedule(
#'   team  = 21, 
#'   month = '2025-12'
#' )
#' @export

team_month_schedule <- function(team = 1, month = 'now') {
  tryCatch(
    expr = {
      games <- nhl_api(
        path = sprintf(
          'v1/club-schedule/%s/month/%s', 
          to_team_tri_code(team), 
          month
        ),
        type = 'w'
      )$games
      names(games)[names(games) == 'id']       <- 'gameId'
      names(games)[names(games) == 'season']   <- 'seasonId'
      names(games)[names(games) == 'gameType'] <- 'gameTypeId'
      names(games) <- normalize_locale_names(names(games))
      names(games) <- normalize_team_abbrev_cols(names(games))
      games
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the schedule for a team and week since a date
#'
#' `team_week_schedule()` retrieves the schedule for a team and week since a date as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @inheritParams team_seasons
#' @inheritParams standings
#'
#' @returns data.frame with one row per game
#' @examples
#' COL_schedule_Family_Week_2025 <- team_week_schedule(
#'   team = 21,
#'   date = '2025-10-06'
#' )
#' @export

team_week_schedule <- function(team = 1, date = 'now') {
  tryCatch(
    expr = {
      games <- nhl_api(
        path = sprintf(
          'v1/club-schedule/%s/week/%s', 
          to_team_tri_code(team), 
          date
        ),
        type = 'w'
      )$games
      names(games)[names(games) == 'id']       <- 'gameId'
      names(games)[names(games) == 'season']   <- 'seasonId'
      names(games)[names(games) == 'gameType'] <- 'gameTypeId'
      names(games) <- normalize_locale_names(names(games))
      names(games) <- normalize_team_abbrev_cols(names(games))
      games
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access all the team logos
#'
#' `team_logos()` retrieves all the team logos as a `data.frame` where each row represents logo and includes detail on team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per logo
#' @examples
#' all_team_logos <- team_logos()
#' @export

team_logos <- function() {
  tryCatch({
    logos <- nhl_api(
      path = 'logo',
      type = 'r'
    )$data
    logos$id <- NULL
    logos    <- logos[order(logos$teamId, logos$startSeason), ]
    names(logos)[names(logos) == 'startSeason'] <- 'startSeasonId'
    names(logos)[names(logos) == 'endSeason']   <- 'endSeasonId'
    logos
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

test_that("team_season_report(20242025, 3) returns non-empty data.frame", {
  skip_if_offline()
  test <- team_season_report(20242025, 3)
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

test_that("team_season_report(0) returns message and empty data.frame", {
  skip_if_offline()
  expect_message(
    test <- team_season_report(0),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.data.frame(test) && nrow(test) == 0)
})

#' Access all the seasons
#'
#' `seasons()` retrieves all the seasons as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context.
#'
#' @returns data.frame with one row per season
#' @examples
#' all_seasons <- seasons()
#' @export

seasons <- function() {
  tryCatch({
    seasons <- nhl_api(
      path = 'en/season',
      type = 's'
    )$data
    seasons <- seasons[order(seasons$id), ]
    names(seasons)[names(seasons) == 'id'] <- 'seasonId'
    seasons
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the season as of now
#'
#' `season_now()` retrieves the season as of now and returns a scalar integer used as the current-context default in season/game-type dependent wrappers.
#'
#' @returns integer in YYYYYYYY (e.g., 20242025)
#' @examples
#' season_now <- season_now()
#' @export

season_now <- function() {
  tryCatch({
    nhl_api(
      path = 'en/componentSeason',
      type = 's'
    )$data$seasonId
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    -1
  })
}

#' Access the game type as of now
#'
#' `game_type_now()` retrieves the game type as of now and returns a scalar integer used as the current-context default in season/game-type dependent wrappers.
#'
#' @returns integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season)
#' @examples
#' game_type_now <- game_type_now()
#' @export

game_type_now <- function() {
  tryCatch({
    nhl_api(
      path = 'en/componentSeason',
      type = 's'
    )$data$gameTypeId
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    -1
  })
}

#' Access the standings rules by season
#'
#' `standings_rules()` retrieves the standings rules by season as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context.
#'
#' @returns data.frame with one row per season
#' @examples
#' standings_rules <- standings_rules()
#' @export

standings_rules <- function() {
  tryCatch({
    seasons <- nhl_api(
      path = 'v1/standings-season',
      type = 'w'
    )$seasons
    names(seasons)[names(seasons) == 'id'] <- 'seasonId'
    seasons
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the standings for a date
#'
#' `standings()` retrieves the standings for a date as a `data.frame` where each row represents team and includes detail on date/season filtering windows and chronological context, production, workload, efficiency, and result-level performance outcomes, and ranking movement, points pace, and division/conference position signals.
#'
#' @param date character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see 
#' [seasons()] for reference
#'
#' @returns data.frame with one row per team
#' @examples
#' standings_Halloween_2025 <- standings(date = '2025-10-31')
#' @export

standings <- function(date = 'now') {
  tryCatch(
    expr = {
      standings <- nhl_api(
        path = sprintf('v1/standings/%s', date),
        type = 'w'
      )$standings
      names(standings) <- normalize_locale_names(names(standings))
      names(standings) <- normalize_team_abbrev_cols(names(standings))
      standings
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the schedule for a date
#'
#' `schedule()` retrieves the schedule for a date as a `data.frame` where each row represents game and includes detail on game timing, matchup state, scoring flow, and situational event detail.
#'
#' @inheritParams standings
#'
#' @returns data.frame with one row per game
#' @examples
#' schedule_Halloween_2025 <- schedule(date = '2025-10-31')
#' @export

schedule <- function(date = Sys.Date()) {
  tryCatch(
    expr = {
      gameWeek <- nhl_api(
        path = sprintf('v1/schedule/%s', date),
        type = 'w'
      )$gameWeek
      gameWeek <- gameWeek[gameWeek$date == date, ]$games[[1]]
      names(gameWeek)[names(gameWeek) == 'id']       <- 'gameId'
      names(gameWeek)[names(gameWeek) == 'season']   <- 'seasonId'
      names(gameWeek)[names(gameWeek) == 'gameType'] <- 'gameTypeId'
      names(gameWeek) <- normalize_locale_names(names(gameWeek))
      names(gameWeek) <- normalize_team_abbrev_cols(names(gameWeek))
      gameWeek
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access all the venues
#'
#' `venues()` retrieves all the venues as a `data.frame` where each row represents venue and includes detail on venue/location geography and regional metadata.
#'
#' @returns data.frame with one row per venue
#' @examples
#' all_venues <- venues()
#' @export

venues <- function() {
  tryCatch({
    venues    <- nhl_api(
      path = 'venue',
      type = 'r'
    )$data
    venues$id <- NULL
    venues[order(venues$venueId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the attendance by season and game type
#'
#' `attendance()` retrieves the attendance by season and game type as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context.
#'
#' @returns data.frame with one row per season
#' @examples
#' all_attendance <- attendance()
#' @export

attendance <- function() {
  tryCatch({
    attendance <- nhl_api(
      path = 'attendance',
      type = 'r'
    )$data
    attendance$id <- NULL
    attendance[order(attendance$seasonId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

test_that("schedule('2025-01-01') returns data.frame", {
  skip_if_offline()
  test <- schedule('2025-01-01')
  expect_true(is.data.frame(test))
})

test_that("schedule(20250101) returns message and empty data.frame", {
  skip_if_offline()
  expect_message(
    test <- schedule(20250101),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.data.frame(test) && nrow(test) == 0)
})

test_that("standings('2025-01-01') returns non-empty data.frame", {
  skip_if_offline()
  test <- standings('2025-01-01')
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

test_that("standings(20250101) returns message and empty data.frame", {
  skip_if_offline()
  expect_message(
    test <- standings(20250101),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.data.frame(test) && nrow(test) == 0)
})

#' Draw a circle on the rink
#'
#' `draw_rink_circle()` draws a circle on the rink.
#'
#' @param x numeric x coordinate of the circle center
#' @param y numeric y coordinate of the circle center
#' @param r numeric radius
#' @param n integer number of points used to approximate circle
#' @param col character color
#' @param lwd numeric line width
#' @returns `NULL`
#' @keywords internal

draw_rink_circle <- function(x, y, r, n = 200, col = 'black', lwd = 1) {
  theta <- seq(0, 2 * pi, length.out = n)
  graphics::lines(x + r * cos(theta), y + r * sin(theta), col = col, lwd = lwd)
}

#' Draw an arc on the rink
#'
#' `draw_rink_arc()` draws an arc on the rink.
#'
#' @param cx numeric x coordinate of the circle center
#' @param cy numeric y coordinate of the circle center
#' @param r numeric radius
#' @param theta1 numeric start angle of arc
#' @param theta2 numeric end angle of arc
#' @param n integer number of points used to approximate arc
#' @param col character color
#' @param lwd numeric line width
#' @returns `NULL`
#' @keywords internal

draw_rink_arc <- function(
  cx, cy, r, theta1, theta2, n = 100, col = 'black', lwd = 1
) {
  theta <- seq(theta1, theta2, length.out = n)
  graphics::lines(
    cx + r * cos(theta), cy + r * sin(theta), 
    col = col, lwd = lwd
  )
}

#' Draw a full NHL rink
#'
#' `draw_NHL_rink()` draws a full NHL rink such that the x and y coordinates 
#' span -100 to 100 and -43 to +43, respectively. Use [graphics::points()] 
#' to create custom graphs; check out an example on the online documentation!
#'
#' @returns `NULL`
#' @examples
#' draw_NHL_rink()
#' @export

draw_NHL_rink <- function() {
  old_par <- graphics::par(
    xaxs = 'r',
    yaxs = 'r',
    mar  = c(1, 1, 3, 1),
    oma  = c(0, 0, 0, 0)
  )
  on.exit(graphics::par(old_par))
  plot(
    NA, NA,
    xlim = c(-100, 100),
    ylim = c(-43, 43),
    asp  = 1,
    xlab = '',
    ylab = '',
    axes = FALSE
  )
  board_col          <- 'black'
  center_line_col    <- 'red'
  blue_line_col      <- 'blue'
  goal_line_col      <- 'red'
  faceoff_center_col <- 'blue'
  faceoff_other_col  <- 'red'
  crease_col         <- 'blue'
  line_lwd <- 1
  x_min <- -100
  x_max <-  100
  y_min <- -43
  y_max <-  43
  corner_r <- 28
  straight_top_bottom_x <- 100 - corner_r
  straight_side_y       <- 43  - corner_r
  goal_line   <- 89
  blue_line   <- 25
  center_line <- 0
  circle_r <- 15
  off_x    <- 69
  off_y    <- 22
  cx_right <- x_max - corner_r
  cx_left  <- x_min + corner_r
  cy_top   <- y_max - corner_r
  cy_bot   <- y_min + corner_r
  graphics::segments(
    center_line, y_min, center_line, y_max,
    col = center_line_col, lwd = line_lwd
  )
  graphics::segments(
    blue_line, y_min,  blue_line, y_max,
    col = blue_line_col, lwd = line_lwd
  )
  graphics::segments(
    -blue_line, y_min, -blue_line, y_max,
    col = blue_line_col, lwd = line_lwd
  )
  dx_goal  <- goal_line - cx_right
  dy_goal  <- sqrt(corner_r^2 - dx_goal^2)
  y_goal_top    <- cy_top + dy_goal
  y_goal_bottom <- cy_bot - dy_goal
  graphics::segments(
    goal_line,  y_goal_bottom,  goal_line,  y_goal_top,
    col = goal_line_col, lwd = line_lwd
  )
  graphics::segments(
    -goal_line,  y_goal_bottom, -goal_line,  y_goal_top,
    col = goal_line_col, lwd = line_lwd
  )
  draw_rink_circle(
    0, 0, circle_r,
    col = faceoff_center_col, lwd = line_lwd
  )
  draw_rink_circle(
    off_x,  off_y, circle_r,
    col = faceoff_other_col, lwd = line_lwd
  )
  draw_rink_circle(
    off_x, -off_y, circle_r,
    col = faceoff_other_col, lwd = line_lwd
  )
  draw_rink_circle(
    -off_x,  off_y, circle_r,
    col = faceoff_other_col, lwd = line_lwd
  )
  draw_rink_circle(
    -off_x, -off_y, circle_r,
    col = faceoff_other_col, lwd = line_lwd
  )
  crease_r <- 6
  theta <- seq(-pi / 2, pi / 2, length.out = 200)
  graphics::lines(
    goal_line - crease_r * cos(theta),
    0         + crease_r * sin(theta),
    col = crease_col,
    lwd = line_lwd
  )
  graphics::lines(
    -goal_line + crease_r * cos(theta),
    0          + crease_r * sin(theta),
    col = crease_col,
    lwd = line_lwd
  )
  graphics::segments(
    -straight_top_bottom_x,  y_max, straight_top_bottom_x,  y_max,
    col = board_col, lwd = line_lwd
  )
  graphics::segments(
    -straight_top_bottom_x,  y_min, straight_top_bottom_x,  y_min,
    col = board_col, lwd = line_lwd
  )
  graphics::segments(
    x_min, -straight_side_y, x_min, straight_side_y,
    col = board_col, lwd = line_lwd
  )
  graphics::segments(
    x_max, -straight_side_y, x_max,  straight_side_y,
    col = board_col, lwd = line_lwd
  )
  draw_rink_arc(
    cx_right, cy_top, corner_r, theta1 = pi / 2, theta2 = 0,
    col = board_col, lwd = line_lwd
  )
  draw_rink_arc(
    cx_left, cy_top, corner_r, theta1 = pi / 2, theta2 = pi,
    col = board_col, lwd = line_lwd
  )
  draw_rink_arc(
    cx_left, cy_bot, corner_r, theta1 = pi, theta2 = 3 * pi / 2,
    col = board_col, lwd = line_lwd
  )
  draw_rink_arc(
    cx_right, cy_bot, corner_r, theta1 = 3 * pi / 2, theta2 = 2 * pi,
    col = board_col, lwd = line_lwd
  )
  invisible(NULL)
}

#' nhlscraper: Scrape, clean, and visualize NHL data
#'
#' `nhlscraper` is a minimum-dependency R package to scrape, clean, and
#' visualize NHL data via the NHL and ESPN APIs. It primarily wraps 125+
#' [endpoints](https://github.com/RentoSaijo/nhlscraper/wiki) from high-level
#' multi-season summaries and award winners to low-level decisecond replays and
#' bookmakers' odds, making them significantly more accessible. It also
#' features cleaning and visualization functions, primarily for play-by-plays,
#' to help analyze the data.
#'
#' @section Prerequisite:
#' - R/RStudio; you can check out my
#'   [tutorial](https://youtu.be/CUXzeuXJCPs?si=bkWFAm-hSXhUsxaN) if you are not
#'   familiar.
#'
#' @section Disclosure:
#' Detailed documentation for each scraping function will be released gradually
#' over time. Because each function can return a large amount of information,
#' users are encouraged to explore the provided examples to discover what is
#' available. Most, if not all, of the endpoints accessed by this package are
#' unofficially documented (i.e., hidden), so it is important to use them
#' responsibly and with respect for the NHL's data servers. Endpoints serving
#' historical or otherwise mostly static data should ideally be queried once and
#' then stored locally (for example, in a MySQL database) for further analysis,
#' rather than being called repeatedly; the
#' [load functions](https://rentosaijo.github.io/nhlscraper/reference/index.html#load)
#' attempt to mitigate this risk by providing pre-scraped data, so please make
#' use of them. The exact rate limits for these APIs are not publicly known, so
#' users are asked to avoid excessive or abusive querying to help ensure
#' continued access for everyone.
#'
#' @section History:
#' Prior to the NHL API rework in 2023, Drew Hynes documented a comprehensive
#' list of known [endpoints](https://gitlab.com/dword4/nhlapi/), and several R
#' packages, such as [nhlapi](https://github.com/jozefhajnala/nhlapi) and
#' [hockeyR](https://github.com/danmorse314/hockeyR), were built to access them.
#' However, after the NHL completely transformed its API structure, all of these
#' packages became mostly defunct as their authors understandably chose not to
#' continue maintaining them. The community gathered around work by Zachary
#' Maludzinski to discover and share new
#' [endpoints](https://github.com/Zmalski/NHL-API-Reference), but progress
#' naturally slowed once most of the "main" endpoints were identified. Over the
#' summer of 2025, I began reverse-engineering many of the remaining,
#' undocumented endpoints while searching for additional data for future
#' research, with a particular focus on NHL EDGE and Records data. After I
#' shared those
#' [findings](https://github.com/Zmalski/NHL-API-Reference/issues/69), the
#' effort to expand and refine this map of the APIs accelerated, and with the
#' addition of David Fleischer's
#' [effort](https://github.com/dfleis/nhl-api-docs), we ultimately identified
#' 400+ new endpoints. In parallel, I also discovered many new endpoints for
#' the ESPN API, extending beyond what Joseph Wilson had already
#' [compiled](https://github.com/pseudo-r/Public-ESPN-API).
#'
#' @name nhlscraper
#' @aliases nhlscraper-package
#' @useDynLib nhlscraper, .registration = TRUE
"_PACKAGE"

#' Access all the playoff series by game
#'
#' `series()` retrieves all the playoff series by game as a `data.frame` where each row represents game per series and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and playoff-series progression, round status, and series leverage.
#'
#' @returns data.frame with one row per game per series
#' @examples
#' # May take >5s, so skip.
#' \donttest{all_series <- series()}
#' @export

series <- function() {
  tryCatch({
    series    <- nhl_api(
      path = 'playoff-series',
      type = 'r'
    )$data
    series$id <- NULL
    series[order(series$seasonId, series$playoffSeriesLetter, series$gameId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the playoff statistics by season
#'
#' `playoff_season_statistics()` retrieves the playoff statistics by season as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context.
#'
#' @returns data.frame with one row per season
#' @examples
#' playoff_season_stats <- playoff_season_statistics()
#' @export

playoff_season_statistics <- function() {
  tryCatch({
    totals    <- nhl_api(
      path = 'league-playoff-year-totals',
      type = 'r'
    )$data
    totals$id <- NULL
    totals[order(totals$seasonId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname playoff_season_statistics
#' @export

playoff_season_stats <- function() {
  playoff_season_statistics()
}

#' Access the playoff bracket for a season
#'
#' `bracket()` retrieves the playoff bracket for a season as a `data.frame` where each row represents series and includes detail on team identity, affiliation, and matchup-side context.
#'
#' @inheritParams roster
#'
#' @returns data.frame with one row per series
#' @examples
#' bracket_20242025 <- bracket(season = 20242025)
#' @export

bracket <- function(season = season_now()){
  tryCatch(
    expr = {
      series <- data.frame(nhl_api(
        path = sprintf(
          'v1/playoff-bracket/%s', 
          suppressWarnings(as.integer(season)) %% 1e4
        ),
        type = 'w'
      )$series)
      names(series) <- normalize_locale_names(names(series))
      names(series) <- normalize_team_abbrev_cols(names(series))
      series
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the playoff schedule for a season and series
#'
#' `series_schedule()` retrieves the playoff schedule for a season and series as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @inheritParams roster
#' @param series one-letter code (e.g., 'O'); see [series()] and/or 
#' [bracket()] for reference
#'
#' @returns data.frame with one row per game
#' @examples
#' SCF_schedule_20212022 <- series_schedule(
#'   season = 20212022, 
#'   series = 'O'
#' )
#' @export

series_schedule <- function(season = season_now() - 10001, series = 'a') {
  tryCatch(
    expr = {
      games <- nhl_api(
        path = sprintf(
          'v1/schedule/playoff-series/%s/%s', 
          season, 
          tolower(series)
        ),
        type = 'w'
      )$games
      names(games)[names(games) == 'id']       <- 'gameId'
      names(games)[names(games) == 'season']   <- 'seasonId'
      names(games)[names(games) == 'gameType'] <- 'gameTypeId'
      names(games) <- normalize_locale_names(names(games))
      names(games) <- normalize_team_abbrev_cols(names(games))
      games
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the glossary
#' 
#' `get_glossary()` is deprecated. Use [glossary()] instead.
#' 
#' @returns data.frame with one row per terminology
#' @export

get_glossary <- function() {
  .Deprecated(
    new     = 'glossary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_glossary()` is deprecated.',
      'Use `glossary()` instead.'
    )
  )
  glossary()
}

#' Access all the countries
#' 
#' `get_countries()` is deprecated. Use [countries()] instead.
#' 
#' @returns data.frame with one row per country
#' @export

get_countries <- function() {
  .Deprecated(
    new     = 'countries()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_countries()` is deprecated.',
      'Use `countries()` instead.'
    )
  )
  countries()
}

#' Access all the streams
#' 
#' `get_streams()` is deprecated. Use [streams()] instead.
#' 
#' @returns data.frame with one row per stream
#' @export

get_streams <- function() {
  .Deprecated(
    new     = 'streams()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_streams()` is deprecated.',
      'Use `streams()` instead.'
    )
  )
  streams()
}

#' Access the NHL Network TV schedule for a date
#' 
#' `get_tv_schedule()` is deprecated. Use [tv_schedule()] instead.
#' 
#' @inheritParams standings
#' @returns data.frame with one row per program
#' @export

get_tv_schedule <- function(date = 'now') {
  .Deprecated(
    new     = 'tv_schedule()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_tv_schedule()` is deprecated.',
      'Use `tv_schedule()` instead.'
    )
  )
  tv_schedule(date)
}

#' Ping
#' 
#' `ping()` is defunct.
#' 
#' @export

ping <- function() {
  .Defunct(
    msg = paste(
      '`get_teams()` is defunct.'
    )
  )
}

#' Access the configurations for team, skater, and goalie reports
#' 
#' `get_configuration()` is defunct. Use [team_report_configurations()], 
#' [skater_report_configurations()], and/or [goalie_report_configurations()] 
#' instead.
#' 
#' @export

get_configuration <- function() {
  .Defunct(
    new     = 'team_report_configurations()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_configuration()` is defunct.',
      'Use `team_report_configurations()`, `skater_report_configurations()`,',
      'and/or `goalie_report_configurations()` instead.'
    )
  )
}

#' Access all the ESPN teams
#'
#' `espn_teams()` retrieves all the ESPN teams as a `data.frame` where each row represents ESPN team and includes detail on team composition, matchup context, and season progression detail.
#'
#' @returns data.frame with one row per ESPN team
#' @examples
#' all_ESPN_teams <- espn_teams()
#' @export

espn_teams <- function() {
  tryCatch({
    page <- 1
    all_teams <- list()
    repeat {
      teams <- espn_api(
        path  = 'teams',
        query = list(limit = 1000, page = page),
        type  = 'c'
      )
      df   <- as.data.frame(teams$items, stringsAsFactors = FALSE)
      all_teams[[length(all_teams) + 1]] <- df
      if (nrow(df) < 1000) break
      page <- page + 1
    }
    out <- do.call(rbind, all_teams)
    id  <- sub('.*teams/([0-9]+)\\?lang.*', '\\1', out[[1]])
    data.frame(espnTeamId = id, stringsAsFactors = FALSE)
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the ESPN summary for a team
#'
#' `espn_team_summary()` retrieves the ESPN summary for a team as a `data.frame` where each row represents one result and includes detail on game timing, matchup state, scoring flow, and situational event detail.
#'
#' @param team integer ID (e.g., 1); see [espn_teams()] for 
#' reference
#'
#' @returns data.frame with one row
#' @examples
#' ESPN_summary_Boston_Bruins <- espn_team_summary(team = 1)
#' @export

espn_team_summary <- function(team = 1) {
  get_or_na <- function(x, ...) {
    tryCatch({
      for (nm in list(...)) {
        if (is.null(x)) return(NA)
        x <- x[[nm]]
      }
      if (is.null(x)) NA else x
    }, error = function(e) NA)
  }
  tryCatch(
    expr = {
      team <- as.integer(team[[1]])
      if (is.na(team)) {
        stop('Invalid team.')
      }
      team <- espn_api(
        path = sprintf('teams/%s', team),
        type = 'c'
      )
      data.frame(
        location      = get_or_na(team, 'location'),
        teamName      = get_or_na(team, 'name'),
        teamFullName  = get_or_na(team, 'displayName'),
        teamTriCode   = get_or_na(team, 'abbreviation'),
        isActive      = get_or_na(team, 'isActive'),
        stringsAsFactors = FALSE
      )
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access all the officials
#' 
#' `get_officials()` is deprecated. Use [officials()] instead.
#' 
#' @returns data.frame with one row per official
#' @export

get_officials <- function() {
  .Deprecated(
    new     = 'officials()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_officials()` is deprecated.',
      'Use `officials()` instead.'
    )
  )
  officials()
}

#' Access all the drafts
#'
#' `drafts()` retrieves all the drafts as a `data.frame` where each row represents draft and includes detail on venue/location geography and regional metadata.
#'
#' @returns data.frame with one row per draft
#' @examples
#' all_drafts <- drafts()
#' @export

drafts <- function() {
  tryCatch({
    master <- nhl_api(
      path = 'draft-master',
      type = 'r'
    )$data
    rounds <- nhl_api(
      path = 'en/draft',
      type = 's'
    )$data
    rounds$id        <- NULL
    drafts           <- merge(master, rounds, by = 'draftYear')
    column_to_move   <- 'id'
    other_columns    <- setdiff(names(drafts), column_to_move)
    new_column_order <- c(column_to_move, other_columns)
    drafts <- drafts[, new_column_order]
    names(drafts)[names(drafts) == 'id']   <- 'draftId'
    drafts
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the draft picks
#'
#' `draft_picks()` retrieves all the draft picks as a `data.frame` where each row represents pick and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per pick
#' @examples
#' # May take >5s, so skip.
#' \donttest{all_draft_picks <- draft_picks()}
#' @export

draft_picks <- function() {
  tryCatch({
    picks <- nhl_api(
      path = 'draft',
      type = 'r'
    )$data
    picks <- picks[order(picks$draftYear), ]
    names(picks)[names(picks) == 'id']        <- 'draftPickId'
    names(picks)[names(picks) == 'firstName'] <- 'playerFirstName'
    names(picks)[names(picks) == 'lastName']  <- 'playerLastName'
    names(picks)[names(picks) == 'position']  <- 'positionCode'
    names(picks)[names(picks) == 'triCode']   <- 'teamTriCode'
    picks
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the draft prospects
#'
#' `draft_prospects()` retrieves all the draft prospects as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus broadcast carriage, media availability, and viewing-link metadata.
#'
#' @returns data.frame with one row per player
#' @examples
#' # May take >5s, so skip.
#' \donttest{all_prospects <- draft_prospects()}
#' @export

draft_prospects <- function() {
  tryCatch({
    prospects <- nhl_api(
      path = 'draft-prospect',
      type = 'r'
    )$data
    prospects <- prospects[order(prospects$firstName, prospects$lastName), ]
    names(prospects)[names(prospects) == 'id']                 <- 'prospectId'
    names(prospects)[names(prospects) == 'playerid']           <- 'playerId'
    names(prospects)[names(prospects) == 'firstName']          <- 'playerFirstName'
    names(prospects)[names(prospects) == 'lastName']           <- 'playerLastName'
    names(prospects)[names(prospects) == 'birthCountry3code']  <- 'birthCountryTriCode'
    names(prospects)[names(prospects) == 'birthStateProvCode'] <- 'birthStateProvinceCode'
    prospects
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the draft rankings for a class and category
#'
#' `draft_rankings()` retrieves the draft rankings for a class and category as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus draft-board context, scouting background, and pick/round progression.
#'
#' @param class integer in YYYY (e.g., 2017); see [drafts()] for reference
#' @param category integer in 1:4 (where 1 = North American Skaters, 
#' 2 = International Skaters, 3 = North American Goalies, and 4 = International 
#' Goalies) OR character of 'NAS'/'NA Skaters'/'North American Skaters', 
#' 'INTLS'/'INTL Skaters'/'International Skaters', 
#' 'NAG'/'NA Goalies'/'North American Goalies',
#' 'INTLG'/'INTL Goalies'/'International Goalies'
#'
#' @returns data.frame with one row per player
#' @examples
#' draft_rankings_INTL_Skaters_2017 <- draft_rankings(
#'   class    = 2017, 
#'   category = 2
#' )
#' @export

draft_rankings <- function(
  class    = season_now() %% 1e4,
  category = 1
) {
  tryCatch(
    expr = {
      category <- switch(
        tolower(category),
        `1`                      = 1,
        nas                      = 1,
        `na skaters`             = 1,
        `north american skaters` = 1,
        `2`                      = 2,
        intls                    = 2,
        `intl skaters`           = 2,
        `international skaters`  = 2,
        `3`                      = 3,
        nag                      = 3,
        `na goalies`             = 3,
        `north american goalies` = 3,
        `4`                      = 4,
        intlg                    = 4,
        `intl goalies`           = 4,
        `international goalies`  = 4
      )
      rankings <- nhl_api(
        path = sprintf('v1/draft/rankings/%s/%s', class, category),
        type = 'w'
      )$rankings
      names(rankings)[names(rankings) == 'firstName'] <- 'playerFirstName'
      names(rankings)[names(rankings) == 'lastName']  <- 'playerLastName'
      rankings
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the draft combine reports
#'
#' `combine_reports()` retrieves the draft combine reports as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player
#' @examples
#' combine_reports <- combine_reports()
#' @export

combine_reports <- function() {
  tryCatch({
    reports    <- nhl_api(
      path = 'combine',
      type = 'r'
    )$data
    reports$id <- NULL
    reports    <- reports[order(reports$draftYear, reports$event), ]
    names(reports)[names(reports) == 'firstName'] <- 'playerFirstName'
    names(reports)[names(reports) == 'lastName']  <- 'playerLastName'
    reports
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the draft lottery odds
#'
#' `lottery_odds()` retrieves the draft lottery odds as a `data.frame` where each row represents draft lottery and includes detail on draft-cycle evaluation, ranking, and selection tracking detail.
#'
#' @returns data.frame with one row per draft lottery
#' @examples
#' lottery_odds <- lottery_odds()
#' @export

lottery_odds <- function() {
  tryCatch({
    lotteries <- nhl_api(
      path = 'draft-lottery-odds',
      type = 'r'
    )$data
    lotteries <- lotteries[order(lotteries$draftYear), ]
    names(lotteries)[names(lotteries) == 'id'] <- 'lotteryId'
    lotteries
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the real-time draft tracker
#'
#' `draft_tracker()` retrieves the real-time draft tracker as a `data.frame` where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and venue/location geography and regional metadata.
#'
#' @returns data.frame with one row per player
#' @examples
#' draft_tracker <- draft_tracker()
#' @export

draft_tracker <- function() {
  tryCatch({
    picks <- nhl_api(
      path = 'v1/draft-tracker/picks/now',
      type = 'w'
    )$picks
    names(picks) <- normalize_locale_names(names(picks))
    names(picks) <- scope_person_name_cols(names(picks), 'player')
    names(picks) <- normalize_team_abbrev_cols(names(picks))
    picks
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the expansion drafts
#'
#' `expansion_drafts()` retrieves all the expansion drafts as a `data.frame` where each row represents expansion draft and includes detail on date/season filtering windows and chronological context plus reference definitions and rules-framework information.
#'
#' @returns data.frame with one row per expansion draft
#' @examples
#' all_expansion_drafts <- expansion_drafts()
#' @export

expansion_drafts <- function() {
  tryCatch({
    drafts    <- nhl_api(
      path = 'expansion-draft-rules',
      type = 'r'
    )$data
    drafts$id <- NULL
    drafts[order(drafts$seasonId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the expansion draft picks
#'
#' `expansion_draft_picks()` retrieves all the expansion draft picks as a `data.frame` where each row represents pick and includes detail on date/season filtering windows and chronological context plus team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per pick
#' @examples
#' all_expansion_draft_picks <- expansion_draft_picks()
#' @export

expansion_draft_picks <- function() {
  tryCatch({
    drafts    <- nhl_api(
      path = 'expansion-draft-picks',
      type = 'r'
    )$data
    drafts$id <- NULL
    drafts[order(drafts$seasonId, drafts$teamId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

test_that("draft_rankings(2025) returns non-empty data.frame", {
  skip_if_offline()
  test <- draft_rankings(2025)
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

test_that("draft_rankings(20242025) returns message and empty data.frame", {
  skip_if_offline()
  expect_message(
    test <- draft_rankings(20242025),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.data.frame(test) && nrow(test) == 0)
})

#' Access the configurations for skater reports
#'
#' `skater_report_configurations()` retrieves the configurations for skater reports as a nested `list` that separates summary and detail blocks for production, workload, efficiency, and result-level performance outcomes, situational splits across home/road, strength state, and overtime/shootout states, and configuration catalogs for valid report categories and filters.
#'
#' @returns list with various items
#' @examples
#' skater_report_configs <- skater_report_configurations()
#' @export

skater_report_configurations <- function() {
  tryCatch({
    nhl_api(
      path = 'en/config',
      type = 's'
    )$playerReportData
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname skater_report_configurations
#' @export

skater_report_configs <- function() {
  skater_report_configurations()
}

#' Access various reports for a season, game type, and category for all 
#' the skaters by season
#'
#' `skater_season_report()` retrieves various reports for a season, game type, and category for all the skaters by season as a `data.frame` where each row represents player and includes detail on date/season filtering windows and chronological context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams roster_statistics
#' @param category character (e.g., 'puckPossessions'); see 
#' [skater_report_configurations()] for reference
#'
#' @returns data.frame with one row per player
#' @examples
#' # May take >5s, so skip.
#' \donttest{possession_skater_season_report_playoff_20212022 <- 
#'   skater_season_report(
#'     season    = 20212022, 
#'     game_type = 3, 
#'     category  = 'puckPossessions'
#'   )}
#' @export

skater_season_report <- function(
  season    = season_now(), 
  game_type = game_type_now(), 
  category  = 'summary'
) {
  tryCatch(
    expr = {
      report <- nhl_api(
        path  = sprintf('en/skater/%s', category),
        query = list(
          limit       = -1,
          isAggregate = FALSE,
          isGame      = FALSE,
          cayenneExp  = sprintf(
            'seasonId = %s and gameTypeId = %s', 
            season,
            game_type
          )
        ),
        type  = 's'
      )$data
      report <- report[order(report$playerId), ]
      names(report)[names(report) == 'lastName'] <- 'skaterLastName'
      names(report) <- normalize_team_abbrev_cols(names(report))
      report
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access various reports for a season, game type, and category for all 
#' the skaters by game
#'
#' `skater_game_report()` retrieves various reports for a season, game type, and category for all the skaters by game as a `data.frame` where each row represents game per player and includes detail on game timeline state, period/clock progression, and matchup flow, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams skater_season_report
#'
#' @returns data.frame with one row per game per player
#' @examples
#' # May take >5s, so skip.
#' \donttest{possession_skater_game_report_playoff_20212022 <- 
#'   skater_game_report(
#'     season    = 20212022, 
#'     game_type = 3, 
#'     category  = 'puckPossessions'
#'   )}
#' @export

skater_game_report <- function(
  season    = season_now(), 
  game_type = game_type_now(), 
  category  = 'summary'
) {
  tryCatch(
    expr = {
      if (game_type != 2) {
        report <- nhl_api(
          path  = sprintf('en/skater/%s', category),
          query = list(
            limit       = -1,
            isAggregate = FALSE,
            isGame      = TRUE,
            cayenneExp  = sprintf(
              'seasonId = %s and gameTypeId = %s',
              season,
              game_type
            )
          ),
          type  = 's'
        )$data
      } else {
        seasons_tbl <- seasons()
        season_row  <- seasons_tbl[seasons_tbl$seasonId == season, ]
        if (!nrow(season_row)) {
          stop('No season metadata found for season: ', season)
        }
        start_date <- as.Date(season_row$startDate)
        end_date   <- as.Date(season_row$regularSeasonEndDate)
        all_dates <- seq(start_date, end_date, by = '1 day')
        date_groups <- split(all_dates, format(all_dates, '%Y-%m'))
        pages <- list()
        for (dates in date_groups) {
          dates_chr <- as.character(dates)
          cayenne <- sprintf(
            'seasonId = %s and gameTypeId = %s and gameDate in (%s)',
            season,
            game_type,
            paste0('\'', dates_chr, '\'', collapse = ',')
          )
          page <- nhl_api(
            path  = sprintf('en/skater/%s', category),
            query = list(
              limit       = -1,
              isAggregate = FALSE,
              isGame      = TRUE,
              cayenneExp  = cayenne
            ),
            type  = 's'
          )$data
          if (!is.null(page)) {
            if (!is.data.frame(page)) {
              page <- as.data.frame(page, stringsAsFactors = FALSE)
            }
            if (nrow(page)) {
              pages[[length(pages) + 1]] <- page
            }
          }
        }
        if (!length(pages)) {
          return(data.frame())
        }
        report <- do.call(rbind, pages)
      }
      if (is.null(report)) {
        return(data.frame())
      }
      if (!is.data.frame(report)) {
        report <- as.data.frame(report, stringsAsFactors = FALSE)
      }
      if (!nrow(report)) {
        return(report)
      }
      report <- report[order(report$playerId, report$gameId), ]
      names(report)[names(report) == 'lastName'] <- 'skaterLastName'
      names(report) <- normalize_team_abbrev_cols(names(report))
      report
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the career statistics for all the skaters
#'
#' `skater_statistics()` retrieves the career statistics for all the skaters as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per player
#' @examples
#' skater_stats <- skater_statistics()
#' @export

skater_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'skater-career-scoring-regular-plus-playoffs',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'skaterFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'skaterLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$playerId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname skater_statistics
#' @export

skater_stats <- function() {
  skater_statistics()
}

#' Access the career regular season statistics for all the skaters
#'
#' `skater_regular_statistics()` retrieves the career regular season statistics for all the skaters as a `data.frame` where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per player
#' @examples
#' skater_regular_stats <- skater_regular_statistics()
#' @export

skater_regular_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'skater-career-scoring-regular-season',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'skaterFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'skaterLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname skater_regular_statistics
#' @export

skater_regular_stats <- function() {
  skater_regular_statistics()
}

#' Access the career playoff statistics for all the skaters
#'
#' `skater_playoff_statistics()` retrieves the career playoff statistics for all the skaters as a `data.frame` where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per player
#' @examples
#' skater_playoff_stats <- skater_playoff_statistics()
#' @export

skater_playoff_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'skater-career-scoring-playoffs',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'skaterFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'skaterLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname skater_playoff_statistics
#' @export

skater_playoff_stats <- function() {
  skater_playoff_statistics()
}

#' Access the statistics for all the skaters by season, game type, and team
#'
#' `skater_season_statistics()` retrieves the statistics for all the skaters by season, game type, and team as a `data.frame` where each row represents player per season per game type, separated by team if applicable and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player per season per game type, 
#' separated by team if applicable
#' @examples
#' # May take >5s, so skip.
#' \donttest{skater_season_stats <- skater_season_statistics()}
#' @export

skater_season_statistics <- function() {
  tryCatch({
    stats                    <- nhl_api(
      path = 'player-stats',
      type = 'r'
    )$data
    stats$`id.db:SEQUENCENO` <- NULL
    stats <- stats[order(stats$`id.db:PLAYERID`, stats$`id.db:SEASON`), ]
    names(stats)[names(stats) == 'id.db:PLAYERID'] <- 'playerId'
    names(stats)[names(stats) == 'id.db:TEAMID']   <- 'teamId'
    names(stats)[names(stats) == 'id.db:SEASON']   <- 'seasonId'
    names(stats)[names(stats) == 'id.db:GAMETYPE'] <- 'gameTypeId'
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname skater_season_statistics
#' @export

skater_season_stats <- function() {
  skater_season_statistics()
}

#' Access the playoff statistics for all the skaters by series
#'
#' `skater_series_statistics()` retrieves the playoff statistics for all the skaters by series as a `data.frame` where each row represents player per series and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player per series
#' @examples
#' # May take >5s, so skip.
#' \donttest{skater_series_stats <- skater_series_statistics()}
#' @export

skater_series_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'playoff-skater-series-stats',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'skaterFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'skaterLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname skater_series_statistics
#' @export

skater_series_stats <- function() {
  skater_series_statistics()
}

#' Access the skater statistics leaders for a season, game type, and category
#'
#' `skater_leaders()` retrieves the skater statistics leaders for a season, game type, and category as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @inheritParams roster_statistics
#' @param category string of 'a'/'assists', 'g'/goals', 
#' 'shg'/'shorthanded goals', 'ppg'/'powerplay goals', 'p'/'points', 
#' 'pim'/penalty minutes'/'penalty infraction minutes', 'toi'/'time on ice', 
#' 'pm'/'plus minus', or 'f'/'faceoffs'
#'
#' @returns data.frame with one row per player
#' @examples
#' TOI_leaders_regular_20242025 <- skater_leaders(
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'TOI'
#' )
#' @export

skater_leaders <- function(
  season    = 'current',
  game_type = '',
  category  = 'points'
) {
  tryCatch(
    expr = {
      category <- switch(
        tolower(category),
        a                            = 'assists',
        assists                      = 'assists',
        g                            = 'goals',
        goals                        = 'goals',
        shg                          = 'goalsSh',
        `shothanded goals`           = 'goalsSh',
        ppg                          = 'goalsPp',
        `powerplay goals`            = 'goalsPp',
        p                            = 'points',
        points                       = 'points',
        pim                          = 'penaltyMins',
        `penalty minutes`            = 'penaltyMins',
        `penalty infraction minutes` = 'penaltyMins',
        toi                          = 'toi',
        `time on ice`                = 'toi',
        pm                           = 'plusMinus',
        `plus minus`                 = 'plusMinus',
        f                            = 'faceoffLeaders',
        faceoffs                     = 'faceoffLeaders'
      )
      skaters <- nhl_api(
        path  = sprintf('v1/skater-stats-leaders/%s/%s', season, game_type),
        type  = 'w'
      )[[category]]
      names(skaters)[names(skaters) == 'id']       <- 'playerId'
      names(skaters)[names(skaters) == 'position'] <- 'positionCode'
      names(skaters) <- normalize_locale_names(names(skaters))
      names(skaters) <- scope_person_name_cols(names(skaters), 'skater')
      names(skaters) <- normalize_team_abbrev_cols(names(skaters))
      skaters
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the skaters on milestone watch
#'
#' `skater_milestones()` retrieves the skaters on milestone watch as a `data.frame` where each row represents player and includes detail on date/season filtering windows and chronological context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per player
#' @examples
#' skater_milestones <- skater_milestones()
#' @export

skater_milestones <- function() {
  tryCatch({
    milestones <- nhl_api(
      path = 'en/milestones/skaters',
      type = 's'
    )$data
    milestones$id <- NULL
    names(milestones)[names(milestones) == 'firstName'] <- 'skaterFirstName'
    names(milestones)[names(milestones) == 'lastName']  <- 'skaterLastName'
    names(milestones) <- normalize_team_abbrev_cols(names(milestones))
    milestones
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the officials
#'
#' `officials()` retrieves all the officials as a `data.frame` where each row represents official and includes detail on player identity, role, handedness, and biographical profile plus coach/management/officiating identity and assignment history.
#'
#' @returns data.frame with one row per official
#' @examples
#' all_officials <- officials()
#' @export

officials <- function() {
  tryCatch({
    officials <- nhl_api(
      path = 'officials',
      type = 'r'
    )$data
    names(officials)[names(officials) == 'id']                   <- 'officialId'
    names(officials)[names(officials) == 'firstName']            <- 'officialFirstName'
    names(officials)[names(officials) == 'lastName']             <- 'officialLastName'
    names(officials)[names(officials) == 'referreeAssociationId'] <- 'refereeAssociationId'
    officials
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Calculate the expected goals for all the shots in (a) play-by-plays
#'
#' `calculate_expected_goals()` scores shot events with `nhlscraper`'s built-in
#' ridge expected-goals model. The runtime model is a fixed six-partition system:
#' `sd` (5v5), `ev` (other even strength), `pp` (power play), `sh`
#' (short-handed), `en` (empty net against), and `ps` (penalty shot; trained on
#' penalty-shot and shootout-style rows). The legacy `model` argument is
#' accepted for backward compatibility but ignored.
#'
#' @param play_by_play data.frame of play-by-play(s) using the current public
#'   schema returned by [gc_play_by_play()], [gc_play_by_plays()],
#'   [wsc_play_by_play()], or [wsc_play_by_plays()]. Legacy alias-only columns
#'   such as `typeDescKey`, `period`, `SOGFor`, `SOGAgainst`, and
#'   `SOGDifferential` are no longer backfilled by the xG scorer.
#' @param model deprecated legacy model selector; ignored
#' @returns data.frame with one row per event (play) and added `xG` column
#' @examples
#' # May take >5s, so skip.
#' \donttest{
#'   pbp <- gc_play_by_play()
#'   pbp_with_xg <- calculate_expected_goals(play_by_play = pbp)
#' }
#' @export
calculate_expected_goals <- function(play_by_play, model = NULL) {
  tryCatch(
    expr = {
      .xg_warn_ignored_model(model, "calculate_expected_goals")

      pbp <- .xg_prepare_play_by_play(play_by_play)
      n <- nrow(pbp)
      xg <- rep(NA_real_, n)

      is_shot <- .shot_event_mask(
        pbp,
        c("goal", "shot-on-goal", "missed-shot")
      )
      if (!any(is_shot)) {
        play_by_play$xG <- xg
        return(play_by_play)
      }

      shot_idx <- which(is_shot)
      shots <- .xg_build_model_frame(
        shots = pbp[shot_idx, , drop = FALSE],
        play_by_play = pbp
      )

      partition <- .xg_partition_shots(shots)
      goalie_ids <- unique(
        as.integer(
          pbp$goaliePlayerIdAgainst[!is.na(pbp$goaliePlayerIdAgainst)]
        )
      )
      shooter_ids <- as.integer(shots$shootingPlayerId)
      score_ok <- is.na(shooter_ids) | !(shooter_ids %in% goalie_ids)

      for (key in names(XG_RIDGE_MODEL_SPECS)) {
        idx <- which(score_ok & !is.na(partition) & partition == key)
        if (!length(idx)) {
          next
        }
        xg[shot_idx[idx]] <- .xg_score_partition(
          shots[idx, , drop = FALSE],
          XG_RIDGE_MODEL_SPECS[[key]]
        )
      }

      play_by_play$xG <- xg
      play_by_play
    },
    error = function(e) {
      message(conditionMessage(e))
      message("Invalid argument(s); refer to help file.")
      play_by_play
    }
  )
}

#' @rdname calculate_expected_goals
#' @export
calculate_xG <- function(play_by_play, model = NULL) {
  calculate_expected_goals(play_by_play, model)
}

.xg_warn_ignored_model <- function(model, fn_name) {
  if (missing(model) || is.null(model)) {
    return(invisible(NULL))
  }
  model_value <- suppressWarnings(as.integer(model[[1L]]))
  if (!is.na(model_value) && identical(model_value, 1L)) {
    return(invisible(NULL))
  }
  warning(
    sprintf(
      "`%s()` now uses one built-in six-partition ridge xG model; `model` is ignored.",
      fn_name
    ),
    call. = FALSE
  )
  invisible(NULL)
}

.xg_require_current_public_schema <- function(play_by_play) {
  legacy_aliases <- c(
    eventTypeDescKey = "typeDescKey",
    periodNumber = "period",
    shotsFor = "SOGFor",
    shotsAgainst = "SOGAgainst",
    shotDifferential = "SOGDifferential"
  )
  legacy_only <- names(legacy_aliases)[vapply(
    names(legacy_aliases),
    function(nm) {
      alt <- legacy_aliases[[nm]]
      !(nm %in% names(play_by_play)) && alt %in% names(play_by_play)
    },
    logical(1)
  )]

  if (length(legacy_only)) {
    replacements <- paste(
      sprintf(
        "%s -> %s",
        unname(legacy_aliases[legacy_only]),
        legacy_only
      ),
      collapse = ", "
    )
    stop(
      paste0(
        "calculate_expected_goals() requires the current public play-by-play schema. ",
        "Replace legacy xG alias columns with their public names: ",
        replacements
      ),
      call. = FALSE
    )
  }

  .require_public_pbp_columns(
    play_by_play,
    c(
      "gameId",
      "eventId",
      "sortOrder",
      "gameTypeId",
      "periodNumber",
      "eventOwnerTeamId",
      "eventTypeDescKey",
      "situationCode"
    ),
    "calculate_expected_goals"
  )

  invisible(NULL)
}

.xg_fill_goalie_against_fallback <- function(play_by_play) {
  if ("goalieInNetId" %in% names(play_by_play)) {
    goalie_in_net <- suppressWarnings(as.integer(play_by_play$goalieInNetId))
    if ("goaliePlayerIdAgainst" %in% names(play_by_play)) {
      goalie_against <- suppressWarnings(as.integer(play_by_play$goaliePlayerIdAgainst))
      goalie_in_net[is.na(goalie_in_net)] <- goalie_against[is.na(goalie_in_net)]
      play_by_play$goaliePlayerIdAgainst <- goalie_in_net
    } else {
      play_by_play$goaliePlayerIdAgainst <- goalie_in_net
    }
  }
  play_by_play
}

.xg_required_shift_cols <- function(play_by_play) {
  id_cols <- grep(
    "^(?:home|away)Skater[0-9]+PlayerId$|^skater[0-9]+PlayerId(?:For|Against)$",
    names(play_by_play),
    value = TRUE
  )
  if (!length(id_cols)) {
    return(character())
  }
  slot_count <- .on_ice_skater_slots(play_by_play = play_by_play)
  if (!slot_count) {
    return(character())
  }
  c(
    paste0("skater", seq_len(slot_count), "SecondsElapsedInShiftFor"),
    paste0("skater", seq_len(slot_count), "SecondsElapsedInShiftAgainst"),
    paste0("skater", seq_len(slot_count), "SecondsElapsedInPeriodSinceLastShiftFor"),
    paste0("skater", seq_len(slot_count), "SecondsElapsedInPeriodSinceLastShiftAgainst")
  )
}

.xg_fetch_shift_data <- function(game_ids) {
  game_ids <- sort(unique(as.integer(game_ids[!is.na(game_ids)])))
  if (!length(game_ids)) {
    return(data.frame())
  }

  if (length(game_ids) <= 4L) {
    out <- vector("list", length(game_ids))
    for (i in seq_along(game_ids)) {
      out[[i]] <- tryCatch(
        shift_chart(game_ids[[i]]),
        error = function(e) data.frame()
      )
    }
    out <- Filter(
      function(x) is.data.frame(x) && nrow(x) > 0L,
      out
    )
    if (!length(out)) {
      return(data.frame())
    }
    return(do.call(rbind, out))
  }

  season_ids <- sort(unique(game_ids %/% 1e6 * 1e4 + game_ids %/% 1e6 + 1L))
  out <- vector("list", length(season_ids))
  for (i in seq_along(season_ids)) {
    sc <- tryCatch(
      shift_charts(season_ids[[i]]),
      error = function(e) data.frame()
    )
    if (is.data.frame(sc) && nrow(sc) > 0L && "gameId" %in% names(sc)) {
      sc <- sc[sc$gameId %in% game_ids, , drop = FALSE]
    }
    out[[i]] <- sc
  }
  out <- Filter(
    function(x) is.data.frame(x) && nrow(x) > 0L,
    out
  )
  if (!length(out)) {
    data.frame()
  } else {
    do.call(rbind, out)
  }
}

.xg_prepare_play_by_play <- function(play_by_play) {
  .xg_require_current_public_schema(play_by_play)
  pbp <- .xg_fill_goalie_against_fallback(play_by_play)

  if (!("isHome" %in% names(pbp))) {
    pbp <- .flag_is_home(pbp)
  }

  state_cols <- c(
    "isEmptyNetFor",
    "isEmptyNetAgainst",
    "skaterCountFor",
    "skaterCountAgainst",
    "manDifferential",
    "strengthState"
  )
  if (!all(state_cols %in% names(pbp))) {
    pbp <- .strip_situation_code(pbp)
  }

  need_context <- !all(c(
    "isRush",
    "isRebound",
    "goalsFor",
    "goalsAgainst",
    "shotsFor",
    "shotsAgainst",
    "shotDifferential",
    "fenwickFor",
    "fenwickAgainst",
    "fenwickDifferential",
    "corsiFor",
    "corsiAgainst",
    "corsiDifferential"
  ) %in% names(pbp))
  if (need_context) {
    pbp <- .apply_shot_context(pbp)
  }

  shift_cols <- .xg_required_shift_cols(pbp)
  if (length(shift_cols) && !all(shift_cols %in% names(pbp))) {
    shifts <- .xg_fetch_shift_data(pbp$gameId)
    if (nrow(shifts) > 0L) {
      pbp <- tryCatch(
        add_shift_times(pbp, shifts),
        error = function(e) pbp
      )
    }
  }

  if (!all(.pbp_delta_public_cols() %in% names(pbp))) {
    pbp <- add_deltas(pbp)
  }

  if (!all(c(
    "shooterHeight",
    "shooterWeight",
    "shooterHandCode",
    "shooterPositionCode",
    "shooterAge"
  ) %in% names(pbp))) {
    pbp <- add_shooter_biometrics(pbp)
  }

  if (!all(c(
    "goalieHeight",
    "goalieWeight",
    "goalieHandCode",
    "goalieAge"
  ) %in% names(pbp))) {
    pbp <- add_goalie_biometrics(pbp)
  }

  .xg_fill_goalie_against_fallback(pbp)
}

.xg_normalize_shot_type <- function(x) {
  x <- tolower(trimws(as.character(x)))
  keep <- c("backhand", "deflected", "slap", "snap", "tip-in", "wrist")
  ifelse(!is.na(x) & x %in% keep, x, "other")
}

.xg_normalize_missed_reason <- function(x) {
  x <- tolower(trimws(as.character(x)))
  out <- rep("other", length(x))
  out[!is.na(x) & x %in% c(
    "goalpost", "hit-left-post", "hit-right-post", "hit-crossbar"
  )] <- "post"
  out[!is.na(x) & x %in% c("over-net", "above-crossbar")] <- "high"
  out[!is.na(x) & x %in% c(
    "wide-of-net",
    "high-and-wide-left",
    "high-and-wide-right",
    "wide-left",
    "wide-right"
  )] <- "wide"
  out
}

.xg_make_type_desc_key_prev <- function(
  type_desc_key_prev,
  reason_prev,
  shot_type_prev,
  event_owner_team_id_prev,
  event_owner_team_id
) {
  n <- length(type_desc_key_prev)
  out <- rep(NA_character_, n)
  prev_type <- tolower(as.character(type_desc_key_prev))
  reason_prev <- .xg_normalize_missed_reason(reason_prev)
  shot_type_prev <- .xg_normalize_shot_type(shot_type_prev)
  is_for <- !is.na(event_owner_team_id_prev) &
    !is.na(event_owner_team_id) &
    event_owner_team_id_prev == event_owner_team_id

  idx <- !is.na(prev_type) & prev_type == "faceoff" & is_for
  out[idx] <- "won-faceoff"
  idx <- !is.na(prev_type) & prev_type == "faceoff" & !is_for
  out[idx] <- "lost-faceoff"

  idx <- !is.na(prev_type) & prev_type == "shot-on-goal" & is_for
  out[idx] <- paste0(shot_type_prev[idx], "-shot-on-goal-for")
  idx <- !is.na(prev_type) & prev_type == "shot-on-goal" & !is_for
  out[idx] <- paste0(shot_type_prev[idx], "-shot-on-goal-against")

  idx <- !is.na(prev_type) & prev_type == "hit" & is_for
  out[idx] <- "given-hit"
  idx <- !is.na(prev_type) & prev_type == "hit" & !is_for
  out[idx] <- "taken-hit"

  idx <- !is.na(prev_type) & prev_type == "blocked-shot" & is_for
  out[idx] <- "blocked-shot-for"
  idx <- !is.na(prev_type) & prev_type == "blocked-shot" & !is_for
  out[idx] <- "blocked-shot-against"

  idx <- !is.na(prev_type) & prev_type == "giveaway" & is_for
  out[idx] <- "giveaway-for"
  idx <- !is.na(prev_type) & prev_type == "giveaway" & !is_for
  out[idx] <- "giveaway-against"

  idx <- !is.na(prev_type) & prev_type == "takeaway" & is_for
  out[idx] <- "takeaway-for"
  idx <- !is.na(prev_type) & prev_type == "takeaway" & !is_for
  out[idx] <- "takeaway-against"

  idx <- !is.na(prev_type) & prev_type == "missed-shot" & is_for
  out[idx] <- paste0(reason_prev[idx], "-missed-shot-for")
  idx <- !is.na(prev_type) & prev_type == "missed-shot" & !is_for
  out[idx] <- paste0(reason_prev[idx], "-missed-shot-against")

  other <- is.na(out) & !is.na(prev_type)
  out[other] <- prev_type[other]
  out
}

.xg_extract_slot_indices <- function(data, suffix) {
  matches <- regexec(
    paste0("^skater([0-9]+)", suffix, "$"),
    names(data)
  )
  vals <- regmatches(names(data), matches)
  idx <- vapply(
    vals,
    function(x) {
      if (length(x) < 2L) {
        return(NA_integer_)
      }
      as.integer(x[[2L]])
    },
    integer(1L)
  )
  sort(unique(idx[!is.na(idx)]))
}

.xg_build_skater_matrix <- function(data, suffix, mode = "numeric") {
  idx <- .xg_extract_slot_indices(data, suffix)
  if (!length(idx)) {
    out <- matrix(numeric(0), nrow = nrow(data), ncol = 0L)
    storage.mode(out) <- mode
    return(out)
  }
  cols <- paste0("skater", idx, suffix)
  out <- matrix(NA_real_, nrow = nrow(data), ncol = length(cols))
  for (j in seq_along(cols)) {
    if (cols[[j]] %in% names(data)) {
      out[, j] <- suppressWarnings(as.numeric(data[[cols[[j]]]]))
    }
  }
  if (identical(mode, "integer")) {
    storage.mode(out) <- "integer"
  }
  out
}

.xg_row_min <- function(mat) {
  n <- nrow(mat)
  if (!ncol(mat)) {
    return(rep(NA_real_, n))
  }
  out <- rep(Inf, n)
  any_ok <- rep(FALSE, n)
  for (j in seq_len(ncol(mat))) {
    col <- mat[, j]
    ok <- !is.na(col)
    any_ok[ok] <- TRUE
    out[ok] <- pmin(out[ok], col[ok])
  }
  out[!any_ok] <- NA_real_
  out
}

.xg_row_max <- function(mat) {
  n <- nrow(mat)
  if (!ncol(mat)) {
    return(rep(NA_real_, n))
  }
  out <- rep(-Inf, n)
  any_ok <- rep(FALSE, n)
  for (j in seq_len(ncol(mat))) {
    col <- mat[, j]
    ok <- !is.na(col)
    any_ok[ok] <- TRUE
    out[ok] <- pmax(out[ok], col[ok])
  }
  out[!any_ok] <- NA_real_
  out
}

.xg_row_mean <- function(mat) {
  if (!ncol(mat)) {
    return(rep(NA_real_, nrow(mat)))
  }
  counts <- rowSums(!is.na(mat))
  sums <- rowSums(mat, na.rm = TRUE)
  out <- sums / counts
  out[counts == 0L] <- NA_real_
  out
}

.xg_extract_matched_value <- function(id_mat, value_mat, player_id) {
  n <- nrow(id_mat)
  out <- rep(NA_real_, n)
  if (!ncol(id_mat) || !ncol(value_mat)) {
    return(out)
  }
  player_id <- suppressWarnings(as.integer(player_id))
  for (j in seq_len(ncol(id_mat))) {
    hit <- is.na(out) &
      !is.na(player_id) &
      !is.na(id_mat[, j]) &
      id_mat[, j] == player_id
    if (any(hit)) {
      out[hit] <- value_mat[hit, j]
    }
  }
  out
}

.xg_to_logical <- function(x) {
  if (is.logical(x)) {
    return(x)
  }
  if (is.numeric(x) || is.integer(x)) {
    out <- rep(NA, length(x))
    out[!is.na(x)] <- x[!is.na(x)] != 0
    return(out)
  }
  x <- as.character(x)
  out <- rep(NA, length(x))
  out[!is.na(x) & x %in% c("TRUE", "T", "true", "1", "yes", "YES")] <- TRUE
  out[!is.na(x) & x %in% c("FALSE", "F", "false", "0", "no", "NO")] <- FALSE
  out
}

.xg_build_model_frame <- function(shots, play_by_play) {
  shots <- .xg_fill_goalie_against_fallback(shots)
  play_by_play <- .xg_fill_goalie_against_fallback(play_by_play)
  n <- nrow(shots)

  prev_keys <- if ("eventIdPrev" %in% names(shots)) {
    paste(shots$gameId, shots$eventIdPrev, sep = ":")
  } else {
    rep(NA_character_, n)
  }
  cur_keys <- paste(play_by_play$gameId, play_by_play$eventId, sep = ":")
  prev_idx <- match(prev_keys, cur_keys)

  shot_type_prev <- if ("shotType" %in% names(play_by_play)) {
    .xg_normalize_shot_type(play_by_play$shotType[prev_idx])
  } else {
    rep("other", n)
  }

  shooting_player_id <- if ("shootingPlayerId" %in% names(shots)) {
    suppressWarnings(as.integer(shots$shootingPlayerId))
  } else {
    rep(NA_integer_, n)
  }
  if ("scoringPlayerId" %in% names(shots)) {
    use_scoring <- is.na(shooting_player_id)
    shooting_player_id[use_scoring] <- suppressWarnings(
      as.integer(shots$scoringPlayerId[use_scoring])
    )
  }

  if (!("gameTypeId" %in% names(shots)) && "gameId" %in% names(shots)) {
    shots$gameTypeId <- shots$gameId %/% 1e4 %% 1e2
  }

  shot_type <- if ("shotType" %in% names(shots)) {
    shots$shotType
  } else {
    rep(NA_character_, n)
  }
  period_type <- if ("periodType" %in% names(shots)) {
    shots$periodType
  } else {
    rep(NA_character_, n)
  }
  x_coord_norm <- if ("xCoordNorm" %in% names(shots)) {
    suppressWarnings(as.numeric(shots$xCoordNorm))
  } else {
    rep(NA_real_, n)
  }
  y_coord_norm <- if ("yCoordNorm" %in% names(shots)) {
    suppressWarnings(as.numeric(shots$yCoordNorm))
  } else {
    rep(NA_real_, n)
  }
  d_y_coord_norm <- if ("dYCoordNorm" %in% names(shots)) {
    suppressWarnings(as.numeric(shots$dYCoordNorm))
  } else {
    rep(NA_real_, n)
  }

  shots$shotType <- .xg_normalize_shot_type(shot_type)
  shots$typeDescKeyPrev <- .xg_make_type_desc_key_prev(
    type_desc_key_prev = play_by_play$eventTypeDescKey[prev_idx],
    reason_prev = play_by_play$reason[prev_idx],
    shot_type_prev = shot_type_prev,
    event_owner_team_id_prev = play_by_play$eventOwnerTeamId[prev_idx],
    event_owner_team_id = shots$eventOwnerTeamId
  )
  shots$shootingPlayerId <- shooting_player_id

  is_empty_for <- .xg_to_logical(
    if ("isEmptyNetFor" %in% names(shots)) shots$isEmptyNetFor else rep(NA, n)
  )
  is_empty_against <- .xg_to_logical(
    if ("isEmptyNetAgainst" %in% names(shots)) shots$isEmptyNetAgainst else rep(NA, n)
  )
  is_empty_for[is.na(is_empty_for)] <- FALSE
  is_empty_against[is.na(is_empty_against)] <- FALSE
  shots$isEmptyNetFor <- is_empty_for
  shots$isEmptyNetAgainst <- is_empty_against
  shots$isPlayoff <- !is.na(shots$gameTypeId) & shots$gameTypeId == 3L
  shots$isOvertime <- !is.na(period_type) &
    as.character(period_type) == "OT"
  shots$xCoordNorm <- x_coord_norm
  shots$yCoordNorm <- y_coord_norm
  shots$dYCoordNorm <- d_y_coord_norm
  shots$isBehindNet <- !is.na(x_coord_norm) & x_coord_norm >= 89
  y_prev <- y_coord_norm - d_y_coord_norm
  shots$crossedRoyalRoad <- !is.na(y_coord_norm) &
    !is.na(y_prev) &
    y_coord_norm * y_prev < 0

  shots$zoneCode <- if ("zoneCode" %in% names(shots)) {
    toupper(as.character(shots$zoneCode))
  } else {
    rep(NA_character_, n)
  }
  shots$strengthState <- if ("strengthState" %in% names(shots)) {
    tolower(as.character(shots$strengthState))
  } else {
    rep(NA_character_, n)
  }
  shots$shooterHandCode <- if ("shooterHandCode" %in% names(shots)) {
    toupper(as.character(shots$shooterHandCode))
  } else {
    rep(NA_character_, n)
  }
  shots$goalieHandCode <- if ("goalieHandCode" %in% names(shots)) {
    toupper(as.character(shots$goalieHandCode))
  } else {
    rep(NA_character_, n)
  }
  shots$shooterPositionCode <- if ("shooterPositionCode" %in% names(shots)) {
    toupper(as.character(shots$shooterPositionCode))
  } else {
    rep(NA_character_, n)
  }

  player_ids_for <- .xg_build_skater_matrix(shots, "PlayerIdFor", "integer")
  shift_for <- .xg_build_skater_matrix(shots, "SecondsElapsedInShiftFor")
  shift_against <- .xg_build_skater_matrix(shots, "SecondsElapsedInShiftAgainst")
  rest_for <- .xg_build_skater_matrix(shots, "SecondsElapsedInPeriodSinceLastShiftFor")
  rest_against <- .xg_build_skater_matrix(shots, "SecondsElapsedInPeriodSinceLastShiftAgainst")

  shots$minSecondsElapsedInShiftFor <- .xg_row_min(shift_for)
  shots$maxSecondsElapsedInShiftFor <- .xg_row_max(shift_for)
  shots$avgSecondsElapsedInShiftFor <- .xg_row_mean(shift_for)
  shots$minSecondsElapsedInShiftAgainst <- .xg_row_min(shift_against)
  shots$maxSecondsElapsedInShiftAgainst <- .xg_row_max(shift_against)
  shots$avgSecondsElapsedInShiftAgainst <- .xg_row_mean(shift_against)
  shots$minSecondsElapsedSinceLastShiftFor <- .xg_row_min(rest_for)
  shots$maxSecondsElapsedSinceLastShiftFor <- .xg_row_max(rest_for)
  shots$avgSecondsElapsedSinceLastShiftFor <- .xg_row_mean(rest_for)
  shots$minSecondsElapsedSinceLastShiftAgainst <- .xg_row_min(rest_against)
  shots$maxSecondsElapsedSinceLastShiftAgainst <- .xg_row_max(rest_against)
  shots$avgSecondsElapsedSinceLastShiftAgainst <- .xg_row_mean(rest_against)
  shots$shooterSecondsElapsedInShift <- .xg_extract_matched_value(
    player_ids_for,
    shift_for,
    shots$shootingPlayerId
  )
  shots$shooterSecondsElapsedSinceLastShift <- .xg_extract_matched_value(
    player_ids_for,
    rest_for,
    shots$shootingPlayerId
  )

  shots
}

.xg_partition_shots <- function(shots) {
  sc <- .normalize_situation_code_for_parse(shots$situationCode)
  is_empty_for <- .xg_to_logical(shots$isEmptyNetFor)
  is_empty_against <- .xg_to_logical(shots$isEmptyNetAgainst)
  is_empty_for[is.na(is_empty_for)] <- FALSE
  is_empty_against[is.na(is_empty_against)] <- FALSE

  skater_for <- suppressWarnings(as.integer(shots$skaterCountFor))
  skater_against <- suppressWarnings(as.integer(shots$skaterCountAgainst))

  is_ps <- !is.na(sc) & sc %in% c("1010", "0101")
  is_en <- !is_ps & is_empty_against
  is_sd_standard <- (!is_ps &
    !is_en &
    !is.na(skater_for) &
    !is.na(skater_against) &
    skater_for == 5L &
    skater_against == 5L &
    !is_empty_for &
    !is_empty_against)
  is_ev <- !is_ps &
    !is_en &
    !is.na(skater_for) &
    !is.na(skater_against) &
    skater_for == skater_against &
    !is_sd_standard
  is_pp <- !is_ps &
    !is_en &
    !is.na(skater_for) &
    !is.na(skater_against) &
    skater_for > skater_against
  is_sh <- !is_ps &
    !is_en &
    !is.na(skater_for) &
    !is.na(skater_against) &
    skater_for < skater_against
  is_uncategorizable_partition <- !(
    is_ps |
    is_en |
    is_sd_standard |
    is_ev |
    is_pp |
    is_sh
  )
  is_sd <- is_sd_standard | is_uncategorizable_partition
  is_ps[is.na(is_ps)] <- FALSE
  is_en[is.na(is_en)] <- FALSE
  is_sd[is.na(is_sd)] <- FALSE
  is_ev[is.na(is_ev)] <- FALSE
  is_pp[is.na(is_pp)] <- FALSE
  is_sh[is.na(is_sh)] <- FALSE

  out <- rep(NA_character_, nrow(shots))
  out[is_ps] <- "ps"
  out[is_en] <- "en"
  out[is_sd] <- "sd"
  out[is_ev] <- "ev"
  out[is_pp] <- "pp"
  out[is_sh] <- "sh"
  out
}

.xg_encode_categorical <- function(values, var, spec, n) {
  if (is.null(values)) {
    values <- rep(NA, n)
  }

  logical_var <- var %in% spec$logical_vars
  if (logical_var) {
    values <- .xg_to_logical(values)
    out <- rep(NA_character_, length(values))
    out[!is.na(values) & values] <- "yes"
    out[!is.na(values) & !values] <- "no"
  } else {
    out <- as.character(values)
  }

  known <- spec$known_levels[[var]]
  known <- if (is.null(known)) character() else known
  allowed <- c(
    known,
    if (isTRUE(spec$has_unknown[[var]])) "unknown",
    if (isTRUE(spec$has_new[[var]])) "new"
  )

  out[is.na(out) & isTRUE(spec$has_unknown[[var]])] <- "unknown"
  novel <- !is.na(out) & !(out %in% allowed)
  if (isTRUE(spec$has_new[[var]])) {
    out[novel] <- "new"
  }

  out
}

.xg_standardize_term <- function(x, term, spec, n) {
  if (is.null(x)) {
    x <- rep(NA_real_, n)
  }
  if (is.factor(x)) {
    x <- as.character(x)
  }
  if (is.logical(x)) {
    x <- ifelse(is.na(x), NA_real_, as.numeric(x))
  } else {
    x <- suppressWarnings(as.numeric(x))
  }

  med <- spec$medians[[term]]
  if (is.null(med) || !length(med) || is.na(med)) {
    med <- 0
  }
  x[is.na(x)] <- med

  center <- spec$normalize_means[[term]]
  scale <- spec$normalize_sds[[term]]
  if (is.null(center) || !length(center) || is.na(center)) {
    center <- 0
  }
  if (is.null(scale) || !length(scale) || !is.finite(scale) || scale == 0) {
    return(rep(0, n))
  }
  (x - center) / scale
}

.xg_score_partition <- function(df, spec) {
  n <- nrow(df)
  if (!n) {
    return(numeric(0))
  }

  coeffs <- spec$coefficients
  eta <- rep(unname(coeffs[["(Intercept)"]]), n)

  if (length(spec$raw_terms)) {
    for (term in spec$raw_terms) {
      eta <- eta + coeffs[[term]] * .xg_standardize_term(
        if (term %in% names(df)) df[[term]] else NULL,
        term,
        spec,
        n
      )
    }
  }

  if (length(spec$active_dummy_by_var)) {
    for (var in names(spec$active_dummy_by_var)) {
      values <- .xg_encode_categorical(
        if (var %in% names(df)) df[[var]] else NULL,
        var,
        spec,
        n
      )
      terms <- spec$active_dummy_by_var[[var]]
      for (term in terms) {
        level <- spec$dummy_term_to_level[[term]]
        dummy <- as.numeric(values == level)
        dummy[is.na(dummy)] <- NA_real_
        eta <- eta + coeffs[[term]] * .xg_standardize_term(
          dummy,
          term,
          spec,
          n
        )
      }
    }
  }

  stats::plogis(eta)
}

test_that("xG ridge partitioning covers the six shot situations", {
  shots <- data.frame(
    situationCode = c("1551", "1441", "1541", "1451", "0650", "1010"),
    isEmptyNetFor = c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE),
    isEmptyNetAgainst = c(FALSE, FALSE, FALSE, FALSE, TRUE, FALSE),
    skaterCountFor = c(5L, 4L, 5L, 4L, 6L, 1L),
    skaterCountAgainst = c(5L, 4L, 4L, 5L, 5L, 1L),
    stringsAsFactors = FALSE
  )

  expect_equal(
    nhlscraper:::.xg_partition_shots(shots),
    c("sd", "ev", "pp", "sh", "en", "ps")
  )
})

test_that("xG ridge partitioning routes uncategorizable rows to sd", {
  shots <- data.frame(
    situationCode = c(NA_character_, "1551", "1541", "1551"),
    isEmptyNetFor = c(FALSE, FALSE, FALSE, FALSE),
    isEmptyNetAgainst = c(FALSE, FALSE, FALSE, FALSE),
    skaterCountFor = c(NA_integer_, 5L, 4L, "oops"),
    skaterCountAgainst = c(NA_integer_, 5L, NA_integer_, 5L),
    stringsAsFactors = FALSE
  )

  expect_equal(
    nhlscraper:::.xg_partition_shots(shots),
    c("sd", "sd", "sd", "sd")
  )
})

test_that("xG ridge partitioning does not force sd on missing situationCode alone", {
  shots <- data.frame(
    situationCode = c(NA_character_, NA_character_),
    isEmptyNetFor = c(FALSE, FALSE),
    isEmptyNetAgainst = c(FALSE, FALSE),
    skaterCountFor = c(5L, 4L),
    skaterCountAgainst = c(4L, 4L),
    stringsAsFactors = FALSE
  )

  expect_equal(
    nhlscraper:::.xg_partition_shots(shots),
    c("pp", "ev")
  )
})

test_that("xG scorer rejects legacy alias-only public columns", {
  play_by_play <- data.frame(
    gameId = 1L,
    eventId = 1L,
    sortOrder = 1L,
    gameTypeId = 2L,
    period = 1L,
    eventOwnerTeamId = 1L,
    typeDescKey = "shot-on-goal",
    situationCode = "1551",
    stringsAsFactors = FALSE
  )

  expect_error(
    nhlscraper:::.xg_require_current_public_schema(play_by_play),
    "requires the current public play-by-play schema"
  )
})

test_that("xG categorical encoding keeps known baseline levels out of new bucket", {
  spec <- nhlscraper:::XG_RIDGE_MODEL_SPECS$sd

  encoded <- nhlscraper:::.xg_encode_categorical(
    values = c("backhand", "wrist", "mystery", NA),
    var = "shotType",
    spec = spec,
    n = 4L
  )

  expect_equal(encoded, c("backhand", "wrist", "new", "unknown"))
})

test_that("xG partition scoring returns finite probabilities with sparse inputs", {
  spec <- nhlscraper:::XG_RIDGE_MODEL_SPECS$ps
  df <- data.frame(
    shotType = c("backhand", "snap"),
    shooterHandCode = c("L", "R"),
    shooterPositionCode = c("C", "D"),
    goalieHandCode = c("L", "R"),
    isPlayoff = c(FALSE, TRUE),
    isHome = c(TRUE, FALSE),
    xCoordNorm = c(75, 78),
    yCoordNorm = c(0, 3),
    distance = c(13, 11),
    angle = c(12, 18),
    goalsFor = c(3, 4),
    goalsAgainst = c(3, 2),
    goalDifferential = c(0, 2),
    shotsFor = c(31, 34),
    shotsAgainst = c(32, 29),
    shotDifferential = c(-1, 5),
    fenwickFor = c(46, 50),
    fenwickAgainst = c(46, 44),
    fenwickDifferential = c(0, 6),
    corsiFor = c(64, 69),
    corsiAgainst = c(64, 61),
    corsiDifferential = c(0, 8),
    shooterHeight = c(73, 72),
    shooterWeight = c(198, 190),
    shooterAge = c(27, 25),
    goalieHeight = c(75, 74),
    goalieWeight = c(201, 198),
    goalieAge = c(28, 29),
    stringsAsFactors = FALSE
  )

  probs <- nhlscraper:::.xg_score_partition(df, spec)

  expect_length(probs, 2L)
  expect_true(all(is.finite(probs)))
  expect_true(all(probs > 0 & probs < 1))
})

test_that("xG scorer prefers goalieInNetId over goaliePlayerIdAgainst", {
  play_by_play <- data.frame(
    eventTypeDescKey = "shot-on-goal",
    periodNumber = 1L,
    shotsFor = 10L,
    shotsAgainst = 8L,
    shotDifferential = 2L,
    goaliePlayerIdAgainst = c(66L, NA_integer_),
    goalieInNetId = c(77L, 88L),
    stringsAsFactors = FALSE
  )

  out <- nhlscraper:::.xg_fill_goalie_against_fallback(play_by_play)

  expect_equal(out$goaliePlayerIdAgainst, c(77L, 88L))
})

#' Access all contracts from packaged internal data
#'
#' `contracts()` loads preprocessed contract records bundled with the package and returns a cleaned `data.frame` with package-consistent column names, season IDs, numeric money fields, and team/player identifiers.
#'
#' @returns data.frame with one row per contract
#' @examples
#' \donttest{all_contracts <- contracts()}
#' @export

contracts <- function() {
  tryCatch(
    expr = {
      normalize_person_key <- function(x) {
        x <- as.character(x)
        x <- iconv(x, from = '', to = 'ASCII//TRANSLIT')
        x[is.na(x)] <- ''
        x <- tolower(trimws(x))
        x <- gsub('[^a-z0-9]', '', x)
        x
      }
      allowed_positions <- function(pos) {
        switch(
          toupper(trimws(as.character(pos))),
          C  = 'C',
          D  = 'D',
          G  = 'G',
          LW = c('L', 'LW'),
          RW = c('R', 'RW'),
          F  = c('C', 'L', 'R', 'F'),
          character(0)
        )
      }
      out <- get('.contracts_base', inherits = TRUE)
      dropped_duplicate_rows <- attr(out, 'droppedDuplicateRows')
      dropped_invalid_season <- attr(out, 'droppedInvalidSeasonRows')
      if (is.null(dropped_duplicate_rows)) dropped_duplicate_rows <- NA_integer_
      if (is.null(dropped_invalid_season)) dropped_invalid_season <- NA_integer_
      if (!is.data.frame(out) || !nrow(out)) {
        out <- data.frame(
          playerId = integer(),
          playerFullName = character(),
          positionCode = character(),
          ageAtSigning = integer(),
          signedWithTeamId = integer(),
          signedWithTeamTriCode = character(),
          startSeasonId = integer(),
          endSeasonId = integer(),
          term = integer(),
          aav = numeric(),
          value = numeric(),
          bonus = numeric(),
          stringsAsFactors = FALSE
        )
        attr(out, 'droppedDuplicateRows') <- dropped_duplicate_rows
        attr(out, 'droppedInvalidSeasonRows') <- dropped_invalid_season
        attr(out, 'droppedUnresolvedPlayers') <- NA_integer_
        return(out)
      }
      players_tbl  <- players()
      out$playerId <- NA_integer_
      dropped_unresolved <- NA_integer_
      if (is.data.frame(players_tbl) && nrow(players_tbl) > 0L) {
        req_cols <- c(
          'playerId',
          'playerFullName',
          'playerFirstName',
          'playerLastName',
          'positionCode',
          'birthDate',
          'currentTeamId',
          'careerTeamId',
          'firstSignedByTeamId',
          'lastNHLTeamId',
          'onRoster'
        )
        req_cols <- req_cols[req_cols %in% names(players_tbl)]
        p <- players_tbl[, req_cols, drop = FALSE]
        p$birthYear <- suppressWarnings(as.integer(substr(as.character(p$birthDate), 1L, 4L)))
        p$key1  <- normalize_person_key(p$playerFullName)
        p$key2  <- normalize_person_key(paste(p$playerFirstName, p$playerLastName))
        idx_tbl <- rbind(
          data.frame(key = p$key1, idx = seq_len(nrow(p)), stringsAsFactors = FALSE),
          data.frame(key = p$key2, idx = seq_len(nrow(p)), stringsAsFactors = FALSE)
        )
        idx_tbl <- idx_tbl[nzchar(idx_tbl$key), , drop = FALSE]
        idx_tbl <- unique(idx_tbl)
        key_map <- split(idx_tbl$idx, idx_tbl$key)
        c_key <- normalize_person_key(out$playerFullName)
        c_pos <- out$positionCode
        c_tid <- out$teamId
        c_start_year <- out$startSeasonId %/% 10000L
        c_age <- out$ageAtSigning
        matched_player_id <- rep(NA_integer_, nrow(out))
        matched_player_full_name <- rep(NA_character_, nrow(out))
        matched_position_code <- rep(NA_character_, nrow(out))
        for (i in seq_len(nrow(out))) {
          idx <- key_map[[c_key[[i]]]]
          if (is.null(idx) || !length(idx)) next
          idx <- unique(idx)
          pos_allowed <- allowed_positions(c_pos[[i]])
          if (length(pos_allowed) && 'positionCode' %in% names(p)) {
            pos_keep <- p$positionCode[idx] %in% pos_allowed
            if (any(pos_keep, na.rm = TRUE)) idx <- idx[pos_keep]
          }
          if ('birthYear' %in% names(p) && !is.na(c_start_year[[i]]) && !is.na(c_age[[i]])) {
            expected <- c(
              c_start_year[[i]] - c_age[[i]] - 1L,
              c_start_year[[i]] - c_age[[i]],
              c_start_year[[i]] - c_age[[i]] + 1L
            )
            by_keep <- p$birthYear[idx] %in% expected
            if (any(by_keep, na.rm = TRUE)) idx <- idx[by_keep]
          }
          if (!is.na(c_tid[[i]])) {
            team_hits <- rep(FALSE, length(idx))
            team_cols <- c('currentTeamId', 'careerTeamId', 'firstSignedByTeamId', 'lastNHLTeamId')
            team_cols <- team_cols[team_cols %in% names(p)]
            if (length(team_cols)) {
              for (tc in team_cols) {
                team_hits <- team_hits | (p[[tc]][idx] == c_tid[[i]])
              }
              if (any(team_hits, na.rm = TRUE)) idx <- idx[team_hits]
            }
          }
          if (length(idx) > 1L && 'onRoster' %in% names(p) && !is.na(c_start_year[[i]]) && c_start_year[[i]] >= 2020L) {
            roster_keep <- p$onRoster[idx] == 'Y'
            if (any(roster_keep, na.rm = TRUE)) idx <- idx[roster_keep]
          }
          idx <- unique(idx)
          if (length(idx) == 1L) {
            matched_player_id[[i]] <- suppressWarnings(as.integer(p$playerId[idx]))
            matched_player_full_name[[i]] <- as.character(p$playerFullName[idx])
            if ('positionCode' %in% names(p)) {
              matched_position_code[[i]] <- as.character(p$positionCode[idx])
            }
          }
        }
        out$playerId <- matched_player_id
        out$playerFullName <- ifelse(
          is.na(matched_player_full_name) | matched_player_full_name == '',
          out$playerFullName,
          matched_player_full_name
        )
        out$positionCode <- ifelse(
          is.na(matched_position_code) | matched_position_code == '',
          out$positionCode,
          matched_position_code
        )
        unresolved <- is.na(out$playerId)
        dropped_unresolved <- sum(unresolved)
        out <- out[!unresolved, , drop = FALSE]
        message(
          sprintf(
            'Dropped %s row(s) with unresolved/ambiguous playerId matches.',
            dropped_unresolved
          )
        )
      } else {
        message(
          'Unable to access players() for playerId matching; ',
          'returning cleaned contracts with NA playerId.'
        )
      }
      out <- out[, c(
        'playerId',
        'playerFullName',
        'positionCode',
        'ageAtSigning',
        'signedWithTeamId',
        'signedWithTriCode',
        'startSeasonId',
        'endSeasonId',
        'contractYears',
        'contractAAV',
        'contractValue',
        'signingBonus'
      )]
      names(out)[names(out) == 'signedWithTriCode'] <- 'signedWithTeamTriCode'
      names(out)[names(out) == 'contractYears'] <- 'term'
      names(out)[names(out) == 'contractAAV']   <- 'aav'
      names(out)[names(out) == 'contractValue'] <- 'value'
      names(out)[names(out) == 'signingBonus']  <- 'bonus'
      out <- out[order(out$playerFullName, out$startSeasonId), , drop = FALSE]
      rownames(out) <- NULL
      attr(out, 'droppedDuplicateRows') <- dropped_duplicate_rows
      attr(out, 'droppedInvalidSeasonRows') <- dropped_invalid_season
      attr(out, 'droppedUnresolvedPlayers') <- dropped_unresolved
      out
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the raw GameCenter (GC) play-by-plays for a season
#' 
#' `gc_play_by_plays_raw()` loads the raw GC play-by-plays for a given
#' `season`.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per raw event (play) per game
#' @examples
#' # May take >5s, so skip.
#' \donttest{gc_pbps_raw_20212022 <- gc_play_by_plays_raw(season = 20212022)}
#' @export

gc_play_by_plays_raw <- function(season = 20242025) {
  tryCatch(
    expr = {
      u <- paste0(
        'https://huggingface.co/datasets/RentoSaijo/NHL_DB/resolve/main/',
        'data/game/pbps/gc/NHL_PBPS_GC_Raw_',
        season,
        '.parquet'
      )
      tmp <- tempfile(fileext = '.parquet')
      utils::download.file(u, tmp, mode = 'wb', quiet = TRUE)
      as.data.frame(arrow::read_parquet(tmp), stringsAsFactors = FALSE)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname gc_play_by_plays_raw
#' @export

gc_pbps_raw <- function(season = 20242025) {
  gc_play_by_plays_raw(season)
}

#' Access the GameCenter (GC) play-by-plays for a season
#' 
#' `gc_play_by_plays()` loads the GC play-by-plays for a given `season`.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per event (play) per game
#' @examples
#' # May take >5s, so skip.
#' \donttest{gc_pbps_20212022 <- gc_play_by_plays(season = 20212022)}
#' @export

gc_play_by_plays <- function(season = 20242025) {
  tryCatch(
    expr = {
      u <- paste0(
        'https://huggingface.co/datasets/RentoSaijo/NHL_DB/resolve/main/',
        'data/game/pbps/gc/NHL_PBPS_GC_',
        season,
        '.parquet'
      )
      tmp <- tempfile(fileext = '.parquet')
      utils::download.file(u, tmp, mode = 'wb', quiet = TRUE)
      pbps <- as.data.frame(arrow::read_parquet(tmp), stringsAsFactors = FALSE)
      raw_situation <- pbps[['situationCode']]
      situation_chr <- as.character(raw_situation)
      situation_pad <- rep(NA_character_, length(situation_chr))
      valid         <- !is.na(situation_chr) & nchar(situation_chr) > 0
      if (any(valid)) {
        situation_pad[valid] <- sprintf('%04d', as.integer(situation_chr[valid]))
      }
      pbps[['situationCode']] <- situation_pad
      pbps
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname gc_play_by_plays
#' @export

gc_pbps <- function(season = 20242025) {
  gc_play_by_plays(season)
}

#' Access the raw World Showcase (WSC) play-by-plays for a season
#' 
#' `wsc_play_by_plays_raw()` loads the raw WSC play-by-plays for a given
#' `season`.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per raw event (play) per game
#' @examples
#' # May take >5s, so skip.
#' \donttest{wsc_pbps_raw_20212022 <- wsc_play_by_plays_raw(season = 20212022)}
#' @export

wsc_play_by_plays_raw <- function(season = 20242025) {
  tryCatch(
    expr = {
      u <- paste0(
        'https://huggingface.co/datasets/RentoSaijo/NHL_DB/resolve/main/',
        'data/game/pbps/wsc/NHL_PBPS_WSC_Raw_',
        season,
        '.parquet'
      )
      tmp <- tempfile(fileext = '.parquet')
      utils::download.file(u, tmp, mode = 'wb', quiet = TRUE)
      as.data.frame(arrow::read_parquet(tmp), stringsAsFactors = FALSE)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname wsc_play_by_plays_raw
#' @export

wsc_pbps_raw <- function(season = 20242025) {
  wsc_play_by_plays_raw(season)
}

#' Access the World Showcase (WSC) play-by-plays for a season
#' 
#' `wsc_play_by_plays()` loads the WSC play-by-plays for a given `season`.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per event (play) per game
#' @examples
#' # May take >5s, so skip.
#' \donttest{wsc_pbps_20212022 <- wsc_play_by_plays(season = 20212022)}
#' @export

wsc_play_by_plays <- function(season = 20242025) {
  tryCatch(
    expr = {
      u <- paste0(
        'https://huggingface.co/datasets/RentoSaijo/NHL_DB/resolve/main/',
        'data/game/pbps/wsc/NHL_PBPS_WSC_',
        season,
        '.parquet'
      )
      tmp <- tempfile(fileext = '.parquet')
      utils::download.file(u, tmp, mode = 'wb', quiet = TRUE)
      pbps <- as.data.frame(arrow::read_parquet(tmp), stringsAsFactors = FALSE)
      raw_situation <- pbps[['situationCode']]
      situation_chr <- as.character(raw_situation)
      situation_pad <- rep(NA_character_, length(situation_chr))
      valid         <- !is.na(situation_chr) & nchar(situation_chr) > 0
      if (any(valid)) {
        situation_pad[valid] <- sprintf('%04d', as.integer(situation_chr[valid]))
      }
      pbps[['situationCode']] <- situation_pad
      pbps
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname wsc_play_by_plays
#' @export

wsc_pbps <- function(season = 20242025) {
  wsc_play_by_plays(season)
}

#' Access the shift charts for a season
#' 
#' `shift_charts()` loads the shift charts for a given `season`.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per event (play) per game
#' @examples
#' # May take >5s, so skip.
#' \donttest{shift_charts_20212022 <- shift_charts(season = 20212022)}
#' @export

shift_charts <- function(season = 20242025) {
  tryCatch(
    expr = {
      u <- paste0(
        'https://huggingface.co/datasets/RentoSaijo/NHL_DB/resolve/main/',
        'data/game/scs/NHL_SCS_',
        season,
        '.parquet'
      )
      tmp <- tempfile(fileext = '.parquet')
      utils::download.file(u, tmp, mode = 'wb', quiet = TRUE)
      shifts <- as.data.frame(arrow::read_parquet(tmp), stringsAsFactors = FALSE)
      shifts$id <- NULL
      shifts
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the replays for a season
#' 
#' `replays()` loads the replays for a given `season`.
#' 
#' @inheritParams roster
#' @returns data.frame with one row per decisecond
#' @examples
#' # May take >5s, so skip.
#' \donttest{replays_20252026 <- replays(season = 20252026)}
#' @export

replays <- function(season = 20242025) {
  old_timeout <- getOption('timeout')
  on.exit(options(timeout = old_timeout), add = TRUE)
  options(timeout = 600)
  tryCatch(
    expr = {
      u <- paste0(
        'https://huggingface.co/datasets/RentoSaijo/NHL_DB/resolve/main/',
        'data/event/replays/NHL_REPLAYS_', 
        season, 
        '.parquet'
      )
      tmp <- tempfile(fileext = '.parquet')
      utils::download.file(u, tmp, mode = 'wb', quiet = TRUE, method = 'libcurl')
      as.data.frame(arrow::read_parquet(tmp), stringsAsFactors = FALSE)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

test_that("gc_play_by_plays_raw() returns the stored raw parquet without padding situationCode", {
  src <- tempfile(fileext = ".parquet")
  raw_pbp <- data.frame(
    gameId = 202120220001,
    eventId = 1L,
    situationCode = "651",
    stringsAsFactors = FALSE
  )
  arrow::write_parquet(raw_pbp, src)

  local_mocked_bindings(
    download.file = function(url, destfile, mode = "wb", quiet = TRUE, ...) {
      expect_true(grepl("NHL_PBPS_GC_(Raw_)?20212022\\.parquet$", url))
      expect_true(file.copy(src, destfile, overwrite = TRUE))
      0L
    },
    .package = "utils"
  )

  raw_out <- gc_play_by_plays_raw(20212022)
  clean_out <- gc_play_by_plays(20212022)

  expect_equal(raw_out$situationCode, "651")
  expect_equal(clean_out$situationCode, "0651")
})

test_that("wsc_play_by_plays_raw() returns the stored raw parquet without padding situationCode", {
  src <- tempfile(fileext = ".parquet")
  raw_pbp <- data.frame(
    gameId = 202120220001,
    eventId = 1L,
    situationCode = "551",
    stringsAsFactors = FALSE
  )
  arrow::write_parquet(raw_pbp, src)

  local_mocked_bindings(
    download.file = function(url, destfile, mode = "wb", quiet = TRUE, ...) {
      expect_true(grepl("NHL_PBPS_WSC_(Raw_)?20212022\\.parquet$", url))
      expect_true(file.copy(src, destfile, overwrite = TRUE))
      0L
    },
    .package = "utils"
  )

  raw_out <- wsc_play_by_plays_raw(20212022)
  clean_out <- wsc_play_by_plays(20212022)

  expect_equal(raw_out$situationCode, "551")
  expect_equal(clean_out$situationCode, "0551")
})

#' Access all the coaches
#'
#' `coaches()` retrieves all the coaches as a `data.frame` where each row represents coach and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per coach
#' @examples
#' all_coaches <- coaches()
#' @export

coaches <- function() {
  tryCatch({
    coaches <- nhl_api(
      path = 'coach',
      type = 'r'
    )$data
    names(coaches)[names(coaches) == 'id']                <- 'coachId'
    names(coaches)[names(coaches) == 'firstName']         <- 'coachFirstName'
    names(coaches)[names(coaches) == 'fullName']          <- 'coachFullName'
    names(coaches)[names(coaches) == 'lastName']          <- 'coachLastName'
    names(coaches)[names(coaches) == 'birthCountry3code'] <- 'birthCountryTriCode'
    coaches
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the career statistics for all the coaches
#'
#' `coach_career_statistics()` retrieves the career statistics for all the coaches as a `data.frame` where each row represents coach and includes detail on ranking movement, points pace, and division/conference position signals.
#'
#' @returns data.frame with one row per coach
#' @examples
#' coach_career_stats <- coach_career_statistics()
#' @export

coach_career_statistics <- function() {
  tryCatch({
    results <- nhl_api(
      path = 'coach-career-records-regular-plus-playoffs',
      type = 'r'
    )$data
    results$id <- NULL
    results    <- results[order(results$coachId), ]
    names(results)[names(results) == 'startSeason'] <- 'startSeasonId'
    names(results)[names(results) == 'endSeason']   <- 'endSeasonId'
    names(results) <- normalize_team_abbrev_cols(names(results))
    results
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname coach_career_statistics
#' @export

coach_career_stats <- function() {
  coach_career_statistics()
}

#' Access the statistics for all the coaches by franchise and game type
#'
#' `coach_franchise_statistics()` retrieves the statistics for all the coaches by franchise and game type as a `data.frame` where each row represents franchise per coach per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per franchise per coach per game type
#' @examples
#' coach_franchise_stats <- coach_franchise_statistics()
#' @export

coach_franchise_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'coach-franchise-records',
      type = 'r'
    )$data
    stats$id <- NULL
    stats    <- stats[order(stats$coachName, stats$firstCoachedDate), ]
    names(stats)[names(stats) == 'startSeason'] <- 'startSeasonId'
    names(stats)[names(stats) == 'endSeason']   <- 'endSeasonId'
    names(stats)[names(stats) == 'firstName']   <- 'coachFirstName'
    names(stats)[names(stats) == 'lastName']    <- 'coachLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname coach_franchise_statistics
#' @export

coach_franchise_stats <- function() {
  coach_franchise_statistics()
}

#' Access all the players
#' 
#' `get_players()` is deprecated. Use [players()] instead.
#' 
#' @returns data.frame with one row per player
#' @export

get_players <- function() {
  .Deprecated(
    new     = 'players()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_players()` is deprecated.',
      'Use `players()` instead.'
    )
  )
  players()
}

#' Access the summary for a player
#' 
#' `get_player_landing()` is deprecated. Use [player_summary()] instead.
#' 
#' @inheritParams player_seasons
#' @returns list with various items
#' @export

get_player_landing <- function(player = 8478402) {
  .Deprecated(
    new     = 'player_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_player_landing()` is deprecated.',
      'Use `player_summary()` instead.'
    )
  )
  player_summary(player)
}

#' Access the game log for a player, season, and game type
#' 
#' `get_player_game_log()` is deprecated. Use [player_game_log()] instead.
#' 
#' @inheritParams player_seasons
#' @inheritParams roster_statistics
#' @returns data.frame with one row per game
#' @export

get_player_game_log <- function(
  player    = 8478402, 
  season    = 'now', 
  game_type = ''
) {
  .Deprecated(
    new     = 'player_game_log()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_player_game_log()` is deprecated.',
      'Use `player_game_log()` instead.'
    )
  )
  player_game_log(player, season, game_type)
}

#' Access the spotlight players
#' 
#' `get_spotlight_players()` is deprecated. Use [spotlight_players()] 
#' instead.
#' 
#' @returns data.frame with one row per player
#' @export

get_spotlight_players <- function() {
  .Deprecated(
    new     = 'spotlight_players()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_spotlight_players()` is deprecated.',
      'Use `spotlight_players()` instead.'
    )
  )
  spotlight_players()
}

#' Access all the franchises
#'
#' `franchises()` retrieves all the franchises as a `data.frame` where each row represents franchise and includes detail on team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per franchise
#' @examples
#' all_franchises <- franchises()
#' @export

franchises <- function() {
  tryCatch({
    franchises <- nhl_api(
      path = 'franchise',
      type = 'r'
    )$data
    details    <- nhl_api(
      path = 'franchise-detail',
      type = 'r'
    )$data
    details$firstSeasonId    <- NULL
    details$mostRecentTeamId <- NULL
    details$teamAbbrev       <- NULL
    franchises <- merge(franchises[order(franchises$id), ], details, by = 'id')
    names(franchises)[names(franchises) == 'id']       <- 'franchiseId'
    names(franchises)[names(franchises) == 'fullName'] <- 'franchiseFullName'
    names(franchises) <- normalize_team_abbrev_cols(names(franchises))
    franchises
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the all-time statistics for all the franchises by game type
#'
#' `franchise_statistics()` retrieves the all-time statistics for all the franchises by game type as a `data.frame` where each row represents franchise per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per franchise per game type
#' @examples
#' franchise_stats <- franchise_statistics()
#' @export

franchise_statistics <- function() {
  tryCatch({
    stats <- nhl_api(
      path = 'franchise-totals',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'teamAbbrev'] <- 'teamTriCode'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$franchiseId, stats$gameTypeId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname franchise_statistics
#' @export

franchise_stats <- function() {
  franchise_statistics()
}

#' Access the all-time statistics for all the franchises by team and game type
#'
#' `franchise_team_statistics()` retrieves the all-time statistics for all the franchises by team and game type as a `data.frame` where each row represents team per franchise per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per team per franchise per game type
#' @examples
#' franchise_team_stats <- franchise_team_statistics()
#' @export

franchise_team_statistics <- function() {
  tryCatch({
    stats <- nhl_api(
      path = 'franchise-team-totals',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'triCode'] <- 'teamTriCode'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$franchiseId, stats$teamId, stats$gameTypeId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname franchise_team_statistics
#' @export

franchise_team_stats <- function() {
  franchise_team_statistics()
}

#' Access the statistics for all the franchises by season and game type
#'
#' `franchise_season_statistics()` retrieves the statistics for all the franchises by season and game type as a `data.frame` where each row represents franchise per season per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per franchise per season per game type
#' @examples
#' # May take >5s, so skip.
#' \donttest{franchise_season_stats <- franchise_season_statistics()}
#' @export

franchise_season_statistics <- function() {
  tryCatch({
    stats <- nhl_api(
      path = 'franchise-season-results',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'triCode'] <- 'teamTriCode'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$franchiseId, stats$seasonId, stats$gameTypeId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname franchise_season_statistics
#' @export

franchise_season_stats <- function() {
  franchise_season_statistics()
}

#' Access the all-time statistics versus other franchises for all the 
#' franchises by game type
#'
#' `franchise_versus_franchise()` retrieves the all-time statistics versus other franchises for all the franchises by game type as a `data.frame` where each row represents franchise per franchise per game type and includes detail on date/season filtering windows and chronological context plus team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per franchise per franchise per game type
#' @examples
#' # May take >5s, so skip.
#' \donttest{franchise_vs_franchise <- franchise_versus_franchise()}
#' @export

franchise_versus_franchise <- function() {
  tryCatch({
    versus    <- nhl_api(
      path = 'all-time-record-vs-franchise',
      type = 'r'
    )$data
    versus$id <- NULL
    versus    <- versus[order(
      versus$teamFranchiseId, 
      versus$teamId, 
      versus$opponentFranchiseId, 
      versus$opponentTeamId, 
      versus$gameTypeId
    ), ]
    names(versus)[names(versus) == 'teamFranchiseId'] <- 'franchiseId'
    versus
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname franchise_versus_franchise
#' @export
franchise_vs_franchise <- function() {
  franchise_versus_franchise()
}

#' Access the playoff series results for all the franchises by situation
#'
#' `franchise_playoff_situational_results()` retrieves the playoff series results for all the franchises by situation as a `data.frame` where each row represents franchise per situation and includes detail on team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per franchise per situation
#' @examples
#' franchise_playoff_situational_results <- 
#'   franchise_playoff_situational_results()
#' @export

franchise_playoff_situational_results <- function() {
  tryCatch({
    results    <- nhl_api(
      path = 'series-situational-records',
      type = 'r'
    )$data
    results$id <- NULL
    results[order(results$franchiseId, results$seriesSituation), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

test_that("franchise_playoff_situational_results() returns non-empty data.frame", {
  skip_if_offline()
  test <- franchise_playoff_situational_results()
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

#' Access the configurations for goalie reports
#'
#' `goalie_report_configurations()` retrieves the configurations for goalie reports as a nested `list` that separates summary and detail blocks for situational splits across home/road, strength state, and overtime/shootout states plus configuration catalogs for valid report categories and filters.
#'
#' @returns list with various items
#' @examples
#' goalie_report_configs <- goalie_report_configurations()
#' @export

goalie_report_configurations <- function() {
  tryCatch({
    nhl_api(
      path = 'en/config',
      type = 's'
    )$goalieReportData
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    list()
  })
}

#' @rdname goalie_report_configurations
#' @export

goalie_report_configs <- function() {
  goalie_report_configurations()
}

#' Access various reports for a season, game type, and category for all 
#' the goalies by season
#'
#' `goalie_season_report()` retrieves various reports for a season, game type, and category for all the goalies by season as a `data.frame` where each row represents player and includes detail on date/season filtering windows and chronological context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams roster_statistics
#' @param category character (e.g., 'advanced'); see 
#' [goalie_report_configurations()] for reference
#'
#' @returns data.frame with one row per player
#' @examples
#' # May take >5s, so skip.
#' \donttest{advanced_goalie_season_report_playoffs_20212022 <- 
#'   goalie_season_report(
#'     season    = 20212022, 
#'     game_type = 3, 
#'     category  = 'advanced'
#'   )}
#' @export

goalie_season_report <- function(
  season    = season_now(), 
  game_type = game_type_now(), 
  category  = 'summary'
) {
  tryCatch(
    expr = {
      report <- nhl_api(
        path  = sprintf('en/goalie/%s', category),
        query = list(
          limit       = -1,
          isAggregate = FALSE,
          isGame      = FALSE,
          cayenneExp  = sprintf(
            'seasonId = %s and gameTypeId = %s', 
            season,
            game_type
          )
        ),
        type  = 's'
      )$data
      report <- report[order(report$playerId), ]
      names(report)[names(report) == 'lastName'] <- 'goalieLastName'
      names(report) <- normalize_team_abbrev_cols(names(report))
      report
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access various reports for a season, game type, and category for all 
#' the goalies by game
#'
#' `goalie_game_report()` retrieves various reports for a season, game type, and category for all the goalies by game as a `data.frame` where each row represents game per goalie and includes detail on game timeline state, period/clock progression, and matchup flow, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams goalie_season_report
#'
#' @returns data.frame with one row per game per goalie
#' @examples
#' # May take >5s, so skip.
#' \donttest{advanced_goalie_game_report_playoffs_20212022 <- 
#'   goalie_game_report(
#'     season    = 20212022, 
#'     game_type = 3, 
#'     category  = 'advanced'
#'   )}
#' @export

goalie_game_report <- function(
    season    = season_now(), 
    game_type = game_type_now(), 
    category  = 'summary'
) {
  tryCatch(
    expr = {
      report <- nhl_api(
        path  = sprintf('en/goalie/%s', category),
        query = list(
          limit       = -1,
          isAggregate = FALSE,
          isGame      = TRUE,
          cayenneExp  = sprintf(
            'seasonId = %s and gameTypeId = %s', 
            season,
            game_type
          )
        ),
        type  = 's'
      )$data
      report <- report[order(report$playerId, report$gameId), ]
      names(report)[names(report) == 'lastName'] <- 'goalieLastName'
      names(report) <- normalize_team_abbrev_cols(names(report))
      report
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the career statistics for all the goalies
#'
#' `goalie_statistics()` retrieves the career statistics for all the goalies as a `data.frame` where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
#'
#' @returns data.frame with one row per player
#' @examples
#' goalie_stats <- goalie_statistics()
#' @export

goalie_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'goalie_career_stats_incl_playoffs',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'goalieFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'goalieLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$playerId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname goalie_statistics
#' @export

goalie_stats <- function() {
  goalie_statistics()
}

#' Access the career regular season statistics for all the goalies
#'
#' `goalie_regular_statistics()` retrieves the career regular season statistics for all the goalies as a `data.frame` where each row represents goalie and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per goalie
#' @examples
#' goalie_career_regular_statistics <- goalie_regular_statistics()
#' @export

goalie_regular_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'goalie-career-stats',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'goalieFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'goalieLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$playerId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname goalie_regular_statistics
#' @export

goalie_regular_stats <- function() {
  goalie_regular_statistics()
}


#' Access the statistics for all the goalies by season, game type, and team.
#'
#' `goalie_season_statistics()` retrieves the statistics for all the goalies by season, game type, and team as a `data.frame` where each row represents player per season per game type, separated by team if applicable and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player per season per game type, 
#' separated by team if applicable
#' @examples
#' goalie_season_stats <- goalie_season_statistics()
#' @export

goalie_season_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'goalie-season-stats',
      type = 'r'
    )$data
    stats$id <- NULL
    stats    <- stats[order(stats$playerId, stats$seasonId, stats$gameType), ]
    names(stats)[names(stats) == 'firstName'] <- 'goalieFirstName'
    names(stats)[names(stats) == 'gameType']  <- 'gameTypeId'
    names(stats)[names(stats) == 'lastName']  <- 'goalieLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname goalie_season_statistics
#' @export

goalie_season_stats <- function() {
  goalie_season_statistics()
}

#' Access the statistics for all the goalies by game
#'
#' `goalie_game_statistics()` retrieves the statistics for all the goalies by game as a `data.frame` with detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per goalie per game
#' @examples
#' \donttest{goalie_game_stats <- goalie_game_statistics()}
#' @export

goalie_game_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'goalie-game-stats',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName']       <- 'goalieFirstName'
    names(stats)[names(stats) == 'lastName']        <- 'goalieLastName'
    names(stats)[names(stats) == 'triCode']         <- 'teamTriCode'
    names(stats)[names(stats) == 'opponentTriCode'] <- 'opponentTeamTriCode'
    names(stats)[names(stats) == 'opponentName']    <- 'opponentTeamName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats[order(stats$playerId, stats$gameId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname goalie_game_statistics
#' @export

goalie_game_stats <- function() {
  goalie_game_statistics()
}

#' Access the playoff statistics for all the goalies by series
#'
#' `goalie_series_statistics()` retrieves the playoff statistics for all the goalies by series as a `data.frame` where each row represents player per series and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player per series
#' @examples
#' goalie_series_stats <- goalie_series_statistics()
#' @export

goalie_series_statistics <- function() {
  tryCatch({
    stats    <- nhl_api(
      path = 'playoff-goalie-series-stats',
      type = 'r'
    )$data
    stats$id <- NULL
    names(stats)[names(stats) == 'firstName'] <- 'goalieFirstName'
    names(stats)[names(stats) == 'lastName']  <- 'goalieLastName'
    names(stats) <- normalize_team_abbrev_cols(names(stats))
    stats
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname goalie_series_statistics
#' @export

goalie_series_stats <- function() {
  goalie_series_statistics()
}

#' Access the career scoring statistics for all the goalies
#'
#' `goalie_scoring()` retrieves the career scoring statistics for all the goalies as a `data.frame` where each row represents player and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per player
#' @examples
#' goalie_scoring <- goalie_scoring()
#' @export

goalie_scoring <- function() {
  tryCatch({
    scoring    <- nhl_api(
      path = 'goalie-career-scoring',
      type = 'r'
    )$data
    scoring$id <- NULL
    names(scoring)[names(scoring) == 'firstName'] <- 'goalieFirstName'
    names(scoring)[names(scoring) == 'lastName']  <- 'goalieLastName'
    names(scoring) <- normalize_team_abbrev_cols(names(scoring))
    scoring[order(scoring$playerId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the scoring statistics for all the goalies by game
#'
#' `goalie_game_scoring()` retrieves the scoring statistics for all the goalies by game as a `data.frame` with detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @returns data.frame with one row per player per game
#' @examples
#' goalie_game_scoring <- goalie_game_scoring()
#' @export

goalie_game_scoring <- function() {
  tryCatch({
    scoring    <- nhl_api(
      path = 'goalie-game-scoring',
      type = 'r'
    )$data
    scoring$id <- NULL
    names(scoring)[names(scoring) == 'firstName'] <- 'goalieFirstName'
    names(scoring)[names(scoring) == 'lastName']  <- 'goalieLastName'
    names(scoring) <- normalize_team_abbrev_cols(names(scoring))
    scoring[order(scoring$playerId, scoring$gameId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the goalie statistics leaders for a season, game type, and category
#'
#' `goalie_leaders()` retrieves the goalie statistics leaders for a season, game type, and category as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @inheritParams roster_statistics
#' @param category character of 'w'/'wins', 's'/shutouts', 
#' 's%'/'sP'/'save %'/'save percentage', or 'gaa'/'goals against average'
#'
#' @returns data.frame with one row per player
#' @examples
#' GAA_leaders_regular_20242025 <- goalie_leaders(
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'GAA'
#' )
#' @export

goalie_leaders <- function(
  season    = 'current',
  game_type = '',
  category  = 'wins'
) {
  tryCatch(
    expr = {
      category <- switch(
        tolower(category),
        w                       = 'wins',
        wins                    = 'wins',
        s                       = 'shutouts',
        shutouts                = 'shutouts',
        `s%`                    = 'savePctg',
        sp                      = 'savePctg',
        `save %`                = 'savePctg',
        `save percentage`       = 'savePctg',
        gaa                     = 'goalsAgainstAverage',
        `goals against average` = 'goalsAgainstAverage'
      )
      goalies <- nhl_api(
        path  = sprintf('v1/goalie-stats-leaders/%s/%s', season, game_type),
        type  = 'w'
      )[[category]]
      names(goalies)[names(goalies) == 'id']       <- 'playerId'
      names(goalies)[names(goalies) == 'position'] <- 'positionCode'
      names(goalies) <- normalize_locale_names(names(goalies))
      names(goalies) <- scope_person_name_cols(names(goalies), 'goalie')
      names(goalies) <- normalize_team_abbrev_cols(names(goalies))
      goalies
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the goalies on milestone watch
#'
#' `goalie_milestones()` retrieves the goalies on milestone watch as a `data.frame` where each row represents player and includes detail on date/season filtering windows and chronological context, player identity, role, handedness, and biographical profile, and ranking movement, points pace, and division/conference position signals.
#'
#' @returns data.frame with one row per player
#' @examples
#' goalie_milestones <- goalie_milestones()
#' @export

goalie_milestones <- function() {
  tryCatch({
    milestones    <- nhl_api(
      path = 'en/milestones/goalies',
      type = 's'
    )$data
    milestones$id <- NULL
    names(milestones)[names(milestones) == 'firstName'] <- 'goalieFirstName'
    names(milestones)[names(milestones) == 'lastName']  <- 'goalieLastName'
    names(milestones) <- normalize_team_abbrev_cols(names(milestones))
    milestones
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the games
#' 
#' `get_games()` is deprecated. Use [games()] instead.
#' 
#' @returns data.frame with one row per game
#' @export

get_games <- function() {
  .Deprecated(
    new     = 'games()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_games()` is deprecated.',
      'Use `games()` instead.'
    )
  )
  games()
}

#' Access the scores for a date
#' 
#' `get_scores()` is deprecated. Use [scores()] instead.
#' 
#' @inheritParams standings
#' @returns data.frame with one row per game
#' @export

get_scores <- function(date = 'now') {
  .Deprecated(
    new     = 'scores()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_scores()` is deprecated.',
      'Use `scores()` instead.'
    )
  )
  scores(date)
}

#' Access the scoreboards for a date
#' 
#' `get_scoreboards()` is deprecated. Use [scores()] instead.
#' 
#' @inheritParams standings
#' @returns data.frame with one row per game
#' @export

get_scoreboards <- function(date = 'now') {
  .Deprecated(
    new     = 'scores()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_scoreboards()` is deprecated.',
      'Use `scores()` instead.'
    )
  )
  scores(date)
}

#' Access the GameCenter (GC) summary for a game
#' 
#' `get_game_landing()` is deprecated. Use [gc_summary()] instead.
#'
#' @inheritParams gc_summary
#' @returns list of various items
#' @export

get_game_landing <- function(game = 2023030417) {
  .Deprecated(
    new     = 'gc_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_game_landing()` is deprecated.',
      'Use `gc_summary()` instead.'
    )
  )
  gc_summary(game)
}

#' Access the World Showcase (WSC) summary for a game
#' 
#' `get_game_story()` is deprecated. Use [wsc_summary()] instead.
#'
#' @inheritParams gc_summary
#' @returns list of various items
#' @export

get_game_story <- function(game = 2023030417) {
  .Deprecated(
    new     = 'wsc_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_game_story()` is deprecated.',
      'Use `wsc_summary()` instead.'
    )
  )
  wsc_summary(game)
}

#' Access the boxscore for a game, team, and player type
#' 
#' `get_game_boxscore()` is deprecated. Use [boxscore()] instead.
#'
#' @inheritParams gc_summary
#' @param team character of 'home' or 'away'
#' @param player_type character of 'forwards', 'defense', or 'goalies'
#' @returns data.frame with one row per player
#' @export

get_game_boxscore <- function(
    game        = 2023030417,
    team        = 'home',
    player_type = 'forwards'
) {
  .Deprecated(
    new     = 'boxscore()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_game_boxscore()` is deprecated.',
      'Use `boxscore()` instead.'
    )
  )
  boxscore(game, team, player_type)
}

#' Access the GameCenter (GC) play-by-play for a game
#' 
#' `get_gc_play_by_play()` is deprecated. Use [gc_play_by_play()] instead.
#'
#' @inheritParams gc_summary
#' @returns data.frame with one row per event (play)
#' @export

get_gc_play_by_play <- function(game = 2023030417) {
  .Deprecated(
    new     = 'gc_play_by_play()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_gc_play_by_play()` is deprecated.',
      'Use `gc_play_by_play()` instead.'
    )
  )
  gc_play_by_play(game)
}

#' Access the World Showcase (WSC) play-by-play for a game
#' 
#' `get_wsc_play_by_play()` is deprecated. Use [wsc_play_by_play()] instead.
#'
#' @inheritParams gc_summary
#' @returns data.frame with one row per event (play)
#' @export

get_wsc_play_by_play <- function(game = 2023030417) {
  .Deprecated(
    new     = 'wsc_play_by_play()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_wsc_play_by_play()` is deprecated.',
      'Use `wsc_play_by_play()` instead.'
    )
  )
  wsc_play_by_play(game)
}

#' Access the shift charts for a game
#' 
#' `get_shift_charts()` is deprecated. Use [shift_chart()] instead.
#'
#' @inheritParams gc_summary
#' @returns data.frame with one row per shift
#' @export

get_shift_charts <- function(game = 2023030417) {
  .Deprecated(
    new     = 'shift_chart()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_shift_charts()` is deprecated.',
      'Use `shift_chart()` instead.'
    )
  )
  shift_chart(game)
}

#' Access the shift charts for a game
#' 
#' `shifts()` is deprecated. Use [shift_chart()] instead.
#'
#' @inheritParams gc_summary
#' @returns data.frame with one row per shift
#' @export

shifts <- function(game = 2023030417) {
  .Deprecated(
    new     = 'shift_chart()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_shift_charts()` is deprecated.',
      'Use `shift_chart()` instead.'
    )
  )
  shift_chart(game)
}

#' Access the real-time game odds for a country by partnered bookmaker
#' 
#' `get_partner_odds()` is deprecated. Use [game_odds()] instead.
#' 
#' @inheritParams game_odds
#' @returns data.frame with one row per game
#' @export

get_partner_odds <- function(country = 'US') {
  .Deprecated(
    new     = 'game_odds()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_partner_odds()` is deprecated.',
      'Use `game_odds()` instead.'
    )
  )
  game_odds(country)
}

#' Access all the awards
#' 
#' `get_awards()` is deprecated. Use [awards()] instead.
#' 
#' @returns data.frame with one row per award
#' @export

get_awards <- function() {
  .Deprecated(
    new     = 'awards()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_awards()` is deprecated.',
      'Use `awards()` instead.'
    )
  )
  awards()
}

#' Access all the award winners/finalists
#' 
#' `get_award_winners()` is deprecated. Use [award_winners()] instead.
#' 
#' @returns data.frame with one row per winner/finalist
#' @export

get_award_winners <- function() {
  .Deprecated(
    new     = 'award_winners()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_award_winners()` is deprecated.',
      'Use `award_winners()` instead.'
    )
  )
  award_winners()
}

#' Access the glossary
#'
#' `glossary()` retrieves the glossary as a `data.frame` where each row represents terminology and includes detail on reference definitions and rules-framework information.
#'
#' @returns data.frame with one row per terminology
#' @examples
#' glossary <- glossary()
#' @export

glossary <- function() {
  tryCatch({
    terms <- nhl_api(
      path = 'en/glossary',
      type = 's'
    )$data
    names(terms)[names(terms) == 'id']       <- 'terminologyId'
    names(terms)[names(terms) == 'fullName'] <- 'terminologyFullName'
    terms
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the countries
#'
#' `countries()` retrieves all the countries as a `data.frame` where each row represents country and includes detail on reference metadata, regional context, and media availability detail.
#'
#' @returns data.frame with one row per country
#' @examples
#' all_countries <- countries()
#' @export

countries <- function() {
  tryCatch({
    countries <- nhl_api(
      path = 'en/country',
      type = 's'
    )$data
    names(countries)[names(countries) == 'id']           <- 'countryId'
    names(countries)[names(countries) == 'country3Code'] <- 'countryTriCode'
    countries
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the location for a zip code
#'
#' `location()` retrieves the location for a zip code as a `data.frame` where each row represents team and includes detail on venue/location geography and regional metadata.
#'
#' @param zip integer (e.g., 48304)
#'
#' @returns data.frame with one row per team
#' @examples
#' Cranbrook_Schools <- location(48304)
#' @export

location <- function(zip = 10001) {
  tryCatch(
    expr = {
      location <- nhl_api(
        path = sprintf('v1/postal-lookup/%s', zip),
        type = 'w'
      )
      location[0, ]
      names(location)[names(location) == 'country'] <- 'countryCode'
      names(location) <- normalize_locale_names(names(location))
      names(location) <- normalize_team_abbrev_cols(names(location))
      location
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access all the streams
#'
#' `streams()` retrieves all the streams as a `data.frame` where each row represents stream and includes detail on reference metadata, regional context, and media availability detail.
#'
#' @returns data.frame with one row per stream
#' @examples
#' all_streams <- streams()
#' @export

streams <- function() {
  tryCatch({
    nhl_api(
      path = 'v1/where-to-watch',
      type = 'w'
    )
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the NHL Network TV schedule for a date
#'
#' `tv_schedule()` retrieves the NHL Network TV schedule for a date as a `data.frame` where each row represents program and includes detail on date/season filtering windows and chronological context.
#'
#' @inheritParams standings
#'
#' @returns data.frame with one row per program
#' @examples
#' tv_schedule_Halloween_2025 <- tv_schedule(date = '2025-10-31')
#' @export

tv_schedule <- function(date = 'now') {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf('v1/network/tv-schedule/%s', date),
        type = 'w'
      )$broadcasts
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the skater statistics leaders for a season, game type, and category
#' 
#' `get_skater_leaders()` is deprecated. Use [skater_leaders()] instead.
#' 
#' @inheritParams roster_statistics
#' @param category character of 'assists', 'goals', 'goalsSh', 'goalsPp', 
#' 'points, 'penaltyMins', 'toi', 'plusMinus', or 'faceoffLeaders'
#' @returns data.frame with one row per player
#' @export

get_skater_leaders <- function(
  season    = 'current',
  game_type = '',
  category  = 'points'
) {
  .Deprecated(
    new     = 'skater_leaders()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_skater_leaders()` is deprecated.',
      'Use `skater_leaders()` instead.'
    )
  )
  category <- switch(
    category,
    assists        = 'a',
    goals          = 'g',
    goalsSh        = 'shg',
    goalsPp        = 'ppg',
    points         = 'p',
    penaltyMins    = 'pim',
    toi            = 'toi',
    plusMinus      = 'pm',
    faceoffLeaders = 'f'
  )
  skater_leaders(season, game_type, category)
}

#' Access the skaters on milestone watch
#' 
#' `get_skater_milestones()` is deprecated. Use [skater_milestones()] 
#' instead.
#'
#' @returns data.frame with one row per player
#' @export

get_skater_milestones <- function() {
  .Deprecated(
    new     = 'skater_milestones()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_skater_milestones()` is deprecated.',
      'Use `skater_milestones()` instead.'
    )
  )
  skater_milestones()
}

#' Access all the skaters for a range of seasons
#' 
#' `get_skaters()` is defunct. Use [players()] instead.
#'
#' @export

get_skaters <- function() {
  .Defunct(
    new     = 'players()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_skaters()` is defunct.',
      'Use `players()` instead.'
    )
  )
}

#' Access various reports for all the skaters by season or game
#' 
#' `get_skater_statistics()` is defunct. Use [skater_season_report()] or 
#' [skater_game_report()] instead.
#' 
#' @export

get_skater_statistics <- function() {
  .Defunct(
    new     = 'skater_season_report()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_skater_statistics()` is defunct.',
      'Use `skater_season_report()` and/or `skater_game_report()` instead.'
    )
  )
}

#' Access all the penalty shots
#' 
#' `ps()` is deprecated. Use [penalty_shots()] instead.
#'
#' @returns data.frame with one row per penalty shot
#' @export

ps <- function() {
  .Deprecated(
    new     = 'penalty_shots()',
    package = 'nhlscraper',
    msg     = paste(
      '`ps()` is deprecated.',
      'Use `penalty_shots()` instead.'
    )
  )
  penalty_shots()
}

#' Access all the teams
#' 
#' `get_teams()` is deprecated. Use [teams()] instead.
#' 
#' @returns data.frame with one row per team
#' @export

get_teams <- function() {
  .Deprecated(
    new     = 'teams()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_teams()` is deprecated.',
      'Use `teams()` instead.'
    )
  )
  teams()
}

#' Access the season(s) and game type(s) in which a team played
#' 
#' `get_team_seasons()` is deprecated. Use [team_seasons()] instead.
#' 
#' @param team three-letter code (e.g., 'COL'); see [teams()] for reference
#' @returns data.frame with one row per season
#' @export

get_team_seasons <- function(team = 'NJD') {
  .Deprecated(
    new     = 'team_seasons()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_team_seasons()` is deprecated.',
      'Use `team_seasons()` instead.'
    )
  )
  team_seasons(team)
}

#' Access the roster for a team, season, and player type
#' 
#' `get_team_roster()` is deprecated. Use [roster()] instead.
#' 
#' @inheritParams get_team_seasons
#' @param season integer in YYYYYYYY (e.g., 20242025); see [seasons()] for 
#' reference
#' @param player_type character of 'forwards', 'defensemen', or 'goalies'
#' @returns data.frame with one row per player
#' @export

get_team_roster <- function(
  team        = 'NJD',
  season      = 'current',
  player_type = 'forwards'
) {
  .Deprecated(
    new     = 'roster()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_team_roster()` is deprecated.',
      'Use `roster()` instead.'
    )
  )
  roster(team, season, player_type)
}

#' Access the roster statistics for a team, season, game type, and player type
#' 
#' `get_team_roster_statistics()` is deprecated. Use [roster_statistics()] 
#' instead.
#' 
#' @inheritParams get_team_roster
#' @param game_type integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season) OR character of 'pre', 'regular', or 
#' playoff'/'post'; see [seasons()] for reference; most functions will NOT 
#' support pre-season
#' @param player_type character of 'skaters' or 'goalies'
#' @returns data.frame with one row per player
#' @export

get_team_roster_statistics <- function(
  team        = 'NJD',
  season      = 'now',
  game_type   = 2,
  player_type = 'skaters'
) {
  .Deprecated(
    new     = 'roster_statistics()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_team_roster_statistics()` is deprecated.',
      'Use `roster_statistics()` instead.'
    )
  )
  roster_statistics(team, season, game_type, player_type)
}

#' Access the prospects for a team and position
#' 
#' `get_team_prospects()` is deprecated. Use [team_prospects()] instead.
#' 
#' @inheritParams get_team_roster
#' @returns data.frame with one row per player
#' @export

get_team_prospects <- function(team = 'NJD', player_type = 'forwards') {
  .Deprecated(
    new     = 'team_prospects()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_team_prospects()` is deprecated.',
      'Use `team_prospects()` instead.'
    )
  )
  team_prospects(team, player_type)
}

#' Access the schedule for a team and season
#' 
#' `get_team_schedule()` is deprecated. Use [team_season_schedule()] instead.
#' 
#' @inheritParams get_team_roster
#' @returns data.frame with one row per game
#' @export

get_team_schedule <- function(team = 'NJD', season = 'now') {
  .Deprecated(
    new     = 'team_season_schedule()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_team_schedule()` is deprecated.',
      'Use `team_season_schedule()` instead.'
    )
  )
  team_season_schedule(team, season)
}

#' Access various reports for all the teams by season or game
#' 
#' `get_team_statistics()` is defunct. Use [team_season_report()] and/or 
#' [team_game_report()] instead.
#' 
#' @export

get_team_statistics <- function() {
  .Defunct(
    new     = 'team_season_report()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_team_statistics()` is defunct.',
      'Use `team_season_report()` and/or `team_game_report` instead.'
    )
  )
}

#' Access the team scoreboard as of now
#' 
#' `get_team_scoreboard()` is defunct.
#' 
#' @export

get_team_scoreboard <- function() {
  .Defunct(
    msg = paste(
      '`get_team_scoreboard()` is defunct.'
    )
  )
}

#' Access all the general managers
#'
#' `general_managers()` retrieves all the general managers as a `data.frame` where each row represents general manager and includes detail on player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per general manager
#' @examples
#' all_GMs <- general_managers()
#' @export

general_managers <- function() {
  tryCatch({
    gms <- nhl_api(
      path = 'general-manager',
      type = 'r'
    )$data
    names(gms)[names(gms) == 'id']        <- 'generalManagerId'
    names(gms)[names(gms) == 'firstName'] <- 'generalManagerFirstName'
    names(gms)[names(gms) == 'fullName']  <- 'generalManagerFullName'
    names(gms)[names(gms) == 'lastName']  <- 'generalManagerLastName'
    gms
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname general_managers
#' @export

gms <- function() {
  general_managers()
}

#' Access all the drafts
#' 
#' `get_drafts()` is deprecated. Use [drafts()] instead.
#'
#' @returns data.frame with one row per draft
#' @export

get_drafts <- function() {
  .Deprecated(
    new     = 'drafts()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_drafts()` is deprecated.',
      'Use `drafts()` instead.'
    )
  )
  drafts()
}

#' Access all the draft picks
#' 
#' `get_draft_picks()` is deprecated. Use [draft_picks()] instead.
#' 
#' @returns data.frame with one row per pick
#' @export

get_draft_picks <- function() {
  .Deprecated(
    new     = 'draft_picks()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_draft_picks()` is deprecated.',
      'Use `draft_picks()` instead.'
    )
  )
  draft_picks()
}

#' Access the draft rankings for a year and player type
#' 
#' `get_draft_rankings()` is deprecated. Use [draft_rankings()] instead.
#' 
#' @param year integer in YYYY (e.g., 2017); see [drafts()] for reference
#' @param player_type integer in 1:4 (where 1 = North American Skaters, 
#' 2 = International Skaters, 3 = North American Goalies, and 4 = International 
#' Goalies)
#' @returns data.frame with one row per player
#' @export

get_draft_rankings <- function(
  year        = season_now() %/% 1e4,
  player_type = 1
) {
  .Deprecated(
    new     = 'draft_rankings()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_draft_rankings()` is deprecated.',
      'Use `draft_rankings()` instead.'
    )
  )
  draft_rankings(year, player_type)
}

#' Access the real-time draft tracker
#' 
#' `get_draft_tracker()` is deprecated. Use [draft_tracker()] instead.
#' 
#' @returns data.frame with one row per player
#' @export

get_draft_tracker <- function() {
  .Deprecated(
    new     = 'draft_tracker()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_draft_tracker()` is deprecated.',
      'Use `draft_tracker()` instead.'
    )
  )
  draft_tracker()
}

#' Access all the franchises
#' 
#' `get_franchises()` is deprecated. Use [franchises()] instead.
#' 
#' @returns data.frame with one row per franchise
#' @export

get_franchises <- function() {
  .Deprecated(
    new     = 'franchises()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_franchises()` is deprecated.',
      'Use `franchises()` instead.'
    )
  )
  franchises()
}

#' Access the all-time statistics for all the franchises by team and game type
#' 
#' `get_franchise_team_totals()` is deprecated. Use 
#' [franchise_team_statistics()] instead.
#'
#' @returns data.frame with one row per team per franchise per game type
#' @export

get_franchise_team_totals <- function() {
  .Deprecated(
    new     = 'franchise_team_statistics()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_franchise_team_totals()` is deprecated.',
      'Use `franchise_team_statistics()` instead.'
    )
  )
  franchise_team_statistics()
}

#' Access the statistics for all the franchises by season and game type
#' 
#' `get_franchise_season_by_season()` is deprecated. Use 
#' [franchise_season_statistics()] instead.
#' 
#' @returns data.frame with one row per franchise per season per game type
#' @export

get_franchise_season_by_season <- function() {
  .Deprecated(
    new     = 'franchise_season_statistics()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_franchise_season_by_season()` is deprecated.',
      'Use `franchise_season_statistics()` instead.'
    )
  )
  franchise_season_statistics()
}

#' Access the all-time statistics versus other franchises for all the 
#' franchises by game type
#' 
#' `get_franchise_vs_franchise()` is deprecated. Use 
#' [franchise_versus_franchise()] instead.
#' 
#' @returns data.frame with one row per franchise per franchise per game type
#' @export

get_franchise_vs_franchise <- function() {
  .Deprecated(
    new     = 'franchise_versus_franchise()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_franchise_vs_franchise()` is deprecated.',
      'Use `franchise_versus_franchise()` instead.'
    )
  )
  franchise_versus_franchise()
}

#' Access the ESPN summary for an event (game)
#' 
#' `get_espn_event()` is deprecated. Use [espn_game_summary()] instead.
#' 
#' @param event integer ID (e.g., 401777460); see [espn_games()] for 
#' reference
#' @return data.frame with one row per event (game)
#' @export

get_espn_event <- function(event = 401777460) {
  .Deprecated(
    new     = 'espn_game_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_event()` is dprecated.',
      'Use `espn_game_summary()` instead.'
    )
  )
  espn_game_summary(event)
}

#' Access the ESPN play-by-play for an event (game)
#' 
#' `get_espn_event_play_by_play()` is deprecated. Use [espn_play_by_play()] 
#' instead.
#' 
#' @inheritParams get_espn_event
#' @returns data.frame with one row per play
#' @export

get_espn_event_play_by_play <- function(event = 401777460) {
  .Deprecated(
    new     = 'espn_play_by_play()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_event_play_by_play()` is deprecated.',
      'Use `espn_play_by_play()` instead.'
    )
  )
  espn_play_by_play(event)
}

#' Access the ESPN odds for an event (game)
#' 
#' `get_espn_event_odds()` is deprecated. Use [espn_game_odds()] 
#' instead.
#' 
#' @inheritParams get_espn_event
#' @returns data.frame with one row per provider
#' @export

get_espn_event_odds <- function(event = 401777460) {
  .Deprecated(
    new     = 'espn_game_odds()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_event_odds()` is deprecated.',
      'Use `espn_game_odds()` instead.'
    )
  )
  espn_game_odds(event)
}

#' Access the ESPN events (games) by start and end dates
#' 
#' `get_espn_events()` is defunct. Use [espn_games()] instead.
#' 
#' @export

get_espn_events <- function() {
  .Defunct(
    new     = 'espn_games()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_events()` is defunct.',
      'Use `espn_games()` instead.'
    )
  )
}

#' Access the three stars for an ESPN event (game)
#' 
#' `get_espn_event_stars()` is defunct. Use [gc_summary()] and/or 
#' [wsc_summary()] instead.
#' 
#' @export

get_espn_event_stars <- function() {
  .Defunct(
    new     = 'gc_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_event_stars()` is defunct.',
      'Use `gc_summary()` and/or `wsc_summary()` instead.'
    )
  )
}

#' Access the officials for an ESPN event (game)
#' 
#' `get_espn_event_officials()` is defunct. Use [gc_summary()] and/or 
#' [wsc_summary()] instead.
#' 
#' @export

get_espn_event_officials <- function() {
  .Defunct(
    new     = 'gc_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_event_officials()` is defunct.',
      'Use `gc_summary()` and/or `wsc_summary()` instead.'
    )
  )
}

#' Save an Instagram (IG) share-able shot-location plot for a game
#'
#' `ig_game_shot_locations()` saves an IG share-able shot location plot for a 
#' given `game`.
#' 
#' @inheritParams boxscore
#' @param model deprecated legacy model selector; ignored
#' @param save logical only FALSE for tests
#' @returns `NULL`
#' @examples
#' # May take >5s, so skip.
#' \donttest{ig_game_shot_locations(
#'   game  = 2023030417, 
#'   team  = 'H', 
#'   save  = FALSE
#' )}
#' @export

ig_game_shot_locations <- function(
  game  = 2023030417,
  team  = 'home',
  model = NULL,
  save  = TRUE
) {
  tryCatch(
    expr = {
      .xg_warn_ignored_model(model, 'ig_game_shot_locations')
      team <- switch(
        substring(tolower(team), 1, 1),
        h = 'home',
        a = 'away'
      )
      file_name <- sprintf(
        'ig_shot_locs_%s_%s.png',
        as.character(game),
        team
      )
      game_sum <- gc_summary(game)
      home_abbrev <- tryCatch(
        game_sum$homeTeam$abbrev,
        error = function(e) 'HOME'
      )
      away_abbrev <- tryCatch(
        game_sum$awayTeam$abbrev,
        error = function(e) 'AWAY'
      )
      game_date <- tryCatch(
        as.character(game_sum$gameDate),
        error = function(e) ''
      )
      if (team == 'home') {
        shooting_abbrev <- home_abbrev
        opp_abbrev      <- away_abbrev
      } else {
        shooting_abbrev <- away_abbrev
        opp_abbrev      <- home_abbrev
      }
      if (nzchar(game_date)) {
        plot_title <- sprintf(
          '%s %s Shots vs. %s by Outcome and xG, jittered',
          game_date,
          shooting_abbrev,
          opp_abbrev
        )
      } else {
        plot_title <- sprintf(
          '%s Shots vs. %s by Outcome and xG, jittered',
          shooting_abbrev,
          opp_abbrev
        )
      }
      if (isTRUE(save)) {
        grDevices::png(
          filename = file_name,
          width    = 1080 * 1.25,
          height   = 566 * 1.25,
          res      = 144
        )
        on.exit(grDevices::dev.off(), add = TRUE)
      }
      pbp <- gc_play_by_play(game)
      pbp <- calculate_expected_goals(pbp)
      x_col <- 'xCoordNorm'
      y_col <- 'yCoordNorm'
      xg_col <- 'xG'
      type <- as.character(pbp[['eventTypeDescKey']])
      shot_types <- c('goal', 'shot-on-goal', 'missed-shot', 'blocked-shot')
      idx_shot <- .shot_event_mask(pbp, shot_types)
      if (!any(idx_shot)) {
        message('No shot attempts found for this game.')
        return(invisible(NULL))
      }
      shots <- pbp[idx_shot, , drop = FALSE]
      is_home_vec <- as.logical(shots[['isHome']])
      if (team == 'home') {
        keep <- !is.na(is_home_vec) & is_home_vec
      } else {
        keep <- !is.na(is_home_vec) & !is_home_vec
      }
      if (!any(keep)) {
        return(invisible(NULL))
      }
      shots <- shots[keep, , drop = FALSE]
      type_shot <- as.character(shots[['eventTypeDescKey']])
      x <- as.numeric(shots[[x_col]])
      y <- as.numeric(shots[[y_col]])
      x_j <- x + stats::runif(length(x), -0.6, 0.6)
      y_j <- y + stats::runif(length(y), -0.3, 0.3)
      xg <- as.numeric(shots[[xg_col]])
      xg[!is.finite(xg) | xg < 0] <- 0
      xg[xg > 1] <- 1
      pos_xg <- xg[is.finite(xg) & xg > 0]
      xg_cap <- suppressWarnings(
        stats::quantile(
          pos_xg,
          probs = 0.98,
          na.rm = TRUE,
          names = FALSE
        )
      )
      if (!is.finite(xg_cap) || xg_cap <= 0) {
        xg_cap <- 0.20
      }
      xg_cap <- min(1, xg_cap)
      scale_xg <- function(v) {
        vv <- pmax(0, pmin(v, xg_cap))
        log1p(99 * (vv / xg_cap)) / log(100)
      }
      pal_cont <- grDevices::colorRampPalette(
        c('#2166AC', '#67A9CF', '#D1E5F0', '#FDAE61', '#B2182B')
      )(256)
      col_idx <- 1L + floor(scale_xg(xg) * 255)
      col_idx[!is.finite(col_idx)] <- 1L
      col_idx <- pmax(1L, pmin(256L, as.integer(col_idx)))
      col_vec <- pal_cont[col_idx]
      n_ticks <- 4L
      legend_scaled <- seq(0, 1, length.out = n_ticks)
      legend_vals <- xg_cap * (exp(legend_scaled * log(100)) - 1) / 99
      legend_vals[1] <- 0
      legend_vals[n_ticks] <- xg_cap
      label_fmt <- if (xg_cap < 0.1) '%.3f xG' else '%.2f xG'
      top_fmt <- if (xg_cap < 0.1) '%.3f+ xG' else '%.2f+ xG'
      color_labels <- sprintf(label_fmt, legend_vals)
      color_labels[length(color_labels)] <- sprintf(top_fmt, xg_cap)
      pch_vec <- rep(16L, length(type_shot))
      pch_vec[type_shot == 'goal']         <- 8
      pch_vec[type_shot == 'shot-on-goal'] <- 16
      pch_vec[type_shot == 'missed-shot']  <- 17
      pch_vec[type_shot == 'blocked-shot'] <- 15
      draw_NHL_rink()
      graphics::title(main = plot_title, line = 2.8, cex.main = 1.2)
      graphics::points(
        x_j,
        y_j,
        pch = pch_vec,
        col = col_vec
      )
      usr   <- graphics::par('usr')
      rng_y <- usr[4] - usr[3]
      x_mid <- (usr[1] + usr[2]) / 2
      y_top_shapes <- usr[4] + 0.12 * rng_y
      y_top_colors <- usr[4] + 0.06 * rng_y
      old_xpd <- graphics::par('xpd')
      graphics::par(xpd = NA)
      graphics::legend(
        x      = x_mid,
        y      = y_top_shapes - 3,
        horiz  = TRUE,
        xjust  = 0.5,
        legend = c('Goal', 'SOG', 'Missed', 'Blocked'),
        pch    = c(8, 16, 17, 15),
        col    = 'black',
        pt.cex = 0.75,
        bty    = 'n',
        cex    = 0.8
      )
      y_bar_center <- y_top_colors - 7
      rng_x <- usr[2] - usr[1]
      bar_half_w <- 0.20 * rng_x
      bar_h <- 1.7
      x_left <- x_mid - bar_half_w
      x_right <- x_mid + bar_half_w
      y_bottom <- y_bar_center - bar_h / 2
      y_top <- y_bar_center + bar_h / 2
      bar_img <- grDevices::as.raster(matrix(pal_cont, nrow = 1L))
      graphics::rasterImage(
        image       = bar_img,
        xleft       = x_left,
        ybottom     = y_bottom,
        xright      = x_right,
        ytop        = y_top,
        interpolate = FALSE
      )
      graphics::rect(
        xleft   = x_left,
        ybottom = y_bottom,
        xright  = x_right,
        ytop    = y_top,
        border  = 'black',
        lwd     = 0.75
      )
      tick_x <- x_left + legend_scaled * (x_right - x_left)
      graphics::segments(
        x0  = tick_x,
        y0  = y_bottom,
        x1  = tick_x,
        y1  = y_bottom - 0.8
      )
      graphics::text(
        x      = tick_x,
        y      = y_bottom - 1.4,
        labels = color_labels,
        cex    = 0.7,
        adj    = c(0.5, 1)
      )
      graphics::text(
        x      = 65,
        y      = -49,
        labels = 'Data acquired and modeled via R package \'nhlscraper\'',
        cex    = 0.7
      )
      graphics::par(xpd = old_xpd)
      invisible(NULL)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      invisible(NULL)
    }
  )
}

#' @rdname ig_game_shot_locations
#' @export

ig_game_shot_locs <- function(game = 2023030417, team = 'home', model = NULL) {
  ig_game_shot_locations(game, team, model)
}

#' Save an X (Twitter) share-able shot-location plot for a game
#'
#' `x_game_shot_locations()` saves an X share-able shot-location plot for a 
#' given `game`.
#' 
#' @inheritParams ig_game_shot_locations
#' @returns `NULL`
#' @examples
#' # May take >5s, so skip.
#' \donttest{x_game_shot_locations(
#'   game  = 2023030417, 
#'   team  = 'H',
#'   save  = FALSE
#' )}
#' @export

x_game_shot_locations <- function(
  game  = 2023030417,
  team  = 'home',
  model = NULL,
  save  = TRUE
) {
  tryCatch(
    expr = {
      .xg_warn_ignored_model(model, 'x_game_shot_locations')
      team <- switch(
        substring(tolower(team), 1, 1),
        h = 'home',
        a = 'away'
      )
      file_name <- sprintf(
        'x_shot_locs_%s_%s.png',
        as.character(game),
        team
      )
      game_sum <- gc_summary(game)
      home_abbrev <- tryCatch(
        game_sum$homeTeam$abbrev,
        error = function(e) 'HOME'
      )
      away_abbrev <- tryCatch(
        game_sum$awayTeam$abbrev,
        error = function(e) 'AWAY'
      )
      game_date <- tryCatch(
        as.character(game_sum$gameDate),
        error = function(e) ''
      )
      if (team == 'home') {
        shooting_abbrev <- home_abbrev
        opp_abbrev      <- away_abbrev
      } else {
        shooting_abbrev <- away_abbrev
        opp_abbrev      <- home_abbrev
      }
      if (nzchar(game_date)) {
        plot_title <- sprintf(
          '%s %s Shots vs. %s by Outcome and xG, jittered',
          game_date,
          shooting_abbrev,
          opp_abbrev
        )
      } else {
        plot_title <- sprintf(
          '%s Shots vs. %s by Outcome and xG, jittered',
          shooting_abbrev,
          opp_abbrev
        )
      }
      if (isTRUE(save)) {
        grDevices::png(
          filename = file_name,
          width    = 1200 * 1.25,
          height   = 675 * 1.25,
          res      = 144
        )
        on.exit(grDevices::dev.off(), add = TRUE)
      }
      pbp <- gc_play_by_play(game)
      pbp <- calculate_expected_goals(pbp)
      x_col <- 'xCoordNorm'
      y_col <- 'yCoordNorm'
      xg_col <- 'xG'
      type <- as.character(pbp[['eventTypeDescKey']])
      shot_types <- c('goal', 'shot-on-goal', 'missed-shot', 'blocked-shot')
      idx_shot <- .shot_event_mask(pbp, shot_types)
      if (!any(idx_shot)) {
        return(invisible(NULL))
      }
      shots <- pbp[idx_shot, , drop = FALSE]
      is_home_vec <- as.logical(shots[['isHome']])
      if (team == 'home') {
        keep <- !is.na(is_home_vec) & is_home_vec
      } else {
        keep <- !is.na(is_home_vec) & !is_home_vec
      }
      if (!any(keep)) {
        return(invisible(NULL))
      }
      shots <- shots[keep, , drop = FALSE]
      type_shot <- as.character(shots[['eventTypeDescKey']])
      x <- as.numeric(shots[[x_col]])
      y <- as.numeric(shots[[y_col]])
      x_j <- x + stats::runif(length(x), -0.6, 0.6)
      y_j <- y + stats::runif(length(y), -0.3, 0.3)
      xg <- as.numeric(shots[[xg_col]])
      xg[!is.finite(xg) | xg < 0] <- 0
      xg[xg > 1] <- 1
      pos_xg <- xg[is.finite(xg) & xg > 0]
      xg_cap <- suppressWarnings(
        stats::quantile(
          pos_xg,
          probs = 0.98,
          na.rm = TRUE,
          names = FALSE
        )
      )
      if (!is.finite(xg_cap) || xg_cap <= 0) {
        xg_cap <- 0.20
      }
      xg_cap <- min(1, xg_cap)
      scale_xg <- function(v) {
        vv <- pmax(0, pmin(v, xg_cap))
        log1p(99 * (vv / xg_cap)) / log(100)
      }
      pal_cont <- grDevices::colorRampPalette(
        c('#2166AC', '#67A9CF', '#D1E5F0', '#FDAE61', '#B2182B')
      )(256)
      col_idx <- 1L + floor(scale_xg(xg) * 255)
      col_idx[!is.finite(col_idx)] <- 1L
      col_idx <- pmax(1L, pmin(256L, as.integer(col_idx)))
      col_vec <- pal_cont[col_idx]
      n_ticks <- 4L
      legend_scaled <- seq(0, 1, length.out = n_ticks)
      legend_vals <- xg_cap * (exp(legend_scaled * log(100)) - 1) / 99
      legend_vals[1] <- 0
      legend_vals[n_ticks] <- xg_cap
      label_fmt <- if (xg_cap < 0.1) '%.3f xG' else '%.2f xG'
      top_fmt <- if (xg_cap < 0.1) '%.3f+ xG' else '%.2f+ xG'
      color_labels <- sprintf(label_fmt, legend_vals)
      color_labels[length(color_labels)] <- sprintf(top_fmt, xg_cap)
      pch_vec <- rep(16L, length(type_shot))
      pch_vec[type_shot == 'goal']         <- 8
      pch_vec[type_shot == 'shot-on-goal'] <- 16
      pch_vec[type_shot == 'missed-shot']  <- 17
      pch_vec[type_shot == 'blocked-shot'] <- 15
      draw_NHL_rink()
      graphics::title(main = plot_title, line = 2.8, cex.main = 1.2)
      graphics::points(
        x_j,
        y_j,
        pch = pch_vec,
        col = col_vec
      )
      usr   <- graphics::par('usr')
      rng_y <- usr[4] - usr[3]
      x_mid <- (usr[1] + usr[2]) / 2
      y_top_shapes <- usr[4] + 0.12 * rng_y
      y_top_colors <- usr[4] + 0.06 * rng_y
      old_xpd <- graphics::par('xpd')
      graphics::par(xpd = NA)
      graphics::legend(
        x      = x_mid,
        y      = y_top_shapes - 8,
        horiz  = TRUE,
        xjust  = 0.5,
        legend = c('Goal', 'SOG', 'Missed', 'Blocked'),
        pch    = c(8, 16, 17, 15),
        col    = 'black',
        pt.cex = 0.75,
        bty    = 'n',
        cex    = 0.8
      )
      y_bar_center <- y_top_colors - 13
      rng_x <- usr[2] - usr[1]
      bar_half_w <- 0.20 * rng_x
      bar_h <- 1.7
      x_left <- x_mid - bar_half_w
      x_right <- x_mid + bar_half_w
      y_bottom <- y_bar_center - bar_h / 2
      y_top <- y_bar_center + bar_h / 2
      bar_img <- grDevices::as.raster(matrix(pal_cont, nrow = 1L))
      graphics::rasterImage(
        image       = bar_img,
        xleft       = x_left,
        ybottom     = y_bottom,
        xright      = x_right,
        ytop        = y_top,
        interpolate = FALSE
      )
      graphics::rect(
        xleft   = x_left,
        ybottom = y_bottom,
        xright  = x_right,
        ytop    = y_top,
        border  = 'black',
        lwd     = 0.75
      )
      tick_x <- x_left + legend_scaled * (x_right - x_left)
      graphics::segments(
        x0  = tick_x,
        y0  = y_bottom,
        x1  = tick_x,
        y1  = y_bottom - 0.8
      )
      graphics::text(
        x      = tick_x,
        y      = y_bottom - 1.4,
        labels = color_labels,
        cex    = 0.7,
        adj    = c(0.5, 1)
      )
      graphics::text(
        x      = 65,
        y      = -51,
        labels = 'Data acquired and modeled via R package \'nhlscraper\'',
        cex    = 0.7
      )
      graphics::par(xpd = old_xpd)
      invisible(NULL)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      invisible(NULL)
    }
  )
}

#' @rdname x_game_shot_locations
#' @export

x_game_shot_locs <- function(game = 2023030417, team = 'home', model = NULL) {
  x_game_shot_locations(game, team, model)
}

#' Save an Instagram (IG) share-able cumulative expected goals (xG) time-series 
#' plot for a game
#'
#' `ig_game_cumulative_expected_goals()` saves an IG share-able cumulative xG 
#' time-series plot for a given `game` as a PNG.
#'
#' @inheritParams ig_game_shot_locations
#' @returns `NULL`
#' @examples
#' # May take >5s, so skip.
#' \donttest{ig_game_cumulative_expected_goals(
#'   game  = 2023030417, 
#'   save  = FALSE
#' )}
#' @export

ig_game_cumulative_expected_goals <- function(
  game  = 2023030417,
  model = NULL,
  save  = TRUE
) {
  tryCatch(
    expr = {
      .xg_warn_ignored_model(model, 'ig_game_cumulative_expected_goals')
      file_name <- sprintf(
        'ig_cum_xG_%s.png',
        as.character(game)
      )
      game_sum <- gc_summary(game)
      home_abbrev <- tryCatch(
        game_sum$homeTeam$abbrev,
        error = function(e) 'HOME'
      )
      away_abbrev <- tryCatch(
        game_sum$awayTeam$abbrev,
        error = function(e) 'AWAY'
      )
      game_date <- tryCatch(
        as.character(game_sum$gameDate),
        error = function(e) ''
      )
      if (nzchar(game_date)) {
        plot_title <- sprintf(
          '%s %s @ %s xG over Seconds Elapsed',
          game_date,
          away_abbrev,
          home_abbrev
        )
      } else {
        plot_title <- sprintf(
          '%s @ %s xG over Seconds Elapsed',
          away_abbrev,
          home_abbrev
        )
      }
      if (isTRUE(save)) {
        grDevices::png(
          filename = file_name,
          width    = 1080 * 1.25,
          height   = 566 * 1.25,
          res      = 144
        )
        on.exit(grDevices::dev.off(), add = TRUE)
      }
      pbp <- gc_play_by_play(game)
      pbp <- calculate_expected_goals(pbp)
      xg_col <- 'xG'
      type <- as.character(pbp[['eventTypeDescKey']])
      shot_types <- c('goal', 'shot-on-goal', 'missed-shot', 'blocked-shot')
      idx_shot <- .shot_event_mask(pbp, shot_types)
      if (!any(idx_shot)) {
        return(invisible(NULL))
      }
      shots <- pbp[idx_shot, , drop = FALSE]
      sec   <- as.numeric(shots[['secondsElapsedInGame']])
      xg    <- as.numeric(shots[[xg_col]])
      is_h  <- as.logical(shots[['isHome']])
      valid <- !is.na(sec) & !is.na(xg) & !is.na(is_h)
      if (!any(valid)) {
        return(invisible(NULL))
      }
      sec  <- sec[valid]
      xg   <- xg[valid]
      is_h <- is_h[valid]
      home_sec <- sec[is_h]
      home_xg  <- xg[is_h]
      away_sec <- sec[!is_h]
      away_xg  <- xg[!is_h]
      if (length(home_sec) > 0L) {
        o_h <- order(home_sec)
        home_sec <- home_sec[o_h]
        home_xg  <- home_xg[o_h]
        home_t   <- c(0, home_sec)
        home_cum <- c(0, cumsum(home_xg))
      } else {
        home_t   <- 0
        home_cum <- 0
      }
      if (length(away_sec) > 0L) {
        o_a <- order(away_sec)
        away_sec <- away_sec[o_a]
        away_xg  <- away_xg[o_a]
        away_t   <- c(0, away_sec)
        away_cum <- c(0, cumsum(away_xg))
      } else {
        away_t   <- 0
        away_cum <- 0
      }
      game_end_sec <- suppressWarnings(
        max(as.numeric(pbp[['secondsElapsedInGame']]), na.rm = TRUE)
      )
      if (!is.finite(game_end_sec) || game_end_sec <= 0) {
        game_end_sec <- 3600
      }
      if (utils::tail(home_t, 1L) < game_end_sec) {
        home_t   <- c(home_t, game_end_sec)
        home_cum <- c(home_cum, utils::tail(home_cum, 1L))
      }
      if (utils::tail(away_t, 1L) < game_end_sec) {
        away_t   <- c(away_t, game_end_sec)
        away_cum <- c(away_cum, utils::tail(away_cum, 1L))
      }
      max_time <- game_end_sec
      max_xg   <- max(c(home_cum, away_cum), na.rm = TRUE)
      if (!is.finite(max_xg) || max_xg <= 0) {
        max_xg <- 1
      }
      graphics::plot(
        NA_real_, NA_real_,
        xlim = c(0, max_time),
        ylim = c(0, max_xg),
        xlab = 'Seconds Elapsed in Game',
        ylab = 'Cumulative Expected Goals',
        xaxs = 'i',
        yaxs = 'i',
        xaxt = 'n'
      )
      graphics::axis(
        side = 1,
        at   = seq(0, max_time, by = 600)
      )
      graphics::title(main = plot_title, line = 1.8, cex.main = 1.2)
      if (length(home_t) > 1L) {
        graphics::lines(home_t, home_cum, col = 'red', lwd = 2)
      }
      if (length(away_t) > 1L) {
        graphics::lines(away_t, away_cum, col = 'blue', lwd = 2)
      }
      graphics::legend(
        'topleft',
        legend = c(home_abbrev, away_abbrev),
        col    = c('red', 'blue'),
        lty    = 1,
        lwd    = 2,
        bty    = 'n',
        cex    = 0.9
      )
      usr   <- graphics::par('usr')
      rng_y <- usr[4] - usr[3]
      x_mid <- (usr[1] + usr[2]) / 2
      old_xpd <- graphics::par('xpd')
      graphics::par(xpd = NA)
      graphics::text(
        x      = x_mid,
        y      = usr[4] + 0.04 * rng_y,
        labels = "Data acquired and modeled via R package 'nhlscraper'",
        cex    = 0.7,
        adj    = c(0.5, 0)
      )
      graphics::par(xpd = old_xpd)
      invisible(NULL)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      invisible(NULL)
    }
  )
}

#' @rdname ig_game_cumulative_expected_goals
#' @export

ig_game_cum_xG <- function(game = 2023030417, model = NULL) {
  ig_game_cumulative_expected_goals(game, model)
}

#' Save an X (Twitter) share-able cumulative expected goals (xG) time-series 
#' plot for a game
#'
#' `x_game_cumulative_expected_goals()` saves an X share-able cumulative xG 
#' time-series plot for a given `game` as a PNG.
#'
#' @inheritParams ig_game_shot_locations
#' @returns `NULL`
#' @examples
#' # May take >5s, so skip.
#' \donttest{x_game_cumulative_expected_goals(
#'   game  = 2023030417, 
#'   save  = FALSE
#' )}
#' @export

x_game_cumulative_expected_goals <- function(
  game  = 2023030417,
  model = NULL,
  save  = TRUE
) {
  tryCatch(
    expr = {
      .xg_warn_ignored_model(model, 'x_game_cumulative_expected_goals')
      file_name <- sprintf(
        'x_cum_xG_%s.png',
        as.character(game)
      )
      game_sum <- gc_summary(game)
      home_abbrev <- tryCatch(
        game_sum$homeTeam$abbrev,
        error = function(e) 'HOME'
      )
      away_abbrev <- tryCatch(
        game_sum$awayTeam$abbrev,
        error = function(e) 'AWAY'
      )
      game_date <- tryCatch(
        as.character(game_sum$gameDate),
        error = function(e) ''
      )
      if (nzchar(game_date)) {
        plot_title <- sprintf(
          '%s %s @ %s xG over Seconds Elapsed',
          game_date,
          away_abbrev,
          home_abbrev
        )
      } else {
        plot_title <- sprintf(
          '%s @ %s xG over Seconds Elapsed',
          away_abbrev,
          home_abbrev
        )
      }
      if (isTRUE(save)) {
        grDevices::png(
          filename = file_name,
          width    = 1200 * 1.25,
          height   = 675 * 1.25,
          res      = 144
        )
        on.exit(grDevices::dev.off(), add = TRUE)
      }
      pbp <- gc_play_by_play(game)
      pbp <- calculate_expected_goals(pbp)
      xg_col <- 'xG'
      type <- as.character(pbp[['eventTypeDescKey']])
      shot_types <- c('goal', 'shot-on-goal', 'missed-shot', 'blocked-shot')
      idx_shot <- .shot_event_mask(pbp, shot_types)
      if (!any(idx_shot)) {
        return(invisible(NULL))
      }
      shots <- pbp[idx_shot, , drop = FALSE]
      sec   <- as.numeric(shots[['secondsElapsedInGame']])
      xg    <- as.numeric(shots[[xg_col]])
      is_h  <- as.logical(shots[['isHome']])
      valid <- !is.na(sec) & !is.na(xg) & !is.na(is_h)
      if (!any(valid)) {
        return(invisible(NULL))
      }
      sec  <- sec[valid]
      xg   <- xg[valid]
      is_h <- is_h[valid]
      home_sec <- sec[is_h]
      home_xg  <- xg[is_h]
      away_sec <- sec[!is_h]
      away_xg  <- xg[!is_h]
      if (length(home_sec) > 0L) {
        o_h <- order(home_sec)
        home_sec <- home_sec[o_h]
        home_xg  <- home_xg[o_h]
        home_t   <- c(0, home_sec)
        home_cum <- c(0, cumsum(home_xg))
      } else {
        home_t   <- 0
        home_cum <- 0
      }
      if (length(away_sec) > 0L) {
        o_a <- order(away_sec)
        away_sec <- away_sec[o_a]
        away_xg  <- away_xg[o_a]
        away_t   <- c(0, away_sec)
        away_cum <- c(0, cumsum(away_xg))
      } else {
        away_t   <- 0
        away_cum <- 0
      }
      game_end_sec <- suppressWarnings(
        max(as.numeric(pbp[['secondsElapsedInGame']]), na.rm = TRUE)
      )
      if (!is.finite(game_end_sec) || game_end_sec <= 0) {
        game_end_sec <- 3600
      }
      if (utils::tail(home_t, 1L) < game_end_sec) {
        home_t   <- c(home_t, game_end_sec)
        home_cum <- c(home_cum, utils::tail(home_cum, 1L))
      }
      if (utils::tail(away_t, 1L) < game_end_sec) {
        away_t   <- c(away_t, game_end_sec)
        away_cum <- c(away_cum, utils::tail(away_cum, 1L))
      }
      max_time <- game_end_sec
      max_xg   <- max(c(home_cum, away_cum), na.rm = TRUE)
      if (!is.finite(max_xg) || max_xg <= 0) {
        max_xg <- 1
      }
      graphics::plot(
        NA_real_, NA_real_,
        xlim = c(0, max_time),
        ylim = c(0, max_xg),
        xlab = 'Seconds Elapsed in Game',
        ylab = 'Cumulative Expected Goals',
        xaxs = 'i',
        yaxs = 'i',
        xaxt = 'n'
      )
      graphics::axis(
        side = 1,
        at   = seq(0, max_time, by = 600)
      )
      graphics::title(main = plot_title, line = 1.8, cex.main = 1.2)
      if (length(home_t) > 1L) {
        graphics::lines(home_t, home_cum, col = 'red', lwd = 2)
      }
      if (length(away_t) > 1L) {
        graphics::lines(away_t, away_cum, col = 'blue', lwd = 2)
      }
      graphics::legend(
        'topleft',
        legend = c(home_abbrev, away_abbrev),
        col    = c('red', 'blue'),
        lty    = 1,
        lwd    = 2,
        bty    = 'n',
        cex    = 0.9
      )
      usr   <- graphics::par('usr')
      rng_y <- usr[4] - usr[3]
      x_mid <- (usr[1] + usr[2]) / 2
      old_xpd <- graphics::par('xpd')
      graphics::par(xpd = NA)
      graphics::text(
        x      = x_mid,
        y      = usr[4] + 0.04 * rng_y,
        labels = "Data acquired and modeled via R package 'nhlscraper'",
        cex    = 0.7,
        adj    = c(0.5, 0)
      )
      graphics::par(xpd = old_xpd)
      invisible(NULL)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      invisible(NULL)
    }
  )
}

#' @rdname x_game_cumulative_expected_goals
#' @export

x_game_cum_xG <- function(game = 2023030417, model = NULL) {
  x_game_cumulative_expected_goals(game, model)
}

#' Access the real-time ESPN injury reports
#' 
#' `get_espn_injuries()` is deprecated. Use [espn_injuries()] instead.
#' 
#' @returns nested data.frame with one row per team (outer) and player (inner)
#' @export

get_espn_injuries <- function() {
  .Deprecated(
    new     = 'espn_injuries()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_injuries()` is deprecated.',
      'Use `espn_injuries()` instead.'
    )
  )
  espn_injuries()
}

#' Access the ESPN transactions by start and end dates
#' 
#' `get_espn_transactions()` is defunct. Use [espn_transactions()] instead.
#' 
#' @export

get_espn_transactions <- function() {
  .Defunct(
    new     = 'espn_transactions()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_transactions()` is defunct.',
      'Use `espn_transactions()` instead.'
    )
  )
}

#' Access the ESPN futures for a season
#' 
#' `get_espn_futures()` is defunct. Use [espn_futures()] instead.
#' 
#' @export

get_espn_futures <- function() {
  .Defunct(
    new     = 'espn_futures()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_futures()` is defunct.',
      'Use `espn_futures()` instead.'
    )
  )
}

# Generated by data-raw/build_xg_ridge_specs.R. Do not edit by hand.

XG_RIDGE_MODEL_SPECS <-
list(sd = list(coefficients = c(`(Intercept)` = -3.22787367141254, 
distance = -0.844397672863668, angle = -0.437501387022007, xCoordNorm = 0.382296310281, 
shotType_snap = 0.217161232990873, shotType_slap = 0.145733315294426, 
shotType_tip.in = -0.13813603966555, dDistancePerSecond = 0.134836427724351, 
dDistance = -0.132126819850828, dAnglePerSecond = 0.123949457168983, 
shotType_wrist = 0.113686185991433, dXCoordNormPerSecond = -0.102806393143214, 
dAngle = -0.0889687439678296, avgSecondsElapsedInShiftFor = 0.0884868022674456, 
minSecondsElapsedInShiftFor = 0.0802266848926595, typeDescKeyPrev_giveaway.against = 0.0702133378300738, 
crossedRoyalRoad_yes = 0.0681967805345412, zoneCode_N = -0.0633204449743235, 
shotType_deflected = -0.0429071127228005, typeDescKeyPrev_takeaway.for = 0.0421277255495227, 
zoneCode_O = 0.0411923859370841, typeDescKeyPrev_given.hit = -0.0404519703310082, 
typeDescKeyPrev_takeaway.against = -0.0375644395488907, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0352867327824331, 
shooterSecondsElapsedSinceLastShift = -0.035112566546137, typeDescKeyPrev_post.missed.shot.for = 0.0344203147638533, 
avgSecondsElapsedInShiftAgainst = 0.033720487370499, typeDescKeyPrev_won.faceoff = -0.0334709291097645, 
typeDescKeyPrev_lost.faceoff = -0.030131186541664, shooterPositionCode_D = -0.0286202481343203, 
typeDescKeyPrev_taken.hit = -0.0285319909038956, fenwickFor = -0.0281533031382358, 
goalDifferential = 0.0270142915900183, shooterSecondsElapsedInShift = -0.0263349221114251, 
typeDescKeyPrev_wrist.shot.on.goal.for = 0.02602013197317, typeDescKeyPrev_wide.missed.shot.against = 0.0255046483641207, 
fenwickDifferential = -0.0247828807747691, goalsFor = 0.0247594881961234, 
maxSecondsElapsedSinceLastShiftAgainst = 0.0244603204179417, 
secondsElapsedInSequence = 0.0244228360648075, shooterAge = -0.0238311761771997, 
goalieAge = -0.022549263746433, typeDescKeyPrev_slap.shot.on.goal.for = 0.0212588950623637, 
typeDescKeyPrev_post.missed.shot.against = -0.0208901510785224, 
minSecondsElapsedInShiftAgainst = 0.0195462135690771, dSecondsElapsedInSequence = -0.0193140139095899, 
goalieWeight = -0.0186592927322434, typeDescKeyPrev_blocked.shot.against = 0.017976944607056, 
dYCoordNormPerSecond = 0.0171236848669715, isRebound_yes = -0.0165133656374864, 
typeDescKeyPrev_blocked.shot.for = -0.0158735257481945, shooterHandCode_R = 0.0150376845568847, 
shooterWeight = 0.014444865405606, secondsElapsedInPeriod = -0.0141852198665849, 
typeDescKeyPrev_other.shot.on.goal.against = 0.0140008988900591, 
corsiDifferential = 0.0133534605788409, periodNumber = 0.0131925230169041, 
minSecondsElapsedSinceLastShiftAgainst = -0.0120736472366711, 
shooterHeight = -0.0115324408919261, minSecondsElapsedSinceLastShiftFor = 0.0111797494599343, 
typeDescKeyPrev_other.missed.shot.against = 0.0111167920596775, 
typeDescKeyPrev_wrist.shot.on.goal.against = 0.0110954905039129, 
goalieHandCode_R = -0.0109638245914837, typeDescKeyPrev_giveaway.for = 0.0108764295886965, 
shooterPositionCode_R = 0.010532602488184, fenwickAgainst = -0.0105081795194286, 
typeDescKeyPrev_snap.shot.on.goal.against = -0.0100832987473275, 
avgSecondsElapsedSinceLastShiftFor = -0.00961730810323688, dYCoordNorm = -0.00959851243408236, 
typeDescKeyPrev_backhand.shot.on.goal.for = -0.00929347666737915, 
typeDescKeyPrev_other.shot.on.goal.for = 0.00922835094399347, 
isHome_yes = 0.00827747447052165, shooterPositionCode_L = 0.00825648614273755, 
typeDescKeyPrev_high.missed.shot.for = 0.00821946846552225, secondsElapsedInGame = 0.00744617590756035, 
goalsAgainst = -0.00718147985175807, isRush_yes = 0.00704239044427655, 
typeDescKeyPrev_tip.in.shot.on.goal.against = 0.00699134666579523, 
typeDescKeyPrev_high.missed.shot.against = 0.00668033204783943, 
maxSecondsElapsedInShiftAgainst = -0.0063570601797447, corsiAgainst = -0.00621187245589276, 
typeDescKeyPrev_snap.shot.on.goal.for = 0.00611463421246356, 
typeDescKeyPrev_deflected.shot.on.goal.against = 0.00576625439102478, 
isBehindNet_yes = -0.0043988220776942, typeDescKeyPrev_wide.missed.shot.for = -0.00415544985638108, 
isPlayoff_yes = -0.00374612055559669, shotsFor = 0.00368301470260814, 
maxSecondsElapsedInShiftFor = 0.00327183376046586, corsiFor = 0.00301798843405002, 
dXCoordNorm = -0.00287523645397623, shotDifferential = 0.00285323849626911, 
typeDescKeyPrev_deflected.shot.on.goal.for = 0.0020058672830322, 
yCoordNorm = 0.00196555392677064, shotType_other = -0.00175697635451625, 
isOvertime_yes = 0.00161471916964574, shotsAgainst = 0.0012302520886071, 
goalieHeight = 0.00114844451058863, typeDescKeyPrev_slap.shot.on.goal.against = -0.00113110575857476, 
maxSecondsElapsedSinceLastShiftFor = 0.000810198009182974, avgSecondsElapsedSinceLastShiftAgainst = -0.000689695743102425, 
typeDescKeyPrev_other.missed.shot.for = 0.000228850375413098), 
    raw_terms = c("distance", "angle", "xCoordNorm", "dDistancePerSecond", 
    "dDistance", "dAnglePerSecond", "dXCoordNormPerSecond", "dAngle", 
    "avgSecondsElapsedInShiftFor", "minSecondsElapsedInShiftFor", 
    "shooterSecondsElapsedSinceLastShift", "avgSecondsElapsedInShiftAgainst", 
    "fenwickFor", "goalDifferential", "shooterSecondsElapsedInShift", 
    "fenwickDifferential", "goalsFor", "maxSecondsElapsedSinceLastShiftAgainst", 
    "secondsElapsedInSequence", "shooterAge", "goalieAge", "minSecondsElapsedInShiftAgainst", 
    "dSecondsElapsedInSequence", "goalieWeight", "dYCoordNormPerSecond", 
    "shooterWeight", "secondsElapsedInPeriod", "corsiDifferential", 
    "periodNumber", "minSecondsElapsedSinceLastShiftAgainst", 
    "shooterHeight", "minSecondsElapsedSinceLastShiftFor", "fenwickAgainst", 
    "avgSecondsElapsedSinceLastShiftFor", "dYCoordNorm", "secondsElapsedInGame", 
    "goalsAgainst", "maxSecondsElapsedInShiftAgainst", "corsiAgainst", 
    "shotsFor", "maxSecondsElapsedInShiftFor", "corsiFor", "dXCoordNorm", 
    "shotDifferential", "yCoordNorm", "shotsAgainst", "goalieHeight", 
    "maxSecondsElapsedSinceLastShiftFor", "avgSecondsElapsedSinceLastShiftAgainst"
    ), active_dummy_terms = c("shotType_snap", "shotType_slap", 
    "shotType_tip.in", "shotType_wrist", "typeDescKeyPrev_giveaway.against", 
    "crossedRoyalRoad_yes", "zoneCode_N", "shotType_deflected", 
    "typeDescKeyPrev_takeaway.for", "zoneCode_O", "typeDescKeyPrev_given.hit", 
    "typeDescKeyPrev_takeaway.against", "typeDescKeyPrev_tip.in.shot.on.goal.for", 
    "typeDescKeyPrev_post.missed.shot.for", "typeDescKeyPrev_won.faceoff", 
    "typeDescKeyPrev_lost.faceoff", "shooterPositionCode_D", 
    "typeDescKeyPrev_taken.hit", "typeDescKeyPrev_wrist.shot.on.goal.for", 
    "typeDescKeyPrev_wide.missed.shot.against", "typeDescKeyPrev_slap.shot.on.goal.for", 
    "typeDescKeyPrev_post.missed.shot.against", "typeDescKeyPrev_blocked.shot.against", 
    "isRebound_yes", "typeDescKeyPrev_blocked.shot.for", "shooterHandCode_R", 
    "typeDescKeyPrev_other.shot.on.goal.against", "typeDescKeyPrev_other.missed.shot.against", 
    "typeDescKeyPrev_wrist.shot.on.goal.against", "goalieHandCode_R", 
    "typeDescKeyPrev_giveaway.for", "shooterPositionCode_R", 
    "typeDescKeyPrev_snap.shot.on.goal.against", "typeDescKeyPrev_backhand.shot.on.goal.for", 
    "typeDescKeyPrev_other.shot.on.goal.for", "isHome_yes", "shooterPositionCode_L", 
    "typeDescKeyPrev_high.missed.shot.for", "isRush_yes", "typeDescKeyPrev_tip.in.shot.on.goal.against", 
    "typeDescKeyPrev_high.missed.shot.against", "typeDescKeyPrev_snap.shot.on.goal.for", 
    "typeDescKeyPrev_deflected.shot.on.goal.against", "isBehindNet_yes", 
    "typeDescKeyPrev_wide.missed.shot.for", "isPlayoff_yes", 
    "typeDescKeyPrev_deflected.shot.on.goal.for", "shotType_other", 
    "isOvertime_yes", "typeDescKeyPrev_slap.shot.on.goal.against", 
    "typeDescKeyPrev_other.missed.shot.for"), active_dummy_by_var = list(
        crossedRoyalRoad = "crossedRoyalRoad_yes", goalieHandCode = "goalieHandCode_R", 
        isBehindNet = "isBehindNet_yes", isHome = "isHome_yes", 
        isOvertime = "isOvertime_yes", isPlayoff = "isPlayoff_yes", 
        isRebound = "isRebound_yes", isRush = "isRush_yes", shooterHandCode = "shooterHandCode_R", 
        shooterPositionCode = c("shooterPositionCode_D", "shooterPositionCode_R", 
        "shooterPositionCode_L"), shotType = c("shotType_snap", 
        "shotType_slap", "shotType_tip.in", "shotType_wrist", 
        "shotType_deflected", "shotType_other"), typeDescKeyPrev = c("typeDescKeyPrev_giveaway.against", 
        "typeDescKeyPrev_takeaway.for", "typeDescKeyPrev_given.hit", 
        "typeDescKeyPrev_takeaway.against", "typeDescKeyPrev_tip.in.shot.on.goal.for", 
        "typeDescKeyPrev_post.missed.shot.for", "typeDescKeyPrev_won.faceoff", 
        "typeDescKeyPrev_lost.faceoff", "typeDescKeyPrev_taken.hit", 
        "typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_wide.missed.shot.against", 
        "typeDescKeyPrev_slap.shot.on.goal.for", "typeDescKeyPrev_post.missed.shot.against", 
        "typeDescKeyPrev_blocked.shot.against", "typeDescKeyPrev_blocked.shot.for", 
        "typeDescKeyPrev_other.shot.on.goal.against", "typeDescKeyPrev_other.missed.shot.against", 
        "typeDescKeyPrev_wrist.shot.on.goal.against", "typeDescKeyPrev_giveaway.for", 
        "typeDescKeyPrev_snap.shot.on.goal.against", "typeDescKeyPrev_backhand.shot.on.goal.for", 
        "typeDescKeyPrev_other.shot.on.goal.for", "typeDescKeyPrev_high.missed.shot.for", 
        "typeDescKeyPrev_tip.in.shot.on.goal.against", "typeDescKeyPrev_high.missed.shot.against", 
        "typeDescKeyPrev_snap.shot.on.goal.for", "typeDescKeyPrev_deflected.shot.on.goal.against", 
        "typeDescKeyPrev_wide.missed.shot.for", "typeDescKeyPrev_deflected.shot.on.goal.for", 
        "typeDescKeyPrev_slap.shot.on.goal.against", "typeDescKeyPrev_other.missed.shot.for"
        ), zoneCode = c("zoneCode_N", "zoneCode_O")), dummy_term_to_var = c(isPlayoff_yes = "isPlayoff", 
    isPlayoff_unknown = "isPlayoff", isPlayoff_new = "isPlayoff", 
    isHome_yes = "isHome", isHome_unknown = "isHome", isHome_new = "isHome", 
    isOvertime_yes = "isOvertime", isOvertime_unknown = "isOvertime", 
    isOvertime_new = "isOvertime", zoneCode_N = "zoneCode", zoneCode_O = "zoneCode", 
    zoneCode_unknown = "zoneCode", zoneCode_new = "zoneCode", 
    isBehindNet_yes = "isBehindNet", isBehindNet_unknown = "isBehindNet", 
    isBehindNet_new = "isBehindNet", crossedRoyalRoad_yes = "crossedRoyalRoad", 
    crossedRoyalRoad_unknown = "crossedRoyalRoad", crossedRoyalRoad_new = "crossedRoyalRoad", 
    typeDescKeyPrev_backhand.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_blocked.shot.against = "typeDescKeyPrev", 
    typeDescKeyPrev_blocked.shot.for = "typeDescKeyPrev", typeDescKeyPrev_deflected.shot.on.goal.against = "typeDescKeyPrev", 
    typeDescKeyPrev_deflected.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_giveaway.against = "typeDescKeyPrev", typeDescKeyPrev_giveaway.for = "typeDescKeyPrev", 
    typeDescKeyPrev_given.hit = "typeDescKeyPrev", typeDescKeyPrev_high.missed.shot.against = "typeDescKeyPrev", 
    typeDescKeyPrev_high.missed.shot.for = "typeDescKeyPrev", 
    typeDescKeyPrev_lost.faceoff = "typeDescKeyPrev", typeDescKeyPrev_other.missed.shot.against = "typeDescKeyPrev", 
    typeDescKeyPrev_other.missed.shot.for = "typeDescKeyPrev", 
    typeDescKeyPrev_other.shot.on.goal.against = "typeDescKeyPrev", 
    typeDescKeyPrev_other.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_post.missed.shot.against = "typeDescKeyPrev", 
    typeDescKeyPrev_post.missed.shot.for = "typeDescKeyPrev", 
    typeDescKeyPrev_slap.shot.on.goal.against = "typeDescKeyPrev", 
    typeDescKeyPrev_slap.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_snap.shot.on.goal.against = "typeDescKeyPrev", 
    typeDescKeyPrev_snap.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_takeaway.against = "typeDescKeyPrev", typeDescKeyPrev_takeaway.for = "typeDescKeyPrev", 
    typeDescKeyPrev_taken.hit = "typeDescKeyPrev", typeDescKeyPrev_tip.in.shot.on.goal.against = "typeDescKeyPrev", 
    typeDescKeyPrev_tip.in.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_wide.missed.shot.against = "typeDescKeyPrev", 
    typeDescKeyPrev_wide.missed.shot.for = "typeDescKeyPrev", 
    typeDescKeyPrev_won.faceoff = "typeDescKeyPrev", typeDescKeyPrev_wrist.shot.on.goal.against = "typeDescKeyPrev", 
    typeDescKeyPrev_wrist.shot.on.goal.for = "typeDescKeyPrev", 
    typeDescKeyPrev_unknown = "typeDescKeyPrev", typeDescKeyPrev_new = "typeDescKeyPrev", 
    shotType_deflected = "shotType", shotType_other = "shotType", 
    shotType_slap = "shotType", shotType_snap = "shotType", shotType_tip.in = "shotType", 
    shotType_wrist = "shotType", shotType_unknown = "shotType", 
    shotType_new = "shotType", isRebound_yes = "isRebound", isRebound_unknown = "isRebound", 
    isRebound_new = "isRebound", isRush_yes = "isRush", isRush_unknown = "isRush", 
    isRush_new = "isRush", shooterHandCode_R = "shooterHandCode", 
    shooterHandCode_unknown = "shooterHandCode", shooterHandCode_new = "shooterHandCode", 
    shooterPositionCode_D = "shooterPositionCode", shooterPositionCode_L = "shooterPositionCode", 
    shooterPositionCode_R = "shooterPositionCode", shooterPositionCode_unknown = "shooterPositionCode", 
    shooterPositionCode_new = "shooterPositionCode", goalieHandCode_R = "goalieHandCode", 
    goalieHandCode_unknown = "goalieHandCode", goalieHandCode_new = "goalieHandCode"
    ), dummy_term_to_level = c(isPlayoff_yes = "yes", isPlayoff_unknown = "unknown", 
    isPlayoff_new = "new", isHome_yes = "yes", isHome_unknown = "unknown", 
    isHome_new = "new", isOvertime_yes = "yes", isOvertime_unknown = "unknown", 
    isOvertime_new = "new", zoneCode_N = "N", zoneCode_O = "O", 
    zoneCode_unknown = "unknown", zoneCode_new = "new", isBehindNet_yes = "yes", 
    isBehindNet_unknown = "unknown", isBehindNet_new = "new", 
    crossedRoyalRoad_yes = "yes", crossedRoyalRoad_unknown = "unknown", 
    crossedRoyalRoad_new = "new", typeDescKeyPrev_backhand.shot.on.goal.for = "backhand-shot-on-goal-for", 
    typeDescKeyPrev_blocked.shot.against = "blocked-shot-against", 
    typeDescKeyPrev_blocked.shot.for = "blocked-shot-for", typeDescKeyPrev_deflected.shot.on.goal.against = "deflected-shot-on-goal-against", 
    typeDescKeyPrev_deflected.shot.on.goal.for = "deflected-shot-on-goal-for", 
    typeDescKeyPrev_giveaway.against = "giveaway-against", typeDescKeyPrev_giveaway.for = "giveaway-for", 
    typeDescKeyPrev_given.hit = "given-hit", typeDescKeyPrev_high.missed.shot.against = "high-missed-shot-against", 
    typeDescKeyPrev_high.missed.shot.for = "high-missed-shot-for", 
    typeDescKeyPrev_lost.faceoff = "lost-faceoff", typeDescKeyPrev_other.missed.shot.against = "other-missed-shot-against", 
    typeDescKeyPrev_other.missed.shot.for = "other-missed-shot-for", 
    typeDescKeyPrev_other.shot.on.goal.against = "other-shot-on-goal-against", 
    typeDescKeyPrev_other.shot.on.goal.for = "other-shot-on-goal-for", 
    typeDescKeyPrev_post.missed.shot.against = "post-missed-shot-against", 
    typeDescKeyPrev_post.missed.shot.for = "post-missed-shot-for", 
    typeDescKeyPrev_slap.shot.on.goal.against = "slap-shot-on-goal-against", 
    typeDescKeyPrev_slap.shot.on.goal.for = "slap-shot-on-goal-for", 
    typeDescKeyPrev_snap.shot.on.goal.against = "snap-shot-on-goal-against", 
    typeDescKeyPrev_snap.shot.on.goal.for = "snap-shot-on-goal-for", 
    typeDescKeyPrev_takeaway.against = "takeaway-against", typeDescKeyPrev_takeaway.for = "takeaway-for", 
    typeDescKeyPrev_taken.hit = "taken-hit", typeDescKeyPrev_tip.in.shot.on.goal.against = "tip-in-shot-on-goal-against", 
    typeDescKeyPrev_tip.in.shot.on.goal.for = "tip-in-shot-on-goal-for", 
    typeDescKeyPrev_wide.missed.shot.against = "wide-missed-shot-against", 
    typeDescKeyPrev_wide.missed.shot.for = "wide-missed-shot-for", 
    typeDescKeyPrev_won.faceoff = "won-faceoff", typeDescKeyPrev_wrist.shot.on.goal.against = "wrist-shot-on-goal-against", 
    typeDescKeyPrev_wrist.shot.on.goal.for = "wrist-shot-on-goal-for", 
    typeDescKeyPrev_unknown = "unknown", typeDescKeyPrev_new = "new", 
    shotType_deflected = "deflected", shotType_other = "other", 
    shotType_slap = "slap", shotType_snap = "snap", shotType_tip.in = "tip-in", 
    shotType_wrist = "wrist", shotType_unknown = "unknown", shotType_new = "new", 
    isRebound_yes = "yes", isRebound_unknown = "unknown", isRebound_new = "new", 
    isRush_yes = "yes", isRush_unknown = "unknown", isRush_new = "new", 
    shooterHandCode_R = "R", shooterHandCode_unknown = "unknown", 
    shooterHandCode_new = "new", shooterPositionCode_D = "D", 
    shooterPositionCode_L = "L", shooterPositionCode_R = "R", 
    shooterPositionCode_unknown = "unknown", shooterPositionCode_new = "new", 
    goalieHandCode_R = "R", goalieHandCode_unknown = "unknown", 
    goalieHandCode_new = "new"), dummy_map = list(crossedRoyalRoad = c(yes = "crossedRoyalRoad_yes", 
    unknown = "crossedRoyalRoad_unknown", new = "crossedRoyalRoad_new"
    ), goalieHandCode = c(R = "goalieHandCode_R", unknown = "goalieHandCode_unknown", 
    new = "goalieHandCode_new"), isBehindNet = c(yes = "isBehindNet_yes", 
    unknown = "isBehindNet_unknown", new = "isBehindNet_new"), 
        isHome = c(yes = "isHome_yes", unknown = "isHome_unknown", 
        new = "isHome_new"), isOvertime = c(yes = "isOvertime_yes", 
        unknown = "isOvertime_unknown", new = "isOvertime_new"
        ), isPlayoff = c(yes = "isPlayoff_yes", unknown = "isPlayoff_unknown", 
        new = "isPlayoff_new"), isRebound = c(yes = "isRebound_yes", 
        unknown = "isRebound_unknown", new = "isRebound_new"), 
        isRush = c(yes = "isRush_yes", unknown = "isRush_unknown", 
        new = "isRush_new"), shooterHandCode = c(R = "shooterHandCode_R", 
        unknown = "shooterHandCode_unknown", new = "shooterHandCode_new"
        ), shooterPositionCode = c(D = "shooterPositionCode_D", 
        L = "shooterPositionCode_L", R = "shooterPositionCode_R", 
        unknown = "shooterPositionCode_unknown", new = "shooterPositionCode_new"
        ), shotType = c(deflected = "shotType_deflected", other = "shotType_other", 
        slap = "shotType_slap", snap = "shotType_snap", `tip-in` = "shotType_tip.in", 
        wrist = "shotType_wrist", unknown = "shotType_unknown", 
        new = "shotType_new"), typeDescKeyPrev = c(`backhand-shot-on-goal-for` = "typeDescKeyPrev_backhand.shot.on.goal.for", 
        `blocked-shot-against` = "typeDescKeyPrev_blocked.shot.against", 
        `blocked-shot-for` = "typeDescKeyPrev_blocked.shot.for", 
        `deflected-shot-on-goal-against` = "typeDescKeyPrev_deflected.shot.on.goal.against", 
        `deflected-shot-on-goal-for` = "typeDescKeyPrev_deflected.shot.on.goal.for", 
        `giveaway-against` = "typeDescKeyPrev_giveaway.against", 
        `giveaway-for` = "typeDescKeyPrev_giveaway.for", `given-hit` = "typeDescKeyPrev_given.hit", 
        `high-missed-shot-against` = "typeDescKeyPrev_high.missed.shot.against", 
        `high-missed-shot-for` = "typeDescKeyPrev_high.missed.shot.for", 
        `lost-faceoff` = "typeDescKeyPrev_lost.faceoff", `other-missed-shot-against` = "typeDescKeyPrev_other.missed.shot.against", 
        `other-missed-shot-for` = "typeDescKeyPrev_other.missed.shot.for", 
        `other-shot-on-goal-against` = "typeDescKeyPrev_other.shot.on.goal.against", 
        `other-shot-on-goal-for` = "typeDescKeyPrev_other.shot.on.goal.for", 
        `post-missed-shot-against` = "typeDescKeyPrev_post.missed.shot.against", 
        `post-missed-shot-for` = "typeDescKeyPrev_post.missed.shot.for", 
        `slap-shot-on-goal-against` = "typeDescKeyPrev_slap.shot.on.goal.against", 
        `slap-shot-on-goal-for` = "typeDescKeyPrev_slap.shot.on.goal.for", 
        `snap-shot-on-goal-against` = "typeDescKeyPrev_snap.shot.on.goal.against", 
        `snap-shot-on-goal-for` = "typeDescKeyPrev_snap.shot.on.goal.for", 
        `takeaway-against` = "typeDescKeyPrev_takeaway.against", 
        `takeaway-for` = "typeDescKeyPrev_takeaway.for", `taken-hit` = "typeDescKeyPrev_taken.hit", 
        `tip-in-shot-on-goal-against` = "typeDescKeyPrev_tip.in.shot.on.goal.against", 
        `tip-in-shot-on-goal-for` = "typeDescKeyPrev_tip.in.shot.on.goal.for", 
        `wide-missed-shot-against` = "typeDescKeyPrev_wide.missed.shot.against", 
        `wide-missed-shot-for` = "typeDescKeyPrev_wide.missed.shot.for", 
        `won-faceoff` = "typeDescKeyPrev_won.faceoff", `wrist-shot-on-goal-against` = "typeDescKeyPrev_wrist.shot.on.goal.against", 
        `wrist-shot-on-goal-for` = "typeDescKeyPrev_wrist.shot.on.goal.for", 
        unknown = "typeDescKeyPrev_unknown", new = "typeDescKeyPrev_new"
        ), zoneCode = c(N = "zoneCode_N", O = "zoneCode_O", unknown = "zoneCode_unknown", 
        new = "zoneCode_new")), known_levels = list(isPlayoff = c("no", 
    "yes"), isHome = c("no", "yes"), isOvertime = c("no", "yes"
    ), zoneCode = c("D", "N", "O"), isBehindNet = c("no", "yes"
    ), crossedRoyalRoad = c("no", "yes"), typeDescKeyPrev = c("backhand-shot-on-goal-against", 
    "backhand-shot-on-goal-for", "blocked-shot-against", "blocked-shot-for", 
    "deflected-shot-on-goal-against", "deflected-shot-on-goal-for", 
    "giveaway-against", "giveaway-for", "given-hit", "high-missed-shot-against", 
    "high-missed-shot-for", "lost-faceoff", "other-missed-shot-against", 
    "other-missed-shot-for", "other-shot-on-goal-against", "other-shot-on-goal-for", 
    "post-missed-shot-against", "post-missed-shot-for", "slap-shot-on-goal-against", 
    "slap-shot-on-goal-for", "snap-shot-on-goal-against", "snap-shot-on-goal-for", 
    "takeaway-against", "takeaway-for", "taken-hit", "tip-in-shot-on-goal-against", 
    "tip-in-shot-on-goal-for", "wide-missed-shot-against", "wide-missed-shot-for", 
    "won-faceoff", "wrist-shot-on-goal-against", "wrist-shot-on-goal-for"
    ), shotType = c("backhand", "deflected", "other", "slap", 
    "snap", "tip-in", "wrist"), isRebound = c("no", "yes"), isRush = c("no", 
    "yes"), shooterHandCode = c("L", "R"), shooterPositionCode = c("C", 
    "D", "L", "R"), goalieHandCode = c("L", "R")), logical_vars = c("crossedRoyalRoad", 
    "isBehindNet", "isHome", "isOvertime", "isPlayoff", "isRebound", 
    "isRush"), has_unknown = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
    isBehindNet = TRUE, isHome = TRUE, isOvertime = TRUE, isPlayoff = TRUE, 
    isRebound = TRUE, isRush = TRUE, shooterHandCode = TRUE, 
    shooterPositionCode = TRUE, shotType = TRUE, typeDescKeyPrev = TRUE, 
    zoneCode = TRUE), has_new = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
    isBehindNet = TRUE, isHome = TRUE, isOvertime = TRUE, isPlayoff = TRUE, 
    isRebound = TRUE, isRush = TRUE, shooterHandCode = TRUE, 
    shooterPositionCode = TRUE, shotType = TRUE, typeDescKeyPrev = TRUE, 
    zoneCode = TRUE), medians = c(distance = 33.9411254969543, 
    angle = 31.8659776936037, xCoordNorm = 64, shotType_snap = 0, 
    shotType_slap = 0, shotType_tip.in = 0, dDistancePerSecond = -1.34742835543301, 
    dDistance = -14.9899563782146, dAnglePerSecond = 0.302473865727572, 
    shotType_wrist = 1, dXCoordNormPerSecond = 1.16666666666667, 
    dAngle = 4.49800470013907, avgSecondsElapsedInShiftFor = 26.2, 
    minSecondsElapsedInShiftFor = 13, typeDescKeyPrev_giveaway.against = 0, 
    crossedRoyalRoad_yes = 0, zoneCode_N = 0, shotType_deflected = 0, 
    typeDescKeyPrev_takeaway.for = 0, zoneCode_O = 1, typeDescKeyPrev_given.hit = 0, 
    typeDescKeyPrev_takeaway.against = 0, typeDescKeyPrev_tip.in.shot.on.goal.for = 0, 
    shooterSecondsElapsedSinceLastShift = 164, typeDescKeyPrev_post.missed.shot.for = 0, 
    avgSecondsElapsedInShiftAgainst = 31, typeDescKeyPrev_won.faceoff = 0, 
    typeDescKeyPrev_lost.faceoff = 0, shooterPositionCode_D = 0, 
    typeDescKeyPrev_taken.hit = 0, fenwickFor = 20, goalDifferential = 0, 
    shooterSecondsElapsedInShift = 26, typeDescKeyPrev_wrist.shot.on.goal.for = 0, 
    typeDescKeyPrev_wide.missed.shot.against = 0, fenwickDifferential = 0, 
    goalsFor = 1, maxSecondsElapsedSinceLastShiftAgainst = 221, 
    secondsElapsedInSequence = 47, shooterAge = 28, goalieAge = 28, 
    typeDescKeyPrev_slap.shot.on.goal.for = 0, typeDescKeyPrev_post.missed.shot.against = 0, 
    minSecondsElapsedInShiftAgainst = 18, dSecondsElapsedInSequence = 10, 
    goalieWeight = 201, typeDescKeyPrev_blocked.shot.against = 0, 
    dYCoordNormPerSecond = 0, isRebound_yes = 0, typeDescKeyPrev_blocked.shot.for = 0, 
    shooterHandCode_R = 0, shooterWeight = 201, secondsElapsedInPeriod = 574, 
    typeDescKeyPrev_other.shot.on.goal.against = 0, corsiDifferential = 1, 
    periodNumber = 2, minSecondsElapsedSinceLastShiftAgainst = 112, 
    shooterHeight = 73, minSecondsElapsedSinceLastShiftFor = 105, 
    typeDescKeyPrev_other.missed.shot.against = 0, typeDescKeyPrev_wrist.shot.on.goal.against = 0, 
    goalieHandCode_R = 0, typeDescKeyPrev_giveaway.for = 0, shooterPositionCode_R = 0, 
    fenwickAgainst = 20, typeDescKeyPrev_snap.shot.on.goal.against = 0, 
    avgSecondsElapsedSinceLastShiftFor = 163.6, dYCoordNorm = 0, 
    typeDescKeyPrev_backhand.shot.on.goal.for = 0, typeDescKeyPrev_other.shot.on.goal.for = 0, 
    isHome_yes = 1, shooterPositionCode_L = 0, typeDescKeyPrev_high.missed.shot.for = 0, 
    secondsElapsedInGame = 1736, goalsAgainst = 1, isRush_yes = 0, 
    typeDescKeyPrev_tip.in.shot.on.goal.against = 0, typeDescKeyPrev_high.missed.shot.against = 0, 
    maxSecondsElapsedInShiftAgainst = 43, corsiAgainst = 27, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0, typeDescKeyPrev_deflected.shot.on.goal.against = 0, 
    isBehindNet_yes = 0, typeDescKeyPrev_wide.missed.shot.for = 0, 
    isPlayoff_yes = 0, shotsFor = 13, maxSecondsElapsedInShiftFor = 38, 
    corsiFor = 28, dXCoordNorm = 12, shotDifferential = 0, typeDescKeyPrev_deflected.shot.on.goal.for = 0, 
    yCoordNorm = 0, shotType_other = 0, isOvertime_yes = 0, shotsAgainst = 13, 
    goalieHeight = 75, typeDescKeyPrev_slap.shot.on.goal.against = 0, 
    maxSecondsElapsedSinceLastShiftFor = 214, avgSecondsElapsedSinceLastShiftAgainst = 169.6, 
    typeDescKeyPrev_other.missed.shot.for = 0), normalize_means = c(distance = 36.3007994468372, 
    angle = 34.3283512744358, xCoordNorm = 59.2567478984105, 
    shotType_snap = 0.174876369376314, shotType_slap = 0.105616999407412, 
    shotType_tip.in = 0.0870999375947223, dDistancePerSecond = -7.20523105149532, 
    dDistance = -32.9401886746547, dAnglePerSecond = 0.405452701172801, 
    shotType_wrist = 0.521839225125728, dXCoordNormPerSecond = 7.01635321671675, 
    dAngle = 1.38695676867213, avgSecondsElapsedInShiftFor = 27.3430355185671, 
    minSecondsElapsedInShiftFor = 15.9005501103892, typeDescKeyPrev_giveaway.against = 0.0792966558112948, 
    crossedRoyalRoad_yes = 0.441913272396415, zoneCode_N = 0.0287955781403227, 
    shotType_deflected = 0.0186429211962934, typeDescKeyPrev_takeaway.for = 0.0402540261892401, 
    zoneCode_O = 0.960831510994342, typeDescKeyPrev_given.hit = 0.0871156700176728, 
    typeDescKeyPrev_takeaway.against = 0.0146993271767118, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0100163092784586, 
    shooterSecondsElapsedSinceLastShift = 195.380750856106, typeDescKeyPrev_post.missed.shot.for = 0.00592587931133941, 
    avgSecondsElapsedInShiftAgainst = 32.5713774785119, typeDescKeyPrev_won.faceoff = 0.146217138901562, 
    typeDescKeyPrev_lost.faceoff = 0.0723691455721096, shooterPositionCode_D = 0.303688204353686, 
    typeDescKeyPrev_taken.hit = 0.114290808594098, fenwickFor = 21.0978556707519, 
    goalDifferential = -0.0893916272045058, shooterSecondsElapsedInShift = 28.6568863437325, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.0686510496148178, 
    typeDescKeyPrev_wide.missed.shot.against = 0.0242646403305906, 
    fenwickDifferential = 0.623444456680773, goalsFor = 1.29255489304574, 
    maxSecondsElapsedSinceLastShiftAgainst = 259.220426977959, 
    secondsElapsedInSequence = 65.9611671360173, shooterAge = 27.7450193771009, 
    goalieAge = 28.6089076978745, typeDescKeyPrev_slap.shot.on.goal.for = 0.0118832234685797, 
    typeDescKeyPrev_post.missed.shot.against = 0.00190886731798898, 
    minSecondsElapsedInShiftAgainst = 21.247061970014, dSecondsElapsedInSequence = 14.4934212251362, 
    goalieWeight = 201.744442521593, typeDescKeyPrev_blocked.shot.against = 0.0398344949105612, 
    dYCoordNormPerSecond = 0.118797340663537, isRebound_yes = 0.102423317548469, 
    typeDescKeyPrev_blocked.shot.for = 0.0963820671354929, shooterHandCode_R = 0.375254996355322, 
    shooterWeight = 201.209293666651, secondsElapsedInPeriod = 581.135828495613, 
    typeDescKeyPrev_other.shot.on.goal.against = 0.00052441409834862, 
    corsiDifferential = 0.944842125135692, periodNumber = 1.97567767411859, 
    minSecondsElapsedSinceLastShiftAgainst = 137.961539470027, 
    shooterHeight = 73.2677763268988, minSecondsElapsedSinceLastShiftFor = 129.907178704592, 
    typeDescKeyPrev_other.missed.shot.against = 0.00144738291144219, 
    typeDescKeyPrev_wrist.shot.on.goal.against = 0.0227071304584952, 
    goalieHandCode_R = 0.0529343590873097, typeDescKeyPrev_giveaway.for = 0.0284756855403301, 
    shooterPositionCode_R = 0.174205119330428, fenwickAgainst = 20.4744112140711, 
    typeDescKeyPrev_snap.shot.on.goal.against = 0.00771937552769169, 
    avgSecondsElapsedSinceLastShiftFor = 189.955641384665, dYCoordNorm = 0.202214076323228, 
    typeDescKeyPrev_backhand.shot.on.goal.for = 0.0107190241702458, 
    typeDescKeyPrev_other.shot.on.goal.for = 0.00287378925895044, 
    isHome_yes = 0.508172993722763, shooterPositionCode_L = 0.201831254031433, 
    typeDescKeyPrev_high.missed.shot.for = 0.0042162893507229, 
    secondsElapsedInGame = 1751.94903743792, goalsAgainst = 1.38194652025025, 
    isRush_yes = 0.0751275637294233, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.00301013692452108, 
    typeDescKeyPrev_high.missed.shot.against = 0.00190886731798898, 
    maxSecondsElapsedInShiftAgainst = 45.287704062636, corsiAgainst = 28.4937883150051, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0.0215848842880292, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0.00071844731473761, 
    isBehindNet_yes = 0.0210604701896806, typeDescKeyPrev_wide.missed.shot.for = 0.0684727488213793, 
    isPlayoff_yes = 0.0639522992936142, shotsFor = 14.1932675718054, 
    maxSecondsElapsedInShiftFor = 39.578580830567, corsiFor = 29.4386304401408, 
    dXCoordNorm = 32.2281568417685, shotDifferential = 0.340003880664328, 
    typeDescKeyPrev_deflected.shot.on.goal.for = 0.00234413101961833, 
    yCoordNorm = 0.559964130075673, shotType_other = 0.0170382140553467, 
    isOvertime_yes = 0.00318843771795961, shotsAgainst = 13.8532636911411, 
    goalieHeight = 74.9203100336149, typeDescKeyPrev_slap.shot.on.goal.against = 0.00436312529826052, 
    maxSecondsElapsedSinceLastShiftFor = 252.561972636072, avgSecondsElapsedSinceLastShiftAgainst = 196.324328094438, 
    typeDescKeyPrev_other.missed.shot.for = 0.00285805683599998
    ), normalize_sds = c(distance = 23.7965266074563, angle = 22.5379315511933, 
    xCoordNorm = 23.698385334075, shotType_snap = 0.379862319157004, 
    shotType_slap = 0.30734759510749, shotType_tip.in = 0.281982189948842, 
    dDistancePerSecond = 26.6574028380376, dDistance = 60.3035150311097, 
    dAnglePerSecond = 14.0976613368947, shotType_wrist = 0.499524130335086, 
    dXCoordNormPerSecond = 27.9095640411456, dAngle = 37.7462322925232, 
    avgSecondsElapsedInShiftFor = 13.3763328136724, minSecondsElapsedInShiftFor = 13.2996730659043, 
    typeDescKeyPrev_giveaway.against = 0.270201922750787, crossedRoyalRoad_yes = 0.496615772428559, 
    zoneCode_N = 0.167231992992742, shotType_deflected = 0.135260706155197, 
    typeDescKeyPrev_takeaway.for = 0.196554934219564, zoneCode_O = 0.19399617479603, 
    typeDescKeyPrev_given.hit = 0.282005225316335, typeDescKeyPrev_takeaway.against = 0.120346719564196, 
    typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0995792891517513, 
    shooterSecondsElapsedSinceLastShift = 114.117774308454, typeDescKeyPrev_post.missed.shot.for = 0.0767515091571383, 
    avgSecondsElapsedInShiftAgainst = 16.781545415302, typeDescKeyPrev_won.faceoff = 0.353324697499104, 
    typeDescKeyPrev_lost.faceoff = 0.25909883132144, shooterPositionCode_D = 0.459850832151485, 
    typeDescKeyPrev_taken.hit = 0.318164973754328, fenwickFor = 13.726685831327, 
    goalDifferential = 1.64550956996529, shooterSecondsElapsedInShift = 18.0840546271316, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.252860472007036, 
    typeDescKeyPrev_wide.missed.shot.against = 0.153870048158856, 
    fenwickDifferential = 9.26362662352026, goalsFor = 1.37955533874177, 
    maxSecondsElapsedSinceLastShiftAgainst = 123.470202092802, 
    secondsElapsedInSequence = 62.041377883281, shooterAge = 4.1775147554224, 
    goalieAge = 3.73765834905587, typeDescKeyPrev_slap.shot.on.goal.for = 0.10836085107495, 
    typeDescKeyPrev_post.missed.shot.against = 0.0436489809144454, 
    minSecondsElapsedInShiftAgainst = 15.6667165744879, dSecondsElapsedInSequence = 14.6259743066056, 
    goalieWeight = 15.6274708019463, typeDescKeyPrev_blocked.shot.against = 0.195570725067292, 
    dYCoordNormPerSecond = 11.3885682057416, isRebound_yes = 0.303204986242277, 
    typeDescKeyPrev_blocked.shot.for = 0.295115267307653, shooterHandCode_R = 0.48418995601075, 
    shooterWeight = 15.4987256298557, secondsElapsedInPeriod = 342.168747397364, 
    typeDescKeyPrev_other.shot.on.goal.against = 0.0228941441612783, 
    corsiDifferential = 12.3909219329857, periodNumber = 0.815371370627598, 
    minSecondsElapsedSinceLastShiftAgainst = 96.4462724200546, 
    shooterHeight = 2.1603020920171, minSecondsElapsedSinceLastShiftFor = 93.5658440505107, 
    typeDescKeyPrev_other.missed.shot.against = 0.0380170431975969, 
    typeDescKeyPrev_wrist.shot.on.goal.against = 0.148968564002121, 
    goalieHandCode_R = 0.223903049594715, typeDescKeyPrev_giveaway.for = 0.166327886875221, 
    shooterPositionCode_R = 0.379286765052536, fenwickAgainst = 13.4425846646873, 
    typeDescKeyPrev_snap.shot.on.goal.against = 0.0875204372611295, 
    avgSecondsElapsedSinceLastShiftFor = 88.3246457971612, dYCoordNorm = 31.6148668192733, 
    typeDescKeyPrev_backhand.shot.on.goal.for = 0.102976610455359, 
    typeDescKeyPrev_other.shot.on.goal.for = 0.0535307913408025, 
    isHome_yes = 0.499934508575746, shooterPositionCode_L = 0.401367965512214, 
    typeDescKeyPrev_high.missed.shot.for = 0.0647960976643637, 
    secondsElapsedInGame = 1027.24828127115, goalsAgainst = 1.44212169204754, 
    isRush_yes = 0.263597756591535, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.0547822210060903, 
    typeDescKeyPrev_high.missed.shot.against = 0.0436489809144611, 
    maxSecondsElapsedInShiftAgainst = 24.3221059695226, corsiAgainst = 18.436756204371, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0.145324078561471, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0.0267943074764607, 
    isBehindNet_yes = 0.143586332580056, typeDescKeyPrev_wide.missed.shot.for = 0.252556065033873, 
    isPlayoff_yes = 0.244668585309574, shotsFor = 9.46069044666863, 
    maxSecondsElapsedInShiftFor = 20.3630538539914, corsiFor = 18.8919707756306, 
    dXCoordNorm = 63.6463974344396, shotDifferential = 6.72140058907313, 
    typeDescKeyPrev_deflected.shot.on.goal.for = 0.0483595733396074, 
    yCoordNorm = 20.9149926069331, shotType_other = 0.129414068578287, 
    isOvertime_yes = 0.0563763092997843, shotsAgainst = 9.32646008471507, 
    goalieHeight = 1.61614230921098, typeDescKeyPrev_slap.shot.on.goal.against = 0.0659098719237357, 
    maxSecondsElapsedSinceLastShiftFor = 121.777046220713, avgSecondsElapsedSinceLastShiftAgainst = 90.4660313322339, 
    typeDescKeyPrev_other.missed.shot.for = 0.0533844855029514
    )), ev = list(coefficients = c(`(Intercept)` = -2.25155785904054, 
distance = -0.347012229429611, xCoordNorm = 0.229999108551498, 
angle = -0.174994363438993, shotType_snap = 0.145936085503686, 
avgSecondsElapsedInShiftFor = 0.0898990756364799, minSecondsElapsedInShiftFor = 0.0720914431918606, 
dSecondsElapsedInSequence = -0.0696285655862927, dAngle = -0.0678013689318382, 
typeDescKeyPrev_blocked.shot.for = -0.0666122449768758, dYCoordNorm = -0.0638430600428616, 
typeDescKeyPrev_blocked.shot.against = 0.0636371112323108, skaterCountAgainst = -0.0528435860439824, 
skaterCountFor = -0.0527855486675944, maxSecondsElapsedSinceLastShiftAgainst = -0.0495230428049249, 
avgSecondsElapsedInShiftAgainst = 0.0493007292224438, typeDescKeyPrev_giveaway.for = -0.0492839712471866, 
shooterPositionCode_D = -0.049031024183541, goalieHeight = -0.0473969637256579, 
typeDescKeyPrev_snap.shot.on.goal.for = 0.0464459550897196, isOvertime_yes = 0.0452560520475805, 
shotType_other = 0.0410296776203072, typeDescKeyPrev_wide.missed.shot.for = 0.0397997077971771, 
shotType_slap = 0.0397204841984181, shooterAge = 0.038703571493319, 
yCoordNorm = 0.0378832804932377, minSecondsElapsedInShiftAgainst = -0.0377906726576768, 
typeDescKeyPrev_other.missed.shot.for = 0.0365039439199717, goalDifferential = 0.0357258666252162, 
shooterHeight = -0.0350959752911113, typeDescKeyPrev_slap.shot.on.goal.for = -0.0332647259629092, 
periodNumber = 0.0327434254012294, typeDescKeyPrev_post.missed.shot.against = 0.0318868242011377, 
secondsElapsedInPeriod = -0.0316863631429213, typeDescKeyPrev_other.shot.on.goal.against = -0.0312787235006244, 
goalieAge = -0.0311850034308387, isPlayoff_yes = 0.0311549466779367, 
shooterWeight = 0.0310840996328743, typeDescKeyPrev_other.missed.shot.against = -0.0303716771271543, 
typeDescKeyPrev_taken.hit = -0.029281756109804, secondsElapsedInGame = 0.0289393307626774, 
secondsElapsedInSequence = 0.0288725476196743, shooterSecondsElapsedInShift = -0.028651118078253, 
typeDescKeyPrev_deflected.shot.on.goal.against = 0.028558684185321, 
shooterHandCode_R = -0.028056280811865, isRebound_yes = 0.0274926666377523, 
typeDescKeyPrev_backhand.shot.on.goal.for = 0.0273577300523961, 
typeDescKeyPrev_wrist.shot.on.goal.for = -0.0262563696596323, 
dAnglePerSecond = 0.0231147942223715, shooterPositionCode_R = -0.0213845904825497, 
typeDescKeyPrev_high.missed.shot.against = 0.0208008275446357, 
typeDescKeyPrev_high.missed.shot.for = -0.0206650575069266, manDifferential = 0.0202193700845522, 
isEmptyNetFor_yes = -0.0202141432077523, strengthState_penalty.kill = -0.0202088288953459, 
crossedRoyalRoad_yes = 0.0192980466396642, goalsAgainst = -0.0186230791850835, 
shotDifferential = -0.0185359608649035, zoneCode_N = 0.0184457101547008, 
dDistance = -0.0176378458648437, typeDescKeyPrev_tip.in.shot.on.goal.against = -0.0160982096257462, 
avgSecondsElapsedSinceLastShiftFor = -0.0157103199246219, goalieWeight = -0.0152153593170259, 
goalsFor = 0.0150770515532062, typeDescKeyPrev_slap.shot.on.goal.against = 0.0150504779729718, 
isRush_yes = 0.014948055888642, shooterSecondsElapsedSinceLastShift = 0.0141531119762345, 
zoneCode_O = -0.0138612803071167, typeDescKeyPrev_wrist.shot.on.goal.against = 0.0138083614985251, 
typeDescKeyPrev_lost.faceoff = -0.013563344406401, dYCoordNormPerSecond = 0.0130970192951924, 
maxSecondsElapsedSinceLastShiftFor = -0.0123017284800889, typeDescKeyPrev_takeaway.for = 0.012009113841486, 
dDistancePerSecond = 0.0116636565547126, typeDescKeyPrev_takeaway.against = -0.0111772776346153, 
avgSecondsElapsedSinceLastShiftAgainst = 0.0111355704578853, 
minSecondsElapsedSinceLastShiftAgainst = 0.010842317888819, shooterPositionCode_L = 0.0106325588754725, 
typeDescKeyPrev_other.shot.on.goal.for = -0.0105668226674142, 
maxSecondsElapsedInShiftAgainst = 0.0101818505831751, shotType_tip.in = -0.00988285488611214, 
shotType_wrist = -0.00936561921955064, maxSecondsElapsedInShiftFor = 0.0090398975683844, 
shotsFor = -0.00873401704679281, typeDescKeyPrev_giveaway.against = 0.00861189558962555, 
typeDescKeyPrev_deflected.shot.on.goal.for = -0.00854385624926819, 
isBehindNet_yes = 0.00844849209793505, typeDescKeyPrev_post.missed.shot.for = 0.00842037504275693, 
shotType_deflected = -0.0078876736972527, typeDescKeyPrev_wide.missed.shot.against = -0.00785643985547647, 
typeDescKeyPrev_given.hit = -0.00751605764518238, goalieHandCode_R = 0.00697824891204184, 
shotsAgainst = 0.00619507992616124, corsiAgainst = 0.00613275107026916, 
corsiDifferential = -0.0057903029006726, typeDescKeyPrev_snap.shot.on.goal.against = -0.0054970538655398, 
typeDescKeyPrev_won.faceoff = 0.00404869186959835, fenwickDifferential = 0.00291784666289579, 
dXCoordNorm = -0.00197008239839765, dXCoordNormPerSecond = 0.00192187690157279, 
fenwickFor = 0.00191725567402822, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.00167054654121044, 
isHome_yes = 0.00149003273855938, corsiFor = 0.00121776686340797, 
fenwickAgainst = -0.000407966324717708, minSecondsElapsedSinceLastShiftFor = 0.000317195784304764
), raw_terms = c("distance", "xCoordNorm", "angle", "avgSecondsElapsedInShiftFor", 
"minSecondsElapsedInShiftFor", "dSecondsElapsedInSequence", "dAngle", 
"dYCoordNorm", "skaterCountAgainst", "skaterCountFor", "maxSecondsElapsedSinceLastShiftAgainst", 
"avgSecondsElapsedInShiftAgainst", "goalieHeight", "shooterAge", 
"yCoordNorm", "minSecondsElapsedInShiftAgainst", "goalDifferential", 
"shooterHeight", "periodNumber", "secondsElapsedInPeriod", "goalieAge", 
"shooterWeight", "secondsElapsedInGame", "secondsElapsedInSequence", 
"shooterSecondsElapsedInShift", "dAnglePerSecond", "manDifferential", 
"goalsAgainst", "shotDifferential", "dDistance", "avgSecondsElapsedSinceLastShiftFor", 
"goalieWeight", "goalsFor", "shooterSecondsElapsedSinceLastShift", 
"dYCoordNormPerSecond", "maxSecondsElapsedSinceLastShiftFor", 
"dDistancePerSecond", "avgSecondsElapsedSinceLastShiftAgainst", 
"minSecondsElapsedSinceLastShiftAgainst", "maxSecondsElapsedInShiftAgainst", 
"maxSecondsElapsedInShiftFor", "shotsFor", "shotsAgainst", "corsiAgainst", 
"corsiDifferential", "fenwickDifferential", "dXCoordNorm", "dXCoordNormPerSecond", 
"fenwickFor", "corsiFor", "fenwickAgainst", "minSecondsElapsedSinceLastShiftFor"
), active_dummy_terms = c("shotType_snap", "typeDescKeyPrev_blocked.shot.for", 
"typeDescKeyPrev_blocked.shot.against", "typeDescKeyPrev_giveaway.for", 
"shooterPositionCode_D", "typeDescKeyPrev_snap.shot.on.goal.for", 
"isOvertime_yes", "shotType_other", "typeDescKeyPrev_wide.missed.shot.for", 
"shotType_slap", "typeDescKeyPrev_other.missed.shot.for", "typeDescKeyPrev_slap.shot.on.goal.for", 
"typeDescKeyPrev_post.missed.shot.against", "typeDescKeyPrev_other.shot.on.goal.against", 
"isPlayoff_yes", "typeDescKeyPrev_other.missed.shot.against", 
"typeDescKeyPrev_taken.hit", "typeDescKeyPrev_deflected.shot.on.goal.against", 
"shooterHandCode_R", "isRebound_yes", "typeDescKeyPrev_backhand.shot.on.goal.for", 
"typeDescKeyPrev_wrist.shot.on.goal.for", "shooterPositionCode_R", 
"typeDescKeyPrev_high.missed.shot.against", "typeDescKeyPrev_high.missed.shot.for", 
"isEmptyNetFor_yes", "strengthState_penalty.kill", "crossedRoyalRoad_yes", 
"zoneCode_N", "typeDescKeyPrev_tip.in.shot.on.goal.against", 
"typeDescKeyPrev_slap.shot.on.goal.against", "isRush_yes", "zoneCode_O", 
"typeDescKeyPrev_wrist.shot.on.goal.against", "typeDescKeyPrev_lost.faceoff", 
"typeDescKeyPrev_takeaway.for", "typeDescKeyPrev_takeaway.against", 
"shooterPositionCode_L", "typeDescKeyPrev_other.shot.on.goal.for", 
"shotType_tip.in", "shotType_wrist", "typeDescKeyPrev_giveaway.against", 
"typeDescKeyPrev_deflected.shot.on.goal.for", "isBehindNet_yes", 
"typeDescKeyPrev_post.missed.shot.for", "shotType_deflected", 
"typeDescKeyPrev_wide.missed.shot.against", "typeDescKeyPrev_given.hit", 
"goalieHandCode_R", "typeDescKeyPrev_snap.shot.on.goal.against", 
"typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_tip.in.shot.on.goal.for", 
"isHome_yes"), active_dummy_by_var = list(crossedRoyalRoad = "crossedRoyalRoad_yes", 
    goalieHandCode = "goalieHandCode_R", isBehindNet = "isBehindNet_yes", 
    isEmptyNetFor = "isEmptyNetFor_yes", isHome = "isHome_yes", 
    isOvertime = "isOvertime_yes", isPlayoff = "isPlayoff_yes", 
    isRebound = "isRebound_yes", isRush = "isRush_yes", shooterHandCode = "shooterHandCode_R", 
    shooterPositionCode = c("shooterPositionCode_D", "shooterPositionCode_R", 
    "shooterPositionCode_L"), shotType = c("shotType_snap", "shotType_other", 
    "shotType_slap", "shotType_tip.in", "shotType_wrist", "shotType_deflected"
    ), strengthState = "strengthState_penalty.kill", typeDescKeyPrev = c("typeDescKeyPrev_blocked.shot.for", 
    "typeDescKeyPrev_blocked.shot.against", "typeDescKeyPrev_giveaway.for", 
    "typeDescKeyPrev_snap.shot.on.goal.for", "typeDescKeyPrev_wide.missed.shot.for", 
    "typeDescKeyPrev_other.missed.shot.for", "typeDescKeyPrev_slap.shot.on.goal.for", 
    "typeDescKeyPrev_post.missed.shot.against", "typeDescKeyPrev_other.shot.on.goal.against", 
    "typeDescKeyPrev_other.missed.shot.against", "typeDescKeyPrev_taken.hit", 
    "typeDescKeyPrev_deflected.shot.on.goal.against", "typeDescKeyPrev_backhand.shot.on.goal.for", 
    "typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_high.missed.shot.against", 
    "typeDescKeyPrev_high.missed.shot.for", "typeDescKeyPrev_tip.in.shot.on.goal.against", 
    "typeDescKeyPrev_slap.shot.on.goal.against", "typeDescKeyPrev_wrist.shot.on.goal.against", 
    "typeDescKeyPrev_lost.faceoff", "typeDescKeyPrev_takeaway.for", 
    "typeDescKeyPrev_takeaway.against", "typeDescKeyPrev_other.shot.on.goal.for", 
    "typeDescKeyPrev_giveaway.against", "typeDescKeyPrev_deflected.shot.on.goal.for", 
    "typeDescKeyPrev_post.missed.shot.for", "typeDescKeyPrev_wide.missed.shot.against", 
    "typeDescKeyPrev_given.hit", "typeDescKeyPrev_snap.shot.on.goal.against", 
    "typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_tip.in.shot.on.goal.for"
    ), zoneCode = c("zoneCode_N", "zoneCode_O")), dummy_term_to_var = c(isPlayoff_yes = "isPlayoff", 
isPlayoff_unknown = "isPlayoff", isPlayoff_new = "isPlayoff", 
isHome_yes = "isHome", isHome_unknown = "isHome", isHome_new = "isHome", 
isOvertime_yes = "isOvertime", isOvertime_unknown = "isOvertime", 
isOvertime_new = "isOvertime", zoneCode_N = "zoneCode", zoneCode_O = "zoneCode", 
zoneCode_unknown = "zoneCode", zoneCode_new = "zoneCode", isBehindNet_yes = "isBehindNet", 
isBehindNet_unknown = "isBehindNet", isBehindNet_new = "isBehindNet", 
crossedRoyalRoad_yes = "crossedRoyalRoad", crossedRoyalRoad_unknown = "crossedRoyalRoad", 
crossedRoyalRoad_new = "crossedRoyalRoad", typeDescKeyPrev_backhand.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_blocked.shot.against = "typeDescKeyPrev", typeDescKeyPrev_blocked.shot.for = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_giveaway.against = "typeDescKeyPrev", typeDescKeyPrev_giveaway.for = "typeDescKeyPrev", 
typeDescKeyPrev_given.hit = "typeDescKeyPrev", typeDescKeyPrev_high.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_high.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_lost.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_other.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_post.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_post.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_slap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_slap.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_snap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_snap.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_takeaway.against = "typeDescKeyPrev", 
typeDescKeyPrev_takeaway.for = "typeDescKeyPrev", typeDescKeyPrev_taken.hit = "typeDescKeyPrev", 
typeDescKeyPrev_tip.in.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_tip.in.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_won.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_unknown = "typeDescKeyPrev", 
typeDescKeyPrev_new = "typeDescKeyPrev", shotType_deflected = "shotType", 
shotType_other = "shotType", shotType_slap = "shotType", shotType_snap = "shotType", 
shotType_tip.in = "shotType", shotType_wrist = "shotType", shotType_unknown = "shotType", 
shotType_new = "shotType", isRebound_yes = "isRebound", isRebound_unknown = "isRebound", 
isRebound_new = "isRebound", isRush_yes = "isRush", isRush_unknown = "isRush", 
isRush_new = "isRush", shooterHandCode_R = "shooterHandCode", 
shooterHandCode_unknown = "shooterHandCode", shooterHandCode_new = "shooterHandCode", 
shooterPositionCode_D = "shooterPositionCode", shooterPositionCode_L = "shooterPositionCode", 
shooterPositionCode_R = "shooterPositionCode", shooterPositionCode_unknown = "shooterPositionCode", 
shooterPositionCode_new = "shooterPositionCode", goalieHandCode_R = "goalieHandCode", 
goalieHandCode_unknown = "goalieHandCode", goalieHandCode_new = "goalieHandCode", 
isEmptyNetFor_yes = "isEmptyNetFor", isEmptyNetFor_unknown = "isEmptyNetFor", 
isEmptyNetFor_new = "isEmptyNetFor", strengthState_penalty.kill = "strengthState", 
strengthState_unknown = "strengthState", strengthState_new = "strengthState"
), dummy_term_to_level = c(isPlayoff_yes = "yes", isPlayoff_unknown = "unknown", 
isPlayoff_new = "new", isHome_yes = "yes", isHome_unknown = "unknown", 
isHome_new = "new", isOvertime_yes = "yes", isOvertime_unknown = "unknown", 
isOvertime_new = "new", zoneCode_N = "N", zoneCode_O = "O", zoneCode_unknown = "unknown", 
zoneCode_new = "new", isBehindNet_yes = "yes", isBehindNet_unknown = "unknown", 
isBehindNet_new = "new", crossedRoyalRoad_yes = "yes", crossedRoyalRoad_unknown = "unknown", 
crossedRoyalRoad_new = "new", typeDescKeyPrev_backhand.shot.on.goal.for = "backhand-shot-on-goal-for", 
typeDescKeyPrev_blocked.shot.against = "blocked-shot-against", 
typeDescKeyPrev_blocked.shot.for = "blocked-shot-for", typeDescKeyPrev_deflected.shot.on.goal.against = "deflected-shot-on-goal-against", 
typeDescKeyPrev_deflected.shot.on.goal.for = "deflected-shot-on-goal-for", 
typeDescKeyPrev_giveaway.against = "giveaway-against", typeDescKeyPrev_giveaway.for = "giveaway-for", 
typeDescKeyPrev_given.hit = "given-hit", typeDescKeyPrev_high.missed.shot.against = "high-missed-shot-against", 
typeDescKeyPrev_high.missed.shot.for = "high-missed-shot-for", 
typeDescKeyPrev_lost.faceoff = "lost-faceoff", typeDescKeyPrev_other.missed.shot.against = "other-missed-shot-against", 
typeDescKeyPrev_other.missed.shot.for = "other-missed-shot-for", 
typeDescKeyPrev_other.shot.on.goal.against = "other-shot-on-goal-against", 
typeDescKeyPrev_other.shot.on.goal.for = "other-shot-on-goal-for", 
typeDescKeyPrev_post.missed.shot.against = "post-missed-shot-against", 
typeDescKeyPrev_post.missed.shot.for = "post-missed-shot-for", 
typeDescKeyPrev_slap.shot.on.goal.against = "slap-shot-on-goal-against", 
typeDescKeyPrev_slap.shot.on.goal.for = "slap-shot-on-goal-for", 
typeDescKeyPrev_snap.shot.on.goal.against = "snap-shot-on-goal-against", 
typeDescKeyPrev_snap.shot.on.goal.for = "snap-shot-on-goal-for", 
typeDescKeyPrev_takeaway.against = "takeaway-against", typeDescKeyPrev_takeaway.for = "takeaway-for", 
typeDescKeyPrev_taken.hit = "taken-hit", typeDescKeyPrev_tip.in.shot.on.goal.against = "tip-in-shot-on-goal-against", 
typeDescKeyPrev_tip.in.shot.on.goal.for = "tip-in-shot-on-goal-for", 
typeDescKeyPrev_wide.missed.shot.against = "wide-missed-shot-against", 
typeDescKeyPrev_wide.missed.shot.for = "wide-missed-shot-for", 
typeDescKeyPrev_won.faceoff = "won-faceoff", typeDescKeyPrev_wrist.shot.on.goal.against = "wrist-shot-on-goal-against", 
typeDescKeyPrev_wrist.shot.on.goal.for = "wrist-shot-on-goal-for", 
typeDescKeyPrev_unknown = "unknown", typeDescKeyPrev_new = "new", 
shotType_deflected = "deflected", shotType_other = "other", shotType_slap = "slap", 
shotType_snap = "snap", shotType_tip.in = "tip-in", shotType_wrist = "wrist", 
shotType_unknown = "unknown", shotType_new = "new", isRebound_yes = "yes", 
isRebound_unknown = "unknown", isRebound_new = "new", isRush_yes = "yes", 
isRush_unknown = "unknown", isRush_new = "new", shooterHandCode_R = "R", 
shooterHandCode_unknown = "unknown", shooterHandCode_new = "new", 
shooterPositionCode_D = "D", shooterPositionCode_L = "L", shooterPositionCode_R = "R", 
shooterPositionCode_unknown = "unknown", shooterPositionCode_new = "new", 
goalieHandCode_R = "R", goalieHandCode_unknown = "unknown", goalieHandCode_new = "new", 
isEmptyNetFor_yes = "yes", isEmptyNetFor_unknown = "unknown", 
isEmptyNetFor_new = "new", strengthState_penalty.kill = "penalty-kill", 
strengthState_unknown = "unknown", strengthState_new = "new"), 
    dummy_map = list(crossedRoyalRoad = c(yes = "crossedRoyalRoad_yes", 
    unknown = "crossedRoyalRoad_unknown", new = "crossedRoyalRoad_new"
    ), goalieHandCode = c(R = "goalieHandCode_R", unknown = "goalieHandCode_unknown", 
    new = "goalieHandCode_new"), isBehindNet = c(yes = "isBehindNet_yes", 
    unknown = "isBehindNet_unknown", new = "isBehindNet_new"), 
        isEmptyNetFor = c(yes = "isEmptyNetFor_yes", unknown = "isEmptyNetFor_unknown", 
        new = "isEmptyNetFor_new"), isHome = c(yes = "isHome_yes", 
        unknown = "isHome_unknown", new = "isHome_new"), isOvertime = c(yes = "isOvertime_yes", 
        unknown = "isOvertime_unknown", new = "isOvertime_new"
        ), isPlayoff = c(yes = "isPlayoff_yes", unknown = "isPlayoff_unknown", 
        new = "isPlayoff_new"), isRebound = c(yes = "isRebound_yes", 
        unknown = "isRebound_unknown", new = "isRebound_new"), 
        isRush = c(yes = "isRush_yes", unknown = "isRush_unknown", 
        new = "isRush_new"), shooterHandCode = c(R = "shooterHandCode_R", 
        unknown = "shooterHandCode_unknown", new = "shooterHandCode_new"
        ), shooterPositionCode = c(D = "shooterPositionCode_D", 
        L = "shooterPositionCode_L", R = "shooterPositionCode_R", 
        unknown = "shooterPositionCode_unknown", new = "shooterPositionCode_new"
        ), shotType = c(deflected = "shotType_deflected", other = "shotType_other", 
        slap = "shotType_slap", snap = "shotType_snap", `tip-in` = "shotType_tip.in", 
        wrist = "shotType_wrist", unknown = "shotType_unknown", 
        new = "shotType_new"), strengthState = c(`penalty-kill` = "strengthState_penalty.kill", 
        unknown = "strengthState_unknown", new = "strengthState_new"
        ), typeDescKeyPrev = c(`backhand-shot-on-goal-for` = "typeDescKeyPrev_backhand.shot.on.goal.for", 
        `blocked-shot-against` = "typeDescKeyPrev_blocked.shot.against", 
        `blocked-shot-for` = "typeDescKeyPrev_blocked.shot.for", 
        `deflected-shot-on-goal-against` = "typeDescKeyPrev_deflected.shot.on.goal.against", 
        `deflected-shot-on-goal-for` = "typeDescKeyPrev_deflected.shot.on.goal.for", 
        `giveaway-against` = "typeDescKeyPrev_giveaway.against", 
        `giveaway-for` = "typeDescKeyPrev_giveaway.for", `given-hit` = "typeDescKeyPrev_given.hit", 
        `high-missed-shot-against` = "typeDescKeyPrev_high.missed.shot.against", 
        `high-missed-shot-for` = "typeDescKeyPrev_high.missed.shot.for", 
        `lost-faceoff` = "typeDescKeyPrev_lost.faceoff", `other-missed-shot-against` = "typeDescKeyPrev_other.missed.shot.against", 
        `other-missed-shot-for` = "typeDescKeyPrev_other.missed.shot.for", 
        `other-shot-on-goal-against` = "typeDescKeyPrev_other.shot.on.goal.against", 
        `other-shot-on-goal-for` = "typeDescKeyPrev_other.shot.on.goal.for", 
        `post-missed-shot-against` = "typeDescKeyPrev_post.missed.shot.against", 
        `post-missed-shot-for` = "typeDescKeyPrev_post.missed.shot.for", 
        `slap-shot-on-goal-against` = "typeDescKeyPrev_slap.shot.on.goal.against", 
        `slap-shot-on-goal-for` = "typeDescKeyPrev_slap.shot.on.goal.for", 
        `snap-shot-on-goal-against` = "typeDescKeyPrev_snap.shot.on.goal.against", 
        `snap-shot-on-goal-for` = "typeDescKeyPrev_snap.shot.on.goal.for", 
        `takeaway-against` = "typeDescKeyPrev_takeaway.against", 
        `takeaway-for` = "typeDescKeyPrev_takeaway.for", `taken-hit` = "typeDescKeyPrev_taken.hit", 
        `tip-in-shot-on-goal-against` = "typeDescKeyPrev_tip.in.shot.on.goal.against", 
        `tip-in-shot-on-goal-for` = "typeDescKeyPrev_tip.in.shot.on.goal.for", 
        `wide-missed-shot-against` = "typeDescKeyPrev_wide.missed.shot.against", 
        `wide-missed-shot-for` = "typeDescKeyPrev_wide.missed.shot.for", 
        `won-faceoff` = "typeDescKeyPrev_won.faceoff", `wrist-shot-on-goal-against` = "typeDescKeyPrev_wrist.shot.on.goal.against", 
        `wrist-shot-on-goal-for` = "typeDescKeyPrev_wrist.shot.on.goal.for", 
        unknown = "typeDescKeyPrev_unknown", new = "typeDescKeyPrev_new"
        ), zoneCode = c(N = "zoneCode_N", O = "zoneCode_O", unknown = "zoneCode_unknown", 
        new = "zoneCode_new")), known_levels = list(isPlayoff = c("no", 
    "yes"), isHome = c("no", "yes"), isOvertime = c("no", "yes"
    ), zoneCode = c("D", "N", "O"), isBehindNet = c("no", "yes"
    ), crossedRoyalRoad = c("no", "yes"), typeDescKeyPrev = c("backhand-shot-on-goal-against", 
    "backhand-shot-on-goal-for", "blocked-shot-against", "blocked-shot-for", 
    "deflected-shot-on-goal-against", "deflected-shot-on-goal-for", 
    "giveaway-against", "giveaway-for", "given-hit", "high-missed-shot-against", 
    "high-missed-shot-for", "lost-faceoff", "other-missed-shot-against", 
    "other-missed-shot-for", "other-shot-on-goal-against", "other-shot-on-goal-for", 
    "post-missed-shot-against", "post-missed-shot-for", "slap-shot-on-goal-against", 
    "slap-shot-on-goal-for", "snap-shot-on-goal-against", "snap-shot-on-goal-for", 
    "takeaway-against", "takeaway-for", "taken-hit", "tip-in-shot-on-goal-against", 
    "tip-in-shot-on-goal-for", "wide-missed-shot-against", "wide-missed-shot-for", 
    "won-faceoff", "wrist-shot-on-goal-against", "wrist-shot-on-goal-for"
    ), shotType = c("backhand", "deflected", "other", "slap", 
    "snap", "tip-in", "wrist"), isRebound = c("no", "yes"), isRush = c("no", 
    "yes"), shooterHandCode = c("L", "R"), shooterPositionCode = c("C", 
    "D", "L", "R"), goalieHandCode = c("L", "R"), isEmptyNetFor = c("no", 
    "yes"), strengthState = c("even-strength", "penalty-kill"
    )), logical_vars = c("crossedRoyalRoad", "isBehindNet", "isEmptyNetFor", 
    "isHome", "isOvertime", "isPlayoff", "isRebound", "isRush"
    ), has_unknown = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
    isBehindNet = TRUE, isEmptyNetFor = TRUE, isHome = TRUE, 
    isOvertime = TRUE, isPlayoff = TRUE, isRebound = TRUE, isRush = TRUE, 
    shooterHandCode = TRUE, shooterPositionCode = TRUE, shotType = TRUE, 
    strengthState = TRUE, typeDescKeyPrev = TRUE, zoneCode = TRUE
    ), has_new = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
    isBehindNet = TRUE, isEmptyNetFor = TRUE, isHome = TRUE, 
    isOvertime = TRUE, isPlayoff = TRUE, isRebound = TRUE, isRush = TRUE, 
    shooterHandCode = TRUE, shooterPositionCode = TRUE, shotType = TRUE, 
    strengthState = TRUE, typeDescKeyPrev = TRUE, zoneCode = TRUE
    ), medians = c(distance = 27.6586333718787, xCoordNorm = 67, 
    angle = 29.9816393688493, shotType_snap = 0, avgSecondsElapsedInShiftFor = 27.6666666666667, 
    minSecondsElapsedInShiftFor = 15, dSecondsElapsedInSequence = 16, 
    dAngle = 7.12001497155098, typeDescKeyPrev_blocked.shot.for = 0, 
    dYCoordNorm = 0, typeDescKeyPrev_blocked.shot.against = 0, 
    skaterCountAgainst = 4, skaterCountFor = 4, maxSecondsElapsedSinceLastShiftAgainst = 249, 
    avgSecondsElapsedInShiftAgainst = 32, typeDescKeyPrev_giveaway.for = 0, 
    shooterPositionCode_D = 0, goalieHeight = 75, typeDescKeyPrev_snap.shot.on.goal.for = 0, 
    isOvertime_yes = 0, shotType_other = 0, typeDescKeyPrev_wide.missed.shot.for = 0, 
    shotType_slap = 0, shooterAge = 27, yCoordNorm = 0, minSecondsElapsedInShiftAgainst = 20, 
    typeDescKeyPrev_other.missed.shot.for = 0, goalDifferential = 0, 
    shooterHeight = 73, typeDescKeyPrev_slap.shot.on.goal.for = 0, 
    periodNumber = 3, typeDescKeyPrev_post.missed.shot.against = 0, 
    secondsElapsedInPeriod = 251, typeDescKeyPrev_other.shot.on.goal.against = 0, 
    goalieAge = 28, isPlayoff_yes = 0, shooterWeight = 200, typeDescKeyPrev_other.missed.shot.against = 0, 
    typeDescKeyPrev_taken.hit = 0, secondsElapsedInGame = 3357, 
    secondsElapsedInSequence = 43, shooterSecondsElapsedInShift = 26, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0, shooterHandCode_R = 0, 
    isRebound_yes = 0, typeDescKeyPrev_backhand.shot.on.goal.for = 0, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0, dAnglePerSecond = 0.333179278037169, 
    shooterPositionCode_R = 0, typeDescKeyPrev_high.missed.shot.against = 0, 
    typeDescKeyPrev_high.missed.shot.for = 0, manDifferential = 0, 
    isEmptyNetFor_yes = 0, strengthState_penalty.kill = 0, crossedRoyalRoad_yes = 0, 
    goalsAgainst = 2, shotDifferential = 0, zoneCode_N = 0, dDistance = -16.3988213205148, 
    typeDescKeyPrev_tip.in.shot.on.goal.against = 0, avgSecondsElapsedSinceLastShiftFor = 161.5, 
    goalieWeight = 201, goalsFor = 2, typeDescKeyPrev_slap.shot.on.goal.against = 0, 
    isRush_yes = 0, shooterSecondsElapsedSinceLastShift = 153.5, 
    zoneCode_O = 1, typeDescKeyPrev_wrist.shot.on.goal.against = 0, 
    typeDescKeyPrev_lost.faceoff = 0, dYCoordNormPerSecond = 0, 
    maxSecondsElapsedSinceLastShiftFor = 236, typeDescKeyPrev_takeaway.for = 0, 
    dDistancePerSecond = -1.1423923064826, typeDescKeyPrev_takeaway.against = 0, 
    avgSecondsElapsedSinceLastShiftAgainst = 167.291666666667, 
    minSecondsElapsedSinceLastShiftAgainst = 100, shooterPositionCode_L = 0, 
    typeDescKeyPrev_other.shot.on.goal.for = 0, maxSecondsElapsedInShiftAgainst = 42, 
    shotType_tip.in = 0, shotType_wrist = 1, maxSecondsElapsedInShiftFor = 38, 
    shotsFor = 23, typeDescKeyPrev_giveaway.against = 0, typeDescKeyPrev_deflected.shot.on.goal.for = 0, 
    isBehindNet_yes = 0, typeDescKeyPrev_post.missed.shot.for = 0, 
    shotType_deflected = 0, typeDescKeyPrev_wide.missed.shot.against = 0, 
    typeDescKeyPrev_given.hit = 0, goalieHandCode_R = 0, shotsAgainst = 23, 
    corsiAgainst = 47, corsiDifferential = 1, typeDescKeyPrev_snap.shot.on.goal.against = 0, 
    typeDescKeyPrev_won.faceoff = 0, fenwickDifferential = 1, 
    dXCoordNorm = 14, dXCoordNormPerSecond = 1.0377358490566, 
    fenwickFor = 35, typeDescKeyPrev_tip.in.shot.on.goal.for = 0, 
    isHome_yes = 1, corsiFor = 49, fenwickAgainst = 34, minSecondsElapsedSinceLastShiftFor = 91
    ), normalize_means = c(distance = 27.8836425441254, xCoordNorm = 65.8612236628849, 
    angle = 31.8111673046011, shotType_snap = 0.204213938411669, 
    avgSecondsElapsedInShiftFor = 29.1636311453268, minSecondsElapsedInShiftFor = 17.996961102107, 
    dSecondsElapsedInSequence = 20.5970421393841, dAngle = 5.65760894239027, 
    typeDescKeyPrev_blocked.shot.for = 0.0569286871961102, dYCoordNorm = -0.298622366288493, 
    typeDescKeyPrev_blocked.shot.against = 0.0423419773095624, 
    skaterCountAgainst = 3.55388978930308, skaterCountFor = 3.55388978930308, 
    maxSecondsElapsedSinceLastShiftAgainst = 274.257293354943, 
    avgSecondsElapsedInShiftAgainst = 34.0889282820097, typeDescKeyPrev_giveaway.for = 0.016612641815235, 
    shooterPositionCode_D = 0.31969205834684, goalieHeight = 74.9143030794165, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0.0235008103727715, 
    isOvertime_yes = 0.475688816855754, shotType_other = 0.016612641815235, 
    typeDescKeyPrev_wide.missed.shot.for = 0.0601701782820097, 
    shotType_slap = 0.0682739059967585, shooterAge = 27.3869529983793, 
    yCoordNorm = 0.309967585089141, minSecondsElapsedInShiftAgainst = 23.2878849270665, 
    typeDescKeyPrev_other.missed.shot.for = 0.00141815235008104, 
    goalDifferential = -0.0415316045380875, shooterHeight = 73.115680713128, 
    typeDescKeyPrev_slap.shot.on.goal.for = 0.00607779578606159, 
    periodNumber = 2.98136142625608, typeDescKeyPrev_post.missed.shot.against = 0.00688816855753647, 
    secondsElapsedInPeriod = 401.321515397083, typeDescKeyPrev_other.shot.on.goal.against = 0.00121555915721232, 
    goalieAge = 28.4390194489465, isPlayoff_yes = 0.0425445705024311, 
    shooterWeight = 199.879862236629, typeDescKeyPrev_other.missed.shot.against = 0.0010129659643436, 
    typeDescKeyPrev_taken.hit = 0.0362641815235008, secondsElapsedInGame = 2778.95522690438, 
    secondsElapsedInSequence = 58.2838330632091, shooterSecondsElapsedInShift = 29.9902755267423, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0.000810372771474878, 
    shooterHandCode_R = 0.388371150729335, isRebound_yes = 0.0644246353322528, 
    typeDescKeyPrev_backhand.shot.on.goal.for = 0.0119529983792545, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.0769854132901134, 
    dAnglePerSecond = 0.504619403742005, shooterPositionCode_R = 0.17240680713128, 
    typeDescKeyPrev_high.missed.shot.against = 0.00486223662884927, 
    typeDescKeyPrev_high.missed.shot.for = 0.00486223662884927, 
    manDifferential = -0.00668557536466775, isEmptyNetFor_yes = 0.00668557536466775, 
    strengthState_penalty.kill = 0.00668557536466775, crossedRoyalRoad_yes = 0.413695299837925, 
    goalsAgainst = 2.06138573743922, shotDifferential = 0.461507293354943, 
    zoneCode_N = 0.00567260940032415, dDistance = -37.5734424587598, 
    typeDescKeyPrev_tip.in.shot.on.goal.against = 0.00324149108589951, 
    avgSecondsElapsedSinceLastShiftFor = 206.954186925986, goalieWeight = 201.355145867099, 
    goalsFor = 2.01985413290113, typeDescKeyPrev_slap.shot.on.goal.against = 0.00648298217179903, 
    isRush_yes = 0.0285656401944895, shooterSecondsElapsedSinceLastShift = 207.469205834684, 
    zoneCode_O = 0.991693679092382, typeDescKeyPrev_wrist.shot.on.goal.against = 0.0547001620745543, 
    typeDescKeyPrev_lost.faceoff = 0.103727714748784, dYCoordNormPerSecond = -0.13209881857716, 
    maxSecondsElapsedSinceLastShiftFor = 268.787884927066, typeDescKeyPrev_takeaway.for = 0.0478119935170178, 
    dDistancePerSecond = -4.10845918884632, typeDescKeyPrev_takeaway.against = 0.0170178282009724, 
    avgSecondsElapsedSinceLastShiftAgainst = 214.34705902215, 
    minSecondsElapsedSinceLastShiftAgainst = 162, shooterPositionCode_L = 0.153768233387358, 
    typeDescKeyPrev_other.shot.on.goal.for = 0.00182333873581848, 
    maxSecondsElapsedInShiftAgainst = 46.2674230145867, shotType_tip.in = 0.0492301458670989, 
    shotType_wrist = 0.566653160453809, maxSecondsElapsedInShiftFor = 41.7013776337115, 
    shotsFor = 22.3243517017828, typeDescKeyPrev_giveaway.against = 0.053484602917342, 
    typeDescKeyPrev_deflected.shot.on.goal.for = 0.00162074554294976, 
    isBehindNet_yes = 0.0143841166936791, typeDescKeyPrev_post.missed.shot.for = 0.010129659643436, 
    shotType_deflected = 0.0119529983792545, typeDescKeyPrev_wide.missed.shot.against = 0.038290113452188, 
    typeDescKeyPrev_given.hit = 0.0334278768233387, goalieHandCode_R = 0.0518638573743922, 
    shotsAgainst = 21.8628444084279, corsiAgainst = 44.9564424635332, 
    corsiDifferential = 1.11932739059968, typeDescKeyPrev_snap.shot.on.goal.against = 0.0196515397082658, 
    typeDescKeyPrev_won.faceoff = 0.243517017828201, fenwickDifferential = 0.731158833063209, 
    dXCoordNorm = 37.512358184765, dXCoordNormPerSecond = 3.97051237542973, 
    fenwickFor = 32.9850081037277, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.00506482982171799, 
    isHome_yes = 0.508914100486224, corsiFor = 46.0757698541329, 
    fenwickAgainst = 32.2538492706645, minSecondsElapsedSinceLastShiftFor = 151.900729335494
    ), normalize_sds = c(distance = 15.9300473972311, xCoordNorm = 15.967075662423, 
    angle = 21.5469685697741, shotType_snap = 0.403166883540809, 
    avgSecondsElapsedInShiftFor = 14.7399565542128, minSecondsElapsedInShiftFor = 14.3953650821835, 
    dSecondsElapsedInSequence = 18.0205979006272, dAngle = 32.7977598710415, 
    typeDescKeyPrev_blocked.shot.for = 0.231729779612907, dYCoordNorm = 24.6269834574023, 
    typeDescKeyPrev_blocked.shot.against = 0.201388557049044, 
    skaterCountAgainst = 0.51120481009207, skaterCountFor = 0.51120481009207, 
    maxSecondsElapsedSinceLastShiftAgainst = 134.627863568148, 
    avgSecondsElapsedInShiftAgainst = 17.4968959064147, typeDescKeyPrev_giveaway.for = 0.127827901157436, 
    shooterPositionCode_D = 0.466404456352566, goalieHeight = 1.5879963991926, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0.15150304432943, 
    isOvertime_yes = 0.499459212720901, shotType_other = 0.127827901157434, 
    typeDescKeyPrev_wide.missed.shot.for = 0.237825959135982, 
    shotType_slap = 0.252240896453758, shooterAge = 4.08250573106685, 
    yCoordNorm = 15.519871561703, minSecondsElapsedInShiftAgainst = 16.4535385003685, 
    typeDescKeyPrev_other.missed.shot.for = 0.037635464029222, 
    goalDifferential = 1.37442894256115, shooterHeight = 2.12463460415575, 
    typeDescKeyPrev_slap.shot.on.goal.for = 0.0777308192982554, 
    periodNumber = 1.11217679992388, typeDescKeyPrev_post.missed.shot.against = 0.082717034859428, 
    secondsElapsedInPeriod = 356.227677486903, typeDescKeyPrev_other.shot.on.goal.against = 0.0348472034408927, 
    goalieAge = 3.7276492276995, isPlayoff_yes = 0.201848418953271, 
    shooterWeight = 15.5092139055743, typeDescKeyPrev_other.missed.shot.against = 0.0318142250883039, 
    typeDescKeyPrev_taken.hit = 0.186965698844831, secondsElapsedInGame = 1123.07200665801, 
    secondsElapsedInSequence = 50.5981873719287, shooterSecondsElapsedInShift = 20.4840884364671, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0.0284583932021896, 
    shooterHandCode_R = 0.487429106174922, isRebound_yes = 0.245532717354864, 
    typeDescKeyPrev_backhand.shot.on.goal.for = 0.108685405389147, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.266595308189183, 
    dAnglePerSecond = 8.77034775526019, shooterPositionCode_R = 0.377771905238188, 
    typeDescKeyPrev_high.missed.shot.against = 0.0695670593666077, 
    typeDescKeyPrev_high.missed.shot.for = 0.0695670593666078, 
    manDifferential = 0.0814998412029298, isEmptyNetFor_yes = 0.0814998412029298, 
    strengthState_penalty.kill = 0.0814998412029298, crossedRoyalRoad_yes = 0.49254507201858, 
    goalsAgainst = 1.47222857085925, shotDifferential = 8.51832384996687, 
    zoneCode_N = 0.0751104110451705, dDistance = 55.9620289934552, 
    typeDescKeyPrev_tip.in.shot.on.goal.against = 0.0568475024025425, 
    avgSecondsElapsedSinceLastShiftFor = 113.798898663227, goalieWeight = 15.6479145526495, 
    goalsFor = 1.44090300791271, typeDescKeyPrev_slap.shot.on.goal.against = 0.0802636796542013, 
    isRush_yes = 0.166599121915498, shooterSecondsElapsedSinceLastShift = 137.618028053179, 
    zoneCode_O = 0.0907689104535364, typeDescKeyPrev_wrist.shot.on.goal.against = 0.227417088553474, 
    typeDescKeyPrev_lost.faceoff = 0.304937886295219, dYCoordNormPerSecond = 5.53933835742204, 
    maxSecondsElapsedSinceLastShiftFor = 134.041195648219, typeDescKeyPrev_takeaway.for = 0.213389858992813, 
    dDistancePerSecond = 13.3533983330833, typeDescKeyPrev_takeaway.against = 0.129350730322169, 
    avgSecondsElapsedSinceLastShiftAgainst = 116.279083696836, 
    minSecondsElapsedSinceLastShiftAgainst = 131.146455948308, 
    shooterPositionCode_L = 0.360762984906197, typeDescKeyPrev_other.shot.on.goal.for = 0.0426659462437079, 
    maxSecondsElapsedInShiftAgainst = 26.0848238402378, shotType_tip.in = 0.216370107021805, 
    shotType_wrist = 0.495587645660225, maxSecondsElapsedInShiftFor = 24.0877063013396, 
    shotsFor = 10.5823897696679, typeDescKeyPrev_giveaway.against = 0.225020573113164, 
    typeDescKeyPrev_deflected.shot.on.goal.for = 0.0402299218629799, 
    isBehindNet_yes = 0.119080169086438, typeDescKeyPrev_post.missed.shot.for = 0.100145301749902, 
    shotType_deflected = 0.108685405389145, typeDescKeyPrev_wide.missed.shot.against = 0.191915196020359, 
    typeDescKeyPrev_given.hit = 0.179769299600504, goalieHandCode_R = 0.221774574756755, 
    shotsAgainst = 10.2695203766714, corsiAgainst = 20.2983303705439, 
    corsiDifferential = 16.5137179545448, typeDescKeyPrev_snap.shot.on.goal.against = 0.138813761984803, 
    typeDescKeyPrev_won.faceoff = 0.429247956807713, fenwickDifferential = 12.1691576839932, 
    dXCoordNorm = 58.1427604865045, dXCoordNormPerSecond = 13.5488794054465, 
    fenwickFor = 15.2859995184002, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0709943549198634, 
    isHome_yes = 0.499971180441693, corsiFor = 21.1061660565772, 
    fenwickAgainst = 14.7693714045207, minSecondsElapsedSinceLastShiftFor = 128.445152162144
    )), pp = list(coefficients = c(`(Intercept)` = -2.40873593672909, 
distance = -0.54159493018628, angle = -0.215705911687938, shotType_snap = 0.19716907987083, 
xCoordNorm = 0.194974538661697, shotType_slap = 0.194897084952875, 
dDistancePerSecond = 0.114059176484837, skaterCountAgainst = -0.0805266678362151, 
isBehindNet_yes = -0.0749412656456376, fenwickDifferential = -0.0740644436487372, 
shotType_wrist = 0.0729746072456859, dAnglePerSecond = 0.0710273591130374, 
manDifferential = 0.0657540478232958, strengthState_power.play = -0.0651437396330941, 
shotType_tip.in = -0.0609302322226729, shooterSecondsElapsedInShift = 0.0604697258659946, 
dDistance = -0.0589946431545966, zoneCode_O = 0.0539114498200749, 
dXCoordNormPerSecond = -0.0519980536155228, crossedRoyalRoad_yes = 0.0476524205784241, 
shotDifferential = 0.046996053829332, corsiDifferential = 0.0466833242616486, 
typeDescKeyPrev_tip.in.shot.on.goal.against = -0.0443450271047899, 
avgSecondsElapsedInShiftAgainst = 0.0433982208658065, zoneCode_N = -0.0410018037204976, 
fenwickAgainst = 0.0405827524508978, dYCoordNormPerSecond = 0.0396071988167495, 
corsiAgainst = -0.0381725166417229, typeDescKeyPrev_other.missed.shot.against = -0.0370122592006922, 
shooterPositionCode_D = -0.0367724318156815, maxSecondsElapsedInShiftFor = -0.0367723514667785, 
typeDescKeyPrev_given.hit = -0.0359689768879517, goalsFor = 0.0338128533319229, 
secondsElapsedInPeriod = -0.0337400734552623, typeDescKeyPrev_backhand.shot.on.goal.for = -0.0330243547728242, 
goalieWeight = -0.0298369965872576, shotsAgainst = -0.0283553099429725, 
dXCoordNorm = -0.0274249327181236, shooterHandCode_R = 0.0265374425530595, 
avgSecondsElapsedSinceLastShiftAgainst = 0.025669819990372, typeDescKeyPrev_post.missed.shot.for = 0.0250219036651624, 
shotType_deflected = -0.0249885318857282, secondsElapsedInSequence = 0.0246819478374614, 
shooterSecondsElapsedSinceLastShift = -0.024266482507998, typeDescKeyPrev_wide.missed.shot.for = 0.024162116492808, 
typeDescKeyPrev_blocked.shot.for = -0.0241612316591937, goalieHandCode_R = -0.0240330305437289, 
maxSecondsElapsedSinceLastShiftAgainst = -0.0238816171555811, 
minSecondsElapsedInShiftAgainst = 0.023858561313903, typeDescKeyPrev_takeaway.against = 0.0225148172847808, 
goalDifferential = 0.0209467327052098, avgSecondsElapsedSinceLastShiftFor = 0.0208718465485786, 
typeDescKeyPrev_takeaway.for = 0.0208645584122001, typeDescKeyPrev_blocked.shot.against = -0.0200057949928402, 
typeDescKeyPrev_giveaway.for = -0.0198367624489711, typeDescKeyPrev_slap.shot.on.goal.against = -0.0191438988694104, 
typeDescKeyPrev_high.missed.shot.against = -0.0183676156559531, 
avgSecondsElapsedInShiftFor = 0.0178401601545399, periodNumber = 0.0176194837502181, 
typeDescKeyPrev_snap.shot.on.goal.against = -0.015866366866815, 
typeDescKeyPrev_slap.shot.on.goal.for = 0.0158540972697513, typeDescKeyPrev_deflected.shot.on.goal.against = 0.0156373458043986, 
shooterWeight = 0.0142853769189897, goalieAge = 0.0142374621400034, 
shooterPositionCode_L = -0.0140278286365682, dAngle = -0.0135530645839663, 
fenwickFor = -0.0134179545964687, yCoordNorm = 0.0127126548438979, 
shotType_other = 0.0123638354488715, isRush_yes = 0.0121347520256905, 
typeDescKeyPrev_high.missed.shot.for = 0.0121263670150074, skaterCountFor = -0.0120734981294251, 
typeDescKeyPrev_other.shot.on.goal.for = -0.0119337859318244, 
isPlayoff_yes = 0.0118545831958862, typeDescKeyPrev_giveaway.against = 0.0118150476882019, 
typeDescKeyPrev_wide.missed.shot.against = -0.0116482137834756, 
minSecondsElapsedInShiftFor = -0.0111270089676604, typeDescKeyPrev_post.missed.shot.against = -0.0104187024962581, 
isRebound_yes = -0.010148209868158, typeDescKeyPrev_other.missed.shot.for = 0.00982366700478954, 
typeDescKeyPrev_snap.shot.on.goal.for = 0.00956480712515684, 
maxSecondsElapsedInShiftAgainst = -0.00907648698578467, shooterPositionCode_R = -0.00907187449060681, 
isEmptyNetFor_yes = 0.00888728972455235, typeDescKeyPrev_other.shot.on.goal.against = 0.00837537319072141, 
isOvertime_yes = 0.00809371103567771, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0080170918474623, 
typeDescKeyPrev_won.faceoff = 0.00785244249448494, typeDescKeyPrev_wrist.shot.on.goal.against = 0.00781691479964121, 
shotsFor = 0.00761786303739305, isHome_yes = -0.00715136385551856, 
dSecondsElapsedInSequence = -0.00659622587702555, goalsAgainst = 0.00657488948434205, 
typeDescKeyPrev_taken.hit = 0.00622893212610536, shooterAge = 0.00622430309461466, 
maxSecondsElapsedSinceLastShiftFor = -0.00516577829045611, secondsElapsedInGame = 0.00460628671009295, 
corsiFor = -0.00438434350624629, goalieHeight = -0.00405974799713595, 
typeDescKeyPrev_deflected.shot.on.goal.for = -0.00356581981912205, 
minSecondsElapsedSinceLastShiftFor = -0.00274894577959475, dYCoordNorm = -0.00256418142314409, 
typeDescKeyPrev_wrist.shot.on.goal.for = 0.0020828784867278, 
shooterHeight = -0.00157456935417754, typeDescKeyPrev_lost.faceoff = -0.000759496272344574, 
minSecondsElapsedSinceLastShiftAgainst = -0.000364320801870089
), raw_terms = c("distance", "angle", "xCoordNorm", "dDistancePerSecond", 
"skaterCountAgainst", "fenwickDifferential", "dAnglePerSecond", 
"manDifferential", "shooterSecondsElapsedInShift", "dDistance", 
"dXCoordNormPerSecond", "shotDifferential", "corsiDifferential", 
"avgSecondsElapsedInShiftAgainst", "fenwickAgainst", "dYCoordNormPerSecond", 
"corsiAgainst", "maxSecondsElapsedInShiftFor", "goalsFor", "secondsElapsedInPeriod", 
"goalieWeight", "shotsAgainst", "dXCoordNorm", "avgSecondsElapsedSinceLastShiftAgainst", 
"secondsElapsedInSequence", "shooterSecondsElapsedSinceLastShift", 
"maxSecondsElapsedSinceLastShiftAgainst", "minSecondsElapsedInShiftAgainst", 
"goalDifferential", "avgSecondsElapsedSinceLastShiftFor", "avgSecondsElapsedInShiftFor", 
"periodNumber", "shooterWeight", "goalieAge", "dAngle", "fenwickFor", 
"yCoordNorm", "skaterCountFor", "minSecondsElapsedInShiftFor", 
"maxSecondsElapsedInShiftAgainst", "shotsFor", "dSecondsElapsedInSequence", 
"goalsAgainst", "shooterAge", "maxSecondsElapsedSinceLastShiftFor", 
"secondsElapsedInGame", "corsiFor", "goalieHeight", "minSecondsElapsedSinceLastShiftFor", 
"dYCoordNorm", "shooterHeight", "minSecondsElapsedSinceLastShiftAgainst"
), active_dummy_terms = c("shotType_snap", "shotType_slap", "isBehindNet_yes", 
"shotType_wrist", "strengthState_power.play", "shotType_tip.in", 
"zoneCode_O", "crossedRoyalRoad_yes", "typeDescKeyPrev_tip.in.shot.on.goal.against", 
"zoneCode_N", "typeDescKeyPrev_other.missed.shot.against", "shooterPositionCode_D", 
"typeDescKeyPrev_given.hit", "typeDescKeyPrev_backhand.shot.on.goal.for", 
"shooterHandCode_R", "typeDescKeyPrev_post.missed.shot.for", 
"shotType_deflected", "typeDescKeyPrev_wide.missed.shot.for", 
"typeDescKeyPrev_blocked.shot.for", "goalieHandCode_R", "typeDescKeyPrev_takeaway.against", 
"typeDescKeyPrev_takeaway.for", "typeDescKeyPrev_blocked.shot.against", 
"typeDescKeyPrev_giveaway.for", "typeDescKeyPrev_slap.shot.on.goal.against", 
"typeDescKeyPrev_high.missed.shot.against", "typeDescKeyPrev_snap.shot.on.goal.against", 
"typeDescKeyPrev_slap.shot.on.goal.for", "typeDescKeyPrev_deflected.shot.on.goal.against", 
"shooterPositionCode_L", "shotType_other", "isRush_yes", "typeDescKeyPrev_high.missed.shot.for", 
"typeDescKeyPrev_other.shot.on.goal.for", "isPlayoff_yes", "typeDescKeyPrev_giveaway.against", 
"typeDescKeyPrev_wide.missed.shot.against", "typeDescKeyPrev_post.missed.shot.against", 
"isRebound_yes", "typeDescKeyPrev_other.missed.shot.for", "typeDescKeyPrev_snap.shot.on.goal.for", 
"shooterPositionCode_R", "isEmptyNetFor_yes", "typeDescKeyPrev_other.shot.on.goal.against", 
"isOvertime_yes", "typeDescKeyPrev_tip.in.shot.on.goal.for", 
"typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_wrist.shot.on.goal.against", 
"isHome_yes", "typeDescKeyPrev_taken.hit", "typeDescKeyPrev_deflected.shot.on.goal.for", 
"typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_lost.faceoff"
), active_dummy_by_var = list(crossedRoyalRoad = "crossedRoyalRoad_yes", 
    goalieHandCode = "goalieHandCode_R", isBehindNet = "isBehindNet_yes", 
    isEmptyNetFor = "isEmptyNetFor_yes", isHome = "isHome_yes", 
    isOvertime = "isOvertime_yes", isPlayoff = "isPlayoff_yes", 
    isRebound = "isRebound_yes", isRush = "isRush_yes", shooterHandCode = "shooterHandCode_R", 
    shooterPositionCode = c("shooterPositionCode_D", "shooterPositionCode_L", 
    "shooterPositionCode_R"), shotType = c("shotType_snap", "shotType_slap", 
    "shotType_wrist", "shotType_tip.in", "shotType_deflected", 
    "shotType_other"), strengthState = "strengthState_power.play", 
    typeDescKeyPrev = c("typeDescKeyPrev_tip.in.shot.on.goal.against", 
    "typeDescKeyPrev_other.missed.shot.against", "typeDescKeyPrev_given.hit", 
    "typeDescKeyPrev_backhand.shot.on.goal.for", "typeDescKeyPrev_post.missed.shot.for", 
    "typeDescKeyPrev_wide.missed.shot.for", "typeDescKeyPrev_blocked.shot.for", 
    "typeDescKeyPrev_takeaway.against", "typeDescKeyPrev_takeaway.for", 
    "typeDescKeyPrev_blocked.shot.against", "typeDescKeyPrev_giveaway.for", 
    "typeDescKeyPrev_slap.shot.on.goal.against", "typeDescKeyPrev_high.missed.shot.against", 
    "typeDescKeyPrev_snap.shot.on.goal.against", "typeDescKeyPrev_slap.shot.on.goal.for", 
    "typeDescKeyPrev_deflected.shot.on.goal.against", "typeDescKeyPrev_high.missed.shot.for", 
    "typeDescKeyPrev_other.shot.on.goal.for", "typeDescKeyPrev_giveaway.against", 
    "typeDescKeyPrev_wide.missed.shot.against", "typeDescKeyPrev_post.missed.shot.against", 
    "typeDescKeyPrev_other.missed.shot.for", "typeDescKeyPrev_snap.shot.on.goal.for", 
    "typeDescKeyPrev_other.shot.on.goal.against", "typeDescKeyPrev_tip.in.shot.on.goal.for", 
    "typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_wrist.shot.on.goal.against", 
    "typeDescKeyPrev_taken.hit", "typeDescKeyPrev_deflected.shot.on.goal.for", 
    "typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_lost.faceoff"
    ), zoneCode = c("zoneCode_O", "zoneCode_N")), dummy_term_to_var = c(isPlayoff_yes = "isPlayoff", 
isPlayoff_unknown = "isPlayoff", isPlayoff_new = "isPlayoff", 
isHome_yes = "isHome", isHome_unknown = "isHome", isHome_new = "isHome", 
isOvertime_yes = "isOvertime", isOvertime_unknown = "isOvertime", 
isOvertime_new = "isOvertime", zoneCode_N = "zoneCode", zoneCode_O = "zoneCode", 
zoneCode_unknown = "zoneCode", zoneCode_new = "zoneCode", isBehindNet_yes = "isBehindNet", 
isBehindNet_unknown = "isBehindNet", isBehindNet_new = "isBehindNet", 
crossedRoyalRoad_yes = "crossedRoyalRoad", crossedRoyalRoad_unknown = "crossedRoyalRoad", 
crossedRoyalRoad_new = "crossedRoyalRoad", typeDescKeyPrev_backhand.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_blocked.shot.against = "typeDescKeyPrev", typeDescKeyPrev_blocked.shot.for = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_giveaway.against = "typeDescKeyPrev", typeDescKeyPrev_giveaway.for = "typeDescKeyPrev", 
typeDescKeyPrev_given.hit = "typeDescKeyPrev", typeDescKeyPrev_high.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_high.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_lost.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_other.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_post.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_post.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_slap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_slap.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_snap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_snap.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_takeaway.against = "typeDescKeyPrev", 
typeDescKeyPrev_takeaway.for = "typeDescKeyPrev", typeDescKeyPrev_taken.hit = "typeDescKeyPrev", 
typeDescKeyPrev_tip.in.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_tip.in.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_won.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_unknown = "typeDescKeyPrev", 
typeDescKeyPrev_new = "typeDescKeyPrev", shotType_deflected = "shotType", 
shotType_other = "shotType", shotType_slap = "shotType", shotType_snap = "shotType", 
shotType_tip.in = "shotType", shotType_wrist = "shotType", shotType_unknown = "shotType", 
shotType_new = "shotType", isRebound_yes = "isRebound", isRebound_unknown = "isRebound", 
isRebound_new = "isRebound", isRush_yes = "isRush", isRush_unknown = "isRush", 
isRush_new = "isRush", shooterHandCode_R = "shooterHandCode", 
shooterHandCode_unknown = "shooterHandCode", shooterHandCode_new = "shooterHandCode", 
shooterPositionCode_D = "shooterPositionCode", shooterPositionCode_L = "shooterPositionCode", 
shooterPositionCode_R = "shooterPositionCode", shooterPositionCode_unknown = "shooterPositionCode", 
shooterPositionCode_new = "shooterPositionCode", goalieHandCode_R = "goalieHandCode", 
goalieHandCode_unknown = "goalieHandCode", goalieHandCode_new = "goalieHandCode", 
isEmptyNetFor_yes = "isEmptyNetFor", isEmptyNetFor_unknown = "isEmptyNetFor", 
isEmptyNetFor_new = "isEmptyNetFor", strengthState_power.play = "strengthState", 
strengthState_unknown = "strengthState", strengthState_new = "strengthState"
), dummy_term_to_level = c(isPlayoff_yes = "yes", isPlayoff_unknown = "unknown", 
isPlayoff_new = "new", isHome_yes = "yes", isHome_unknown = "unknown", 
isHome_new = "new", isOvertime_yes = "yes", isOvertime_unknown = "unknown", 
isOvertime_new = "new", zoneCode_N = "N", zoneCode_O = "O", zoneCode_unknown = "unknown", 
zoneCode_new = "new", isBehindNet_yes = "yes", isBehindNet_unknown = "unknown", 
isBehindNet_new = "new", crossedRoyalRoad_yes = "yes", crossedRoyalRoad_unknown = "unknown", 
crossedRoyalRoad_new = "new", typeDescKeyPrev_backhand.shot.on.goal.for = "backhand-shot-on-goal-for", 
typeDescKeyPrev_blocked.shot.against = "blocked-shot-against", 
typeDescKeyPrev_blocked.shot.for = "blocked-shot-for", typeDescKeyPrev_deflected.shot.on.goal.against = "deflected-shot-on-goal-against", 
typeDescKeyPrev_deflected.shot.on.goal.for = "deflected-shot-on-goal-for", 
typeDescKeyPrev_giveaway.against = "giveaway-against", typeDescKeyPrev_giveaway.for = "giveaway-for", 
typeDescKeyPrev_given.hit = "given-hit", typeDescKeyPrev_high.missed.shot.against = "high-missed-shot-against", 
typeDescKeyPrev_high.missed.shot.for = "high-missed-shot-for", 
typeDescKeyPrev_lost.faceoff = "lost-faceoff", typeDescKeyPrev_other.missed.shot.against = "other-missed-shot-against", 
typeDescKeyPrev_other.missed.shot.for = "other-missed-shot-for", 
typeDescKeyPrev_other.shot.on.goal.against = "other-shot-on-goal-against", 
typeDescKeyPrev_other.shot.on.goal.for = "other-shot-on-goal-for", 
typeDescKeyPrev_post.missed.shot.against = "post-missed-shot-against", 
typeDescKeyPrev_post.missed.shot.for = "post-missed-shot-for", 
typeDescKeyPrev_slap.shot.on.goal.against = "slap-shot-on-goal-against", 
typeDescKeyPrev_slap.shot.on.goal.for = "slap-shot-on-goal-for", 
typeDescKeyPrev_snap.shot.on.goal.against = "snap-shot-on-goal-against", 
typeDescKeyPrev_snap.shot.on.goal.for = "snap-shot-on-goal-for", 
typeDescKeyPrev_takeaway.against = "takeaway-against", typeDescKeyPrev_takeaway.for = "takeaway-for", 
typeDescKeyPrev_taken.hit = "taken-hit", typeDescKeyPrev_tip.in.shot.on.goal.against = "tip-in-shot-on-goal-against", 
typeDescKeyPrev_tip.in.shot.on.goal.for = "tip-in-shot-on-goal-for", 
typeDescKeyPrev_wide.missed.shot.against = "wide-missed-shot-against", 
typeDescKeyPrev_wide.missed.shot.for = "wide-missed-shot-for", 
typeDescKeyPrev_won.faceoff = "won-faceoff", typeDescKeyPrev_wrist.shot.on.goal.against = "wrist-shot-on-goal-against", 
typeDescKeyPrev_wrist.shot.on.goal.for = "wrist-shot-on-goal-for", 
typeDescKeyPrev_unknown = "unknown", typeDescKeyPrev_new = "new", 
shotType_deflected = "deflected", shotType_other = "other", shotType_slap = "slap", 
shotType_snap = "snap", shotType_tip.in = "tip-in", shotType_wrist = "wrist", 
shotType_unknown = "unknown", shotType_new = "new", isRebound_yes = "yes", 
isRebound_unknown = "unknown", isRebound_new = "new", isRush_yes = "yes", 
isRush_unknown = "unknown", isRush_new = "new", shooterHandCode_R = "R", 
shooterHandCode_unknown = "unknown", shooterHandCode_new = "new", 
shooterPositionCode_D = "D", shooterPositionCode_L = "L", shooterPositionCode_R = "R", 
shooterPositionCode_unknown = "unknown", shooterPositionCode_new = "new", 
goalieHandCode_R = "R", goalieHandCode_unknown = "unknown", goalieHandCode_new = "new", 
isEmptyNetFor_yes = "yes", isEmptyNetFor_unknown = "unknown", 
isEmptyNetFor_new = "new", strengthState_power.play = "power-play", 
strengthState_unknown = "unknown", strengthState_new = "new"), 
    dummy_map = list(crossedRoyalRoad = c(yes = "crossedRoyalRoad_yes", 
    unknown = "crossedRoyalRoad_unknown", new = "crossedRoyalRoad_new"
    ), goalieHandCode = c(R = "goalieHandCode_R", unknown = "goalieHandCode_unknown", 
    new = "goalieHandCode_new"), isBehindNet = c(yes = "isBehindNet_yes", 
    unknown = "isBehindNet_unknown", new = "isBehindNet_new"), 
        isEmptyNetFor = c(yes = "isEmptyNetFor_yes", unknown = "isEmptyNetFor_unknown", 
        new = "isEmptyNetFor_new"), isHome = c(yes = "isHome_yes", 
        unknown = "isHome_unknown", new = "isHome_new"), isOvertime = c(yes = "isOvertime_yes", 
        unknown = "isOvertime_unknown", new = "isOvertime_new"
        ), isPlayoff = c(yes = "isPlayoff_yes", unknown = "isPlayoff_unknown", 
        new = "isPlayoff_new"), isRebound = c(yes = "isRebound_yes", 
        unknown = "isRebound_unknown", new = "isRebound_new"), 
        isRush = c(yes = "isRush_yes", unknown = "isRush_unknown", 
        new = "isRush_new"), shooterHandCode = c(R = "shooterHandCode_R", 
        unknown = "shooterHandCode_unknown", new = "shooterHandCode_new"
        ), shooterPositionCode = c(D = "shooterPositionCode_D", 
        L = "shooterPositionCode_L", R = "shooterPositionCode_R", 
        unknown = "shooterPositionCode_unknown", new = "shooterPositionCode_new"
        ), shotType = c(deflected = "shotType_deflected", other = "shotType_other", 
        slap = "shotType_slap", snap = "shotType_snap", `tip-in` = "shotType_tip.in", 
        wrist = "shotType_wrist", unknown = "shotType_unknown", 
        new = "shotType_new"), strengthState = c(`power-play` = "strengthState_power.play", 
        unknown = "strengthState_unknown", new = "strengthState_new"
        ), typeDescKeyPrev = c(`backhand-shot-on-goal-for` = "typeDescKeyPrev_backhand.shot.on.goal.for", 
        `blocked-shot-against` = "typeDescKeyPrev_blocked.shot.against", 
        `blocked-shot-for` = "typeDescKeyPrev_blocked.shot.for", 
        `deflected-shot-on-goal-against` = "typeDescKeyPrev_deflected.shot.on.goal.against", 
        `deflected-shot-on-goal-for` = "typeDescKeyPrev_deflected.shot.on.goal.for", 
        `giveaway-against` = "typeDescKeyPrev_giveaway.against", 
        `giveaway-for` = "typeDescKeyPrev_giveaway.for", `given-hit` = "typeDescKeyPrev_given.hit", 
        `high-missed-shot-against` = "typeDescKeyPrev_high.missed.shot.against", 
        `high-missed-shot-for` = "typeDescKeyPrev_high.missed.shot.for", 
        `lost-faceoff` = "typeDescKeyPrev_lost.faceoff", `other-missed-shot-against` = "typeDescKeyPrev_other.missed.shot.against", 
        `other-missed-shot-for` = "typeDescKeyPrev_other.missed.shot.for", 
        `other-shot-on-goal-against` = "typeDescKeyPrev_other.shot.on.goal.against", 
        `other-shot-on-goal-for` = "typeDescKeyPrev_other.shot.on.goal.for", 
        `post-missed-shot-against` = "typeDescKeyPrev_post.missed.shot.against", 
        `post-missed-shot-for` = "typeDescKeyPrev_post.missed.shot.for", 
        `slap-shot-on-goal-against` = "typeDescKeyPrev_slap.shot.on.goal.against", 
        `slap-shot-on-goal-for` = "typeDescKeyPrev_slap.shot.on.goal.for", 
        `snap-shot-on-goal-against` = "typeDescKeyPrev_snap.shot.on.goal.against", 
        `snap-shot-on-goal-for` = "typeDescKeyPrev_snap.shot.on.goal.for", 
        `takeaway-against` = "typeDescKeyPrev_takeaway.against", 
        `takeaway-for` = "typeDescKeyPrev_takeaway.for", `taken-hit` = "typeDescKeyPrev_taken.hit", 
        `tip-in-shot-on-goal-against` = "typeDescKeyPrev_tip.in.shot.on.goal.against", 
        `tip-in-shot-on-goal-for` = "typeDescKeyPrev_tip.in.shot.on.goal.for", 
        `wide-missed-shot-against` = "typeDescKeyPrev_wide.missed.shot.against", 
        `wide-missed-shot-for` = "typeDescKeyPrev_wide.missed.shot.for", 
        `won-faceoff` = "typeDescKeyPrev_won.faceoff", `wrist-shot-on-goal-against` = "typeDescKeyPrev_wrist.shot.on.goal.against", 
        `wrist-shot-on-goal-for` = "typeDescKeyPrev_wrist.shot.on.goal.for", 
        unknown = "typeDescKeyPrev_unknown", new = "typeDescKeyPrev_new"
        ), zoneCode = c(N = "zoneCode_N", O = "zoneCode_O", unknown = "zoneCode_unknown", 
        new = "zoneCode_new")), known_levels = list(isPlayoff = c("no", 
    "yes"), isHome = c("no", "yes"), isOvertime = c("no", "yes"
    ), zoneCode = c("D", "N", "O"), isBehindNet = c("no", "yes"
    ), crossedRoyalRoad = c("no", "yes"), typeDescKeyPrev = c("backhand-shot-on-goal-against", 
    "backhand-shot-on-goal-for", "blocked-shot-against", "blocked-shot-for", 
    "deflected-shot-on-goal-against", "deflected-shot-on-goal-for", 
    "giveaway-against", "giveaway-for", "given-hit", "high-missed-shot-against", 
    "high-missed-shot-for", "lost-faceoff", "other-missed-shot-against", 
    "other-missed-shot-for", "other-shot-on-goal-against", "other-shot-on-goal-for", 
    "post-missed-shot-against", "post-missed-shot-for", "slap-shot-on-goal-against", 
    "slap-shot-on-goal-for", "snap-shot-on-goal-against", "snap-shot-on-goal-for", 
    "takeaway-against", "takeaway-for", "taken-hit", "tip-in-shot-on-goal-against", 
    "tip-in-shot-on-goal-for", "wide-missed-shot-against", "wide-missed-shot-for", 
    "won-faceoff", "wrist-shot-on-goal-against", "wrist-shot-on-goal-for"
    ), shotType = c("backhand", "deflected", "other", "slap", 
    "snap", "tip-in", "wrist"), isRebound = c("no", "yes"), isRush = c("no", 
    "yes"), shooterHandCode = c("L", "R"), shooterPositionCode = c("C", 
    "D", "L", "R"), goalieHandCode = c("L", "R"), isEmptyNetFor = c("no", 
    "yes"), strengthState = c("even-strength", "power-play")), 
    logical_vars = c("crossedRoyalRoad", "isBehindNet", "isEmptyNetFor", 
    "isHome", "isOvertime", "isPlayoff", "isRebound", "isRush"
    ), has_unknown = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
    isBehindNet = TRUE, isEmptyNetFor = TRUE, isHome = TRUE, 
    isOvertime = TRUE, isPlayoff = TRUE, isRebound = TRUE, isRush = TRUE, 
    shooterHandCode = TRUE, shooterPositionCode = TRUE, shotType = TRUE, 
    strengthState = TRUE, typeDescKeyPrev = TRUE, zoneCode = TRUE
    ), has_new = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
    isBehindNet = TRUE, isEmptyNetFor = TRUE, isHome = TRUE, 
    isOvertime = TRUE, isPlayoff = TRUE, isRebound = TRUE, isRush = TRUE, 
    shooterHandCode = TRUE, shooterPositionCode = TRUE, shotType = TRUE, 
    strengthState = TRUE, typeDescKeyPrev = TRUE, zoneCode = TRUE
    ), medians = c(distance = 29.0688837074973, angle = 30.1413855520753, 
    shotType_snap = 0, xCoordNorm = 67, shotType_slap = 0, dDistancePerSecond = -0.710599964774739, 
    skaterCountAgainst = 4, isBehindNet_yes = 0, fenwickDifferential = 1, 
    shotType_wrist = 0, dAnglePerSecond = 0.129172955109414, 
    manDifferential = 1, strengthState_power.play = 1, shotType_tip.in = 0, 
    shooterSecondsElapsedInShift = 43, dDistance = -9.0351168524732, 
    zoneCode_O = 1, dXCoordNormPerSecond = 0.625, crossedRoyalRoad_yes = 0, 
    shotDifferential = 1, corsiDifferential = 2, typeDescKeyPrev_tip.in.shot.on.goal.against = 0, 
    avgSecondsElapsedInShiftAgainst = 34.5, zoneCode_N = 0, fenwickAgainst = 23, 
    dYCoordNormPerSecond = -0.0434782608695652, corsiAgainst = 32, 
    typeDescKeyPrev_other.missed.shot.against = 0, shooterPositionCode_D = 0, 
    maxSecondsElapsedInShiftFor = 61, typeDescKeyPrev_given.hit = 0, 
    goalsFor = 1, secondsElapsedInPeriod = 695, typeDescKeyPrev_backhand.shot.on.goal.for = 0, 
    goalieWeight = 201, shotsAgainst = 15, dXCoordNorm = 8, shooterHandCode_R = 0, 
    avgSecondsElapsedSinceLastShiftAgainst = 130.75, typeDescKeyPrev_post.missed.shot.for = 0, 
    shotType_deflected = 0, secondsElapsedInSequence = 38, shooterSecondsElapsedSinceLastShift = 142, 
    typeDescKeyPrev_wide.missed.shot.for = 0, typeDescKeyPrev_blocked.shot.for = 0, 
    goalieHandCode_R = 0, maxSecondsElapsedSinceLastShiftAgainst = 191, 
    minSecondsElapsedInShiftAgainst = 24, typeDescKeyPrev_takeaway.against = 0, 
    goalDifferential = 0, avgSecondsElapsedSinceLastShiftFor = 151.2, 
    typeDescKeyPrev_takeaway.for = 0, typeDescKeyPrev_blocked.shot.against = 0, 
    typeDescKeyPrev_giveaway.for = 0, typeDescKeyPrev_slap.shot.on.goal.against = 0, 
    typeDescKeyPrev_high.missed.shot.against = 0, avgSecondsElapsedInShiftFor = 44.2, 
    periodNumber = 2, typeDescKeyPrev_snap.shot.on.goal.against = 0, 
    typeDescKeyPrev_slap.shot.on.goal.for = 0, typeDescKeyPrev_deflected.shot.on.goal.against = 0, 
    shooterWeight = 200, goalieAge = 28, shooterPositionCode_L = 0, 
    dAngle = 2.43439897191186, fenwickFor = 24, yCoordNorm = 0, 
    shotType_other = 0, isRush_yes = 0, typeDescKeyPrev_high.missed.shot.for = 0, 
    skaterCountFor = 5, typeDescKeyPrev_other.shot.on.goal.for = 0, 
    isPlayoff_yes = 0, typeDescKeyPrev_giveaway.against = 0, 
    typeDescKeyPrev_wide.missed.shot.against = 0, minSecondsElapsedInShiftFor = 30, 
    typeDescKeyPrev_post.missed.shot.against = 0, isRebound_yes = 0, 
    typeDescKeyPrev_other.missed.shot.for = 0, typeDescKeyPrev_snap.shot.on.goal.for = 0, 
    maxSecondsElapsedInShiftAgainst = 46, shooterPositionCode_R = 0, 
    isEmptyNetFor_yes = 0, typeDescKeyPrev_other.shot.on.goal.against = 0, 
    isOvertime_yes = 0, typeDescKeyPrev_tip.in.shot.on.goal.for = 0, 
    typeDescKeyPrev_won.faceoff = 0, typeDescKeyPrev_wrist.shot.on.goal.against = 0, 
    shotsFor = 16, isHome_yes = 1, dSecondsElapsedInSequence = 13, 
    goalsAgainst = 1, typeDescKeyPrev_taken.hit = 0, shooterAge = 28, 
    maxSecondsElapsedSinceLastShiftFor = 219, secondsElapsedInGame = 2024, 
    corsiFor = 34, goalieHeight = 75, typeDescKeyPrev_deflected.shot.on.goal.for = 0, 
    minSecondsElapsedSinceLastShiftFor = 90, dYCoordNorm = -1, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0, shooterHeight = 73, 
    typeDescKeyPrev_lost.faceoff = 0, minSecondsElapsedSinceLastShiftAgainst = 75
    ), normalize_means = c(distance = 29.1825308626806, angle = 31.8740842498832, 
    shotType_snap = 0.182333358745648, xCoordNorm = 64.7012020024904, 
    shotType_slap = 0.190262000965668, dDistancePerSecond = -3.92513227181903, 
    skaterCountAgainst = 4.08754542451272, isBehindNet_yes = 0.0124774465706081, 
    fenwickDifferential = 1.43399151228685, shotType_wrist = 0.433178318213006, 
    dAnglePerSecond = 0.394836951007524, manDifferential = 0.8954791491957, 
    strengthState_power.play = 0.870981677721024, shotType_tip.in = 0.112297019135473, 
    shooterSecondsElapsedInShift = 49.4412848466367, dDistance = -30.5685391359683, 
    zoneCode_O = 0.994383878427486, dXCoordNormPerSecond = 3.74943871134502, 
    crossedRoyalRoad_yes = 0.432543010342812, shotDifferential = 0.795278391908719, 
    corsiDifferential = 2.3450229981449, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.000432009351731849, 
    avgSecondsElapsedInShiftAgainst = 37.6706661414787, zoneCode_N = 0.0045488043505883, 
    fenwickAgainst = 23.3634215140657, dYCoordNormPerSecond = -0.126969700004367, 
    corsiAgainst = 32.3850982185967, typeDescKeyPrev_other.missed.shot.against = 0.000228710833269803, 
    shooterPositionCode_D = 0.18230794643084, maxSecondsElapsedInShiftFor = 67.6061599451094, 
    typeDescKeyPrev_given.hit = 0.0182460420319687, goalsFor = 1.36741124749053, 
    secondsElapsedInPeriod = 676.414779802292, typeDescKeyPrev_backhand.shot.on.goal.for = 0.0112830677746436, 
    goalieWeight = 201.753983380346, shotsAgainst = 15.8849584508653, 
    dXCoordNorm = 30.0671647480369, shooterHandCode_R = 0.405936316739092, 
    avgSecondsElapsedSinceLastShiftAgainst = 148.424941339906, 
    typeDescKeyPrev_post.missed.shot.for = 0.0130365174963787, 
    shotType_deflected = 0.0219308276790933, secondsElapsedInSequence = 46.8849838631801, 
    shooterSecondsElapsedSinceLastShift = 167.745698965719, typeDescKeyPrev_wide.missed.shot.for = 0.0988284922873625, 
    typeDescKeyPrev_blocked.shot.for = 0.138370054128231, goalieHandCode_R = 0.0564407511880257, 
    maxSecondsElapsedSinceLastShiftAgainst = 218.025666437956, 
    minSecondsElapsedInShiftAgainst = 27.7192955706335, typeDescKeyPrev_takeaway.against = 0.0156285736067698, 
    goalDifferential = -0.324743970928312, avgSecondsElapsedSinceLastShiftFor = 166.560267845798, 
    typeDescKeyPrev_takeaway.for = 0.0166704785138878, typeDescKeyPrev_blocked.shot.against = 0.0071916850905949, 
    typeDescKeyPrev_giveaway.for = 0.0265050443444893, typeDescKeyPrev_slap.shot.on.goal.against = 0.00180427435135066, 
    typeDescKeyPrev_high.missed.shot.against = 0.000533658610962873, 
    avgSecondsElapsedInShiftFor = 49.0909896402464, periodNumber = 2.12324972681762, 
    typeDescKeyPrev_snap.shot.on.goal.against = 0.00355772407308582, 
    typeDescKeyPrev_slap.shot.on.goal.for = 0.0311300856395009, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0.000177886203654291, 
    shooterWeight = 199.48771314579, goalieAge = 28.5687276053976, 
    shooterPositionCode_L = 0.220528067901705, dAngle = 2.46539410588522, 
    fenwickFor = 24.7974130263526, yCoordNorm = -0.153363319864806, 
    shotType_other = 0.0129348682371477, isRush_yes = 0.0254377271225636, 
    typeDescKeyPrev_high.missed.shot.for = 0.00820817768290514, 
    skaterCountFor = 5.12955198088994, typeDescKeyPrev_other.shot.on.goal.for = 0.0028207669436609, 
    isPlayoff_yes = 0.0642677441488145, typeDescKeyPrev_giveaway.against = 0.0311554979543087, 
    typeDescKeyPrev_wide.missed.shot.against = 0.00752204518309573, 
    minSecondsElapsedInShiftFor = 34.7852913521893, typeDescKeyPrev_post.missed.shot.against = 0.00106731722192575, 
    isRebound_yes = 0.117608192930294, typeDescKeyPrev_other.missed.shot.for = 0.00447256740616503, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0.0345099235089324, 
    maxSecondsElapsedInShiftAgainst = 51.318975375467, shooterPositionCode_R = 0.210744326700719, 
    isEmptyNetFor_yes = 0.14652740718152, typeDescKeyPrev_other.shot.on.goal.against = 0.00030494777769307, 
    isOvertime_yes = 0.00942796879367742, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0201773779573581, 
    typeDescKeyPrev_won.faceoff = 0.218571319661508, typeDescKeyPrev_wrist.shot.on.goal.against = 0.018423928235623, 
    shotsFor = 16.680236842774, isHome_yes = 0.524561002261696, 
    dSecondsElapsedInSequence = 17.7462580366446, goalsAgainst = 1.69215521841885, 
    typeDescKeyPrev_taken.hit = 0.0452847449874209, shooterAge = 27.7757617341364, 
    maxSecondsElapsedSinceLastShiftFor = 242.675433915275, secondsElapsedInGame = 2024.31445198343, 
    corsiFor = 34.7301212167416, goalieHeight = 74.919493786689, 
    typeDescKeyPrev_deflected.shot.on.goal.for = 0.00424385657289523, 
    minSecondsElapsedSinceLastShiftFor = 104.568117709842, dYCoordNorm = -0.922594089095576, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.0937968539554268, 
    shooterHeight = 72.9694798099159, typeDescKeyPrev_lost.faceoff = 0.12342761302127, 
    minSecondsElapsedSinceLastShiftAgainst = 93.0578130161876
    ), normalize_sds = c(distance = 17.1583418484218, angle = 22.0452025613815, 
    shotType_snap = 0.386123935801959, xCoordNorm = 17.1798785507251, 
    shotType_slap = 0.392512785951225, dDistancePerSecond = 15.0300317182372, 
    skaterCountAgainst = 0.398963269337967, isBehindNet_yes = 0.111004833363643, 
    fenwickDifferential = 10.0858000156472, shotType_wrist = 0.49552104154212, 
    dAnglePerSecond = 11.498787437631, manDifferential = 0.377617278689708, 
    strengthState_power.play = 0.335224477798939, shotType_tip.in = 0.31573554116574, 
    shooterSecondsElapsedInShift = 34.4353479956146, dDistance = 57.1889220770317, 
    zoneCode_O = 0.0747310020789636, dXCoordNormPerSecond = 15.4006870048607, 
    crossedRoyalRoad_yes = 0.49543495248741, shotDifferential = 7.32490965589123, 
    corsiDifferential = 13.6752390553583, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.0207806085942067, 
    avgSecondsElapsedInShiftAgainst = 21.4078116651204, zoneCode_N = 0.0672921080240239, 
    fenwickAgainst = 13.8396973446054, dYCoordNormPerSecond = 6.31585281634119, 
    corsiAgainst = 18.8722248204091, typeDescKeyPrev_other.missed.shot.against = 0.0151216512165405, 
    shooterPositionCode_D = 0.386103026993762, maxSecondsElapsedInShiftFor = 39.4366029021034, 
    typeDescKeyPrev_given.hit = 0.133841619863337, goalsFor = 1.34481892692647, 
    secondsElapsedInPeriod = 354.534291015798, typeDescKeyPrev_backhand.shot.on.goal.for = 0.105622174078812, 
    goalieWeight = 15.8038237963044, shotsAgainst = 9.67362991872764, 
    dXCoordNorm = 59.3483643223646, shooterHandCode_R = 0.491078559782661, 
    avgSecondsElapsedSinceLastShiftAgainst = 76.2167035701549, 
    typeDescKeyPrev_post.missed.shot.for = 0.113432330865189, 
    shotType_deflected = 0.146459590266464, secondsElapsedInSequence = 38.2816876532339, 
    shooterSecondsElapsedSinceLastShift = 103.030093210562, typeDescKeyPrev_wide.missed.shot.for = 0.298435394539607, 
    typeDescKeyPrev_blocked.shot.for = 0.345292357398837, goalieHandCode_R = 0.230773798697999, 
    maxSecondsElapsedSinceLastShiftAgainst = 114.978414202739, 
    minSecondsElapsedInShiftAgainst = 18.867155017967, typeDescKeyPrev_takeaway.against = 0.124035125085396, 
    goalDifferential = 1.6851929845146, avgSecondsElapsedSinceLastShiftFor = 74.897713499265, 
    typeDescKeyPrev_takeaway.for = 0.128035113323765, typeDescKeyPrev_blocked.shot.against = 0.0844993858190727, 
    typeDescKeyPrev_giveaway.for = 0.160633691009623, typeDescKeyPrev_slap.shot.on.goal.against = 0.0424389527985708, 
    typeDescKeyPrev_high.missed.shot.against = 0.0230951807539552, 
    avgSecondsElapsedInShiftFor = 28.9302031044488, periodNumber = 0.81936906375636, 
    typeDescKeyPrev_snap.shot.on.goal.against = 0.0595412190263552, 
    typeDescKeyPrev_slap.shot.on.goal.for = 0.173671442350077, 
    typeDescKeyPrev_deflected.shot.on.goal.against = 0.013336381816806, 
    shooterWeight = 15.2288879916244, goalieAge = 3.74472564683238, 
    shooterPositionCode_L = 0.414608016735528, dAngle = 32.2121187715559, 
    fenwickFor = 14.4166896230049, yCoordNorm = 16.1379090363397, 
    shotType_other = 0.112995052468134, isRush_yes = 0.157452466367448, 
    typeDescKeyPrev_high.missed.shot.for = 0.0902275478108235, 
    skaterCountFor = 0.377699315051126, typeDescKeyPrev_other.shot.on.goal.for = 0.0530366071628462, 
    isPlayoff_yes = 0.245232398920371, typeDescKeyPrev_giveaway.against = 0.173740035649566, 
    typeDescKeyPrev_wide.missed.shot.against = 0.0864040145994382, 
    minSecondsElapsedInShiftFor = 25.7503886187896, typeDescKeyPrev_post.missed.shot.against = 0.0326527969799588, 
    isRebound_yes = 0.322147703940644, typeDescKeyPrev_other.missed.shot.for = 0.0667283800176815, 
    typeDescKeyPrev_snap.shot.on.goal.for = 0.182537216540878, 
    maxSecondsElapsedInShiftAgainst = 31.620876052298, shooterPositionCode_R = 0.407842349972714, 
    isEmptyNetFor_yes = 0.35363866332461, typeDescKeyPrev_other.shot.on.goal.against = 0.0174603130501346, 
    isOvertime_yes = 0.0966401548623797, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.140608512532968, 
    typeDescKeyPrev_won.faceoff = 0.413282274436155, typeDescKeyPrev_wrist.shot.on.goal.against = 0.134480283626792, 
    shotsFor = 9.92035042136552, isHome_yes = 0.499402738356446, 
    dSecondsElapsedInSequence = 16.5505773566519, goalsAgainst = 1.56223513285134, 
    typeDescKeyPrev_taken.hit = 0.207930602758942, shooterAge = 4.33663898086633, 
    maxSecondsElapsedSinceLastShiftFor = 110.630508651649, secondsElapsedInGame = 1054.49872281385, 
    corsiFor = 20.0597708432829, goalieHeight = 1.5926527327773, 
    typeDescKeyPrev_deflected.shot.on.goal.for = 0.065007335321074, 
    minSecondsElapsedSinceLastShiftFor = 74.5816495106049, dYCoordNorm = 25.4658354094814, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.291549591355914, 
    shooterHeight = 2.05312137394645, typeDescKeyPrev_lost.faceoff = 0.328931583882874, 
    minSecondsElapsedSinceLastShiftAgainst = 72.0762813257111
    )), sh = list(coefficients = c(`(Intercept)` = -3.63153798338081, 
distance = -0.966884796033344, xCoordNorm = 0.712287276359287, 
angle = -0.299721153736643, minSecondsElapsedInShiftFor = 0.270697844828597, 
dDistance = -0.270424342210878, shotType_tip.in = -0.18748895339856, 
shotType_snap = 0.181593017886797, zoneCode_O = 0.170362375447486, 
typeDescKeyPrev_tip.in.shot.on.goal.against = -0.159885480263973, 
typeDescKeyPrev_won.faceoff = -0.153910966629935, avgSecondsElapsedInShiftAgainst = 0.150788932169999, 
typeDescKeyPrev_post.missed.shot.against = -0.14744703173948, 
zoneCode_N = -0.122759549835248, minSecondsElapsedSinceLastShiftFor = 0.119446699869042, 
avgSecondsElapsedSinceLastShiftFor = -0.117335141414068, secondsElapsedInGame = -0.11386113608954, 
goalieHeight = 0.112983719002759, dXCoordNorm = 0.111461702734452, 
corsiDifferential = -0.109646778465175, isOvertime_yes = -0.108763551334847, 
shooterHandCode_R = 0.105819579002731, shooterPositionCode_D = -0.104538563426477, 
fenwickDifferential = 0.104244256422268, typeDescKeyPrev_blocked.shot.for = 0.0999938116499298, 
manDifferential = 0.0975666891266696, typeDescKeyPrev_takeaway.for = 0.0969216503133783, 
dDistancePerSecond = 0.0949942661470541, typeDescKeyPrev_lost.faceoff = 0.0942728934704784, 
shooterSecondsElapsedSinceLastShift = 0.0930235664484371, typeDescKeyPrev_post.missed.shot.for = 0.0900011186416822, 
shooterWeight = 0.089343021300825, periodNumber = -0.0892117065006534, 
dAngle = -0.0876346390397818, goalieWeight = -0.0867592980503014, 
dXCoordNormPerSecond = -0.0860803941318523, shooterHeight = -0.0849093889814402, 
goalieAge = 0.0839532813572198, shooterSecondsElapsedInShift = -0.0801969245795928, 
dYCoordNormPerSecond = 0.0795518417492723, corsiAgainst = 0.0755924357662071, 
secondsElapsedInPeriod = -0.0720412805370922, shooterPositionCode_R = 0.0636785445524335, 
maxSecondsElapsedSinceLastShiftAgainst = 0.0624502944569121, 
isRebound_yes = 0.0608391175688252, shotType_other = -0.0605088294240121, 
typeDescKeyPrev_deflected.shot.on.goal.against = -0.0602968646760614, 
dYCoordNorm = -0.0588440238072053, typeDescKeyPrev_slap.shot.on.goal.for = 0.0576767301623944, 
typeDescKeyPrev_other.shot.on.goal.for = -0.0567382282269528, 
dSecondsElapsedInSequence = -0.0559515265150227, typeDescKeyPrev_high.missed.shot.for = -0.052714401553124, 
avgSecondsElapsedInShiftFor = 0.0514710784218389, typeDescKeyPrev_snap.shot.on.goal.against = 0.0488086779809238, 
maxSecondsElapsedInShiftFor = 0.0483594781430845, crossedRoyalRoad_yes = 0.0472410451795902, 
typeDescKeyPrev_deflected.shot.on.goal.for = -0.0443317884863132, 
fenwickFor = 0.0425205663259666, typeDescKeyPrev_giveaway.against = 0.0412982762614141, 
typeDescKeyPrev_slap.shot.on.goal.against = 0.0396858019227885, 
typeDescKeyPrev_other.missed.shot.for = -0.0390228667144192, 
shotsFor = 0.0386766739245762, fenwickAgainst = -0.0369460332199352, 
maxSecondsElapsedSinceLastShiftFor = -0.0351154256210093, skaterCountFor = 0.0343034423586185, 
isHome_yes = -0.0340077143330081, shotType_slap = 0.0337787048249246, 
typeDescKeyPrev_giveaway.for = -0.0336722418317461, typeDescKeyPrev_taken.hit = -0.0330078733225434, 
skaterCountAgainst = -0.0316860879633458, typeDescKeyPrev_other.shot.on.goal.against = 0.0313133374320686, 
minSecondsElapsedSinceLastShiftAgainst = -0.0312521695137525, 
typeDescKeyPrev_tip.in.shot.on.goal.for = 0.031104374441944, 
typeDescKeyPrev_wrist.shot.on.goal.against = -0.0310693612044417, 
isBehindNet_yes = 0.0306816035521421, shooterPositionCode_L = -0.0299236912805296, 
yCoordNorm = 0.0289363151396329, typeDescKeyPrev_other.missed.shot.against = 0.0287245706785486, 
typeDescKeyPrev_blocked.shot.against = 0.0285931746315683, goalsFor = 0.0285508077234967, 
maxSecondsElapsedInShiftAgainst = -0.0285023865466986, typeDescKeyPrev_snap.shot.on.goal.for = -0.0269096392869296, 
isRush_yes = -0.026745104240082, minSecondsElapsedInShiftAgainst = -0.0238173645997878, 
shotsAgainst = 0.0234209182283407, avgSecondsElapsedSinceLastShiftAgainst = -0.0226078467351798, 
secondsElapsedInSequence = -0.0211884910172206, shotDifferential = 0.0209405477431257, 
typeDescKeyPrev_wide.missed.shot.against = 0.0205110639444662, 
shotType_deflected = -0.0203621698906344, shotType_wrist = -0.0201989463446854, 
goalsAgainst = 0.0200509202816621, typeDescKeyPrev_wrist.shot.on.goal.for = 0.0196509176599285, 
dAnglePerSecond = 0.0196432283184315, shooterAge = 0.0163202459346931, 
typeDescKeyPrev_high.missed.shot.against = 0.0159443789044939, 
isEmptyNetFor_yes = -0.0153174831542125, goalieHandCode_R = -0.0119914869246025, 
typeDescKeyPrev_given.hit = -0.0109804652957711, goalDifferential = 0.00742708264657712, 
corsiFor = -0.0062692423970964, isPlayoff_yes = 0.0060658315479522, 
typeDescKeyPrev_takeaway.against = -0.00548341528040623, typeDescKeyPrev_wide.missed.shot.for = -0.0022947453173044, 
typeDescKeyPrev_backhand.shot.on.goal.for = 0.00164041606227585
), raw_terms = c("distance", "xCoordNorm", "angle", "minSecondsElapsedInShiftFor", 
"dDistance", "avgSecondsElapsedInShiftAgainst", "minSecondsElapsedSinceLastShiftFor", 
"avgSecondsElapsedSinceLastShiftFor", "secondsElapsedInGame", 
"goalieHeight", "dXCoordNorm", "corsiDifferential", "fenwickDifferential", 
"manDifferential", "dDistancePerSecond", "shooterSecondsElapsedSinceLastShift", 
"shooterWeight", "periodNumber", "dAngle", "goalieWeight", "dXCoordNormPerSecond", 
"shooterHeight", "goalieAge", "shooterSecondsElapsedInShift", 
"dYCoordNormPerSecond", "corsiAgainst", "secondsElapsedInPeriod", 
"maxSecondsElapsedSinceLastShiftAgainst", "dYCoordNorm", "dSecondsElapsedInSequence", 
"avgSecondsElapsedInShiftFor", "maxSecondsElapsedInShiftFor", 
"fenwickFor", "shotsFor", "fenwickAgainst", "maxSecondsElapsedSinceLastShiftFor", 
"skaterCountFor", "skaterCountAgainst", "minSecondsElapsedSinceLastShiftAgainst", 
"yCoordNorm", "goalsFor", "maxSecondsElapsedInShiftAgainst", 
"minSecondsElapsedInShiftAgainst", "shotsAgainst", "avgSecondsElapsedSinceLastShiftAgainst", 
"secondsElapsedInSequence", "shotDifferential", "goalsAgainst", 
"dAnglePerSecond", "shooterAge", "goalDifferential", "corsiFor"
), active_dummy_terms = c("shotType_tip.in", "shotType_snap", 
"zoneCode_O", "typeDescKeyPrev_tip.in.shot.on.goal.against", 
"typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_post.missed.shot.against", 
"zoneCode_N", "isOvertime_yes", "shooterHandCode_R", "shooterPositionCode_D", 
"typeDescKeyPrev_blocked.shot.for", "typeDescKeyPrev_takeaway.for", 
"typeDescKeyPrev_lost.faceoff", "typeDescKeyPrev_post.missed.shot.for", 
"shooterPositionCode_R", "isRebound_yes", "shotType_other", "typeDescKeyPrev_deflected.shot.on.goal.against", 
"typeDescKeyPrev_slap.shot.on.goal.for", "typeDescKeyPrev_other.shot.on.goal.for", 
"typeDescKeyPrev_high.missed.shot.for", "typeDescKeyPrev_snap.shot.on.goal.against", 
"crossedRoyalRoad_yes", "typeDescKeyPrev_deflected.shot.on.goal.for", 
"typeDescKeyPrev_giveaway.against", "typeDescKeyPrev_slap.shot.on.goal.against", 
"typeDescKeyPrev_other.missed.shot.for", "isHome_yes", "shotType_slap", 
"typeDescKeyPrev_giveaway.for", "typeDescKeyPrev_taken.hit", 
"typeDescKeyPrev_other.shot.on.goal.against", "typeDescKeyPrev_tip.in.shot.on.goal.for", 
"typeDescKeyPrev_wrist.shot.on.goal.against", "isBehindNet_yes", 
"shooterPositionCode_L", "typeDescKeyPrev_other.missed.shot.against", 
"typeDescKeyPrev_blocked.shot.against", "typeDescKeyPrev_snap.shot.on.goal.for", 
"isRush_yes", "typeDescKeyPrev_wide.missed.shot.against", "shotType_deflected", 
"shotType_wrist", "typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_high.missed.shot.against", 
"isEmptyNetFor_yes", "goalieHandCode_R", "typeDescKeyPrev_given.hit", 
"isPlayoff_yes", "typeDescKeyPrev_takeaway.against", "typeDescKeyPrev_wide.missed.shot.for", 
"typeDescKeyPrev_backhand.shot.on.goal.for"), active_dummy_by_var = list(
    crossedRoyalRoad = "crossedRoyalRoad_yes", goalieHandCode = "goalieHandCode_R", 
    isBehindNet = "isBehindNet_yes", isEmptyNetFor = "isEmptyNetFor_yes", 
    isHome = "isHome_yes", isOvertime = "isOvertime_yes", isPlayoff = "isPlayoff_yes", 
    isRebound = "isRebound_yes", isRush = "isRush_yes", shooterHandCode = "shooterHandCode_R", 
    shooterPositionCode = c("shooterPositionCode_D", "shooterPositionCode_R", 
    "shooterPositionCode_L"), shotType = c("shotType_tip.in", 
    "shotType_snap", "shotType_other", "shotType_slap", "shotType_deflected", 
    "shotType_wrist"), typeDescKeyPrev = c("typeDescKeyPrev_tip.in.shot.on.goal.against", 
    "typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_post.missed.shot.against", 
    "typeDescKeyPrev_blocked.shot.for", "typeDescKeyPrev_takeaway.for", 
    "typeDescKeyPrev_lost.faceoff", "typeDescKeyPrev_post.missed.shot.for", 
    "typeDescKeyPrev_deflected.shot.on.goal.against", "typeDescKeyPrev_slap.shot.on.goal.for", 
    "typeDescKeyPrev_other.shot.on.goal.for", "typeDescKeyPrev_high.missed.shot.for", 
    "typeDescKeyPrev_snap.shot.on.goal.against", "typeDescKeyPrev_deflected.shot.on.goal.for", 
    "typeDescKeyPrev_giveaway.against", "typeDescKeyPrev_slap.shot.on.goal.against", 
    "typeDescKeyPrev_other.missed.shot.for", "typeDescKeyPrev_giveaway.for", 
    "typeDescKeyPrev_taken.hit", "typeDescKeyPrev_other.shot.on.goal.against", 
    "typeDescKeyPrev_tip.in.shot.on.goal.for", "typeDescKeyPrev_wrist.shot.on.goal.against", 
    "typeDescKeyPrev_other.missed.shot.against", "typeDescKeyPrev_blocked.shot.against", 
    "typeDescKeyPrev_snap.shot.on.goal.for", "typeDescKeyPrev_wide.missed.shot.against", 
    "typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_high.missed.shot.against", 
    "typeDescKeyPrev_given.hit", "typeDescKeyPrev_takeaway.against", 
    "typeDescKeyPrev_wide.missed.shot.for", "typeDescKeyPrev_backhand.shot.on.goal.for"
    ), zoneCode = c("zoneCode_O", "zoneCode_N")), dummy_term_to_var = c(isPlayoff_yes = "isPlayoff", 
isPlayoff_unknown = "isPlayoff", isPlayoff_new = "isPlayoff", 
isHome_yes = "isHome", isHome_unknown = "isHome", isHome_new = "isHome", 
isOvertime_yes = "isOvertime", isOvertime_unknown = "isOvertime", 
isOvertime_new = "isOvertime", zoneCode_N = "zoneCode", zoneCode_O = "zoneCode", 
zoneCode_unknown = "zoneCode", zoneCode_new = "zoneCode", isBehindNet_yes = "isBehindNet", 
isBehindNet_unknown = "isBehindNet", isBehindNet_new = "isBehindNet", 
crossedRoyalRoad_yes = "crossedRoyalRoad", crossedRoyalRoad_unknown = "crossedRoyalRoad", 
crossedRoyalRoad_new = "crossedRoyalRoad", typeDescKeyPrev_backhand.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_blocked.shot.against = "typeDescKeyPrev", typeDescKeyPrev_blocked.shot.for = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_giveaway.against = "typeDescKeyPrev", typeDescKeyPrev_giveaway.for = "typeDescKeyPrev", 
typeDescKeyPrev_given.hit = "typeDescKeyPrev", typeDescKeyPrev_high.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_high.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_lost.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_other.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_post.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_post.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_slap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_slap.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_snap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_snap.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_takeaway.against = "typeDescKeyPrev", 
typeDescKeyPrev_takeaway.for = "typeDescKeyPrev", typeDescKeyPrev_taken.hit = "typeDescKeyPrev", 
typeDescKeyPrev_tip.in.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_tip.in.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_won.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_unknown = "typeDescKeyPrev", 
typeDescKeyPrev_new = "typeDescKeyPrev", shotType_deflected = "shotType", 
shotType_other = "shotType", shotType_slap = "shotType", shotType_snap = "shotType", 
shotType_tip.in = "shotType", shotType_wrist = "shotType", shotType_unknown = "shotType", 
shotType_new = "shotType", isRebound_yes = "isRebound", isRebound_unknown = "isRebound", 
isRebound_new = "isRebound", isRush_yes = "isRush", isRush_unknown = "isRush", 
isRush_new = "isRush", shooterHandCode_R = "shooterHandCode", 
shooterHandCode_unknown = "shooterHandCode", shooterHandCode_new = "shooterHandCode", 
shooterPositionCode_D = "shooterPositionCode", shooterPositionCode_L = "shooterPositionCode", 
shooterPositionCode_R = "shooterPositionCode", shooterPositionCode_unknown = "shooterPositionCode", 
shooterPositionCode_new = "shooterPositionCode", goalieHandCode_R = "goalieHandCode", 
goalieHandCode_unknown = "goalieHandCode", goalieHandCode_new = "goalieHandCode", 
isEmptyNetFor_yes = "isEmptyNetFor", isEmptyNetFor_unknown = "isEmptyNetFor", 
isEmptyNetFor_new = "isEmptyNetFor", strengthState_unknown = "strengthState", 
strengthState_new = "strengthState"), dummy_term_to_level = c(isPlayoff_yes = "yes", 
isPlayoff_unknown = "unknown", isPlayoff_new = "new", isHome_yes = "yes", 
isHome_unknown = "unknown", isHome_new = "new", isOvertime_yes = "yes", 
isOvertime_unknown = "unknown", isOvertime_new = "new", zoneCode_N = "N", 
zoneCode_O = "O", zoneCode_unknown = "unknown", zoneCode_new = "new", 
isBehindNet_yes = "yes", isBehindNet_unknown = "unknown", isBehindNet_new = "new", 
crossedRoyalRoad_yes = "yes", crossedRoyalRoad_unknown = "unknown", 
crossedRoyalRoad_new = "new", typeDescKeyPrev_backhand.shot.on.goal.for = "backhand-shot-on-goal-for", 
typeDescKeyPrev_blocked.shot.against = "blocked-shot-against", 
typeDescKeyPrev_blocked.shot.for = "blocked-shot-for", typeDescKeyPrev_deflected.shot.on.goal.against = "deflected-shot-on-goal-against", 
typeDescKeyPrev_deflected.shot.on.goal.for = "deflected-shot-on-goal-for", 
typeDescKeyPrev_giveaway.against = "giveaway-against", typeDescKeyPrev_giveaway.for = "giveaway-for", 
typeDescKeyPrev_given.hit = "given-hit", typeDescKeyPrev_high.missed.shot.against = "high-missed-shot-against", 
typeDescKeyPrev_high.missed.shot.for = "high-missed-shot-for", 
typeDescKeyPrev_lost.faceoff = "lost-faceoff", typeDescKeyPrev_other.missed.shot.against = "other-missed-shot-against", 
typeDescKeyPrev_other.missed.shot.for = "other-missed-shot-for", 
typeDescKeyPrev_other.shot.on.goal.against = "other-shot-on-goal-against", 
typeDescKeyPrev_other.shot.on.goal.for = "other-shot-on-goal-for", 
typeDescKeyPrev_post.missed.shot.against = "post-missed-shot-against", 
typeDescKeyPrev_post.missed.shot.for = "post-missed-shot-for", 
typeDescKeyPrev_slap.shot.on.goal.against = "slap-shot-on-goal-against", 
typeDescKeyPrev_slap.shot.on.goal.for = "slap-shot-on-goal-for", 
typeDescKeyPrev_snap.shot.on.goal.against = "snap-shot-on-goal-against", 
typeDescKeyPrev_snap.shot.on.goal.for = "snap-shot-on-goal-for", 
typeDescKeyPrev_takeaway.against = "takeaway-against", typeDescKeyPrev_takeaway.for = "takeaway-for", 
typeDescKeyPrev_taken.hit = "taken-hit", typeDescKeyPrev_tip.in.shot.on.goal.against = "tip-in-shot-on-goal-against", 
typeDescKeyPrev_tip.in.shot.on.goal.for = "tip-in-shot-on-goal-for", 
typeDescKeyPrev_wide.missed.shot.against = "wide-missed-shot-against", 
typeDescKeyPrev_wide.missed.shot.for = "wide-missed-shot-for", 
typeDescKeyPrev_won.faceoff = "won-faceoff", typeDescKeyPrev_wrist.shot.on.goal.against = "wrist-shot-on-goal-against", 
typeDescKeyPrev_wrist.shot.on.goal.for = "wrist-shot-on-goal-for", 
typeDescKeyPrev_unknown = "unknown", typeDescKeyPrev_new = "new", 
shotType_deflected = "deflected", shotType_other = "other", shotType_slap = "slap", 
shotType_snap = "snap", shotType_tip.in = "tip-in", shotType_wrist = "wrist", 
shotType_unknown = "unknown", shotType_new = "new", isRebound_yes = "yes", 
isRebound_unknown = "unknown", isRebound_new = "new", isRush_yes = "yes", 
isRush_unknown = "unknown", isRush_new = "new", shooterHandCode_R = "R", 
shooterHandCode_unknown = "unknown", shooterHandCode_new = "new", 
shooterPositionCode_D = "D", shooterPositionCode_L = "L", shooterPositionCode_R = "R", 
shooterPositionCode_unknown = "unknown", shooterPositionCode_new = "new", 
goalieHandCode_R = "R", goalieHandCode_unknown = "unknown", goalieHandCode_new = "new", 
isEmptyNetFor_yes = "yes", isEmptyNetFor_unknown = "unknown", 
isEmptyNetFor_new = "new", strengthState_unknown = "unknown", 
strengthState_new = "new"), dummy_map = list(crossedRoyalRoad = c(yes = "crossedRoyalRoad_yes", 
unknown = "crossedRoyalRoad_unknown", new = "crossedRoyalRoad_new"
), goalieHandCode = c(R = "goalieHandCode_R", unknown = "goalieHandCode_unknown", 
new = "goalieHandCode_new"), isBehindNet = c(yes = "isBehindNet_yes", 
unknown = "isBehindNet_unknown", new = "isBehindNet_new"), isEmptyNetFor = c(yes = "isEmptyNetFor_yes", 
unknown = "isEmptyNetFor_unknown", new = "isEmptyNetFor_new"), 
    isHome = c(yes = "isHome_yes", unknown = "isHome_unknown", 
    new = "isHome_new"), isOvertime = c(yes = "isOvertime_yes", 
    unknown = "isOvertime_unknown", new = "isOvertime_new"), 
    isPlayoff = c(yes = "isPlayoff_yes", unknown = "isPlayoff_unknown", 
    new = "isPlayoff_new"), isRebound = c(yes = "isRebound_yes", 
    unknown = "isRebound_unknown", new = "isRebound_new"), isRush = c(yes = "isRush_yes", 
    unknown = "isRush_unknown", new = "isRush_new"), shooterHandCode = c(R = "shooterHandCode_R", 
    unknown = "shooterHandCode_unknown", new = "shooterHandCode_new"
    ), shooterPositionCode = c(D = "shooterPositionCode_D", L = "shooterPositionCode_L", 
    R = "shooterPositionCode_R", unknown = "shooterPositionCode_unknown", 
    new = "shooterPositionCode_new"), shotType = c(deflected = "shotType_deflected", 
    other = "shotType_other", slap = "shotType_slap", snap = "shotType_snap", 
    `tip-in` = "shotType_tip.in", wrist = "shotType_wrist", unknown = "shotType_unknown", 
    new = "shotType_new"), strengthState = c(unknown = "strengthState_unknown", 
    new = "strengthState_new"), typeDescKeyPrev = c(`backhand-shot-on-goal-for` = "typeDescKeyPrev_backhand.shot.on.goal.for", 
    `blocked-shot-against` = "typeDescKeyPrev_blocked.shot.against", 
    `blocked-shot-for` = "typeDescKeyPrev_blocked.shot.for", 
    `deflected-shot-on-goal-against` = "typeDescKeyPrev_deflected.shot.on.goal.against", 
    `deflected-shot-on-goal-for` = "typeDescKeyPrev_deflected.shot.on.goal.for", 
    `giveaway-against` = "typeDescKeyPrev_giveaway.against", 
    `giveaway-for` = "typeDescKeyPrev_giveaway.for", `given-hit` = "typeDescKeyPrev_given.hit", 
    `high-missed-shot-against` = "typeDescKeyPrev_high.missed.shot.against", 
    `high-missed-shot-for` = "typeDescKeyPrev_high.missed.shot.for", 
    `lost-faceoff` = "typeDescKeyPrev_lost.faceoff", `other-missed-shot-against` = "typeDescKeyPrev_other.missed.shot.against", 
    `other-missed-shot-for` = "typeDescKeyPrev_other.missed.shot.for", 
    `other-shot-on-goal-against` = "typeDescKeyPrev_other.shot.on.goal.against", 
    `other-shot-on-goal-for` = "typeDescKeyPrev_other.shot.on.goal.for", 
    `post-missed-shot-against` = "typeDescKeyPrev_post.missed.shot.against", 
    `post-missed-shot-for` = "typeDescKeyPrev_post.missed.shot.for", 
    `slap-shot-on-goal-against` = "typeDescKeyPrev_slap.shot.on.goal.against", 
    `slap-shot-on-goal-for` = "typeDescKeyPrev_slap.shot.on.goal.for", 
    `snap-shot-on-goal-against` = "typeDescKeyPrev_snap.shot.on.goal.against", 
    `snap-shot-on-goal-for` = "typeDescKeyPrev_snap.shot.on.goal.for", 
    `takeaway-against` = "typeDescKeyPrev_takeaway.against", 
    `takeaway-for` = "typeDescKeyPrev_takeaway.for", `taken-hit` = "typeDescKeyPrev_taken.hit", 
    `tip-in-shot-on-goal-against` = "typeDescKeyPrev_tip.in.shot.on.goal.against", 
    `tip-in-shot-on-goal-for` = "typeDescKeyPrev_tip.in.shot.on.goal.for", 
    `wide-missed-shot-against` = "typeDescKeyPrev_wide.missed.shot.against", 
    `wide-missed-shot-for` = "typeDescKeyPrev_wide.missed.shot.for", 
    `won-faceoff` = "typeDescKeyPrev_won.faceoff", `wrist-shot-on-goal-against` = "typeDescKeyPrev_wrist.shot.on.goal.against", 
    `wrist-shot-on-goal-for` = "typeDescKeyPrev_wrist.shot.on.goal.for", 
    unknown = "typeDescKeyPrev_unknown", new = "typeDescKeyPrev_new"
    ), zoneCode = c(N = "zoneCode_N", O = "zoneCode_O", unknown = "zoneCode_unknown", 
    new = "zoneCode_new")), known_levels = list(isPlayoff = c("no", 
"yes"), isHome = c("no", "yes"), isOvertime = c("no", "yes"), 
    zoneCode = c("D", "N", "O"), isBehindNet = c("no", "yes"), 
    crossedRoyalRoad = c("no", "yes"), typeDescKeyPrev = c("backhand-shot-on-goal-against", 
    "backhand-shot-on-goal-for", "blocked-shot-against", "blocked-shot-for", 
    "deflected-shot-on-goal-against", "deflected-shot-on-goal-for", 
    "giveaway-against", "giveaway-for", "given-hit", "high-missed-shot-against", 
    "high-missed-shot-for", "lost-faceoff", "other-missed-shot-against", 
    "other-missed-shot-for", "other-shot-on-goal-against", "other-shot-on-goal-for", 
    "post-missed-shot-against", "post-missed-shot-for", "slap-shot-on-goal-against", 
    "slap-shot-on-goal-for", "snap-shot-on-goal-against", "snap-shot-on-goal-for", 
    "takeaway-against", "takeaway-for", "taken-hit", "tip-in-shot-on-goal-against", 
    "tip-in-shot-on-goal-for", "wide-missed-shot-against", "wide-missed-shot-for", 
    "won-faceoff", "wrist-shot-on-goal-against", "wrist-shot-on-goal-for"
    ), shotType = c("backhand", "deflected", "other", "slap", 
    "snap", "tip-in", "wrist"), isRebound = c("no", "yes"), isRush = c("no", 
    "yes"), shooterHandCode = c("L", "R"), shooterPositionCode = c("C", 
    "D", "L", "R"), goalieHandCode = c("L", "R"), isEmptyNetFor = c("no", 
    "yes"), strengthState = "penalty-kill"), logical_vars = c("crossedRoyalRoad", 
"isBehindNet", "isEmptyNetFor", "isHome", "isOvertime", "isPlayoff", 
"isRebound", "isRush"), has_unknown = c(crossedRoyalRoad = TRUE, 
goalieHandCode = TRUE, isBehindNet = TRUE, isEmptyNetFor = TRUE, 
isHome = TRUE, isOvertime = TRUE, isPlayoff = TRUE, isRebound = TRUE, 
isRush = TRUE, shooterHandCode = TRUE, shooterPositionCode = TRUE, 
shotType = TRUE, strengthState = TRUE, typeDescKeyPrev = TRUE, 
zoneCode = TRUE), has_new = c(crossedRoyalRoad = TRUE, goalieHandCode = TRUE, 
isBehindNet = TRUE, isEmptyNetFor = TRUE, isHome = TRUE, isOvertime = TRUE, 
isPlayoff = TRUE, isRebound = TRUE, isRush = TRUE, shooterHandCode = TRUE, 
shooterPositionCode = TRUE, shotType = TRUE, strengthState = TRUE, 
typeDescKeyPrev = TRUE, zoneCode = TRUE), medians = c(distance = 36.1247837363769, 
xCoordNorm = 59, angle = 21.8014094863518, minSecondsElapsedInShiftFor = 12, 
dDistance = -8.43579169116009, shotType_tip.in = 0, shotType_snap = 0, 
zoneCode_O = 1, typeDescKeyPrev_tip.in.shot.on.goal.against = 0, 
typeDescKeyPrev_won.faceoff = 0, avgSecondsElapsedInShiftAgainst = 40.4, 
typeDescKeyPrev_post.missed.shot.against = 0, zoneCode_N = 0, 
minSecondsElapsedSinceLastShiftFor = 62, avgSecondsElapsedSinceLastShiftFor = 119.5, 
secondsElapsedInGame = 1815, goalieHeight = 75, dXCoordNorm = 7, 
corsiDifferential = 0, isOvertime_yes = 0, shooterHandCode_R = 0, 
shooterPositionCode_D = 0, fenwickDifferential = 0, typeDescKeyPrev_blocked.shot.for = 0, 
manDifferential = -1, typeDescKeyPrev_takeaway.for = 0, dDistancePerSecond = -0.57176654128078, 
typeDescKeyPrev_lost.faceoff = 0, shooterSecondsElapsedSinceLastShift = 113, 
typeDescKeyPrev_post.missed.shot.for = 0, shooterWeight = 201, 
periodNumber = 2, dAngle = 1.43318147538709, goalieWeight = 201, 
dXCoordNormPerSecond = 0.454545454545455, shooterHeight = 73, 
goalieAge = 28, shooterSecondsElapsedInShift = 24, dYCoordNormPerSecond = 0.0535714285714286, 
corsiAgainst = 29, secondsElapsedInPeriod = 616, shooterPositionCode_R = 0, 
maxSecondsElapsedSinceLastShiftAgainst = 220, isRebound_yes = 0, 
shotType_other = 0, typeDescKeyPrev_deflected.shot.on.goal.against = 0, 
dYCoordNorm = 1, typeDescKeyPrev_slap.shot.on.goal.for = 0, typeDescKeyPrev_other.shot.on.goal.for = 0, 
dSecondsElapsedInSequence = 9, typeDescKeyPrev_high.missed.shot.for = 0, 
avgSecondsElapsedInShiftFor = 23.75, typeDescKeyPrev_snap.shot.on.goal.against = 0, 
maxSecondsElapsedInShiftFor = 35, crossedRoyalRoad_yes = 0, typeDescKeyPrev_deflected.shot.on.goal.for = 0, 
fenwickFor = 21, typeDescKeyPrev_giveaway.against = 0, typeDescKeyPrev_slap.shot.on.goal.against = 0, 
typeDescKeyPrev_other.missed.shot.for = 0, shotsFor = 14, fenwickAgainst = 21, 
maxSecondsElapsedSinceLastShiftFor = 183, skaterCountFor = 4, 
isHome_yes = 0, shotType_slap = 0, typeDescKeyPrev_giveaway.for = 0, 
typeDescKeyPrev_taken.hit = 0, skaterCountAgainst = 5, typeDescKeyPrev_other.shot.on.goal.against = 0, 
minSecondsElapsedSinceLastShiftAgainst = 89, typeDescKeyPrev_tip.in.shot.on.goal.for = 0, 
typeDescKeyPrev_wrist.shot.on.goal.against = 0, isBehindNet_yes = 0, 
shooterPositionCode_L = 0, yCoordNorm = 1, typeDescKeyPrev_other.missed.shot.against = 0, 
typeDescKeyPrev_blocked.shot.against = 0, goalsFor = 1, maxSecondsElapsedInShiftAgainst = 57, 
typeDescKeyPrev_snap.shot.on.goal.for = 0, isRush_yes = 0, minSecondsElapsedInShiftAgainst = 24, 
shotsAgainst = 14, avgSecondsElapsedSinceLastShiftAgainst = 150.8, 
secondsElapsedInSequence = 39, shotDifferential = 0, typeDescKeyPrev_wide.missed.shot.against = 0, 
shotType_deflected = 0, shotType_wrist = 1, goalsAgainst = 1, 
typeDescKeyPrev_wrist.shot.on.goal.for = 0, dAnglePerSecond = 0.1090902697479, 
shooterAge = 28, typeDescKeyPrev_high.missed.shot.against = 0, 
isEmptyNetFor_yes = 0, goalieHandCode_R = 0, typeDescKeyPrev_given.hit = 0, 
goalDifferential = 0, corsiFor = 29, isPlayoff_yes = 0, typeDescKeyPrev_takeaway.against = 0, 
typeDescKeyPrev_wide.missed.shot.for = 0, typeDescKeyPrev_backhand.shot.on.goal.for = 0
), normalize_means = c(distance = 63.1809537417715, xCoordNorm = 30.3849470091611, 
angle = 25.9799070157217, minSecondsElapsedInShiftFor = 13.9951499910185, 
dDistance = -15.168903996702, shotType_tip.in = 0.0226333752469912, 
shotType_snap = 0.152326208011496, zoneCode_O = 0.711514280582001, 
typeDescKeyPrev_tip.in.shot.on.goal.against = 0.00952038800071852, 
typeDescKeyPrev_won.faceoff = 0.216992994431471, avgSecondsElapsedInShiftAgainst = 43.4482126818753, 
typeDescKeyPrev_post.missed.shot.against = 0.00844260822705227, 
zoneCode_N = 0.0506556493623136, minSecondsElapsedSinceLastShiftFor = 81.1692114244656, 
avgSecondsElapsedSinceLastShiftFor = 137.533923118376, secondsElapsedInGame = 1817.75678103108, 
goalieHeight = 74.9342554338064, dXCoordNorm = 14.591880725705, 
corsiDifferential = -0.10185018861146, isOvertime_yes = 0.00359259924555416, 
shooterHandCode_R = 0.364109933536914, shooterPositionCode_D = 0.267828273756062, 
fenwickDifferential = -0.124483563858452, typeDescKeyPrev_blocked.shot.for = 0.0111370576612179, 
manDifferential = -1.00467037901922, typeDescKeyPrev_takeaway.for = 0.0934075803844081, 
dDistancePerSecond = -1.2452145492716, typeDescKeyPrev_lost.faceoff = 0.105263157894737, 
shooterSecondsElapsedSinceLastShift = 144.58828812646, typeDescKeyPrev_post.missed.shot.for = 0.00287407939644333, 
shooterWeight = 201.8785701455, periodNumber = 2.00125740973594, 
dAngle = 1.01807593296847, goalieWeight = 201.932459134184, dXCoordNormPerSecond = 0.989775309528442, 
shooterHeight = 73.4576971438836, goalieAge = 28.5611640021556, 
shooterSecondsElapsedInShift = 28.0046703790192, dYCoordNormPerSecond = 0.20206839262364, 
corsiAgainst = 30.1780132926172, secondsElapsedInPeriod = 616.247889347943, 
shooterPositionCode_R = 0.143703969822166, maxSecondsElapsedSinceLastShiftAgainst = 243.927788755164, 
isRebound_yes = 0.0465241602299264, shotType_other = 0.0113166876234956, 
typeDescKeyPrev_deflected.shot.on.goal.against = 0.00125740973594396, 
dYCoordNorm = 0.985270343093228, typeDescKeyPrev_slap.shot.on.goal.for = 0.0021555595473325, 
typeDescKeyPrev_other.shot.on.goal.for = 0.000538889886833124, 
dSecondsElapsedInSequence = 15.5194898509071, typeDescKeyPrev_high.missed.shot.for = 0.00089814981138854, 
avgSecondsElapsedInShiftFor = 25.578025267948, typeDescKeyPrev_snap.shot.on.goal.against = 0.0176037363032154, 
maxSecondsElapsedInShiftFor = 39.9904796119993, crossedRoyalRoad_yes = 0.437398958146219, 
typeDescKeyPrev_deflected.shot.on.goal.for = 0.000538889886833124, 
fenwickFor = 21.6030177833663, typeDescKeyPrev_giveaway.against = 0.128255793066283, 
typeDescKeyPrev_slap.shot.on.goal.against = 0.0197592958505479, 
typeDescKeyPrev_other.missed.shot.for = 0.000359259924555416, 
shotsFor = 14.6671456798994, fenwickAgainst = 21.7275013472247, 
maxSecondsElapsedSinceLastShiftFor = 208.098796479253, skaterCountFor = 3.99335369139572, 
isHome_yes = 0.486617567810311, shotType_slap = 0.0635890066463086, 
typeDescKeyPrev_giveaway.for = 0.014729656906772, typeDescKeyPrev_taken.hit = 0.0219148553978804, 
skaterCountAgainst = 4.99766481049039, typeDescKeyPrev_other.shot.on.goal.against = 0.00197592958505479, 
minSecondsElapsedSinceLastShiftAgainst = 105.816777438477, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.00251481947188791, 
typeDescKeyPrev_wrist.shot.on.goal.against = 0.0441889707203161, 
isBehindNet_yes = 0.0113166876234956, shooterPositionCode_L = 0.172265133824322, 
yCoordNorm = 0.92635171546614, typeDescKeyPrev_other.missed.shot.against = 0.0044907490569427, 
typeDescKeyPrev_blocked.shot.against = 0.0946649901203521, goalsFor = 1.37039698221663, 
maxSecondsElapsedInShiftAgainst = 62.3698580923298, typeDescKeyPrev_snap.shot.on.goal.for = 0.00754445841566373, 
isRush_yes = 0.084246452308245, minSecondsElapsedInShiftAgainst = 28.7937848033052, 
shotsAgainst = 14.7253457876774, avgSecondsElapsedSinceLastShiftAgainst = 168.054158433627, 
secondsElapsedInSequence = 44.1534039877852, shotDifferential = -0.0582001077779774, 
typeDescKeyPrev_wide.missed.shot.against = 0.0528112089096461, 
shotType_deflected = 0.00700556852883061, shotType_wrist = 0.638045626010419, 
goalsAgainst = 1.34506915753548, typeDescKeyPrev_wrist.shot.on.goal.for = 0.0440093407580384, 
dAnglePerSecond = 0.00429356766359049, shooterAge = 28.4118915035028, 
typeDescKeyPrev_high.missed.shot.against = 0.00395185917010957, 
isEmptyNetFor_yes = 0.000359259924555416, goalieHandCode_R = 0.0574815879288665, 
typeDescKeyPrev_given.hit = 0.0481408298904257, goalDifferential = 0.0253278246811568, 
corsiFor = 30.0761631040057, isPlayoff_yes = 0.0610741871744207, 
typeDescKeyPrev_takeaway.against = 0.0111370576612179, typeDescKeyPrev_wide.missed.shot.for = 0.0176037363032154, 
typeDescKeyPrev_backhand.shot.on.goal.for = 0.0068259385665529
), normalize_sds = c(distance = 58.6677538893601, xCoordNorm = 59.8407427368433, 
angle = 21.1496411383108, minSecondsElapsedInShiftFor = 12.8123838893787, 
dDistance = 76.8631528193355, shotType_tip.in = 0.148745016384372, 
shotType_snap = 0.35936907613108, zoneCode_O = 0.453098870984346, 
typeDescKeyPrev_tip.in.shot.on.goal.against = 0.0971156237844054, 
typeDescKeyPrev_won.faceoff = 0.412234836806347, avgSecondsElapsedInShiftAgainst = 25.1049188403975, 
typeDescKeyPrev_post.missed.shot.against = 0.0915031945089388, 
zoneCode_N = 0.21931323362902, minSecondsElapsedSinceLastShiftFor = 70.7051219197829, 
avgSecondsElapsedSinceLastShiftFor = 73.5945589068384, secondsElapsedInGame = 938.507272145994, 
goalieHeight = 1.63295304921384, dXCoordNorm = 79.1491366609479, 
corsiDifferential = 12.8502475815637, isOvertime_yes = 0.059835905706905, 
shooterHandCode_R = 0.481222908570225, shooterPositionCode_D = 0.442867384925269, 
fenwickDifferential = 9.63784324618641, typeDescKeyPrev_blocked.shot.for = 0.104952380784357, 
manDifferential = 0.0681865217685952, typeDescKeyPrev_takeaway.for = 0.291028896464069, 
dDistancePerSecond = 29.490560588503, typeDescKeyPrev_lost.faceoff = 0.306919772223429, 
shooterSecondsElapsedSinceLastShift = 103.38073003469, typeDescKeyPrev_post.missed.shot.for = 0.0535381540914937, 
shooterWeight = 14.9852359573519, periodNumber = 0.778920911406041, 
dAngle = 30.9008544780758, goalieWeight = 15.4345639522836, dXCoordNormPerSecond = 30.8451887863004, 
shooterHeight = 2.09270363261508, goalieAge = 3.69529626080676, 
shooterSecondsElapsedInShift = 19.4431733162066, dYCoordNormPerSecond = 10.365677123193, 
corsiAgainst = 17.0954633408751, secondsElapsedInPeriod = 337.523589602319, 
shooterPositionCode_R = 0.350820818768588, maxSecondsElapsedSinceLastShiftAgainst = 110.176783572376, 
isRebound_yes = 0.210636256378092, shotType_other = 0.105785775873895, 
typeDescKeyPrev_deflected.shot.on.goal.against = 0.0354408561082151, 
dYCoordNorm = 29.5408047182312, typeDescKeyPrev_slap.shot.on.goal.for = 0.0463821037494733, 
typeDescKeyPrev_other.shot.on.goal.for = 0.0232098309022316, 
dSecondsElapsedInSequence = 16.4104273669405, typeDescKeyPrev_high.missed.shot.for = 0.029958377075181, 
avgSecondsElapsedInShiftFor = 14.0698717367964, typeDescKeyPrev_snap.shot.on.goal.against = 0.131517876435811, 
maxSecondsElapsedInShiftFor = 25.9106736061195, crossedRoyalRoad_yes = 0.496110190422404, 
typeDescKeyPrev_deflected.shot.on.goal.for = 0.0232098309022309, 
fenwickFor = 12.9887916336736, typeDescKeyPrev_giveaway.against = 0.334404443707038, 
typeDescKeyPrev_slap.shot.on.goal.against = 0.139184575049848, 
typeDescKeyPrev_other.missed.shot.for = 0.0189524504778257, shotsFor = 9.02788649845418, 
fenwickAgainst = 12.5672436343953, maxSecondsElapsedSinceLastShiftFor = 110.294115393318, 
skaterCountFor = 0.106181634243063, isHome_yes = 0.499865775875685, 
shotType_slap = 0.244041273041567, typeDescKeyPrev_giveaway.for = 0.120479465043808, 
typeDescKeyPrev_taken.hit = 0.146418733425504, skaterCountAgainst = 0.0857945076631566, 
typeDescKeyPrev_other.shot.on.goal.against = 0.0444114803378586, 
minSecondsElapsedSinceLastShiftAgainst = 76.9201922404384, typeDescKeyPrev_tip.in.shot.on.goal.for = 0.0500893784824438, 
typeDescKeyPrev_wrist.shot.on.goal.against = 0.205533194049181, 
isBehindNet_yes = 0.105785775873894, shooterPositionCode_L = 0.377644641833166, 
yCoordNorm = 20.4044349038331, typeDescKeyPrev_other.missed.shot.against = 0.0668684187389328, 
typeDescKeyPrev_blocked.shot.against = 0.292777949056241, goalsFor = 1.41473696227119, 
maxSecondsElapsedInShiftAgainst = 36.6084679402859, typeDescKeyPrev_snap.shot.on.goal.for = 0.0865383428947544, 
isRush_yes = 0.27778201587083, minSecondsElapsedInShiftAgainst = 23.6446141914994, 
shotsAgainst = 8.83447675113178, avgSecondsElapsedSinceLastShiftAgainst = 76.0260045129353, 
secondsElapsedInSequence = 32.9875437705848, shotDifferential = 7.02138746912391, 
typeDescKeyPrev_wide.missed.shot.against = 0.223676490093908, 
shotType_deflected = 0.0834130706607534, shotType_wrist = 0.480608881469885, 
goalsAgainst = 1.38482368618566, typeDescKeyPrev_wrist.shot.on.goal.for = 0.205134291445436, 
dAnglePerSecond = 12.9392293300766, shooterAge = 3.83247214686059, 
typeDescKeyPrev_high.missed.shot.against = 0.0627451127438685, 
isEmptyNetFor_yes = 0.0189524504778259, goalieHandCode_R = 0.232781418117279, 
typeDescKeyPrev_given.hit = 0.214082981814658, goalDifferential = 1.7152393806928, 
corsiFor = 17.8397853770828, isPlayoff_yes = 0.239487856498219, 
typeDescKeyPrev_takeaway.against = 0.104952380784357, typeDescKeyPrev_wide.missed.shot.for = 0.131517876435809, 
typeDescKeyPrev_backhand.shot.on.goal.for = 0.0823441747881165
)), en = list(coefficients = c(`(Intercept)` = 0.358992590226874, 
distance = -0.339259719257798, angle = -0.322322770061326, xCoordNorm = 0.234621243387866, 
zoneCode_O = 0.21534584416679, shotType_wrist = 0.181597924633319, 
dDistancePerSecond = -0.153153415641791, typeDescKeyPrev_takeaway.against = -0.138881176805622, 
shotType_snap = 0.138188123096424, dDistance = -0.13801366597764, 
corsiFor = 0.135638265681385, shotType_other = 0.129322136271426, 
typeDescKeyPrev_blocked.shot.for = -0.128445353256351, dAngle = -0.125034687868652, 
typeDescKeyPrev_blocked.shot.against = 0.110808240565555, dYCoordNormPerSecond = -0.104422517312544, 
dXCoordNormPerSecond = 0.09845469169962, isRush_yes = -0.0971751230311377, 
shotType_deflected = 0.0927685209789503, typeDescKeyPrev_high.missed.shot.against = 0.0884401808524156, 
typeDescKeyPrev_takeaway.for = 0.0881819866655735, minSecondsElapsedInShiftFor = 0.072896424098742, 
typeDescKeyPrev_high.missed.shot.for = -0.0718080015191672, isHome_yes = -0.0712966389442882, 
zoneCode_N = -0.0687102939677219, maxSecondsElapsedInShiftAgainst = 0.0684352635817362, 
maxSecondsElapsedInShiftFor = 0.0672820377865518, isPlayoff_yes = 0.0671836564327116, 
typeDescKeyPrev_deflected.shot.on.goal.against = 0.0671695907051486, 
shotType_slap = 0.0657160014861867, crossedRoyalRoad_yes = -0.061731021473829, 
shooterPositionCode_R = 0.0598889101623524, shooterPositionCode_L = -0.0583103923007061, 
dXCoordNorm = 0.0566962948009291, corsiDifferential = 0.0566395103440728, 
avgSecondsElapsedInShiftFor = 0.0553996501594768, secondsElapsedInPeriod = -0.054318403297663, 
shooterWeight = 0.0495921017692478, typeDescKeyPrev_given.hit = -0.0476581923611311, 
secondsElapsedInGame = -0.0443119003674056, typeDescKeyPrev_other.shot.on.goal.against = -0.0425469800798999, 
skaterCountAgainst = 0.0412794646706844, typeDescKeyPrev_giveaway.against = 0.0395032838511861, 
fenwickFor = -0.0388249241472325, typeDescKeyPrev_post.missed.shot.for = 0.0377128045820563, 
isRebound_yes = -0.0363963007464233, corsiAgainst = 0.0338868984092933, 
avgSecondsElapsedSinceLastShiftFor = -0.0334931649244189, isOvertime_yes = 0.0332468981879563, 
periodNumber = 0.0332427613358308, typeDescKeyPrev_lost.faceoff = -0.0329052664237163, 
maxSecondsElapsedSinceLastShiftFor = 0.032385486765552, shooterAge = 0.0316700916540132, 
goalDifferential = -0.0316304951877376, fenwickDifferential = -0.0310776265813251, 
maxSecondsElapsedSinceLastShiftAgainst = 0.0309954611251061, 
dAnglePerSecond = 0.0300889688595502, shooterSecondsElapsedSinceLastShift = -0.0299043130556757, 
goalsAgainst = 0.028867370763428, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.0280193698069366, 
dYCoordNorm = 0.02704421220711, typeDescKeyPrev_won.faceoff = -0.0264136334899004, 
typeDescKeyPrev_wide.missed.shot.for = -0.0260846134307286, minSecondsElapsedInShiftAgainst = 0.0258828774869373, 
shooterHeight = -0.0254869214514688, typeDescKeyPrev_snap.shot.on.goal.against = -0.0254745741172278, 
manDifferential = -0.0253481208172621, minSecondsElapsedSinceLastShiftFor = 0.024985176822828, 
shooterPositionCode_D = 0.0237316157078179, shotsAgainst = -0.0229905732024914, 
typeDescKeyPrev_wide.missed.shot.against = -0.0227472174191103, 
typeDescKeyPrev_wrist.shot.on.goal.against = -0.0215450881694877, 
shotType_tip.in = -0.0197091411616381, shooterSecondsElapsedInShift = -0.0182158560675903, 
fenwickAgainst = 0.0144940715332119, strengthState_penalty.kill = -0.0129773384245651, 
secondsElapsedInSequence = -0.0123082172822855, typeDescKeyPrev_other.missed.shot.against = 0.0122735653613287, 
avgSecondsElapsedInShiftAgainst = 0.0116870089929426, yCoordNorm = 0.0113480484338671, 
typeDescKeyPrev_wrist.shot.on.goal.for = -0.0105914774885016, 
strengthState_power.play = 0.0102648585563287, shotsFor = -0.0101001219082209, 
isBehindNet_yes = 0.0100330680989801, typeDescKeyPrev_slap.shot.on.goal.against = 0.0099315727329714, 
goalsFor = 0.0095755716860291, minSecondsElapsedSinceLastShiftAgainst = -0.00887533754159351, 
shotDifferential = 0.00864494268147159, avgSecondsElapsedSinceLastShiftAgainst = -0.0073785340518477, 
shooterHandCode_R = -0.00442281829414478, skaterCountFor = 0.00254772706094163, 
typeDescKeyPrev_post.missed.shot.against = -0.0019381584795818, 
typeDescKeyPrev_giveaway.for = 0.00188063209656073, typeDescKeyPrev_backhand.shot.on.goal.for = -0.00187168118066526, 
dSecondsElapsedInSequence = 0.000997763377465499, typeDescKeyPrev_taken.hit = -0.00013632266227135
), raw_terms = c("distance", "angle", "xCoordNorm", "dDistancePerSecond", 
"dDistance", "corsiFor", "dAngle", "dYCoordNormPerSecond", "dXCoordNormPerSecond", 
"minSecondsElapsedInShiftFor", "maxSecondsElapsedInShiftAgainst", 
"maxSecondsElapsedInShiftFor", "dXCoordNorm", "corsiDifferential", 
"avgSecondsElapsedInShiftFor", "secondsElapsedInPeriod", "shooterWeight", 
"secondsElapsedInGame", "skaterCountAgainst", "fenwickFor", "corsiAgainst", 
"avgSecondsElapsedSinceLastShiftFor", "periodNumber", "maxSecondsElapsedSinceLastShiftFor", 
"shooterAge", "goalDifferential", "fenwickDifferential", "maxSecondsElapsedSinceLastShiftAgainst", 
"dAnglePerSecond", "shooterSecondsElapsedSinceLastShift", "goalsAgainst", 
"dYCoordNorm", "minSecondsElapsedInShiftAgainst", "shooterHeight", 
"manDifferential", "minSecondsElapsedSinceLastShiftFor", "shotsAgainst", 
"shooterSecondsElapsedInShift", "fenwickAgainst", "secondsElapsedInSequence", 
"avgSecondsElapsedInShiftAgainst", "yCoordNorm", "shotsFor", 
"goalsFor", "minSecondsElapsedSinceLastShiftAgainst", "shotDifferential", 
"avgSecondsElapsedSinceLastShiftAgainst", "skaterCountFor", "dSecondsElapsedInSequence"
), active_dummy_terms = c("zoneCode_O", "shotType_wrist", "typeDescKeyPrev_takeaway.against", 
"shotType_snap", "shotType_other", "typeDescKeyPrev_blocked.shot.for", 
"typeDescKeyPrev_blocked.shot.against", "isRush_yes", "shotType_deflected", 
"typeDescKeyPrev_high.missed.shot.against", "typeDescKeyPrev_takeaway.for", 
"typeDescKeyPrev_high.missed.shot.for", "isHome_yes", "zoneCode_N", 
"isPlayoff_yes", "typeDescKeyPrev_deflected.shot.on.goal.against", 
"shotType_slap", "crossedRoyalRoad_yes", "shooterPositionCode_R", 
"shooterPositionCode_L", "typeDescKeyPrev_given.hit", "typeDescKeyPrev_other.shot.on.goal.against", 
"typeDescKeyPrev_giveaway.against", "typeDescKeyPrev_post.missed.shot.for", 
"isRebound_yes", "isOvertime_yes", "typeDescKeyPrev_lost.faceoff", 
"typeDescKeyPrev_tip.in.shot.on.goal.against", "typeDescKeyPrev_won.faceoff", 
"typeDescKeyPrev_wide.missed.shot.for", "typeDescKeyPrev_snap.shot.on.goal.against", 
"shooterPositionCode_D", "typeDescKeyPrev_wide.missed.shot.against", 
"typeDescKeyPrev_wrist.shot.on.goal.against", "shotType_tip.in", 
"strengthState_penalty.kill", "typeDescKeyPrev_other.missed.shot.against", 
"typeDescKeyPrev_wrist.shot.on.goal.for", "strengthState_power.play", 
"isBehindNet_yes", "typeDescKeyPrev_slap.shot.on.goal.against", 
"shooterHandCode_R", "typeDescKeyPrev_post.missed.shot.against", 
"typeDescKeyPrev_giveaway.for", "typeDescKeyPrev_backhand.shot.on.goal.for", 
"typeDescKeyPrev_taken.hit"), active_dummy_by_var = list(crossedRoyalRoad = "crossedRoyalRoad_yes", 
    isBehindNet = "isBehindNet_yes", isHome = "isHome_yes", isOvertime = "isOvertime_yes", 
    isPlayoff = "isPlayoff_yes", isRebound = "isRebound_yes", 
    isRush = "isRush_yes", shooterHandCode = "shooterHandCode_R", 
    shooterPositionCode = c("shooterPositionCode_R", "shooterPositionCode_L", 
    "shooterPositionCode_D"), shotType = c("shotType_wrist", 
    "shotType_snap", "shotType_other", "shotType_deflected", 
    "shotType_slap", "shotType_tip.in"), strengthState = c("strengthState_penalty.kill", 
    "strengthState_power.play"), typeDescKeyPrev = c("typeDescKeyPrev_takeaway.against", 
    "typeDescKeyPrev_blocked.shot.for", "typeDescKeyPrev_blocked.shot.against", 
    "typeDescKeyPrev_high.missed.shot.against", "typeDescKeyPrev_takeaway.for", 
    "typeDescKeyPrev_high.missed.shot.for", "typeDescKeyPrev_deflected.shot.on.goal.against", 
    "typeDescKeyPrev_given.hit", "typeDescKeyPrev_other.shot.on.goal.against", 
    "typeDescKeyPrev_giveaway.against", "typeDescKeyPrev_post.missed.shot.for", 
    "typeDescKeyPrev_lost.faceoff", "typeDescKeyPrev_tip.in.shot.on.goal.against", 
    "typeDescKeyPrev_won.faceoff", "typeDescKeyPrev_wide.missed.shot.for", 
    "typeDescKeyPrev_snap.shot.on.goal.against", "typeDescKeyPrev_wide.missed.shot.against", 
    "typeDescKeyPrev_wrist.shot.on.goal.against", "typeDescKeyPrev_other.missed.shot.against", 
    "typeDescKeyPrev_wrist.shot.on.goal.for", "typeDescKeyPrev_slap.shot.on.goal.against", 
    "typeDescKeyPrev_post.missed.shot.against", "typeDescKeyPrev_giveaway.for", 
    "typeDescKeyPrev_backhand.shot.on.goal.for", "typeDescKeyPrev_taken.hit"
    ), zoneCode = c("zoneCode_O", "zoneCode_N")), dummy_term_to_var = c(isPlayoff_yes = "isPlayoff", 
isPlayoff_unknown = "isPlayoff", isPlayoff_new = "isPlayoff", 
isHome_yes = "isHome", isHome_unknown = "isHome", isHome_new = "isHome", 
isOvertime_yes = "isOvertime", isOvertime_unknown = "isOvertime", 
isOvertime_new = "isOvertime", zoneCode_N = "zoneCode", zoneCode_O = "zoneCode", 
zoneCode_unknown = "zoneCode", zoneCode_new = "zoneCode", isBehindNet_yes = "isBehindNet", 
isBehindNet_unknown = "isBehindNet", isBehindNet_new = "isBehindNet", 
crossedRoyalRoad_yes = "crossedRoyalRoad", crossedRoyalRoad_unknown = "crossedRoyalRoad", 
crossedRoyalRoad_new = "crossedRoyalRoad", typeDescKeyPrev_backhand.shot.on.goal.for = "typeDescKeyPrev", 
typeDescKeyPrev_blocked.shot.against = "typeDescKeyPrev", typeDescKeyPrev_blocked.shot.for = "typeDescKeyPrev", 
typeDescKeyPrev_deflected.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_giveaway.against = "typeDescKeyPrev", typeDescKeyPrev_giveaway.for = "typeDescKeyPrev", 
typeDescKeyPrev_given.hit = "typeDescKeyPrev", typeDescKeyPrev_high.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_high.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_lost.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_other.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_other.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_post.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_post.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_slap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_snap.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_takeaway.against = "typeDescKeyPrev", typeDescKeyPrev_takeaway.for = "typeDescKeyPrev", 
typeDescKeyPrev_taken.hit = "typeDescKeyPrev", typeDescKeyPrev_tip.in.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.against = "typeDescKeyPrev", 
typeDescKeyPrev_wide.missed.shot.for = "typeDescKeyPrev", typeDescKeyPrev_won.faceoff = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.against = "typeDescKeyPrev", 
typeDescKeyPrev_wrist.shot.on.goal.for = "typeDescKeyPrev", typeDescKeyPrev_unknown = "typeDescKeyPrev", 
typeDescKeyPrev_new = "typeDescKeyPrev", shotType_deflected = "shotType", 
shotType_other = "shotType", shotType_slap = "shotType", shotType_snap = "shotType", 
shotType_tip.in = "shotType", shotType_wrist = "shotType", shotType_unknown = "shotType", 
shotType_new = "shotType", isRebound_yes = "isRebound", isRebound_unknown = "isRebound", 
isRebound_new = "isRebound", isRush_yes = "isRush", isRush_unknown = "isRush", 
isRush_new = "isRush", shooterHandCode_R = "shooterHandCode", 
shooterHandCode_unknown = "shooterHandCode", shooterHandCode_new = "shooterHandCode", 
shooterPositionCode_D = "shooterPositionCode", shooterPositionCode_L = "shooterPositionCode", 
shooterPositionCode_R = "shooterPositionCode", shooterPositionCode_unknown = "shooterPositionCode", 
shooterPositionCode_new = "shooterPositionCode", isEmptyNetFor_yes = "isEmptyNetFor", 
isEmptyNetFor_unknown = "isEmptyNetFor", isEmptyNetFor_new = "isEmptyNetFor", 
strengthState_penalty.kill = "strengthState", strengthState_power.play = "strengthState", 
strengthState_unknown = "strengthState", strengthState_new = "strengthState"
), dummy_term_to_level = c(isPlayoff_yes = "yes", isPlayoff_unknown = "unknown", 
isPlayoff_new = "new", isHome_yes = "yes", isHome_unknown = "unknown", 
isHome_new = "new", isOvertime_yes = "yes", isOvertime_unknown = "unknown", 
isOvertime_new = "new", zoneCode_N = "N", zoneCode_O = "O", zoneCode_unknown = "unknown", 
zoneCode_new = "new", isBehindNet_yes = "yes", isBehindNet_unknown = "unknown", 
isBehindNet_new = "new", crossedRoyalRoad_yes = "yes", crossedRoyalRoad_unknown = "unknown", 
crossedRoyalRoad_new = "new", typeDescKeyPrev_backhand.shot.on.goal.for = "backhand-shot-on-goal-for", 
typeDescKeyPrev_blocked.shot.against = "blocked-shot-against", 
typeDescKeyPrev_blocked.shot.for = "blocked-shot-for", typeDescKeyPrev_deflected.shot.on.goal.against = "deflected-shot-on-goal-against", 
typeDescKeyPrev_giveaway.against = "giveaway-against", typeDescKeyPrev_giveaway.for = "giveaway-for", 
typeDescKeyPrev_given.hit = "given-hit", typeDescKeyPrev_high.missed.shot.against = "high-missed-shot-against", 
typeDescKeyPrev_high.missed.shot.for = "high-missed-shot-for", 
typeDescKeyPrev_lost.faceoff = "lost-faceoff", typeDescKeyPrev_other.missed.shot.against = "other-missed-shot-against", 
typeDescKeyPrev_other.shot.on.goal.against = "other-shot-on-goal-against", 
typeDescKeyPrev_post.missed.shot.against = "post-missed-shot-against", 
typeDescKeyPrev_post.missed.shot.for = "post-missed-shot-for", 
typeDescKeyPrev_slap.shot.on.goal.against = "slap-shot-on-goal-against", 
typeDescKeyPrev_snap.shot.on.goal.against = "snap-shot-on-goal-against", 
typeDescKeyPrev_takeaway.against = "takeaway-against", typeDescKeyPrev_takeaway.for = "takeaway-for", 
typeDescKeyPrev_taken.hit = "taken-hit", typeDescKeyPrev_tip.in.shot.on.goal.against = "tip-in-shot-on-goal-against", 
typeDescKeyPrev_wide.missed.shot.against = "wide-missed-shot-against", 
typeDescKeyPrev_wide.missed.shot.for = "wide-missed-shot-for", 
typeDescKeyPrev_won.faceoff = "won-faceoff", typeDescKeyPrev_wrist.shot.on.goal.against = "wrist-shot-on-goal-against", 
typeDescKeyPrev_wrist.shot.on.goal.for = "wrist-shot-on-goal-for", 
typeDescKeyPrev_unknown = "unknown", typeDescKeyPrev_new = "new", 
shotType_deflected = "deflected", shotType_other = "other", shotType_slap = "slap", 
shotType_snap = "snap", shotType_tip.in = "tip-in", shotType_wrist = "wrist", 
shotType_unknown = "unknown", shotType_new = "new", isRebound_yes = "yes", 
isRebound_unknown = "unknown", isRebound_new = "new", isRush_yes = "yes", 
isRush_unknown = "unknown", isRush_new = "new", shooterHandCode_R = "R", 
shooterHandCode_unknown = "unknown", shooterHandCode_new = "new", 
shooterPositionCode_D = "D", shooterPositionCode_L = "L", shooterPositionCode_R = "R", 
shooterPositionCode_unknown = "unknown", shooterPositionCode_new = "new", 
isEmptyNetFor_yes = "yes", isEmptyNetFor_unknown = "unknown", 
isEmptyNetFor_new = "new", strengthState_penalty.kill = "penalty-kill", 
strengthState_power.play = "power-play", strengthState_unknown = "unknown", 
strengthState_new = "new"), dummy_map = list(crossedRoyalRoad = c(yes = "crossedRoyalRoad_yes", 
unknown = "crossedRoyalRoad_unknown", new = "crossedRoyalRoad_new"
), isBehindNet = c(yes = "isBehindNet_yes", unknown = "isBehindNet_unknown", 
new = "isBehindNet_new"), isEmptyNetFor = c(yes = "isEmptyNetFor_yes", 
unknown = "isEmptyNetFor_unknown", new = "isEmptyNetFor_new"), 
    isHome = c(yes = "isHome_yes", unknown = "isHome_unknown", 
    new = "isHome_new"), isOvertime = c(yes = "isOvertime_yes", 
    unknown = "isOvertime_unknown", new = "isOvertime_new"), 
    isPlayoff = c(yes = "isPlayoff_yes", unknown = "isPlayoff_unknown", 
    new = "isPlayoff_new"), isRebound = c(yes = "isRebound_yes", 
    unknown = "isRebound_unknown", new = "isRebound_new"), isRush = c(yes = "isRush_yes", 
    unknown = "isRush_unknown", new = "isRush_new"), shooterHandCode = c(R = "shooterHandCode_R", 
    unknown = "shooterHandCode_unknown", new = "shooterHandCode_new"
    ), shooterPositionCode = c(D = "shooterPositionCode_D", L = "shooterPositionCode_L", 
    R = "shooterPositionCode_R", unknown = "shooterPositionCode_unknown", 
    new = "shooterPositionCode_new"), shotType = c(deflected = "shotType_deflected", 
    other = "shotType_other", slap = "shotType_slap", snap = "shotType_snap", 
    `tip-in` = "shotType_tip.in", wrist = "shotType_wrist", unknown = "shotType_unknown", 
    new = "shotType_new"), strengthState = c(`penalty-kill` = "strengthState_penalty.kill", 
    `power-play` = "strengthState_power.play", unknown = "strengthState_unknown", 
    new = "strengthState_new"), typeDescKeyPrev = c(`backhand-shot-on-goal-for` = "typeDescKeyPrev_backhand.shot.on.goal.for", 
    `blocked-shot-against` = "typeDescKeyPrev_blocked.shot.against", 
    `blocked-shot-for` = "typeDescKeyPrev_blocked.shot.for", 
    `deflected-shot-on-goal-against` = "typeDescKeyPrev_deflected.shot.on.goal.against", 
    `giveaway-against` = "typeDescKeyPrev_giveaway.against", 
    `giveaway-for` = "typeDescKeyPrev_giveaway.for", `given-hit` = "typeDescKeyPrev_given.hit", 
    `high-missed-shot-against` = "typeDescKeyPrev_high.missed.shot.against", 
    `high-missed-shot-for` = "typeDescKeyPrev_high.missed.shot.for", 
    `lost-faceoff` = "typeDescKeyPrev_lost.faceoff", `other-missed-shot-against` = "typeDescKeyPrev_other.missed.shot.against", 
    `other-shot-on-goal-against` = "typeDescKeyPrev_other.shot.on.goal.against", 
    `post-missed-shot-against` = "typeDescKeyPrev_post.missed.shot.against", 
    `post-missed-shot-for` = "typeDescKeyPrev_post.missed.shot.for", 
    `slap-shot-on-goal-against` = "typeDescKeyPrev_slap.shot.on.goal.against", 
    `snap-shot-on-goal-against` = "typeDescKeyPrev_snap.shot.on.goal.against", 
    `takeaway-against` = "typeDescKeyPrev_takeaway.against", 
    `takeaway-for` = "typeDescKeyPrev_takeaway.for", `taken-hit` = "typeDescKeyPrev_taken.hit", 
    `tip-in-shot-on-goal-against` = "typeDescKeyPrev_tip.in.shot.on.goal.against", 
    `wide-missed-shot-against` = "typeDescKeyPrev_wide.missed.shot.against", 
    `wide-missed-shot-for` = "typeDescKeyPrev_wide.missed.shot.for", 
    `won-faceoff` = "typeDescKeyPrev_won.faceoff", `wrist-shot-on-goal-against` = "typeDescKeyPrev_wrist.shot.on.goal.against", 
    `wrist-shot-on-goal-for` = "typeDescKeyPrev_wrist.shot.on.goal.for", 
    unknown = "typeDescKeyPrev_unknown", new = "typeDescKeyPrev_new"
    ), zoneCode = c(N = "zoneCode_N", O = "zoneCode_O", unknown = "zoneCode_unknown", 
    new = "zoneCode_new")), known_levels = list(isPlayoff = c("no", 
"yes"), isHome = c("no", "yes"), isOvertime = c("no", "yes"), 
    zoneCode = c("D", "N", "O"), isBehindNet = c("no", "yes"), 
    crossedRoyalRoad = c("no", "yes"), typeDescKeyPrev = c("backhand-shot-on-goal-against", 
    "backhand-shot-on-goal-for", "blocked-shot-against", "blocked-shot-for", 
    "deflected-shot-on-goal-against", "giveaway-against", "giveaway-for", 
    "given-hit", "high-missed-shot-against", "high-missed-shot-for", 
    "lost-faceoff", "other-missed-shot-against", "other-shot-on-goal-against", 
    "post-missed-shot-against", "post-missed-shot-for", "slap-shot-on-goal-against", 
    "snap-shot-on-goal-against", "takeaway-against", "takeaway-for", 
    "taken-hit", "tip-in-shot-on-goal-against", "wide-missed-shot-against", 
    "wide-missed-shot-for", "won-faceoff", "wrist-shot-on-goal-against", 
    "wrist-shot-on-goal-for"), shotType = c("backhand", "deflected", 
    "other", "slap", "snap", "tip-in", "wrist"), isRebound = c("no", 
    "yes"), isRush = c("no", "yes"), shooterHandCode = c("L", 
    "R"), shooterPositionCode = c("C", "D", "L", "R"), isEmptyNetFor = c("no", 
    "yes"), strengthState = c("even-strength", "penalty-kill", 
    "power-play")), logical_vars = c("crossedRoyalRoad", "isBehindNet", 
"isEmptyNetFor", "isHome", "isOvertime", "isPlayoff", "isRebound", 
"isRush"), has_unknown = c(crossedRoyalRoad = TRUE, isBehindNet = TRUE, 
isEmptyNetFor = TRUE, isHome = TRUE, isOvertime = TRUE, isPlayoff = TRUE, 
isRebound = TRUE, isRush = TRUE, shooterHandCode = TRUE, shooterPositionCode = TRUE, 
shotType = TRUE, strengthState = TRUE, typeDescKeyPrev = TRUE, 
zoneCode = TRUE), has_new = c(crossedRoyalRoad = TRUE, isBehindNet = TRUE, 
isEmptyNetFor = TRUE, isHome = TRUE, isOvertime = TRUE, isPlayoff = TRUE, 
isRebound = TRUE, isRush = TRUE, shooterHandCode = TRUE, shooterPositionCode = TRUE, 
shotType = TRUE, strengthState = TRUE, typeDescKeyPrev = TRUE, 
zoneCode = TRUE), medians = c(distance = 83.1955405552883, angle = 15.8380000013696, 
xCoordNorm = 11, zoneCode_O = 0, shotType_wrist = 1, dDistancePerSecond = 1.18128575642463, 
typeDescKeyPrev_takeaway.against = 0, shotType_snap = 0, dDistance = 15.6124665490398, 
corsiFor = 56, shotType_other = 0, typeDescKeyPrev_blocked.shot.for = 0, 
dAngle = -3.36402503261127, typeDescKeyPrev_blocked.shot.against = 0, 
dYCoordNormPerSecond = 0.149839743589744, dXCoordNormPerSecond = -1.22649572649573, 
isRush_yes = 0, shotType_deflected = 0, typeDescKeyPrev_high.missed.shot.against = 0, 
typeDescKeyPrev_takeaway.for = 0, minSecondsElapsedInShiftFor = 24, 
typeDescKeyPrev_high.missed.shot.for = 0, isHome_yes = 1, zoneCode_N = 0, 
maxSecondsElapsedInShiftAgainst = 83, maxSecondsElapsedInShiftFor = 62, 
isPlayoff_yes = 0, typeDescKeyPrev_deflected.shot.on.goal.against = 0, 
shotType_slap = 0, crossedRoyalRoad_yes = 0, shooterPositionCode_R = 0, 
shooterPositionCode_L = 0, dXCoordNorm = -17, corsiDifferential = -4, 
avgSecondsElapsedInShiftFor = 42.9, secondsElapsedInPeriod = 1121, 
shooterWeight = 200, typeDescKeyPrev_given.hit = 0, secondsElapsedInGame = 3521, 
typeDescKeyPrev_other.shot.on.goal.against = 0, skaterCountAgainst = 6, 
typeDescKeyPrev_giveaway.against = 0, fenwickFor = 41, typeDescKeyPrev_post.missed.shot.for = 0, 
isRebound_yes = 0, corsiAgainst = 59, avgSecondsElapsedSinceLastShiftFor = 130.9, 
isOvertime_yes = 0, periodNumber = 3, typeDescKeyPrev_lost.faceoff = 0, 
maxSecondsElapsedSinceLastShiftFor = 182, shooterAge = 28, goalDifferential = 2, 
fenwickDifferential = -1, maxSecondsElapsedSinceLastShiftAgainst = 206.5, 
dAnglePerSecond = -0.273920225223822, shooterSecondsElapsedSinceLastShift = 129, 
goalsAgainst = 2, typeDescKeyPrev_tip.in.shot.on.goal.against = 0, 
dYCoordNorm = 2, typeDescKeyPrev_won.faceoff = 0, typeDescKeyPrev_wide.missed.shot.for = 0, 
minSecondsElapsedInShiftAgainst = 26, shooterHeight = 73, typeDescKeyPrev_snap.shot.on.goal.against = 0, 
manDifferential = 0, minSecondsElapsedSinceLastShiftFor = 84, 
shooterPositionCode_D = 0, shotsAgainst = 28, typeDescKeyPrev_wide.missed.shot.against = 0, 
typeDescKeyPrev_wrist.shot.on.goal.against = 0, shotType_tip.in = 0, 
shooterSecondsElapsedInShift = 39, fenwickAgainst = 42, strengthState_penalty.kill = 0, 
secondsElapsedInSequence = 40, typeDescKeyPrev_other.missed.shot.against = 0, 
avgSecondsElapsedInShiftAgainst = 56, yCoordNorm = 2, typeDescKeyPrev_wrist.shot.on.goal.for = 0, 
strengthState_power.play = 0, shotsFor = 27, isBehindNet_yes = 0, 
typeDescKeyPrev_slap.shot.on.goal.against = 0, goalsFor = 3, 
minSecondsElapsedSinceLastShiftAgainst = 79.5, shotDifferential = 0, 
avgSecondsElapsedSinceLastShiftAgainst = 141.75, shooterHandCode_R = 0, 
skaterCountFor = 5, typeDescKeyPrev_post.missed.shot.against = 0, 
typeDescKeyPrev_giveaway.for = 0, typeDescKeyPrev_backhand.shot.on.goal.for = 0, 
dSecondsElapsedInSequence = 9, typeDescKeyPrev_taken.hit = 0), 
    normalize_means = c(distance = 88.1871627898831, angle = 20.1315121152732, 
    xCoordNorm = 5.43409586056645, zoneCode_O = 0.434640522875817, 
    shotType_wrist = 0.769063180827887, dDistancePerSecond = 3.73172164667088, 
    typeDescKeyPrev_takeaway.against = 0.0163398692810458, shotType_snap = 0.0964052287581699, 
    dDistance = 15.8014687880717, corsiFor = 56.2075163398693, 
    shotType_other = 0.0136165577342048, typeDescKeyPrev_blocked.shot.for = 0.0484749455337691, 
    dAngle = -9.13232699145456, typeDescKeyPrev_blocked.shot.against = 0.157952069716776, 
    dYCoordNormPerSecond = 0.466647317036255, dXCoordNormPerSecond = -4.15172162831412, 
    isRush_yes = 0.0833333333333333, shotType_deflected = 0.00163398692810458, 
    typeDescKeyPrev_high.missed.shot.against = 0.00435729847494553, 
    typeDescKeyPrev_takeaway.for = 0.0615468409586057, minSecondsElapsedInShiftFor = 28.6437908496732, 
    typeDescKeyPrev_high.missed.shot.for = 0.000544662309368192, 
    isHome_yes = 0.550653594771242, zoneCode_N = 0.218954248366013, 
    maxSecondsElapsedInShiftAgainst = 90.1350762527233, maxSecondsElapsedInShiftFor = 68.1541394335512, 
    isPlayoff_yes = 0.0718954248366013, typeDescKeyPrev_deflected.shot.on.goal.against = 0.000544662309368192, 
    shotType_slap = 0.0174291938997821, crossedRoyalRoad_yes = 0.467864923747277, 
    shooterPositionCode_R = 0.202069716775599, shooterPositionCode_L = 0.254901960784314, 
    dXCoordNorm = -17.5697167755991, corsiDifferential = -3.6318082788671, 
    avgSecondsElapsedInShiftFor = 46.0640341321714, secondsElapsedInPeriod = 1103.83387799564, 
    shooterWeight = 200.181917211329, typeDescKeyPrev_given.hit = 0.037037037037037, 
    secondsElapsedInGame = 3504.48747276688, typeDescKeyPrev_other.shot.on.goal.against = 0.000544662309368192, 
    skaterCountAgainst = 5.96296296296296, typeDescKeyPrev_giveaway.against = 0.104030501089325, 
    fenwickFor = 41.0076252723312, typeDescKeyPrev_post.missed.shot.for = 0.0103485838779956, 
    isRebound_yes = 0.0206971677559913, corsiAgainst = 59.8393246187364, 
    avgSecondsElapsedSinceLastShiftFor = 136.430001815541, isOvertime_yes = 0.000544662309368192, 
    periodNumber = 3.00054466230937, typeDescKeyPrev_lost.faceoff = 0.0964052287581699, 
    maxSecondsElapsedSinceLastShiftFor = 195.557189542484, shooterAge = 28.328431372549, 
    goalDifferential = 1.70152505446623, fenwickDifferential = -1.01307189542484, 
    maxSecondsElapsedSinceLastShiftAgainst = 220.966775599129, 
    dAnglePerSecond = -2.03163545092558, shooterSecondsElapsedSinceLastShift = 141.507625272331, 
    goalsAgainst = 1.80174291938998, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.00544662309368192, 
    dYCoordNorm = 2.02124183006536, typeDescKeyPrev_won.faceoff = 0.148148148148148, 
    typeDescKeyPrev_wide.missed.shot.for = 0.0419389978213508, 
    minSecondsElapsedInShiftAgainst = 34.2478213507625, shooterHeight = 73.1399782135076, 
    typeDescKeyPrev_snap.shot.on.goal.against = 0.0163398692810458, 
    manDifferential = -0.0392156862745098, minSecondsElapsedSinceLastShiftFor = 89.4242919389978, 
    shooterPositionCode_D = 0.13562091503268, shotsAgainst = 27.7184095860566, 
    typeDescKeyPrev_wide.missed.shot.against = 0.065359477124183, 
    typeDescKeyPrev_wrist.shot.on.goal.against = 0.062636165577342, 
    shotType_tip.in = 0.00871459694989107, shooterSecondsElapsedInShift = 43.9635076252723, 
    fenwickAgainst = 42.020697167756, strengthState_penalty.kill = 0.0610021786492375, 
    secondsElapsedInSequence = 59.4651416122004, typeDescKeyPrev_other.missed.shot.against = 0.00925925925925926, 
    avgSecondsElapsedInShiftAgainst = 62.3196169208425, yCoordNorm = 2.26579520697168, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.00326797385620915, 
    strengthState_power.play = 0.0206971677559913, shotsFor = 27.8126361655773, 
    isBehindNet_yes = 0.00217864923747277, typeDescKeyPrev_slap.shot.on.goal.against = 0.0174291938997821, 
    goalsFor = 3.50326797385621, minSecondsElapsedSinceLastShiftAgainst = 86.5065359477124, 
    shotDifferential = 0.0942265795206972, avgSecondsElapsedSinceLastShiftAgainst = 146.344480755265, 
    shooterHandCode_R = 0.342047930283224, skaterCountFor = 4.92374727668845, 
    typeDescKeyPrev_post.missed.shot.against = 0.00599128540305011, 
    typeDescKeyPrev_giveaway.for = 0.0245098039215686, typeDescKeyPrev_backhand.shot.on.goal.for = 0.00108932461873638, 
    dSecondsElapsedInSequence = 14.4673202614379, typeDescKeyPrev_taken.hit = 0.0561002178649237
    ), normalize_sds = c(distance = 53.6100444542208, angle = 17.4929574635285, 
    xCoordNorm = 55.2599752450788, zoneCode_O = 0.495844784728335, 
    shotType_wrist = 0.421546903904294, dDistancePerSecond = 24.7978688935219, 
    typeDescKeyPrev_takeaway.against = 0.126813394460044, shotType_snap = 0.295226578584803, 
    dDistance = 76.7243925540284, corsiFor = 11.216131246985, 
    shotType_other = 0.115924400002255, typeDescKeyPrev_blocked.shot.for = 0.214826119215686, 
    dAngle = 33.405874605476, typeDescKeyPrev_blocked.shot.against = 0.364795414861761, 
    dYCoordNormPerSecond = 10.9751342567029, dXCoordNormPerSecond = 26.7917681875398, 
    isRush_yes = 0.276460698312024, shotType_deflected = 0.040400569499963, 
    typeDescKeyPrev_high.missed.shot.against = 0.0658838115769126, 
    typeDescKeyPrev_takeaway.for = 0.24039613875186, minSecondsElapsedInShiftFor = 21.916599897594, 
    typeDescKeyPrev_high.missed.shot.for = 0.0233380014004673, 
    isHome_yes = 0.497563116468556, zoneCode_N = 0.413650191272091, 
    maxSecondsElapsedInShiftAgainst = 48.1635525424483, maxSecondsElapsedInShiftFor = 36.1888580753221, 
    isPlayoff_yes = 0.258385053601326, typeDescKeyPrev_deflected.shot.on.goal.against = 0.0233380014004672, 
    shotType_slap = 0.130899769870523, crossedRoyalRoad_yes = 0.499102207866331, 
    shooterPositionCode_R = 0.401653350802023, shooterPositionCode_L = 0.435924825627467, 
    dXCoordNorm = 80.2549210804602, corsiDifferential = 19.4077585316867, 
    avgSecondsElapsedInShiftFor = 23.732073682097, secondsElapsedInPeriod = 85.6869797109394, 
    shooterWeight = 15.5681719547181, typeDescKeyPrev_given.hit = 0.188904026035916, 
    secondsElapsedInGame = 83.4117808440893, typeDescKeyPrev_other.shot.on.goal.against = 0.023338001400467, 
    skaterCountAgainst = 0.19736890793884, typeDescKeyPrev_giveaway.against = 0.305383284689608, 
    fenwickFor = 8.58970593737442, typeDescKeyPrev_post.missed.shot.for = 0.101227821687282, 
    isRebound_yes = 0.142407305527925, corsiAgainst = 11.9041494554423, 
    avgSecondsElapsedSinceLastShiftFor = 45.2518971818069, isOvertime_yes = 0.0233380014004672, 
    periodNumber = 0.0233380014004672, typeDescKeyPrev_lost.faceoff = 0.295226578584803, 
    maxSecondsElapsedSinceLastShiftFor = 84.7116120613761, shooterAge = 4.02640845649108, 
    goalDifferential = 0.689522889127525, fenwickDifferential = 14.2781952292712, 
    maxSecondsElapsedSinceLastShiftAgainst = 102.413238280516, 
    dAnglePerSecond = 12.4040747046006, shooterSecondsElapsedSinceLastShift = 77.0580689536785, 
    goalsAgainst = 1.14359594021359, typeDescKeyPrev_tip.in.shot.on.goal.against = 0.0736200340324108, 
    dYCoordNorm = 34.1493607567096, typeDescKeyPrev_won.faceoff = 0.355343563808172, 
    typeDescKeyPrev_wide.missed.shot.for = 0.200504400962461, 
    minSecondsElapsedInShiftAgainst = 29.4002932503624, shooterHeight = 2.09098566110518, 
    typeDescKeyPrev_snap.shot.on.goal.against = 0.126813394460044, 
    manDifferential = 0.288920258638785, minSecondsElapsedSinceLastShiftFor = 37.8365945764699, 
    shooterPositionCode_D = 0.342478856052427, shotsAgainst = 6.64512760208197, 
    typeDescKeyPrev_wide.missed.shot.against = 0.247226426840092, 
    typeDescKeyPrev_wrist.shot.on.goal.against = 0.242373415326444, 
    shotType_tip.in = 0.0929696749621195, shooterSecondsElapsedInShift = 28.9907676275631, 
    fenwickAgainst = 8.98813063849919, strengthState_penalty.kill = 0.239399516719656, 
    secondsElapsedInSequence = 61.8256173330889, typeDescKeyPrev_other.missed.shot.against = 0.0958046166613085, 
    avgSecondsElapsedInShiftAgainst = 34.7772901431202, yCoordNorm = 24.6856894852434, 
    typeDescKeyPrev_wrist.shot.on.goal.for = 0.0570882588216965, 
    strengthState_power.play = 0.142407305527925, shotsFor = 6.32858429705112, 
    isBehindNet_yes = 0.0466378324228981, typeDescKeyPrev_slap.shot.on.goal.against = 0.130899769870524, 
    goalsFor = 1.17024424944335, minSecondsElapsedSinceLastShiftAgainst = 45.055623639711, 
    shotDifferential = 9.8483514658523, avgSecondsElapsedSinceLastShiftAgainst = 53.2909390438439, 
    shooterHandCode_R = 0.474524801614421, skaterCountFor = 0.267519251120546, 
    typeDescKeyPrev_post.missed.shot.against = 0.0771921974448055, 
    typeDescKeyPrev_giveaway.for = 0.154667717708335, typeDescKeyPrev_backhand.shot.on.goal.for = 0.0329959237079672, 
    dSecondsElapsedInSequence = 15.2618032084928, typeDescKeyPrev_taken.hit = 0.230177845658759
    )), ps = list(coefficients = c(`(Intercept)` = -0.780853160081028, 
shotType_other = 0.0374854886988706, goalieWeight = 0.0299253741734925, 
yCoordNorm = 0.0284818528565124, shotType_snap = 0.0249259290776625, 
shooterHandCode_R = -0.0236758893185456, shooterPositionCode_D = -0.0223388143191858, 
isPlayoff_yes = 0.0199466856355296, goalsAgainst = -0.0183804473385284, 
goalsFor = -0.0164524460835742, distance = -0.0156122702729602, 
fenwickDifferential = 0.0149245382082734, fenwickAgainst = -0.014513922004296, 
isHome_yes = 0.013654982345578, angle = -0.0127221762612523, 
shotType_deflected = -0.0109112865771066, shooterPositionCode_L = 0.0101605184964408, 
shooterHeight = -0.00913965125477881, goalieHandCode_R = 0.00882681470881424, 
xCoordNorm = 0.00820518594658108, shooterWeight = -0.00802568774346184, 
corsiAgainst = -0.00766575385917949, shotType_slap = -0.00709296640090848, 
corsiDifferential = 0.00685855166229266, goalieHeight = 0.00601482698906022, 
shooterAge = -0.0058236330928557, goalDifferential = 0.00549980338474507, 
fenwickFor = 0.00525385588164898, goalieAge = 0.00504448533018919, 
shotsAgainst = -0.00438268295869598, shotsFor = -0.00260723728893, 
corsiFor = 0.0015887603073083, shotType_wrist = -0.00147520997341168, 
shotDifferential = 0.0014140598744273, shooterPositionCode_R = -0.000304683893460496
), raw_terms = c("goalieWeight", "yCoordNorm", "goalsAgainst", 
"goalsFor", "distance", "fenwickDifferential", "fenwickAgainst", 
"angle", "shooterHeight", "xCoordNorm", "shooterWeight", "corsiAgainst", 
"corsiDifferential", "goalieHeight", "shooterAge", "goalDifferential", 
"fenwickFor", "goalieAge", "shotsAgainst", "shotsFor", "corsiFor", 
"shotDifferential"), active_dummy_terms = c("shotType_other", 
"shotType_snap", "shooterHandCode_R", "shooterPositionCode_D", 
"isPlayoff_yes", "isHome_yes", "shotType_deflected", "shooterPositionCode_L", 
"goalieHandCode_R", "shotType_slap", "shotType_wrist", "shooterPositionCode_R"
), active_dummy_by_var = list(goalieHandCode = "goalieHandCode_R", 
    isHome = "isHome_yes", isPlayoff = "isPlayoff_yes", shooterHandCode = "shooterHandCode_R", 
    shooterPositionCode = c("shooterPositionCode_D", "shooterPositionCode_L", 
    "shooterPositionCode_R"), shotType = c("shotType_other", 
    "shotType_snap", "shotType_deflected", "shotType_slap", "shotType_wrist"
    )), dummy_term_to_var = c(isPlayoff_yes = "isPlayoff", isPlayoff_unknown = "isPlayoff", 
isPlayoff_new = "isPlayoff", isHome_yes = "isHome", isHome_unknown = "isHome", 
isHome_new = "isHome", shotType_deflected = "shotType", shotType_other = "shotType", 
shotType_slap = "shotType", shotType_snap = "shotType", shotType_wrist = "shotType", 
shotType_unknown = "shotType", shotType_new = "shotType", shooterHandCode_R = "shooterHandCode", 
shooterHandCode_unknown = "shooterHandCode", shooterHandCode_new = "shooterHandCode", 
shooterPositionCode_D = "shooterPositionCode", shooterPositionCode_L = "shooterPositionCode", 
shooterPositionCode_R = "shooterPositionCode", shooterPositionCode_unknown = "shooterPositionCode", 
shooterPositionCode_new = "shooterPositionCode", goalieHandCode_R = "goalieHandCode", 
goalieHandCode_unknown = "goalieHandCode", goalieHandCode_new = "goalieHandCode"
), dummy_term_to_level = c(isPlayoff_yes = "yes", isPlayoff_unknown = "unknown", 
isPlayoff_new = "new", isHome_yes = "yes", isHome_unknown = "unknown", 
isHome_new = "new", shotType_deflected = "deflected", shotType_other = "other", 
shotType_slap = "slap", shotType_snap = "snap", shotType_wrist = "wrist", 
shotType_unknown = "unknown", shotType_new = "new", shooterHandCode_R = "R", 
shooterHandCode_unknown = "unknown", shooterHandCode_new = "new", 
shooterPositionCode_D = "D", shooterPositionCode_L = "L", shooterPositionCode_R = "R", 
shooterPositionCode_unknown = "unknown", shooterPositionCode_new = "new", 
goalieHandCode_R = "R", goalieHandCode_unknown = "unknown", goalieHandCode_new = "new"
), dummy_map = list(goalieHandCode = c(R = "goalieHandCode_R", 
unknown = "goalieHandCode_unknown", new = "goalieHandCode_new"
), isHome = c(yes = "isHome_yes", unknown = "isHome_unknown", 
new = "isHome_new"), isPlayoff = c(yes = "isPlayoff_yes", unknown = "isPlayoff_unknown", 
new = "isPlayoff_new"), shooterHandCode = c(R = "shooterHandCode_R", 
unknown = "shooterHandCode_unknown", new = "shooterHandCode_new"
), shooterPositionCode = c(D = "shooterPositionCode_D", L = "shooterPositionCode_L", 
R = "shooterPositionCode_R", unknown = "shooterPositionCode_unknown", 
new = "shooterPositionCode_new"), shotType = c(deflected = "shotType_deflected", 
other = "shotType_other", slap = "shotType_slap", snap = "shotType_snap", 
wrist = "shotType_wrist", unknown = "shotType_unknown", new = "shotType_new"
)), known_levels = list(isPlayoff = c("no", "yes"), isHome = c("no", 
"yes"), shotType = c("backhand", "deflected", "other", "slap", 
"snap", "wrist"), shooterHandCode = c("L", "R"), shooterPositionCode = c("C", 
"D", "L", "R"), goalieHandCode = c("L", "R")), logical_vars = c("isHome", 
"isPlayoff"), has_unknown = c(goalieHandCode = TRUE, isHome = TRUE, 
isPlayoff = TRUE, shooterHandCode = TRUE, shooterPositionCode = TRUE, 
shotType = TRUE), has_new = c(goalieHandCode = TRUE, isHome = TRUE, 
isPlayoff = TRUE, shooterHandCode = TRUE, shooterPositionCode = TRUE, 
shotType = TRUE), medians = c(shotType_other = 0, goalieWeight = 201, 
yCoordNorm = 0, shotType_snap = 0, shooterHandCode_R = 0, shooterPositionCode_D = 0, 
isPlayoff_yes = 0, goalsAgainst = 3, goalsFor = 3, distance = 13.0384048104053, 
fenwickDifferential = 0, fenwickAgainst = 47, isHome_yes = 1, 
angle = 10.6196552761551, shotType_deflected = 0, shooterPositionCode_L = 0, 
shooterHeight = 73, goalieHandCode_R = 0, xCoordNorm = 76, shooterWeight = 198, 
corsiAgainst = 65, shotType_slap = 0, corsiDifferential = 0, 
goalieHeight = 75, shooterAge = 27, goalDifferential = 0, fenwickFor = 46, 
goalieAge = 28, shotsAgainst = 32, shotsFor = 31, corsiFor = 65, 
shotType_wrist = 1, shotDifferential = 0, shooterPositionCode_R = 0
), normalize_means = c(shotType_other = 0.00251677852348993, 
goalieWeight = 200.947986577181, yCoordNorm = -0.0276845637583893, 
shotType_snap = 0.140939597315436, shooterHandCode_R = 0.417785234899329, 
shooterPositionCode_D = 0.0562080536912752, isPlayoff_yes = 0.000838926174496644, 
goalsAgainst = 3.20721476510067, goalsFor = 3.06208053691275, 
distance = 14.1162288604514, fenwickDifferential = -0.326342281879195, 
fenwickAgainst = 46.9093959731544, isHome_yes = 0.518456375838926, 
angle = 15.8529645126295, shotType_deflected = 0.000838926174496644, 
shooterPositionCode_L = 0.202181208053691, shooterHeight = 72.7298657718121, 
goalieHandCode_R = 0.0553691275167785, xCoordNorm = 75.6870805369127, 
shooterWeight = 197.462248322148, corsiAgainst = 64.3464765100671, 
shotType_slap = 0.00419463087248322, corsiDifferential = -0.278523489932886, 
goalieHeight = 74.8859060402685, shooterAge = 27.3951342281879, 
goalDifferential = -0.145134228187919, fenwickFor = 46.5830536912752, 
goalieAge = 28.4840604026846, shotsAgainst = 31.8724832214765, 
shotsFor = 31.5444630872483, corsiFor = 64.0679530201342, shotType_wrist = 0.656040268456376, 
shotDifferential = -0.328020134228188, shooterPositionCode_R = 0.275167785234899
), normalize_sds = c(shotType_other = 0.0501253647882481, goalieWeight = 16.1798969079781, 
yCoordNorm = 4.24284713524948, shotType_snap = 0.348105280082041, 
shooterHandCode_R = 0.493401423541137, shooterPositionCode_D = 0.230419725085298, 
isPlayoff_yes = 0.0289642223181745, goalsAgainst = 1.49721766948129, 
goalsFor = 1.446326607353, distance = 5.90127586207415, fenwickDifferential = 14.9971744568645, 
fenwickAgainst = 11.4025721745455, isHome_yes = 0.499868966651878, 
angle = 16.2204685269784, shotType_deflected = 0.0289642223181745, 
shooterPositionCode_L = 0.401795225089013, shooterHeight = 2.11249864561555, 
goalieHandCode_R = 0.228795329404374, xCoordNorm = 6.23499246320141, 
shooterWeight = 15.0218508961002, corsiAgainst = 14.5048677431509, 
shotType_slap = 0.0646571195721434, corsiDifferential = 19.4936216441126, 
goalieHeight = 1.59777427619789, shooterAge = 4.43943209529498, 
goalDifferential = 0.677013604807505, fenwickFor = 11.096166828729, 
goalieAge = 3.7245297515799, shotsAgainst = 8.1947648424196, 
shotsFor = 8.01246108938845, corsiFor = 14.1606915577925, shotType_wrist = 0.475227207197835, 
shotDifferential = 10.6112412917972, shooterPositionCode_R = 0.446786235148888
)))

#' Access all the awards
#'
#' `awards()` retrieves all the awards as a `data.frame` where each row represents award and includes detail on recognition, leaderboard, and milestone-watch context.
#'
#' @returns data.frame with one row per award
#' @examples
#' all_awards <- awards()
#' @export

awards <- function() {
  tryCatch({
    trophies <- nhl_api(
      path = 'trophy',
      type = 'r'
    )$data
    names(trophies)[names(trophies) == 'id']        <- 'trophyId'
    names(trophies)[names(trophies) == 'name']      <- 'trophyName'
    names(trophies)[names(trophies) == 'shortName'] <- 'trophyShortName'
    trophies
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access all the award winners/finalists
#'
#' `award_winners()` retrieves all the award winners/finalists as a `data.frame` where each row represents winner/finalist and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
#'
#' @returns data.frame with one row per winner/finalist
#' @examples
#' all_award_winners <- award_winners()
#' @export

award_winners <- function() {
  tryCatch({
    winners    <- nhl_api(
      path = 'award-details',
      type = 'r'
    )$data
    winners$id <- NULL
    names(winners)[names(winners) == 'fullName'] <- 'playerFullName'
    winners[order(winners$trophyId, winners$seasonId, winners$status), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the season(s) and game type(s) in which there exists team EDGE 
#' statistics
#'
#' `team_edge_seasons()` retrieves the season(s) and game type(s) in which there exists team EDGE statistics as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @returns data.frame with one row per season
#' @examples
#' team_EDGE_seasons <- team_edge_seasons()
#' @export

team_edge_seasons <- function() {
  tryCatch(
    expr = {
      seasons <- nhl_api(
        path = sprintf('v1/edge/team-landing/now'),
        type = 'w'
      )$seasonsWithEdgeStats
      names(seasons)[names(seasons) == 'id']        <- 'seasonId'
      names(seasons)[names(seasons) == 'gameTypes'] <- 'gameTypeIds'
      seasons
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the team EDGE statistics leaders for a season and game type
#'
#' `team_edge_leaders()` retrieves the team EDGE statistics leaders for a season and game type as a nested `list` that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
#'
#' @param season integer in YYYYYYYY (e.g., 20242025); see 
#' [team_edge_seasons()] for reference
#' @param game_type integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season) OR character of 'pre', 'regular', or 
#' 'playoff'/'post'; see [team_edge_seasons()] for reference; most functions 
#' will NOT support pre-season
#'
#' @returns list of various items
#' @examples
#' team_EDGE_leaders_regular_20242025 <- team_edge_leaders(
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

team_edge_leaders <- function(season = 'now', game_type = '') {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf(
          'v1/edge/team-landing/%s/%s', 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )$leaders
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the EDGE summary for a team, season, and game type
#'
#' `team_edge_summary()` retrieves the EDGE summary for a team, season, and game type as a nested `list` that separates summary and detail blocks for team identity, affiliation, and matchup-side context plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams team_seasons
#' @inheritParams team_edge_leaders
#'
#' @returns list of various items
#' @examples
#' COL_EDGE_summary_regular_20242025 <- team_edge_summary(
#'   team      = 21, 
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

team_edge_summary <- function(team = 1, season = 'now', game_type = '') {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf(
          'v1/edge/team-detail/%s/%s/%s', 
          to_team_id(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the EDGE zone time statistics for a team, season, game type, and 
#' category
#'
#' `team_edge_zone_time()` retrieves the EDGE zone time statistics for a team, season, game type, and category as a `data.frame` where each row represents strength state and includes detail on NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams team_edge_summary
#' @param category character of 'd'/'details' or 
#' 'dS'/'dSOG'/'dShot'/'shot differential'
#'
#' @returns data.frame with one row per strength state 
#' (category = 'details') or list with four items (category = 'shot 
#' differential')
#' @examples
#' COL_dS_regular_20242025 <- team_edge_zone_time(
#'   team      = 21,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'dS'
#' )
#' @export

team_edge_zone_time <- function(
  team      = 1, 
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        tolower(category),
        d                   = 'zoneTimeDetails',
        details             = 'zoneTimeDetails',
        ds                  = 'shotDifferential',
        dsog                = 'shotDifferential',
        dshot               = 'shotDifferential',
        `shot differential` = 'shotDifferential'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/team-zone-time-details/%s/%s/%s', 
          to_team_id(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE skating distance statistics for a team, season, game type, 
#' and category
#'
#' `team_edge_skating_distance()` retrieves the EDGE skating distance statistics for a team, season, game type, and category as a `data.frame` where each row represents combination of strength state and position and includes detail on team identity, affiliation, and matchup-side context, ranking movement, points pace, and division/conference position signals, and NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams team_edge_summary
#' @param category character of 'd'/'details' or 'l'/'l10'/'last 10'
#'
#' @returns data.frame with one row per combination of strength state and 
#' position (category = 'details') or game (category = 'last 10')
#' game
#' @examples
#' COL_L10_skating_distance_regular_20242025 <- team_edge_skating_distance(
#'   team      = 21,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'L'
#' )
#' @export

team_edge_skating_distance <- function(
  team      = 1, 
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'skatingDistanceDetails',
        l = 'skatingDistanceLast10'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/team-skating-distance-detail/%s/%s/%s', 
          to_team_id(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE skating speed statistics for a team, season, game type, and 
#' category
#'
#' `team_edge_skating_speed()` retrieves the EDGE skating speed statistics for a team, season, game type, and category as a `data.frame` where each row represents position and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and ranking movement, points pace, and division/conference position signals.
#'
#' @inheritParams team_edge_summary
#' @param category character of 'd'/'details' or 't'/'top'/'top speeds'
#'
#' @returns data.frame with one row per position (category = 'details') or 
#' burst (category = 'top speeds')
#' @examples
#' COL_top_speeds_regular_20242025 <- team_edge_skating_speed(
#'   team      = 21,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'T'
#' )
#' @export

team_edge_skating_speed <- function(
  team      = 1, 
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'skatingSpeedDetails',
        t = 'topSkatingSpeeds'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/team-skating-speed-detail/%s/%s/%s', 
          to_team_id(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE shot location statistics for a team, season, game type, and 
#' category
#'
#' `team_edge_shot_location()` retrieves the EDGE shot location statistics for a team, season, game type, and category as a `data.frame` where each row represents location and includes detail on production, workload, efficiency, and result-level performance outcomes plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams team_edge_summary
#' @param category character of 'd'/details' or 't'/'totals'
#'
#' @returns data.frame with one row per location (category = 'details') or 
#' combination of strength state and position (category = 'totals')
#' @examples
#' COL_shot_location_totals_regular_20242025 <- team_edge_shot_location(
#'   team      = 21,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'T'
#' )
#' @export

team_edge_shot_location <- function(
  team      = 1, 
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'shotLocationDetails',
        t = 'shotLocationTotals'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/team-shot-location-detail/%s/%s/%s', 
          to_team_id(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE shot speed statistics for a team, season, game type, and 
#' category
#'
#' `team_edge_shot_speed()` retrieves the EDGE shot speed statistics for a team, season, game type, and category as a `data.frame` where each row represents position and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and ranking movement, points pace, and division/conference position signals.
#'
#' @inheritParams team_edge_summary
#' @param category character of 'd'/'details' or 'h'/'hardest'
#'
#' @returns data.frame with one row per position (category = 'details') or 
#' shot (category = 'hardest')
#' @examples
#' COL_hardest_shots_regular_20242025 <- team_edge_shot_speed(
#'   team      = 21,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'H'
#' )
#' @export

team_edge_shot_speed <- function(
  team      = 1, 
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'shotSpeedDetails',
        h = 'hardestShots',
      )
      nhl_api(
        path = sprintf(
          'v1/edge/team-shot-speed-detail/%s/%s/%s', 
          to_team_id(team), 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access all the games
#'
#' `games()` retrieves all the games as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow plus date/season filtering windows and chronological context.
#'
#' @returns data.frame with one row per game
#' @examples
#' # May take >5s, so skip.
#' \donttest{all_games <- games()}
#' @export

games <- function() {
  tryCatch({
    games <- nhl_api(
      path = 'en/game',
      type = 's'
    )$data
    names(games)[names(games) == 'id']       <- 'gameId'
    names(games)[names(games) == 'season']   <- 'seasonId'
    names(games)[names(games) == 'gameType'] <- 'gameTypeId'
    games[order(games$gameId), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Access the scores for a date
#'
#' `scores()` retrieves the scores for a date as a `data.frame` where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @inheritParams standings
#'
#' @returns data.frame with one row per game
#' @examples
#' scores_Halloween_2025 <- scores(date = '2025-10-31')
#' @export

scores <- function(date = 'now') {
  tryCatch(
    expr = {
      games <- nhl_api(
        path = sprintf('v1/score/%s', date),
        type = 'w'
      )$games
      names(games)[names(games) == 'id']       <- 'gameId'
      names(games)[names(games) == 'season']   <- 'seasonId'
      names(games)[names(games) == 'gameType'] <- 'gameTypeId'
      names(games) <- normalize_locale_names(names(games))
      names(games) <- normalize_team_abbrev_cols(names(games))
      games[order(games$gameId), ]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the GameCenter (GC) summary for a game
#'
#' `gc_summary()` retrieves the GameCenter (GC) summary for a game as a nested `list` that separates summary and detail blocks for game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and venue/location geography and regional metadata.
#'
#' @param game integer ID (e.g., 2025020275); see [games()] for reference
#'
#' @returns list of various items
#' @examples
#' gc_summary_Martin_Necas_legacy_game <- gc_summary(game = 2025020275)
#' @export

gc_summary <- function(game = 2023030417) {
  tryCatch(
    expr = {
      landing    <- nhl_api(
        path = sprintf('v1/gamecenter/%s/landing', game),
        type = 'w'
      )
      right_rail <- nhl_api(
        path = sprintf('v1/gamecenter/%s/right-rail', game),
        type = 'w'
      )
      summary <- c(landing, right_rail)
      names(summary)[names(summary) == 'id']       <- 'gameId'
      names(summary)[names(summary) == 'season']   <- 'seasonId'
      names(summary)[names(summary) == 'gameType'] <- 'gameTypeId'
      summary
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the World Showcase (WSC) summary for a game
#'
#' `wsc_summary()` retrieves the World Showcase (WSC) summary for a game as a nested `list` that separates summary and detail blocks for game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and venue/location geography and regional metadata.
#'
#' @inheritParams gc_summary
#'
#' @returns list of various items
#' @examples
#' wsc_summary_Martin_Necas_legacy_game <- wsc_summary(game = 2025020275)
#' @export

wsc_summary <- function(game = 2023030417) {
  tryCatch(
    expr = {
      summary <- nhl_api(
        path = sprintf('v1/wsc/game-story/%s', game),
        type = 'w'
      )
      names(summary)[names(summary) == 'id']       <- 'gameId'
      names(summary)[names(summary) == 'season']   <- 'seasonId'
      names(summary)[names(summary) == 'gameType'] <- 'gameTypeId'
      summary
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the boxscore for a game, team, and position
#'
#' `boxscore()` retrieves the boxscore for a game, team, and position as a `data.frame` where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus production, workload, efficiency, and result-level performance outcomes.
#'
#' @inheritParams gc_summary
#' @inheritParams roster
#' @param team character of 'h'/'home' or 'a'/'away'
#'
#' @returns data.frame with one row per player
#' @examples
#' boxscore_COL_forwards_Martin_Necas_legacy_game <- boxscore(
#'   game     = 2025020275,
#'   team     = 'H',
#'   position = 'F'
#' )
#' @export

boxscore <- function(
  game     = 2023030417,
  team     = 'home',
  position = 'forwards'
) {
  tryCatch(
    expr = {
      team <- switch(
        substring(tolower(team), 1, 1),
        h = 'home',
        a = 'away'
      )
      position <- switch(
        substring(tolower(position), 1, 1),
        f = 'forwards',
        d = 'defense',
        g = 'goalies'
      )
      boxscore <- nhl_api(
        path = sprintf('v1/gamecenter/%s/boxscore', game),
        type = 'w'
      )$playerByGameStats
      boxscore <- boxscore[[paste0(team, 'Team')]][[position]]
      names(boxscore)[names(boxscore) == 'position'] <- 'positionCode'
      names(boxscore) <- normalize_locale_names(names(boxscore))
      names(boxscore) <- scope_person_name_cols(names(boxscore), 'player')
      boxscore
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the rosters for a game
#'
#' `game_rosters()` retrieves the rosters for a game as a `data.frame` where each row represents player and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
#'
#' @inheritParams gc_summary
#'
#' @returns data.frame with one row per player
#' @examples
#' rosters_Martin_Necas_legacy_game <- game_rosters(game = 2025020275)
#' @export

game_rosters <- function(game = 2023030417) {
  tryCatch(
    expr = {
      rosters <- nhl_api(
        path = sprintf('v1/gamecenter/%s/play-by-play', game),
        type = 'w'
      )$rosterSpots
      names(rosters) <- normalize_locale_names(names(rosters))
      names(rosters) <- scope_person_name_cols(names(rosters), 'player')
      rosters <- rosters[order(rosters$sweaterNumber), ]
      rosters[order(rosters$teamId), ]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' List strength-context event types
#'
#' `.supported_strength_event_types()` returns the event types that support the
#' HTML row-matching pipeline.
#'
#' @returns character vector of supported event type keys
#' @keywords internal
.supported_strength_event_types <- function() {
  c(
    'faceoff',
    'hit',
    'shot-on-goal',
    'giveaway',
    'missed-shot',
    'blocked-shot',
    'penalty',
    'goal',
    'delayed-penalty',
    'takeaway',
    'failed-shot-attempt'
  )
}

#' List HTML on-ice player-ID event types
#'
#' `.supported_html_on_ice_id_event_types()` returns the event types that are
#' allowed to receive HTML-derived on-ice player IDs.
#'
#' @returns character vector of supported on-ice player-ID event type keys
#' @keywords internal
.supported_html_on_ice_id_event_types <- function() {
  .supported_strength_event_types()
}

#' Return the tracked number of on-ice skater slots
#'
#' `.on_ice_skater_slots()` returns the current number of skater slots tracked
#' per team in play-by-play outputs. It guarantees the standard five skater
#' columns and expands further when the input already contains overflow skater
#' slots or when a caller requests a larger slot count.
#'
#' @param play_by_play optional data.frame whose existing on-ice columns should
#'   be inspected
#' @param slot_count optional integer scalar minimum slot count requested by the
#'   caller
#' @param min_slots integer scalar default floor for standard outputs
#' @returns integer scalar
#' @keywords internal
.on_ice_skater_slots <- function(
  play_by_play = NULL,
  slot_count = NULL,
  min_slots = 5L
) {
  extract_slots <- function(nms, pattern) {
    matches <- regexec(pattern, nms, perl = TRUE)
    parts <- regmatches(nms, matches)
    vals <- suppressWarnings(as.integer(vapply(
      parts,
      function(x) {
        if (length(x) >= 2L) x[2L] else NA_character_
      },
      character(1)
    )))
    vals[!is.na(vals) & vals > 0L]
  }

  detected <- integer()
  if (!is.null(play_by_play)) {
    nms <- if (is.data.frame(play_by_play)) {
      names(play_by_play)
    } else {
      as.character(play_by_play)
    }
    detected <- unique(c(
      extract_slots(
        nms,
        '^(?:home|away)Skater([0-9]+)(?:PlayerId|SecondsRemainingInShift|SecondsElapsedInShift|SecondsElapsedInPeriodSinceLastShift)$'
      ),
      extract_slots(
        nms,
        '^skater([0-9]+)(?:PlayerId(?:For|Against)|SecondsRemainingInShift(?:For|Against)|SecondsElapsedInShift(?:For|Against)|SecondsElapsedInPeriodSinceLastShift(?:For|Against))$'
      )
    ))
  }

  candidates <- c(as.integer(min_slots), as.integer(slot_count), detected)
  candidates <- candidates[!is.na(candidates) & candidates >= 0L]
  if (!length(candidates)) {
    return(0L)
  }
  max(candidates)
}

#' List on-ice player ID columns
#'
#' `.on_ice_id_scalar_column_names()` returns the public scalar goalie and
#' skater ID column names used in enriched play-by-play outputs.
#'
#' @param play_by_play optional data.frame whose existing on-ice columns should
#'   be inspected
#' @param slot_count optional integer scalar minimum slot count requested by the
#'   caller
#' @returns character vector of on-ice player ID columns
#' @keywords internal
.on_ice_id_scalar_column_names <- function(
  play_by_play = NULL,
  slot_count = NULL
) {
  skater_slots <- seq_len(.on_ice_skater_slots(
    play_by_play = play_by_play,
    slot_count = slot_count
  ))
  c(
    'homeGoaliePlayerId',
    'awayGoaliePlayerId',
    'goaliePlayerIdFor',
    'goaliePlayerIdAgainst',
    as.vector(rbind(
      paste0('homeSkater', skater_slots, 'PlayerId'),
      paste0('awaySkater', skater_slots, 'PlayerId'),
      paste0('skater', skater_slots, 'PlayerIdFor'),
      paste0('skater', skater_slots, 'PlayerIdAgainst')
    ))
  )
}

#' List all scalar on-ice columns
#'
#' `.on_ice_scalar_column_names()` returns the scalar on-ice player ID and
#' shift-timing columns used by the public play-by-play schema.
#'
#' @param play_by_play optional data.frame whose existing on-ice columns should
#'   be inspected
#' @param slot_count optional integer scalar minimum slot count requested by the
#'   caller
#' @returns character vector of scalar on-ice columns
#' @keywords internal
.on_ice_scalar_column_names <- function(
  play_by_play = NULL,
  slot_count = NULL
) {
  c(
    .on_ice_id_scalar_column_names(
      play_by_play = play_by_play,
      slot_count = slot_count
    ),
    .on_ice_timing_scalar_column_names(
      'SecondsRemainingInShift',
      play_by_play = play_by_play,
      slot_count = slot_count
    ),
    .on_ice_timing_scalar_column_names(
      'SecondsElapsedInShift',
      play_by_play = play_by_play,
      slot_count = slot_count
    ),
    .on_ice_timing_scalar_column_names(
      'SecondsElapsedInPeriodSinceLastShift',
      play_by_play = play_by_play,
      slot_count = slot_count
    )
  )
}

#' Initialize empty on-ice columns
#'
#' `.add_empty_html_on_ice_columns()` allocates empty scalar on-ice player ID
#' and timing columns before HTML and shift-chart enrichment.
#'
#' @param play_by_play data.frame play-by-play object
#' @param slot_count optional integer scalar minimum slot count requested by the
#'   caller
#' @returns data.frame with empty on-ice columns added
#' @keywords internal
.add_empty_html_on_ice_columns <- function(play_by_play, slot_count = NULL) {
  slot_count <- .on_ice_skater_slots(
    play_by_play = play_by_play,
    slot_count = slot_count
  )
  id_cols <- .on_ice_id_scalar_column_names(slot_count = slot_count)
  time_cols <- setdiff(
    .on_ice_scalar_column_names(slot_count = slot_count),
    id_cols
  )
  for (nm in id_cols) {
    if (!(nm %in% names(play_by_play))) {
      play_by_play[[nm]] <- rep(NA_integer_, nrow(play_by_play))
    }
  }
  for (nm in time_cols) {
    if (!(nm %in% names(play_by_play))) {
      play_by_play[[nm]] <- rep(NA_real_, nrow(play_by_play))
    }
  }
  play_by_play
}

#' Mask HTML on-ice columns on unsupported events
#'
#' `.mask_strength_context_block()` sets HTML-derived on-ice player-ID and
#' timing columns to `NA` on event types that are not supported by the HTML
#' on-ice player pipeline.
#'
#' @param play_by_play data.frame play-by-play object
#' @returns data.frame with unsupported rows masked
#' @keywords internal
.mask_strength_context_block <- function(play_by_play) {
  type_desc_key <- if ('eventTypeDescKey' %in% names(play_by_play)) {
    as.character(play_by_play$eventTypeDescKey)
  } else {
    rep(NA_character_, nrow(play_by_play))
  }
  supported <- !is.na(type_desc_key) &
    type_desc_key %in% .supported_html_on_ice_id_event_types()
  cols <- intersect(
    .on_ice_scalar_column_names(play_by_play = play_by_play),
    names(play_by_play)
  )
  for (nm in cols) {
    x <- play_by_play[[nm]]
    if (is.integer(x)) {
      x[!supported] <- NA_integer_
    } else if (is.numeric(x)) {
      x[!supported] <- NA_real_
    } else if (is.character(x)) {
      x[!supported] <- NA_character_
    } else if (is.logical(x)) {
      x[!supported] <- NA
    }
    play_by_play[[nm]] <- x
  }
  play_by_play
}

#' Normalize roster person labels
#'
#' `.normalize_roster_person()` standardizes roster names for stable matching by
#' uppercasing, trimming, collapsing whitespace, and transliterating accents.
#'
#' @param x character vector of player names
#' @returns normalized character vector
#' @keywords internal
.normalize_roster_person <- function(x) {
  x <- gsub('\\s+', ' ', x)
  x <- trimws(x)
  x <- toupper(x)
  x <- iconv(x, from = '', to = 'ASCII//TRANSLIT')
  x[is.na(x)] <- ''
  x
}

#' Build a game-roster lookup table
#'
#' `.build_game_roster_lookup()` standardizes roster columns and prepares a
#' lookup table used to resolve HTML sweater numbers to player IDs.
#'
#' @param rosters roster data.frame from the game play-by-play metadata
#' @returns standardized roster lookup data.frame
#' @keywords internal
.build_game_roster_lookup <- function(rosters) {
  if (is.null(rosters) || !nrow(rosters)) {
    return(data.frame())
  }
  rosters <- as.data.frame(rosters, stringsAsFactors = FALSE)
  if (!'playerId' %in% names(rosters) && 'id' %in% names(rosters)) {
    names(rosters)[names(rosters) == 'id'] <- 'playerId'
  }
  if (!'positionCode' %in% names(rosters) && 'position' %in% names(rosters)) {
    names(rosters)[names(rosters) == 'position'] <- 'positionCode'
  }
  if (!'playerLastName' %in% names(rosters) && 'lastName.default' %in% names(rosters)) {
    rosters$playerLastName <- rosters[['lastName.default']]
  }
  if (!'playerFirstName' %in% names(rosters) && 'firstName.default' %in% names(rosters)) {
    rosters$playerFirstName <- rosters[['firstName.default']]
  }
  rosters$sweaterNumber <- suppressWarnings(as.integer(rosters$sweaterNumber))
  rosters$playerLabel <- .normalize_roster_person(
    paste(rosters$playerLastName, rosters$playerFirstName, sep = ', ')
  )
  rosters
}

#' Resolve a roster player ID
#'
#' `.lookup_roster_player_id()` resolves a player ID from team, sweater number,
#' and optional name information within a roster lookup table.
#'
#' @param roster_lookup standardized roster lookup data.frame
#' @param team_id team ID
#' @param sweater_number sweater number
#' @param player_name optional player name
#' @returns integer player ID or `NA_integer_`
#' @keywords internal
.lookup_roster_player_id <- function(
  roster_lookup,
  team_id,
  sweater_number,
  player_name = NA_character_
) {
  if (!nrow(roster_lookup) || is.na(team_id)) {
    return(NA_integer_)
  }
  idx <- which(roster_lookup$teamId == team_id)
  if (!length(idx)) {
    return(NA_integer_)
  }
  if (!is.na(sweater_number)) {
    idx_num <- idx[roster_lookup$sweaterNumber[idx] == sweater_number]
    if (length(idx_num) == 1L) {
      return(as.integer(roster_lookup$playerId[idx_num]))
    }
    if (length(idx_num) > 1L && !is.na(player_name) && nzchar(player_name)) {
      target <- .normalize_roster_person(player_name)
      idx_name <- idx_num[roster_lookup$playerLabel[idx_num] == target]
      if (length(idx_name)) {
        return(as.integer(roster_lookup$playerId[idx_name[1L]]))
      }
    }
    if (length(idx_num)) {
      return(as.integer(roster_lookup$playerId[idx_num[1L]]))
    }
  }
  if (!is.na(player_name) && nzchar(player_name)) {
    target <- .normalize_roster_person(player_name)
    idx_name <- idx[roster_lookup$playerLabel[idx] == target]
    if (length(idx_name)) {
      return(as.integer(roster_lookup$playerId[idx_name[1L]]))
    }
  }
  NA_integer_
}

#' Parse an HTML period label
#'
#' `.parse_html_period_label()` converts HTML report period labels such as `OT`
#' and `SO` into integer period numbers.
#'
#' @param x HTML period label
#' @param is_playoffs logical; whether the game is a playoff game
#' @returns integer period number or `NA_integer_`
#' @keywords internal
.parse_html_period_label <- function(x, is_playoffs = FALSE) {
  x <- trimws(toupper(as.character(x)))
  if (!length(x) || is.na(x) || !nzchar(x)) {
    return(NA_integer_)
  }
  if (grepl('^[0-9]+$', x)) {
    return(as.integer(x))
  }
  if (x == 'OT') {
    return(4L)
  }
  if (grepl('^[0-9]+OT$', x)) {
    return(3L + as.integer(sub('OT$', '', x)))
  }
  if (x == 'SO') {
    return(if (is_playoffs) NA_integer_ else 5L)
  }
  NA_integer_
}

#' Parse an HTML elapsed clock
#'
#' `.parse_html_elapsed_clock()` extracts the elapsed `MM:SS` clock from an HTML
#' report time field and converts it to elapsed seconds in period.
#'
#' @param x HTML time field
#' @returns integer seconds elapsed in period or `NA_integer_`
#' @keywords internal
.parse_html_elapsed_clock <- function(x) {
  x <- trimws(as.character(x))
  if (!length(x) || is.na(x)) {
    return(NA_integer_)
  }
  m <- regexpr('^[0-9]{1,2}:[0-9]{2}', x)
  if (m[1] < 0L) {
    return(NA_integer_)
  }
  clock <- substr(x, m[1], m[1] + attr(m, 'match.length')[1] - 1L)
  parts <- strsplit(clock, ':', fixed = TRUE)[[1]]
  if (length(parts) != 2L) {
    return(NA_integer_)
  }
  mins <- suppressWarnings(as.integer(parts[1]))
  secs <- suppressWarnings(as.integer(parts[2]))
  if (is.na(mins) || is.na(secs)) {
    return(NA_integer_)
  }
  60L * mins + secs
}

#' Normalize an HTML team abbreviation
#'
#' `.normalize_html_team_abbrev()` uppercases and strips punctuation from HTML
#' team abbreviations so dotted legacy abbreviations still match.
#'
#' @param x character vector of team abbreviations
#' @returns normalized character vector
#' @keywords internal
.normalize_html_team_abbrev <- function(x) {
  x <- toupper(as.character(x))
  x <- gsub('[^A-Z]', '', x)
  x[is.na(x)] <- ''
  x
}

#' Map HTML event codes to event type keys
#'
#' `.html_event_code_to_type_desc()` converts HTML report event codes and select
#' description fallbacks into internal play-by-play event type keys.
#'
#' @param event_code HTML event code
#' @param description optional HTML event description
#' @returns event type key or `NA_character_`
#' @keywords internal
.html_event_code_to_type_desc <- function(event_code, description = NA_character_) {
  code_map <- c(
    FAC = 'faceoff',
    HIT = 'hit',
    SHOT = 'shot-on-goal',
    GIVE = 'giveaway',
    MISS = 'missed-shot',
    BLOCK = 'blocked-shot',
    PENL = 'penalty',
    GOAL = 'goal',
    TAKE = 'takeaway',
    DELPEN = 'delayed-penalty',
    DLPEN = 'delayed-penalty',
    DPEN = 'delayed-penalty'
  )
  event_code <- trimws(toupper(as.character(event_code)))
  out <- unname(code_map[event_code])
  if (!is.na(out)) {
    return(out)
  }
  desc <- trimws(toupper(as.character(description)))
  if (grepl('DELAYED PENALTY', desc, fixed = TRUE)) {
    return('delayed-penalty')
  }
  if (grepl('FAILED SHOT ATTEMPT', desc, fixed = TRUE)) {
    return('failed-shot-attempt')
  }
  NA_character_
}

#' Resolve the owner team from an HTML description
#'
#' `.html_desc_owner_team_id()` infers the event owner team ID from the leading
#' team abbreviation embedded in an HTML event description.
#'
#' @param description HTML event description
#' @param home_abbrev home-team abbreviation
#' @param away_abbrev away-team abbreviation
#' @param home_team_id home-team ID
#' @param away_team_id away-team ID
#' @returns integer team ID or `NA_integer_`
#' @keywords internal
.html_desc_owner_team_id <- function(
  description,
  home_abbrev,
  away_abbrev,
  home_team_id,
  away_team_id
) {
  desc <- trimws(toupper(as.character(description)))
  m <- regexpr('^[A-Z.]{2,4}', desc)
  if (m[1] < 0L) {
    return(NA_integer_)
  }
  team <- .normalize_html_team_abbrev(
    substr(desc, m[1], m[1] + attr(m, 'match.length')[1] - 1L)
  )
  home_norm <- .normalize_html_team_abbrev(home_abbrev)
  away_norm <- .normalize_html_team_abbrev(away_abbrev)
  if (
    nzchar(home_norm) &&
      (team == home_norm || startsWith(home_norm, team) || startsWith(team, home_norm))
  ) {
    return(as.integer(home_team_id))
  }
  if (
    nzchar(away_norm) &&
      (team == away_norm || startsWith(away_norm, team) || startsWith(team, away_norm))
  ) {
    return(as.integer(away_team_id))
  }
  NA_integer_
}

#' Extract team-tagged player references from HTML
#'
#' `.html_extract_tagged_players()` resolves explicitly team-tagged HTML player
#' references such as `TOR #42` into ordered team/player pairs.
#'
#' @param description HTML event description
#' @param home_team_id home-team ID
#' @param away_team_id away-team ID
#' @param home_abbrev home-team abbreviation
#' @param away_abbrev away-team abbreviation
#' @param roster_lookup standardized roster lookup data.frame
#' @returns data.frame of ordered team/player references
#' @keywords internal
.html_extract_tagged_players <- function(
  description,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev,
  roster_lookup
) {
  desc <- toupper(as.character(description))
  matches <- regmatches(
    desc,
    gregexpr('([A-Z.]{2,4})\\s+#([0-9]{1,2})', desc, perl = TRUE)
  )[[1]]
  if (!length(matches)) {
    return(data.frame(teamId = integer(), playerId = integer(), stringsAsFactors = FALSE))
  }
  home_norm <- .normalize_html_team_abbrev(home_abbrev)
  away_norm <- .normalize_html_team_abbrev(away_abbrev)
  team_codes <- .normalize_html_team_abbrev(sub('\\s+#.*$', '', matches))
  sweaters <- suppressWarnings(as.integer(sub('^.*#([0-9]{1,2})$', '\\1', matches, perl = TRUE)))
  team_ids <- ifelse(
    nzchar(home_norm) &
      (team_codes == home_norm | startsWith(home_norm, team_codes) | startsWith(team_codes, home_norm)),
    as.integer(home_team_id),
    ifelse(
      nzchar(away_norm) &
        (team_codes == away_norm | startsWith(away_norm, team_codes) | startsWith(team_codes, away_norm)),
      as.integer(away_team_id),
      NA_integer_
    )
  )
  player_ids <- vapply(
    seq_along(sweaters),
    function(i) {
      if (is.na(team_ids[i])) {
        return(NA_integer_)
      }
      .lookup_roster_player_id(roster_lookup, team_ids[i], sweaters[i])
    },
    integer(1)
  )
  data.frame(
    teamId = as.integer(team_ids),
    playerId = as.integer(player_ids),
    stringsAsFactors = FALSE
  )
}

#' Extract actor player IDs from an HTML event
#'
#' `.html_extract_actor_player_ids()` resolves primary, secondary, and tertiary
#' actor player IDs from HTML event descriptions using the roster lookup.
#'
#' @param description HTML event description
#' @param type_desc_key internal event type key
#' @param owner_team_id owner team ID
#' @param home_team_id home-team ID
#' @param away_team_id away-team ID
#' @param home_abbrev home-team abbreviation
#' @param away_abbrev away-team abbreviation
#' @param roster_lookup standardized roster lookup data.frame
#' @returns named list of actor player IDs
#' @keywords internal
.html_extract_actor_player_ids <- function(
  description,
  type_desc_key,
  owner_team_id,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev,
  roster_lookup
) {
  nums_raw <- regmatches(
    description,
    gregexpr('#[0-9]{1,2}', description, perl = TRUE)
  )[[1]]
  sweaters <- suppressWarnings(as.integer(sub('^#', '', nums_raw)))
  opponent_team_id <- if (!is.na(owner_team_id)) {
    if (owner_team_id == home_team_id) {
      away_team_id
    } else if (owner_team_id == away_team_id) {
      home_team_id
    } else {
      NA_integer_
    }
  } else {
    NA_integer_
  }
  tagged <- .html_extract_tagged_players(
    description = description,
    home_team_id = home_team_id,
    away_team_id = away_team_id,
    home_abbrev = home_abbrev,
    away_abbrev = away_abbrev,
    roster_lookup = roster_lookup
  )
  owner_tagged <- if (!is.na(owner_team_id)) {
    tagged[tagged$teamId == owner_team_id, , drop = FALSE]
  } else {
    tagged[0, , drop = FALSE]
  }
  opponent_tagged <- if (!is.na(opponent_team_id)) {
    tagged[tagged$teamId == opponent_team_id, , drop = FALSE]
  } else {
    tagged[0, , drop = FALSE]
  }
  owner_pick_fallback <- function(pos) {
    if (length(sweaters) < pos || is.na(owner_team_id)) {
      return(NA_integer_)
    }
    .lookup_roster_player_id(roster_lookup, owner_team_id, sweaters[pos])
  }
  opp_pick_fallback <- function(pos) {
    if (length(sweaters) < pos || is.na(opponent_team_id)) {
      return(NA_integer_)
    }
    .lookup_roster_player_id(roster_lookup, opponent_team_id, sweaters[pos])
  }
  owner_pick <- function(pos) {
    if (nrow(owner_tagged) >= pos) {
      picked <- owner_tagged$playerId[pos]
      if (!is.na(picked)) {
        return(as.integer(picked))
      }
    }
    owner_pick_fallback(pos)
  }
  opp_pick <- function(pos) {
    if (nrow(opponent_tagged) >= pos) {
      picked <- opponent_tagged$playerId[pos]
      if (!is.na(picked)) {
        return(as.integer(picked))
      }
    }
    opp_pick_fallback(pos)
  }
  primary <- secondary <- tertiary <- NA_integer_
  if (is.na(type_desc_key)) {
    return(list(primaryPlayerId = primary, secondaryPlayerId = secondary, tertiaryPlayerId = tertiary))
  }
  switch(
    type_desc_key,
    faceoff = {
      primary <- owner_pick(1L)
      secondary <- opp_pick(1L)
    },
    hit = {
      primary <- owner_pick(1L)
      secondary <- opp_pick(1L)
    },
    `shot-on-goal` = {
      primary <- owner_pick(1L)
    },
    `missed-shot` = {
      primary <- owner_pick(1L)
    },
    `failed-shot-attempt` = {
      primary <- owner_pick(1L)
    },
    `blocked-shot` = {
      primary <- owner_pick(1L)
      secondary <- opp_pick(1L)
    },
    penalty = {
      primary <- owner_pick(1L)
      secondary <- opp_pick(1L)
    },
    `delayed-penalty` = {
      primary <- owner_pick(1L)
      secondary <- opp_pick(1L)
    },
    goal = {
      primary <- owner_pick(1L)
      secondary <- owner_pick(2L)
      tertiary <- owner_pick(3L)
    },
    giveaway = {
      primary <- owner_pick(1L)
    },
    takeaway = {
      primary <- owner_pick(1L)
    }
  )
  list(
    primaryPlayerId = as.integer(primary),
    secondaryPlayerId = as.integer(secondary),
    tertiaryPlayerId = as.integer(tertiary)
  )
}

#' Parse an HTML on-ice cell
#'
#' `.parse_html_on_ice_cell()` resolves a goalie and skater player-ID list from
#' a single team on-ice HTML cell.
#'
#' @param text on-ice cell text
#' @param team_id team ID for the cell
#' @param roster_lookup standardized roster lookup data.frame
#' @returns named list containing goalie and skater player IDs
#' @keywords internal
.parse_html_on_ice_cell <- function(text, team_id, roster_lookup) {
  txt <- gsub('\u00A0', ' ', as.character(text), fixed = TRUE)
  txt <- gsub('\\s+', ' ', txt)
  txt <- trimws(txt)
  if (!nzchar(txt)) {
    return(list(goaliePlayerId = NA_integer_, skaterPlayerIds = integer()))
  }
  tokens <- regmatches(txt, gregexpr('([0-9]{1,2})\\s+([A-Z]{1,3})', txt, perl = TRUE))[[1]]
  if (!length(tokens)) {
    return(list(goaliePlayerId = NA_integer_, skaterPlayerIds = integer()))
  }
  sweaters <- suppressWarnings(as.integer(sub('\\s+.*$', '', tokens)))
  pos <- sub('^[0-9]{1,2}\\s+', '', tokens)
  player_ids <- vapply(
    sweaters,
    function(sw) .lookup_roster_player_id(roster_lookup, team_id, sw),
    integer(1)
  )
  goalie_idx <- which(pos == 'G')
  goalie_id <- if (length(goalie_idx)) player_ids[goalie_idx[1L]] else NA_integer_
  skater_ids <- player_ids[setdiff(seq_along(player_ids), goalie_idx)]
  list(
    goaliePlayerId = as.integer(goalie_id),
    skaterPlayerIds = as.integer(skater_ids)
  )
}

#' Parse an HTML play-by-play document
#'
#' `.parse_html_pbp_doc()` extracts event rows, actor IDs, and on-ice player
#' lists from an NHL HTML play-by-play report.
#'
#' @param doc parsed HTML document
#' @param roster_lookup standardized roster lookup data.frame
#' @param home_team_id home-team ID
#' @param away_team_id away-team ID
#' @param home_abbrev home-team abbreviation
#' @param away_abbrev away-team abbreviation
#' @param is_playoffs logical; whether the game is a playoff game
#' @returns data.frame of parsed HTML play-by-play rows
#' @keywords internal
.parse_html_pbp_doc <- function(
  doc,
  roster_lookup,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev,
  is_playoffs = FALSE
) {
  rows <- xml2::xml_find_all(doc, './/tr')
  out <- vector('list', length(rows))
  n_out <- 0L
  for (row in rows) {
    cells <- xml2::xml_find_all(row, './th|./td')
    if (length(cells) < 8L) {
      next
    }
    txt <- xml2::xml_text(cells, trim = TRUE)
    txt <- gsub('\u00A0', ' ', txt, fixed = TRUE)
    txt <- trimws(txt)
    if (!length(txt) || !grepl('^[0-9]+$', txt[1])) {
      next
    }
    period_number <- .parse_html_period_label(txt[2], is_playoffs = is_playoffs)
    event_code <- trimws(toupper(txt[5]))
    description <- txt[6]
    type_desc_key <- .html_event_code_to_type_desc(event_code, description)
    owner_team_id <- .html_desc_owner_team_id(
      description,
      home_abbrev = home_abbrev,
      away_abbrev = away_abbrev,
      home_team_id = home_team_id,
      away_team_id = away_team_id
    )
    actors <- .html_extract_actor_player_ids(
      description = description,
      type_desc_key = type_desc_key,
      owner_team_id = owner_team_id,
      home_team_id = home_team_id,
      away_team_id = away_team_id,
      home_abbrev = home_abbrev,
      away_abbrev = away_abbrev,
      roster_lookup = roster_lookup
    )
    away_on_ice <- .parse_html_on_ice_cell(
      txt[length(txt) - 1L],
      team_id = away_team_id,
      roster_lookup = roster_lookup
    )
    home_on_ice <- .parse_html_on_ice_cell(
      txt[length(txt)],
      team_id = home_team_id,
      roster_lookup = roster_lookup
    )
    n_out <- n_out + 1L
    out[[n_out]] <- data.frame(
      htmlEventNumber = suppressWarnings(as.integer(txt[1])),
      period = as.integer(period_number),
      strengthCodeHtml = trimws(txt[3]),
      secondsElapsedInPeriod = .parse_html_elapsed_clock(txt[4]),
      htmlEventCode = event_code,
      typeDescKey = type_desc_key,
      description = description,
      ownerTeamId = as.integer(owner_team_id),
      primaryPlayerId = actors$primaryPlayerId,
      secondaryPlayerId = actors$secondaryPlayerId,
      tertiaryPlayerId = actors$tertiaryPlayerId,
      homeGoaliePlayerId = home_on_ice$goaliePlayerId,
      awayGoaliePlayerId = away_on_ice$goaliePlayerId,
      stringsAsFactors = FALSE
    )
    out[[n_out]]$homeSkaterPlayerIds <- list(home_on_ice$skaterPlayerIds)
    out[[n_out]]$awaySkaterPlayerIds <- list(away_on_ice$skaterPlayerIds)
  }
  if (!n_out) {
    return(data.frame())
  }
  out <- do.call(rbind, out[seq_len(n_out)])
  out[order(out$period, out$secondsElapsedInPeriod, out$htmlEventNumber), ]
}

#' Check HTML on-ice counts against a situation code
#'
#' `.html_on_ice_matches_situation_code()` validates that a parsed HTML on-ice
#' row is compatible with the source API situation code. Missing or unparseable
#' situation codes are treated as compatible.
#'
#' @param situation_code raw API situation code
#' @param home_goalie_player_id parsed home goalie player ID
#' @param away_goalie_player_id parsed away goalie player ID
#' @param home_skater_player_ids parsed home skater player IDs
#' @param away_skater_player_ids parsed away skater player IDs
#' @returns logical scalar
#' @keywords internal
.html_on_ice_matches_situation_code <- function(
  situation_code,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
) {
  sc_parts <- .parse_situation_code_components(situation_code)
  if (!nrow(sc_parts)) {
    return(TRUE)
  }
  api_sig <- as.integer(sc_parts[1L, c(
    'awayGoalieCount',
    'awaySkaterCount',
    'homeSkaterCount',
    'homeGoalieCount'
  )])
  if (any(is.na(api_sig))) {
    return(TRUE)
  }
  html_sig <- c(
    if (is.na(away_goalie_player_id)) 0L else 1L,
    length(away_skater_player_ids),
    length(home_skater_player_ids),
    if (is.na(home_goalie_player_id)) 0L else 1L
  )
  identical(as.integer(html_sig), api_sig)
}

#' Build the HTML play-by-play report URL
#'
#' `.html_pbp_report_url()` returns the NHL HTML play-by-play report URL for a
#' game.
#'
#' @param game game ID
#' @returns character scalar URL
#' @keywords internal
.html_pbp_report_url <- function(game) {
  season <- paste0(game %/% 1e6, game %/% 1e6 + 1L)
  game_code <- sprintf('%06d', game %% 10^6)
  sprintf(
    'https://www.nhl.com/scores/htmlreports/%s/PL%s.HTM',
    season,
    game_code
  )
}

#' Parse an HTML on-ice play-by-play response
#'
#' `.parse_html_pbp_response()` converts an already-fetched HTML play-by-play
#' response into parsed on-ice rows.
#'
#' @param resp httr2 response object for the HTML report
#' @inheritParams .fetch_html_pbp_on_ice
#' @returns data.frame of parsed HTML play-by-play rows
#' @keywords internal
.parse_html_pbp_response <- function(
  resp,
  game,
  rosters,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev
) {
  doc <- xml2::read_html(httr2::resp_body_raw(resp))
  .parse_html_pbp_doc(
    doc = doc,
    roster_lookup = .build_game_roster_lookup(rosters),
    home_team_id = as.integer(home_team_id),
    away_team_id = as.integer(away_team_id),
    home_abbrev = home_abbrev,
    away_abbrev = away_abbrev,
    is_playoffs = game %/% 1e4 %% 1e2 == 3L
  )
}

#' Fetch HTML on-ice play-by-play rows
#'
#' `.fetch_html_pbp_on_ice()` downloads an NHL HTML play-by-play report and
#' parses its on-ice content into a structured data frame.
#'
#' @param game game ID
#' @param rosters roster data.frame from play-by-play metadata
#' @param home_team_id home-team ID
#' @param away_team_id away-team ID
#' @param home_abbrev home-team abbreviation
#' @param away_abbrev away-team abbreviation
#' @returns data.frame of parsed HTML play-by-play rows
#' @keywords internal
.fetch_html_pbp_on_ice <- function(
  game,
  rosters,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev
) {
  resp <- httr2::req_perform(
    .html_report_request(.html_pbp_report_url(as.integer(game)))
  )
  .parse_html_pbp_response(
    resp = resp,
    game = as.integer(game),
    home_team_id = as.integer(home_team_id),
    away_team_id = as.integer(away_team_id),
    rosters = rosters,
    home_abbrev = home_abbrev,
    away_abbrev = away_abbrev
  )
}

#' Build API-side HTML matching features
#'
#' `.build_api_html_match_table()` derives the API-side matching features used
#' to align HTML play-by-play rows back to API play-by-play rows.
#'
#' @param play_by_play data.frame play-by-play object
#' @returns data.frame of API-side matching features
#' @keywords internal
.build_api_html_match_table <- function(play_by_play) {
  n <- nrow(play_by_play)
  out <- data.frame(
    apiIndex = seq_len(n),
    apiSeq = seq_len(n),
    periodNumber = if ('periodNumber' %in% names(play_by_play)) as.integer(play_by_play$periodNumber) else rep(NA_integer_, n),
    secondsElapsedInPeriod = if ('secondsElapsedInPeriod' %in% names(play_by_play)) as.integer(play_by_play$secondsElapsedInPeriod) else rep(NA_integer_, n),
    eventTypeDescKey = if ('eventTypeDescKey' %in% names(play_by_play)) as.character(play_by_play$eventTypeDescKey) else rep(NA_character_, n),
    ownerTeamId = if ('eventOwnerTeamId' %in% names(play_by_play)) as.integer(play_by_play$eventOwnerTeamId) else rep(NA_integer_, n),
    situationCode = if ('situationCode' %in% names(play_by_play)) as.character(play_by_play$situationCode) else rep(NA_character_, n),
    primaryPlayerId = rep(NA_integer_, n),
    secondaryPlayerId = rep(NA_integer_, n),
    tertiaryPlayerId = rep(NA_integer_, n),
    stringsAsFactors = FALSE
  )
  for (i in seq_len(n)) {
    type_desc_key <- out$eventTypeDescKey[i]
    if (is.na(type_desc_key)) {
      next
    }
    if (type_desc_key == 'faceoff') {
      out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'winningPlayerId')[i]
      out$secondaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'losingPlayerId')[i]
    } else if (type_desc_key == 'hit') {
      out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'hittingPlayerId')[i]
      out$secondaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'hitteePlayerId')[i]
    } else if (type_desc_key %in% c('shot-on-goal', 'missed-shot', 'failed-shot-attempt')) {
      shooter <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[i]
      scorer <- .on_ice_int_col(play_by_play, 'scoringPlayerId')[i]
      out$primaryPlayerId[i] <- ifelse(!is.na(scorer), scorer, shooter)
    } else if (type_desc_key == 'blocked-shot') {
      out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[i]
      out$secondaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'blockingPlayerId')[i]
    } else if (type_desc_key == 'penalty' || type_desc_key == 'delayed-penalty') {
      out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'committedByPlayerId')[i]
      if (is.na(out$primaryPlayerId[i])) {
        out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'playerId')[i]
      }
      out$secondaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'drawnByPlayerId')[i]
    } else if (type_desc_key == 'goal') {
      out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'scoringPlayerId')[i]
      if (is.na(out$primaryPlayerId[i])) {
        out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[i]
      }
      out$secondaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'assist1PlayerId')[i]
      out$tertiaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'assist2PlayerId')[i]
    } else if (type_desc_key %in% c('giveaway', 'takeaway')) {
      out$primaryPlayerId[i] <- .on_ice_int_col(play_by_play, 'playerId')[i]
    }
  }
  out <- out[out$eventTypeDescKey %in% .supported_strength_event_types(), , drop = FALSE]
  out$apiSeq <- seq_len(nrow(out))
  out
}

#' Score HTML-to-API matching candidates
#'
#' `.score_html_api_candidates()` scores candidate API rows for a single HTML
#' row using time, type, team, player, and local sequence agreement.
#'
#' @param api API-side matching table
#' @param html HTML-side matching table
#' @param cand integer candidate API indices
#' @param h_idx HTML row index being scored
#' @param last_api_seq optional last matched API sequence number
#' @returns named numeric vector of candidate scores
#' @keywords internal
.score_html_api_candidates <- function(api, html, cand, h_idx, last_api_seq = NULL) {
  if (!length(cand)) {
    return(stats::setNames(numeric(), character()))
  }
  sec_diff <- abs(api$secondsElapsedInPeriod[cand] - html$secondsElapsedInPeriod[h_idx])
  if (any(sec_diff == 0L, na.rm = TRUE)) {
    cand <- cand[sec_diff == 0L]
    sec_diff <- sec_diff[sec_diff == 0L]
  } else if (any(sec_diff <= 1L, na.rm = TRUE)) {
    cand <- cand[sec_diff <= 1L]
    sec_diff <- sec_diff[sec_diff <= 1L]
  } else if (any(sec_diff <= 2L, na.rm = TRUE)) {
    cand <- cand[sec_diff <= 2L]
    sec_diff <- sec_diff[sec_diff <= 2L]
  } else {
    return(stats::setNames(numeric(), character()))
  }
  scores <- rep(0, length(cand))
  scores <- scores + ifelse(sec_diff == 0L, 500, ifelse(sec_diff <= 1L, 300, 150))
  exact_type <- api$eventTypeDescKey[cand] == html$typeDescKey[h_idx]
  scores <- scores + ifelse(exact_type, 250, 180)
  if (!is.na(html$ownerTeamId[h_idx])) {
    scores <- scores + ifelse(
      !is.na(api$ownerTeamId[cand]) & api$ownerTeamId[cand] == html$ownerTeamId[h_idx],
      140,
      -40
    )
  }
  if (!is.na(html$primaryPlayerId[h_idx])) {
    scores <- scores + ifelse(
      !is.na(api$primaryPlayerId[cand]) & api$primaryPlayerId[cand] == html$primaryPlayerId[h_idx],
      160,
      0
    )
  }
  if (!is.na(html$secondaryPlayerId[h_idx])) {
    scores <- scores + ifelse(
      !is.na(api$secondaryPlayerId[cand]) & api$secondaryPlayerId[cand] == html$secondaryPlayerId[h_idx],
      100,
      0
    )
  }
  if (!is.na(html$tertiaryPlayerId[h_idx])) {
    scores <- scores + ifelse(
      !is.na(api$tertiaryPlayerId[cand]) & api$tertiaryPlayerId[cand] == html$tertiaryPlayerId[h_idx],
      50,
      0
    )
  }
  if (all(c(
    'homeGoaliePlayerId',
    'awayGoaliePlayerId',
    'homeSkaterPlayerIds',
    'awaySkaterPlayerIds'
  ) %in% names(html))) {
    html_home_skaters <- unlist(html$homeSkaterPlayerIds[h_idx], use.names = FALSE)
    html_away_skaters <- unlist(html$awaySkaterPlayerIds[h_idx], use.names = FALSE)
    situation_score <- vapply(
      cand,
      function(idx) {
        sc <- api$situationCode[idx]
        sc_parse <- .normalize_situation_code_for_parse(sc)
        if (is.na(sc_parse)) {
          return(0)
        }
        if (.html_on_ice_matches_situation_code(
          situation_code = sc,
          home_goalie_player_id = html$homeGoaliePlayerId[h_idx],
          away_goalie_player_id = html$awayGoaliePlayerId[h_idx],
          home_skater_player_ids = html_home_skaters,
          away_skater_player_ids = html_away_skaters
        )) {
          120
        } else {
          -60
        }
      },
      numeric(1)
    )
    scores <- scores + situation_score
  }
  scores <- scores - 8 * abs(api$apiSeq[cand] - html$htmlSeq[h_idx])
  if (!is.null(last_api_seq) && !is.na(last_api_seq)) {
    scores <- scores - 75 * pmax(0L, last_api_seq - api$apiSeq[cand])
  }
  names(scores) <- as.character(cand)
  scores
}

#' Match HTML play-by-play rows to API rows
#'
#' `.match_html_pbp_to_api()` aligns parsed HTML play-by-play rows back to API
#' play-by-play rows using exact keys, greedy scoring, and a reciprocal-best
#' fallback for duplicate clusters.
#'
#' @param play_by_play data.frame API play-by-play object
#' @param html_rows data.frame parsed HTML play-by-play rows
#' @returns data.frame of matched HTML rows with `apiIndex`
#' @keywords internal
.match_html_pbp_to_api <- function(play_by_play, html_rows) {
  api <- .build_api_html_match_table(play_by_play)
  if (!nrow(api) || is.null(html_rows) || !nrow(html_rows)) {
    return(data.frame())
  }
  html <- html_rows[html_rows$typeDescKey %in% .supported_strength_event_types(), , drop = FALSE]
  if (!nrow(html)) {
    return(data.frame())
  }
  html$htmlSeq <- seq_len(nrow(html))
  html$apiIndex <- NA_integer_
  api_taken <- rep(FALSE, nrow(api))

  html_key <- paste(
    html$period,
    html$secondsElapsedInPeriod,
    html$typeDescKey,
    ifelse(is.na(html$ownerTeamId), 'NA', html$ownerTeamId),
    sep = '|'
  )
  api_key <- paste(
    api$periodNumber,
    api$secondsElapsedInPeriod,
    api$eventTypeDescKey,
    ifelse(is.na(api$ownerTeamId), 'NA', api$ownerTeamId),
    sep = '|'
  )
  for (key in intersect(unique(html_key), unique(api_key))) {
    h_idx <- which(is.na(html$apiIndex) & html_key == key)
    a_idx <- which(!api_taken & api_key == key)
    if (length(h_idx) == 1L && length(a_idx) == 1L) {
      html$apiIndex[h_idx] <- api$apiIndex[a_idx]
      api_taken[a_idx] <- TRUE
    }
  }

  last_api_seq <- 0L
  for (h_idx in which(is.na(html$apiIndex))) {
    cand <- which(
      !api_taken &
        api$periodNumber == html$period[h_idx] &
        (
          api$eventTypeDescKey == html$typeDescKey[h_idx] |
            (
              api$eventTypeDescKey == 'failed-shot-attempt' &
                html$typeDescKey[h_idx] %in% c('shot-on-goal', 'missed-shot', 'goal')
            )
        )
    )
    if (!length(cand)) {
      next
    }
    scores <- .score_html_api_candidates(api, html, cand, h_idx, last_api_seq = last_api_seq)
    if (!length(scores) || !any(is.finite(scores))) {
      next
    }
    cand <- as.integer(names(scores))
    best_pos <- which.max(scores)
    if (length(best_pos) && is.finite(scores[best_pos]) && scores[best_pos] >= 180) {
      html$apiIndex[h_idx] <- api$apiIndex[cand[best_pos]]
      api_taken[cand[best_pos]] <- TRUE
      last_api_seq <- api$apiSeq[cand[best_pos]]
    }
  }

  unmatched_html <- which(is.na(html$apiIndex))
  unmatched_api <- which(!api_taken)
  if (length(unmatched_html) && length(unmatched_api)) {
    html_best <- rep(NA_integer_, length(unmatched_html))
    html_best_score <- rep(-Inf, length(unmatched_html))
    for (i in seq_along(unmatched_html)) {
      h_idx <- unmatched_html[i]
      cand <- unmatched_api[
        api$periodNumber[unmatched_api] == html$period[h_idx] &
          (
            api$eventTypeDescKey[unmatched_api] == html$typeDescKey[h_idx] |
              (
                api$eventTypeDescKey[unmatched_api] == 'failed-shot-attempt' &
                  html$typeDescKey[h_idx] %in% c('shot-on-goal', 'missed-shot', 'goal')
              )
          )
      ]
      scores <- .score_html_api_candidates(api, html, cand, h_idx)
      if (!length(scores) || !any(is.finite(scores))) {
        next
      }
      best_pos <- which.max(scores)
      html_best[i] <- as.integer(names(scores)[best_pos])
      html_best_score[i] <- scores[best_pos]
    }

    api_best <- rep(NA_integer_, length(unmatched_api))
    api_best_score <- rep(-Inf, length(unmatched_api))
    for (i in seq_along(unmatched_api)) {
      a_idx <- unmatched_api[i]
      cand <- unmatched_html[
        html$period[unmatched_html] == api$periodNumber[a_idx] &
          (
            api$eventTypeDescKey[a_idx] == html$typeDescKey[unmatched_html] |
              (
                api$eventTypeDescKey[a_idx] == 'failed-shot-attempt' &
                  html$typeDescKey[unmatched_html] %in% c('shot-on-goal', 'missed-shot', 'goal')
              )
          )
      ]
      if (!length(cand)) {
        next
      }
      scores <- vapply(
        cand,
        function(h_idx) {
          out <- .score_html_api_candidates(api, html, a_idx, h_idx)
          if (!length(out)) {
            return(-Inf)
          }
          unname(out[[1L]])
        },
        numeric(1)
      )
      if (!any(is.finite(scores))) {
        next
      }
      best_pos <- which.max(scores)
      api_best[i] <- cand[best_pos]
      api_best_score[i] <- scores[best_pos]
    }

    for (i in seq_along(unmatched_html)) {
      h_idx <- unmatched_html[i]
      a_idx <- html_best[i]
      if (is.na(a_idx) || html_best_score[i] < 180) {
        next
      }
      api_slot <- match(a_idx, unmatched_api)
      if (is.na(api_slot) || is.na(api_best[api_slot])) {
        next
      }
      if (api_best[api_slot] != h_idx || api_best_score[api_slot] < 180) {
        next
      }
      html$apiIndex[h_idx] <- api$apiIndex[a_idx]
      api_taken[a_idx] <- TRUE
    }
  }

  html[!is.na(html$apiIndex), , drop = FALSE]
}

#' Override derived strength context from resolved on-ice players
#'
#' `.override_strength_context_from_html()` updates the derived strength/count
#' columns from a resolved HTML on-ice signature while leaving the raw
#' `situationCode` column untouched.
#'
#' @param play_by_play data.frame play-by-play object
#' @param idx row index to update
#' @param home_goalie resolved home goalie player ID
#' @param away_goalie resolved away goalie player ID
#' @param home_skaters integer vector of resolved home skater IDs
#' @param away_skaters integer vector of resolved away skater IDs
#' @returns data.frame with derived strength context updated for `idx`
#' @keywords internal
.override_strength_context_from_html <- function(
  play_by_play,
  idx,
  home_goalie,
  away_goalie,
  home_skaters,
  away_skaters
) {
  home_skater_count <- as.integer(length(home_skaters))
  away_skater_count <- as.integer(length(away_skaters))
  home_is_empty_net <- is.na(home_goalie)
  away_is_empty_net <- is.na(away_goalie)

  play_by_play$homeIsEmptyNet[idx] <- home_is_empty_net
  play_by_play$awayIsEmptyNet[idx] <- away_is_empty_net
  play_by_play$homeSkaterCount[idx] <- home_skater_count
  play_by_play$awaySkaterCount[idx] <- away_skater_count

  if (play_by_play$isHome[idx] %in% TRUE) {
    skaters_for <- home_skater_count
    skaters_against <- away_skater_count
    empty_for <- home_is_empty_net
    empty_against <- away_is_empty_net
    men_for <- skaters_for + if (home_is_empty_net) 0L else 1L
    men_against <- skaters_against + if (away_is_empty_net) 0L else 1L
  } else if (play_by_play$isHome[idx] %in% FALSE) {
    skaters_for <- away_skater_count
    skaters_against <- home_skater_count
    empty_for <- away_is_empty_net
    empty_against <- home_is_empty_net
    men_for <- skaters_for + if (away_is_empty_net) 0L else 1L
    men_against <- skaters_against + if (home_is_empty_net) 0L else 1L
  } else {
    skaters_for <- NA_integer_
    skaters_against <- NA_integer_
    empty_for <- NA
    empty_against <- NA
    men_for <- NA_integer_
    men_against <- NA_integer_
  }

  play_by_play$isEmptyNetFor[idx] <- empty_for
  play_by_play$isEmptyNetAgainst[idx] <- empty_against
  play_by_play$skaterCountFor[idx] <- skaters_for
  play_by_play$skaterCountAgainst[idx] <- skaters_against
  play_by_play$manDifferential[idx] <- if (
    is.na(men_for) || is.na(men_against)
  ) {
    NA_integer_
  } else {
    men_for - men_against
  }
  play_by_play$strengthState[idx] <- if (is.na(play_by_play$manDifferential[idx])) {
    NA_character_
  } else if (play_by_play$manDifferential[idx] > 0L) {
    'power-play'
  } else if (play_by_play$manDifferential[idx] < 0L) {
    'penalty-kill'
  } else {
    'even-strength'
  }

  play_by_play
}

#' Reconstruct skater counts from active penalties
#'
#' `.reconstruct_skater_counts_from_penalties()` walks a cleaned play-by-play in
#' event order and reconstructs home/away skater counts from active strength-
#' affecting penalties. The reconstruction intentionally targets regulation and
#' playoff-strength states; regular-season overtime is left unsupported because
#' its 3-on-3 penalty rules require a separate ruleset.
#'
#' @param play_by_play data.frame cleaned internal play-by-play object
#' @returns data.frame with reconstructed home/away skater counts and metadata
#' @keywords internal
.reconstruct_skater_counts_from_penalties <- function(play_by_play) {
  n <- nrow(play_by_play)
  out <- data.frame(
    homeSkaterCountRecon = rep(NA_integer_, n),
    awaySkaterCountRecon = rep(NA_integer_, n),
    strengthReconAvailable = rep(FALSE, n),
    activePenaltyCountHome = rep(NA_integer_, n),
    activePenaltyCountAway = rep(NA_integer_, n),
    stringsAsFactors = FALSE
  )
  if (
      !n ||
      !all(c(
        'gameId', 'gameTypeId', 'periodNumber', 'sortOrder', 'secondsElapsedInGame',
        'eventTypeDescKey', 'isHome'
      ) %in% names(play_by_play))
  ) {
    return(out)
  }

  team_side <- ifelse(
    play_by_play$isHome %in% TRUE,
    'home',
    ifelse(play_by_play$isHome %in% FALSE, 'away', NA_character_)
  )
  penalty_type_code <- toupper(as.character(
    if ('penaltyTypeCode' %in% names(play_by_play)) play_by_play$penaltyTypeCode else NA_character_
  ))
  penalty_duration <- suppressWarnings(as.integer(
    if ('penaltyDuration' %in% names(play_by_play)) play_by_play$penaltyDuration else NA_integer_
  ))
  type_desc_key <- as.character(play_by_play$eventTypeDescKey)
  period <- suppressWarnings(as.integer(play_by_play$periodNumber))
  seconds_elapsed_in_game <- suppressWarnings(as.integer(play_by_play$secondsElapsedInGame))
  sort_order <- suppressWarnings(as.integer(play_by_play$sortOrder))
  game_type_id <- suppressWarnings(as.integer(play_by_play$gameTypeId))

  is_strength_penalty <- function(i) {
    if (!identical(type_desc_key[i], 'penalty') || is.na(team_side[i])) {
      return(FALSE)
    }
    dur <- penalty_duration[i]
    type_code <- penalty_type_code[i]
    if (is.na(dur) || dur <= 0L || dur >= 10L) {
      return(FALSE)
    }
    !is.na(type_code) && type_code %in% c('MIN', 'BEN', 'MAJ')
  }

  new_penalty <- function(team, start_time, type_code, duration_min) {
    durations <- if (
      !is.na(type_code) &&
        type_code %in% c('MIN', 'BEN') &&
        !is.na(duration_min) &&
        duration_min == 4L
    ) {
      c(120L, 120L)
    } else {
      rep(as.integer(duration_min) * 60L, 1L)
    }
    list(
      team = team,
      releasable = !is.na(type_code) && type_code %in% c('MIN', 'BEN'),
      segments = durations,
      segmentStart = as.integer(start_time),
      segmentEnd = as.integer(start_time) + durations[1L]
    )
  }

  advance_penalties <- function(active, current_time, inclusive = FALSE) {
    if (!length(active) || is.na(current_time)) {
      return(active)
    }
    keep <- rep(TRUE, length(active))
    for (j in seq_along(active)) {
      pen <- active[[j]]
      while (
        length(pen$segments) &&
          !is.na(pen$segmentEnd) &&
          if (inclusive) pen$segmentEnd <= current_time else pen$segmentEnd < current_time
      ) {
        if (length(pen$segments) == 1L) {
          pen$segments <- integer()
          break
        }
        segment_end <- pen$segmentEnd
        pen$segments <- pen$segments[-1L]
        pen$segmentStart <- segment_end
        pen$segmentEnd <- segment_end + pen$segments[1L]
      }
      if (!length(pen$segments)) {
        keep[j] <- FALSE
      } else {
        active[[j]] <- pen
      }
    }
    active[keep]
  }

  release_after_goal <- function(active, scoring_team, current_time, home_count, away_count) {
    if (
      !length(active) ||
        is.na(scoring_team) ||
        is.na(current_time) ||
        is.na(home_count) ||
        is.na(away_count) ||
        identical(home_count, away_count)
    ) {
      return(active)
    }
    penalized_team <- if (identical(scoring_team, 'home') && home_count > away_count) {
      'away'
    } else if (identical(scoring_team, 'away') && away_count > home_count) {
      'home'
    } else {
      NA_character_
    }
    if (is.na(penalized_team)) {
      return(active)
    }
    candidate_idx <- which(vapply(
      active,
      function(pen) identical(pen$team, penalized_team) && isTRUE(pen$releasable),
      logical(1)
    ))
    if (!length(candidate_idx)) {
      return(active)
    }
    release_idx <- candidate_idx[which.min(vapply(
      active[candidate_idx],
      function(pen) pen$segmentEnd,
      numeric(1)
    ))]
    pen <- active[[release_idx]]
    if (length(pen$segments) == 1L) {
      active <- active[-release_idx]
    } else {
      pen$segments <- pen$segments[-1L]
      pen$segmentStart <- as.integer(current_time)
      pen$segmentEnd <- as.integer(current_time) + pen$segments[1L]
      active[[release_idx]] <- pen
    }
    active
  }

  major_pair_skip <- rep(FALSE, n)
  for (g in unique(play_by_play$gameId)) {
    idx_game <- which(play_by_play$gameId == g)
    idx_pen <- idx_game[vapply(idx_game, is_strength_penalty, logical(1))]
    if (!length(idx_pen)) {
      next
    }
    key <- paste(seconds_elapsed_in_game[idx_pen], penalty_duration[idx_pen], sep = '|')
    for (grp in split(idx_pen, key)) {
      if (
        length(grp) >= 2L &&
          all(!is.na(penalty_type_code[grp]) & penalty_type_code[grp] == 'MAJ') &&
          any(team_side[grp] == 'home') &&
          any(team_side[grp] == 'away')
      ) {
        home_grp <- grp[team_side[grp] == 'home']
        away_grp <- grp[team_side[grp] == 'away']
        n_pair <- min(length(home_grp), length(away_grp))
        if (n_pair > 0L) {
          major_pair_skip[home_grp[seq_len(n_pair)]] <- TRUE
          major_pair_skip[away_grp[seq_len(n_pair)]] <- TRUE
        }
      }
    }
  }

  for (g in unique(play_by_play$gameId)) {
    idx_game <- which(play_by_play$gameId == g)
    idx_game <- idx_game[order(sort_order[idx_game], seq_along(idx_game), na.last = TRUE)]
    active <- list()

    for (i in idx_game) {
      current_time <- seconds_elapsed_in_game[i]
      active <- advance_penalties(
        active,
        current_time = current_time,
        inclusive = type_desc_key[i] %in% c('period-start', 'stoppage', 'faceoff')
      )

      home_active <- sum(vapply(active, function(pen) identical(pen$team, 'home'), logical(1)))
      away_active <- sum(vapply(active, function(pen) identical(pen$team, 'away'), logical(1)))
      out$activePenaltyCountHome[i] <- as.integer(home_active)
      out$activePenaltyCountAway[i] <- as.integer(away_active)

      if (!(game_type_id[i] == 2L && period[i] == 4L) && period[i] < 5L) {
        out$homeSkaterCountRecon[i] <- as.integer(max(3L, 5L - home_active))
        out$awaySkaterCountRecon[i] <- as.integer(max(3L, 5L - away_active))
        out$strengthReconAvailable[i] <- TRUE
      }

      if (is_strength_penalty(i) && !major_pair_skip[i] && !is.na(current_time)) {
        active[[length(active) + 1L]] <- new_penalty(
          team = team_side[i],
          start_time = current_time,
          type_code = penalty_type_code[i],
          duration_min = penalty_duration[i]
        )
      }

      if (identical(type_desc_key[i], 'goal') && !is.na(current_time)) {
        active <- release_after_goal(
          active = active,
          scoring_team = team_side[i],
          current_time = current_time,
          home_count = out$homeSkaterCountRecon[i],
          away_count = out$awaySkaterCountRecon[i]
        )
      }
    }
  }

  out
}

#' Check HTML skater counts against reconstructed penalty state
#'
#' `.html_on_ice_matches_reconstructed_strength()` validates the HTML skater
#' counts against the penalty-based reconstruction. Goalie presence is validated
#' separately when deciding whether the HTML row can override stale source
#' strength context.
#'
#' @param reconstruction data.frame from `.reconstruct_skater_counts_from_penalties()`
#' @param idx row index in the play-by-play
#' @param home_skater_player_ids parsed home skater IDs
#' @param away_skater_player_ids parsed away skater IDs
#' @returns logical scalar
#' @keywords internal
.html_on_ice_matches_reconstructed_strength <- function(
  reconstruction,
  idx,
  home_skater_player_ids,
  away_skater_player_ids
) {
  if (
    is.null(reconstruction) ||
      !nrow(reconstruction) ||
      is.na(idx) ||
      idx < 1L ||
      idx > nrow(reconstruction) ||
      !isTRUE(reconstruction$strengthReconAvailable[idx])
  ) {
    return(FALSE)
  }
  identical(
    c(
      as.integer(length(home_skater_player_ids)),
      as.integer(length(away_skater_player_ids))
    ),
    c(
      as.integer(reconstruction$homeSkaterCountRecon[idx]),
      as.integer(reconstruction$awaySkaterCountRecon[idx])
    )
  )
}

#' Check required event actors against resolved HTML on-ice IDs
#'
#' `.html_on_ice_matches_event_actors()` validates that the key API event actors
#' for a matched row are present on the corresponding HTML on-ice team sets. It
#' is intentionally narrow and only returns `TRUE` when at least one informative
#' actor field is available for the event type.
#'
#' @param play_by_play data.frame cleaned internal play-by-play object
#' @param idx row index in the play-by-play
#' @param home_goalie_player_id parsed home goalie player ID
#' @param away_goalie_player_id parsed away goalie player ID
#' @param home_skater_player_ids parsed home skater IDs
#' @param away_skater_player_ids parsed away skater IDs
#' @returns logical scalar
#' @keywords internal
.html_on_ice_matches_event_actors <- function(
  play_by_play,
  idx,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
) {
  if (
    is.null(play_by_play) ||
      !nrow(play_by_play) ||
      is.na(idx) ||
      idx < 1L ||
      idx > nrow(play_by_play)
  ) {
    return(FALSE)
  }

  type_desc_key <- as.character(play_by_play$eventTypeDescKey[idx])
  if (
    is.na(type_desc_key) ||
      !(type_desc_key %in% .supported_html_on_ice_id_event_types())
  ) {
    return(FALSE)
  }

  home_ids <- c(
    if (!is.na(home_goalie_player_id)) as.integer(home_goalie_player_id),
    as.integer(home_skater_player_ids)
  )
  away_ids <- c(
    if (!is.na(away_goalie_player_id)) as.integer(away_goalie_player_id),
    as.integer(away_skater_player_ids)
  )
  team_for <- if (play_by_play$isHome[idx] %in% TRUE) {
    home_ids
  } else if (play_by_play$isHome[idx] %in% FALSE) {
    away_ids
  } else {
    integer()
  }
  team_against <- if (play_by_play$isHome[idx] %in% TRUE) {
    away_ids
  } else if (play_by_play$isHome[idx] %in% FALSE) {
    home_ids
  } else {
    integer()
  }
  if (!length(team_for) || !length(team_against)) {
    return(FALSE)
  }

  players <- list(primary = NA_integer_, secondary = NA_integer_, tertiary = NA_integer_)
  if (type_desc_key == 'faceoff') {
    players$primary <- .on_ice_int_col(play_by_play, 'winningPlayerId')[idx]
    players$secondary <- .on_ice_int_col(play_by_play, 'losingPlayerId')[idx]
  } else if (type_desc_key == 'hit') {
    players$primary <- .on_ice_int_col(play_by_play, 'hittingPlayerId')[idx]
    players$secondary <- .on_ice_int_col(play_by_play, 'hitteePlayerId')[idx]
  } else if (type_desc_key %in% c('shot-on-goal', 'missed-shot', 'failed-shot-attempt')) {
    players$primary <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[idx]
    if (is.na(players$primary)) {
      players$primary <- .on_ice_int_col(play_by_play, 'scoringPlayerId')[idx]
    }
    if (is.na(players$primary)) {
      players$primary <- .on_ice_int_col(play_by_play, 'playerId')[idx]
    }
  } else if (type_desc_key == 'blocked-shot') {
    players$primary <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[idx]
    players$secondary <- .on_ice_int_col(play_by_play, 'blockingPlayerId')[idx]
  } else if (type_desc_key == 'goal') {
    players$primary <- .on_ice_int_col(play_by_play, 'scoringPlayerId')[idx]
    if (is.na(players$primary)) {
      players$primary <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[idx]
    }
    players$secondary <- .on_ice_int_col(play_by_play, 'assist1PlayerId')[idx]
    players$tertiary <- .on_ice_int_col(play_by_play, 'assist2PlayerId')[idx]
  } else if (type_desc_key == 'delayed-penalty') {
    players$primary <- .on_ice_int_col(play_by_play, 'committedByPlayerId')[idx]
    if (is.na(players$primary)) {
      players$primary <- .on_ice_int_col(play_by_play, 'playerId')[idx]
    }
    players$secondary <- .on_ice_int_col(play_by_play, 'drawnByPlayerId')[idx]
  } else if (type_desc_key %in% c('giveaway', 'takeaway')) {
    players$primary <- .on_ice_int_col(play_by_play, 'playerId')[idx]
  }

  required <- unlist(players, use.names = FALSE)
  if (!any(!is.na(required))) {
    return(FALSE)
  }

  in_for <- function(player_id) is.na(player_id) || player_id %in% team_for
  in_against <- function(player_id) is.na(player_id) || player_id %in% team_against

  if (type_desc_key == 'goal') {
    return(
      in_for(players$primary) &&
        in_for(players$secondary) &&
        in_for(players$tertiary)
    )
  }
  if (type_desc_key %in% c('faceoff', 'hit', 'blocked-shot', 'delayed-penalty')) {
    return(
      in_for(players$primary) &&
        in_against(players$secondary)
    )
  }

  in_for(players$primary)
}

#' Check HTML on-ice state against a late empty-net heuristic
#'
#' `.html_on_ice_matches_late_empty_net()` accepts late-game 6-on-5 HTML rows
#' when the trailing team has clearly pulled its goalie and there are no active
#' reconstructed penalties that would already explain the manpower mismatch.
#'
#' @param play_by_play data.frame cleaned internal play-by-play object
#' @param reconstruction data.frame from `.reconstruct_skater_counts_from_penalties()`
#' @param idx row index in the play-by-play
#' @param home_goalie_player_id parsed home goalie player ID
#' @param away_goalie_player_id parsed away goalie player ID
#' @param home_skater_player_ids parsed home skater IDs
#' @param away_skater_player_ids parsed away skater IDs
#' @returns logical scalar
#' @keywords internal
.html_on_ice_matches_late_empty_net <- function(
  play_by_play,
  reconstruction,
  idx,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
) {
  if (
    is.null(reconstruction) ||
      !nrow(reconstruction) ||
      is.na(idx) ||
      idx < 1L ||
      idx > nrow(reconstruction)
  ) {
    return(FALSE)
  }
  if (
    is.na(play_by_play$periodNumber[idx]) ||
      play_by_play$periodNumber[idx] < 3L ||
      play_by_play$periodNumber[idx] >= 5L ||
      is.na(reconstruction$activePenaltyCountHome[idx]) ||
      is.na(reconstruction$activePenaltyCountAway[idx]) ||
      reconstruction$activePenaltyCountHome[idx] > 0L ||
      reconstruction$activePenaltyCountAway[idx] > 0L
  ) {
    return(FALSE)
  }
  home_score <- if ('homeGoals' %in% names(play_by_play)) {
    play_by_play$homeGoals[idx]
  } else if ('homeScore' %in% names(play_by_play)) {
    play_by_play$homeScore[idx]
  } else {
    NA_integer_
  }
  away_score <- if ('awayGoals' %in% names(play_by_play)) {
    play_by_play$awayGoals[idx]
  } else if ('awayScore' %in% names(play_by_play)) {
    play_by_play$awayScore[idx]
  } else {
    NA_integer_
  }
  if (is.na(home_score) || is.na(away_score)) {
    return(FALSE)
  }
  home_skater_count <- length(home_skater_player_ids)
  away_skater_count <- length(away_skater_player_ids)
  home_goalie_present <- !is.na(home_goalie_player_id)
  away_goalie_present <- !is.na(away_goalie_player_id)

  (
    home_score < away_score &&
      home_skater_count == 6L &&
      away_skater_count == 5L &&
      !home_goalie_present &&
      away_goalie_present
  ) || (
    away_score < home_score &&
      away_skater_count == 6L &&
      home_skater_count == 5L &&
      !away_goalie_present &&
      home_goalie_present
  )
}

#' Check whether an HTML on-ice signature can override derived strength context
#'
#' `.html_on_ice_can_override_strength_context()` allows the HTML on-ice skater
#' and goalie counts to replace derived strength context when the resolved
#' signature still looks like a plausible hockey state. Overflow rows with more
#' than six skaters are preserved in the scalar player-ID output but do not
#' rewrite the derived count columns.
#'
#' @param situation_code raw API situation code
#' @param home_goalie_player_id parsed home goalie player ID
#' @param away_goalie_player_id parsed away goalie player ID
#' @param home_skater_player_ids parsed home skater IDs
#' @param away_skater_player_ids parsed away skater IDs
#' @returns logical scalar
#' @keywords internal
.html_on_ice_can_override_strength_context <- function(
  situation_code,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
) {
  if (!is.na(situation_code) && situation_code %in% c('0101', '1010')) {
    return(FALSE)
  }
  home_goalie_count <- if (is.na(home_goalie_player_id)) 0L else 1L
  away_goalie_count <- if (is.na(away_goalie_player_id)) 0L else 1L
  home_skater_count <- length(home_skater_player_ids)
  away_skater_count <- length(away_skater_player_ids)
  home_goalie_count %in% 0:1 &&
    away_goalie_count %in% 0:1 &&
    home_skater_count >= 0L &&
    home_skater_count <= 6L &&
    away_skater_count >= 0L &&
    away_skater_count <= 6L
}

#' Add one-on-one shooter/goalie assignments
#'
#' `.add_one_on_one_on_ice_players()` populates on-ice player-ID columns for
#' penalty-shot and shootout rows whose `situationCode` implies a one-skater
#' versus one-goalie state (`0101`/`1010`). The raw `situationCode` remains the
#' source column; this helper only synthesizes the compatible scalar player-ID
#' assignments when the general HTML on-ice signature is not usable.
#'
#' @param play_by_play data.frame play-by-play object
#' @param matched optional matched HTML rows from `.match_html_pbp_to_api()`
#' @returns data.frame enriched with one-on-one shooter/goalie assignments
#' @keywords internal
.add_one_on_one_on_ice_players <- function(play_by_play, matched = data.frame()) {
  n <- nrow(play_by_play)
  if (!n) {
    return(play_by_play)
  }

  matched_lookup <- rep(NA_integer_, n)
  if (nrow(matched)) {
    ok_idx <- !is.na(matched$apiIndex) &
      matched$apiIndex >= 1L &
      matched$apiIndex <= n
    matched_lookup[matched$apiIndex[ok_idx]] <- seq_len(nrow(matched))[ok_idx]
  }

  slot_count <- .on_ice_skater_slots(play_by_play = play_by_play)
  type_desc_key <- as.character(play_by_play$eventTypeDescKey)
  situation_code <- as.character(play_by_play$situationCode)
  shooter_id <- .on_ice_int_col(play_by_play, 'scoringPlayerId')
  missing_shooter <- is.na(shooter_id)
  shooter_id[missing_shooter] <- .on_ice_int_col(play_by_play, 'shootingPlayerId')[missing_shooter]
  missing_shooter <- is.na(shooter_id)
  shooter_id[missing_shooter] <- .on_ice_int_col(play_by_play, 'playerId')[missing_shooter]
  goalie_id <- .on_ice_int_col(play_by_play, 'goalieInNetId')
  goalie_id_against <- .on_ice_int_col(play_by_play, 'goaliePlayerIdAgainst')
  goalie_id[is.na(goalie_id)] <- goalie_id_against[is.na(goalie_id)]

  is_one_on_one <- !is.na(situation_code) &
    situation_code %in% c('0101', '1010') &
    !is.na(type_desc_key) &
    type_desc_key %in% c('shot-on-goal', 'missed-shot', 'goal', 'failed-shot-attempt')

  for (idx in which(is_one_on_one)) {
    if (is.na(play_by_play$isHome[idx]) || is.na(shooter_id[idx])) {
      next
    }
    if (is.na(goalie_id[idx])) {
      matched_pos <- matched_lookup[idx]
      if (!is.na(matched_pos)) {
        if (play_by_play$isHome[idx] %in% TRUE) {
          goalie_id[idx] <- as.integer(matched$awayGoaliePlayerId[matched_pos])
          if (is.na(goalie_id[idx])) {
            goalie_id[idx] <- as.integer(matched$homeGoaliePlayerId[matched_pos])
          }
        } else {
          goalie_id[idx] <- as.integer(matched$homeGoaliePlayerId[matched_pos])
          if (is.na(goalie_id[idx])) {
            goalie_id[idx] <- as.integer(matched$awayGoaliePlayerId[matched_pos])
          }
        }
      }
    }
    if (is.na(goalie_id[idx])) {
      next
    }

    play_by_play$homeGoaliePlayerId[idx] <- NA_integer_
    play_by_play$awayGoaliePlayerId[idx] <- NA_integer_
    play_by_play$goaliePlayerIdFor[idx] <- NA_integer_
    play_by_play$goaliePlayerIdAgainst[idx] <- goalie_id[idx]
    for (k in seq_len(slot_count)) {
      play_by_play[[paste0('homeSkater', k, 'PlayerId')]][idx] <- NA_integer_
      play_by_play[[paste0('awaySkater', k, 'PlayerId')]][idx] <- NA_integer_
      play_by_play[[paste0('skater', k, 'PlayerIdFor')]][idx] <- NA_integer_
      play_by_play[[paste0('skater', k, 'PlayerIdAgainst')]][idx] <- NA_integer_
    }
    if (play_by_play$isHome[idx] %in% TRUE) {
      play_by_play$awayGoaliePlayerId[idx] <- goalie_id[idx]
      play_by_play$homeSkater1PlayerId[idx] <- shooter_id[idx]
    } else {
      play_by_play$homeGoaliePlayerId[idx] <- goalie_id[idx]
      play_by_play$awaySkater1PlayerId[idx] <- shooter_id[idx]
    }
    play_by_play$skater1PlayerIdFor[idx] <- shooter_id[idx]
  }

  play_by_play
}

#' Assign resolved HTML on-ice IDs to scalar columns
#'
#' `.assign_html_on_ice_player_ids()` copies resolved HTML goalie and skater IDs
#' into the public scalar on-ice player-ID columns for one play-by-play row.
#'
#' @param play_by_play data.frame play-by-play object
#' @param idx row index to update
#' @param home_goalie resolved home goalie player ID
#' @param away_goalie resolved away goalie player ID
#' @param home_skaters integer vector of resolved home skater IDs
#' @param away_skaters integer vector of resolved away skater IDs
#' @returns data.frame with scalar on-ice player-ID columns assigned for `idx`
#' @keywords internal
.assign_html_on_ice_player_ids <- function(
  play_by_play,
  idx,
  home_goalie,
  away_goalie,
  home_skaters,
  away_skaters
) {
  slot_count <- .on_ice_skater_slots(
    play_by_play = play_by_play,
    slot_count = max(length(home_skaters), length(away_skaters))
  )
  play_by_play <- .add_empty_html_on_ice_columns(
    play_by_play,
    slot_count = slot_count
  )
  home_slots <- rep(NA_integer_, slot_count)
  away_slots <- rep(NA_integer_, slot_count)
  if (length(home_skaters)) {
    take <- seq_len(min(slot_count, length(home_skaters)))
    home_slots[take] <- home_skaters[take]
  }
  if (length(away_skaters)) {
    take <- seq_len(min(slot_count, length(away_skaters)))
    away_slots[take] <- away_skaters[take]
  }

  play_by_play$homeGoaliePlayerId[idx] <- home_goalie
  play_by_play$awayGoaliePlayerId[idx] <- away_goalie
  for (k in seq_len(slot_count)) {
    play_by_play[[paste0('homeSkater', k, 'PlayerId')]][idx] <- home_slots[k]
    play_by_play[[paste0('awaySkater', k, 'PlayerId')]][idx] <- away_slots[k]
  }
  if (play_by_play$isHome[idx] %in% TRUE) {
    play_by_play$goaliePlayerIdFor[idx] <- home_goalie
    play_by_play$goaliePlayerIdAgainst[idx] <- away_goalie
    for (k in seq_len(slot_count)) {
      play_by_play[[paste0('skater', k, 'PlayerIdFor')]][idx] <- home_slots[k]
      play_by_play[[paste0('skater', k, 'PlayerIdAgainst')]][idx] <- away_slots[k]
    }
  } else if (play_by_play$isHome[idx] %in% FALSE) {
    play_by_play$goaliePlayerIdFor[idx] <- away_goalie
    play_by_play$goaliePlayerIdAgainst[idx] <- home_goalie
    for (k in seq_len(slot_count)) {
      play_by_play[[paste0('skater', k, 'PlayerIdFor')]][idx] <- away_slots[k]
      play_by_play[[paste0('skater', k, 'PlayerIdAgainst')]][idx] <- home_slots[k]
    }
  }

  play_by_play
}

#' Identify populated on-ice player rows
#'
#' `.has_any_on_ice_player_ids()` returns a logical vector indicating whether a
#' play-by-play row already has any scalar on-ice player IDs assigned.
#'
#' @param play_by_play data.frame play-by-play object
#' @returns logical vector
#' @keywords internal
.has_any_on_ice_player_ids <- function(play_by_play) {
  id_cols <- intersect(
    .on_ice_id_scalar_column_names(play_by_play = play_by_play),
    names(play_by_play)
  )
  if (!length(id_cols)) {
    return(rep(FALSE, nrow(play_by_play)))
  }
  rowSums(!is.na(play_by_play[, id_cols, drop = FALSE])) > 0L
}

#' Backfill unmatched delayed-penalty rows from nearby populated rows
#'
#' `.backfill_delayed_penalty_on_ice_players()` fills unmatched delayed-penalty
#' rows from the nearest previous populated supported row in the same game and
#' period when the raw `situationCode` and goalie/skater counts are unchanged.
#'
#' @param play_by_play data.frame play-by-play object
#' @param max_gap_seconds integer scalar time window
#' @returns data.frame with delayed-penalty rows backfilled where possible
#' @keywords internal
.backfill_delayed_penalty_on_ice_players <- function(
  play_by_play,
  max_gap_seconds = 15L
) {
  n <- nrow(play_by_play)
  if (
      !n ||
      !all(c(
        'gameId', 'periodNumber', 'secondsElapsedInPeriod', 'sortOrder', 'eventTypeDescKey',
        'situationCode', 'homeIsEmptyNet', 'awayIsEmptyNet',
        'homeSkaterCount', 'awaySkaterCount'
      ) %in% names(play_by_play))
  ) {
    return(play_by_play)
  }

  populated <- .has_any_on_ice_player_ids(play_by_play)
  id_cols <- .on_ice_id_scalar_column_names(play_by_play = play_by_play)
  candidate_idx <- which(
    !populated &
      !is.na(play_by_play$eventTypeDescKey) &
      play_by_play$eventTypeDescKey == 'delayed-penalty'
  )
  if (!length(candidate_idx)) {
    return(play_by_play)
  }

  slot_count <- .on_ice_skater_slots(play_by_play = play_by_play)
  row_signature_matches <- function(src_idx, dst_idx) {
    home_goalie_present <- !is.na(play_by_play$homeGoaliePlayerId[src_idx])
    away_goalie_present <- !is.na(play_by_play$awayGoaliePlayerId[src_idx])
    home_skater_count <- sum(!is.na(play_by_play[src_idx, paste0(
      'homeSkater',
      seq_len(slot_count),
      'PlayerId'
    )]))
    away_skater_count <- sum(!is.na(play_by_play[src_idx, paste0(
      'awaySkater',
      seq_len(slot_count),
      'PlayerId'
    )]))
    identical(
      c(
        if (home_goalie_present) 1L else 0L,
        home_skater_count,
        away_skater_count,
        if (away_goalie_present) 1L else 0L
      ),
      c(
        if (play_by_play$homeIsEmptyNet[dst_idx] %in% TRUE) 0L else 1L,
        as.integer(play_by_play$homeSkaterCount[dst_idx]),
        as.integer(play_by_play$awaySkaterCount[dst_idx]),
        if (play_by_play$awayIsEmptyNet[dst_idx] %in% TRUE) 0L else 1L
      )
    )
  }

  for (idx in candidate_idx) {
    prev_idx <- which(
        populated &
        play_by_play$gameId == play_by_play$gameId[idx] &
        play_by_play$periodNumber == play_by_play$periodNumber[idx] &
        play_by_play$secondsElapsedInPeriod <= play_by_play$secondsElapsedInPeriod[idx] &
        play_by_play$secondsElapsedInPeriod >= (play_by_play$secondsElapsedInPeriod[idx] - max_gap_seconds) &
        !is.na(play_by_play$situationCode) &
        play_by_play$situationCode == play_by_play$situationCode[idx]
    )
    if (!length(prev_idx)) {
      next
    }
    prev_idx <- prev_idx[order(
      play_by_play$secondsElapsedInPeriod[prev_idx],
      play_by_play$sortOrder[prev_idx],
      decreasing = TRUE
    )]
    prev_idx <- prev_idx[vapply(prev_idx, row_signature_matches, logical(1), dst_idx = idx)]
    if (!length(prev_idx)) {
      next
    }
    src_idx <- prev_idx[1L]
    for (nm in id_cols) {
      play_by_play[[nm]][idx] <- play_by_play[[nm]][src_idx]
    }
  }

  play_by_play
}

#' Add HTML-derived on-ice players to play-by-play
#'
#' `.add_html_on_ice_players()` resolves HTML on-ice goalie and skater IDs,
#' matches them to API rows, and injects the resulting scalar on-ice columns
#' into play-by-play output.
#'
#' @param play_by_play data.frame play-by-play object
#' @param game game ID
#' @param rosters optional roster data.frame
#' @param home_team optional home-team metadata
#' @param away_team optional away-team metadata
#' @param html_rows optional parsed HTML on-ice rows
#' @returns data.frame enriched with HTML-derived on-ice columns
#' @keywords internal
.add_html_on_ice_players <- function(
  play_by_play,
  game,
  rosters = NULL,
  home_team = NULL,
  away_team = NULL,
  html_rows = NULL
) {
  play_by_play <- .add_empty_html_on_ice_columns(play_by_play)
  if (!nrow(play_by_play)) {
    return(play_by_play)
  }
  if (is.null(rosters) || is.null(home_team) || is.null(away_team)) {
    meta <- nhl_api(
      path = sprintf('v1/gamecenter/%s/play-by-play', as.integer(game)),
      type = 'w'
    )
    if (is.null(rosters)) {
      rosters <- meta$rosterSpots
    }
    if (is.null(home_team)) {
      home_team <- meta$homeTeam
    }
    if (is.null(away_team)) {
      away_team <- meta$awayTeam
    }
  }
  if (is.null(html_rows)) {
    html_rows <- tryCatch(
      .fetch_html_pbp_on_ice(
        game = as.integer(game),
        rosters = rosters,
        home_team_id = home_team$id,
        away_team_id = away_team$id,
        home_abbrev = home_team$abbrev,
        away_abbrev = away_team$abbrev
      ),
      error = function(e) data.frame()
    )
  }
  matched <- .match_html_pbp_to_api(play_by_play, html_rows)
  reconstruction <- .reconstruct_skater_counts_from_penalties(play_by_play)
  if (!nrow(matched)) {
    play_by_play <- .add_one_on_one_on_ice_players(play_by_play)
    play_by_play <- .backfill_delayed_penalty_on_ice_players(play_by_play)
    return(.mask_strength_context_block(play_by_play))
  }
  for (i in seq_len(nrow(matched))) {
    idx <- matched$apiIndex[i]
    home_goalie <- as.integer(matched$homeGoaliePlayerId[i])
    away_goalie <- as.integer(matched$awayGoaliePlayerId[i])
    home_skaters <- unlist(matched$homeSkaterPlayerIds[i], use.names = FALSE)
    away_skaters <- unlist(matched$awaySkaterPlayerIds[i], use.names = FALSE)
    type_desc_key <- as.character(play_by_play$eventTypeDescKey[idx])
    if (
      is.na(type_desc_key) ||
        !(type_desc_key %in% .supported_html_on_ice_id_event_types())
    ) {
      next
    }
    if (!is.na(play_by_play$situationCode[idx]) && play_by_play$situationCode[idx] %in% c('0101', '1010')) {
      next
    }
    if (.html_on_ice_can_override_strength_context(
      situation_code = play_by_play$situationCode[idx],
      home_goalie_player_id = home_goalie,
      away_goalie_player_id = away_goalie,
      home_skater_player_ids = home_skaters,
      away_skater_player_ids = away_skaters
    )) {
      play_by_play <- .override_strength_context_from_html(
        play_by_play = play_by_play,
        idx = idx,
        home_goalie = home_goalie,
        away_goalie = away_goalie,
        home_skaters = home_skaters,
        away_skaters = away_skaters
      )
    }
    play_by_play <- .assign_html_on_ice_player_ids(
      play_by_play = play_by_play,
      idx = idx,
      home_goalie = home_goalie,
      away_goalie = away_goalie,
      home_skaters = home_skaters,
      away_skaters = away_skaters
    )
  }
  play_by_play <- .add_one_on_one_on_ice_players(play_by_play, matched = matched)
  play_by_play <- .backfill_delayed_penalty_on_ice_players(play_by_play)
  .mask_strength_context_block(play_by_play)
}

#' Allocate a missing public play-by-play column
#'
#' `.empty_public_pbp_column()` returns a typed `NA` vector for a public
#' play-by-play column that is absent in the upstream source data.
#'
#' @param name public play-by-play column name
#' @param n output row count
#' @returns typed vector of missing values
#' @keywords internal
.empty_public_pbp_column <- function(name, n) {
  if (grepl(
    '^(?:home|away)GoaliePlayerId$|^goaliePlayerId(?:For|Against)$|^(?:home|away)Skater[0-9]+PlayerId$|^skater[0-9]+PlayerId(?:For|Against)$',
    name
  )) {
    return(rep(NA_integer_, n))
  }
  if (grepl(
    '^(?:home|away)GoalieSeconds(?:RemainingInShift|ElapsedIn(?:Shift|PeriodSinceLastShift))$|^goalieSeconds(?:RemainingInShift|ElapsedIn(?:Shift|PeriodSinceLastShift))(?:For|Against)$|^(?:home|away)Skater[0-9]+Seconds(?:RemainingInShift|ElapsedIn(?:Shift|PeriodSinceLastShift))$|^skater[0-9]+Seconds(?:RemainingInShift|ElapsedIn(?:Shift|PeriodSinceLastShift))(?:For|Against)$',
    name
  )) {
    return(rep(NA_real_, n))
  }
  character_cols <- c(
    'periodType', 'eventTypeDescKey', 'situationCode', 'strengthState',
    'homeTeamDefendingSide', 'zoneCode', 'shotType', 'penaltyTypeCode',
    'penaltyTypeDescKey', 'reason', 'secondaryReason', 'discreteClip',
    'highlightClip', 'highlightClipSharingUrl', 'pptReplayUrl', 'utc'
  )
  logical_cols <- c(
    'isHome', 'homeIsEmptyNet', 'awayIsEmptyNet', 'isEmptyNetFor',
    'isEmptyNetAgainst', 'isRush', 'isRebound', 'createdRebound'
  )
  numeric_cols <- c(
    'xCoord', 'yCoord', 'xCoordNorm', 'yCoordNorm', 'distance', 'angle',
    .on_ice_timing_scalar_column_names('SecondsRemainingInShift'),
    .on_ice_timing_scalar_column_names('SecondsElapsedInShift'),
    .on_ice_timing_scalar_column_names('SecondsElapsedInPeriodSinceLastShift')
  )
  if (name %in% character_cols) {
    return(rep(NA_character_, n))
  }
  if (name %in% logical_cols) {
    return(rep(NA, n))
  }
  if (name %in% numeric_cols) {
    return(rep(NA_real_, n))
  }
  rep(NA_integer_, n)
}

#' Finalize public play-by-play output
#'
#' `.finalize_pbp_output()` selects and orders the final GC or WSC public
#' play-by-play column set.
#'
#' @param play_by_play data.frame play-by-play object
#' @param source output source, either `"gc"` or `"wsc"`
#' @returns data.frame with the finalized public play-by-play schema
#' @keywords internal
.finalize_pbp_output <- function(play_by_play, source = c('gc', 'wsc')) {
  source <- match.arg(source)
  on_ice_cols <- .on_ice_id_scalar_column_names(play_by_play = play_by_play)

  if (all(c('eventTypeDescKey', 'shootingPlayerId', 'scoringPlayerId') %in% names(play_by_play))) {
    goal_idx <- play_by_play$eventTypeDescKey == 'goal' &
      is.na(play_by_play$shootingPlayerId) &
      !is.na(play_by_play$scoringPlayerId)
    play_by_play$shootingPlayerId[goal_idx] <- play_by_play$scoringPlayerId[goal_idx]
  }

  gc_cols <- c(
    'gameId', 'eventId', 'seasonId', 'gameTypeId', 'gameNumber', 'sortOrder',
    'periodNumber', 'periodType', 'secondsElapsedInPeriod',
    'secondsElapsedInGame', 'eventOwnerTeamId', 'isHome', 'eventTypeCode',
    'eventTypeDescKey', 'situationCode', 'homeIsEmptyNet', 'awayIsEmptyNet',
    'isEmptyNetFor', 'isEmptyNetAgainst', 'homeSkaterCount',
    'awaySkaterCount', 'skaterCountFor', 'skaterCountAgainst',
    'manDifferential', 'strengthState', on_ice_cols, 'homeTeamDefendingSide', 'zoneCode',
    'xCoord', 'yCoord', 'xCoordNorm', 'yCoordNorm', 'distance', 'angle',
    'shotType', 'isRush', 'isRebound', 'createdRebound', 'homeGoals',
    'awayGoals', 'goalsFor', 'goalsAgainst', 'homeShots', 'awayShots',
    'shotsFor', 'shotsAgainst', 'homeFenwick', 'awayFenwick', 'fenwickFor',
    'fenwickAgainst', 'homeCorsi', 'awayCorsi', 'corsiFor', 'corsiAgainst',
    'goalDifferential', 'shotDifferential', 'fenwickDifferential',
    'corsiDifferential', 'playerId', 'winningPlayerId', 'losingPlayerId',
    'hittingPlayerId', 'hitteePlayerId', 'committedByPlayerId',
    'drawnByPlayerId', 'servedByPlayerId', 'blockingPlayerId', 'goalieInNetId',
    'shootingPlayerId',
    'scoringPlayerId', 'assist1PlayerId', 'assist2PlayerId',
    'scoringPlayerTotal', 'assist1PlayerTotal', 'assist2PlayerTotal',
    'penaltyTypeCode', 'penaltyTypeDescKey', 'penaltyDuration', 'reason',
    'secondaryReason', 'discreteClip', 'highlightClip',
    'highlightClipSharingUrl', 'pptReplayUrl'
  )
  wsc_cols <- c(
    'gameId', 'eventId', 'seasonId', 'gameTypeId', 'gameNumber', 'sortOrder',
    'periodNumber', 'secondsElapsedInPeriod', 'secondsElapsedInGame', 'utc',
    'eventOwnerTeamId', 'isHome', 'eventTypeCode', 'eventTypeDescKey',
    'situationCode', 'homeIsEmptyNet', 'awayIsEmptyNet', 'isEmptyNetFor',
    'isEmptyNetAgainst', 'homeSkaterCount', 'awaySkaterCount',
    'skaterCountFor', 'skaterCountAgainst', 'manDifferential',
    'strengthState', on_ice_cols, 'homeTeamDefendingSide', 'zoneCode', 'xCoord', 'yCoord',
    'xCoordNorm', 'yCoordNorm', 'distance', 'angle', 'shotType', 'isRush',
    'isRebound', 'createdRebound', 'homeGoals', 'awayGoals', 'goalsFor',
    'goalsAgainst', 'homeShots', 'awayShots', 'shotsFor', 'shotsAgainst',
    'homeFenwick', 'awayFenwick', 'fenwickFor', 'fenwickAgainst',
    'homeCorsi', 'awayCorsi', 'corsiFor', 'corsiAgainst',
    'goalDifferential', 'shotDifferential', 'fenwickDifferential',
    'corsiDifferential', 'playerId', 'winningPlayerId', 'losingPlayerId',
    'hittingPlayerId', 'hitteePlayerId', 'committedByPlayerId',
    'drawnByPlayerId', 'servedByPlayerId', 'blockingPlayerId', 'goalieInNetId',
    'shootingPlayerId',
    'scoringPlayerId', 'assist1PlayerId', 'assist2PlayerId',
    'scoringPlayerTotal', 'assist1PlayerTotal', 'assist2PlayerTotal',
    'penaltyTypeCode', 'penaltyTypeDescKey', 'penaltyDuration', 'reason'
  )

  keep <- switch(source, gc = gc_cols, wsc = wsc_cols)
  missing_cols <- setdiff(keep, names(play_by_play))
  if (length(missing_cols)) {
    for (nm in missing_cols) {
      play_by_play[[nm]] <- .empty_public_pbp_column(nm, nrow(play_by_play))
    }
  }
  play_by_play[, keep, drop = FALSE]
}

#' Check whether a parallel request failed
#'
#' `.parallel_request_failed()` returns `TRUE` when a response object from
#' `.perform_parallel_requests()` is an httr2 failure.
#'
#' @param resp response or failure object
#' @returns logical scalar
#' @keywords internal
.parallel_request_failed <- function(resp) {
  inherits(resp, 'httr2_failure')
}

#' Parse optional HTML on-ice rows from a parallel response
#'
#' `.optional_html_pbp_rows_from_response()` parses a fetched HTML play-by-play
#' response when it succeeded and otherwise returns an empty data frame.
#'
#' @param resp response or failure object
#' @inheritParams .fetch_html_pbp_on_ice
#' @returns data.frame of parsed HTML rows or an empty data.frame
#' @keywords internal
.optional_html_pbp_rows_from_response <- function(
  resp,
  game,
  rosters,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev
) {
  if (.parallel_request_failed(resp)) {
    return(data.frame())
  }
  tryCatch(
    .parse_html_pbp_response(
      resp = resp,
      game = game,
      rosters = rosters,
      home_team_id = home_team_id,
      away_team_id = away_team_id,
      home_abbrev = home_abbrev,
      away_abbrev = away_abbrev
    ),
    error = function(e) data.frame()
  )
}

#' Access the raw GameCenter (GC) play-by-play for a game
#'
#' `gc_play_by_play_raw()` returns the raw flattened GameCenter play-by-play as
#' served by the NHL API for one game. Use [gc_play_by_play()] for the cleaned
#' public schema that repairs common clock/order defects and appends the derived
#' public columns.
#'
#' @inheritParams gc_summary
#'
#' @returns data.frame with one row per event (play)
#' @examples
#' gc_raw_Martin_Necas_legacy_game <- gc_play_by_play_raw(game = 2025020275)
#' @export

gc_play_by_play_raw <- function(game = 2023030417) {
  tryCatch(
    expr = {
      game <- as.integer(game)
      if (length(game) != 1L || is.na(game) || game <= 0L) {
        stop('invalid game')
      }
      pbp_meta <- nhl_api(
        path = sprintf('v1/gamecenter/%s/play-by-play', game),
        type = 'w'
      )
      plays <- if (is.null(pbp_meta$plays)) {
        data.frame()
      } else {
        as.data.frame(pbp_meta$plays, stringsAsFactors = FALSE)
      }
      if (!nrow(plays)) {
        return(plays)
      }
      plays$gameId <- game
      plays[, c('gameId', setdiff(names(plays), 'gameId')), drop = FALSE]
    },
    error = function(e) {
      message(paste(
        'Invalid argument(s); refer to help file.',
        '\nProvided game:',
        game
      ))
      data.frame()
    }
  )
}

#' @rdname gc_play_by_play_raw
#' @export

gc_pbp_raw <- function(game = 2023030417) {
  gc_play_by_play_raw(game)
}

#' Access the GameCenter (GC) play-by-play for a game
#'
#' `gc_play_by_play()` retrieves the GameCenter (GC) play-by-play for a game as
#' a `data.frame` where each row represents an event. The returned schema is the
#' cleaned, public-facing play-by-play schema, including canonical names such as
#' `periodNumber`, `eventTypeCode`, `eventTypeDescKey`, `homeShots`,
#' `shotsFor`, `penaltyTypeDescKey`, `penaltyDuration`, `servedByPlayerId`,
#' `goalieInNetId`, and HTML-report-derived on-ice player ID columns such as
#' `homeGoaliePlayerId`, `awayGoaliePlayerId`, `homeSkater1PlayerId`, and any
#' additional overflow skater slots required by the game. HTML report skater and
#' goalie IDs are returned whenever they can be matched back to a supported row,
#' even when the raw `situationCode` is stale. Use [add_shift_times()] with
#' [shift_chart()] (or [shift_charts()]) to add on-ice shift timing columns.
#'
#' @inheritParams gc_summary
#'
#' @returns data.frame with one row per event (play)
#' @examples
#' # May take >5s, so skip.
#' \donttest{gc_pbp_Martin_Necas_legacy_game <- gc_play_by_play(game = 2025020275)}
#' @export

gc_play_by_play <- function(game = 2023030417) {
  tryCatch(
    expr  = {
      game <- as.integer(game)
      responses <- .perform_parallel_requests(
        list(
          pbp_meta = .nhl_request(
            path = sprintf('v1/gamecenter/%s/play-by-play', game),
            type = 'w'
          ),
          html_pbp = .html_report_request(.html_pbp_report_url(game))
        ),
        on_error = 'return'
      )
      if (.parallel_request_failed(responses$pbp_meta)) {
        stop(conditionMessage(responses$pbp_meta))
      }
      pbp_meta <- .nhl_json_from_response(responses$pbp_meta)
      html_rows <- .optional_html_pbp_rows_from_response(
        resp = responses$html_pbp,
        game = game,
        rosters = pbp_meta$rosterSpots,
        home_team_id = pbp_meta$homeTeam$id,
        away_team_id = pbp_meta$awayTeam$id,
        home_abbrev = pbp_meta$homeTeam$abbrev,
        away_abbrev = pbp_meta$awayTeam$abbrev
      )
      plays <- pbp_meta$plays
      # Rename columns.
      plays$gameId <- game
      plays        <- plays[, c('gameId', setdiff(names(plays), 'gameId'))]
      nms <- names(plays)
      nms[nms == 'details.typeCode'] <- 'penaltyTypeCode'
      idx          <- grepl('\\.', nms)
      nms[idx]     <- sub('^[^.]*\\.', '', nms[idx])
      nms[nms == 'number'] <- 'period'
      names(plays) <- nms
      if (
        all(c('typeDescKey', 'shootingPlayerId', 'scoringPlayerId') %in% names(plays))
      ) {
        goal_idx <- plays$typeDescKey == 'goal' &
          is.na(plays$shootingPlayerId) &
          !is.na(plays$scoringPlayerId)
        plays$shootingPlayerId[goal_idx] <- plays$scoringPlayerId[goal_idx]
      }
      # Fix zoneCode for blocked shots.
      idx <- plays$typeDescKey == 'blocked-shot' &
        plays$zoneCode %in% c('O', 'D')
      plays$zoneCode[idx] <- ifelse(
        plays$zoneCode[idx] == 'O',
        'D',
        'O'
      )
      # Remove inconsistent columns.
      plays$timeRemaining <- NULL
      # Clean.
      plays <- .strip_game_id(plays) |>
        .normalize_public_pbp_names() |>
        .strip_time_period() |>
        .drop_illogical_ordered_events() |>
        .repair_public_pbp_sequence() |>
        .flag_is_home() |>
        .strip_situation_code() |>
        .normalize_coordinates() |>
        .calculate_distance() |>
        .calculate_angle() |>
        .apply_shot_context() |>
        .add_html_on_ice_players(
          game = game,
          rosters = pbp_meta$rosterSpots,
          home_team = pbp_meta$homeTeam,
          away_team = pbp_meta$awayTeam,
          html_rows = html_rows
        )
      .finalize_pbp_output(plays, source = 'gc')
    },
    error = function(e) {
      message(paste(
        'Invalid argument(s); refer to help file.',
        '\nProvided game:',
        game
      ))
      data.frame()
    }
  )
}

#' @rdname gc_play_by_play
#' @export

gc_pbp <- function(game = 2023030417) {
  gc_play_by_play(game)
}

#' Access the raw World Showcase (WSC) play-by-play for a game
#'
#' `wsc_play_by_play_raw()` returns the raw flattened World Showcase play-by-play
#' as served by the NHL API for one game. Use [wsc_play_by_play()] for the
#' cleaned public schema that repairs common clock/order defects and appends the
#' derived public columns.
#'
#' @inheritParams gc_summary
#'
#' @returns data.frame with one row per event (play)
#' @examples
#' wsc_raw_Martin_Necas_legacy_game <- wsc_play_by_play_raw(game = 2025020275)
#' @export

wsc_play_by_play_raw <- function(game = 2023030417) {
  tryCatch(
    expr = {
      game <- as.integer(game)
      if (length(game) != 1L || is.na(game) || game <= 0L) {
        stop('invalid game')
      }
      plays <- nhl_api(
        path = sprintf('v1/wsc/play-by-play/%s', game),
        type = 'w'
      )
      plays <- as.data.frame(plays, stringsAsFactors = FALSE)
      if (!nrow(plays)) {
        return(plays)
      }
      if ('id' %in% names(plays)) {
        plays$id <- NULL
      }
      plays$gameId <- game
      plays[, c('gameId', setdiff(names(plays), 'gameId')), drop = FALSE]
    },
    error = function(e) {
      message(paste(
        'Invalid argument(s); refer to help file.',
        '\nProvided game:',
        game
      ))
      data.frame()
    }
  )
}

#' @rdname wsc_play_by_play_raw
#' @export

wsc_pbp_raw <- function(game = 2023030417) {
  wsc_play_by_play_raw(game)
}

#' Access the World Showcase (WSC) play-by-play for a game
#'
#' `wsc_play_by_play()` retrieves the World Showcase (WSC) play-by-play for a
#' game as a `data.frame` where each row represents an event. The returned
#' schema follows the same cleaned public-facing naming as `gc_play_by_play()`,
#' including `servedByPlayerId`, `goalieInNetId`, and `utc` immediately after
#' `secondsElapsedInGame` while omitting GC-only clip fields. It also includes
#' the same HTML-report-derived on-ice player ID columns added to the GC output,
#' including dynamically expanded overflow skater slots when needed. HTML report
#' skater and goalie IDs are returned whenever they can be matched back to a
#' supported row, even when the raw `situationCode` is stale. Use [add_shift_times()]
#' with [shift_chart()] (or [shift_charts()]) to add on-ice shift timing columns.
#'
#' @inheritParams gc_summary
#'
#' @returns data.frame with one row per event (play)
#' @examples
#' # May take >5s, so skip.
#' \donttest{wsc_pbp_Martin_Necas_legacy_game <- wsc_play_by_play(game = 2025020275)}
#' @export

wsc_play_by_play <- function(game = 2023030417) {
  tryCatch(
    expr = {
      game <- as.integer(game)
      responses <- .perform_parallel_requests(
        list(
          pbp_meta = .nhl_request(
            path = sprintf('v1/gamecenter/%s/play-by-play', game),
            type = 'w'
          ),
          wsc = .nhl_request(
            path = sprintf('v1/wsc/play-by-play/%s', game),
            type = 'w'
          ),
          html_pbp = .html_report_request(.html_pbp_report_url(game))
        ),
        on_error = 'return'
      )
      if (.parallel_request_failed(responses$pbp_meta)) {
        stop(conditionMessage(responses$pbp_meta))
      }
      if (.parallel_request_failed(responses$wsc)) {
        stop(conditionMessage(responses$wsc))
      }
      pbp_meta <- .nhl_json_from_response(responses$pbp_meta)
      plays <- .nhl_json_from_response(responses$wsc)
      html_rows <- .optional_html_pbp_rows_from_response(
        resp = responses$html_pbp,
        game = game,
        rosters = pbp_meta$rosterSpots,
        home_team_id = pbp_meta$homeTeam$id,
        away_team_id = pbp_meta$awayTeam$id,
        home_abbrev = pbp_meta$homeTeam$abbrev,
        away_abbrev = pbp_meta$awayTeam$abbrev
      )
      # Rename column.
      plays$id     <- NULL
      plays$gameId <- game
      plays        <- plays[, c('gameId', setdiff(names(plays), 'gameId'))]
      if (
        all(c('typeDescKey', 'shootingPlayerId', 'scoringPlayerId') %in% names(plays))
      ) {
        goal_idx <- plays$typeDescKey == 'goal' &
          is.na(plays$shootingPlayerId) &
          !is.na(plays$scoringPlayerId)
        plays$shootingPlayerId[goal_idx] <- plays$scoringPlayerId[goal_idx]
      }
      # Fix zoneCode for blocked shots.
      idx <- plays$typeDescKey == 'blocked-shot' &
        plays$zoneCode %in% c('O', 'D')
      plays$zoneCode[idx] <- ifelse(
        plays$zoneCode[idx] == 'O',
        'D',
        'O'
      )
      # Remove inconsistent columns.
      plays$goalModifier     <- NULL
      plays$strength         <- NULL
      plays$strengthCode     <- NULL
      plays$goalCode         <- NULL
      plays$secondsRemaining <- NULL
      # Clean.
      plays <- .strip_game_id(plays) |>
        .normalize_public_pbp_names() |>
        .strip_time_period() |>
        .drop_illogical_ordered_events() |>
        .repair_public_pbp_sequence() |>
        .flag_is_home() |>
        .strip_situation_code() |>
        .normalize_coordinates() |>
        .calculate_distance() |>
        .calculate_angle() |>
        .apply_shot_context() |>
        .add_html_on_ice_players(
          game = game,
          rosters = pbp_meta$rosterSpots,
          home_team = pbp_meta$homeTeam,
          away_team = pbp_meta$awayTeam,
          html_rows = html_rows
        )
      .finalize_pbp_output(plays, source = 'wsc')
    },
    error = function(e) {
      message(paste(
        'Invalid argument(s); refer to help file.',
        '\nProvided game:',
        game
      ))
      data.frame()
    }
  )
}

#' @rdname wsc_play_by_play
#' @export

wsc_pbp <- function(game = 2023030417) {
  wsc_play_by_play(game)
}

#' Access the shift chart for a game
#'
#' `shift_chart()` retrieves the shift chart for a game as a `data.frame` where each row represents shift and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
#'
#' @inheritParams gc_summary
#'
#' @returns data.frame with one row per shift
#' @examples
#' shifts_Martin_Necas_legacy_game <- shift_chart(game = 2025020275)
#' @export

shift_chart <- function(game = 2023030417) {
  game <- as.integer(game)
  if (length(game) != 1L || is.na(game) || game <= 0L) {
    message(paste(
      'Invalid argument(s); refer to help file.',
      '\nProvided game:',
      paste(game, collapse = ', ')
    ))
    return(data.frame())
  }
  shift_output_cols <- c(
    'gameId',
    'teamId',
    'playerId',
    'shiftNumber',
    'period',
    'startTime',
    'endTime',
    'startSecondsElapsedInPeriod',
    'endSecondsElapsedInPeriod',
    'startSecondsElapsedInGame',
    'endSecondsElapsedInGame',
    'duration'
  )
  empty_shift_output <- function() {
    data.frame(
      gameId = integer(),
      teamId = integer(),
      playerId = integer(),
      shiftNumber = integer(),
      period = integer(),
      startTime = character(),
      endTime = character(),
      startSecondsElapsedInPeriod = integer(),
      endSecondsElapsedInPeriod = integer(),
      startSecondsElapsedInGame = integer(),
      endSecondsElapsedInGame = integer(),
      duration = integer(),
      stringsAsFactors = FALSE
    )
  }
  has_valid_shift_rows <- function(shifts) {
    is.data.frame(shifts) &&
      nrow(shifts) > 0L &&
      all(shift_output_cols %in% names(shifts))
  }

  fetch_from_api <- function(game) {
    shifts <- nhl_api(
      path  = 'en/shiftcharts',
      query = list(cayenneExp = sprintf('gameId = %s', game)),
      type  = 's'
    )$data
    raw_shift_cols <- c(
      'id',
      'gameId',
      'teamId',
      'playerId',
      'shiftNumber',
      'period',
      'startTime',
      'endTime',
      'eventDescription'
    )
    if (!is.data.frame(shifts) || !all(raw_shift_cols %in% names(shifts))) {
      return(empty_shift_output())
    }
    shifts <- shifts[order(shifts$teamId), ]
    shifts <- shifts[is.na(shifts$eventDescription), ]
    if (!nrow(shifts)) {
      return(empty_shift_output())
    }
    shifts <- shifts[, c('id', 'gameId', 'teamId', 'playerId', 'shiftNumber', 'period', 'startTime', 'endTime')]
    is_playoffs <- game %/% 1e4 %% 1e2 == 3
    base <- integer(nrow(shifts))
    reg_idx <- shifts$period <= 3L
    ot_idx <- !reg_idx
    base[reg_idx] <- (shifts$period[reg_idx] - 1L) * 1200L
    if (is_playoffs) {
      base[ot_idx] <- 3600L + (shifts$period[ot_idx] - 4L) * 1200L
    } else {
      base[ot_idx] <- 3600L + (shifts$period[ot_idx] - 4L) * 300L
    }
    tp_s  <- strsplit(shifts$startTime, ':', fixed = TRUE)
    s_min <- as.integer(vapply(tp_s, `[`, '', 1L))
    s_sec <- as.integer(vapply(tp_s, `[`, '', 2L))
    s_elp <- 60L * s_min + s_sec
    tp_e  <- strsplit(shifts$endTime, ':', fixed = TRUE)
    e_min <- as.integer(vapply(tp_e, `[`, '', 1L))
    e_sec <- as.integer(vapply(tp_e, `[`, '', 2L))
    e_elp <- 60L * e_min + e_sec
    shifts$startSecondsElapsedInPeriod <- s_elp
    shifts$endSecondsElapsedInPeriod   <- e_elp
    shifts$startSecondsElapsedInGame   <- base + s_elp
    shifts$endSecondsElapsedInGame     <- base + e_elp
    shifts$duration                    <- shifts$endSecondsElapsedInGame - shifts$startSecondsElapsedInGame
    shifts$id <- NULL
    shifts[, shift_output_cols, drop = FALSE]
  }

  fetch_from_html <- function(game) {
    season    <- paste0(game %/% 1e6, game %/% 1e6 + 1)
    game_code <- sprintf('%06d', game %% 10^6)
    shift_report_url <- function(team_tag) {
      sprintf(
        'https://www.nhl.com/scores/htmlreports/%s/%s%s.HTM',
        season,
        team_tag,
        game_code
      )
    }
    normalize_person <- function(x) {
      x <- gsub('\\s+', ' ', x)
      x <- trimws(x)
      x <- toupper(x)
      x <- iconv(x, from = '', to = 'ASCII//TRANSLIT')
      x[is.na(x)] <- ''
      x
    }
    extract_clock <- function(x) {
      if (is.na(x)) return(NA_character_)
      match <- regexpr('[0-9]{1,2}:[0-9]{2}', x)
      if (match[1] < 0L) return(NA_character_)
      substr(
        x,
        match[1],
        match[1] + attr(match, 'match.length')[1] - 1L
      )
    }
    format_clock <- function(x) {
      if (is.na(x)) return(NA_character_)
      parts <- strsplit(x, ':', fixed = TRUE)[[1]]
      if (length(parts) != 2L) return(NA_character_)
      mins <- suppressWarnings(as.integer(parts[1]))
      secs <- suppressWarnings(as.integer(parts[2]))
      if (is.na(mins) || is.na(secs)) return(NA_character_)
      sprintf('%02d:%02d', mins, secs)
    }
    clocks_to_seconds <- function(x) {
      vapply(
        x,
        function(clock) {
          if (
            is.na(clock) ||
            !grepl('^[0-9]{1,2}:[0-9]{2}$', clock)
          ) {
            return(NA_integer_)
          }
          parts <- strsplit(clock, ':', fixed = TRUE)[[1]]
          as.integer(parts[1]) * 60L + as.integer(parts[2])
        },
        integer(1)
      )
    }
    responses <- .perform_parallel_requests(
      list(
        pbp_meta = .nhl_request(
          path = sprintf('v1/gamecenter/%s/play-by-play', game),
          type = 'w'
        ),
        home_report = .html_report_request(shift_report_url('TH')),
        away_report = .html_report_request(shift_report_url('TV'))
      ),
      on_error = 'return'
    )
    if (.parallel_request_failed(responses$pbp_meta)) {
      stop(conditionMessage(responses$pbp_meta))
    }
    if (.parallel_request_failed(responses$home_report)) {
      stop(conditionMessage(responses$home_report))
    }
    if (.parallel_request_failed(responses$away_report)) {
      stop(conditionMessage(responses$away_report))
    }
    pbp_meta <- .nhl_json_from_response(responses$pbp_meta)
    rosters <- pbp_meta$rosterSpots
    rosters$sweaterNumber <- suppressWarnings(
      as.integer(rosters$sweaterNumber)
    )
    rosters$playerLabel <- normalize_person(
      paste(rosters$lastName.default, rosters$firstName.default, sep = ', ')
    )
    lookup_player_id <- function(team_id, sweater_number, player_name) {
      idx <- which(rosters$teamId == team_id)
      if (!length(idx)) return(NA_integer_)
      if (!is.na(sweater_number)) {
        idx_num <- idx[rosters$sweaterNumber[idx] == sweater_number]
        if (length(idx_num) == 1L) {
          return(as.integer(rosters$playerId[idx_num]))
        }
        if (length(idx_num) > 1L && nzchar(player_name)) {
          target   <- normalize_person(player_name)
          idx_name <- idx_num[rosters$playerLabel[idx_num] == target]
          if (length(idx_name)) {
            return(as.integer(rosters$playerId[idx_name[1L]]))
          }
        }
        if (length(idx_num)) {
          return(as.integer(rosters$playerId[idx_num[1L]]))
        }
      }
      if (!is.na(player_name) && nzchar(player_name)) {
        target   <- normalize_person(player_name)
        idx_name <- idx[rosters$playerLabel[idx] == target]
        if (length(idx_name)) {
          return(as.integer(rosters$playerId[idx_name[1L]]))
        }
      }
      NA_integer_
    }
    parse_period_label <- function(x) {
      x <- trimws(toupper(x))
      if (grepl('^[0-9]+$', x)) return(as.integer(x))
      if (x == 'OT') return(4L)
      if (grepl('^[0-9]+OT$', x)) {
        return(3L + as.integer(sub('OT$', '', x)))
      }
      NA_integer_
    }
    parse_shift_report <- function(report_doc, team_id) {
      report <- report_doc
      rows <- xml2::xml_find_all(report, './/tr')
      out  <- vector('list', length(rows))
      n_out <- 0L
      current_player_id <- NA_integer_
      current_player_nm <- NA_character_
      for (row in rows) {
        cells <- xml2::xml_find_all(row, './th|./td')
        if (!length(cells)) next
        txt <- xml2::xml_text(cells, trim = TRUE)
        txt <- gsub('\u00A0', ' ', txt, fixed = TRUE)
        txt <- trimws(txt)
        if (
          length(txt) == 1L &&
          grepl('^[0-9]+\\s+.+', txt[1]) &&
          !grepl('^Shift\\s*#', txt[1], ignore.case = TRUE)
        ) {
          m <- regexec('^([0-9]+)\\s+(.+)$', txt[1])
          p <- regmatches(txt[1], m)[[1]]
          if (length(p) >= 3L) {
            sweater_number    <- suppressWarnings(as.integer(p[2]))
            current_player_nm <- p[3]
            current_player_id <- lookup_player_id(
              team_id,
              sweater_number,
              current_player_nm
            )
          }
          next
        }
        is_shift_row <- length(cells) == 6L &&
          length(txt) >= 4L &&
          grepl('^[0-9]+$', txt[1]) &&
          !is.na(parse_period_label(txt[2])) &&
          grepl('/', txt[3], fixed = TRUE) &&
          grepl('/', txt[4], fixed = TRUE) &&
          grepl('[0-9]{1,2}:[0-9]{2}', txt[3]) &&
          grepl('[0-9]{1,2}:[0-9]{2}', txt[4])
        if (!is_shift_row) next
        start_time <- format_clock(extract_clock(txt[3]))
        end_time   <- format_clock(extract_clock(txt[4]))
        if (is.na(start_time) || is.na(end_time)) next
        n_out <- n_out + 1L
        out[[n_out]] <- data.frame(
          gameId      = as.integer(game),
          teamId      = as.integer(team_id),
          playerId    = as.integer(current_player_id),
          shiftNumber = as.integer(txt[1]),
          period      = parse_period_label(txt[2]),
          startTime   = start_time,
          endTime     = end_time,
          stringsAsFactors = FALSE
        )
      }
      if (!n_out) return(data.frame())
      do.call(rbind, out[seq_len(n_out)])
    }
    home_team_id <- as.integer(pbp_meta$homeTeam$id)
    away_team_id <- as.integer(pbp_meta$awayTeam$id)
    shifts_home  <- parse_shift_report(
      xml2::read_html(httr2::resp_body_raw(responses$home_report)),
      home_team_id
    )
    shifts_away  <- parse_shift_report(
      xml2::read_html(httr2::resp_body_raw(responses$away_report)),
      away_team_id
    )
    shifts       <- rbind(shifts_home, shifts_away)
    if (!nrow(shifts)) stop('No shift rows parsed from HTML reports.')
    shifts$startSecondsElapsedInPeriod <- clocks_to_seconds(shifts$startTime)
    shifts$endSecondsElapsedInPeriod   <- clocks_to_seconds(shifts$endTime)
    is_playoffs <- game %/% 1e4 %% 1e2 == 3
    base <- integer(nrow(shifts))
    reg_idx <- shifts$period <= 3L
    ot_idx <- !reg_idx
    base[reg_idx] <- (shifts$period[reg_idx] - 1L) * 1200L
    if (is_playoffs) {
      base[ot_idx] <- 3600L + (shifts$period[ot_idx] - 4L) * 1200L
    } else {
      base[ot_idx] <- 3600L + (shifts$period[ot_idx] - 4L) * 300L
    }
    shifts$startSecondsElapsedInGame <- base + shifts$startSecondsElapsedInPeriod
    shifts$endSecondsElapsedInGame   <- base + shifts$endSecondsElapsedInPeriod
    if (!is_playoffs) {
      ot_idx <- shifts$period >= 4L
      if (any(ot_idx)) {
        # HTML OT rows can arrive 900 seconds early; re-anchor them to the
        # regular-season OT period start implied by the period label.
        expected_base <- 3600L + (shifts$period[ot_idx] - 4L) * 300L
        actual_base <- shifts$startSecondsElapsedInGame[ot_idx] -
          shifts$startSecondsElapsedInPeriod[ot_idx]
        adjust <- expected_base - actual_base
        shifts$startSecondsElapsedInGame[ot_idx] <- shifts$startSecondsElapsedInGame[ot_idx] + adjust
        shifts$endSecondsElapsedInGame[ot_idx]   <- shifts$endSecondsElapsedInGame[ot_idx] + adjust
      }
    }
    shifts$duration <- shifts$endSecondsElapsedInGame - shifts$startSecondsElapsedInGame
    shifts <- shifts[
      !is.na(shifts$startSecondsElapsedInPeriod) &
        !is.na(shifts$endSecondsElapsedInPeriod),
    ]
    if (!is_playoffs) {
      shifts <- shifts[shifts$period <= 5L, ]
    }
    shifts <- shifts[shifts$duration > 0L, ]
    shifts <- shifts[order(shifts$teamId, shifts$playerId, shifts$shiftNumber), ]
    shifts[, shift_output_cols, drop = FALSE]
  }

  api_error <- NULL
  html_error <- NULL

  api_shifts <- tryCatch(
    fetch_from_api(game),
    error = function(e) {
      api_error <<- e
      empty_shift_output()
    }
  )
  if (has_valid_shift_rows(api_shifts)) {
    return(api_shifts)
  }

  html_shifts <- tryCatch(
    fetch_from_html(game),
    error = function(e) {
      html_error <<- e
      empty_shift_output()
    }
  )
  if (has_valid_shift_rows(html_shifts)) {
    return(html_shifts)
  }

  if (!is.null(api_error)) {
    message(api_error)
  }
  if (!is.null(html_error)) {
    message(html_error)
  }
  message(paste(
    'Invalid argument(s); refer to help file.',
    '\nProvided game:',
    game
  ))
  data.frame()
}

#' Access the real-time game odds for a country by partnered bookmaker
#'
#' `game_odds()` retrieves the real-time game odds for a country by partnered bookmaker as a `data.frame` where each row represents game and includes detail on betting market lines, prices, and provider-level context.
#'
#' @param country two-letter code (e.g., 'CA'); see [countries()] for reference
#'
#' @returns data.frame with one row per game
#' @examples
#' game_odds_CA <- game_odds(country = 'CA')
#' @export

game_odds <- function(country = 'US') {
  tryCatch(
    expr = {
      games <- nhl_api(
        path = sprintf('v1/partner-game/%s/now', country),
        type = 'w'
      )$games
      games[[1]]
      names(games)[names(games) == 'gameType'] <- 'gameTypeId'
      names(games) <- normalize_locale_names(names(games))
      names(games) <- normalize_team_abbrev_cols(names(games))
      games
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

test_that("shift_chart() returns non-empty data.frame", {
  skip_if_offline()
  test <- shift_chart()
  expect_true(is.data.frame(test) && nrow(test) > 0)
})

test_that("shift_chart(0) returns message and empty data.frame", {
  skip_if_offline()
  expect_message(
    test <- shift_chart(0),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.data.frame(test) && nrow(test) == 0)
})

test_that("shift_chart() anchors overtime shifts to game time", {
  local_mocked_bindings(
    .perform_parallel_requests = function(...) stop("HTML unavailable"),
    nhl_api = function(path, query = NULL, type = NULL) {
      expect_equal(path, "en/shiftcharts")
      game <- as.integer(sub("^gameId = ", "", query$cayenneExp))
      list(data = data.frame(
        id = 1:2,
        gameId = game,
        teamId = c(10L, 10L),
        playerId = c(101L, 101L),
        shiftNumber = c(1L, 2L),
        period = c(1L, 4L),
        startTime = c("00:10", "00:05"),
        endTime = c("00:20", "00:15"),
        eventDescription = c(NA_character_, NA_character_),
        stringsAsFactors = FALSE
      ))
    },
    .package = "nhlscraper"
  )

  reg <- shift_chart(2024020001)
  expect_equal(reg$startSecondsElapsedInGame, c(10L, 3605L))
  expect_equal(reg$endSecondsElapsedInGame, c(20L, 3615L))

  playoff <- shift_chart(2024030001)
  expect_equal(playoff$startSecondsElapsedInGame, c(10L, 3605L))
  expect_equal(playoff$endSecondsElapsedInGame, c(20L, 3615L))
})

test_that("shift_chart() prefers API shifts when the API returns rows", {
  html_calls <- 0L

  local_mocked_bindings(
    .perform_parallel_requests = function(...) {
      html_calls <<- html_calls + 1L
      stop("HTML should not be called")
    },
    nhl_api = function(path, query = NULL, type = NULL) {
      expect_equal(path, "en/shiftcharts")
      game <- as.integer(sub("^gameId = ", "", query$cayenneExp))
      list(data = data.frame(
        id = 1L,
        gameId = game,
        teamId = 10L,
        playerId = 101L,
        shiftNumber = 1L,
        period = 1L,
        startTime = "00:10",
        endTime = "00:20",
        eventDescription = NA_character_,
        stringsAsFactors = FALSE
      ))
    },
    .package = "nhlscraper"
  )

  out <- shift_chart(2024020001)

  expect_equal(html_calls, 0L)
  expect_equal(nrow(out), 1L)
  expect_equal(out$playerId, 101L)
})

test_that("shift_chart() falls back to HTML when the API shift feed is empty", {
  home_html <- paste(
    "<table>",
    "<tr><td>10 DOE, JOHN</td></tr>",
    "<tr><td>1</td><td>1</td><td>00:00 / ON</td><td>00:20 / OFF</td><td></td><td></td></tr>",
    "</table>"
  )
  away_html <- paste(
    "<table>",
    "<tr><td>20 SMITH, JANE</td></tr>",
    "<tr><td>1</td><td>1</td><td>00:05 / ON</td><td>00:25 / OFF</td><td></td><td></td></tr>",
    "</table>"
  )

  local_mocked_bindings(
    nhl_api = function(path, query = NULL, type = NULL) {
      expect_equal(path, "en/shiftcharts")
      list(data = data.frame(
        id = integer(),
        gameId = integer(),
        teamId = integer(),
        playerId = integer(),
        shiftNumber = integer(),
        period = integer(),
        startTime = character(),
        endTime = character(),
        eventDescription = character(),
        stringsAsFactors = FALSE
      ))
    },
    .perform_parallel_requests = function(reqs, on_error = "return") {
      expect_named(reqs, c("pbp_meta", "home_report", "away_report"))
      list(
        pbp_meta = structure(list(kind = "pbp_meta"), class = "mock_resp"),
        home_report = httr2::response(body = charToRaw(home_html)),
        away_report = httr2::response(body = charToRaw(away_html))
      )
    },
    .parallel_request_failed = function(resp) FALSE,
    .nhl_json_from_response = function(resp) {
      expect_identical(resp$kind, "pbp_meta")
      list(
        rosterSpots = data.frame(
          teamId = c(1L, 2L),
          sweaterNumber = c("10", "20"),
          playerId = c(101L, 201L),
          lastName.default = c("Doe", "Smith"),
          firstName.default = c("John", "Jane"),
          stringsAsFactors = FALSE
        ),
        homeTeam = list(id = 1L),
        awayTeam = list(id = 2L)
      )
    },
    .package = "nhlscraper"
  )

  out <- shift_chart(2024020001)

  expect_equal(nrow(out), 2L)
  expect_equal(out$playerId, c(101L, 201L))
  expect_equal(out$startSecondsElapsedInPeriod, c(0L, 5L))
  expect_equal(out$endSecondsElapsedInPeriod, c(20L, 25L))
})

#' Access the ESPN games for a season
#'
#' `espn_games()` retrieves the ESPN games for a season as a `data.frame` where each row represents ESPN and includes detail on game timing, matchup state, scoring flow, and situational event detail.
#'
#' @inheritParams roster
#'
#' @returns data.frame with one row per ESPN game
#' @examples
#' ESPN_games_20242025 <- espn_games(season = 20242025)
#' @export

espn_games <- function(season = season_now()) {
  tryCatch(
    expr = {
      seasons_tbl <- seasons()
      season_row  <- seasons_tbl[seasons_tbl$seasonId == season, , drop = FALSE]
      start_src   <- season_row$preseasonStartdate[[1]]
      if (is.null(start_src) || is.na(start_src)) {
        start_src <- season_row$startDate[[1]]
      }
      end_src     <- season_row$endDate[[1]]
      if (!inherits(start_src, 'Date')) {
        start_date <- as.Date(as.character(start_src))
      } else {
        start_date <- start_src
      }
      if (!inherits(end_src, 'Date')) {
        end_date <- as.Date(as.character(end_src))
      } else {
        end_date <- end_src
      }
      start_str  <- format(start_date, '%Y%m%d')
      end_str    <- format(end_date, '%Y%m%d')
      page       <- 1
      all_events <- list()
      repeat {
        events <- espn_api(
          path  = 'events',
          query = list(
            lang   = 'en',
            region = 'us',
            limit  = 1000,
            page   = page,
            dates  = sprintf('%s-%s', start_str, end_str)
          ),
          type  = 'c'
        )
        df <- as.data.frame(events$items, stringsAsFactors = FALSE)
        all_events[[length(all_events) + 1]] <- df
        if (nrow(df) < 1000) {
          break
        }
        page <- page + 1
      }
      out <- do.call(rbind, all_events)
      id  <- sub('.*events/([0-9]+)\\?lang.*', '\\1', out[[1]])
      data.frame(espnGameId = id, stringsAsFactors = FALSE)
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the ESPN summary for a game
#'
#' `espn_game_summary()` retrieves the ESPN summary for a game as a nested `list` that separates summary and detail blocks for date/season filtering windows and chronological context, venue/location geography and regional metadata, and playoff-series progression, round status, and series leverage.
#'
#' @param game integer ID (e.g., 401777460); see [espn_games()] for 
#' reference
#'
#' @returns list with various items
#' @examples
#' ESPN_summary_SCF_20242025 <- espn_game_summary(game = 401777460)
#' @export

espn_game_summary <- function(game = 401777460) {
  tryCatch(
    expr = {
      game  <- espn_api(
        path  = sprintf('events/%s/competitions/%s', game, game),
        type  = 'c'
      )
      names(game)[names(game) == 'id'] <- 'espnGameId'
      keeps <- setdiff(names(game), c(
        '$ref',
        'guid',
        'uid',
        'previewAvailable',
        'recapAvailable',
        'boxscoreAvailable',
        'lineupAvailable',
        'gamecastAvailable',
        'conversationAvailable',
        'commentaryAvailable',
        'summaryAvailable',
        'liveAvailable',
        'ticketsAvailable',
        'shotChartAvailable',
        'possessionArrowAvailable',
        'onWatchESPN',
        'recent',
        'bracketAvailable',
        'wallclockAvailable',
        'highlightsAvailable',
        'gameSource',
        'boxscoreSource',
        'playByPlaySource',
        'linescoreSource',
        'statsSource',
        'situation',
        'status',
        'odds',
        'broadcasts',
        'officials',
        'details',
        'links',
        ''
      ))
      game[keeps]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the ESPN play-by-play for a game
#'
#' `espn_play_by_play()` retrieves the ESPN play-by-play for a game as a `data.frame` where each row represents event and includes detail on game timeline state, period/clock progression, and matchup flow, team identity, affiliation, and matchup-side context, and situational splits across home/road, strength state, and overtime/shootout states.
#'
#' @inheritParams espn_game_summary
#'
#' @returns data.frame with one row per event (play)
#' @examples
#' ESPN_pbp_SCF_20242025 <- espn_play_by_play(game = 401777460)
#' @export

espn_play_by_play <- function(game = 401777460) {
  tryCatch(
    expr = {
      plays <- espn_api(
        path  = sprintf('events/%s/competitions/%s/plays', game, game),
        query = list(lang = 'en', region = 'us', limit = 1000),
        type  = 'c'
      )$items
      names(plays)[names(plays) == 'id'] <- 'espnEventId'
      names(plays)[names(plays) == 'coordinate.x'] <- 'xCoord'
      names(plays)[names(plays) == 'coordinate.y'] <- 'yCoord'
      old_names <- names(plays)
      keep_ref <- grepl('(^\\$ref$|\\.\\$ref$)', old_names)
      new_names <- dot_to_camel(old_names)
      names(plays) <- ifelse(keep_ref, old_names, new_names)
      names(plays)[names(plays) == 'periodNumber'] <- 'period'
      plays
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname espn_play_by_play
#' @export

espn_pbp <- function(game = 401777460) {
  espn_play_by_play(game)
}

#' Access the ESPN odds for a game
#'
#' `espn_game_odds()` retrieves the ESPN odds for a game as a `data.frame` where each row represents provider and includes detail on team identity, affiliation, and matchup-side context plus betting market snapshots with side/total prices and provider variation.
#'
#' @inheritParams espn_game_summary
#'
#' @returns data.frame with one row per provider
#' @examples
#' ESPN_odds_SCF_20242025 <- espn_game_odds(game = 401777460)
#' @export

espn_game_odds <- function(game = 401777460) {
  tryCatch(
    expr = {
      odds <- espn_api(
        path  = sprintf('events/%s/competitions/%s/odds', game, game),
        type  = 'c'
      )$items
      old_names <- names(odds)
      keep_ref <- grepl('(^\\$ref$|\\.\\$ref$)', old_names)
      new_names <- dot_to_camel(old_names)
      names(odds) <- ifelse(keep_ref, old_names, new_names)
      odds
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access all the ESPN teams for a season
#' 
#' `get_espn_teams()` is defunct. Use [espn_teams()] instead.
#' 
#' @export

get_espn_teams <- function() {
  .Defunct(
    new     = 'espn_teams()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_teams()` is defunct.',
      'Use `espn_teams()` instead.'
    )
  )
}

#' Access the ESPN summary for a team and season
#' 
#' `get_espn_team()` is defunct. Use [espn_team_summary()] instead.
#' 
#' @export

get_espn_team <- function() {
  .Defunct(
    new     = 'espn_team_summary()',
    package = 'nhlscraper',
    msg     = paste(
      '`get_espn_team()` is defunct.',
      'Use `espn_team_summary()` instead.'
    )
  )
}

#' Access the ESPN transactions for a season
#'
#' `espn_transactions()` retrieves the ESPN transactions for a season as a `data.frame` where each row represents transaction and includes detail on availability and transaction-status tracking detail.
#'
#' @param season integer in YYYYYYYY (e.g., 20242025); the summer of the latter 
#' year is included
#'
#' @returns data.frame with one row per transaction
#' @examples
#' ESPN_transactions_20242025 <- espn_transactions(season = 20242025)
#' @export

espn_transactions <- function(season = season_now()) {
  tryCatch(
    expr = {
      season <- as.integer(season[[1]])
      if (is.na(season)) {
        stop('Invalid season.')
      }
      seasons_tbl <- seasons()
      idx         <- which(seasons_tbl$seasonId == season)
      if (!length(idx)) {
        stop('No season metadata found for season: ', season)
      }
      idx         <- idx[[1]]
      this_season <- seasons_tbl[idx, , drop = FALSE]
      start_src   <- this_season$preseasonStartdate[[1]]
      if (is.null(start_src) || is.na(start_src)) {
        start_src <- this_season$startDate[[1]]
      }
      if (idx < nrow(seasons_tbl)) {
        next_season  <- seasons_tbl[idx + 1, , drop = FALSE]
        next_start   <- next_season$preseasonStartdate[[1]]
        if (is.null(next_start) || is.na(next_start)) {
          next_start <- next_season$startDate[[1]]
        }
        if (!inherits(next_start, 'Date')) {
          next_start <- as.Date(as.character(next_start))
        }
        end_src <- next_start - 1
      } else {
        end_src <- this_season$endDate[[1]]
      }
      if (!inherits(start_src, 'Date')) {
        start_date <- as.Date(as.character(start_src))
      } else {
        start_date <- start_src
      }
      if (!inherits(end_src, 'Date')) {
        end_date <- as.Date(as.character(end_src))
      } else {
        end_date <- end_src
      }
      start_str <- format(start_date, '%Y%m%d')
      end_str   <- format(end_date, '%Y%m%d')
      page             <- 1
      all_transactions <- list()
      repeat {
        transactions <- espn_api(
          path  = 'transactions',
          query = list(
            limit = 1000,
            page  = page,
            dates = sprintf('%s-%s', start_str, end_str)
          ),
          type = 'g'
        )
        df <- as.data.frame(transactions$transactions, stringsAsFactors = FALSE)
        all_transactions[[length(all_transactions) + 1]] <- df
        if (nrow(df) < 1000) {
          break
        }
        page <- page + 1
      }
      transactions <- do.call(rbind, all_transactions)
      if (!is.null(transactions) && ncol(transactions)) {
        nms <- names(transactions)
        nms <- normalize_locale_names(nms)
        nms <- normalize_team_abbrev_cols(nms)
        nms <- ifelse(nms == 'teamId', 'espnTeamId', nms)
        names(transactions) <- nms
      }
      transactions
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the ESPN futures for a season
#'
#' `espn_futures()` retrieves the ESPN futures for a season as a `data.frame` where each row represents type and includes detail on betting market snapshots with side/total prices and provider variation.
#'
#' @inheritParams roster
#'
#' @returns nested data.frame with one row per type (outer) and book (inner)
#' @examples
#' ESPN_futures_20252026 <- espn_futures(20252026)
#' @export

espn_futures <- function(season = season_now()) {
  tryCatch(
    expr = {
      season <- as.integer(season[[1]])
      if (is.na(season) || !season %in% seasons()$seasonId) {
        stop('Invalid season.')
      }
      season <- season %% 1e4
      futures <- espn_api(
        path  = sprintf('seasons/%s/futures', season),
        query = list(lang = 'en', region = 'us', limit = 1000),
        type  = 'c'
      )$items
      if (!is.data.frame(futures)) {
        futures <- as.data.frame(futures, stringsAsFactors = FALSE)
      }
      names(futures)[names(futures) == 'id']          <- 'espnFutureId'
      names(futures)[names(futures) == 'name']        <- 'futureName'
      names(futures)[names(futures) == 'displayName'] <- 'futureDisplayName'
      futures
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the real-time ESPN injury reports
#'
#' `espn_injuries()` retrieves the real-time ESPN injury reports as a `data.frame` where each row represents team and includes detail on availability status tracking for injuries or transactions.
#'
#' @returns nested data.frame with one row per team (outer) and player (inner)
#' @examples
#' ESPN_injuries_now <- espn_injuries()
#' @export

espn_injuries <- function() {
  tryCatch({
    teams <- espn_api(
      path  = 'injuries',
      query = list(limit = 1000),
      type  = 'g'
    )$injuries
    names(teams)[names(teams) == 'id']          <- 'espnTeamId'
    names(teams)[names(teams) == 'displayName'] <- 'teamDisplayName'
    teams[order(as.integer(teams$espnTeamId)), ]
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' Call the NHL API with 429 (rate limit) error-handling
#' 
#' @param path character
#' @param query list
#' @param type character of 'w' for web, 's' for stats, and 'r' for records
#' @returns parsed JSON (i.e., data.frame or list)
#' @keywords internal

nhl_api <- function(path, query = list(), type) {
  .nhl_json_from_response(
    httr2::req_perform(.nhl_request(path = path, query = query, type = type))
  )
}

#' Build an NHL API request
#'
#' `.nhl_request()` constructs an NHL API request with the standard retry policy
#' used throughout the package.
#'
#' @param path character
#' @param query list
#' @param type character of 'w' for web, 's' for stats, and 'r' for records
#' @returns httr2 request object
#' @keywords internal
.nhl_request <- function(path, query = list(), type) {
  base <- switch(
    type,
    w = 'https://api-web.nhle.com/',
    s = 'https://api.nhle.com/stats/rest/',
    r = 'https://records.nhl.com/site/api/'
  )
  req <- httr2::request(paste0(base, path))
  req <- do.call(httr2::req_url_query, c(list(req), query))
  httr2::req_retry(
    req,
    max_tries = 3,
    backoff = function(attempt) 2 ^ (attempt - 1),
    is_transient = function(resp) httr2::resp_status(resp) == 429
  )
}

#' Parse an NHL API response as JSON
#'
#' `.nhl_json_from_response()` converts an `httr2` response object into a parsed
#' JSON object using the package's standard flattening settings.
#'
#' @param resp httr2 response object
#' @returns parsed JSON (i.e., data.frame or list)
#' @keywords internal
.nhl_json_from_response <- function(resp) {
  jsonlite::fromJSON(
    httr2::resp_body_string(resp, encoding = 'UTF-8'),
    simplifyVector = TRUE,
    flatten = TRUE
  )
}

#' Build an HTML report request
#'
#' `.html_report_request()` constructs an HTML report request with the same retry
#' policy used for API calls.
#'
#' @param url character scalar
#' @returns httr2 request object
#' @keywords internal
.html_report_request <- function(url) {
  httr2::req_retry(
    httr2::request(url),
    max_tries = 3,
    backoff = function(attempt) 2 ^ (attempt - 1),
    is_transient = function(resp) httr2::resp_status(resp) == 429
  )
}

#' Perform multiple requests concurrently
#'
#' `.perform_parallel_requests()` executes a list of `httr2` requests with
#' libcurl multi support and preserves the input names on the output.
#'
#' @param reqs list of httr2 request objects
#' @param on_error forwarded to [httr2::req_perform_parallel()]
#' @returns list of responses or httr2 failure objects
#' @keywords internal
.perform_parallel_requests <- function(reqs, on_error = c('stop', 'return', 'continue')) {
  on_error <- match.arg(on_error)
  req_names <- names(reqs)
  out <- httr2::req_perform_parallel(
    unname(reqs),
    on_error = on_error,
    progress = FALSE
  )
  if (!is.null(req_names)) {
    names(out) <- req_names
  }
  out
}

#' Call the ESPN API with 429 (rate limit) error-handling
#' 
#' @param path character
#' @param query list
#' @param type character of 'g' for general and 'c' for core
#' @returns parsed JSON (i.e., data.frame or list)
#' @keywords internal

espn_api <- function(path, query=list(), type) {
  base <- switch(
    type, 
    g = 'https://site.api.espn.com/apis/site/v2/sports/hockey/nhl/',
    c = 'https://sports.core.api.espn.com/v2/sports/hockey/leagues/nhl/'
  )
  req <- httr2::request(paste0(base, path))
  req <- do.call(httr2::req_url_query, c(list(req), query))
  req <- httr2::req_retry(
    req,
    max_tries    = 3,
    backoff      = function(attempt) 2 ^ (attempt - 1),
    is_transient = function(resp) httr2::resp_status(resp) == 429
  )
  resp <- httr2::req_perform(req)
  jsonlite::fromJSON(
    httr2::resp_body_string(resp, encoding = 'UTF-8'),
    simplifyVector = TRUE,
    flatten        = TRUE
  )
}

#' Convert to the appropriate game type ID
#' 
#' @param game_type integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season) OR character of 'pre', 'regular', or 'playoff'/'post'
#' @returns integer in 1:3
#' @keywords internal

to_game_type_id <- function(game_type) {
  switch(
    tolower(as.character(game_type)),
    `1`     = 1,
    pre     = 1,
    `2`     = 2,
    regular = 2,
    `3`     = 3,
    playoff = 3,
    post    = 3,
    ''
  )
}

#' Normalize the team key
#' 
#' @param team integer ID (e.g., 21), character full name (e.g., 'Colorado 
#' Avalanche'), OR three-letter code (e.g., 'COL')
#' @returns integer in 1:68, character full name, OR three-letter code
#' @keywords internal

normalize_team_key <- function(team) {
  gsub('[^a-z0-9]', '', tolower(trimws(as.character(team))))
}

#' Convert to the appropriate team three-letter code
#'
#' @param team integer ID (e.g., 21), character full name (e.g., 'Colorado 
#' Avalanche'), OR three-letter code (e.g., 'COL')
#' @returns three-letter code
#' @keywords internal

to_team_tri_code <- function(team, lookup = .to_team_tri_code) {
  unname(lookup[normalize_team_key(team)])
}

#' Convert to the appropriate team ID
#'
#' @param team integer ID (e.g., 21), character full name (e.g., 'Colorado 
#' Avalanche'), OR three-letter code (e.g., 'COL')
#' @returns integer in 1:68
#' @keywords internal

to_team_id <- function(team, lookup = .to_team_id) {
  unname(lookup[normalize_team_key(team)])
}

#' Convert dot-delimited names to camelCase
#' 
#' @param x character vector
#' @returns character vector
#' @keywords internal

dot_to_camel <- function(x) {
  parts <- strsplit(x, '\\.')
  vapply(parts, function(p) {
    if (length(p) == 1L) return(p)
    paste0(
      p[1],
      paste0(
        toupper(substr(p[-1], 1, 1)),
        substr(p[-1], 2, nchar(p[-1])),
        collapse = ''
      )
    )
  }, character(1))
}

#' Normalize locale-style dotted columns to camelCase
#' 
#' Converts dotted names (e.g., firstName.default, name.cs) to camelCase and
#' removes trailing `Default` from default-language fields.
#'
#' @param x character vector
#' @returns character vector
#' @keywords internal

normalize_locale_names <- function(x) {
  x <- dot_to_camel(x)
  sub('Default$', '', x)
}

#' Scope generic person-name columns to an entity
#'
#' @param x character vector
#' @param prefix character scalar (e.g., 'player', 'goalie', 'skater')
#' @returns character vector
#' @keywords internal

scope_person_name_cols <- function(x, prefix) {
  x <- gsub('^firstName([A-Z].*)?$', paste0(prefix, 'FirstName\\1'), x, perl = TRUE)
  x <- gsub('^lastName([A-Z].*)?$', paste0(prefix, 'LastName\\1'), x, perl = TRUE)
  x <- gsub('^fullName([A-Z].*)?$', paste0(prefix, 'FullName\\1'), x, perl = TRUE)
  x <- gsub('^name([A-Z].*)?$', paste0(prefix, 'Name\\1'), x, perl = TRUE)
  x
}

#' Normalize team abbreviation columns to team tri-code names
#'
#' @param x character vector
#' @returns character vector
#' @keywords internal

normalize_team_abbrev_cols <- function(x) {
  x <- sub('TeamAbbreviations$', 'TeamTriCodes', x, perl = TRUE)
  x <- sub('TeamAbbreviation$', 'TeamTriCode', x, perl = TRUE)
  x <- sub('TeamAbbrevs$', 'TeamTriCodes', x, perl = TRUE)
  x <- sub('TeamAbbrev$', 'TeamTriCode', x, perl = TRUE)
  map <- c(
    opponentAbbrev = 'opponentTeamTriCode',
    teamAbbreviation  = 'teamTriCode',
    teamAbbreviations = 'teamTriCodes',
    teamAbbrev     = 'teamTriCode',
    teamAbbrevs    = 'teamTriCodes'
  )
  idx <- match(x, names(map))
  x[!is.na(idx)] <- unname(map[idx[!is.na(idx)]])
  x
}

#' Access the season(s) and game type(s) in which there exists goalie EDGE 
#' statistics
#'
#' `goalie_edge_seasons()` retrieves the season(s) and game type(s) in which there exists goalie EDGE statistics as a `data.frame` where each row represents season and includes detail on date/season filtering windows and chronological context plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @returns data.frame with one row per season
#' @examples
#' goalie_EDGE_seasons <- goalie_edge_seasons()
#' @export

goalie_edge_seasons <- function() {
  tryCatch(
    expr = {
      seasons <- nhl_api(
        path = sprintf('v1/edge/goalie-landing/now'),
        type = 'w'
      )$seasonsWithEdgeStats
      names(seasons)[names(seasons) == 'id']        <- 'seasonId'
      names(seasons)[names(seasons) == 'gameTypes'] <- 'gameTypeIds'
      seasons
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the goalie EDGE statistics leaders for a season and game type
#'
#' `goalie_edge_leaders()` retrieves the goalie EDGE statistics leaders for a season and game type as a nested `list` that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
#'
#' @param season integer in YYYYYYYY (e.g., 20242025); see 
#' [goalie_edge_seasons()] for reference
#' @param game_type integer in 1:3 (where 1 = pre-season, 2 = regular season, 3 
#' = playoff/post-season) OR character of 'pre', 'regular', or 
#' 'playoff'/'post'; see [goalie_edge_seasons()] for reference; most functions 
#' will NOT support pre-season
#'
#' @returns list of various items
#' @examples
#' goalie_EDGE_leaders_regular_20242025 <- goalie_edge_leaders(
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

goalie_edge_leaders <- function(season = 'now', game_type = '') {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf(
          'v1/edge/goalie-landing/%s/%s', 
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )$leaders
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the EDGE summary for a goalie, season, and game type
#'
#' `goalie_edge_summary()` retrieves the EDGE summary for a goalie, season, and game type as a nested `list` that separates summary and detail blocks for player identity, role, handedness, and biographical profile plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams goalie_edge_leaders
#' @param player integer ID (e.g., 8478406)
#'
#' @returns list of various items
#' @examples
#' Mackenzie_Blackwood_EDGE_summary_regular_20242025 <- goalie_edge_summary(
#'   player    = 8478406, 
#'   season    = 20242025,
#'   game_type = 2
#' )
#' @export

goalie_edge_summary <- function(
    player    = 8476945, 
    season    = 'now', 
    game_type = ''
) {
  tryCatch(
    expr = {
      nhl_api(
        path = sprintf(
          'v1/edge/goalie-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      list()
    }
  )
}

#' Access the EDGE save percentage statistics for a goalie, season, game type, 
#' and category
#'
#' `goalie_edge_save_percentage()` retrieves the EDGE save percentage statistics for a goalie, season, game type, and category as a nested `list` that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams goalie_edge_summary
#' @param category character of 'd'/'details' or 'l'/'l10'/'last 10'
#'
#' @returns list with two items (category = 'details') or data.frame with one 
#' row per game (category = 'last 10')
#' @examples
#' Mackenzie_Blackwood_L10_sP_regular_20242025 <- 
#'   goalie_edge_save_percentage(
#'     player    = 8478406,
#'     season    = 20242025,
#'     game_type = 2,
#'     category  = 'L'
#'   )
#' @export

goalie_edge_save_percentage <- function(
  player    = 8476945,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'savePctgDetails',
        l = 'savePctgLast10'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/goalie-save-percentage-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' Access the EDGE 5 vs. 5 statistics for a goalie, season, game type, and 
#' category
#'
#' `goalie_edge_five_versus_five()` retrieves the EDGE 5 vs. 5 statistics for a goalie, season, game type, and category as a nested `list` that separates summary and detail blocks for production, workload, efficiency, and result-level performance outcomes plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams goalie_edge_summary
#' @param category character of 'd'/'details' or 'l'/'l10'/'last 10'
#'
#' @returns list with four items (category = 'details') or data.frame with 
#' one row per game (category = 'last 10')
#' @examples
#' Mackenzie_Blackwood_L10_5_vs_5_regular_20242025 <- goalie_edge_five_versus_five(
#'   player    = 8478406,
#'   season    = 20242025,
#'   game_type = 2,
#'   category  = 'L'
#'  )
#' @export

goalie_edge_five_versus_five <- function(
  player    = 8476945,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'savePctg5v5Details',
        l = 'savePctg5v5Last10'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/goalie-5v5-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

#' @rdname goalie_edge_five_versus_five
#' @export
goalie_edge_5_vs_5 <- function(
  player    = 8476945,
  season    = 'now', 
  game_type = '', 
  category  = 'details'
) {
  goalie_edge_five_versus_five(player, season, game_type, category)
}

#' Access the EDGE shot location statistics for a goalie, season, game type, 
#' and category
#'
#' `goalie_edge_shot_location()` retrieves the EDGE shot location statistics for a goalie, season, game type, and category as a `data.frame` where each row represents shot location and includes detail on production, workload, efficiency, and result-level performance outcomes plus NHL EDGE style tracking outputs and relative-performance context.
#'
#' @inheritParams goalie_edge_summary
#' @param category character of 'd'/details' or 't'/'totals'
#'
#' @returns data.frame with one row per shot location
#' @examples
#' Mackenzie_Blackwood_shot_location_totals_regular_20242025 <- 
#'   goalie_edge_shot_location(
#'     player    = 8478406,
#'     season    = 20242025,
#'     game_type = 2,
#'     category  = 'T'
#'   )
#' @export

goalie_edge_shot_location <- function(
    player    = 8476945,
    season    = 'now', 
    game_type = '', 
    category  = 'details'
) {
  tryCatch(
    expr = {
      category <- switch(
        substring(tolower(category), 1, 1),
        d = 'shotLocationDetails',
        t = 'shotLocationTotals'
      )
      nhl_api(
        path = sprintf(
          'v1/edge/goalie-shot-location-detail/%s/%s/%s', 
          player,
          season, 
          to_game_type_id(game_type)
        ),
        type = 'w'
      )[[category]]
    },
    error = function(e) {
      message('Invalid argument(s); refer to help file.')
      data.frame()
    }
  )
}

test_that("goalie_edge_summary() returns non-empty list", {
  skip_if_offline()
  test <- goalie_edge_summary()
  expect_true(is.list(test) && length(test) > 0)
})

test_that("goalie_edge_summary(0) returns message and empty list", {
  skip_if_offline()
  expect_message(
    test <- goalie_edge_summary(0),
    'Invalid argument\\(s\\); refer to help file\\.'
  )
  expect_true(is.list(test) && length(test) == 0)
})

#' Access the replay for an event
#'
#' `replay()` retrieves the replay for an event as a `data.frame` where each row represents decisecond and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
#'
#' @param game integer ID (e.g., 2025020262); see [games()] for reference
#' @param event integer ID (e.g., 751); see [gc_play_by_play()] and/or 
#' [wsc_play_by_play()] for reference; must be a 'goal' event
#'
#' @returns data.frame with one row per decisecond
#' @examples
#' Gabriel_Landeskog_first_regular_goal_back_replay <- replay(
#'   game  = 2025020262,
#'   event = 751
#' )
#' @export

replay <- function(game = 2023030417, event = 866) {
  tryCatch(
    expr = {
      to_num <- function(x) suppressWarnings(as.numeric(x))
      to_replay_x_coord <- function(x_raw) {
        x_raw <- to_num(x_raw)
        (x_raw / 12) - 100
      }
      to_replay_y_coord <- function(y_raw) {
        y_raw <- to_num(y_raw)
        42.5 - (y_raw / 12)
      }
      is_missing_chr <- function(x) {
        x_chr <- trimws(as.character(x))
        is.na(x_chr) | x_chr == ''
      }
      base   <- 'https://wsr.nhle.com/'
      year   <- game %/% 1e6
      season <- paste0(as.character(year), as.character(year + 1))
      url    <- sprintf(
        '%ssprites/%s/%s/ev%s.json',
        base,
        season,
        game,
        event
      )
      req <- httr2::request(url)
      req <- httr2::req_headers(
        req,
        referer      = 'https://www.nhl.com/',
        'user-agent' = 'Chrome/130.0.0.0'
      )
      req <- httr2::req_retry(
        req,
        max_tries    = 3,
        backoff      = function(attempt) 2 ^ (attempt - 1),
        is_transient = function(resp) httr2::resp_status(resp) == 429
      )
      resp   <- httr2::req_perform(req)
      replay <- jsonlite::fromJSON(
        httr2::resp_body_string(resp, encoding = 'UTF-8'),
        simplifyVector = TRUE,
        flatten        = TRUE
      )
      id_cols <- grep('^onIce\\.[^.]+\\.id$', names(replay), value = TRUE)
      if (!length(id_cols)) {
        return(data.frame(timeStamp = replay$timeStamp))
      }
      slots <- sub('\\.id$', '', id_cols)
      get_slot_col <- function(slot, field) {
        col <- paste0(slot, '.', field)
        if (col %in% names(replay)) replay[[col]] else rep(NA, nrow(replay))
      }
      puck_candidates <- vapply(
        slots,
        function(slot) {
          player_id <- get_slot_col(slot, 'playerId')
          x_raw  <- to_num(get_slot_col(slot, 'x'))
          y_raw  <- to_num(get_slot_col(slot, 'y'))
          id_raw <- to_num(get_slot_col(slot, 'id'))
          all(is_missing_chr(player_id)) &&
            (any(!is.na(x_raw)) || any(!is.na(y_raw))) &&
            (grepl('^onIce\\.1$', slot) || any(id_raw == 1, na.rm = TRUE))
        },
        logical(1)
      )
      if (any(puck_candidates)) {
        puck_slot <- slots[which(puck_candidates)[1]]
      } else {
        fallback <- vapply(
          slots,
          function(slot) {
            player_id <- get_slot_col(slot, 'playerId')
            x_raw <- to_num(get_slot_col(slot, 'x'))
            y_raw <- to_num(get_slot_col(slot, 'y'))
            all(is_missing_chr(player_id)) &&
              (any(!is.na(x_raw)) || any(!is.na(y_raw)))
          },
          logical(1)
        )
        if (any(fallback)) {
          puck_slot <- slots[which(fallback)[1]]
        } else if ('onIce.1' %in% slots) {
          puck_slot <- 'onIce.1'
        } else {
          puck_slot <- slots[1]
        }
      }
      out <- data.frame(timeStamp = replay$timeStamp, stringsAsFactors = FALSE)
      puck_x_raw <- to_num(get_slot_col(puck_slot, 'x'))
      puck_y_raw <- to_num(get_slot_col(puck_slot, 'y'))
      out$puckXCoordRaw <- puck_x_raw
      out$puckYCoordRaw <- puck_y_raw
      out$puckXCoord <- to_replay_x_coord(puck_x_raw)
      out$puckYCoord <- to_replay_y_coord(puck_y_raw)
      player_slots   <- slots[slots != puck_slot]
      for (i in seq_along(player_slots)) {
        slot  <- player_slots[i]
        stem  <- paste0('player', i)
        x_raw <- to_num(get_slot_col(slot, 'x'))
        y_raw <- to_num(get_slot_col(slot, 'y'))
        out[[paste0(stem, 'PPTId')]] <- to_num(get_slot_col(slot, 'id'))
        out[[paste0(stem, 'PlayerId')]]  <- to_num(get_slot_col(slot, 'playerId'))
        out[[paste0(stem, 'XCoordRaw')]] <- x_raw
        out[[paste0(stem, 'YCoordRaw')]] <- y_raw
        out[[paste0(stem, 'XCoord')]] <- to_replay_x_coord(x_raw)
        out[[paste0(stem, 'YCoord')]] <- to_replay_y_coord(y_raw)
        out[[paste0(stem, 'TeamId')]] <- to_num(get_slot_col(slot, 'teamId'))
        out[[paste0(stem, 'TeamTriCode')]] <- as.character(get_slot_col(slot, 'teamAbbrev'))
      }
      out
    },
    error = function(e) {
      message(paste(
        'Invalid argument(s); refer to help file.',
        '\nProvided game:',
        game,
        '\nProvided event:',
        event
      ))
      data.frame()
    }
  )
}

#' Access all the penalty shots
#'
#' `penalty_shots()` retrieves all the penalty shots as a `data.frame` where each row represents penalty shot and includes detail on game timeline state, period/clock progression, and matchup flow plus date/season filtering windows and chronological context.
#'
#' @returns data.frame with one row per penalty shot
#' @examples
#' all_pss <- penalty_shots()
#' @export

penalty_shots <- function() {
  tryCatch({
    pss    <- nhl_api(
      path = 'penalty-shots',
      type = 'r'
    )$data
    pss$id <- NULL
    pss    <- pss[order(pss$gameId), ]
    names(pss)[names(pss) == 'season']   <- 'seasonId'
    names(pss)[names(pss) == 'gameType'] <- 'gameTypeId'
    names(pss) <- normalize_team_abbrev_cols(names(pss))
    pss
  }, error = function(e) {
    message('Unable to create connection; please try again later.')
    data.frame()
  })
}

#' @rdname penalty_shots
#' @export

pss <- function() {
  penalty_shots()
}

---
title: 'How Costly Are Back-to-Backs in the Salary-Cap Era?'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How Costly Are Back-to-Backs in the Salary-Cap Era?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}
```

## Overview

Few schedule complaints are as universal as the back-to-back. Coaches hate them, players complain about them, broadcasters invoke them as a built-in excuse, and fans treat them as a warning label the moment the puck drops. But how big is the penalty, really? Is the back-to-back tax large enough to show up cleanly in results, or is it mostly noise wrapped in narrative? This example uses `nhlscraper::games()` and `nhlscraper::teams()` to build every regular-season team-game from the salary-cap era onward, compute the number of off-days before each game, and compare win rate and goal differential across rest buckets. The result is a compact way to turn a schedule cliché into a league-wide estimate.

## Build Team-Game Table

`nhlscraper::games()` gives us one row per game. To study rest, we need one row per **team-game**, which means expanding each game into a home-team record and an away-team record.

```{r data}
# Pull regular-season game and team records.
games_tbl <- nhlscraper::games()
teams_tbl <- nhlscraper::teams()

# Keep salary-cap regular-season games.
games_tbl <- games_tbl[
  games_tbl[['seasonId']] >= 20052006 &
    games_tbl[['gameTypeId']] == 2,
  c(
    'gameId',
    'seasonId',
    'gameDate',
    'homeTeamId',
    'visitingTeamId',
    'homeScore',
    'visitingScore'
  )
]

# Expand games into team-game rows.
home_games <- data.frame(
  gameId = games_tbl[['gameId']],
  seasonId = games_tbl[['seasonId']],
  gameDate = as.Date(games_tbl[['gameDate']]),
  teamId = games_tbl[['homeTeamId']],
  isHome = TRUE,
  goalsFor = games_tbl[['homeScore']],
  goalsAgainst = games_tbl[['visitingScore']]
)
away_games <- data.frame(
  gameId = games_tbl[['gameId']],
  seasonId = games_tbl[['seasonId']],
  gameDate = as.Date(games_tbl[['gameDate']]),
  teamId = games_tbl[['visitingTeamId']],
  isHome = FALSE,
  goalsFor = games_tbl[['visitingScore']],
  goalsAgainst = games_tbl[['homeScore']]
)
team_games <- rbind(home_games, away_games)

# Sort within team and compute off-days since previous game.
team_games <- team_games[order(
  team_games[['teamId']],
  team_games[['gameDate']],
  team_games[['gameId']]
), ]
team_games[['previousGameDate']] <- ave(
  team_games[['gameDate']],
  team_games[['teamId']],
  FUN = function(x) c(as.Date(NA), utils::head(x, -1))
)
team_games[['restDays']] <-
  as.integer(team_games[['gameDate']] - team_games[['previousGameDate']]) - 1L
team_games <- team_games[!is.na(team_games[['restDays']]), ]

# Bucket rest and compute result metrics.
team_games[['restBucket']] <- ifelse(
  team_games[['restDays']] >= 3,
  '3+',
  as.character(team_games[['restDays']])
)
team_games[['restBucket']] <- factor(
  team_games[['restBucket']],
  levels = c('0', '1', '2', '3+')
)
team_games[['win']] <- team_games[['goalsFor']] > team_games[['goalsAgainst']]
team_games[['goalDiff']] <-
  team_games[['goalsFor']] - team_games[['goalsAgainst']]
```

The key definition here is simple: if a team last played yesterday, the next game gets `restDays = 0`, which is the second night of a back-to-back.

## Start With League-Wide Rest Buckets

Now we can ask the direct question: how do teams perform on zero rest compared with one, two, or three-plus days off?

```{r rest-table}
# Summarize results by rest bucket.
rest_summary <- aggregate(
  cbind(win, goalDiff) ~ restBucket,
  data = team_games,
  FUN = mean
)
rest_counts <- as.data.frame(table(team_games[['restBucket']]))
names(rest_counts) <- c('restBucket', 'games')
rest_summary <- merge(rest_summary, rest_counts, by = 'restBucket')
rest_summary <- rest_summary[
  match(levels(team_games[['restBucket']]), rest_summary[['restBucket']]),
  c('restBucket', 'games', 'win', 'goalDiff')
]

make_table(
  rest_summary,
  caption = 'Win rate and average goal differential by rest bucket.'
)
```

The back-to-back effect is real. Teams on zero rest win about `r sprintf('%.1f', 100 * rest_summary[['win']][rest_summary[['restBucket']] == '0'])` percent of the time, compared with about `r sprintf('%.1f', 100 * rest_summary[['win']][rest_summary[['restBucket']] == '2'])` percent on two days of rest. Goal differential moves the same way: teams on zero rest are underwater on average, while teams with one or two days off are slightly positive. That is already enough to say the schedule complaint has teeth, but the context gets sharper once we split home and road games.

## Plot Rest Curve

```{r rest-plot, fig.cap = 'Win rate across rest buckets in the salary-cap era.'}
graphics::barplot(
  rest_summary[['win']],
  names.arg = rest_summary[['restBucket']],
  col = c('#d62828', '#f77f00', '#fcbf49', '#90be6d'),
  border = NA,
  ylim = c(0, 0.6),
  xlab = 'Days of Rest',
  ylab = 'Win Rate'
)
graphics::abline(
  h = mean(team_games[['win']]),
  lty = 2,
  col = '#4d4d4d'
)
```

The overall curve climbs immediately once teams move off the second night of a back-to-back. It does not rise forever, because three-plus days of rest can include rust or unusual schedule contexts, but the biggest step is clearly from zero days to one.

## Separate Home and Road Context

Travel and venue matter. Back-to-backs on the road should be tougher than back-to-backs at home, and even rested teams usually perform better in their own building.

```{r home-road-table}
# Summarize rest effect by venue.
home_road_summary <- aggregate(
  win ~ restBucket + isHome,
  data = team_games,
  FUN = mean
)
home_wins <- home_road_summary[
  home_road_summary[['isHome']],
  c('restBucket', 'win')
]
away_wins <- home_road_summary[
  !home_road_summary[['isHome']],
  c('restBucket', 'win')
]
names(home_wins)[names(home_wins) == 'win'] <- 'homeWinRate'
names(away_wins)[names(away_wins) == 'win'] <- 'awayWinRate'
home_road_table <- merge(home_wins, away_wins, by = 'restBucket')
home_road_table <- home_road_table[
  match(levels(team_games[['restBucket']]), home_road_table[['restBucket']]),
]
make_table(
  home_road_table,
  caption = 'Home and road win rate by rest bucket.'
)
```

Back-to-backs hurt everywhere, but they hurt more on the road. On zero rest, road teams win only about `r sprintf('%.1f', 100 * home_road_table[['awayWinRate']][home_road_table[['restBucket']] == '0'])` percent of the time, while home teams on zero rest still manage about `r sprintf('%.1f', 100 * home_road_table[['homeWinRate']][home_road_table[['restBucket']] == '0'])` percent.

```{r home-road-plot, fig.cap = 'Home and road win rate across rest buckets.'}
graphics::plot(
  seq_len(nrow(home_road_table)),
  home_road_table[['homeWinRate']],
  type = 'b',
  pch = 19,
  lwd = 2,
  col = '#1d3557',
  xaxt = 'n',
  ylim = c(0.35, 0.60),
  xlab = 'Days of Rest',
  ylab = 'Win Rate'
)
graphics::lines(
  seq_len(nrow(home_road_table)),
  home_road_table[['awayWinRate']],
  type = 'b',
  pch = 19,
  lwd = 2,
  col = '#e63946'
)
graphics::axis(
  side = 1,
  at = seq_len(nrow(home_road_table)),
  labels = home_road_table[['restBucket']]
)
graphics::legend(
  'bottomright',
  legend = c('Home', 'Away'),
  col = c('#1d3557', '#e63946'),
  pch = 19,
  lwd = 2,
  bty = 'n'
)
```

The two lines stay separated almost the whole way through. Rest helps, but home ice helps too, and the schedule tax is harshest when teams have to absorb both fatigue and travel.

## Fit Simple Rest Model

A quick logistic model gives us a clean way to describe those two effects together.

```{r model}
# Fit simple win model with rest and venue.
rest_fit <- stats::glm(
  as.integer(win) ~ restBucket + isHome,
  data = team_games,
  family = stats::binomial()
)
rest_fit_tbl <- as.data.frame(summary(rest_fit)$coefficients)
rest_fit_tbl[['term']] <- rownames(rest_fit_tbl)
rownames(rest_fit_tbl) <- NULL
rest_fit_tbl[['term']] <- c(
  'Intercept',
  'One day versus zero',
  'Two days versus zero',
  'Three-plus days versus zero',
  'Home indicator'
)
rest_fit_tbl <- rest_fit_tbl[, c(
  'term',
  'Estimate',
  'Std. Error',
  'z value',
  'Pr(>|z|)'
)]
make_table(
  rest_fit_tbl,
  caption = 'Logistic model of wins on rest and venue.',
  digits = 4
)
```

The signs all point the same way. More rest helps, and playing at home helps. The biggest rest gains come immediately when teams escape zero-rest games.

## See Which Teams Handle Zero Rest Best

League averages are useful, but back-to-backs also become a fun team question: which clubs have managed them best over a large sample?

```{r team-table}
# Rank teams by back-to-back win rate.
zero_rest_tbl <- team_games[team_games[['restDays']] == 0, c('teamId', 'win')]
zero_rest_tbl <- aggregate(
  win ~ teamId,
  data = zero_rest_tbl,
  FUN = function(x) c(winRate = mean(x), games = length(x))
)
zero_rest_tbl <- data.frame(
  teamId = zero_rest_tbl[['teamId']],
  winRate = zero_rest_tbl[['win']][, 'winRate'],
  games = zero_rest_tbl[['win']][, 'games']
)
zero_rest_tbl <- zero_rest_tbl[zero_rest_tbl[['games']] >= 50, ]
zero_rest_tbl <- merge(
  zero_rest_tbl,
  teams_tbl[, c('teamId', 'teamTriCode')],
  by = 'teamId',
  all.x = TRUE
)
zero_rest_tbl <- zero_rest_tbl[
  order(-zero_rest_tbl[['winRate']]),
  c('teamTriCode', 'games', 'winRate')
]
zero_rest_tbl <- utils::head(zero_rest_tbl, 10)
make_table(
  zero_rest_tbl,
  caption = 'Best back-to-back win rates among teams with at least 50 zero-rest games.'
)
```

This is a nice example of how a broad workflow can still end in a fan-readable leaderboard. Once the team-game table exists, you can pivot from league averages to team-specific back-to-back identities without changing tools.

## What We Learned

Back-to-backs are not just an excuse. In the salary-cap era, teams on zero rest win less often and get outscored on average. The penalty eases quickly once teams have even one off-day, and it is especially sharp for road games. That makes this a strong high-level example for `nhlscraper`: with one all-games endpoint and a little reshaping, you can turn a familiar hockey complaint into a quantified league-wide effect, then keep drilling down to venue splits and team leaderboards.

---
title: 'Did the NHL Draft''s Size Obsession Peak in the 1990s?'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Did the NHL Draft's Size Obsession Peak in the 1990s?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}
```

## Overview

NHL draft discourse has always had a body-type vocabulary. Prospects are praised for "frame", "length", and "pro build" long before anyone knows whether those gifts will actually translate into NHL impact. What changes over time is the **intensity** of that preference. This example asks a historical version of the size question: **when was the NHL draft most tilted toward bigger skaters?** Instead of looking at one ranking board, we use `nhlscraper::draft_picks()` to study actual selections across almost half a century of drafts. The goal is not to claim that teams ever drafted size *instead of* skill. The goal is to see whether certain eras leaned harder on size as a tiebreaker, especially near the top of the board.

## Build Draft Sample

We keep skaters drafted from 1979 onward, drop goalies, and require non-missing height and weight. To keep the comparisons intuitive, we split picks into first-round selections and everyone taken from Rounds 2 through 7.

```{r data}
# Pull draft picks and keep skaters with measured size.
draft_tbl <- nhlscraper::draft_picks()
draft_tbl <- draft_tbl[
  draft_tbl[['draftYear']] >= 1979 &
    draft_tbl[['positionCode']] != 'G' &
    !is.na(draft_tbl[['height']]) &
    !is.na(draft_tbl[['weight']]),
]

# Create era, round, and position buckets.
draft_tbl[['roundBucket']] <- ifelse(
  draft_tbl[['roundNumber']] == 1,
  'Round 1',
  'Rounds 2-7'
)
draft_tbl[['era']] <- cut(
  draft_tbl[['draftYear']],
  breaks = c(1978, 1989, 1999, 2009, 2019, Inf),
  labels = c(
    '1979-1989',
    '1990-1999',
    '2000-2009',
    '2010-2019',
    '2020-2025'
  )
)
draft_tbl[['positionBucket']] <- ifelse(
  draft_tbl[['positionCode']] == 'D',
  'Defense',
  'Forward'
)
draft_tbl[['tallSkater']] <- draft_tbl[['height']] >= 74
nrow(draft_tbl)
```

That gives us `r nrow(draft_tbl)` drafted skaters with usable height and weight measurements. This is a big enough pool to tell an era story rather than a class-of-one story.

## Start With Era Averages

The first pass is simple: average height and weight by era and round bucket.

```{r era-table}
# Summarize size by era and round bucket.
era_summary <- aggregate(
  cbind(height, weight) ~ era + roundBucket,
  data = draft_tbl,
  FUN = mean
)
make_table(
  era_summary,
  caption = 'Average drafted skater size by era and draft bucket.'
)
```

The shape is more nuanced than the usual "the game keeps getting bigger" story. Average first-round size surged in the 1990s, stayed elevated in the 2000s, then fell back in the 2010s. The top of the board in the modern era is still not exactly small, but it does not look like the peak size era. Just as important, first-round skaters are consistently larger than later-round skaters. The draft's size preference shows up not only across eras, but also in *where* teams are willing to spend their most expensive picks.

## Plot the First-Round Size Arc

A year-by-year line makes the rise-and-fall pattern easier to see. To reduce noise from small annual swings, we use a five-draft rolling average.

```{r rolling-data}
# Compute annual mean height by round bucket.
round1_annual <- aggregate(
  height ~ draftYear,
  data = draft_tbl[draft_tbl[['roundBucket']] == 'Round 1', ],
  FUN = mean
)
later_annual <- aggregate(
  height ~ draftYear,
  data = draft_tbl[draft_tbl[['roundBucket']] == 'Rounds 2-7', ],
  FUN = mean
)

# Smooth annual means with five-draft rolling averages.
round1_annual[['rollHeight']] <- as.numeric(stats::filter(
  round1_annual[['height']],
  rep(1 / 5, 5),
  sides = 2
))
later_annual[['rollHeight']] <- as.numeric(stats::filter(
  later_annual[['height']],
  rep(1 / 5, 5),
  sides = 2
))
```

```{r rolling-plot, fig.cap = 'Five-draft rolling average height for first-round skaters and later-round skaters.'}
graphics::plot(
  round1_annual[['draftYear']],
  round1_annual[['rollHeight']],
  type = 'l',
  lwd = 2,
  col = '#0f4c5c',
  ylim = range(
    c(round1_annual[['rollHeight']], later_annual[['rollHeight']]),
    na.rm = TRUE
  ),
  xlab = 'Draft Year',
  ylab = 'Average Height (Inches)'
)
graphics::lines(
  later_annual[['draftYear']],
  later_annual[['rollHeight']],
  lwd = 2,
  col = '#e36414'
)
graphics::legend(
  'topright',
  legend = c('Round 1', 'Rounds 2-7'),
  col = c('#0f4c5c', '#e36414'),
  lwd = 2,
  bty = 'n'
)
```

The picture is striking. The first round visibly bulks up in the 1990s, stays large through the 2000s, and then cools off. Later rounds move in the same direction, but less dramatically. That is exactly what you would expect if size became a stronger premium near the very top of the draft board during that period.

## Ask How Often Teams Chased Tall Skaters

Averages can hide roster mix, so it helps to translate height into a more intuitive marker. Here we ask what share of drafted skaters were at least 6-foot-2.

```{r tall-share}
# Summarize share of taller skaters by era.
tall_share <- aggregate(
  tallSkater ~ era + roundBucket,
  data = draft_tbl,
  FUN = mean
)
tall_counts <- aggregate(
  height ~ era + roundBucket,
  data = draft_tbl,
  FUN = length
)
names(tall_counts)[names(tall_counts) == 'height'] <- 'n'
tall_share <- merge(tall_share, tall_counts, by = c('era', 'roundBucket'))

make_table(
  tall_share,
  caption = 'Share of drafted skaters measuring at least 6-foot-2.'
)
```

The first round is where the story becomes loudest. In the 1990s, about `r sprintf('%.1f', 100 * tall_share[['tallSkater']][tall_share[['era']] == '1990-1999' & tall_share[['roundBucket']] == 'Round 1'])` percent of first-round skaters in the sample were at least 6-foot-2. In the 2010s that share dropped to about `r sprintf('%.1f', 100 * tall_share[['tallSkater']][tall_share[['era']] == '2010-2019' & tall_share[['roundBucket']] == 'Round 1'])` percent. That is a major shift in what the "ideal" first-round skater looked like.

```{r tall-share-plot, fig.cap = 'Share of drafted skaters at least 6-foot-2 by era and round bucket.'}
# Plot tall-skater shares by era.
tall_matrix <- rbind(
  tall_share[['tallSkater']][tall_share[['roundBucket']] == 'Round 1'],
  tall_share[['tallSkater']][tall_share[['roundBucket']] == 'Rounds 2-7']
)
graphics::barplot(
  tall_matrix,
  beside = TRUE,
  col = c('#1b4332', '#95d5b2'),
  ylim = c(0, 0.7),
  names.arg = levels(draft_tbl[['era']]),
  ylab = 'Share At Least 6-Foot-2',
  xlab = 'Draft Era'
)
graphics::legend(
  'topright',
  legend = c('Round 1', 'Rounds 2-7'),
  fill = c('#1b4332', '#95d5b2'),
  bty = 'n'
)
```

## Separate Position From Era

Part of any draft-size story is just that defensemen tend to run bigger than forwards. So it helps to break the sample out by position family.

```{r position-table}
# Summarize size by era and position family.
position_summary <- aggregate(
  cbind(height, weight) ~ era + positionBucket,
  data = draft_tbl,
  FUN = mean
)
make_table(
  position_summary,
  caption = 'Average drafted skater size by era and position family.'
)
```

Defensemen are consistently taller and heavier than forwards, which is exactly what most observers would expect. But the 1990s surge does not disappear once position is split out. The size bulge is still visible, especially for first-round picks, so this is not just a story about drafting more defensemen.

## Estimate First-Round Premium

As a simple check, we can fit a linear model with height as the response and draft year, first-round status, and defense status as predictors.

```{r model}
# Fit simple draft-height model.
draft_fit <- stats::lm(
  height ~ draftYear + I(roundNumber == 1) + I(positionCode == 'D'),
  data = draft_tbl
)
draft_fit_tbl <- as.data.frame(summary(draft_fit)$coefficients)
draft_fit_tbl[['term']] <- rownames(draft_fit_tbl)
rownames(draft_fit_tbl) <- NULL
draft_fit_tbl[['term']] <- c(
  'Intercept',
  'Draft year',
  'First-round indicator',
  'Defense indicator'
)
draft_fit_tbl <- draft_fit_tbl[, c(
  'term',
  'Estimate',
  'Std. Error',
  't value',
  'Pr(>|t|)'
)]
make_table(
  draft_fit_tbl,
  caption = 'Linear model of drafted skater height.',
  digits = 4
)
```

This model says the first-round premium is still real even after accounting for time and position. First-round skaters come in roughly `r sprintf('%.2f', draft_fit_tbl[['Estimate']][draft_fit_tbl[['term']] == 'First-round indicator'])` inches taller on average, while defensemen add about `r sprintf('%.2f', draft_fit_tbl[['Estimate']][draft_fit_tbl[['term']] == 'Defense indicator'])` inches on top of that. The draft-year slope itself is small because the historical pattern is not a straight line. That is the core takeaway of the whole article: draft size preference looks less like a steady march and more like a boom that peaked in the 1990s and early 2000s.

## What We Learned

The draft has never been indifferent to size, especially in the first round. Bigger skaters have consistently been favored near the top of the board, and defensemen have carried their own structural size premium the whole way through. But the strongest form of that bias does not appear to be a modern invention. In this sample, the most size-forward era is the 1990s, with the 2000s not far behind. That makes `nhlscraper::draft_picks()` a useful reminder that draft archetypes move in cycles: the league does not just change, it overcorrects and then changes back.

---
title: 'What Separated Florida and Edmonton in Game 7?'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{What Separated Florida and Edmonton in Game 7?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}
```

## Overview

Florida beat Edmonton 2-1 in Game 7 of the 2024 Stanley Cup Final, and the result instantly entered hockey memory as one of those games that *felt* like it might hinge on a single bounce. That is exactly why it is such a good guided example. A one-goal final score leaves room for multiple stories. Maybe Florida truly carried play and deserved the win. Maybe Edmonton created just as much and lost on finishing. Maybe the game was close on everything, including chance quality. The only way to separate those stories is to walk from the scoreboard down into the event log. This example uses `nhlscraper::gc_summary()`, `nhlscraper::gc_pbp()`, `nhlscraper::game_rosters()`, and `nhlscraper::calculate_expected_goals()` to reconstruct the game at the chance level. We will move from team totals, to period context, to the biggest individual looks, and finally to two visuals that show how the game breathed.

## Pull Game Summary and Play-By-Play

We work from the final itself, Game 7 of the 2024 Final (`2023030417`).

```{r data}
# Pull summary, play-by-play, and roster context.
game_summary <- nhlscraper::gc_summary(2023030417)
pbp_xg <- nhlscraper::calculate_expected_goals(
  nhlscraper::gc_pbp(2023030417)
)
if (!('xG' %in% names(pbp_xg))) {
  stop(
    paste(
      "calculate_expected_goals() did not return an xG column during article build.",
      "This usually means the loaded nhlscraper namespace is stale or the scorer hit",
      "a runtime error and returned the input unchanged."
    )
  )
}
rosters <- nhlscraper::game_rosters(2023030417)

# Build team and player labels.
home_id <- game_summary[['homeTeam']][['id']]
away_id <- game_summary[['awayTeam']][['id']]
home_abbrev <- game_summary[['homeTeam']][['abbrev']]
away_abbrev <- game_summary[['awayTeam']][['abbrev']]

rosters[['playerFullName']] <- paste(
  rosters[['playerFirstName']],
  rosters[['playerLastName']]
)
rosters[['teamTriCode']] <- ifelse(
  rosters[['teamId']] == home_id,
  home_abbrev,
  away_abbrev
)

# Keep shot events with positive xG.
shots <- pbp_xg[!is.na(pbp_xg[['xG']]) & pbp_xg[['xG']] > 0, , drop = FALSE]
roster_match <- match(shots[['shootingPlayerId']], rosters[['playerId']])
shots[['playerFullName']] <- rosters[['playerFullName']][roster_match]
shots[['teamTriCode']] <- rosters[['teamTriCode']][roster_match]
shots[['timeInPeriod']] <- sprintf(
  '%02d:%02d',
  shots[['secondsElapsedInPeriod']] %/% 60,
  shots[['secondsElapsedInPeriod']] %% 60
)
```

This is already a powerful workflow. With one summary call, one play-by-play call, one roster call, and one modeling helper, the game opens up from scoreline to process.

## Start With Scoreboard Context

Before getting fancy, it helps to put the official result and the xG estimate side by side.

```{r team-table}
# Summarize team-level scoreboard and xG results.
team_table <- data.frame(
  team = c(home_abbrev, away_abbrev),
  goals = c(
    game_summary[['homeTeam']][['score']],
    game_summary[['awayTeam']][['score']]
  ),
  shotsOnGoal = c(
    game_summary[['homeTeam']][['sog']],
    game_summary[['awayTeam']][['sog']]
  ),
  xG = c(
    sum(shots[['xG']][shots[['eventOwnerTeamId']] == home_id], na.rm = TRUE),
    sum(shots[['xG']][shots[['eventOwnerTeamId']] == away_id], na.rm = TRUE)
  )
)
team_table[['xGPerShot']] <-
  team_table[['xG']] / team_table[['shotsOnGoal']]
make_table(
  team_table,
  caption = 'Game 7 team context: score, shots on goal, and expected goals.'
)
```

The result and the process point in the same direction, but only narrowly. Florida edges Edmonton in goals, edges Edmonton in raw xG, and edges Edmonton in xG per shot. Nothing in this table looks like domination. It looks like a deserved win in a game where the margin stayed small all night.

## Break the Night Into Periods

Game 7s rarely play with a single rhythm from start to finish. One team can own the first period, another can chase late, and the game can still end within one chance. Period-level xG helps show where the pressure actually accumulated.

```{r period-table}
# Summarize xG by period and team.
period_data <- data.frame(
  periodNumber = shots[['periodNumber']],
  eventOwnerTeamId = shots[['eventOwnerTeamId']],
  xG = shots[['xG']]
)
period_summary <- aggregate(
  xG ~ periodNumber + eventOwnerTeamId,
  data = period_data,
  FUN = sum
)
period_ids <- sort(unique(shots[['periodNumber']]))
period_table <- data.frame(period = period_ids)
home_match <- match(
  period_ids,
  period_summary[['periodNumber']][period_summary[['eventOwnerTeamId']] == home_id]
)
away_match <- match(
  period_ids,
  period_summary[['periodNumber']][period_summary[['eventOwnerTeamId']] == away_id]
)
period_table[[paste0(home_abbrev, '_xG')]] <-
  period_summary[['xG']][period_summary[['eventOwnerTeamId']] == home_id][home_match]
period_table[[paste0(away_abbrev, '_xG')]] <-
  period_summary[['xG']][period_summary[['eventOwnerTeamId']] == away_id][away_match]
period_table[is.na(period_table)] <- 0
make_table(
  period_table,
  caption = 'Expected goals by period in Game 7.'
)
```

Florida's biggest push arrives in the third period. That matters because it matches the feel of the game: Edmonton kept hanging around, but Florida generated the most forceful late stretch and did enough to keep the Oilers from turning their chase into an equalizer.

## Surface the Biggest Chances

Totals are useful, but they can still feel abstract. One of the nicest things about working directly from the event log is that you can ask which specific looks drove the game.

```{r chance-table}
# Show largest individual shot-quality events.
chance_idx <- order(-shots[['xG']])
chance_table <- data.frame(
  player = shots[['playerFullName']][chance_idx],
  team = shots[['teamTriCode']][chance_idx],
  period = shots[['periodNumber']][chance_idx],
  timeInPeriod = shots[['timeInPeriod']][chance_idx],
  xCoordNorm = shots[['xCoordNorm']][chance_idx],
  yCoordNorm = shots[['yCoordNorm']][chance_idx],
  xG = shots[['xG']][chance_idx]
)
chance_table <- utils::head(chance_table, 10)
make_table(
  chance_table,
  caption = 'Highest-xG individual chances in Game 7.',
  digits = 3
)
```

That table is a nice reminder that a one-game study can still be concrete. Instead of saying "Florida had a slight edge," you can point to the exact attempts that shaped the edge and where they came from on the ice.

## Plot the Cumulative xG Race

The cumulative view shows whether the game felt close because both teams traded dangerous looks, or because one team held a long territorial edge that never quite broke the score open.

```{r cum-data}
# Build cumulative xG paths for both teams.
build_cum_path <- function(team_id) {
  team_idx <- shots[['eventOwnerTeamId']] == team_id
  team_shots <- data.frame(
    eventId = shots[['eventId']][team_idx],
    secondsElapsedInGame = shots[['secondsElapsedInGame']][team_idx],
    xG = shots[['xG']][team_idx]
  )
  team_shots <- team_shots[order(
    team_shots[['secondsElapsedInGame']],
    team_shots[['eventId']]
  ), ]
  data.frame(
    minutes = c(0, team_shots[['secondsElapsedInGame']] / 60),
    cumXG = c(0, cumsum(team_shots[['xG']]))
  )
}
home_path <- build_cum_path(home_id)
away_path <- build_cum_path(away_id)
```

```{r cum-plot, fig.cap = 'Cumulative expected goals in Game 7 of the 2024 Stanley Cup Final.'}
graphics::plot(
  home_path[['minutes']],
  home_path[['cumXG']],
  type = 's',
  lwd = 2,
  col = '#c1121f',
  xlim = c(0, 60),
  ylim = c(0, max(c(home_path[['cumXG']], away_path[['cumXG']])) * 1.05),
  xlab = 'Minutes Elapsed',
  ylab = 'Cumulative Expected Goals'
)
graphics::lines(
  away_path[['minutes']],
  away_path[['cumXG']],
  type = 's',
  lwd = 2,
  col = '#003049'
)
graphics::abline(v = c(20, 40), lty = 3, col = '#8d99ae')
graphics::legend(
  'topleft',
  legend = c(home_abbrev, away_abbrev),
  col = c('#c1121f', '#003049'),
  lwd = 2,
  bty = 'n'
)
```

The lines stay close for almost the entire game. Florida pulls slightly ahead, then widens that edge late, but the overall shape is still a race rather than a runaway. That makes this final a clean example of a winner who both *earned* the result and still lived inside a narrow margin.

## Plot Shot Geography

The rink view adds one more layer. A chance model tells us *how dangerous* the looks were. A shot map shows *where* those looks came from.

```{r rink-plot, fig.cap = 'Shot-quality map for Game 7. Point size scales with expected goals.'}
# Split shot map inputs by team.
home_shots <- shots[shots[['eventOwnerTeamId']] == home_id, ]
away_shots <- shots[shots[['eventOwnerTeamId']] == away_id, ]
nhlscraper::draw_NHL_rink()
graphics::points(
  home_shots[['xCoordNorm']],
  home_shots[['yCoordNorm']],
  pch = 19,
  col = grDevices::rgb(0.76, 0.07, 0.12, 0.55),
  cex = 0.6 + 7 * sqrt(home_shots[['xG']])
)
graphics::points(
  away_shots[['xCoordNorm']],
  away_shots[['yCoordNorm']],
  pch = 19,
  col = grDevices::rgb(0.00, 0.19, 0.29, 0.55),
  cex = 0.6 + 7 * sqrt(away_shots[['xG']])
)
graphics::legend(
  'topright',
  legend = c(home_abbrev, away_abbrev),
  pch = 19,
  col = c(
    grDevices::rgb(0.76, 0.07, 0.12, 0.75),
    grDevices::rgb(0.00, 0.19, 0.29, 0.75)
  ),
  bty = 'n'
)
```

The map reinforces the same read as the cumulative chart. Florida did not swamp Edmonton with endless pressure, but the Panthers owned just a little more of the dangerous interior. In a 2-1 Game 7, that is often the whole story.

## What We Learned

Florida's win was narrow, but it was not random. The Panthers generated slightly more xG, carried the stronger third-period push, and owned a small edge in dangerous interior looks. Edmonton stayed within range all the way through, which is why the game felt so tense, but the underlying shot-quality story still leans Florida. This is also a good showcase of the package itself. `nhlscraper` lets you start with a game everybody remembers, scrape the event log, attach an xG model, and turn a familiar score into a richer explanation of *why* the score looked the way it did.

---
title: "How gc_play_by_play() and wsc_play_by_play() Work"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How gc_play_by_play() and wsc_play_by_play() Work}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

`gc_play_by_play()` and `wsc_play_by_play()` return the same cleaned public play-by-play schema, but they do not start from the same raw feed. `gc_play_by_play()` starts from the GameCenter play-by-play feed. `wsc_play_by_play()` starts from the World Showcase feed and uses GameCenter metadata to keep the output aligned with the same roster, team, and HTML report context. The pipeline is intentionally source-aware:

- The API play-by-play is the source of truth for event order, event identity, and raw `situationCode`.
- The HTML play-by-play report is the primary source for on-ice player identities.
- `shift_chart()` is not used to populate on-ice player IDs inside these functions. It is a separate downstream source for shift timing via `add_shift_times()`.

This article walks through the process step by step.

## Step 1: Fetch the raw sources.

For a single game, the functions fetch the raw API play-by-play plus the HTML play-by-play report. `wsc_play_by_play()` also fetches the WSC play-by-play feed itself. Those requests are now issued in parallel, because network latency dominates total runtime much more than the in-memory cleaning steps do.

```{r fetch-example, eval = FALSE}
gc  <- nhlscraper::gc_play_by_play(2023030417)
wsc <- nhlscraper::wsc_play_by_play(2023030417)
```

The HTML report is fetched for every game because it is the only source that consistently exposes the full on-ice player sets.

## Step 2: Standardize the raw play-by-play feed.

Once the raw feed is downloaded, the package standardizes the columns before doing any enrichment. That includes:

- moving the `gameId` into the table
- flattening nested API fields into a consistent tabular shape
- renaming public-facing columns such as `periodNumber`, `eventTypeCode`, and `eventTypeDescKey`
- filling obvious source omissions such as missing `shootingPlayerId` values on goal rows when the scorer is known

The aim here is to make the downstream logic operate on one internal structure, even when the upstream feed formats differ.

## Step 3: Repair obviously impossible event ordering.

The API play-by-play remains authoritative, but not every upstream `sortOrder` is logically consistent. Before any HTML matching happens, the pipeline repairs clear boundary mistakes. The guiding principle is conservative: only fix sequences that are plainly impossible from the game clock and event context.

Examples:

- no event can happen between `period-start` and the opening `faceoff`
- illogically ordered boundary faceoffs are dropped
- blocked shots keep the API event identity, but their directional perspective is normalized to the shooting team so they behave like the other shot events

This matters because HTML matching becomes much more reliable once the API timeline itself is internally coherent.

## Step 4: Derive the game-state columns from `situationCode`.

The raw `situationCode` is parsed into the public state columns:

- `homeIsEmptyNet`
- `awayIsEmptyNet`
- `homeSkaterCount`
- `awaySkaterCount`
- `manDifferential`
- `strengthState`

These are the public summary columns that describe the intended rules context of the play. The raw `situationCode` itself is kept in the output as-is. The HTML report can later supply a different-looking player identity set, but it does not rewrite these `situationCode`-derived state columns.

## Step 5: Add coordinate and shot-context enrichment.

After the structural cleanup, the play-by-play gets the geometric and shot-context features used elsewhere in the package. That includes:

- normalized rink coordinates
- shot distance
- shot angle
- rush flags
- rebound flags
- rebound-creation flags
- cumulative goal, shot, Fenwick, and Corsi counts from the event stream

These features are derived from the cleaned API event log, not from the HTML report.

## Step 6: Parse the HTML play-by-play report.

The HTML report is parsed into a second event table that contains:

- the event type
- the period and clock
- team ownership context
- key player signatures used for matching
- the home and away on-ice player sets

The parser also handles the known HTML-side quirks, including the fact that the HTML report records blocked shots from the defending perspective while the package standardizes blocked shots from the shooting perspective.

## Step 7: Match HTML rows back to API rows.

The package does not join HTML to API rows only on time, and it does not require the HTML on-ice counts to agree with `situationCode`. It builds a richer event signature on both sides and uses that to align the reports. The matching logic uses combinations of:

- period
- elapsed seconds
- event type
- event-owning team
- primary event actors
- supporting player signatures

This step is where the earlier ordering repair pays off. The cleaner the API event sequence is, the safer the HTML match becomes.

## Step 8: Keep state context separate from player identity.

A matched HTML row is used for the on-ice player ID columns. It is not used to recalculate the `situationCode`-derived context columns. This keeps two related but different concepts separate:

- `situationCode`, `homeSkaterCount`, `awaySkaterCount`, empty-net flags, `manDifferential`, and `strengthState` describe the intended rules state.
- `homeGoaliePlayerId`, `awayGoaliePlayerId`, and the skater ID columns describe the players listed in the HTML report for that event.

The HTML report can list more or fewer players than the intended state implies. That can happen around line changes, unusual bench situations, or report noise. The package preserves the listed identities without letting those counts change the intended state context.

## Step 9: Keep one-on-one rows constrained.

Penalty-shot and shootout rows with `0101` or `1010` are different. Those states are genuinely one-on-one plays. Even if the HTML report lists extra players, the package records only the shooter and the defending goalie for those rows.

## Step 10: Populate on-ice player IDs.

Once a matched HTML row is accepted, the package writes the scalar on-ice player ID columns into the play-by-play row. That includes:

- `homeGoaliePlayerId` and `awayGoaliePlayerId`
- `homeSkater1PlayerId` through `homeSkater5PlayerId` by default, with extra
  skater slots added only when the game needs them
- `awaySkater1PlayerId` through `awaySkater5PlayerId` by default, with extra
  skater slots added only when the game needs them
- the corresponding `...For` and `...Against` columns

The base schema tracks the standard five skaters. If the HTML report shows an
extra attacker or any other overflow row, the package expands dynamically to
`skater6`, `skater7`, `skater8`, and so on instead of truncating the row.

## Step 11: Handle one-on-one and delayed-penalty edge cases.

Two edge-case families need their own rules.

### Shootouts and penalty shots

Rows with one-on-one states such as `0101` and `1010` populate only the shooter and the defending goalie. Extra HTML-listed players on those rows are treated as report noise because the play is supposed to be one shooter against one goalie.

### Unmatched delayed-penalty rows

Some supported delayed-penalty rows do not appear in the HTML report at all. In those cases, the package can backfill the on-ice player IDs from the nearest prior populated row in the same period when:

- the state signature is unchanged
- the time gap is very small
- the prior row already has a compatible populated on-ice set

This fixes cases where the HTML report skips the delayed-penalty marker but clearly preserves the same live-play skaters immediately before the whistle.

## Step 12: Finalize the public schema.

The last step is to expose the cleaned public-facing schema and hide the internal staging details. Both `gc_play_by_play()` and `wsc_play_by_play()` return one row per event with:

- the same core event columns
- the same strength and on-ice player ID columns
- the same cumulative game-state columns

The only intentional difference is source-specific metadata such as `utc` in the WSC output and GameCenter clip fields in the GC output.

## How `shift_chart()` Fits In

`shift_chart()` is related, but it solves a different problem. It provides shift windows, not event identities. In practical use:

```{r shift-example, eval = FALSE}
pbp    <- nhlscraper::gc_play_by_play(2023030417)
shifts <- nhlscraper::shift_chart(2023030417)
pbp_with_shift_times <- nhlscraper::add_shift_times(pbp, shifts)
```

This is why the package keeps the HTML play-by-play report as the primary on-ice identity source inside `gc_play_by_play()` and `wsc_play_by_play()`, while `shift_chart()` remains the right tool for shift-timing context after the play-by-play is already built.

## Practical Summary

If you want the shortest mental model, it is this:

- start from the API play-by-play
- repair only the event-order and strength mistakes that are logically supportable
- use `situationCode` for intended manpower context
- use the HTML report to recover listed on-ice player identities
- keep one-on-one rows to shooter plus goalie
- use `shift_chart()` later when you need shift timing rather than event-level player identity

That balance is what lets the final play-by-play stay both practical and auditable.

---
title: 'Do Bigger Skaters Hold Their Scoring in the Playoffs?'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Do Bigger Skaters Hold Their Scoring in the Playoffs?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = '#>',
  fig.align = 'center',
  out.width = '90%',
  fig.width = 7,
  fig.height = 4.5
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}
```

## Overview

Playoff hockey has a stubborn stereotype: once the checking tightens, skill supposedly gives way to mass. Bigger skaters are assumed to survive the grind better, win more net-front battles, and keep producing when series turn meaner. That story sounds plausible, but it bundles two different ideas together. The first is about **absolute playoff scoring**: do heavier skaters actually score more? The second is about **playoff translation**: even if bigger players do not outscore smaller ones outright, do they lose *less* offense when the regular season gives way to the postseason? This example tackles both questions at once. We combine `nhlscraper::skater_playoff_statistics()`, `nhlscraper::skater_statistics()`, and `nhlscraper::players()` to build a salary-cap-era sample of skaters with meaningful regular-season and playoff workloads. From there, we compare regular-season points per game, playoff points per game, and the gap between them.

## Build Analysis Table

We want one row per player with career regular-season scoring, career playoff scoring, and basic biometrics.

```{r data}
# Pull playoff, regular-season, and biometric records.
playoff_stats <- nhlscraper::skater_playoff_statistics()
career_stats <- nhlscraper::skater_statistics()[, c(
  'playerId',
  'rsGamesPlayed',
  'rsPoints',
  'positionCode'
)]
player_bios <- nhlscraper::players()[, c(
  'playerId',
  'playerFullName',
  'height',
  'weight'
)]

# Join player-level tables.
analysis_tbl <- merge(
  playoff_stats,
  career_stats,
  by = c('playerId', 'positionCode'),
  all.x = TRUE
)
analysis_tbl <- merge(
  analysis_tbl,
  player_bios,
  by = 'playerId',
  all.x = TRUE
)

# Keep salary-cap skaters with meaningful samples.
analysis_tbl <- analysis_tbl[
  !is.na(analysis_tbl[['height']]) &
    !is.na(analysis_tbl[['weight']]) &
    analysis_tbl[['firstSeasonForGameType']] >= 20052006 &
    analysis_tbl[['gamesPlayed']] >= 20 &
    analysis_tbl[['rsGamesPlayed']] >= 200,
]

# Fill missing names and compute scoring rates.
analysis_tbl[['playerFullName']] <- ifelse(
  is.na(analysis_tbl[['playerFullName']]) |
    analysis_tbl[['playerFullName']] == '',
  paste(
    analysis_tbl[['skaterFirstName']],
    analysis_tbl[['skaterLastName']]
  ),
  analysis_tbl[['playerFullName']]
)
analysis_tbl[['regularPPG']] <-
  analysis_tbl[['rsPoints']] / analysis_tbl[['rsGamesPlayed']]
analysis_tbl[['playoffPPG']] <-
  analysis_tbl[['points']] / analysis_tbl[['gamesPlayed']]
analysis_tbl[['playoffLift']] <-
  analysis_tbl[['playoffPPG']] - analysis_tbl[['regularPPG']]
analysis_tbl[['positionBucket']] <- ifelse(
  analysis_tbl[['positionCode']] == 'D',
  'Defense',
  'Forward'
)

# Assign equal-count weight quartiles.
weight_share <- rank(
  analysis_tbl[['weight']],
  ties.method = 'first'
) / nrow(analysis_tbl)
analysis_tbl[['weightQuartile']] <- cut(
  weight_share,
  breaks = c(0, 0.25, 0.50, 0.75, 1),
  include.lowest = TRUE,
  labels = c('Lightest', 'Second', 'Third', 'Heaviest')
)
nrow(analysis_tbl)
```

That leaves `r nrow(analysis_tbl)` modern skaters. The sample is large enough to smooth away one-hot playoff runs, but still focused enough to keep the comparison about contemporary hockey. To make the shape of the data concrete, it helps to peek at a few of the strongest playoff scorers in the sample.

```{r sample-table}
# Show sample of strongest playoff scorers.
sample_tbl <- analysis_tbl[
  order(-analysis_tbl[['playoffPPG']], -analysis_tbl[['gamesPlayed']]),
  c(
    'playerFullName',
    'positionCode',
    'weight',
    'rsGamesPlayed',
    'gamesPlayed',
    'regularPPG',
    'playoffPPG',
    'playoffLift'
  )
]
sample_tbl <- utils::head(sample_tbl, 8)
make_table(
  sample_tbl,
  caption = 'Top playoff scoring rates among skaters in the working sample.'
)
```

## Compare Weight Quartiles

Now we can step back from individual names and ask the population question. If heavier skaters really become more dangerous in the playoffs, the heaviest quartile should sit clearly above the lighter groups in playoff scoring rate, or at least show a meaningfully smaller drop from regular-season scoring.

```{r quartiles}
# Summarize scoring levels and playoff lift by quartile.
quartile_summary <- aggregate(
  cbind(regularPPG, playoffPPG, playoffLift) ~ weightQuartile,
  data = analysis_tbl,
  FUN = mean
)
quartile_counts <- as.data.frame(table(analysis_tbl[['weightQuartile']]))
names(quartile_counts) <- c('weightQuartile', 'n')
quartile_summary <- merge(
  quartile_summary,
  quartile_counts,
  by = 'weightQuartile'
)
quartile_summary <- quartile_summary[
  match(levels(analysis_tbl[['weightQuartile']]), quartile_summary[['weightQuartile']]),
  c('weightQuartile', 'n', 'regularPPG', 'playoffPPG', 'playoffLift')
]
make_table(
  quartile_summary,
  caption = 'Regular-season scoring, playoff scoring, and playoff lift by weight quartile.'
)
```

This is where the cliché starts to wobble. The lightest quartile posts the strongest playoff scoring rate in the sample at about `r sprintf('%.3f', quartile_summary[['playoffPPG']][quartile_summary[['weightQuartile']] == 'Lightest'])` points per game, while the heaviest quartile sits at about `r sprintf('%.3f', quartile_summary[['playoffPPG']][quartile_summary[['weightQuartile']] == 'Heaviest'])`. But the second question is more subtle. The heaviest skaters do **not** lead in raw playoff scoring, yet they also show the smallest drop from their own regular-season baseline. Their average playoff lift is about `r sprintf('%.3f', quartile_summary[['playoffLift']][quartile_summary[['weightQuartile']] == 'Heaviest'])`, compared with `r sprintf('%.3f', quartile_summary[['playoffLift']][quartile_summary[['weightQuartile']] == 'Lightest'])` for the lightest group. That is a more interesting result than a simple yes-or-no size verdict. Bigger skaters are not the most productive playoff scorers on average, but they may lose slightly less offense when conditions tighten.

## Visualize Level Versus Translation

A table gives the means. A plot shows whether those means reflect broad tendencies or just a few stars.

```{r quartile-plot, fig.cap = 'Playoff scoring level and playoff lift by weight quartile.'}
# Plot playoff scoring distribution and average lift.
old_par <- graphics::par(no.readonly = TRUE)
graphics::par(mfrow = c(1, 2), mar = c(8, 4, 3, 1))
graphics::boxplot(
  playoffPPG ~ weightQuartile,
  data = analysis_tbl,
  col = c('#d9ed92', '#b5e48c', '#76c893', '#34a0a4'),
  border = '#3a5a40',
  las = 2,
  ylab = 'Playoff Points Per Game',
  xlab = ''
)
graphics::barplot(
  quartile_summary[['playoffLift']],
  names.arg = quartile_summary[['weightQuartile']],
  col = c('#f4d35e', '#ee964b', '#f95738', '#7b2cbf'),
  border = NA,
  las = 2,
  ylab = 'Playoff Lift',
  xlab = ''
)
graphics::abline(h = 0, lty = 2, col = '#4d4d4d')
graphics::par(old_par)
```

The left panel reinforces the first point: heavier groups do not shift upward as playoff scorers. The right panel reinforces the second: every group scores less in the playoffs than in the regular season on average, but the drop is a little shallower for the heaviest skaters. That combination makes intuitive hockey sense. Many larger players are not offense-first stars, so their absolute scoring rates start lower. But their game may translate a bit more cleanly into playoff conditions than the regular-season scoring records alone suggest.

## See Who Gains Most

One way to make the translation question feel real is to look at the players with the biggest positive playoff bump.

```{r risers}
# Show largest positive playoff lifts among larger-sample skaters.
risers_tbl <- analysis_tbl[
  analysis_tbl[['gamesPlayed']] >= 40,
  c(
    'playerFullName',
    'positionBucket',
    'weight',
    'regularPPG',
    'playoffPPG',
    'playoffLift'
  )
]
risers_tbl <- risers_tbl[order(-risers_tbl[['playoffLift']]), ]
risers_tbl <- utils::head(risers_tbl, 10)
make_table(
  risers_tbl,
  caption = 'Largest playoff scoring lifts among skaters with at least 40 playoff games.'
)
```

This table also hints at an important caveat: some of the best playoff translators are defensemen. That matters because defense is a position where the scoring baseline is lower to begin with. A small absolute scoring gain there can look like a meaningful playoff bump.

## Fit Simple Model

To separate size from position a little more directly, we can fit a simple model with playoff lift as the response and height, weight, and a defense indicator as predictors.

```{r model}
# Fit simple playoff-lift model.
lift_fit <- stats::lm(
  playoffLift ~ height + weight + I(positionCode == 'D'),
  data = analysis_tbl
)
lift_fit_tbl <- as.data.frame(summary(lift_fit)$coefficients)
lift_fit_tbl[['term']] <- rownames(lift_fit_tbl)
rownames(lift_fit_tbl) <- NULL
lift_fit_tbl[['term']] <- c(
  'Intercept',
  'Height',
  'Weight',
  'Defense indicator'
)
lift_fit_tbl <- lift_fit_tbl[, c(
  'term',
  'Estimate',
  'Std. Error',
  't value',
  'Pr(>|t|)'
)]
make_table(
  lift_fit_tbl,
  caption = 'Linear model of playoff scoring lift on height, weight, and position.',
  digits = 4
)
```

In this specification, neither height nor weight carries a strong standalone slope once position is accounted for. The cleanest signal is positional: defensemen, on average, lose less scoring in the playoffs than forwards do. That does not mean size is irrelevant. It means the popular claim that bigger skaters *as such* become better playoff scorers is too blunt. Role and position explain more than a simple bigger-is-better story.

## What We Learned

The modern playoff-size cliché is only half right. Bigger skaters do **not** post the best playoff scoring rates on average. The lightest quartile still scores more. But bigger skaters also do **not** collapse when the postseason starts; if anything, their scoring translates a bit more cleanly relative to regular-season expectations. That is a useful lesson for exploratory work with `nhlscraper`: the most interesting answer is often a split answer. A single workflow can move from headline myth, to player-level table, to population comparison, to a more careful interpretation of what the data really say.

---
title: "How nhlscraper's Expected Goals Model Works"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How nhlscraper's Expected Goals Model Works}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

make_table <- function(x, caption, digits = 3) {
  knitr::kable(x, caption = caption, digits = digits)
}
```

## Overview

Expected goals, or xG, is an attempt to answer a simple question more carefully than the box score can: *how likely was this shot to become a goal?* A long point wrister through traffic, a rebound from the top of the crease, a backdoor one-timer, and an empty-net clear all count as shot attempts, but they are not equally dangerous. xG tries to put those attempts on the same probability scale. That broad idea is familiar. The harder part is building a model that is useful inside a package. `nhlscraper` has to do more than fit well in a notebook. It has to run on public play-by-play columns, stay light on runtime dependencies, and score rows quickly enough to be practical inside analysis and plotting helpers. That is why the current package model is not a heavy gradient-boosting system. It is a partitioned ridge logistic regression rebuild that can be scored with base-R math once the preprocessing rules and coefficients are frozen. This article explains the model in the order that matters most for package users: what it is trying to estimate, how the shot space is partitioned, what data it was trained on, what information it uses, how the ridge architecture works at runtime, and what the current evaluation results look like.

## One Model, Six Situations

The first thing to understand is that `nhlscraper` no longer treats xG as a menu of version numbers. There is one built-in xG system, but that system is really six separate ridge models applied to six mutually exclusive game states. Those partitions are:

```{r partition-table}
partition_table <- data.frame(
  partition = c("sd", "ev", "pp", "sh", "en", "ps"),
  meaning = c(
    "Regulation 5v5 without empty nets",
    "Other even-strength states outside standard 5v5",
    "Shooting team has a skater advantage",
    "Shooting team is short-handed",
    "Opponent net is empty",
    "Penalty-shot and shootout-style situations"
  ),
  stringsAsFactors = FALSE
)
make_table(
  partition_table,
  caption = "The six shot partitions used by nhlscraper's xG model."
)
```

That split is not cosmetic. It reflects the fact that a 5v5 wrist shot, a 4v4 rush chance, a power-play seam pass, and an empty-net try do not live in the same statistical environment. The package therefore partitions the shot first and only then applies the relevant ridge model. In package terms, the decision rules are explicit:

1. Penalty-shot and shootout-style states (`1010` and `0101`) go to `ps`.
2. Empty-net-against shots go to `en`.
3. Standard 5v5 non-empty-net shots go to `sd`.
4. Remaining even-strength shots go to `ev`.
5. Skater-advantage shots go to `pp`.
6. Skater-disadvantage shots go to `sh`.

That matters analytically too. When someone says "the xG model," what the package is actually doing is choosing among six different coefficient sets that were trained on six different shot environments.

## Training Data

The ridge rebuild was trained on the current public `nhlscraper` play-by-play schema rather than on a private one-off table. That decision keeps the runtime implementation honest, because the package scorer has to reproduce the same feature engineering from columns that package users can actually obtain. The training window covers the `2023-24` and `2024-25` seasons. The preparation pipeline starts from full play-by-play data, then adds the context needed for shot-quality modeling:

```r
pbp <- nhlscraper::gc_pbps(season) |>
  nhlscraper::add_shift_times(nhlscraper::shift_charts(season)) |>
  nhlscraper::add_deltas() |>
  nhlscraper::add_shooter_biometrics() |>
  nhlscraper::add_goalie_biometrics()
```

That pipeline matters because the model is not just a location model. It depends on event-to-event movement, score and attempt context, previous-event information, shift burden, and player biometrics. The package scorer therefore mirrors the same preparation steps before it scores a row. The training volumes are also uneven across partitions, which is exactly what you would expect from NHL data. Standard 5v5 dominates the sample, while empty-net and shootout situations are much smaller.

```{r train-table}
train_summary <- data.frame(
  partition = c("sd", "ev", "pp", "sh", "en", "ps"),
  games = c(2798, 1280, 2793, 2241, 1245, 230),
  rows = c(188930, 4907, 38903, 5539, 1828, 1188),
  goal_rate = c(0.0593, 0.1113, 0.0973, 0.0738, 0.5739, 0.3157)
)
make_table(
  train_summary,
  caption = "Training sample size and goal rate by partition.",
  digits = 4
)
```

That table explains why the package should not promise identical stability across every state. The `sd` model gets to learn from a very large 5v5 sample. The `ps` model does not.

## What the Model Uses

The package model is rich, but the inputs fall into a few intuitive families.

### Shot Geometry

Every partition starts with the spatial basics: normalized x and y coordinates, shot distance, and shot angle. Those remain the backbone of the model because location still carries a large share of shot-quality signal.

### Event-to-Event Movement

`nhlscraper` also tracks how the puck and shot location moved relative to the prior event. That includes raw and per-second deltas in normalized x, normalized y, distance, angle, and sequence time. These movement features help separate a static outside shot from a chance that developed through rapid lateral or downhill movement.

### Game Context

The ridge models also see state variables such as period, overtime, score differential, shots/Fenwick/Corsi context, skater counts, and strength state. Those features help the model understand whether a shot happened in a settled 5v5 environment, a special-teams sequence, a tied game late, or a tilted score state after a long run of pressure.

### Chance Descriptors

Some features are deliberately interpretable hockey flags rather than generic numerics:

- `isBehindNet`
- `crossedRoyalRoad`
- `isRebound`
- `isRush`
- previous-event context through `typeDescKeyPrev`

Those features capture patterns that hockey analysts already describe in words, but the model still estimates their value from data rather than imposing it by hand.

### Player and Shift Context

The package model also includes shooter and goalie biometrics plus shift-timing features. That means the scorer can distinguish not only *where* a shot came from, but also something about *who* took it, *who* faced it, and how taxed the skaters were when it happened.

This is the main reason the runtime scorer now tries to add shift-time context before scoring when those columns are missing. The ridge model was trained with that information, so the package should use it when it can.

## Why Ridge Logistic Regression

The architectural choice is straightforward: ridge logistic regression is the compromise that best fits package reality. It offers three practical advantages:

1. The model is still expressive once the feature engineering is rich.
2. The fitted scorer can be frozen into coefficients plus preprocessing constants.
3. The runtime package code does not need `glmnet`, `tidymodels`, or any other modeling dependency just to score a play-by-play.

The price is that preprocessing matters. The package cannot stop at "here are the coefficients." It also has to preserve the training-time dummy maps, median imputations, normalization constants, and zero-variance removals. Those frozen artifacts are trained upstream in `rentosrink/models/xG/nhlscraper/` and then copied into the package; `nhlscraper` itself is only packaging and scoring them at runtime, not retraining them locally. That frozen preprocessing contract is exactly what the current package implementation now carries internally. In other words, the runtime path is:

1. Engineer the same public-schema features used at training time.
2. Partition the shot into one of six states.
3. Apply the partition-specific preprocessing rules.
4. Compute the linear predictor with the frozen ridge coefficients.
5. Convert that score to a probability with the logistic link.

## How It Was Trained

Training used grouped cross-validation by `gameId` across the full `2023-24` and `2024-25` pool. That grouping matters because hockey shots from the same game are not independent in the way ordinary row-wise cross-validation would pretend they are. Grouped folds make the tuning step more realistic by holding out whole games together. After choosing the ridge penalty from grouped cross-validation, each partition was refit on all available rows from the training window. That means the cross-validation results are tuning diagnostics, not unseen-future proof. The future-facing claim should come from the external tests, not from the grouped CV table. For reference, the grouped-CV summary at the selected penalty looks like this:

```{r cv-table}
cv_summary <- data.frame(
  partition = c("sd", "ev", "pp", "sh", "en", "ps"),
  cv_log_loss = c(0.1986, 0.3314, 0.3036, 0.2211, 0.6191, 0.6241),
  cv_roc_auc = c(0.7718, 0.6728, 0.6693, 0.7960, 0.7002, 0.5264),
  cv_brier = c(0.0525, 0.0953, 0.0852, 0.0628, 0.2161, 0.2163)
)

make_table(
  cv_summary,
  caption = "Grouped cross-validation diagnostics at the selected ridge penalty.",
  digits = 4
)
```

The broad reading is sensible. `sd` dominates the sample and has the steadiest large-sample behavior. `sh` discriminates well but from a much smaller base. `ps` is the least stable partition because it is both structurally different and much smaller.

## External Results

The more interesting question is how the model behaves away from the training fold selection step. The external evaluation script scores the saved ridge workflows on `2021-22`, `2023-24`, and `2025-26`, with `2025-26` acting as the genuine future season relative to the `2023-24` and `2024-25` training window. Overall external results:

```{r overall-table}
overall_results <- data.frame(
  season = c("2021-22", "2023-24", "2025-26"),
  rows = c(122341, 122180, 74169),
  goal_rate = c(0.0730, 0.0718, 0.0744),
  xg_rate = c(0.0757, 0.0715, 0.0779),
  log_loss = c(0.2316, 0.2222, 0.2319),
  roc_auc = c(0.7463, 0.7775, 0.7617),
  calibration_ratio = c(1.0363, 0.9958, 1.0465)
)
make_table(
  overall_results,
  caption = "External evaluation summary by season.",
  digits = 4
)
```

The `2025-26` row is the one to focus on. It says the model remained usable on a future season, with overall calibration slightly high and ROC AUC still in a respectable range for a public-data xG model. The `2025-26` partition results tell the same story in more detail:

```{r future-partition-table}
future_partition_results <- data.frame(
  partition = c("sd", "ev", "pp", "sh", "en", "ps"),
  rows = c(57157, 1750, 12489, 1610, 604, 559),
  log_loss = c(0.2056, 0.3109, 0.3045, 0.2198, 0.5959, 0.6336),
  roc_auc = c(0.7615, 0.7021, 0.6517, 0.7844, 0.7400, 0.5131),
  calibration_ratio = c(1.0324, 1.1482, 1.0818, 1.1837, 1.0115, 0.9623)
)
make_table(
  future_partition_results,
  caption = "Future-season (`2025-26`) external results by partition.",
  digits = 4
)
```

That table is a good reminder that xG should be interpreted with the structure of the game state in mind. The 5v5 `sd` model is the workhorse. Empty-net scoring behaves like its own world. Shootout scoring is much noisier. None of that is a flaw in the package implementation. It is the underlying data-generating process telling you that some states are more predictable and better sampled than others.

## Practical Takeaways

If you want the short version of what changed in the package, it is this:

1. `nhlscraper` no longer exposes xG as a set of model versions.
2. The built-in scorer is now a single six-partition ridge system.
3. The package mirrors the training-time preprocessing instead of relying on a runtime modeling dependency.
4. The model uses more than shot location: it also uses movement, state, previous-event context, biometrics, and shift burden.

That makes the package xG path more coherent. The implementation is lighter, the modeling contract is explicit, and the article story is easier to tell honestly: this is not one monolithic probability model pretending all shots are alike. It is a practical package-facing system that first asks *what kind of shot environment is this?* and only then asks *how likely is this attempt to score?*

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.require_shift_chart_columns}
\alias{.require_shift_chart_columns}
\title{Validate required shift-chart columns}
\usage{
.require_shift_chart_columns(shift_chart, cols, fn_name)
}
\arguments{
\item{shift_chart}{data.frame shift chart object}

\item{cols}{character vector of required columns}

\item{fn_name}{calling function name used in the error message}
}
\value{
\code{NULL}, invisibly, or an error if required columns are missing
}
\description{
\code{.require_shift_chart_columns()} errors when a helper is called with a shift
chart missing required public schema columns.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Event.R
\name{replay}
\alias{replay}
\title{Access the replay for an event}
\usage{
replay(game = 2023030417, event = 866)
}
\arguments{
\item{game}{integer ID (e.g., 2025020262); see \code{\link[=games]{games()}} for reference}

\item{event}{integer ID (e.g., 751); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or
\code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference; must be a 'goal' event}
}
\value{
data.frame with one row per decisecond
}
\description{
\code{replay()} retrieves the replay for an event as a \code{data.frame} where each row represents decisecond and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
}
\examples{
Gabriel_Landeskog_first_regular_goal_back_replay <- replay(
  game  = 2025020262,
  event = 751
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{to_team_id}
\alias{to_team_id}
\title{Convert to the appropriate team ID}
\usage{
to_team_id(team, lookup = .to_team_id)
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL')}
}
\value{
integer in 1:68
}
\description{
Convert to the appropriate team ID
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayerOld.R
\name{get_spotlight_players}
\alias{get_spotlight_players}
\title{Access the spotlight players}
\usage{
get_spotlight_players()
}
\value{
data.frame with one row per player
}
\description{
\code{get_spotlight_players()} is deprecated. Use \code{\link[=spotlight_players]{spotlight_players()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNTeamOld.R
\name{get_espn_team}
\alias{get_espn_team}
\title{Access the ESPN summary for a team and season}
\usage{
get_espn_team()
}
\description{
\code{get_espn_team()} is defunct. Use \code{\link[=espn_team_summary]{espn_team_summary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_on_ice_matches_situation_code}
\alias{.html_on_ice_matches_situation_code}
\title{Check HTML on-ice counts against a situation code}
\usage{
.html_on_ice_matches_situation_code(
  situation_code,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
)
}
\arguments{
\item{situation_code}{raw API situation code}

\item{home_goalie_player_id}{parsed home goalie player ID}

\item{away_goalie_player_id}{parsed away goalie player ID}

\item{home_skater_player_ids}{parsed home skater player IDs}

\item{away_skater_player_ids}{parsed away skater player IDs}
}
\value{
logical scalar
}
\description{
\code{.html_on_ice_matches_situation_code()} validates that a parsed HTML on-ice
row is compatible with the source API situation code. Missing or unparseable
situation codes are treated as compatible.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{expansion_draft_picks}
\alias{expansion_draft_picks}
\title{Access all the expansion draft picks}
\usage{
expansion_draft_picks()
}
\value{
data.frame with one row per pick
}
\description{
\code{expansion_draft_picks()} retrieves all the expansion draft picks as a \code{data.frame} where each row represents pick and includes detail on date/season filtering windows and chronological context plus team identity, affiliation, and matchup-side context.
}
\examples{
all_expansion_draft_picks <- expansion_draft_picks()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.lookup_roster_player_id}
\alias{.lookup_roster_player_id}
\title{Resolve a roster player ID}
\usage{
.lookup_roster_player_id(
  roster_lookup,
  team_id,
  sweater_number,
  player_name = NA_character_
)
}
\arguments{
\item{roster_lookup}{standardized roster lookup data.frame}

\item{team_id}{team ID}

\item{sweater_number}{sweater number}

\item{player_name}{optional player name}
}
\value{
integer player ID or \code{NA_integer_}
}
\description{
\code{.lookup_roster_player_id()} resolves a player ID from team, sweater number,
and optional name information within a roster lookup table.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNCoachOld.R
\name{get_espn_coach_career}
\alias{get_espn_coach_career}
\title{Access the career ESPN statistics for a coach}
\usage{
get_espn_coach_career()
}
\description{
\code{get_espn_coach_career()} is defunct. Use \code{\link[=coach_career_statistics]{coach_career_statistics()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_team_roster}
\alias{get_team_roster}
\title{Access the roster for a team, season, and player type}
\usage{
get_team_roster(team = "NJD", season = "current", player_type = "forwards")
}
\arguments{
\item{team}{three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for reference}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{player_type}{character of 'forwards', 'defensemen', or 'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{get_team_roster()} is deprecated. Use \code{\link[=roster]{roster()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.add_one_on_one_on_ice_players}
\alias{.add_one_on_one_on_ice_players}
\title{Add one-on-one shooter/goalie assignments}
\usage{
.add_one_on_one_on_ice_players(play_by_play, matched = data.frame())
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{matched}{optional matched HTML rows from \code{.match_html_pbp_to_api()}}
}
\value{
data.frame enriched with one-on-one shooter/goalie assignments
}
\description{
\code{.add_one_on_one_on_ice_players()} populates on-ice player-ID columns for
penalty-shot and shootout rows whose \code{situationCode} implies a one-skater
versus one-goalie state (\code{0101}/\code{1010}). The raw \code{situationCode} remains the
source column; this helper only synthesizes the compatible scalar player-ID
assignments when the general HTML on-ice signature is not usable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.normalize_html_team_abbrev}
\alias{.normalize_html_team_abbrev}
\title{Normalize an HTML team abbreviation}
\usage{
.normalize_html_team_abbrev(x)
}
\arguments{
\item{x}{character vector of team abbreviations}
}
\value{
normalized character vector
}
\description{
\code{.normalize_html_team_abbrev()} uppercases and strips punctuation from HTML
team abbreviations so dotted legacy abbreviations still match.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.parse_html_on_ice_cell}
\alias{.parse_html_on_ice_cell}
\title{Parse an HTML on-ice cell}
\usage{
.parse_html_on_ice_cell(text, team_id, roster_lookup)
}
\arguments{
\item{text}{on-ice cell text}

\item{team_id}{team ID for the cell}

\item{roster_lookup}{standardized roster lookup data.frame}
}
\value{
named list containing goalie and skater player IDs
}
\description{
\code{.parse_html_on_ice_cell()} resolves a goalie and skater player-ID list from
a single team on-ice HTML cell.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DraftOld.R
\name{get_drafts}
\alias{get_drafts}
\title{Access all the drafts}
\usage{
get_drafts()
}
\value{
data.frame with one row per draft
}
\description{
\code{get_drafts()} is deprecated. Use \code{\link[=drafts]{drafts()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GoalieEDGE.R
\name{goalie_edge_shot_location}
\alias{goalie_edge_shot_location}
\title{Access the EDGE shot location statistics for a goalie, season, game type,
and category}
\usage{
goalie_edge_shot_location(
  player = 8476945,
  season = "now",
  game_type = "",
  category = "details"
)
}
\arguments{
\item{player}{integer ID (e.g., 8478406)}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see
\code{\link[=goalie_edge_seasons]{goalie_edge_seasons()}} for reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
'playoff'/'post'; see \code{\link[=goalie_edge_seasons]{goalie_edge_seasons()}} for reference; most functions
will NOT support pre-season}

\item{category}{character of 'd'/details' or 't'/'totals'}
}
\value{
data.frame with one row per shot location
}
\description{
\code{goalie_edge_shot_location()} retrieves the EDGE shot location statistics for a goalie, season, game type, and category as a \code{data.frame} where each row represents shot location and includes detail on production, workload, efficiency, and result-level performance outcomes plus NHL EDGE style tracking outputs and relative-performance context.
}
\examples{
Mackenzie_Blackwood_shot_location_totals_regular_20242025 <- 
  goalie_edge_shot_location(
    player    = 8478406,
    season    = 20242025,
    game_type = 2,
    category  = 'T'
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.assign_html_on_ice_player_ids}
\alias{.assign_html_on_ice_player_ids}
\title{Assign resolved HTML on-ice IDs to scalar columns}
\usage{
.assign_html_on_ice_player_ids(
  play_by_play,
  idx,
  home_goalie,
  away_goalie,
  home_skaters,
  away_skaters
)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{idx}{row index to update}

\item{home_goalie}{resolved home goalie player ID}

\item{away_goalie}{resolved away goalie player ID}

\item{home_skaters}{integer vector of resolved home skater IDs}

\item{away_skaters}{integer vector of resolved away skater IDs}
}
\value{
data.frame with scalar on-ice player-ID columns assigned for \code{idx}
}
\description{
\code{.assign_html_on_ice_player_ids()} copies resolved HTML goalie and skater IDs
into the public scalar on-ice player-ID columns for one play-by-play row.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Coach.R
\name{coach_franchise_statistics}
\alias{coach_franchise_statistics}
\alias{coach_franchise_stats}
\title{Access the statistics for all the coaches by franchise and game type}
\usage{
coach_franchise_statistics()

coach_franchise_stats()
}
\value{
data.frame with one row per franchise per coach per game type
}
\description{
\code{coach_franchise_statistics()} retrieves the statistics for all the coaches by franchise and game type as a \code{data.frame} where each row represents franchise per coach per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
}
\examples{
coach_franchise_stats <- coach_franchise_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.covid_round_robin_shootout_game_ids}
\alias{.covid_round_robin_shootout_game_ids}
\title{Identify Return-to-Play round-robin shootout games}
\usage{
.covid_round_robin_shootout_game_ids()
}
\value{
integer vector of game IDs
}
\description{
\code{.covid_round_robin_shootout_game_ids()} returns the 2019-20 round-robin
seeding games that used regular-season overtime/shootout rules despite
carrying playoff game IDs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Player.R
\name{player_summary}
\alias{player_summary}
\title{Access the summary for a player}
\usage{
player_summary(player = 8478402)
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}
}
\value{
list with various items
}
\description{
\code{player_summary()} retrieves the summary for a player as a nested \code{list} that separates summary and detail blocks for player identity, role, handedness, and biographical profile.
}
\examples{
Martin_Necas_summary <- player_summary(player = 8480039)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GameOld.R
\name{get_games}
\alias{get_games}
\title{Access all the games}
\usage{
get_games()
}
\value{
data.frame with one row per game
}
\description{
\code{get_games()} is deprecated. Use \code{\link[=games]{games()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.assign_on_ice_shift_metric}
\alias{.assign_on_ice_shift_metric}
\title{Assign scalar on-ice timing metrics}
\usage{
.assign_on_ice_shift_metric(
  play_by_play,
  home_matrix,
  away_matrix,
  metric_suffix
)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{home_matrix}{numeric matrix for home on-ice players}

\item{away_matrix}{numeric matrix for away on-ice players}

\item{metric_suffix}{scalar timing suffix}
}
\value{
data.frame with timing columns assigned
}
\description{
\code{.assign_on_ice_shift_metric()} copies goalie and skater timing matrices into
home, away, for, and against scalar play-by-play columns.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Player.R
\name{player_seasons}
\alias{player_seasons}
\title{Access the season(s) and game type(s) in which a player played}
\usage{
player_seasons(player = 8478402)
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}
}
\value{
data.frame with one row per season
}
\description{
\code{player_seasons()} retrieves the season(s) and game type(s) in which a player played as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context.
}
\examples{
Martin_Necas_seasons <- player_seasons(player = 8480039)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_scoring}
\alias{goalie_scoring}
\title{Access the career scoring statistics for all the goalies}
\usage{
goalie_scoring()
}
\value{
data.frame with one row per player
}
\description{
\code{goalie_scoring()} retrieves the career scoring statistics for all the goalies as a \code{data.frame} where each row represents player and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
}
\examples{
goalie_scoring <- goalie_scoring()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FranchiseOld.R
\name{get_franchise_season_by_season}
\alias{get_franchise_season_by_season}
\title{Access the statistics for all the franchises by season and game type}
\usage{
get_franchise_season_by_season()
}
\value{
data.frame with one row per franchise per season per game type
}
\description{
\code{get_franchise_season_by_season()} is deprecated. Use
\code{\link[=franchise_season_statistics]{franchise_season_statistics()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{add_goalie_biometrics}
\alias{add_goalie_biometrics}
\title{Add goalie biometrics to (a) play-by-play(s)}
\usage{
add_goalie_biometrics(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}}, \code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}}
}
\value{
data.frame with one row per event (play) and added columns: \code{goalieHeight}, \code{goalieWeight}, \code{goalieHandCode}, and \code{goalieAge}
}
\description{
\code{add_goalie_biometrics()} adds goalie biometrics (height, weight, hand, and age at game date) to (a) play-by-play(s) for shot attempts. Goalie identity is resolved from \code{goalieInNetId} first and \code{goaliePlayerIdAgainst} second. If neither is available on a row, the added goalie biometrics are left as \code{NA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterEDGE.R
\name{skater_edge_skating_distance}
\alias{skater_edge_skating_distance}
\title{Access the EDGE skating distance statistics for a skater, season, game type,
and category}
\usage{
skater_edge_skating_distance(
  player = 8478402,
  season = "now",
  game_type = "",
  category = "details"
)
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see
\code{\link[=skater_edge_seasons]{skater_edge_seasons()}} for reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
'playoff'/'post'; see \code{\link[=skater_edge_seasons]{skater_edge_seasons()}} for reference; most functions
will NOT support pre-season}

\item{category}{character of 'd'/'details' or 'l'/'l10'/'last 10'}
}
\value{
data.frame with one row per strength state (category = 'details')
or game (category = 'last 10')
}
\description{
\code{skater_edge_skating_distance()} retrieves the EDGE skating distance statistics for a skater, season, game type, and category as a \code{data.frame} where each row represents strength state and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and tracking/spatial detail such as location, speed, distance, and zone distribution.
}
\examples{
Martin_Necas_L10_skating_distance_regular_20242025 <- 
  skater_edge_skating_distance(
    player    = 8480039,
    season    = 20242025,
    game_type = 2,
    category  = 'L'
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Manager.R
\name{general_managers}
\alias{general_managers}
\alias{gms}
\title{Access all the general managers}
\usage{
general_managers()

gms()
}
\value{
data.frame with one row per general manager
}
\description{
\code{general_managers()} retrieves all the general managers as a \code{data.frame} where each row represents general manager and includes detail on player identity, role, handedness, and biographical profile.
}
\examples{
all_GMs <- general_managers()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNOther.R
\name{espn_futures}
\alias{espn_futures}
\title{Access the ESPN futures for a season}
\usage{
espn_futures(season = season_now())
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
nested data.frame with one row per type (outer) and book (inner)
}
\description{
\code{espn_futures()} retrieves the ESPN futures for a season as a \code{data.frame} where each row represents type and includes detail on betting market snapshots with side/total prices and provider variation.
}
\examples{
ESPN_futures_20252026 <- espn_futures(20252026)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Other.R
\name{tv_schedule}
\alias{tv_schedule}
\title{Access the NHL Network TV schedule for a date}
\usage{
tv_schedule(date = "now")
}
\arguments{
\item{date}{character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see
\code{\link[=seasons]{seasons()}} for reference}
}
\value{
data.frame with one row per program
}
\description{
\code{tv_schedule()} retrieves the NHL Network TV schedule for a date as a \code{data.frame} where each row represents program and includes detail on date/season filtering windows and chronological context.
}
\examples{
tv_schedule_Halloween_2025 <- tv_schedule(date = '2025-10-31')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_seasons}
\alias{team_seasons}
\title{Access the season(s) and game type(s) in which a team played}
\usage{
team_seasons(team = 1)
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for
reference; ID is preferable as there now exists duplicate three-letter codes
(i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')}
}
\value{
data.frame with one row per season
}
\description{
\code{team_seasons()} retrieves the season(s) and game type(s) in which a team played as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context.
}
\examples{
COL_seasons <- team_seasons(team = 21)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OtherOld.R
\name{get_glossary}
\alias{get_glossary}
\title{Access the glossary}
\usage{
get_glossary()
}
\value{
data.frame with one row per terminology
}
\description{
\code{get_glossary()} is deprecated. Use \code{\link[=glossary]{glossary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNOther.R
\name{espn_injuries}
\alias{espn_injuries}
\title{Access the real-time ESPN injury reports}
\usage{
espn_injuries()
}
\value{
nested data.frame with one row per team (outer) and player (inner)
}
\description{
\code{espn_injuries()} retrieves the real-time ESPN injury reports as a \code{data.frame} where each row represents team and includes detail on availability status tracking for injuries or transactions.
}
\examples{
ESPN_injuries_now <- espn_injuries()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Franchise.R
\name{franchise_versus_franchise}
\alias{franchise_versus_franchise}
\alias{franchise_vs_franchise}
\title{Access the all-time statistics versus other franchises for all the
franchises by game type}
\usage{
franchise_versus_franchise()

franchise_vs_franchise()
}
\value{
data.frame with one row per franchise per franchise per game type
}
\description{
\code{franchise_versus_franchise()} retrieves the all-time statistics versus other franchises for all the franchises by game type as a \code{data.frame} where each row represents franchise per franchise per game type and includes detail on date/season filtering windows and chronological context plus team identity, affiliation, and matchup-side context.
}
\examples{
# May take >5s, so skip.
\donttest{franchise_vs_franchise <- franchise_versus_franchise()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNOtherOld.R
\name{get_espn_futures}
\alias{get_espn_futures}
\title{Access the ESPN futures for a season}
\usage{
get_espn_futures()
}
\description{
\code{get_espn_futures()} is defunct. Use \code{\link[=espn_futures]{espn_futures()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_season_statistics}
\alias{team_season_statistics}
\alias{team_season_stats}
\title{Access the statistics for all the teams by season and game type}
\usage{
team_season_statistics()

team_season_stats()
}
\value{
data.frame with one row per team per season per game type
}
\description{
\code{team_season_statistics()} retrieves the statistics for all the teams by season and game type as a \code{data.frame} where each row represents team per season per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
# May take >5s, so skip.
\donttest{team_season_statistics <- team_season_statistics()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_desc_owner_team_id}
\alias{.html_desc_owner_team_id}
\title{Resolve the owner team from an HTML description}
\usage{
.html_desc_owner_team_id(
  description,
  home_abbrev,
  away_abbrev,
  home_team_id,
  away_team_id
)
}
\arguments{
\item{description}{HTML event description}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}
}
\value{
integer team ID or \code{NA_integer_}
}
\description{
\code{.html_desc_owner_team_id()} infers the event owner team ID from the leading
team abbreviation embedded in an HTML event description.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_season_schedule}
\alias{team_season_schedule}
\title{Access the schedule for a team and season}
\usage{
team_season_schedule(team = 1, season = "now")
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for
reference; ID is preferable as there now exists duplicate three-letter codes
(i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
data.frame with one row per game
}
\description{
\code{team_season_schedule()} retrieves the schedule for a team and season as a \code{data.frame} where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
COL_schedule_20252026 <- team_season_schedule(
  team   = 21, 
  season = 20252026
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.reconstruct_skater_counts_from_penalties}
\alias{.reconstruct_skater_counts_from_penalties}
\title{Reconstruct skater counts from active penalties}
\usage{
.reconstruct_skater_counts_from_penalties(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame cleaned internal play-by-play object}
}
\value{
data.frame with reconstructed home/away skater counts and metadata
}
\description{
\code{.reconstruct_skater_counts_from_penalties()} walks a cleaned play-by-play in
event order and reconstructs home/away skater counts from active strength-
affecting penalties. The reconstruction intentionally targets regulation and
playoff-strength states; regular-season overtime is left unsupported because
its 3-on-3 penalty rules require a separate ruleset.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.match_html_pbp_to_api}
\alias{.match_html_pbp_to_api}
\title{Match HTML play-by-play rows to API rows}
\usage{
.match_html_pbp_to_api(play_by_play, html_rows)
}
\arguments{
\item{play_by_play}{data.frame API play-by-play object}

\item{html_rows}{data.frame parsed HTML play-by-play rows}
}
\value{
data.frame of matched HTML rows with \code{apiIndex}
}
\description{
\code{.match_html_pbp_to_api()} aligns parsed HTML play-by-play rows back to API
play-by-play rows using exact keys, greedy scoring, and a reciprocal-best
fallback for duplicate clusters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeagueOld.R
\name{get_schedule}
\alias{get_schedule}
\title{Access the schedule for a date}
\usage{
get_schedule(date = "2025-01-01")
}
\arguments{
\item{date}{character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see
\code{\link[=seasons]{seasons()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{get_schedule()} is deprecated. Use \code{\link[=schedule]{schedule()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_on_ice_matches_late_empty_net}
\alias{.html_on_ice_matches_late_empty_net}
\title{Check HTML on-ice state against a late empty-net heuristic}
\usage{
.html_on_ice_matches_late_empty_net(
  play_by_play,
  reconstruction,
  idx,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
)
}
\arguments{
\item{play_by_play}{data.frame cleaned internal play-by-play object}

\item{reconstruction}{data.frame from \code{.reconstruct_skater_counts_from_penalties()}}

\item{idx}{row index in the play-by-play}

\item{home_goalie_player_id}{parsed home goalie player ID}

\item{away_goalie_player_id}{parsed away goalie player ID}

\item{home_skater_player_ids}{parsed home skater IDs}

\item{away_skater_player_ids}{parsed away skater IDs}
}
\value{
logical scalar
}
\description{
\code{.html_on_ice_matches_late_empty_net()} accepts late-game 6-on-5 HTML rows
when the trailing team has clearly pulled its goalie and there are no active
reconstructed penalties that would already explain the manpower mismatch.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.supported_strength_event_types}
\alias{.supported_strength_event_types}
\title{List strength-context event types}
\usage{
.supported_strength_event_types()
}
\value{
character vector of supported event type keys
}
\description{
\code{.supported_strength_event_types()} returns the event types that support the
HTML row-matching pipeline.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.supported_html_on_ice_id_event_types}
\alias{.supported_html_on_ice_id_event_types}
\title{List HTML on-ice player-ID event types}
\usage{
.supported_html_on_ice_id_event_types()
}
\value{
character vector of supported on-ice player-ID event type keys
}
\description{
\code{.supported_html_on_ice_id_event_types()} returns the event types that are
allowed to receive HTML-derived on-ice player IDs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.parse_html_period_label}
\alias{.parse_html_period_label}
\title{Parse an HTML period label}
\usage{
.parse_html_period_label(x, is_playoffs = FALSE)
}
\arguments{
\item{x}{HTML period label}

\item{is_playoffs}{logical; whether the game is a playoff game}
}
\value{
integer period number or \code{NA_integer_}
}
\description{
\code{.parse_html_period_label()} converts HTML report period labels such as \code{OT}
and \code{SO} into integer period numbers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.compute_shot_context}
\alias{.compute_shot_context}
\title{Compute shot context summaries}
\usage{
.compute_shot_context(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}
}
\value{
named list of logical and integer shot-context vectors
}
\description{
\code{.compute_shot_context()} derives rush and rebound flags plus running goal,
shot, Fenwick, and Corsi counts, using native code when available.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DraftOld.R
\name{get_draft_picks}
\alias{get_draft_picks}
\title{Access all the draft picks}
\usage{
get_draft_picks()
}
\value{
data.frame with one row per pick
}
\description{
\code{get_draft_picks()} is deprecated. Use \code{\link[=draft_picks]{draft_picks()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGameOld.R
\name{get_espn_event_stars}
\alias{get_espn_event_stars}
\title{Access the three stars for an ESPN event (game)}
\usage{
get_espn_event_stars()
}
\description{
\code{get_espn_event_stars()} is defunct. Use \code{\link[=gc_summary]{gc_summary()}} and/or
\code{\link[=wsc_summary]{wsc_summary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{gc_summary}
\alias{gc_summary}
\title{Access the GameCenter (GC) summary for a game}
\usage{
gc_summary(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
list of various items
}
\description{
\code{gc_summary()} retrieves the GameCenter (GC) summary for a game as a nested \code{list} that separates summary and detail blocks for game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and venue/location geography and regional metadata.
}
\examples{
gc_summary_Martin_Necas_legacy_game <- gc_summary(game = 2025020275)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Playoff.R
\name{bracket}
\alias{bracket}
\title{Access the playoff bracket for a season}
\usage{
bracket(season = season_now())
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
data.frame with one row per series
}
\description{
\code{bracket()} retrieves the playoff bracket for a season as a \code{data.frame} where each row represents series and includes detail on team identity, affiliation, and matchup-side context.
}
\examples{
bracket_20242025 <- bracket(season = 20242025)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayerOld.R
\name{get_player_game_log}
\alias{get_player_game_log}
\title{Access the game log for a player, season, and game type}
\usage{
get_player_game_log(player = 8478402, season = "now", game_type = "")
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}
}
\value{
data.frame with one row per game
}
\description{
\code{get_player_game_log()} is deprecated. Use \code{\link[=player_game_log]{player_game_log()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.backfill_delayed_penalty_on_ice_players}
\alias{.backfill_delayed_penalty_on_ice_players}
\title{Backfill unmatched delayed-penalty rows from nearby populated rows}
\usage{
.backfill_delayed_penalty_on_ice_players(play_by_play, max_gap_seconds = 15L)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{max_gap_seconds}{integer scalar time window}
}
\value{
data.frame with delayed-penalty rows backfilled where possible
}
\description{
\code{.backfill_delayed_penalty_on_ice_players()} fills unmatched delayed-penalty
rows from the nearest previous populated supported row in the same game and
period when the raw \code{situationCode} and goalie/skater counts are unchanged.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_game_scoring}
\alias{goalie_game_scoring}
\title{Access the scoring statistics for all the goalies by game}
\usage{
goalie_game_scoring()
}
\value{
data.frame with one row per player per game
}
\description{
\code{goalie_game_scoring()} retrieves the scoring statistics for all the goalies by game as a \code{data.frame} with detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
goalie_game_scoring <- goalie_game_scoring()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OtherOld.R
\name{get_countries}
\alias{get_countries}
\title{Access all the countries}
\usage{
get_countries()
}
\value{
data.frame with one row per country
}
\description{
\code{get_countries()} is deprecated. Use \code{\link[=countries]{countries()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_extract_actor_player_ids}
\alias{.html_extract_actor_player_ids}
\title{Extract actor player IDs from an HTML event}
\usage{
.html_extract_actor_player_ids(
  description,
  type_desc_key,
  owner_team_id,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev,
  roster_lookup
)
}
\arguments{
\item{description}{HTML event description}

\item{type_desc_key}{internal event type key}

\item{owner_team_id}{owner team ID}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}

\item{roster_lookup}{standardized roster lookup data.frame}
}
\value{
named list of actor player IDs
}
\description{
\code{.html_extract_actor_player_ids()} resolves primary, secondary, and tertiary
actor player IDs from HTML event descriptions using the roster lookup.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.calculate_distance}
\alias{.calculate_distance}
\title{Calculate the Euclidean distance from the attacking net for all the events (plays) in a play-by-play}
\usage{
.calculate_distance(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with one row per event (play) and added \code{distance} column
}
\description{
\code{.calculate_distance()} calculates the Euclidean distance from the attacking net for all the events (plays) in a play-by-play.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.normalize_roster_person}
\alias{.normalize_roster_person}
\title{Normalize roster person labels}
\usage{
.normalize_roster_person(x)
}
\arguments{
\item{x}{character vector of player names}
}
\value{
normalized character vector
}
\description{
\code{.normalize_roster_person()} standardizes roster names for stable matching by
uppercasing, trimming, collapsing whitespace, and transliterating accents.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.compute_pbp_deltas_in_r}
\alias{.compute_pbp_deltas_in_r}
\title{Compute play-by-play deltas in R}
\usage{
.compute_pbp_deltas_in_r(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}}, \code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}}
}
\value{
named list containing delta context vectors
}
\description{
\code{.compute_pbp_deltas_in_r()} is the pure-R fallback for event-to-event delta
calculations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{lottery_odds}
\alias{lottery_odds}
\title{Access the draft lottery odds}
\usage{
lottery_odds()
}
\value{
data.frame with one row per draft lottery
}
\description{
\code{lottery_odds()} retrieves the draft lottery odds as a \code{data.frame} where each row represents draft lottery and includes detail on draft-cycle evaluation, ranking, and selection tracking detail.
}
\examples{
lottery_odds <- lottery_odds()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Playoff.R
\name{series}
\alias{series}
\title{Access all the playoff series by game}
\usage{
series()
}
\value{
data.frame with one row per game per series
}
\description{
\code{series()} retrieves all the playoff series by game as a \code{data.frame} where each row represents game per series and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and playoff-series progression, round status, and series leverage.
}
\examples{
# May take >5s, so skip.
\donttest{all_series <- series()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_event_code_to_type_desc}
\alias{.html_event_code_to_type_desc}
\title{Map HTML event codes to event type keys}
\usage{
.html_event_code_to_type_desc(event_code, description = NA_character_)
}
\arguments{
\item{event_code}{HTML event code}

\item{description}{optional HTML event description}
}
\value{
event type key or \code{NA_character_}
}
\description{
\code{.html_event_code_to_type_desc()} converts HTML report event codes and select
description fallbacks into internal play-by-play event type keys.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{.html_report_request}
\alias{.html_report_request}
\title{Build an HTML report request}
\usage{
.html_report_request(url)
}
\arguments{
\item{url}{character scalar}
}
\value{
httr2 request object
}
\description{
\code{.html_report_request()} constructs an HTML report request with the same retry
policy used for API calls.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Skater.R
\name{skater_leaders}
\alias{skater_leaders}
\title{Access the skater statistics leaders for a season, game type, and category}
\usage{
skater_leaders(season = "current", game_type = "", category = "points")
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}

\item{category}{string of 'a'/'assists', 'g'/goals',
'shg'/'shorthanded goals', 'ppg'/'powerplay goals', 'p'/'points',
'pim'/penalty minutes'/'penalty infraction minutes', 'toi'/'time on ice',
'pm'/'plus minus', or 'f'/'faceoffs'}
}
\value{
data.frame with one row per player
}
\description{
\code{skater_leaders()} retrieves the skater statistics leaders for a season, game type, and category as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
}
\examples{
TOI_leaders_regular_20242025 <- skater_leaders(
  season    = 20242025,
  game_type = 2,
  category  = 'TOI'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{scope_person_name_cols}
\alias{scope_person_name_cols}
\title{Scope generic person-name columns to an entity}
\usage{
scope_person_name_cols(x, prefix)
}
\arguments{
\item{x}{character vector}

\item{prefix}{character scalar (e.g., 'player', 'goalie', 'skater')}
}
\value{
character vector
}
\description{
Scope generic person-name columns to an entity
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Load.R
\name{contracts}
\alias{contracts}
\title{Access all contracts from packaged internal data}
\usage{
contracts()
}
\value{
data.frame with one row per contract
}
\description{
\code{contracts()} loads preprocessed contract records bundled with the package and returns a cleaned \code{data.frame} with package-consistent column names, season IDs, numeric money fields, and team/player identifiers.
}
\examples{
\donttest{all_contracts <- contracts()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.parse_situation_code_components}
\alias{.parse_situation_code_components}
\title{Parse situation-code components}
\usage{
.parse_situation_code_components(situation_code)
}
\arguments{
\item{situation_code}{vector of raw situation codes}
}
\value{
integer matrix with away/home goalie and skater counts
}
\description{
\code{.parse_situation_code_components()} converts situation codes into away/home
goalie and skater counts without mutating the original source vector.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Load.R
\name{gc_play_by_plays_raw}
\alias{gc_play_by_plays_raw}
\alias{gc_pbps_raw}
\title{Access the raw GameCenter (GC) play-by-plays for a season}
\usage{
gc_play_by_plays_raw(season = 20242025)

gc_pbps_raw(season = 20242025)
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
data.frame with one row per raw event (play) per game
}
\description{
\code{gc_play_by_plays_raw()} loads the raw GC play-by-plays for a given
\code{season}.
}
\examples{
# May take >5s, so skip.
\donttest{gc_pbps_raw_20212022 <- gc_play_by_plays_raw(season = 20212022)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.normalize_situation_code_for_parse}
\alias{.normalize_situation_code_for_parse}
\title{Normalize situation codes for parsing}
\usage{
.normalize_situation_code_for_parse(situation_code)
}
\arguments{
\item{situation_code}{vector of raw situation codes}
}
\value{
character vector of parse-ready situation codes
}
\description{
\code{.normalize_situation_code_for_parse()} pads 1-4 digit situation codes to
four characters for internal parsing without rewriting the source column.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{draft_tracker}
\alias{draft_tracker}
\title{Access the real-time draft tracker}
\usage{
draft_tracker()
}
\value{
data.frame with one row per player
}
\description{
\code{draft_tracker()} retrieves the real-time draft tracker as a \code{data.frame} where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and venue/location geography and regional metadata.
}
\examples{
draft_tracker <- draft_tracker()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeagueOld.R
\name{get_standings_information}
\alias{get_standings_information}
\title{Access the standings rules by season}
\usage{
get_standings_information()
}
\value{
data.frame with one row per season
}
\description{
\code{get_standings_information()} is deprecated. Use \code{\link[=standings_rules]{standings_rules()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GameOld.R
\name{get_game_boxscore}
\alias{get_game_boxscore}
\title{Access the boxscore for a game, team, and player type}
\usage{
get_game_boxscore(game = 2023030417, team = "home", player_type = "forwards")
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}

\item{team}{character of 'home' or 'away'}

\item{player_type}{character of 'forwards', 'defense', or 'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{get_game_boxscore()} is deprecated. Use \code{\link[=boxscore]{boxscore()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Graph.R
\name{draw_NHL_rink}
\alias{draw_NHL_rink}
\title{Draw a full NHL rink}
\usage{
draw_NHL_rink()
}
\value{
\code{NULL}
}
\description{
\code{draw_NHL_rink()} draws a full NHL rink such that the x and y coordinates
span -100 to 100 and -43 to +43, respectively. Use \code{\link[graphics:points]{graphics::points()}}
to create custom graphs; check out an example on the online documentation!
}
\examples{
draw_NHL_rink()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterOld.R
\name{get_skaters}
\alias{get_skaters}
\title{Access all the skaters for a range of seasons}
\usage{
get_skaters()
}
\description{
\code{get_skaters()} is defunct. Use \code{\link[=players]{players()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNOtherOld.R
\name{get_espn_injuries}
\alias{get_espn_injuries}
\title{Access the real-time ESPN injury reports}
\usage{
get_espn_injuries()
}
\value{
nested data.frame with one row per team (outer) and player (inner)
}
\description{
\code{get_espn_injuries()} is deprecated. Use \code{\link[=espn_injuries]{espn_injuries()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeagueOld.R
\name{get_season_now}
\alias{get_season_now}
\title{Access the season and game type as of now}
\usage{
get_season_now()
}
\description{
\code{get_season_now()} is defunct. Use \code{\link[=season_now]{season_now()}} and/or \code{\link[=game_type_now]{game_type_now()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{combine_reports}
\alias{combine_reports}
\title{Access the draft combine reports}
\usage{
combine_reports()
}
\value{
data.frame with one row per player
}
\description{
\code{combine_reports()} retrieves the draft combine reports as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
}
\examples{
combine_reports <- combine_reports()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_season_report}
\alias{team_season_report}
\title{Access various reports for a season, game type, and category for all
the teams by season}
\usage{
team_season_report(
  season = season_now(),
  game_type = game_type_now(),
  category = "summary"
)
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}

\item{category}{character (e.g., 'leadingtrailing'); see
\code{\link[=team_report_configurations]{team_report_configurations()}} for reference}
}
\value{
data.frame with one row per team
}
\description{
\code{team_season_report()} retrieves various reports for a season, game type, and category for all the teams by season as a \code{data.frame} where each row represents team and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
situational_team_season_report_playoffs_20212022 <- team_season_report(
  season    = 20212022, 
  game_type = 3, 
  category  = 'leadingtrailing'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.on_ice_timing_scalar_column_names}
\alias{.on_ice_timing_scalar_column_names}
\title{Build scalar on-ice timing column names}
\usage{
.on_ice_timing_scalar_column_names(
  metric_suffix,
  play_by_play = NULL,
  slot_count = NULL
)
}
\arguments{
\item{metric_suffix}{scalar timing suffix}

\item{play_by_play}{optional data.frame whose existing on-ice columns should
be inspected}

\item{slot_count}{optional integer scalar minimum slot count requested by the
caller}
}
\value{
character vector of column names
}
\description{
\code{.on_ice_timing_scalar_column_names()} returns the scalar goalie and skater
timing column names for a timing metric suffix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DraftOld.R
\name{get_draft_tracker}
\alias{get_draft_tracker}
\title{Access the real-time draft tracker}
\usage{
get_draft_tracker()
}
\value{
data.frame with one row per player
}
\description{
\code{get_draft_tracker()} is deprecated. Use \code{\link[=draft_tracker]{draft_tracker()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.score_html_api_candidates}
\alias{.score_html_api_candidates}
\title{Score HTML-to-API matching candidates}
\usage{
.score_html_api_candidates(api, html, cand, h_idx, last_api_seq = NULL)
}
\arguments{
\item{api}{API-side matching table}

\item{html}{HTML-side matching table}

\item{cand}{integer candidate API indices}

\item{h_idx}{HTML row index being scored}

\item{last_api_seq}{optional last matched API sequence number}
}
\value{
named numeric vector of candidate scores
}
\description{
\code{.score_html_api_candidates()} scores candidate API rows for a single HTML
row using time, type, team, player, and local sequence agreement.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_report_configurations}
\alias{team_report_configurations}
\alias{team_report_configs}
\title{Access the configurations for team reports}
\usage{
team_report_configurations()

team_report_configs()
}
\value{
list with various items
}
\description{
\code{team_report_configurations()} retrieves the configurations for team reports as a nested \code{list} that separates summary and detail blocks for situational splits across home/road, strength state, and overtime/shootout states plus configuration catalogs for valid report categories and filters.
}
\examples{
team_report_configs <- team_report_configurations()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Other.R
\name{streams}
\alias{streams}
\title{Access all the streams}
\usage{
streams()
}
\value{
data.frame with one row per stream
}
\description{
\code{streams()} retrieves all the streams as a \code{data.frame} where each row represents stream and includes detail on reference metadata, regional context, and media availability detail.
}
\examples{
all_streams <- streams()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Other.R
\name{location}
\alias{location}
\title{Access the location for a zip code}
\usage{
location(zip = 10001)
}
\arguments{
\item{zip}{integer (e.g., 48304)}
}
\value{
data.frame with one row per team
}
\description{
\code{location()} retrieves the location for a zip code as a \code{data.frame} where each row represents team and includes detail on venue/location geography and regional metadata.
}
\examples{
Cranbrook_Schools <- location(48304)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Share.R
\name{x_game_shot_locations}
\alias{x_game_shot_locations}
\alias{x_game_shot_locs}
\title{Save an X (Twitter) share-able shot-location plot for a game}
\usage{
x_game_shot_locations(
  game = 2023030417,
  team = "home",
  model = NULL,
  save = TRUE
)

x_game_shot_locs(game = 2023030417, team = "home", model = NULL)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}

\item{team}{character of 'h'/'home' or 'a'/'away'}

\item{model}{deprecated legacy model selector; ignored}

\item{save}{logical only FALSE for tests}
}
\value{
\code{NULL}
}
\description{
\code{x_game_shot_locations()} saves an X share-able shot-location plot for a
given \code{game}.
}
\examples{
# May take >5s, so skip.
\donttest{x_game_shot_locations(
  game  = 2023030417, 
  team  = 'H',
  save  = FALSE
)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Graph.R
\name{draw_rink_circle}
\alias{draw_rink_circle}
\title{Draw a circle on the rink}
\usage{
draw_rink_circle(x, y, r, n = 200, col = "black", lwd = 1)
}
\arguments{
\item{x}{numeric x coordinate of the circle center}

\item{y}{numeric y coordinate of the circle center}

\item{r}{numeric radius}

\item{n}{integer number of points used to approximate circle}

\item{col}{character color}

\item{lwd}{numeric line width}
}
\value{
\code{NULL}
}
\description{
\code{draw_rink_circle()} draws a circle on the rink.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.add_empty_html_on_ice_columns}
\alias{.add_empty_html_on_ice_columns}
\title{Initialize empty on-ice columns}
\usage{
.add_empty_html_on_ice_columns(play_by_play, slot_count = NULL)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{slot_count}{optional integer scalar minimum slot count requested by the
caller}
}
\value{
data.frame with empty on-ice columns added
}
\description{
\code{.add_empty_html_on_ice_columns()} allocates empty scalar on-ice player ID
and timing columns before HTML and shift-chart enrichment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{standings}
\alias{standings}
\title{Access the standings for a date}
\usage{
standings(date = "now")
}
\arguments{
\item{date}{character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see
\code{\link[=seasons]{seasons()}} for reference}
}
\value{
data.frame with one row per team
}
\description{
\code{standings()} retrieves the standings for a date as a \code{data.frame} where each row represents team and includes detail on date/season filtering windows and chronological context, production, workload, efficiency, and result-level performance outcomes, and ranking movement, points pace, and division/conference position signals.
}
\examples{
standings_Halloween_2025 <- standings(date = '2025-10-31')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_pbp_report_url}
\alias{.html_pbp_report_url}
\title{Build the HTML play-by-play report URL}
\usage{
.html_pbp_report_url(game)
}
\arguments{
\item{game}{game ID}
}
\value{
character scalar URL
}
\description{
\code{.html_pbp_report_url()} returns the NHL HTML play-by-play report URL for a
game.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Skater.R
\name{skater_report_configurations}
\alias{skater_report_configurations}
\alias{skater_report_configs}
\title{Access the configurations for skater reports}
\usage{
skater_report_configurations()

skater_report_configs()
}
\value{
list with various items
}
\description{
\code{skater_report_configurations()} retrieves the configurations for skater reports as a nested \code{list} that separates summary and detail blocks for production, workload, efficiency, and result-level performance outcomes, situational splits across home/road, strength state, and overtime/shootout states, and configuration catalogs for valid report categories and filters.
}
\examples{
skater_report_configs <- skater_report_configurations()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_on_ice_can_override_strength_context}
\alias{.html_on_ice_can_override_strength_context}
\title{Check whether an HTML on-ice signature can override derived strength context}
\usage{
.html_on_ice_can_override_strength_context(
  situation_code,
  home_goalie_player_id,
  away_goalie_player_id,
  home_skater_player_ids,
  away_skater_player_ids
)
}
\arguments{
\item{situation_code}{raw API situation code}

\item{home_goalie_player_id}{parsed home goalie player ID}

\item{away_goalie_player_id}{parsed away goalie player ID}

\item{home_skater_player_ids}{parsed home skater IDs}

\item{away_skater_player_ids}{parsed away skater IDs}
}
\value{
logical scalar
}
\description{
\code{.html_on_ice_can_override_strength_context()} allows the HTML on-ice skater
and goalie counts to replace derived strength context when the resolved
signature still looks like a plausible hockey state. Overflow rows with more
than six skaters are preserved in the scalar player-ID output but do not
rewrite the derived count columns.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{gc_play_by_play}
\alias{gc_play_by_play}
\alias{gc_pbp}
\title{Access the GameCenter (GC) play-by-play for a game}
\usage{
gc_play_by_play(game = 2023030417)

gc_pbp(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
data.frame with one row per event (play)
}
\description{
\code{gc_play_by_play()} retrieves the GameCenter (GC) play-by-play for a game as
a \code{data.frame} where each row represents an event. The returned schema is the
cleaned, public-facing play-by-play schema, including canonical names such as
\code{periodNumber}, \code{eventTypeCode}, \code{eventTypeDescKey}, \code{homeShots},
\code{shotsFor}, \code{penaltyTypeDescKey}, \code{penaltyDuration}, \code{servedByPlayerId},
\code{goalieInNetId}, and HTML-report-derived on-ice player ID columns such as
\code{homeGoaliePlayerId}, \code{awayGoaliePlayerId}, \code{homeSkater1PlayerId}, and any
additional overflow skater slots required by the game. HTML report skater and
goalie IDs are returned whenever they can be matched back to a supported row,
even when the raw \code{situationCode} is stale. Use \code{\link[=add_shift_times]{add_shift_times()}} with
\code{\link[=shift_chart]{shift_chart()}} (or \code{\link[=shift_charts]{shift_charts()}}) to add on-ice shift timing columns.
}
\examples{
# May take >5s, so skip.
\donttest{gc_pbp_Martin_Necas_legacy_game <- gc_play_by_play(game = 2025020275)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.add_html_on_ice_players}
\alias{.add_html_on_ice_players}
\title{Add HTML-derived on-ice players to play-by-play}
\usage{
.add_html_on_ice_players(
  play_by_play,
  game,
  rosters = NULL,
  home_team = NULL,
  away_team = NULL,
  html_rows = NULL
)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{game}{game ID}

\item{rosters}{optional roster data.frame}

\item{home_team}{optional home-team metadata}

\item{away_team}{optional away-team metadata}

\item{html_rows}{optional parsed HTML on-ice rows}
}
\value{
data.frame enriched with HTML-derived on-ice columns
}
\description{
\code{.add_html_on_ice_players()} resolves HTML on-ice goalie and skater IDs,
matches them to API rows, and injects the resulting scalar on-ice columns
into play-by-play output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GameOld.R
\name{shifts}
\alias{shifts}
\title{Access the shift charts for a game}
\usage{
shifts(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
data.frame with one row per shift
}
\description{
\code{shifts()} is deprecated. Use \code{\link[=shift_chart]{shift_chart()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGame.R
\name{espn_game_summary}
\alias{espn_game_summary}
\title{Access the ESPN summary for a game}
\usage{
espn_game_summary(game = 401777460)
}
\arguments{
\item{game}{integer ID (e.g., 401777460); see \code{\link[=espn_games]{espn_games()}} for
reference}
}
\value{
list with various items
}
\description{
\code{espn_game_summary()} retrieves the ESPN summary for a game as a nested \code{list} that separates summary and detail blocks for date/season filtering windows and chronological context, venue/location geography and regional metadata, and playoff-series progression, round status, and series leverage.
}
\examples{
ESPN_summary_SCF_20242025 <- espn_game_summary(game = 401777460)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{seasons}
\alias{seasons}
\title{Access all the seasons}
\usage{
seasons()
}
\value{
data.frame with one row per season
}
\description{
\code{seasons()} retrieves all the seasons as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context.
}
\examples{
all_seasons <- seasons()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{to_game_type_id}
\alias{to_game_type_id}
\title{Convert to the appropriate game type ID}
\usage{
to_game_type_id(game_type)
}
\arguments{
\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or 'playoff'/'post'}
}
\value{
integer in 1:3
}
\description{
Convert to the appropriate game type ID
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{roster_statistics}
\alias{roster_statistics}
\alias{roster_stats}
\title{Access the roster statistics for a team, season, game type, and position}
\usage{
roster_statistics(
  team = 1,
  season = "now",
  game_type = "",
  position = "skaters"
)

roster_stats(team = 1, season = "now", game_type = "", position = "skaters")
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for
reference; ID is preferable as there now exists duplicate three-letter codes
(i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}

\item{position}{character of 's'/'skaters' or 'g'/'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{roster_statistics()} retrieves the roster statistics for a team, season, game type, and position as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus production, workload, efficiency, and result-level performance outcomes.
}
\examples{
COL_goalies_statistics_regular_20242025 <- roster_statistics(
  team      = 21,
  season    = 20242025,
  game_type = 2,
  position  = 'G'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Official.R
\name{officials}
\alias{officials}
\title{Access all the officials}
\usage{
officials()
}
\value{
data.frame with one row per official
}
\description{
\code{officials()} retrieves all the officials as a \code{data.frame} where each row represents official and includes detail on player identity, role, handedness, and biographical profile plus coach/management/officiating identity and assignment history.
}
\examples{
all_officials <- officials()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.on_ice_int_col}
\alias{.on_ice_int_col}
\title{Extract an integer on-ice column}
\usage{
.on_ice_int_col(df, name)
}
\arguments{
\item{df}{data.frame source object}

\item{name}{column name to extract}
}
\value{
integer vector
}
\description{
\code{.on_ice_int_col()} safely extracts an integer column from a data frame and
returns \code{NA} values when the column is absent.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nhlscraper-package.R
\docType{package}
\name{nhlscraper}
\alias{nhlscraper}
\alias{nhlscraper-package}
\title{nhlscraper: Scrape, clean, and visualize NHL data}
\description{
\code{nhlscraper} is a minimum-dependency R package to scrape, clean, and
visualize NHL data via the NHL and ESPN APIs. It primarily wraps 125+
\href{https://github.com/RentoSaijo/nhlscraper/wiki}{endpoints} from high-level
multi-season summaries and award winners to low-level decisecond replays and
bookmakers' odds, making them significantly more accessible. It also
features cleaning and visualization functions, primarily for play-by-plays,
to help analyze the data.
}
\section{Prerequisite}{

\itemize{
\item R/RStudio; you can check out my
\href{https://youtu.be/CUXzeuXJCPs?si=bkWFAm-hSXhUsxaN}{tutorial} if you are not
familiar.
}
}

\section{Disclosure}{

Detailed documentation for each scraping function will be released gradually
over time. Because each function can return a large amount of information,
users are encouraged to explore the provided examples to discover what is
available. Most, if not all, of the endpoints accessed by this package are
unofficially documented (i.e., hidden), so it is important to use them
responsibly and with respect for the NHL's data servers. Endpoints serving
historical or otherwise mostly static data should ideally be queried once and
then stored locally (for example, in a MySQL database) for further analysis,
rather than being called repeatedly; the
\href{https://rentosaijo.github.io/nhlscraper/reference/index.html#load}{load functions}
attempt to mitigate this risk by providing pre-scraped data, so please make
use of them. The exact rate limits for these APIs are not publicly known, so
users are asked to avoid excessive or abusive querying to help ensure
continued access for everyone.
}

\section{History}{

Prior to the NHL API rework in 2023, Drew Hynes documented a comprehensive
list of known \href{https://gitlab.com/dword4/nhlapi/}{endpoints}, and several R
packages, such as \href{https://github.com/jozefhajnala/nhlapi}{nhlapi} and
\href{https://github.com/danmorse314/hockeyR}{hockeyR}, were built to access them.
However, after the NHL completely transformed its API structure, all of these
packages became mostly defunct as their authors understandably chose not to
continue maintaining them. The community gathered around work by Zachary
Maludzinski to discover and share new
\href{https://github.com/Zmalski/NHL-API-Reference}{endpoints}, but progress
naturally slowed once most of the "main" endpoints were identified. Over the
summer of 2025, I began reverse-engineering many of the remaining,
undocumented endpoints while searching for additional data for future
research, with a particular focus on NHL EDGE and Records data. After I
shared those
\href{https://github.com/Zmalski/NHL-API-Reference/issues/69}{findings}, the
effort to expand and refine this map of the APIs accelerated, and with the
addition of David Fleischer's
\href{https://github.com/dfleis/nhl-api-docs}{effort}, we ultimately identified
400+ new endpoints. In parallel, I also discovered many new endpoints for
the ESPN API, extending beyond what Joseph Wilson had already
\href{https://github.com/pseudo-r/Public-ESPN-API}{compiled}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://rentosaijo.github.io/nhlscraper/}
  \item \url{https://github.com/RentoSaijo/nhlscraper}
  \item Report bugs at \url{https://github.com/RentoSaijo/nhlscraper/issues}
}

}
\author{
\strong{Maintainer}: Rento Saijo \email{rentosaijo0527@gmail.com} (\href{https://orcid.org/0009-0008-4919-7349}{ORCID}) [copyright holder]

Other contributors:
\itemize{
  \item Lars Skytte \email{lars.sunesen.skytte@gmail.com} [contributor]
  \item Jack Pallotta \email{jackjpallotta@gmail.com} (\href{https://orcid.org/0009-0004-0744-6977}{ORCID}) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{normalize_team_abbrev_cols}
\alias{normalize_team_abbrev_cols}
\title{Normalize team abbreviation columns to team tri-code names}
\usage{
normalize_team_abbrev_cols(x)
}
\arguments{
\item{x}{character vector}
}
\value{
character vector
}
\description{
Normalize team abbreviation columns to team tri-code names
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Skater.R
\name{skater_series_statistics}
\alias{skater_series_statistics}
\alias{skater_series_stats}
\title{Access the playoff statistics for all the skaters by series}
\usage{
skater_series_statistics()

skater_series_stats()
}
\value{
data.frame with one row per player per series
}
\description{
\code{skater_series_statistics()} retrieves the playoff statistics for all the skaters by series as a \code{data.frame} where each row represents player per series and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
}
\examples{
# May take >5s, so skip.
\donttest{skater_series_stats <- skater_series_statistics()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayerOld.R
\name{get_player_landing}
\alias{get_player_landing}
\title{Access the summary for a player}
\usage{
get_player_landing(player = 8478402)
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}
}
\value{
list with various items
}
\description{
\code{get_player_landing()} is deprecated. Use \code{\link[=player_summary]{player_summary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{season_now}
\alias{season_now}
\title{Access the season as of now}
\usage{
season_now()
}
\value{
integer in YYYYYYYY (e.g., 20242025)
}
\description{
\code{season_now()} retrieves the season as of now and returns a scalar integer used as the current-context default in season/game-type dependent wrappers.
}
\examples{
season_now <- season_now()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_logos}
\alias{team_logos}
\title{Access all the team logos}
\usage{
team_logos()
}
\value{
data.frame with one row per logo
}
\description{
\code{team_logos()} retrieves all the team logos as a \code{data.frame} where each row represents logo and includes detail on team identity, affiliation, and matchup-side context.
}
\examples{
all_team_logos <- team_logos()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GoalieEDGE.R
\name{goalie_edge_save_percentage}
\alias{goalie_edge_save_percentage}
\title{Access the EDGE save percentage statistics for a goalie, season, game type,
and category}
\usage{
goalie_edge_save_percentage(
  player = 8476945,
  season = "now",
  game_type = "",
  category = "details"
)
}
\arguments{
\item{player}{integer ID (e.g., 8478406)}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see
\code{\link[=goalie_edge_seasons]{goalie_edge_seasons()}} for reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
'playoff'/'post'; see \code{\link[=goalie_edge_seasons]{goalie_edge_seasons()}} for reference; most functions
will NOT support pre-season}

\item{category}{character of 'd'/'details' or 'l'/'l10'/'last 10'}
}
\value{
list with two items (category = 'details') or data.frame with one
row per game (category = 'last 10')
}
\description{
\code{goalie_edge_save_percentage()} retrieves the EDGE save percentage statistics for a goalie, season, game type, and category as a nested \code{list} that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
}
\examples{
Mackenzie_Blackwood_L10_sP_regular_20242025 <- 
  goalie_edge_save_percentage(
    player    = 8478406,
    season    = 20242025,
    game_type = 2,
    category  = 'L'
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OtherOld.R
\name{get_streams}
\alias{get_streams}
\title{Access all the streams}
\usage{
get_streams()
}
\value{
data.frame with one row per stream
}
\description{
\code{get_streams()} is deprecated. Use \code{\link[=streams]{streams()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{wsc_play_by_play_raw}
\alias{wsc_play_by_play_raw}
\alias{wsc_pbp_raw}
\title{Access the raw World Showcase (WSC) play-by-play for a game}
\usage{
wsc_play_by_play_raw(game = 2023030417)

wsc_pbp_raw(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
data.frame with one row per event (play)
}
\description{
\code{wsc_play_by_play_raw()} returns the raw flattened World Showcase play-by-play
as served by the NHL API for one game. Use \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for the
cleaned public schema that repairs common clock/order defects and appends the
derived public columns.
}
\examples{
wsc_raw_Martin_Necas_legacy_game <- wsc_play_by_play_raw(game = 2025020275)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_game_statistics}
\alias{goalie_game_statistics}
\alias{goalie_game_stats}
\title{Access the statistics for all the goalies by game}
\usage{
goalie_game_statistics()

goalie_game_stats()
}
\value{
data.frame with one row per goalie per game
}
\description{
\code{goalie_game_statistics()} retrieves the statistics for all the goalies by game as a \code{data.frame} with detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
\donttest{goalie_game_stats <- goalie_game_statistics()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNTeamOld.R
\name{get_espn_teams}
\alias{get_espn_teams}
\title{Access all the ESPN teams for a season}
\usage{
get_espn_teams()
}
\description{
\code{get_espn_teams()} is defunct. Use \code{\link[=espn_teams]{espn_teams()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterEDGE.R
\name{skater_edge_zone_time}
\alias{skater_edge_zone_time}
\title{Access the EDGE zone time statistics for a skater, season, game type, and
category}
\usage{
skater_edge_zone_time(
  player = 8478402,
  season = "now",
  game_type = "",
  category = "details"
)
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see
\code{\link[=skater_edge_seasons]{skater_edge_seasons()}} for reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
'playoff'/'post'; see \code{\link[=skater_edge_seasons]{skater_edge_seasons()}} for reference; most functions
will NOT support pre-season}

\item{category}{character of 'd'/'details' or 's'/'starts'}
}
\value{
data.frame with one row per strength state (category = 'details')
or list with six items (category = 'starts')
}
\description{
\code{skater_edge_zone_time()} retrieves the EDGE zone time statistics for a skater, season, game type, and category as a \code{data.frame} where each row represents strength state and includes detail on NHL EDGE style tracking outputs and relative-performance context.
}
\examples{
Martin_Necas_starts_regular_20242025 <- skater_edge_zone_time(
  player    = 8480039,
  season    = 20242025,
  game_type = 2,
  category  = 'S'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_team_statistics}
\alias{get_team_statistics}
\title{Access various reports for all the teams by season or game}
\usage{
get_team_statistics()
}
\description{
\code{get_team_statistics()} is defunct. Use \code{\link[=team_season_report]{team_season_report()}} and/or
\code{\link[=team_game_report]{team_game_report()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.sort_shift_chart_for_timing}
\alias{.sort_shift_chart_for_timing}
\title{Sort shift data for timing lookups}
\usage{
.sort_shift_chart_for_timing(shift_data)
}
\arguments{
\item{shift_data}{data.frame shift chart data}
}
\value{
filtered and sorted shift data.frame
}
\description{
\code{.sort_shift_chart_for_timing()} filters incomplete shift rows and sorts the
remainder into the order expected by the timing resolvers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Franchise.R
\name{franchise_statistics}
\alias{franchise_statistics}
\alias{franchise_stats}
\title{Access the all-time statistics for all the franchises by game type}
\usage{
franchise_statistics()

franchise_stats()
}
\value{
data.frame with one row per franchise per game type
}
\description{
\code{franchise_statistics()} retrieves the all-time statistics for all the franchises by game type as a \code{data.frame} where each row represents franchise per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
franchise_stats <- franchise_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGame.R
\name{espn_games}
\alias{espn_games}
\title{Access the ESPN games for a season}
\usage{
espn_games(season = season_now())
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
data.frame with one row per ESPN game
}
\description{
\code{espn_games()} retrieves the ESPN games for a season as a \code{data.frame} where each row represents ESPN and includes detail on game timing, matchup state, scoring flow, and situational event detail.
}
\examples{
ESPN_games_20242025 <- espn_games(season = 20242025)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.compute_on_ice_shift_timing_matrices}
\alias{.compute_on_ice_shift_timing_matrices}
\title{Compute on-ice shift timing matrices}
\usage{
.compute_on_ice_shift_timing_matrices(play_by_play, shift_data)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object with on-ice player IDs}

\item{shift_data}{data.frame shift chart data}
}
\value{
list containing home and away timing matrices
}
\description{
\code{.compute_on_ice_shift_timing_matrices()} uses the native timing resolver
when available and falls back to the R implementation otherwise.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.empty_public_pbp_column}
\alias{.empty_public_pbp_column}
\title{Allocate a missing public play-by-play column}
\usage{
.empty_public_pbp_column(name, n)
}
\arguments{
\item{name}{public play-by-play column name}

\item{n}{output row count}
}
\value{
typed vector of missing values
}
\description{
\code{.empty_public_pbp_column()} returns a typed \code{NA} vector for a public
play-by-play column that is absent in the upstream source data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{dot_to_camel}
\alias{dot_to_camel}
\title{Convert dot-delimited names to camelCase}
\usage{
dot_to_camel(x)
}
\arguments{
\item{x}{character vector}
}
\value{
character vector
}
\description{
Convert dot-delimited names to camelCase
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.fetch_html_pbp_on_ice}
\alias{.fetch_html_pbp_on_ice}
\title{Fetch HTML on-ice play-by-play rows}
\usage{
.fetch_html_pbp_on_ice(
  game,
  rosters,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev
)
}
\arguments{
\item{game}{game ID}

\item{rosters}{roster data.frame from play-by-play metadata}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}
}
\value{
data.frame of parsed HTML play-by-play rows
}
\description{
\code{.fetch_html_pbp_on_ice()} downloads an NHL HTML play-by-play report and
parses its on-ice content into a structured data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.drop_illogical_ordered_events}
\alias{.drop_illogical_ordered_events}
\title{Remove illogically ordered boundary events from a play-by-play}
\usage{
.drop_illogical_ordered_events(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with illogically ordered boundary events removed
}
\description{
\code{.drop_illogical_ordered_events()} removes stray boundary rows that can
appear between a period's \code{period-end} and the following period's
\code{period-start}. In practice, these are most often duplicated \code{0:00}
\code{faceoff}/\code{stoppage} rows that are still tagged to the previous period even
though the actual opening sequence for the next period follows immediately
after.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Skater.R
\name{skater_milestones}
\alias{skater_milestones}
\title{Access the skaters on milestone watch}
\usage{
skater_milestones()
}
\value{
data.frame with one row per player
}
\description{
\code{skater_milestones()} retrieves the skaters on milestone watch as a \code{data.frame} where each row represents player and includes detail on date/season filtering windows and chronological context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
skater_milestones <- skater_milestones()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{schedule}
\alias{schedule}
\title{Access the schedule for a date}
\usage{
schedule(date = Sys.Date())
}
\arguments{
\item{date}{character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see
\code{\link[=seasons]{seasons()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{schedule()} retrieves the schedule for a date as a \code{data.frame} where each row represents game and includes detail on game timing, matchup state, scoring flow, and situational event detail.
}
\examples{
schedule_Halloween_2025 <- schedule(date = '2025-10-31')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.add_on_ice_shift_timing_context}
\alias{.add_on_ice_shift_timing_context}
\title{Add on-ice shift timing context}
\usage{
.add_on_ice_shift_timing_context(play_by_play, game, shift_data = NULL)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object with on-ice player IDs}

\item{game}{game ID}

\item{shift_data}{optional shift chart data.frame}
}
\value{
data.frame with scalar on-ice timing columns added
}
\description{
\code{.add_on_ice_shift_timing_context()} enriches play-by-play rows with scalar
shift elapsed and since-last-shift columns for the on-ice goalies and
skaters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Franchise.R
\name{franchises}
\alias{franchises}
\title{Access all the franchises}
\usage{
franchises()
}
\value{
data.frame with one row per franchise
}
\description{
\code{franchises()} retrieves all the franchises as a \code{data.frame} where each row represents franchise and includes detail on team identity, affiliation, and matchup-side context.
}
\examples{
all_franchises <- franchises()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{.perform_parallel_requests}
\alias{.perform_parallel_requests}
\title{Perform multiple requests concurrently}
\usage{
.perform_parallel_requests(reqs, on_error = c("stop", "return", "continue"))
}
\arguments{
\item{reqs}{list of httr2 request objects}

\item{on_error}{forwarded to \code{\link[httr2:req_perform_parallel]{httr2::req_perform_parallel()}}}
}
\value{
list of responses or httr2 failure objects
}
\description{
\code{.perform_parallel_requests()} executes a list of \code{httr2} requests with
libcurl multi support and preserves the input names on the output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{draft_picks}
\alias{draft_picks}
\title{Access all the draft picks}
\usage{
draft_picks()
}
\value{
data.frame with one row per pick
}
\description{
\code{draft_picks()} retrieves all the draft picks as a \code{data.frame} where each row represents pick and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
}
\examples{
# May take >5s, so skip.
\donttest{all_draft_picks <- draft_picks()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGame.R
\name{espn_game_odds}
\alias{espn_game_odds}
\title{Access the ESPN odds for a game}
\usage{
espn_game_odds(game = 401777460)
}
\arguments{
\item{game}{integer ID (e.g., 401777460); see \code{\link[=espn_games]{espn_games()}} for
reference}
}
\value{
data.frame with one row per provider
}
\description{
\code{espn_game_odds()} retrieves the ESPN odds for a game as a \code{data.frame} where each row represents provider and includes detail on team identity, affiliation, and matchup-side context plus betting market snapshots with side/total prices and provider variation.
}
\examples{
ESPN_odds_SCF_20242025 <- espn_game_odds(game = 401777460)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Load.R
\name{replays}
\alias{replays}
\title{Access the replays for a season}
\usage{
replays(season = 20242025)
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
data.frame with one row per decisecond
}
\description{
\code{replays()} loads the replays for a given \code{season}.
}
\examples{
# May take >5s, so skip.
\donttest{replays_20252026 <- replays(season = 20252026)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.parse_html_elapsed_clock}
\alias{.parse_html_elapsed_clock}
\title{Parse an HTML elapsed clock}
\usage{
.parse_html_elapsed_clock(x)
}
\arguments{
\item{x}{HTML time field}
}
\value{
integer seconds elapsed in period or \code{NA_integer_}
}
\description{
\code{.parse_html_elapsed_clock()} extracts the elapsed \code{MM:SS} clock from an HTML
report time field and converts it to elapsed seconds in period.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.public_pbp_legal_period_seconds}
\alias{.public_pbp_legal_period_seconds}
\title{Derive the legal period length in seconds}
\usage{
.public_pbp_legal_period_seconds(game_id, game_type_id, period)
}
\arguments{
\item{game_id}{integer game IDs}

\item{game_type_id}{integer game type IDs}

\item{period}{integer period numbers}
}
\value{
integer vector of legal period lengths in seconds
}
\description{
\code{.public_pbp_legal_period_seconds()} returns the nominal legal length for each
play-by-play row's period under the relevant game context.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.has_any_on_ice_player_ids}
\alias{.has_any_on_ice_player_ids}
\title{Identify populated on-ice player rows}
\usage{
.has_any_on_ice_player_ids(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}
}
\value{
logical vector
}
\description{
\code{.has_any_on_ice_player_ids()} returns a logical vector indicating whether a
play-by-play row already has any scalar on-ice player IDs assigned.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{scores}
\alias{scores}
\title{Access the scores for a date}
\usage{
scores(date = "now")
}
\arguments{
\item{date}{character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see
\code{\link[=seasons]{seasons()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{scores()} retrieves the scores for a date as a \code{data.frame} where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
scores_Halloween_2025 <- scores(date = '2025-10-31')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNPlayerOld.R
\name{get_espn_athlete}
\alias{get_espn_athlete}
\title{Access the ESPN summary for an athlete (player) and season}
\usage{
get_espn_athlete()
}
\description{
\code{get_espn_athlete()} is defunct. Use \code{\link[=espn_player_summary]{espn_player_summary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{nhl_api}
\alias{nhl_api}
\title{Call the NHL API with 429 (rate limit) error-handling}
\usage{
nhl_api(path, query = list(), type)
}
\arguments{
\item{path}{character}

\item{query}{list}

\item{type}{character of 'w' for web, 's' for stats, and 'r' for records}
}
\value{
parsed JSON (i.e., data.frame or list)
}
\description{
Call the NHL API with 429 (rate limit) error-handling
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{standings_rules}
\alias{standings_rules}
\title{Access the standings rules by season}
\usage{
standings_rules()
}
\value{
data.frame with one row per season
}
\description{
\code{standings_rules()} retrieves the standings rules by season as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context.
}
\examples{
standings_rules <- standings_rules()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.ensure_local_native_symbol}
\alias{.ensure_local_native_symbol}
\title{Ensure a native symbol is available}
\usage{
.ensure_local_native_symbol(symbol)
}
\arguments{
\item{symbol}{native symbol name}
}
\value{
logical scalar indicating whether the symbol is available
}
\description{
\code{.ensure_local_native_symbol()} checks whether a native symbol is already
registered and, if needed, tries to load the installed package shared object.
It does not trust the current working directory.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{game_rosters}
\alias{game_rosters}
\title{Access the rosters for a game}
\usage{
game_rosters(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
data.frame with one row per player
}
\description{
\code{game_rosters()} retrieves the rosters for a game as a \code{data.frame} where each row represents player and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
}
\examples{
rosters_Martin_Necas_legacy_game <- game_rosters(game = 2025020275)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_series_statistics}
\alias{goalie_series_statistics}
\alias{goalie_series_stats}
\title{Access the playoff statistics for all the goalies by series}
\usage{
goalie_series_statistics()

goalie_series_stats()
}
\value{
data.frame with one row per player per series
}
\description{
\code{goalie_series_statistics()} retrieves the playoff statistics for all the goalies by series as a \code{data.frame} where each row represents player per series and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
}
\examples{
goalie_series_stats <- goalie_series_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{boxscore}
\alias{boxscore}
\title{Access the boxscore for a game, team, and position}
\usage{
boxscore(game = 2023030417, team = "home", position = "forwards")
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}

\item{team}{character of 'h'/'home' or 'a'/'away'}

\item{position}{character of 'f'/'forwards', 'd'/'defensemen', or
'g'/'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{boxscore()} retrieves the boxscore for a game, team, and position as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus production, workload, efficiency, and result-level performance outcomes.
}
\examples{
boxscore_COL_forwards_Martin_Necas_legacy_game <- boxscore(
  game     = 2025020275,
  team     = 'H',
  position = 'F'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Graph.R
\name{draw_rink_arc}
\alias{draw_rink_arc}
\title{Draw an arc on the rink}
\usage{
draw_rink_arc(cx, cy, r, theta1, theta2, n = 100, col = "black", lwd = 1)
}
\arguments{
\item{cx}{numeric x coordinate of the circle center}

\item{cy}{numeric y coordinate of the circle center}

\item{r}{numeric radius}

\item{theta1}{numeric start angle of arc}

\item{theta2}{numeric end angle of arc}

\item{n}{integer number of points used to approximate arc}

\item{col}{character color}

\item{lwd}{numeric line width}
}
\value{
\code{NULL}
}
\description{
\code{draw_rink_arc()} draws an arc on the rink.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_extract_tagged_players}
\alias{.html_extract_tagged_players}
\title{Extract team-tagged player references from HTML}
\usage{
.html_extract_tagged_players(
  description,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev,
  roster_lookup
)
}
\arguments{
\item{description}{HTML event description}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}

\item{roster_lookup}{standardized roster lookup data.frame}
}
\value{
data.frame of ordered team/player references
}
\description{
\code{.html_extract_tagged_players()} resolves explicitly team-tagged HTML player
references such as \code{TOR #42} into ordered team/player pairs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.on_ice_scalar_column_names}
\alias{.on_ice_scalar_column_names}
\title{List all scalar on-ice columns}
\usage{
.on_ice_scalar_column_names(play_by_play = NULL, slot_count = NULL)
}
\arguments{
\item{play_by_play}{optional data.frame whose existing on-ice columns should
be inspected}

\item{slot_count}{optional integer scalar minimum slot count requested by the
caller}
}
\value{
character vector of scalar on-ice columns
}
\description{
\code{.on_ice_scalar_column_names()} returns the scalar on-ice player ID and
shift-timing columns used by the public play-by-play schema.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GoalieOld.R
\name{get_goalies}
\alias{get_goalies}
\title{Access all the goalies for a range of seasons}
\usage{
get_goalies()
}
\description{
\code{get_goalies()} is defunct. Use \code{\link[=players]{players()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Model.R
\name{calculate_expected_goals}
\alias{calculate_expected_goals}
\alias{calculate_xG}
\title{Calculate the expected goals for all the shots in (a) play-by-plays}
\usage{
calculate_expected_goals(play_by_play, model = NULL)

calculate_xG(play_by_play, model = NULL)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public
schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}},
\code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}. Legacy alias-only columns
such as \code{typeDescKey}, \code{period}, \code{SOGFor}, \code{SOGAgainst}, and
\code{SOGDifferential} are no longer backfilled by the xG scorer.}

\item{model}{deprecated legacy model selector; ignored}
}
\value{
data.frame with one row per event (play) and added \code{xG} column
}
\description{
\code{calculate_expected_goals()} scores shot events with \code{nhlscraper}'s built-in
ridge expected-goals model. The runtime model is a fixed six-partition system:
\code{sd} (5v5), \code{ev} (other even strength), \code{pp} (power play), \code{sh}
(short-handed), \code{en} (empty net against), and \code{ps} (penalty shot; trained on
penalty-shot and shootout-style rows). The legacy \code{model} argument is
accepted for backward compatibility but ignored.
}
\examples{
# May take >5s, so skip.
\donttest{
  pbp <- gc_play_by_play()
  pbp_with_xg <- calculate_expected_goals(play_by_play = pbp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.override_strength_context_from_html}
\alias{.override_strength_context_from_html}
\title{Override derived strength context from resolved on-ice players}
\usage{
.override_strength_context_from_html(
  play_by_play,
  idx,
  home_goalie,
  away_goalie,
  home_skaters,
  away_skaters
)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{idx}{row index to update}

\item{home_goalie}{resolved home goalie player ID}

\item{away_goalie}{resolved away goalie player ID}

\item{home_skaters}{integer vector of resolved home skater IDs}

\item{away_skaters}{integer vector of resolved away skater IDs}
}
\value{
data.frame with derived strength context updated for \code{idx}
}
\description{
\code{.override_strength_context_from_html()} updates the derived strength/count
columns from a resolved HTML on-ice signature while leaving the raw
\code{situationCode} column untouched.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_season_statistics}
\alias{goalie_season_statistics}
\alias{goalie_season_stats}
\title{Access the statistics for all the goalies by season, game type, and team.}
\usage{
goalie_season_statistics()

goalie_season_stats()
}
\value{
data.frame with one row per player per season per game type,
separated by team if applicable
}
\description{
\code{goalie_season_statistics()} retrieves the statistics for all the goalies by season, game type, and team as a \code{data.frame} where each row represents player per season per game type, separated by team if applicable and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
}
\examples{
goalie_season_stats <- goalie_season_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FranchiseOld.R
\name{get_franchise_vs_franchise}
\alias{get_franchise_vs_franchise}
\title{Access the all-time statistics versus other franchises for all the
franchises by game type}
\usage{
get_franchise_vs_franchise()
}
\value{
data.frame with one row per franchise per franchise per game type
}
\description{
\code{get_franchise_vs_franchise()} is deprecated. Use
\code{\link[=franchise_versus_franchise]{franchise_versus_franchise()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.require_public_pbp_columns}
\alias{.require_public_pbp_columns}
\title{Validate required public play-by-play columns}
\usage{
.require_public_pbp_columns(play_by_play, cols, fn_name)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{cols}{character vector of required columns}

\item{fn_name}{calling function name used in the error message}
}
\value{
\code{NULL}, invisibly, or an error if required columns are missing
}
\description{
\code{.require_public_pbp_columns()} errors when a public helper is called with a
play-by-play missing required public schema columns.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_team_scoreboard}
\alias{get_team_scoreboard}
\title{Access the team scoreboard as of now}
\usage{
get_team_scoreboard()
}
\description{
\code{get_team_scoreboard()} is defunct.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{draft_prospects}
\alias{draft_prospects}
\title{Access all the draft prospects}
\usage{
draft_prospects()
}
\value{
data.frame with one row per player
}
\description{
\code{draft_prospects()} retrieves all the draft prospects as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus broadcast carriage, media availability, and viewing-link metadata.
}
\examples{
# May take >5s, so skip.
\donttest{all_prospects <- draft_prospects()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Share.R
\name{ig_game_cumulative_expected_goals}
\alias{ig_game_cumulative_expected_goals}
\alias{ig_game_cum_xG}
\title{Save an Instagram (IG) share-able cumulative expected goals (xG) time-series
plot for a game}
\usage{
ig_game_cumulative_expected_goals(game = 2023030417, model = NULL, save = TRUE)

ig_game_cum_xG(game = 2023030417, model = NULL)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}

\item{model}{deprecated legacy model selector; ignored}

\item{save}{logical only FALSE for tests}
}
\value{
\code{NULL}
}
\description{
\code{ig_game_cumulative_expected_goals()} saves an IG share-able cumulative xG
time-series plot for a given \code{game} as a PNG.
}
\examples{
# May take >5s, so skip.
\donttest{ig_game_cumulative_expected_goals(
  game  = 2023030417, 
  save  = FALSE
)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{roster}
\alias{roster}
\title{Access the roster for a team, season, and position}
\usage{
roster(team = 1, season = "current", position = "forwards")
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for
reference; ID is preferable as there now exists duplicate three-letter codes
(i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{position}{character of 'f'/'forwards', 'd'/'defensemen', or
'g'/'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{roster()} retrieves the roster for a team, season, and position as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
}
\examples{
COL_defensemen_20242025 <- roster(
  team     = 21,
  season   = 20242025,
  position = 'D'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{add_deltas}
\alias{add_deltas}
\title{Add event-to-event deltas to a play-by-play}
\usage{
add_deltas(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}}, \code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}}
}
\value{
data.frame with one row per event (play) and added columns: \code{eventIdPrev}, \code{secondsElapsedInSequence}, \code{dSecondsElapsedInSequence}, \code{dXCoord}, \code{dYCoord}, \code{dXCoordNorm}, \code{dYCoordNorm}, \code{dDistance}, \code{dAngle}, \code{dXCoordPerSecond}, \code{dYCoordPerSecond}, \code{dXCoordNormPerSecond}, \code{dYCoordNormPerSecond}, \code{dDistancePerSecond}, \code{dAnglePerSecond}
}
\description{
\code{add_deltas()} adds event-to-event deltas in raw and normalized x/y,
distance, and angle for a play-by-play. Sequences are bounded by faceoffs:
each sequence begins at a faceoff, faceoff rows do not look backward across
the boundary, and subsequent events are compared to the most recent prior
valid-spatial event in the same faceoff-bounded sequence. Shootout and
penalty-shot rows (\code{0101}/\code{1010}) are left as \code{NA} and do not serve as
anchors for later rows. When multiple events in a sequence share the same
recorded second, zero-time denominators used internally are replaced by
\code{1 / n}, where \code{n} is the number of events in that same second within the
sequence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.repair_public_pbp_sequence}
\alias{.repair_public_pbp_sequence}
\title{Repair public play-by-play clock/order defects in boundary-row feeds}
\usage{
.repair_public_pbp_sequence(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with repaired clocks and ordering where feasible
}
\description{
\code{.repair_public_pbp_sequence()} removes rows with impossible clocks and
reorders explicit-boundary feeds into a more coherent public sequence. The
repair is intentionally scoped to the modern boundary-row era, where the
audited issues are dominated by stale sort-order markers, early \code{period-end}
rows, and opening faceoffs that arrive after live-play rows.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.build_game_roster_lookup}
\alias{.build_game_roster_lookup}
\title{Build a game-roster lookup table}
\usage{
.build_game_roster_lookup(rosters)
}
\arguments{
\item{rosters}{roster data.frame from the game play-by-play metadata}
}
\value{
standardized roster lookup data.frame
}
\description{
\code{.build_game_roster_lookup()} standardizes roster columns and prepares a
lookup table used to resolve HTML sweater numbers to player IDs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{game_odds}
\alias{game_odds}
\title{Access the real-time game odds for a country by partnered bookmaker}
\usage{
game_odds(country = "US")
}
\arguments{
\item{country}{two-letter code (e.g., 'CA'); see \code{\link[=countries]{countries()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{game_odds()} retrieves the real-time game odds for a country by partnered bookmaker as a \code{data.frame} where each row represents game and includes detail on betting market lines, prices, and provider-level context.
}
\examples{
game_odds_CA <- game_odds(country = 'CA')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.apply_pbp_delta_columns}
\alias{.apply_pbp_delta_columns}
\title{Attach delta context columns to play-by-play}
\usage{
.apply_pbp_delta_columns(play_by_play, delta_ctx)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}}, \code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}}

\item{delta_ctx}{named list returned by \code{.compute_pbp_deltas()}}
}
\value{
data.frame with public delta columns inserted
}
\description{
\code{.apply_pbp_delta_columns()} inserts computed delta columns into a
play-by-play and reorders them into the public schema.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeagueOld.R
\name{get_attendance}
\alias{get_attendance}
\title{Access the attendance by season and game type}
\usage{
get_attendance()
}
\value{
data.frame with one row per season
}
\description{
\code{get_attendance()} is deprecated. Use \code{\link[=attendance]{attendance()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{game_type_now}
\alias{game_type_now}
\title{Access the game type as of now}
\usage{
game_type_now()
}
\value{
integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season)
}
\description{
\code{game_type_now()} retrieves the game type as of now and returns a scalar integer used as the current-context default in season/game-type dependent wrappers.
}
\examples{
game_type_now <- game_type_now()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNCoachOld.R
\name{get_espn_coaches}
\alias{get_espn_coaches}
\title{Access the ESPN coaches for a season}
\usage{
get_espn_coaches()
}
\description{
\code{get_espn_coaches()} is defunct. Use \code{\link[=coaches]{coaches()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{add_shooter_biometrics}
\alias{add_shooter_biometrics}
\title{Add shooter biometrics to (a) play-by-play(s)}
\usage{
add_shooter_biometrics(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}}, \code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}}
}
\value{
data.frame with one row per event (play) and added columns: \code{shooterHeight}, \code{shooterWeight}, \code{shooterHandCode}, \code{shooterAge}, and \code{shooterPositionCode}
}
\description{
\code{add_shooter_biometrics()} adds shooter biometrics (height, weight, hand, age at game date, and position) to (a) play-by-play(s) for shot attempts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Player.R
\name{player_game_log}
\alias{player_game_log}
\title{Access the game log for a player, season, and game type}
\usage{
player_game_log(player = 8478402, season = "now", game_type = "")
}
\arguments{
\item{player}{integer ID (e.g., 8480039); see \code{\link[=players]{players()}} for reference}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}
}
\value{
data.frame with one row per game
}
\description{
\code{player_game_log()} retrieves the game log for a player, season, and game type as a \code{data.frame} where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow plus production, workload, efficiency, and result-level performance outcomes.
}
\examples{
Martin_Necas_game_log_regular_20242025 <- player_game_log(
  player    = 8480039,
  season    = 20242025,
  game_type = 2
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.normalize_coordinates}
\alias{.normalize_coordinates}
\title{Normalize the x and y coordinates for all the events (plays) in a play-by-play}
\usage{
.normalize_coordinates(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with one row per event (play) and added columns \code{xCoordNorm} and \code{yCoordNorm}
}
\description{
\code{.normalize_coordinates()} normalizes the x and y coordinates for all the events (plays) in a play-by-play such that they all attack towards +x. If \code{homeTeamDefendingSide} is not available, the home defending side in period 1 is inferred using \code{zoneCode}, \code{isHome}, and \code{xCoord}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_prospects}
\alias{team_prospects}
\title{Access the prospects for a team and position}
\usage{
team_prospects(team = 1, position = "forwards")
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for
reference; ID is preferable as there now exists duplicate three-letter codes
(i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')}

\item{position}{character of 'f'/'forwards', 'd'/'defensemen', or
'g'/'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{team_prospects()} retrieves the prospects for a team and position as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
}
\examples{
COL_forward_prospects <- team_prospects(
  team     = 21,
  position = 'F'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Franchise.R
\name{franchise_playoff_situational_results}
\alias{franchise_playoff_situational_results}
\title{Access the playoff series results for all the franchises by situation}
\usage{
franchise_playoff_situational_results()
}
\value{
data.frame with one row per franchise per situation
}
\description{
\code{franchise_playoff_situational_results()} retrieves the playoff series results for all the franchises by situation as a \code{data.frame} where each row represents franchise per situation and includes detail on team identity, affiliation, and matchup-side context.
}
\examples{
franchise_playoff_situational_results <- 
  franchise_playoff_situational_results()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OfficialOld.R
\name{get_officials}
\alias{get_officials}
\title{Access all the officials}
\usage{
get_officials()
}
\value{
data.frame with one row per official
}
\description{
\code{get_officials()} is deprecated. Use \code{\link[=officials]{officials()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.build_api_html_match_table}
\alias{.build_api_html_match_table}
\title{Build API-side HTML matching features}
\usage{
.build_api_html_match_table(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}
}
\value{
data.frame of API-side matching features
}
\description{
\code{.build_api_html_match_table()} derives the API-side matching features used
to align HTML play-by-play rows back to API play-by-play rows.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Event.R
\name{penalty_shots}
\alias{penalty_shots}
\alias{pss}
\title{Access all the penalty shots}
\usage{
penalty_shots()

pss()
}
\value{
data.frame with one row per penalty shot
}
\description{
\code{penalty_shots()} retrieves all the penalty shots as a \code{data.frame} where each row represents penalty shot and includes detail on game timeline state, period/clock progression, and matchup flow plus date/season filtering windows and chronological context.
}
\examples{
all_pss <- penalty_shots()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNPlayerOld.R
\name{get_espn_athletes}
\alias{get_espn_athletes}
\title{Access all the ESPN athletes (players)}
\usage{
get_espn_athletes()
}
\value{
data.frame with one row per ESPN athlete (player)
}
\description{
\code{get_espn_athletes()} is deprecated. Use \code{\link[=espn_players]{espn_players()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.finalize_pbp_output}
\alias{.finalize_pbp_output}
\title{Finalize public play-by-play output}
\usage{
.finalize_pbp_output(play_by_play, source = c("gc", "wsc"))
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{source}{output source, either \code{"gc"} or \code{"wsc"}}
}
\value{
data.frame with the finalized public play-by-play schema
}
\description{
\code{.finalize_pbp_output()} selects and orders the final GC or WSC public
play-by-play column set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterOld.R
\name{get_skater_statistics}
\alias{get_skater_statistics}
\title{Access various reports for all the skaters by season or game}
\usage{
get_skater_statistics()
}
\description{
\code{get_skater_statistics()} is defunct. Use \code{\link[=skater_season_report]{skater_season_report()}} or
\code{\link[=skater_game_report]{skater_game_report()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FranchiseOld.R
\name{get_franchise_team_totals}
\alias{get_franchise_team_totals}
\title{Access the all-time statistics for all the franchises by team and game type}
\usage{
get_franchise_team_totals()
}
\value{
data.frame with one row per team per franchise per game type
}
\description{
\code{get_franchise_team_totals()} is deprecated. Use
\code{\link[=franchise_team_statistics]{franchise_team_statistics()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNTeam.R
\name{espn_team_summary}
\alias{espn_team_summary}
\title{Access the ESPN summary for a team}
\usage{
espn_team_summary(team = 1)
}
\arguments{
\item{team}{integer ID (e.g., 1); see \code{\link[=espn_teams]{espn_teams()}} for
reference}
}
\value{
data.frame with one row
}
\description{
\code{espn_team_summary()} retrieves the ESPN summary for a team as a \code{data.frame} where each row represents one result and includes detail on game timing, matchup state, scoring flow, and situational event detail.
}
\examples{
ESPN_summary_Boston_Bruins <- espn_team_summary(team = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{draft_rankings}
\alias{draft_rankings}
\title{Access the draft rankings for a class and category}
\usage{
draft_rankings(class = season_now()\%\%10000, category = 1)
}
\arguments{
\item{class}{integer in YYYY (e.g., 2017); see \code{\link[=drafts]{drafts()}} for reference}

\item{category}{integer in 1:4 (where 1 = North American Skaters,
2 = International Skaters, 3 = North American Goalies, and 4 = International
Goalies) OR character of 'NAS'/'NA Skaters'/'North American Skaters',
'INTLS'/'INTL Skaters'/'International Skaters',
'NAG'/'NA Goalies'/'North American Goalies',
'INTLG'/'INTL Goalies'/'International Goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{draft_rankings()} retrieves the draft rankings for a class and category as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile plus draft-board context, scouting background, and pick/round progression.
}
\examples{
draft_rankings_INTL_Skaters_2017 <- draft_rankings(
  class    = 2017, 
  category = 2
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeagueOld.R
\name{get_seasons}
\alias{get_seasons}
\title{Access all the seasons}
\usage{
get_seasons()
}
\value{
data.frame with one row per season
}
\description{
\code{get_seasons()} is deprecated. Use \code{\link[=seasons]{seasons()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.apply_shot_context}
\alias{.apply_shot_context}
\title{Attach shot context columns to play-by-play}
\usage{
.apply_shot_context(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}
}
\value{
data.frame with shot context columns added
}
\description{
\code{.apply_shot_context()} appends rush, rebound, and running score/attempt
context columns to a play-by-play.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{teams}
\alias{teams}
\title{Access all the teams}
\usage{
teams()
}
\value{
data.frame with one row per team
}
\description{
\code{teams()} retrieves all the teams as a \code{data.frame} where each row represents team and includes detail on team identity, affiliation, and matchup-side context.
}
\examples{
all_teams <- teams()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{add_shift_times}
\alias{add_shift_times}
\title{Add on-ice shift times to a play-by-play}
\usage{
add_shift_times(play_by_play, shift_chart)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s) using the current public schema returned by \code{\link[=gc_play_by_play]{gc_play_by_play()}}, \code{\link[=gc_play_by_plays]{gc_play_by_plays()}}, \code{\link[=wsc_play_by_play]{wsc_play_by_play()}}, or \code{\link[=wsc_play_by_plays]{wsc_play_by_plays()}}}

\item{shift_chart}{data.frame returned by \code{\link[=shift_chart]{shift_chart()}} or \code{\link[=shift_charts]{shift_charts()}}}
}
\value{
data.frame with one row per event (play) and added or updated scalar on-ice shift timing columns
}
\description{
\code{add_shift_times()} adds \code{SecondsRemainingInShift},
\code{SecondsElapsedInShift}, and \code{SecondsElapsedInPeriodSinceLastShift} columns
for the on-ice goalies and skaters already present in a public play-by-play.
It accepts either a single game play-by-play plus \code{\link[=shift_chart]{shift_chart()}} data or a
season aggregate plus \code{\link[=shift_charts]{shift_charts()}} data.
}
\examples{
\donttest{
  pbp <- gc_play_by_play(game = 2023030417)
  sc  <- shift_chart(game = 2023030417)
  pbp <- add_shift_times(pbp, sc)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Other.R
\name{glossary}
\alias{glossary}
\title{Access the glossary}
\usage{
glossary()
}
\value{
data.frame with one row per terminology
}
\description{
\code{glossary()} retrieves the glossary as a \code{data.frame} where each row represents terminology and includes detail on reference definitions and rules-framework information.
}
\examples{
glossary <- glossary()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.pbp_delta_public_cols}
\alias{.pbp_delta_public_cols}
\title{List public delta column names}
\usage{
.pbp_delta_public_cols()
}
\value{
character vector of public delta column names
}
\description{
\code{.pbp_delta_public_cols()} returns the public-facing delta columns inserted
by \code{\link[=add_deltas]{add_deltas()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{normalize_team_key}
\alias{normalize_team_key}
\title{Normalize the team key}
\usage{
normalize_team_key(team)
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL')}
}
\value{
integer in 1:68, character full name, OR three-letter code
}
\description{
Normalize the team key
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{.nhl_json_from_response}
\alias{.nhl_json_from_response}
\title{Parse an NHL API response as JSON}
\usage{
.nhl_json_from_response(resp)
}
\arguments{
\item{resp}{httr2 response object}
}
\value{
parsed JSON (i.e., data.frame or list)
}
\description{
\code{.nhl_json_from_response()} converts an \code{httr2} response object into a parsed
JSON object using the package's standard flattening settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.empty_pbp_delta_context}
\alias{.empty_pbp_delta_context}
\title{Build an empty play-by-play delta context}
\usage{
.empty_pbp_delta_context(n)
}
\arguments{
\item{n}{integer row count}
}
\value{
named list of empty delta vectors
}
\description{
\code{.empty_pbp_delta_context()} allocates an empty delta context sized to the
number of play-by-play rows.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_milestones}
\alias{goalie_milestones}
\title{Access the goalies on milestone watch}
\usage{
goalie_milestones()
}
\value{
data.frame with one row per player
}
\description{
\code{goalie_milestones()} retrieves the goalies on milestone watch as a \code{data.frame} where each row represents player and includes detail on date/season filtering windows and chronological context, player identity, role, handedness, and biographical profile, and ranking movement, points pace, and division/conference position signals.
}
\examples{
goalie_milestones <- goalie_milestones()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNOtherOld.R
\name{get_espn_transactions}
\alias{get_espn_transactions}
\title{Access the ESPN transactions by start and end dates}
\usage{
get_espn_transactions()
}
\description{
\code{get_espn_transactions()} is defunct. Use \code{\link[=espn_transactions]{espn_transactions()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.shot_event_mask}
\alias{.shot_event_mask}
\title{Build a shot-event mask}
\usage{
.shot_event_mask(play_by_play, types)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}

\item{types}{character vector of event type keys to keep}
}
\value{
logical vector
}
\description{
\code{.shot_event_mask()} flags play-by-play rows matching selected event types.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EventOld.R
\name{ps}
\alias{ps}
\title{Access all the penalty shots}
\usage{
ps()
}
\value{
data.frame with one row per penalty shot
}
\description{
\code{ps()} is deprecated. Use \code{\link[=penalty_shots]{penalty_shots()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGameOld.R
\name{get_espn_event}
\alias{get_espn_event}
\title{Access the ESPN summary for an event (game)}
\usage{
get_espn_event(event = 401777460)
}
\arguments{
\item{event}{integer ID (e.g., 401777460); see \code{\link[=espn_games]{espn_games()}} for
reference}
}
\value{
data.frame with one row per event (game)
}
\description{
\code{get_espn_event()} is deprecated. Use \code{\link[=espn_game_summary]{espn_game_summary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OtherOld.R
\name{get_configuration}
\alias{get_configuration}
\title{Access the configurations for team, skater, and goalie reports}
\usage{
get_configuration()
}
\description{
\code{get_configuration()} is defunct. Use \code{\link[=team_report_configurations]{team_report_configurations()}},
\code{\link[=skater_report_configurations]{skater_report_configurations()}}, and/or \code{\link[=goalie_report_configurations]{goalie_report_configurations()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.on_ice_skater_slots}
\alias{.on_ice_skater_slots}
\title{Return the tracked number of on-ice skater slots}
\usage{
.on_ice_skater_slots(play_by_play = NULL, slot_count = NULL, min_slots = 5L)
}
\arguments{
\item{play_by_play}{optional data.frame whose existing on-ice columns should
be inspected}

\item{slot_count}{optional integer scalar minimum slot count requested by the
caller}

\item{min_slots}{integer scalar default floor for standard outputs}
}
\value{
integer scalar
}
\description{
\code{.on_ice_skater_slots()} returns the current number of skater slots tracked
per team in play-by-play outputs. It guarantees the standard five skater
columns and expands further when the input already contains overflow skater
slots or when a caller requests a larger slot count.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FranchiseOld.R
\name{get_franchises}
\alias{get_franchises}
\title{Access all the franchises}
\usage{
get_franchises()
}
\value{
data.frame with one row per franchise
}
\description{
\code{get_franchises()} is deprecated. Use \code{\link[=franchises]{franchises()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GameOld.R
\name{get_scoreboards}
\alias{get_scoreboards}
\title{Access the scoreboards for a date}
\usage{
get_scoreboards(date = "now")
}
\arguments{
\item{date}{character in 'YYYY-MM-DD' (e.g., '2025-01-01'); see
\code{\link[=seasons]{seasons()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{get_scoreboards()} is deprecated. Use \code{\link[=scores]{scores()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterEDGE.R
\name{skater_edge_seasons}
\alias{skater_edge_seasons}
\title{Access the season(s) and game type(s) in which there exists skater EDGE
statistics}
\usage{
skater_edge_seasons()
}
\value{
data.frame with one row per season
}
\description{
\code{skater_edge_seasons()} retrieves the season(s) and game type(s) in which there exists skater EDGE statistics as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context plus NHL EDGE style tracking outputs and relative-performance context.
}
\examples{
skater_EDGE_seasons <- skater_edge_seasons()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Team.R
\name{team_month_schedule}
\alias{team_month_schedule}
\title{Access the schedule for a team and month}
\usage{
team_month_schedule(team = 1, month = "now")
}
\arguments{
\item{team}{integer ID (e.g., 21), character full name (e.g., 'Colorado
Avalanche'), OR three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for
reference; ID is preferable as there now exists duplicate three-letter codes
(i.e., 'UTA' for 'Utah Hockey Club' and 'Utah Mammoth')}

\item{month}{character in 'YYYY-MM' (e.g., '2025-01'); see \code{\link[=seasons]{seasons()}}
for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{team_month_schedule()} retrieves the schedule for a team and month as a \code{data.frame} where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
COL_schedule_December_2025 <- team_month_schedule(
  team  = 21, 
  month = '2025-12'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Other.R
\name{countries}
\alias{countries}
\title{Access all the countries}
\usage{
countries()
}
\value{
data.frame with one row per country
}
\description{
\code{countries()} retrieves all the countries as a \code{data.frame} where each row represents country and includes detail on reference metadata, regional context, and media availability detail.
}
\examples{
all_countries <- countries()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_teams}
\alias{get_teams}
\title{Access all the teams}
\usage{
get_teams()
}
\value{
data.frame with one row per team
}
\description{
\code{get_teams()} is deprecated. Use \code{\link[=teams]{teams()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayoffOld.R
\name{get_series_schedule}
\alias{get_series_schedule}
\title{Access the playoff schedule for a season and series}
\usage{
get_series_schedule(season = season_now(), series = "a")
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{series}{one-letter code (e.g., 'O'); see \code{\link[=series]{series()}} and/or
\code{\link[=bracket]{bracket()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{get_series_schedule()} is deprecated. Use \code{\link[=series_schedule]{series_schedule()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterOld.R
\name{get_skater_milestones}
\alias{get_skater_milestones}
\title{Access the skaters on milestone watch}
\usage{
get_skater_milestones()
}
\value{
data.frame with one row per player
}
\description{
\code{get_skater_milestones()} is deprecated. Use \code{\link[=skater_milestones]{skater_milestones()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GoalieOld.R
\name{get_goalie_milestones}
\alias{get_goalie_milestones}
\title{Access the goalies on milestone watch}
\usage{
get_goalie_milestones()
}
\value{
data.frame with one row per player
}
\description{
\code{get_goalie_milestones()} is deprecated. Use \code{\link[=goalie_milestones]{goalie_milestones()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayerOld.R
\name{get_players}
\alias{get_players}
\title{Access all the players}
\usage{
get_players()
}
\value{
data.frame with one row per player
}
\description{
\code{get_players()} is deprecated. Use \code{\link[=players]{players()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNPlayer.R
\name{espn_player_summary}
\alias{espn_player_summary}
\title{Access the ESPN summary for a player}
\usage{
espn_player_summary(player = 3988803)
}
\arguments{
\item{player}{integer ID (e.g., 3988803); see \code{\link[=espn_players]{espn_players()}} for
reference}
}
\value{
data.frame with one row
}
\description{
\code{espn_player_summary()} retrieves the ESPN summary for a player as a \code{data.frame} where each row represents one result and includes detail on game timing, matchup state, scoring flow, and situational event detail.
}
\examples{
ESPN_summary_Charlie_McAvoy <- espn_player_summary(player = 3988803)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.mask_strength_context_block}
\alias{.mask_strength_context_block}
\title{Mask HTML on-ice columns on unsupported events}
\usage{
.mask_strength_context_block(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}
}
\value{
data.frame with unsupported rows masked
}
\description{
\code{.mask_strength_context_block()} sets HTML-derived on-ice player-ID and
timing columns to \code{NA} on event types that are not supported by the HTML
on-ice player pipeline.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Player.R
\name{players}
\alias{players}
\title{Access all the players}
\usage{
players()
}
\value{
data.frame with one row per player
}
\description{
\code{players()} retrieves all the players as a \code{data.frame} where each row represents player and includes detail on player identity, role, handedness, and biographical profile.
}
\examples{
# May take >5s, so skip.
\donttest{all_players <- players()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.count_goals_shots}
\alias{.count_goals_shots}
\title{Count the as-of-event goal, shots on goal, Fenwick, and Corsi attempts and differentials for all the events (plays) in a play-by-play by perspective}
\usage{
.count_goals_shots(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with one row per event (play) and added columns: \code{homeGoals}, \code{awayGoals}, \code{homeShots}, \code{awayShots}, \code{homeFenwick}, \code{awayFenwick}, \code{homeCorsi}, \code{awayCorsi}, \code{goalsFor}, \code{goalsAgainst}, \code{shotsFor}, \code{shotsAgainst}, \code{fenwickFor}, \code{fenwickAgainst}, \code{corsiFor}, \code{corsiAgainst}, \code{goalDifferential}, \code{shotDifferential}, \code{fenwickDifferential}, and \code{corsiDifferential}
}
\description{
\code{.count_goals_shots()} counts the as-of-event goal, shots on goal, Fenwick, and Corsi attempts and differentials for all the events (plays) in a play-by-play by perspective.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Playoff.R
\name{series_schedule}
\alias{series_schedule}
\title{Access the playoff schedule for a season and series}
\usage{
series_schedule(season = season_now() - 10001, series = "a")
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{series}{one-letter code (e.g., 'O'); see \code{\link[=series]{series()}} and/or
\code{\link[=bracket]{bracket()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{series_schedule()} retrieves the playoff schedule for a season and series as a \code{data.frame} where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
SCF_schedule_20212022 <- series_schedule(
  season = 20212022, 
  series = 'O'
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Skater.R
\name{skater_playoff_statistics}
\alias{skater_playoff_statistics}
\alias{skater_playoff_stats}
\title{Access the career playoff statistics for all the skaters}
\usage{
skater_playoff_statistics()

skater_playoff_stats()
}
\value{
data.frame with one row per player
}
\description{
\code{skater_playoff_statistics()} retrieves the career playoff statistics for all the skaters as a \code{data.frame} where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
skater_playoff_stats <- skater_playoff_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.public_pbp_is_shootout_eligible}
\alias{.public_pbp_is_shootout_eligible}
\title{Identify shootout-eligible games for public play-by-play cleanup}
\usage{
.public_pbp_is_shootout_eligible(game_id, game_type_id)
}
\arguments{
\item{game_id}{integer game IDs}

\item{game_type_id}{integer game type IDs}
}
\value{
logical vector
}
\description{
\code{.public_pbp_is_shootout_eligible()} returns \code{TRUE} for games whose overtime
rules follow the regular-season shootout path.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.parse_html_pbp_response}
\alias{.parse_html_pbp_response}
\title{Parse an HTML on-ice play-by-play response}
\usage{
.parse_html_pbp_response(
  resp,
  game,
  rosters,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev
)
}
\arguments{
\item{resp}{httr2 response object for the HTML report}

\item{game}{game ID}

\item{rosters}{roster data.frame from play-by-play metadata}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}
}
\value{
data.frame of parsed HTML play-by-play rows
}
\description{
\code{.parse_html_pbp_response()} converts an already-fetched HTML play-by-play
response into parsed on-ice rows.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.strip_situation_code}
\alias{.strip_situation_code}
\title{Strip the situation code into goalie and skater counts, man differential, and strength state for all the events (plays) in a play-by-play by perspective}
\usage{
.strip_situation_code(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with one row per event (play) and added columns: \code{homeIsEmptyNet}, \code{awayIsEmptyNet}, \code{homeSkaterCount}, \code{awaySkaterCount}, \code{isEmptyNetFor}, \code{isEmptyNetAgainst}, \code{skaterCountFor}, \code{skaterCountAgainst}, \code{manDifferential}, and \code{strengthState}
}
\description{
\code{.strip_situation_code()} strips the situation code into goalie and skater counts for home and away teams, then (from the event owner's perspective) computes man differential and a strength state classification without rewriting the source \code{situationCode} column.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNOther.R
\name{espn_transactions}
\alias{espn_transactions}
\title{Access the ESPN transactions for a season}
\usage{
espn_transactions(season = season_now())
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); the summer of the latter
year is included}
}
\value{
data.frame with one row per transaction
}
\description{
\code{espn_transactions()} retrieves the ESPN transactions for a season as a \code{data.frame} where each row represents transaction and includes detail on availability and transaction-status tracking detail.
}
\examples{
ESPN_transactions_20242025 <- espn_transactions(season = 20242025)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.html_on_ice_matches_reconstructed_strength}
\alias{.html_on_ice_matches_reconstructed_strength}
\title{Check HTML skater counts against reconstructed penalty state}
\usage{
.html_on_ice_matches_reconstructed_strength(
  reconstruction,
  idx,
  home_skater_player_ids,
  away_skater_player_ids
)
}
\arguments{
\item{reconstruction}{data.frame from \code{.reconstruct_skater_counts_from_penalties()}}

\item{idx}{row index in the play-by-play}

\item{home_skater_player_ids}{parsed home skater IDs}

\item{away_skater_player_ids}{parsed away skater IDs}
}
\value{
logical scalar
}
\description{
\code{.html_on_ice_matches_reconstructed_strength()} validates the HTML skater
counts against the penalty-based reconstruction. Goalie presence is validated
separately when deciding whether the HTML row can override stale source
strength context.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeagueOld.R
\name{get_venues}
\alias{get_venues}
\title{Access all the venues}
\usage{
get_venues()
}
\value{
data.frame with one row per venue
}
\description{
\code{get_venues()} is deprecated. Use \code{\link[=venues]{venues()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/League.R
\name{venues}
\alias{venues}
\title{Access all the venues}
\usage{
venues()
}
\value{
data.frame with one row per venue
}
\description{
\code{venues()} retrieves all the venues as a \code{data.frame} where each row represents venue and includes detail on venue/location geography and regional metadata.
}
\examples{
all_venues <- venues()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNPlayer.R
\name{espn_players}
\alias{espn_players}
\title{Access all the ESPN players}
\usage{
espn_players()
}
\value{
data.frame with one row per ESPN player
}
\description{
\code{espn_players()} retrieves all the ESPN players as a \code{data.frame} where each row represents ESPN player and includes detail on person-level profile context and performance history with situational splits.
}
\examples{
all_ESPN_players <- espn_players()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Award.R
\name{award_winners}
\alias{award_winners}
\title{Access all the award winners/finalists}
\usage{
award_winners()
}
\value{
data.frame with one row per winner/finalist
}
\description{
\code{award_winners()} retrieves all the award winners/finalists as a \code{data.frame} where each row represents winner/finalist and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and player identity, role, handedness, and biographical profile.
}
\examples{
all_award_winners <- award_winners()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{.nhl_request}
\alias{.nhl_request}
\title{Build an NHL API request}
\usage{
.nhl_request(path, query = list(), type)
}
\arguments{
\item{path}{character}

\item{query}{list}

\item{type}{character of 'w' for web, 's' for stats, and 'r' for records}
}
\value{
httr2 request object
}
\description{
\code{.nhl_request()} constructs an NHL API request with the standard retry policy
used throughout the package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.parallel_request_failed}
\alias{.parallel_request_failed}
\title{Check whether a parallel request failed}
\usage{
.parallel_request_failed(resp)
}
\arguments{
\item{resp}{response or failure object}
}
\value{
logical scalar
}
\description{
\code{.parallel_request_failed()} returns \code{TRUE} when a response object from
\code{.perform_parallel_requests()} is an httr2 failure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Player.R
\name{spotlight_players}
\alias{spotlight_players}
\title{Access the spotlight players}
\usage{
spotlight_players()
}
\value{
data.frame with one row per player
}
\description{
\code{spotlight_players()} retrieves the spotlight players as a \code{data.frame} where each row represents player and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
}
\examples{
spotlight_players <- spotlight_players()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GameOld.R
\name{get_game_story}
\alias{get_game_story}
\title{Access the World Showcase (WSC) summary for a game}
\usage{
get_game_story(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
list of various items
}
\description{
\code{get_game_story()} is deprecated. Use \code{\link[=wsc_summary]{wsc_summary()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{wsc_play_by_play}
\alias{wsc_play_by_play}
\alias{wsc_pbp}
\title{Access the World Showcase (WSC) play-by-play for a game}
\usage{
wsc_play_by_play(game = 2023030417)

wsc_pbp(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
data.frame with one row per event (play)
}
\description{
\code{wsc_play_by_play()} retrieves the World Showcase (WSC) play-by-play for a
game as a \code{data.frame} where each row represents an event. The returned
schema follows the same cleaned public-facing naming as \code{gc_play_by_play()},
including \code{servedByPlayerId}, \code{goalieInNetId}, and \code{utc} immediately after
\code{secondsElapsedInGame} while omitting GC-only clip fields. It also includes
the same HTML-report-derived on-ice player ID columns added to the GC output,
including dynamically expanded overflow skater slots when needed. HTML report
skater and goalie IDs are returned whenever they can be matched back to a
supported row, even when the raw \code{situationCode} is stale. Use \code{\link[=add_shift_times]{add_shift_times()}}
with \code{\link[=shift_chart]{shift_chart()}} (or \code{\link[=shift_charts]{shift_charts()}}) to add on-ice shift timing columns.
}
\examples{
# May take >5s, so skip.
\donttest{wsc_pbp_Martin_Necas_legacy_game <- wsc_play_by_play(game = 2025020275)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.parse_html_pbp_doc}
\alias{.parse_html_pbp_doc}
\title{Parse an HTML play-by-play document}
\usage{
.parse_html_pbp_doc(
  doc,
  roster_lookup,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev,
  is_playoffs = FALSE
)
}
\arguments{
\item{doc}{parsed HTML document}

\item{roster_lookup}{standardized roster lookup data.frame}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}

\item{is_playoffs}{logical; whether the game is a playoff game}
}
\value{
data.frame of parsed HTML play-by-play rows
}
\description{
\code{.parse_html_pbp_doc()} extracts event rows, actor IDs, and on-ice player
lists from an NHL HTML play-by-play report.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_team_prospects}
\alias{get_team_prospects}
\title{Access the prospects for a team and position}
\usage{
get_team_prospects(team = "NJD", player_type = "forwards")
}
\arguments{
\item{team}{three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for reference}

\item{player_type}{character of 'forwards', 'defensemen', or 'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{get_team_prospects()} is deprecated. Use \code{\link[=team_prospects]{team_prospects()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GoalieEDGE.R
\name{goalie_edge_seasons}
\alias{goalie_edge_seasons}
\title{Access the season(s) and game type(s) in which there exists goalie EDGE
statistics}
\usage{
goalie_edge_seasons()
}
\value{
data.frame with one row per season
}
\description{
\code{goalie_edge_seasons()} retrieves the season(s) and game type(s) in which there exists goalie EDGE statistics as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context plus NHL EDGE style tracking outputs and relative-performance context.
}
\examples{
goalie_EDGE_seasons <- goalie_edge_seasons()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{.optional_html_pbp_rows_from_response}
\alias{.optional_html_pbp_rows_from_response}
\title{Parse optional HTML on-ice rows from a parallel response}
\usage{
.optional_html_pbp_rows_from_response(
  resp,
  game,
  rosters,
  home_team_id,
  away_team_id,
  home_abbrev,
  away_abbrev
)
}
\arguments{
\item{resp}{response or failure object}

\item{game}{game ID}

\item{rosters}{roster data.frame from play-by-play metadata}

\item{home_team_id}{home-team ID}

\item{away_team_id}{away-team ID}

\item{home_abbrev}{home-team abbreviation}

\item{away_abbrev}{away-team abbreviation}
}
\value{
data.frame of parsed HTML rows or an empty data.frame
}
\description{
\code{.optional_html_pbp_rows_from_response()} parses a fetched HTML play-by-play
response when it succeeded and otherwise returns an empty data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{games}
\alias{games}
\title{Access all the games}
\usage{
games()
}
\value{
data.frame with one row per game
}
\description{
\code{games()} retrieves all the games as a \code{data.frame} where each row represents game and includes detail on game timeline state, period/clock progression, and matchup flow plus date/season filtering windows and chronological context.
}
\examples{
# May take >5s, so skip.
\donttest{all_games <- games()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GameOld.R
\name{get_partner_odds}
\alias{get_partner_odds}
\title{Access the real-time game odds for a country by partnered bookmaker}
\usage{
get_partner_odds(country = "US")
}
\arguments{
\item{country}{two-letter code (e.g., 'CA'); see \code{\link[=countries]{countries()}} for reference}
}
\value{
data.frame with one row per game
}
\description{
\code{get_partner_odds()} is deprecated. Use \code{\link[=game_odds]{game_odds()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DraftOld.R
\name{get_draft_rankings}
\alias{get_draft_rankings}
\title{Access the draft rankings for a year and player type}
\usage{
get_draft_rankings(year = season_now()\%/\%10000, player_type = 1)
}
\arguments{
\item{year}{integer in YYYY (e.g., 2017); see \code{\link[=drafts]{drafts()}} for reference}

\item{player_type}{integer in 1:4 (where 1 = North American Skaters,
2 = International Skaters, 3 = North American Goalies, and 4 = International
Goalies)}
}
\value{
data.frame with one row per player
}
\description{
\code{get_draft_rankings()} is deprecated. Use \code{\link[=draft_rankings]{draft_rankings()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.flag_is_rebound}
\alias{.flag_is_rebound}
\title{Flag if the shot attempt is a rebound attempt or creates a rebound for all the shots in a play-by-play}
\usage{
.flag_is_rebound(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame of play-by-play(s); see \code{\link[=gc_play_by_play]{gc_play_by_play()}} and/or \code{\link[=wsc_play_by_play]{wsc_play_by_play()}} for reference}
}
\value{
data.frame with one row per event (play) and added columns: \code{createdRebound} and \code{isRebound}
}
\description{
\code{.flag_is_rebound()} flags whether a shot attempt is a rebound attempt (i.e., taken within 3 seconds of a prior blocked, missed, or saved attempt with no stoppage in between), and whether a shot attempt creates a rebound under the same definition.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGameOld.R
\name{get_espn_event_play_by_play}
\alias{get_espn_event_play_by_play}
\title{Access the ESPN play-by-play for an event (game)}
\usage{
get_espn_event_play_by_play(event = 401777460)
}
\arguments{
\item{event}{integer ID (e.g., 401777460); see \code{\link[=espn_games]{espn_games()}} for
reference}
}
\value{
data.frame with one row per play
}
\description{
\code{get_espn_event_play_by_play()} is deprecated. Use \code{\link[=espn_play_by_play]{espn_play_by_play()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.format_elapsed_clock}
\alias{.format_elapsed_clock}
\title{Format elapsed seconds as an \code{MM:SS} clock}
\usage{
.format_elapsed_clock(seconds)
}
\arguments{
\item{seconds}{integer/numeric vector of elapsed seconds}
}
\value{
character vector
}
\description{
\code{.format_elapsed_clock()} converts elapsed-in-period seconds back into
zero-padded \code{MM:SS} strings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayoffOld.R
\name{get_series}
\alias{get_series}
\title{Access the playoff series for a season and round}
\usage{
get_series()
}
\description{
\code{get_series()} is defunct.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_statistics}
\alias{goalie_statistics}
\alias{goalie_stats}
\title{Access the career statistics for all the goalies}
\usage{
goalie_statistics()

goalie_stats()
}
\value{
data.frame with one row per player
}
\description{
\code{goalie_statistics()} retrieves the career statistics for all the goalies as a \code{data.frame} where each row represents player and includes detail on team identity, affiliation, and matchup-side context, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
goalie_stats <- goalie_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AwardOld.R
\name{get_award_winners}
\alias{get_award_winners}
\title{Access all the award winners/finalists}
\usage{
get_award_winners()
}
\value{
data.frame with one row per winner/finalist
}
\description{
\code{get_award_winners()} is deprecated. Use \code{\link[=award_winners]{award_winners()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_team_roster_statistics}
\alias{get_team_roster_statistics}
\title{Access the roster statistics for a team, season, game type, and player type}
\usage{
get_team_roster_statistics(
  team = "NJD",
  season = "now",
  game_type = 2,
  player_type = "skaters"
)
}
\arguments{
\item{team}{three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for reference}

\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}

\item{player_type}{character of 'skaters' or 'goalies'}
}
\value{
data.frame with one row per player
}
\description{
\code{get_team_roster_statistics()} is deprecated. Use \code{\link[=roster_statistics]{roster_statistics()}}
instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Clean.R
\name{.normalize_public_pbp_names}
\alias{.normalize_public_pbp_names}
\title{Normalize raw/internal play-by-play names to the public schema}
\usage{
.normalize_public_pbp_names(play_by_play)
}
\arguments{
\item{play_by_play}{data.frame play-by-play object}
}
\value{
data.frame with current public column names where available
}
\description{
\code{.normalize_public_pbp_names()} renames the raw GameCenter/World Showcase
column names used during ingest into the current public play-by-play schema.
This helper is for the package's own ingest pipeline only and is not a
legacy-input compatibility layer for downstream callers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Playoff.R
\name{playoff_season_statistics}
\alias{playoff_season_statistics}
\alias{playoff_season_stats}
\title{Access the playoff statistics by season}
\usage{
playoff_season_statistics()

playoff_season_stats()
}
\value{
data.frame with one row per season
}
\description{
\code{playoff_season_statistics()} retrieves the playoff statistics by season as a \code{data.frame} where each row represents season and includes detail on date/season filtering windows and chronological context.
}
\examples{
playoff_season_stats <- playoff_season_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Coach.R
\name{coach_career_statistics}
\alias{coach_career_statistics}
\alias{coach_career_stats}
\title{Access the career statistics for all the coaches}
\usage{
coach_career_statistics()

coach_career_stats()
}
\value{
data.frame with one row per coach
}
\description{
\code{coach_career_statistics()} retrieves the career statistics for all the coaches as a \code{data.frame} where each row represents coach and includes detail on ranking movement, points pace, and division/conference position signals.
}
\examples{
coach_career_stats <- coach_career_statistics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Award.R
\name{awards}
\alias{awards}
\title{Access all the awards}
\usage{
awards()
}
\value{
data.frame with one row per award
}
\description{
\code{awards()} retrieves all the awards as a \code{data.frame} where each row represents award and includes detail on recognition, leaderboard, and milestone-watch context.
}
\examples{
all_awards <- awards()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AwardOld.R
\name{get_awards}
\alias{get_awards}
\title{Access all the awards}
\usage{
get_awards()
}
\value{
data.frame with one row per award
}
\description{
\code{get_awards()} is deprecated. Use \code{\link[=awards]{awards()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Goalie.R
\name{goalie_game_report}
\alias{goalie_game_report}
\title{Access various reports for a season, game type, and category for all
the goalies by game}
\usage{
goalie_game_report(
  season = season_now(),
  game_type = game_type_now(),
  category = "summary"
)
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}

\item{category}{character (e.g., 'advanced'); see
\code{\link[=goalie_report_configurations]{goalie_report_configurations()}} for reference}
}
\value{
data.frame with one row per game per goalie
}
\description{
\code{goalie_game_report()} retrieves various reports for a season, game type, and category for all the goalies by game as a \code{data.frame} where each row represents game per goalie and includes detail on game timeline state, period/clock progression, and matchup flow, player identity, role, handedness, and biographical profile, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
# May take >5s, so skip.
\donttest{advanced_goalie_game_report_playoffs_20212022 <- 
  goalie_game_report(
    season    = 20212022, 
    game_type = 3, 
    category  = 'advanced'
  )}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlayoffOld.R
\name{get_bracket}
\alias{get_bracket}
\title{Access the playoff bracket for a season}
\usage{
get_bracket(season = season_now())
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}
}
\value{
data.frame with one row per series
}
\description{
\code{get_bracket()} is deprecated. Use \code{\link[=bracket]{bracket()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{drafts}
\alias{drafts}
\title{Access all the drafts}
\usage{
drafts()
}
\value{
data.frame with one row per draft
}
\description{
\code{drafts()} retrieves all the drafts as a \code{data.frame} where each row represents draft and includes detail on venue/location geography and regional metadata.
}
\examples{
all_drafts <- drafts()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGameOld.R
\name{get_espn_events}
\alias{get_espn_events}
\title{Access the ESPN events (games) by start and end dates}
\usage{
get_espn_events()
}
\description{
\code{get_espn_events()} is defunct. Use \code{\link[=espn_games]{espn_games()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Coach.R
\name{coaches}
\alias{coaches}
\title{Access all the coaches}
\usage{
coaches()
}
\value{
data.frame with one row per coach
}
\description{
\code{coaches()} retrieves all the coaches as a \code{data.frame} where each row represents coach and includes detail on team identity, affiliation, and matchup-side context plus player identity, role, handedness, and biographical profile.
}
\examples{
all_coaches <- coaches()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GoalieOld.R
\name{get_goalie_statistics}
\alias{get_goalie_statistics}
\title{Access various reports for all the goalies by season or game}
\usage{
get_goalie_statistics()
}
\description{
\code{get_goalie_statistics()} is defunct. Use \code{\link[=goalie_season_report]{goalie_season_report()}} or
\code{\link[=goalie_game_report]{goalie_game_report()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SkaterOld.R
\name{get_skater_leaders}
\alias{get_skater_leaders}
\title{Access the skater statistics leaders for a season, game type, and category}
\usage{
get_skater_leaders(season = "current", game_type = "", category = "points")
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see \code{\link[=seasons]{seasons()}} for
reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
playoff'/'post'; see \code{\link[=seasons]{seasons()}} for reference; most functions will NOT
support pre-season}

\item{category}{character of 'assists', 'goals', 'goalsSh', 'goalsPp',
'points, 'penaltyMins', 'toi', 'plusMinus', or 'faceoffLeaders'}
}
\value{
data.frame with one row per player
}
\description{
\code{get_skater_leaders()} is deprecated. Use \code{\link[=skater_leaders]{skater_leaders()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNTeam.R
\name{espn_teams}
\alias{espn_teams}
\title{Access all the ESPN teams}
\usage{
espn_teams()
}
\value{
data.frame with one row per ESPN team
}
\description{
\code{espn_teams()} retrieves all the ESPN teams as a \code{data.frame} where each row represents ESPN team and includes detail on team composition, matchup context, and season progression detail.
}
\examples{
all_ESPN_teams <- espn_teams()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Franchise.R
\name{franchise_season_statistics}
\alias{franchise_season_statistics}
\alias{franchise_season_stats}
\title{Access the statistics for all the franchises by season and game type}
\usage{
franchise_season_statistics()

franchise_season_stats()
}
\value{
data.frame with one row per franchise per season per game type
}
\description{
\code{franchise_season_statistics()} retrieves the statistics for all the franchises by season and game type as a \code{data.frame} where each row represents franchise per season per game type and includes detail on date/season filtering windows and chronological context, team identity, affiliation, and matchup-side context, and production, workload, efficiency, and result-level performance outcomes.
}
\examples{
# May take >5s, so skip.
\donttest{franchise_season_stats <- franchise_season_statistics()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamOld.R
\name{get_team_seasons}
\alias{get_team_seasons}
\title{Access the season(s) and game type(s) in which a team played}
\usage{
get_team_seasons(team = "NJD")
}
\arguments{
\item{team}{three-letter code (e.g., 'COL'); see \code{\link[=teams]{teams()}} for reference}
}
\value{
data.frame with one row per season
}
\description{
\code{get_team_seasons()} is deprecated. Use \code{\link[=team_seasons]{team_seasons()}} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utility.R
\name{normalize_locale_names}
\alias{normalize_locale_names}
\title{Normalize locale-style dotted columns to camelCase}
\usage{
normalize_locale_names(x)
}
\arguments{
\item{x}{character vector}
}
\value{
character vector
}
\description{
Converts dotted names (e.g., firstName.default, name.cs) to camelCase and
removes trailing \code{Default} from default-language fields.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Draft.R
\name{expansion_drafts}
\alias{expansion_drafts}
\title{Access all the expansion drafts}
\usage{
expansion_drafts()
}
\value{
data.frame with one row per expansion draft
}
\description{
\code{expansion_drafts()} retrieves all the expansion drafts as a \code{data.frame} where each row represents expansion draft and includes detail on date/season filtering windows and chronological context plus reference definitions and rules-framework information.
}
\examples{
all_expansion_drafts <- expansion_drafts()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TeamEDGE.R
\name{team_edge_leaders}
\alias{team_edge_leaders}
\title{Access the team EDGE statistics leaders for a season and game type}
\usage{
team_edge_leaders(season = "now", game_type = "")
}
\arguments{
\item{season}{integer in YYYYYYYY (e.g., 20242025); see
\code{\link[=team_edge_seasons]{team_edge_seasons()}} for reference}

\item{game_type}{integer in 1:3 (where 1 = pre-season, 2 = regular season, 3
= playoff/post-season) OR character of 'pre', 'regular', or
'playoff'/'post'; see \code{\link[=team_edge_seasons]{team_edge_seasons()}} for reference; most functions
will NOT support pre-season}
}
\value{
list of various items
}
\description{
\code{team_edge_leaders()} retrieves the team EDGE statistics leaders for a season and game type as a nested \code{list} that separates summary and detail blocks for NHL EDGE style tracking outputs and relative-performance context.
}
\examples{
team_EDGE_leaders_regular_20242025 <- team_edge_leaders(
  season    = 20242025,
  game_type = 2
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Game.R
\name{shift_chart}
\alias{shift_chart}
\title{Access the shift chart for a game}
\usage{
shift_chart(game = 2023030417)
}
\arguments{
\item{game}{integer ID (e.g., 2025020275); see \code{\link[=games]{games()}} for reference}
}
\value{
data.frame with one row per shift
}
\description{
\code{shift_chart()} retrieves the shift chart for a game as a \code{data.frame} where each row represents shift and includes detail on game timeline state, period/clock progression, and matchup flow, date/season filtering windows and chronological context, and team identity, affiliation, and matchup-side context.
}
\examples{
shifts_Martin_Necas_legacy_game <- shift_chart(game = 2025020275)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ESPNGame.R
\name{espn_play_by_play}
\alias{espn_play_by_play}
\alias{espn_pbp}
\title{Access the ESPN play-by-play for a game}
\usage{
espn_play_by_play(game = 401777460)

espn_pbp(game = 401777460)
}
\arguments{
\item{game}{integer ID (e.g., 401777460); see \code{\link[=espn_games]{espn_games()}} for
reference}
}
\value{
data.frame with one row per event (play)
}
\description{
\code{espn_play_by_play()} retrieves the ESPN play-by-play for a game as a \code{data.frame} where each row represents event and includes detail on game timeline state, period/clock progression, and matchup flow, team identity, affiliation, and matchup-side context, and situational splits across home/road, strength state, and overtime/shootout states.
}
\examples{
ESPN_pbp_SCF_20242025 <- espn_play_by_play(game = 401777460)
}
